/**********************************************************************
		PORTMAN_PCS - module for serial port MIDI devices
	
			(C) 2002 by Olivier delhaye

		printk messages are written to /var/log/messages

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/soundcard.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/string.h>

#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/serialP.h>

#include <linux/delay.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <linux/wait.h>
#include <linux/compatmac.h>
#include <linux/init.h>
#include <linux/sched.h>
#endif

#define PORTMAN_PCS_SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"

#define PORTMAN_PCS_MAJOR_NUMBER 80				//use for ask
#define PORTMAN_PCS_SEQUENCER_MINOR_NUMBER 0	//use for /dev/sequencer
#define PORTMAN_PCS_SEQUENCER2_MINOR_NUMBER 8	//use for /dev/sequencer2 (/dev/music)
#define PORTMAN_PCS_MIDI_MINOR_NUMBER 1			//use for /dev/midi

#define MC303 1		//if you have an mc303 0 or 1
#define MAX 2048	//buffer for writing
#define RETRYTIMEOUT 400	//X retry for reading midi byte


MODULE_AUTHOR("DELHAYE Olivier");
MODULE_DESCRIPTION("PORTMAN_PCS serial port/midi driver");

static struct file *PORTMAN_PCS_serial_port_file = NULL;
static struct tty_struct *PORTMAN_PCS_serial_port_tty = NULL;

static int PORTMAN_PCS_open_time = 0;
static int PORTMAN_PCS_timer_start = 0; // in jiffies - i.e. 100 ticks per second
static int PORTMAN_PCS_timer_base = 96; // timer ticks per beat
static int PORTMAN_PCS_timer_tempo = 120; // beats per second
static int PORTMAN_PCS_echo_sync_time = 0;

static unsigned char PORTMAN_PCS_outbuffer[MAX];

static int PORTMAN_PCS_outbuffer_pointer = 0;
static int PORTMAN_PCS_outbuffer_length = 0;

static struct timer_list PORTMAN_PCS_buffer_timer;
static wait_queue_head_t PORTMAN_PCS_buffer_wait_queue;

static unsigned char evt_prec = 0x00;	//past event read from tty

static int play_event (unsigned char *q);

/**********************************************************************************
*   Serial Port write/read
*
*	For info on the TTY serial port drivers, see
*	/usr/include/linux/tty_driver.h
*	/usr/src/linux/drivers/char/serial.c
*
*	Because serial port handling is extremely complex, PORTMAN_PCS uses the kernel
*	serial port functions.  Serial port input calls the line discipline functions
*	directly. This is not portable.  Reads from the line discipline do not block, so
*	schedule_timeout is used to poll the read function.
*
**********************************************************************************/

static struct tty_struct *PORTMAN_PCS_open_serial_port(char *serial_port_device_name)
{
	struct tty_struct *tty = NULL;
	struct termios old_termios;

	if (IS_ERR(PORTMAN_PCS_serial_port_file = filp_open(serial_port_device_name, O_WRONLY | O_NONBLOCK, 0)))
		printk(KERN_INFO "PORTMAN_PCS: Failure %d opening serial port %s\n", (int) PTR_ERR(PORTMAN_PCS_serial_port_file), serial_port_device_name);

	else if ((tty = (struct tty_struct *)PORTMAN_PCS_serial_port_file->private_data) == NULL)
		printk(KERN_INFO "PORTMAN_PCS: NULL file private data TTY pointer\n");

	else if (tty->driver.set_termios == NULL)
		printk(KERN_INFO "PORTMAN_PCS: NULL file set_termios function\n");

	else if (tty->termios == NULL)
		printk(KERN_INFO "PORTMAN_PCS: NULL file termios structure\n");

	else if (tty->ldisc.flush_buffer == NULL)
		printk(KERN_ERR "PORTMAN_PCS: NULL TTY ldisc.flush_buffer() function\n");

	else if (tty->driver.write == NULL)
		printk(KERN_ERR "PORTMAN_PCS: NULL TTY driver write function\n");

	else if (tty->ldisc.read == NULL)
		printk(KERN_ERR "PORTMAN_PCS: NULL TTY ldisc.read() function\n");

	else
	{
		old_termios = *tty->termios;
		tty->termios->c_lflag = 0;
		tty->termios->c_lflag = NOFLSH;
		tty->termios->c_iflag = IGNBRK | IGNPAR;
		tty->termios->c_oflag = 0;
		tty->termios->c_cflag =  B115200 | CREAD | CSIZE | CS8 | CRTSCTS | HUPCL ;
		tty->termios->c_cc[VEOL] = 0; /* '\r'; */
		tty->termios->c_cc[VERASE] = 0;
		tty->termios->c_cc[VKILL] = 0;
		tty->termios->c_cc[VMIN] = 0;
		tty->termios->c_cc[VTIME] = 0;
		(*tty->driver.set_termios)(tty, &old_termios);
		return tty;
	}

	return 0;
}

static int PORTMAN_PCS_write_to_port(const unsigned char *data, size_t length)
{
	int scan;
	struct async_struct *info =(struct async_struct *)PORTMAN_PCS_serial_port_tty->driver_data;

	for (scan = 0; scan < length; ++scan)
	{
		int status = -EFAULT;

		info->MCR |= UART_MCR_DTR;
		outb(info->MCR,info->port+UART_MCR);
	
		if ((status = PORTMAN_PCS_serial_port_tty->driver.write(PORTMAN_PCS_serial_port_tty, 0, &data[scan], 1)) < 1)
		{
			printk(KERN_ERR "PORTMAN_PCS: write failure: %d of %d bytes, status %d\n", scan, length, status);
			return -1;
		}
		
		//the udelay is use for my MC303 but you can delete it !
		if (MC303 == 1)
			udelay(550);
		

	}

	return length;
}

int PORTMAN_PCS_read_to_port(char *data)
{
	//DON'T MODIFY THIS FUNCTION PLEASE !!

	struct async_struct *info =(struct async_struct *)PORTMAN_PCS_serial_port_tty->driver_data;
	int count;

	/***** start the bufferin to tty ****/	
	info->MCR &= ~UART_MCR_DTR;
	outb(info->MCR,info->port+UART_MCR);


	count = PORTMAN_PCS_serial_port_tty->ldisc.read(PORTMAN_PCS_serial_port_tty, PORTMAN_PCS_serial_port_file, data,1);
	
	// i must put this minimun delay for reading , i don't know why but if you want good latency ,use it
	udelay(60);	
	
	/***** stop the bufferin to tty ****/	
	info->MCR |= UART_MCR_DTR;
	outb(info->MCR,info->port+UART_MCR);

	
	return count;
}


/**********************************************************************************
*   Soundcard function support  for /dev/sequencer and /dev/music 
**********************************************************************************/

static int seq_chn_voice_event(unsigned char *q)
{
	unsigned char cmd = q[2];
	unsigned char chn = q[3];
	unsigned char note = q[4];
	unsigned char vel = q[5];
	unsigned char prn = q[6];

	unsigned char output[8];

//	printk (KERN_INFO "NOTE ON/OFF :cmd %x chn %x note %x vel %x prn %x\n",cmd,chn,note,vel,prn);

	switch (cmd)
	{
		case MIDI_NOTEOFF:
			output[0] = cmd + chn;
			output[1] = note;
			output[2] = vel; 
			//printk (KERN_INFO "NOTE OFF : %x %x %x\n",output[0],output[1],output[2]);

			PORTMAN_PCS_write_to_port(output,3);
			break;

		case MIDI_NOTEON:
			if(note > 127 && note !=255)
				break;
			output[0] = cmd + chn;
			output[1] = note;
			output[2] = vel; 
			//printk (KERN_INFO "NOTE ON : %x %x %x\n",output[0],output[1],output[2]);

			PORTMAN_PCS_write_to_port(output,3);
			break;

		case MIDI_KEY_PRESSURE:
			output[0] = cmd + chn;
			output[1] = note;
			output[2] = prn; 
			//printk (KERN_INFO "KEY PRESSURE\n");

			PORTMAN_PCS_write_to_port(output,3);
			break;
	}
	return 0;

}

static int seq_chn_common_event(unsigned char *q)
{
	unsigned char cmd = q[2];
	unsigned char chn = q[3];
	unsigned char prn = q[4];
	unsigned char val = q[6];

	unsigned char output[8];

	switch (cmd)
	{
		case MIDI_PGM_CHANGE:
			output[0] = cmd + chn;
			output[1] = prn;

			//printk (KERN_INFO "midi PGM: %x %x\n",output[0],output[1]);
			PORTMAN_PCS_write_to_port(output,2);
			break;

		case MIDI_CTL_CHANGE:
			if (chn > 15 || prn > 127)
				break;

			output[0] = cmd + chn;
			output[1] = prn;
			output[2] = val;

			//printk (KERN_INFO "midi CTL: %x %x %x\n",output[0],output[1],output[2]);
			PORTMAN_PCS_write_to_port(output,3);

			break;

		case MIDI_PITCH_BEND:
			/* pas sur :( */
			output[0] = cmd + chn;
			output[1] = val;

			//printk (KERN_INFO "midi PITCH: %x %x\n",output[0],output[1]);
			PORTMAN_PCS_write_to_port(output,2);

			break;
		default:;
	}
	return 0;
}

static void seq_sysex_message(unsigned char *q)
{
	unsigned char *output = &q[2];
	int i,l =0;
	
	//take from sequencer.c
	//find the lenght of sysex message
	for (i = 0 ; i < 6 && output[i] !=0xff;i++)
		l = l + 1;
	
	PORTMAN_PCS_write_to_port(output,l);

}

/**********************************************************************************
*   Kernel timed serial port input / output functions
*	/dev/sequencer requires MIDI input and output to be handled asynchronously
**********************************************************************************/


static void seq_timing_event(unsigned long argument)
{
	int wait_ticks;
	double wait_jiffies;
	double absolute_jiffies;
	double absolute_ticks;
	int expiration_jiffies;
//	struct sysex_info *sysex;

	unsigned char output[8];
	int i=0; //counter

	for (; PORTMAN_PCS_outbuffer_pointer < PORTMAN_PCS_outbuffer_length; PORTMAN_PCS_outbuffer_pointer += 8)
	{
	//	printk(KERN_INFO "PORTMAN_PCS: timing %x\n",cmd);

	if (PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer] == EV_TIMING)
	{
		switch (PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer+1])
		{
		case TMR_START:
			PORTMAN_PCS_timer_start = jiffies;
		break;

		case TMR_STOP:
			PORTMAN_PCS_timer_start = 0;
		break;

		case TMR_CONTINUE:
			PORTMAN_PCS_timer_start = jiffies;
		break;

		case TMR_TEMPO:
			PORTMAN_PCS_timer_tempo = *((unsigned int*)&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4]);
		break;

		case TMR_ECHO:
			PORTMAN_PCS_echo_sync_time = *((int*)&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4]);
		break;

		case TMR_WAIT_REL:
			wait_ticks = *((unsigned int*)&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4]);
			wait_jiffies = ((double)wait_ticks) / ((double)PORTMAN_PCS_timer_base) * 60.0 * ((double)HZ) / ((double)PORTMAN_PCS_timer_tempo);
			absolute_jiffies = ((double)jiffies) + wait_jiffies - ((double)PORTMAN_PCS_timer_start);
			absolute_ticks = absolute_jiffies * ((double)PORTMAN_PCS_timer_base) / 60.0 / ((double)HZ) * ((double)PORTMAN_PCS_timer_tempo);

			// Convert relative time messages to absolute time for easier handling
			PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 1] = TMR_WAIT_ABS;
			*((unsigned int*)&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4]) = ((int)absolute_ticks);

			// printk(KERN_INFO "PORTMAN_PCS: relative wait %d = absolute %e\n", wait_ticks, absolute_ticks);
			PORTMAN_PCS_outbuffer_pointer -= 8; // so this event is handled again by the loop
		break;

		case TMR_WAIT_ABS:
			wait_ticks = *((unsigned int*)&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4]);
			wait_jiffies = ((double)wait_ticks) / ((double)PORTMAN_PCS_timer_base) * 60.0 * ((double)HZ) / ((double)PORTMAN_PCS_timer_tempo);
			expiration_jiffies = ((int)wait_jiffies) + PORTMAN_PCS_timer_start;

			/* printk(KERN_INFO "PORTMAN_PCS: absolute wait %d - expires %d, it is now %d\n", wait_ticks, expiration_jiffies, (int) jiffies); */

			if (PORTMAN_PCS_timer_start <= 0)
				printk(KERN_ERR "PORTMAN_PCS write: wait with no timer started\n");
			else
			{
				init_timer(&PORTMAN_PCS_buffer_timer);
				PORTMAN_PCS_buffer_timer.expires = expiration_jiffies;
				PORTMAN_PCS_buffer_timer.data = 0;
				PORTMAN_PCS_buffer_timer.function = seq_timing_event;
				add_timer(&PORTMAN_PCS_buffer_timer);
				PORTMAN_PCS_outbuffer_pointer += 8;
				return;
			}
			break;
/*********
			case SYSEX_PATCH:
				sysex = (struct sysex_info*) &PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer];
	
				if ((sysex->len + PORTMAN_PCS_outbuffer_pointer + 8) > PORTMAN_PCS_outbuffer_length)
					sysex->len = PORTMAN_PCS_outbuffer_length - PORTMAN_PCS_outbuffer_pointer - 8;

				if (PORTMAN_PCS_write_to_port(&PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 8], sysex->len) < sysex->len)
					printk(KERN_ERR "PORTMAN_PCS: Failure writing sysex message\n");

				PORTMAN_PCS_outbuffer_pointer += sysex->len;
			break;		
*********/
			default:
				printk(KERN_INFO "PORTMAN_PCS: unsupported sequencer event %x %x %x %x %x %x %x %x\n", PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 1], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 2], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 3], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 4], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 5], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 6], PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer + 7]);

			}//witch
		}//if
		else if (PORTMAN_PCS_timer_start > 0)
		{
			for ( i = 0 ; i < 8 ; i++)
				output[i] = PORTMAN_PCS_outbuffer[PORTMAN_PCS_outbuffer_pointer+i];
			play_event(output);

		}
	}//for

	PORTMAN_PCS_outbuffer_pointer = 0;
	PORTMAN_PCS_outbuffer_length = 0;

	if (waitqueue_active(&PORTMAN_PCS_buffer_wait_queue))
		wake_up_interruptible(&PORTMAN_PCS_buffer_wait_queue);

}

static int play_event (unsigned char *q)
{
	//printk (KERN_INFO "play evt %x %x %x %x %x %x\n",q[0],q[1],q[2],q[3],q[4],q[5]);

	switch (q[0])
	{
		case SEQ_MIDIPUTC:
			PORTMAN_PCS_write_to_port(&q[1], 1);
			//PORTMAN_PCS_outbuffer_pointer -=4;
		break;
		case SEQ_EXTENDED:
			printk(KERN_INFO "extended event not supported\n");
		break;
		case EV_CHN_COMMON:
			seq_chn_common_event(q);
		break;
		case EV_CHN_VOICE:
			seq_chn_voice_event(q);
		break;
		case EV_TIMING:
			seq_timing_event(1);
		break;
		case EV_SYSEX:
			seq_sysex_message(q);
		break;

		default:
			printk(KERN_INFO "event not supported now :(");

	}
	return 0;
}


/****************************************************************************
*	/dev/midiXX file operations
*
*	It is possible to override the standard OSS file_operations functions
*	with module supplied file_operations.  There should be no conflicts
*	with interrupts from /dev/sequencer which calls the OSS midi_operations.
*
*****************************************************************************/


static int PORTMAN_PCS_file_open (struct inode *inode, struct file *file)
{
//	printk(KERN_INFO "PORTMAN_PCS: open minor number %d\n",(int) file->private_data); 

	if (PORTMAN_PCS_open_time > 0)
	{
		printk (KERN_INFO "Timer online\n");
		return -EBUSY;
	}

	file->private_data = (void*) MINOR(inode->i_rdev);
	PORTMAN_PCS_timer_base = 50;
	PORTMAN_PCS_timer_tempo = 120; // default is 100 ticks/second - same as Linux jiffies clock (i.e. HZ)
	PORTMAN_PCS_timer_start = 0;
	PORTMAN_PCS_open_time = jiffies;

	PORTMAN_PCS_outbuffer_pointer = 0;
	PORTMAN_PCS_outbuffer_length = 0;
	init_timer(&PORTMAN_PCS_buffer_timer);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#	init_waitqueue(&PORTMAN_PCS_buffer_wait_queue);
#else
	init_waitqueue_head(&PORTMAN_PCS_buffer_wait_queue);
#endif
	MOD_INC_USE_COUNT;

	return 0;
}

static int PORTMAN_PCS_file_release (struct inode *inode, struct file *file)
{
	printk(KERN_INFO "PORTMAN_PCS: close minor number %d\n", MINOR(inode->i_rdev));

	if (waitqueue_active(&PORTMAN_PCS_buffer_wait_queue))
		wake_up_interruptible(&PORTMAN_PCS_buffer_wait_queue);

	del_timer(&PORTMAN_PCS_buffer_timer);

	PORTMAN_PCS_open_time = 0;
	PORTMAN_PCS_timer_start = 0;

	MOD_DEC_USE_COUNT;

	return 0;
}

static ssize_t PORTMAN_PCS_file_write (struct file *file, const char *data, size_t length, loff_t *offset)
{
//	printk(KERN_INFO "PORTMAN_PCS: file write %d bytes to minor number %d\n",length, (int) file->private_data); 

	if (((int) file->private_data) == PORTMAN_PCS_MIDI_MINOR_NUMBER)
	{
		// ecriture vers /dev/midi sans timer
		copy_from_user(PORTMAN_PCS_outbuffer, data, length);
		return PORTMAN_PCS_write_to_port(PORTMAN_PCS_outbuffer, length);
	}
	else if ((int) file->private_data == (PORTMAN_PCS_SEQUENCER_MINOR_NUMBER | PORTMAN_PCS_SEQUENCER2_MINOR_NUMBER))
	{
		// ecriture vers /dev/sequencer sans timer

		if ((file->f_flags & O_NDELAY) && PORTMAN_PCS_outbuffer_length)
			return -EAGAIN;
		else if (PORTMAN_PCS_outbuffer_length)
			interruptible_sleep_on_timeout(&PORTMAN_PCS_buffer_wait_queue, 1000);

		if (PORTMAN_PCS_outbuffer_length)
		{
			printk(KERN_ERR "PORTMAN_PCS: Timeout while waiting for write buffer to flush\n");
			return -1;
		}	

		/*printk(KERN_INFO "PORTMAN_PCS: copying %d bytes to buffer index %d\n", length, PORTMAN_PCS_outbuffer_pointer); */

		if (length > sizeof(PORTMAN_PCS_outbuffer))
			length = sizeof(PORTMAN_PCS_outbuffer);

		copy_from_user(PORTMAN_PCS_outbuffer, data, length);

		PORTMAN_PCS_outbuffer_length = length;
		PORTMAN_PCS_outbuffer_pointer = 0;

		init_timer(&PORTMAN_PCS_buffer_timer);
		PORTMAN_PCS_buffer_timer.expires = jiffies + 1;
		PORTMAN_PCS_buffer_timer.data = 0;
		PORTMAN_PCS_buffer_timer.function = seq_timing_event;
		add_timer(&PORTMAN_PCS_buffer_timer);
	}

	return length;
}


static ssize_t PORTMAN_PCS_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int scan = 0;
	int select =0;
	
	static int previous_timer = 0;
	int current_jiffies = jiffies - PORTMAN_PCS_timer_start;
	double real_timer = ((double)PORTMAN_PCS_timer_base) * ((double)PORTMAN_PCS_timer_tempo) * ((double)current_jiffies) / 60.0 / ((double)HZ);
	int current_timer = (int) real_timer;
	unsigned char status_byte = 0;
	unsigned char second_byte = 0;
	unsigned char third_byte = 0;
	int retry = RETRYTIMEOUT;

	
//	printk(KERN_INFO "PORTMAN_PCS: file read %d bytes to minor number %d\n",length, (int) file->private_data); 


	if (((int) file->private_data) == PORTMAN_PCS_MIDI_MINOR_NUMBER)
	{
		//printk(KERN_INFO "PORTMAN_PCS: read : /dev/midi03\n");

		for (scan = 0; scan <= length; ++scan)
			return	PORTMAN_PCS_read_to_port(&data[scan]);	
			
		
	}
	else if ((int) file->private_data == (PORTMAN_PCS_SEQUENCER_MINOR_NUMBER | PORTMAN_PCS_SEQUENCER2_MINOR_NUMBER))
	{

		//printk(KERN_INFO "PORTMAN_PCS: read : /dev/sequencer(1-2)\n");

		//clean data buffer
		//copy_to_user(PORTMAN_PCS_inbuffer,0,length);
	

		if (PORTMAN_PCS_echo_sync_time)
		{
		//	printk(KERN_INFO "PORTMAN_PCS: sync echo read\n");
			put_user((unsigned char)EV_TIMING, &data[0]);
			put_user((unsigned char)TMR_ECHO, &data[1]);
			put_user((unsigned char)0, &data[2]);
			put_user((unsigned char)0, &data[3]);
			put_user(current_timer, (int*)&data[4]);
		
			PORTMAN_PCS_echo_sync_time = 0;
			scan += 8;
		}	
		else if ((PORTMAN_PCS_timer_start > 0) && (current_timer != previous_timer))
		{
		//	printk(KERN_INFO "PORTMAN_PCS: read Timer\n");
			put_user((unsigned char)EV_TIMING, &data[0]);
			put_user((unsigned char)TMR_WAIT_ABS, &data[1]);
			put_user((unsigned char)0, &data[2]);
			put_user((unsigned char)0, &data[3]);
			put_user(current_timer, (int*)&data[4]);
		
			previous_timer = current_timer;
			scan += 8;
		}

		
		while (PORTMAN_PCS_read_to_port(&data[scan]) == 1)
		{	
			get_user(status_byte, &data[scan]);
		
			switch (status_byte & 0xf0)  // ignore active sensing and sync time
			{
			
				case SEQ_PRIVATE:
					break;
				case MIDI_SYSTEM_PREFIX:
					put_user((unsigned char) MIDI_SYSTEM_PREFIX, &data[scan]);
					put_user((unsigned char) 0, &data[scan + 1]);
					put_user((unsigned char) status_byte & 0x0f, &data[scan + 2]);
				break;

				default:
					while (retry && (PORTMAN_PCS_read_to_port(&data[scan])) == 0)
						--retry;
					get_user(second_byte, &data[scan]);
		
					if (retry)
						switch (status_byte & 0xf0)
						{
							case MIDI_PGM_CHANGE:
							case MIDI_CHN_PRESSURE:
								put_user((unsigned char) EV_CHN_COMMON, &data[scan]);
								put_user((unsigned char) 0, &data[scan + 1]);
								put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
								put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
								put_user((unsigned char) second_byte, &data[scan + 4]);
								scan += 8;
							break;

							case MIDI_PITCH_BEND:
								select = 1;
							case MIDI_CTL_CHANGE:
								while (retry && (PORTMAN_PCS_read_to_port(&data[scan])) == 0)
									--retry;
								get_user(third_byte, &data[scan]);
			
								put_user((unsigned char) EV_CHN_COMMON, &data[scan]);
								put_user((unsigned char) 0, &data[scan + 1]);
								put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
								put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
								if (select != 1)
									put_user((unsigned char) second_byte, &data[scan + 4]);
								else
									put_user((unsigned char) 0, &data[scan + 4]);
								put_user((unsigned char) 0, &data[scan + 5]);
								put_user((unsigned char) second_byte & 0xf0, &data[scan + 6]);
								put_user((unsigned char) second_byte & 0x0f, &data[scan + 7]);
				
								if (retry)
									scan += 8;
							break;

							case MIDI_NOTEON:
								select = 2;
							case MIDI_NOTEOFF:
								evt_prec = status_byte;
							case MIDI_KEY_PRESSURE:
								while (retry && (PORTMAN_PCS_read_to_port(&data[scan])) == 0) //third byte
									--retry;
								get_user(third_byte, &data[scan]);
					
								if ((select == 2) && (third_byte == 0))
									status_byte = 0x80 & (status_byte & 0xf0);
								put_user((unsigned char) EV_CHN_VOICE, &data[scan]);
								put_user((unsigned char) 0, &data[scan + 1]);
								put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
								put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
								put_user((unsigned char) second_byte, &data[scan + 4]);
								put_user((unsigned char) third_byte, &data[scan + 5]);
					
								if (retry)
									scan += 8;
							break;
									
							default:
								put_user((unsigned char) EV_CHN_VOICE, &data[scan]);
								put_user((unsigned char) 0, &data[scan + 1]);
								put_user((unsigned char) evt_prec & 0xf0, &data[scan + 2]);
								put_user((unsigned char) evt_prec & 0x0f, &data[scan + 3]);
								put_user((unsigned char) status_byte, &data[scan + 4]);
								put_user((unsigned char) second_byte, &data[scan + 5]);
								scan += 8;
	
							//printk(KERN_INFO "PORTMAN_PCS: input event not supported %x %x %x\n",evt_prec, status_byte,second_byte);
				}//switch status byte
		}//switch status byte
				
	}
		

	} //fin du else if


	return scan;
}

static int PORTMAN_PCS_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
//	printk(KERN_INFO "PORTMAN_PCS: ioctl minor number : %d cmd : %x\n",(int) file->private_data,command); 

	int *count;
	struct synth_info info;
	struct synth_info *user_info = (struct synth_info *) data;
	int *timer_resolution;
	struct seq_event_rec seq_evt;
	int *new_base;
	int *new_tempo;
	int *free_memory;

	switch (command)
	{
		case SNDCTL_SEQ_NRSYNTHS:
			count = (int*) data;
			put_user(1, count);
		break;

		case SNDCTL_SEQ_NRMIDIS:
			count = (int*) data;
			put_user(1, count);
		break;

		case SNDCTL_SYNTH_INFO:
			memset(&info, 0, sizeof(info));
			strcpy(info.name, "PORTMAN_PCS");
			info.synth_type = SYNTH_TYPE_MIDI;
			//info.synth_subtype = MIDI_TYPE_MPU401;
			info.nr_voices = 2; /* arbitrarily chosen */
			copy_to_user(user_info, &info, sizeof(info));
		break;

		case SNDCTL_MIDI_INFO:
			memset(&info, 0, sizeof(info));
			strcpy(info.name, "PORTMAN_PCS");
			copy_to_user(user_info, &info, sizeof(info));
		break;

		case SNDCTL_SEQ_RESET:
			/* doesn't actually do anything - but acknowledge so there are no problems */
			PORTMAN_PCS_open_time = jiffies;
			PORTMAN_PCS_timer_start = 0;
		break;

		case SNDCTL_SEQ_SYNC:
			/* doesn't actually do anything - but acknowledge so there are no problems */
			 PORTMAN_PCS_open_time = jiffies;
		break;

		case SNDCTL_SEQ_CTRLRATE:
			timer_resolution = (int*) data;
			put_user(HZ, timer_resolution);
		break;

		case SNDCTL_SEQ_OUTOFBAND:
			//ADD this new function for realtime event :)
						
			copy_from_user(&seq_evt, (char *)data, sizeof(seq_evt));
			play_event(seq_evt.arr);
		break;

		case SNDCTL_TMR_TIMEBASE:
			new_base = (int*) data;
			get_user(PORTMAN_PCS_timer_base, new_base);
			/* printk(KERN_INFO "PORTMAN_PCS: New base %d\n", PORTMAN_PCS_timer_base); */
		break;

		case SNDCTL_TMR_TEMPO:
			new_tempo = (int*) data;
			get_user(PORTMAN_PCS_timer_tempo, new_tempo);
			/* printk(KERN_INFO "PORTMAN_PCS: New tempo %d\n",PORTMAN_PCS_timer_tempo);*/ 
		break;

		case SNDCTL_SYNTH_MEMAVL:
			/* doesn't actually do anything - but acknowledge so there are no problems */
			free_memory = (int*) data;
			put_user(MAX*2, free_memory);
			printk(KERN_INFO "PORTMAN_PCS: Memavl request\n");
		break;

		default:
			printk(KERN_INFO "PORTMAN_PCS: undefined ioctl %x\n", command);
		    return -1;
	}

	return 0;
}


struct file_operations PORTMAN_PCS_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	PORTMAN_PCS_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	PORTMAN_PCS_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	PORTMAN_PCS_file_ioctl, /* int (*ioctl) (struct inode *,struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	PORTMAN_PCS_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	PORTMAN_PCS_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL,  /*ssize_t (*writepage) (struct file *, struct page *,int,size_t, loff_t *, int);*/
#endif

};


/************************ Module entry/exit points ********************************/

int init_module(void)
{
	int registration_status = 0;

	PORTMAN_PCS_open_time = 0;
	PORTMAN_PCS_timer_start = 0;

	PORTMAN_PCS_outbuffer_pointer = 0;
	PORTMAN_PCS_outbuffer_length = 0;

	init_timer(&PORTMAN_PCS_buffer_timer);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#	init_waitqueue(&PORTMAN_PCS_buffer_wait_queue);
#else
	init_waitqueue_head(&PORTMAN_PCS_buffer_wait_queue);
#endif

	if ((registration_status = register_chrdev(PORTMAN_PCS_MAJOR_NUMBER,"PORTMAN_PCS", &PORTMAN_PCS_file_operations)) < 0)
	{
		printk(KERN_ERR "PORTMAN_PCS: register_chrdev() failed code %d\n", registration_status);
		return -1;
	}

	else if ((PORTMAN_PCS_serial_port_tty = PORTMAN_PCS_open_serial_port(PORTMAN_PCS_SERIAL_PORT_DEVICE_NAME)) == NULL)
	{
		unregister_chrdev(PORTMAN_PCS_MAJOR_NUMBER, "PORTMAN_PCS");
		return -1;
	}

	printk(KERN_INFO "PORTMAN_PCS loaded -------\n");

	return 0;
}

void cleanup_module(void)
{
	if (waitqueue_active(&PORTMAN_PCS_buffer_wait_queue))
		wake_up_interruptible(&PORTMAN_PCS_buffer_wait_queue);

	del_timer(&PORTMAN_PCS_buffer_timer);

	if (!IS_ERR(PORTMAN_PCS_serial_port_file))
		filp_close(PORTMAN_PCS_serial_port_file, 0);

	unregister_chrdev(PORTMAN_PCS_MAJOR_NUMBER, "PORTMAN_PCS");

	PORTMAN_PCS_serial_port_file = 0;
	PORTMAN_PCS_serial_port_tty = 0;

	printk(KERN_INFO "PORTMAN_PCS unloaded --------\n");
}


