/*
 *  Copyright (c) 2001 Edouard TISSERANT          <tissered@esstin.uhp-nancy.fr>
 *
 *  USB Acecad "Acecat Flair" tablet support
 *
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb.h>

MODULE_AUTHOR("Edouard TISSERANT <tissered@esstin.uhp-nancy.fr>");
MODULE_DESCRIPTION("USB Acecat Flair tablet driver");
MODULE_LICENSE("GPL");

#define USB_VENDOR_ID_ACECAD	0x0460

struct usb_acecad {
	signed char data[8];
	char name[128];
	struct usb_device *usbdev;
	struct input_dev dev;
	struct urb irq;
	int open;
};

static void usb_acecad_irq(struct urb *urb)
{
	struct usb_acecad *acecad = urb->context;
	unsigned char *data = acecad->data;
	struct input_dev *dev = &acecad->dev;
	int prox;

	if (urb->status) return;

	prox = (data[0] & 0x04) >> 2;

	input_report_key(dev, BTN_TOOL_PEN, prox);
	
	if (prox) {
		int x = data[1] | (data[2] << 8);
		int y = data[3] | (data[4] << 8);
		int pressure = data[5] | (data[6] << 8);
		int touch = data[0] & 0x01;
		int stylus = (data[0] & 0x10) >> 4;
		int stylus2 = (data[0] & 0x20) >> 5;
/*		int i;*/
		input_report_abs(dev, ABS_X, x);
		input_report_abs(dev, ABS_Y, y);
		input_report_abs(dev, ABS_PRESSURE, pressure); 
		input_report_key(dev, BTN_TOUCH, touch);
		input_report_key(dev, BTN_STYLUS, stylus);
		input_report_key(dev, BTN_STYLUS2, stylus2);
/*		printk("X %d\n", x);
		printk("Y %d\n", y);
		printk("%x ", data[6]);
		for(i=7;i>=0;i--)
		printk("%d", (pressure & (1<<i))>>i);
		printk(" %d\n", pressure);
		printk("STYLUS %d\n", stylus);
		printk("STYLUS2 %d\n", stylus2);*/
	}
	
	/* report the id of the current tool */
	/* it must be the last packet sent to indicate the event end */
	input_report_abs(dev, ABS_MISC, 1);
	dev->abs[ABS_MISC] = 0;

}

static int usb_acecad_open(struct input_dev *dev)
{
	struct usb_acecad *acecad = dev->private;

	if (acecad->open++)
		return 0;

	acecad->irq.dev = acecad->usbdev;
	if (usb_submit_urb(&acecad->irq))
		return -EIO;

	return 0;
}

static void usb_acecad_close(struct input_dev *dev)
{
	struct usb_acecad *acecad = dev->private;

	if (!--acecad->open)
		usb_unlink_urb(&acecad->irq);
}

static void *usb_acecad_probe(struct usb_device *dev, unsigned int ifnum,
			     const struct usb_device_id *id)
{
	struct usb_interface *iface;
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct usb_acecad *acecad;
	int pipe, maxp;
	char *buf;

	iface = &dev->actconfig->interface[ifnum];
	interface = &iface->altsetting[iface->act_altsetting];

	if (interface->bNumEndpoints != 1) return NULL;

	endpoint = interface->endpoint + 0;
	if (!(endpoint->bEndpointAddress & 0x80)) return NULL;
	if ((endpoint->bmAttributes & 3) != 3) return NULL;

	pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
	maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));

	usb_set_idle(dev, interface->bInterfaceNumber, 0, 0);

	if (!(acecad = kmalloc(sizeof(struct usb_acecad), GFP_KERNEL))) return NULL;
	memset(acecad, 0, sizeof(struct usb_acecad));

	acecad->usbdev = dev;

	acecad->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	acecad->dev.absbit[0] = BIT(ABS_X) | BIT(ABS_Y) | BIT(ABS_PRESSURE) | BIT(ABS_MISC);
	acecad->dev.keybit[LONG(BTN_LEFT)] = BIT(BTN_LEFT) | BIT(BTN_RIGHT) | BIT(BTN_MIDDLE);
	acecad->dev.keybit[LONG(BTN_DIGI)] = BIT(BTN_TOOL_PEN) |BIT(BTN_TOUCH) | BIT(BTN_STYLUS) | BIT(BTN_STYLUS2);

	acecad->dev.absmax[ABS_X] = 5000;
	acecad->dev.absmax[ABS_Y] = 3750;
	acecad->dev.absmax[ABS_PRESSURE] = 512;
	acecad->dev.absmin[ABS_MISC] = 0;
	acecad->dev.absmax[ABS_MISC] = 4294967295U;
	acecad->dev.absfuzz[ABS_X] = 4;
	acecad->dev.absfuzz[ABS_Y] = 4;

	acecad->dev.private = acecad;
	acecad->dev.open = usb_acecad_open;
	acecad->dev.close = usb_acecad_close;

	acecad->dev.name = acecad->name;
	acecad->dev.idbus = BUS_USB;
	acecad->dev.idvendor = dev->descriptor.idVendor;
	acecad->dev.idproduct = dev->descriptor.idProduct;
	acecad->dev.idversion = dev->descriptor.bcdDevice;

	if (!(buf = kmalloc(63, GFP_KERNEL))) {
		kfree(acecad);
		return NULL;
	}

	if (dev->descriptor.iManufacturer &&
		usb_string(dev, dev->descriptor.iManufacturer, buf, 63) > 0)
			strcat(acecad->name, buf);
	if (dev->descriptor.iProduct &&
		usb_string(dev, dev->descriptor.iProduct, buf, 63) > 0)
			sprintf(acecad->name, "%s %s", acecad->name, buf);

	if (!strlen(acecad->name))
		sprintf(acecad->name, "USB HIDBP Mouse %04x:%04x",
			acecad->dev.idvendor, acecad->dev.idproduct);

	kfree(buf);

	FILL_INT_URB(&acecad->irq, dev, pipe, acecad->data, maxp > 8 ? 8 : maxp,
		usb_acecad_irq, acecad, endpoint->bInterval);

	input_register_device(&acecad->dev);

	printk(KERN_INFO "input%d: %s with packet size %d on usb%d:%d.%d\n",
		 acecad->dev.number, acecad->name, maxp,  dev->bus->busnum, dev->devnum, ifnum);

	return acecad;
}

static void usb_acecad_disconnect(struct usb_device *dev, void *ptr)
{
	struct usb_acecad *acecad = ptr;
	usb_unlink_urb(&acecad->irq);
	input_unregister_device(&acecad->dev);
	kfree(acecad);
}

static struct usb_device_id usb_acecad_id_table [] = {
	{ idVendor: USB_VENDOR_ID_ACECAD, idProduct: 0x0004, driver_info: 0,
	  match_flags: USB_DEVICE_ID_MATCH_VENDOR|USB_DEVICE_ID_MATCH_PRODUCT },
    { }						/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, usb_acecad_id_table);

static struct usb_driver usb_acecad_driver = {
	name:		"usb_acecad",
	probe:		usb_acecad_probe,
	disconnect:	usb_acecad_disconnect,
	id_table:	usb_acecad_id_table,
};

static int __init usb_acecad_init(void)
{
	usb_register(&usb_acecad_driver);
	return 0;
}

static void __exit usb_acecad_exit(void)
{
	usb_deregister(&usb_acecad_driver);
}

module_init(usb_acecad_init);
module_exit(usb_acecad_exit);
