/* $Id: audiofunc.c,v 1.51 2003/06/24 02:24:57 cochran Exp $
 *
 * Driver for the Extigy (tm) USB sound card - generic audio funtions.
 *
 * Copyright (C) 1999, 2000 Alan Cox (alan@lxorguk.ukuu.org.uk) and
 * Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 * Adapted from the generic USB audio driver.
 * Copyright (c) 2003 Richard Cochran (cochran@cs.umass.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */

/* Notes from the audio.c authors:
 *
 * Data IO implementation issues
 *
 * A mmap'able ring buffer per direction is implemented, because
 * almost every OSS app expects it. It is however impractical to
 * transmit/receive USB data directly into and out of the ring buffer,
 * due to alignment and synchronisation issues. Instead, the ring buffer
 * feeds a constant time delay line that handles the USB issues.
 *
 * The AudioStreaming parser simply enumerates all altsettings belonging
 * to the specified interface. It looks for AS_GENERAL and FORMAT_TYPE
 * class specific descriptors to extract the sample format/sample rate
 * data. Only sample format types PCM and PCM8 are supported right now, and
 * only FORMAT_TYPE_I is handled. The isochronous data endpoint needs to
 * be the first endpoint of the interface, and the optional synchronisation
 * isochronous endpoint the second one.
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/module.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/usb.h>

#include "exaudio.h"
#include "mixer.h"
#include "dma.h"

#define DRIVER_AUTHOR  "Richard Cochran, Alan Cox, Thomas Sailer"
#define DRIVER_DESC    "exaudio driver"
#define DRIVER_VERSION "$Revision: 1.51 $"

#define USB_DT_CS_DEVICE                0x21
#define USB_DT_CS_CONFIG                0x22
#define USB_DT_CS_STRING                0x23
#define USB_DT_CS_INTERFACE             0x24
#define USB_DT_CS_ENDPOINT              0x25

#define CS_AUDIO_UNDEFINED		0x20
#define CS_AUDIO_DEVICE			0x21
#define CS_AUDIO_CONFIGURATION		0x22
#define CS_AUDIO_STRING			0x23
#define CS_AUDIO_INTERFACE		0x24
#define CS_AUDIO_ENDPOINT		0x25

#define HEADER				0x01
#define INPUT_TERMINAL			0x02
#define OUTPUT_TERMINAL			0x03
#define MIXER_UNIT			0x04
#define SELECTOR_UNIT			0x05
#define FEATURE_UNIT			0x06
#define PROCESSING_UNIT			0x07
#define EXTENSION_UNIT			0x08

#define AS_GENERAL			0x01
#define FORMAT_TYPE			0x02
#define FORMAT_SPECIFIC			0x03
#define EP_GENERAL			0x01
#define MAX_CHAN   9
#define MAX_FREQ   16
#define MAX_IFACE  8
#define MAX_FORMAT 8
#define MAX_ALT    32 /* Sorry, we need quite a few for the Philips webcams */

#define FLG_URB0RUNNING   1
#define FLG_URB1RUNNING   2
#define FLG_SYNC0RUNNING  4
#define FLG_SYNC1RUNNING  8
#define FLG_RUNNING      16
#define FLG_CONNECTED    32

#define SND_DEV_DSP16 5

#define MAXFORMATS        MAX_ALT

/*
 * This influences:
 * - Latency
 * - Interrupt rate
 * - Synchronisation behaviour
 * Don't touch this if you don't understand all of the above.
 */
#define DESCFRAMES  5
#define SYNCFRAMES  DESCFRAMES

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct iso_packet_descriptor iso_packet_descriptor_t;
#endif

/* Linked list of all audio devices... */
struct list_head alldevices = LIST_HEAD_INIT(alldevices);

DECLARE_MUTEX(open_sem);

/* Wait queue for processes wanting to open an USB audio device */
static DECLARE_WAIT_QUEUE_HEAD(open_wait);

struct audioformat {
  unsigned int format;
  unsigned int sratelo;
  unsigned int sratehi;
  unsigned char altsetting;
  unsigned char attributes;
};

struct my_data_urb {
  struct urb urb;
  iso_packet_descriptor_t isoframe[DESCFRAMES];
};

struct my_sync_urb {
  struct urb urb;
  iso_packet_descriptor_t isoframe[SYNCFRAMES];
};

struct iostream {
  int interface;              /* Interface number, -1 means not used */
  unsigned int format;        /* USB data format */
  unsigned int datapipe;      /* the data input pipe */
  unsigned int syncpipe;      /* the sync pipe for asynchronous mode */
  unsigned int syncinterval;  /* P for asynchronous mode, 0 otherwise */
  unsigned int freqn;   /* nominal sampling rate USB format fs/1000 in Q10.14 */
  unsigned int freqm;   /* momentary sampling rate in USB format, OUT only */
  unsigned int freqmax; /* maximum sampling rate, for buffer management */
  unsigned int phase;         /* phase accumulator */
  unsigned int flags;         /* see FLG_ defines */
  struct my_data_urb durb[2]; /* ISO descriptors for the data endpoint */
  struct my_sync_urb surb[2]; /* ISO sync pipe descriptor if needed */
  struct dmabuf dma;
} usbout;

#define TO_HOST 0
#define TO_DEV  1

/*  A "struct audiodev" is created for each pair of AudioStreaming 
 *  interfaces that we group together, IN and OUT.
 */
struct audiodev {
  struct list_head list; /* points to next ASIF in chain */
  struct extigy *state;  /* points back to the device struct */
  spinlock_t lock;       /* DMA buffer access spinlock */
  struct iostream io[2]; /* one for each direction, IN and OUT */
  unsigned int numfmtin, numfmtout;
  struct audioformat fmtin[MAXFORMATS];
  struct audioformat fmtout[MAXFORMATS];
};

/* private prototypes */

static void* exaudio_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id);

static void exaudio_disconnect(struct usb_device *dev, void *ptr);

/*
 *  USB I/O code.
 */

static void stop_stream(int direction,struct audiodev *as)
{
  struct iostream* u = &as->io[direction];
  unsigned long flags;
  unsigned int i, notkilled = 1;

  spin_lock_irqsave(&as->lock,flags);
  u->flags &= ~FLG_RUNNING;
  i = u->flags;
  spin_unlock_irqrestore(&as->lock,flags);
  while (i &
	 (FLG_URB0RUNNING | FLG_URB1RUNNING | FLG_SYNC0RUNNING |
	  FLG_SYNC1RUNNING)) {
    set_current_state(notkilled ? TASK_INTERRUPTIBLE : TASK_UNINTERRUPTIBLE);
    schedule_timeout(1);
    spin_lock_irqsave(&as->lock,flags);
    i = u->flags;
    spin_unlock_irqrestore(&as->lock,flags);
    if (notkilled && signal_pending(current)) {
      if (i & FLG_URB0RUNNING)
	usb_unlink_urb(&u->durb[0].urb);
      if (i & FLG_URB1RUNNING)
	usb_unlink_urb(&u->durb[1].urb);
      if (i & FLG_SYNC0RUNNING)
	usb_unlink_urb(&u->surb[0].urb);
      if (i & FLG_SYNC1RUNNING)
	usb_unlink_urb(&u->surb[1].urb);
      notkilled = 0;
    }
  }
  set_current_state(TASK_RUNNING);
  if (u->durb[0].urb.transfer_buffer)
    kfree(u->durb[0].urb.transfer_buffer);
  if (u->durb[1].urb.transfer_buffer)
    kfree(u->durb[1].urb.transfer_buffer);
  if (u->surb[0].urb.transfer_buffer)
    kfree(u->surb[0].urb.transfer_buffer);
  if (u->surb[1].urb.transfer_buffer)
    kfree(u->surb[1].urb.transfer_buffer);
  u->durb[0].urb.transfer_buffer = u->durb[1].urb.transfer_buffer =
    u->surb[0].urb.transfer_buffer = u->surb[1].urb.transfer_buffer = NULL;
}

static void disc_stream(int direction,struct audiodev *as)
{
  struct iostream* u = &as->io[direction];
  unsigned long flags;

  spin_lock_irqsave(&as->lock,flags);
  u->flags &= ~(FLG_RUNNING|FLG_CONNECTED);
  spin_unlock_irqrestore(&as->lock,flags);
  stop_stream(direction,as);
}

/*
 *   Prepare, retire, and completion for isochronous transfers:
 */

static int usbin_prepare_desc(struct iostream* u, struct urb* urb)
{
  unsigned int i, maxsize, offs;

  maxsize = (u->freqmax + 0x3fff) >> (14 - AFMT_BYTESSHIFT(u->format));
  for (i = offs = 0; i < DESCFRAMES; i++, offs += maxsize) {
    urb->iso_frame_desc[i].length = maxsize;
    urb->iso_frame_desc[i].offset = offs;
  }
  return 0;
}

static int usbout_prepare_desc(struct iostream* u, struct urb* urb)
{
  unsigned int i, ufmtsh, dfmtsh, err = 0, cnt, scnt, offs;
  unsigned char *cp = urb->transfer_buffer;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  for (i = offs = 0; i < DESCFRAMES; i++) {
    urb->iso_frame_desc[i].offset = offs;
    u->phase = (u->phase & 0x3fff) + u->freqm;
    scnt = u->phase >> 14;
    if (!scnt) {
      urb->iso_frame_desc[i].length = 0;
      continue;
    }
    cnt = scnt << dfmtsh;
    if (!u->dma.mapped) {
      if (cnt > u->dma.count) {
	scnt = u->dma.count >> dfmtsh;
	cnt = scnt << dfmtsh;
	err++;
      }
      u->dma.count -= cnt;
    } else
      u->dma.count += cnt;
    if (u->format == u->dma.format) {
      /* We never do format conversion! */
      dmabuf_copyout(&u->dma, cp, cnt);
    } else {
      ERR("OUT sample format mismatch ios 0x%X dma 0x%X",
	  u->format,u->dma.format);
      dmabuf_copyout(&u->dma, cp, cnt);
    }
    cnt = scnt << ufmtsh;
    urb->iso_frame_desc[i].length = cnt;
    offs += cnt;
    cp += cnt;
  }
  if (err)
    u->dma.error++;
  if (u->dma.mapped) {
    if (u->dma.count >= (signed) u->dma.fragsize)
      wake_up(&u->dma.wait);
  } else {
    if ((signed) u->dma.dmasize >= u->dma.count + (signed) u->dma.fragsize)
      wake_up(&u->dma.wait);
  }
  return err ? -1 : 0;
}

static int prepare_desc(int direction,struct iostream* u,struct urb* urb)
{
  if (TO_HOST == direction) 
    return(usbin_prepare_desc(u,urb));
  else
    return(usbout_prepare_desc(u,urb));
}

/* Returns 0 if descriptor should be restarted, -1 otherwise */
static int usbin_retire_desc(struct iostream* u, struct urb* urb)
{
  unsigned int i, ufmtsh, dfmtsh, err = 0, cnt, scnt, dmafree;
  unsigned char *cp;

  ufmtsh = AFMT_BYTESSHIFT(u->format);
  dfmtsh = AFMT_BYTESSHIFT(u->dma.format);
  for (i = 0; i < DESCFRAMES; i++) {
    cp =
      ((unsigned char *) urb->transfer_buffer) +
      urb->iso_frame_desc[i].offset;
    if (urb->iso_frame_desc[i].status) {
      dbg("usbin_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
    scnt = urb->iso_frame_desc[i].actual_length >> ufmtsh;
    if (!scnt)
      continue;
    cnt = scnt << dfmtsh;
    if (!u->dma.mapped) {
      dmafree = u->dma.dmasize - u->dma.count;
      if (cnt > dmafree) {
	scnt = dmafree >> dfmtsh;
	cnt = scnt << dfmtsh;
	err++;
      }
    }
    u->dma.count += cnt;
    if (u->format == u->dma.format) {
      /* we do not need format conversion */
      dmabuf_copyin(&u->dma, cp, cnt);
    } else {
      ERR("IN sample format mismatch");
      dmabuf_copyin(&u->dma, cp, cnt);
    }
  }
  if (err)
    u->dma.error++;
  if (u->dma.count >= (signed) u->dma.fragsize)
    wake_up(&u->dma.wait);
  return err ? -1 : 0;
}

static int usbout_retire_desc(struct iostream* u, struct urb* urb)
{
  unsigned int i;

  for (i = 0; i < DESCFRAMES; i++) {
    if (urb->iso_frame_desc[i].status) {
      dbg("usbout_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
  }
  return 0;
}

static int retire_desc(int direction,struct iostream* u,struct urb* urb)
{
  if (TO_HOST == direction) 
    return(usbin_retire_desc(u,urb));
  else
    return(usbout_retire_desc(u,urb));
}

static void stream_completion(int direction,struct urb *urb)
{
  struct audiodev *as = (struct audiodev*) urb->context;
  struct iostream* u = &as->io[direction];
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

  if (urb == &u->durb[0].urb) {
    mask = FLG_URB0RUNNING;
  } else if (urb == &u->durb[1].urb) {
    mask = FLG_URB1RUNNING;
  } else {
    mask = 0;
    ERR("completion: panic: unknown URB");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock,flags);
  if ( !retire_desc(direction,u,urb)  &&
       u->flags & FLG_RUNNING         &&
       !prepare_desc(direction,u,urb) &&
       (suret = usb_submit_urb(urb)) == USB_ST_NOERROR ) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask|FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("completion: descriptor not restarted (usb_submit_urb: %d)",suret);
  }
  spin_unlock_irqrestore(&as->lock,flags);
}

static void usbin_completed(struct urb *urb)
{
  stream_completion(TO_HOST,urb);
}

static void usbout_completed(struct urb *urb)
{
  stream_completion(TO_DEV,urb);
}

/*
 *   Prepare, retire, and completion for sync data:
 */

static int sync_prepare_desc(int direction,struct iostream* u,struct urb* urb)
{
  unsigned char *cp = urb->transfer_buffer;
  unsigned int i, offs;

  for (i = offs = 0; i < SYNCFRAMES; i++, offs += 3, cp += 3) {
    urb->iso_frame_desc[i].length = 3;
    urb->iso_frame_desc[i].offset = offs;
    if (TO_HOST == direction) {
      cp[0] = u->freqn;
      cp[1] = u->freqn >> 8;
      cp[2] = u->freqn >> 16;
    }
  }
  return(0);
}

/* Returns 0 if descriptor should be restarted, -1 otherwise */
static int usbin_sync_retire_desc(struct iostream* u, struct urb* urb)
{
  unsigned int i;

  for (i = 0; i < SYNCFRAMES; i++)
    if (urb->iso_frame_desc[0].status)
      dbg("usbin_sync_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
  return 0;
}

/* Returns 0 if descriptor should be restarted, -1 otherwise */
static int usbout_sync_retire_desc(struct iostream* u, struct urb* urb)
{
  unsigned char *cp = urb->transfer_buffer;
  unsigned int f, i;

  for (i = 0; i < SYNCFRAMES; i++, cp += 3) {
    if (urb->iso_frame_desc[i].status) {
      dbg("usbout_sync_retire_desc: frame %u status %d",
	  i,urb->iso_frame_desc[i].status);
      continue;
    }
    if (urb->iso_frame_desc[i].actual_length < 3) {
      dbg("usbout_sync_retire_desc: frame %u length %d",
	  i,urb->iso_frame_desc[i].actual_length);
      continue;
    }
    f = cp[0] | (cp[1] << 8) | (cp[2] << 16);
    if (abs(f - u->freqn) > (u->freqn >> 3) || f > u->freqmax) {
      ERR("sync_retire_desc: requested freq %u (nominal %u) out of range!\n",
	  f,u->freqn);
      continue;
    }
    u->freqm = f;
    //ERR("freqm=%08X or %u.%u",f,f>>14,f&0x3fff);
  }
  return 0;
}

static int sync_retire_desc(int direction,struct iostream* u,struct urb* urb)
{
  if (TO_HOST == direction) 
    return(usbin_sync_retire_desc(u,urb));
  else
    return(usbout_sync_retire_desc(u,urb));
}

static void sync_completed(int direction,struct urb *urb)
{
  struct audiodev *as = (struct audiodev*) urb->context;
  struct iostream* u = &as->io[direction];
  unsigned long flags;
  unsigned int mask;
  int suret = USB_ST_NOERROR;

  if (urb == &u->surb[0].urb) {
    mask = FLG_SYNC0RUNNING;
  } else if (urb == &u->surb[1].urb) {
    mask = FLG_SYNC1RUNNING;
  } else {
    mask = 0;
    ERR("sync_completed: panic: unknown URB");
  }
  urb->dev = as->state->usbdev;
  spin_lock_irqsave(&as->lock,flags);
  if ( !sync_retire_desc(direction,u,urb)  &&
       u->flags & FLG_RUNNING              &&
       !sync_prepare_desc(direction,u,urb) &&
       (suret = usb_submit_urb(urb)) == USB_ST_NOERROR ) {
    u->flags |= mask;
  } else {
    u->flags &= ~(mask|FLG_RUNNING);
    wake_up(&u->dma.wait);
    dbg("sync_completed: descriptor not restarted (usb_submit_urb: %d)",suret);
  }
  spin_unlock_irqrestore(&as->lock,flags);
}

static void usbin_sync_completed(struct urb *urb)
{
  sync_completed(TO_HOST,urb);
}

static void usbout_sync_completed(struct urb *urb)
{
  sync_completed(TO_DEV,urb);
}

static int start_stream(int direction,struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct iostream* u = &as->io[direction];
  struct urb* urb;
  unsigned long flags;
  unsigned int maxsze, bufsz;

  /* allocate USB storage if not already done */
  spin_lock_irqsave(&as->lock,flags);
  if (!(u->flags & FLG_CONNECTED)) {
    spin_unlock_irqrestore(&as->lock,flags);
    return -EIO;
  }

  if (!(u->flags & FLG_RUNNING)) {
    spin_unlock_irqrestore(&as->lock,flags);
    u->freqn = u->freqm = ((u->dma.srate<<11) + 62) / 125;
    /* overflows at approx 2MSPS */
    u->freqmax = u->freqn + (u->freqn >> 2);
    u->phase = 0;
    maxsze = (u->freqmax + 0x3fff) >> (14 - AFMT_BYTESSHIFT(u->format));
    bufsz = DESCFRAMES * maxsze;

    if (u->durb[0].urb.transfer_buffer)
      kfree(u->durb[0].urb.transfer_buffer);
    u->durb[0].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[0].urb.transfer_buffer_length = bufsz;
    if (u->durb[1].urb.transfer_buffer)
      kfree(u->durb[1].urb.transfer_buffer);
    u->durb[1].urb.transfer_buffer = kmalloc(bufsz, GFP_KERNEL);
    u->durb[1].urb.transfer_buffer_length = bufsz;

    if (u->syncpipe) {
      if (u->surb[0].urb.transfer_buffer)
	kfree(u->surb[0].urb.transfer_buffer);
      u->surb[0].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[0].urb.transfer_buffer_length = 3 * SYNCFRAMES;
      if (u->surb[1].urb.transfer_buffer)
	kfree(u->surb[1].urb.transfer_buffer);
      u->surb[1].urb.transfer_buffer = kmalloc(3 * SYNCFRAMES, GFP_KERNEL);
      u->surb[1].urb.transfer_buffer_length = 3 * SYNCFRAMES;
    }

    if (!u->durb[0].urb.transfer_buffer || !u->durb[1].urb.transfer_buffer
	|| (u->syncpipe
	    && (!u->surb[0].urb.transfer_buffer
		|| !u->surb[1].urb.transfer_buffer))) {
      ERR("cannot start playback device %d",dev->devnum);
      return(0);
    }
    spin_lock_irqsave(&as->lock,flags);
  }

  if (TO_HOST == direction) {
    if (u->dma.count >= u->dma.dmasize && !u->dma.mapped) {
      spin_unlock_irqrestore(&as->lock,flags);
      return(0);
    }
  } else {
    if (u->dma.count <= 0 && !u->dma.mapped) {
      spin_unlock_irqrestore(&as->lock,flags);
      return(0);
    }
  }

  u->flags |= FLG_RUNNING;

  if (!(u->flags & FLG_URB0RUNNING)) {
    urb                    = &u->durb[0].urb;
    urb->dev               = dev;
    urb->pipe              = u->datapipe;
    urb->transfer_flags    = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context           = as;
    urb->complete = (TO_HOST == direction) ? usbin_completed : usbout_completed;
    if (!prepare_desc(direction,u,urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB0RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }

  if (u->flags & FLG_RUNNING && !(u->flags & FLG_URB1RUNNING)) {
    urb = &u->durb[1].urb;
    urb->dev               = dev;
    urb->pipe              = u->datapipe;
    urb->transfer_flags    = USB_ISO_ASAP;
    urb->number_of_packets = DESCFRAMES;
    urb->context           = as;
    urb->complete = (TO_HOST == direction) ? usbin_completed : usbout_completed;
    if (!prepare_desc(direction,u,urb) && !usb_submit_urb(urb))
      u->flags |= FLG_URB1RUNNING;
    else
      u->flags &= ~FLG_RUNNING;
  }

  if (u->syncpipe) {

    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC0RUNNING)) {
      urb = &u->surb[0].urb;
      urb->dev               = dev;
      urb->pipe              = u->syncpipe;
      urb->transfer_flags    = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context           = as;
      urb->complete = (TO_HOST == direction) ? 
	usbin_sync_completed : usbout_sync_completed;
      /* stride: u->syncinterval */
      if (!sync_prepare_desc(direction,u,urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC0RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }

    if (u->flags & FLG_RUNNING && !(u->flags & FLG_SYNC1RUNNING)) {
      urb = &u->surb[1].urb;
      urb->dev               = dev;
      urb->pipe              = u->syncpipe;
      urb->transfer_flags    = USB_ISO_ASAP;
      urb->number_of_packets = SYNCFRAMES;
      urb->context           = as;
      urb->complete = (TO_HOST == direction) ? 
	usbin_sync_completed : usbout_sync_completed;
      /* stride: u->syncinterval */
      if (!sync_prepare_desc(direction,u,urb) && !usb_submit_urb(urb))
	u->flags |= FLG_SYNC1RUNNING;
      else
	u->flags &= ~FLG_RUNNING;
    }
  }

  spin_unlock_irqrestore(&as->lock,flags);
  return(0);
}

/*----------------------------------------------------------------------------*/
static unsigned int
format_goodness(struct audioformat *afp, unsigned int fmt, unsigned int srate)
{
  unsigned int g = 0;

  if (srate < afp->sratelo)
    g += afp->sratelo - srate;
  if (srate > afp->sratehi)
    g += srate - afp->sratehi;
  if (AFMT_ISSTEREO(afp->format) && !AFMT_ISSTEREO(fmt))
    g += 0x100000;
  if (!AFMT_ISSTEREO(afp->format) && AFMT_ISSTEREO(fmt))
    g += 0x400000;
  if (AFMT_IS16BIT(afp->format) && !AFMT_IS16BIT(fmt))
    g += 0x100000;
  if (!AFMT_IS16BIT(afp->format) && AFMT_IS16BIT(fmt))
    g += 0x400000;
  return g;
}

static int
find_format(struct audioformat *afp, unsigned int nr, unsigned int fmt,
	    unsigned int srate)
{
  unsigned int i, g, gb = ~0;
  int j = -1;			/* default to failure */

  /* find "best" format (according to format_goodness) */
  for (i = 0; i < nr; i++) {
    g = format_goodness(&afp[i], fmt, srate);
    if (g >= gb)
      continue;
    j = i;
    gb = g;
  }
  return j;
}

static int enable_pitch_srate(struct usb_device *dev,
			      struct dmabuf *d,
			      struct audioformat *fmt,
			      unsigned int ep)
{
  unsigned char data[3];
  int err;

  /* If endpoint has pitch control, enable it. */
  if (fmt->attributes & 0x02) {
    data[0] = 1;
    err = usb_control_msg(dev, usb_sndctrlpipe(dev,0), SET_CUR,
			  USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			  PITCH_CONTROL << 8, ep, data, 1, HZ);
    if (err < 0) {
      err("cannot set pitch ctl device %d endpoint 0x%x to %u",
	  dev->devnum,ep,d->srate);
      return(-1);
    }
  }

  /* If endpoint has sampling rate control, set it. */
  if (fmt->attributes & 0x01) {
    data[0] = d->srate;
    data[1] = d->srate >> 8;
    data[2] = d->srate >> 16;
    err = usb_control_msg(dev, usb_sndctrlpipe(dev, 0), SET_CUR,
			  USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_OUT,
			  SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ);
    if (err < 0) {
      err("cannot set sampling frequency device %d endpoint 0x%x to %u",
	  dev->devnum,ep,d->srate);
      return(-1);
    }
    err = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0), GET_CUR,
			  USB_TYPE_CLASS | USB_RECIP_ENDPOINT | USB_DIR_IN,
			  SAMPLING_FREQ_CONTROL << 8, ep, data, 3, HZ);
    if (err < 0) {
      err("cannot get input sampling frequency device %d endpoint 0x%x\n",
	  dev->devnum,ep);
      return(-1);
    }
    d->srate = data[0] | (data[1] << 8) | (data[2] << 16);
  }
  return(0);
}

static int set_format_in(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usb_config_descriptor *config = dev->actconfig;
  struct usb_interface_descriptor *alts;
  struct usb_interface *iface;
  struct iostream* u = &as->io[TO_HOST];
  struct dmabuf *d = &u->dma;
  struct audioformat *fmt;
  unsigned int ep;
  int fmtnr;

  if (u->interface < 0 || u->interface >= config->bNumInterfaces)
    return 0;
  iface = &config->interface[u->interface];

  fmtnr = find_format(as->fmtin, as->numfmtin, d->format, d->srate);
  if (fmtnr < 0) {
    ERR("set_format_in(): failed to find desired format/speed combination.");
    return -1;
  }

  fmt = as->fmtin + fmtnr;
  alts = &iface->altsetting[fmt->altsetting];
  u->format = fmt->format;
  u->datapipe =
    usb_rcvisocpipe(dev, alts->endpoint[0].bEndpointAddress & 0xf);
  u->syncpipe = u->syncinterval = 0;
  if ((alts->endpoint[0].bmAttributes & 0x0c) == 0x08) {
    if (alts->bNumEndpoints < 2 ||
	alts->endpoint[1].bmAttributes != 0x01 ||
	alts->endpoint[1].bSynchAddress != 0 ||
	alts->endpoint[1].bEndpointAddress !=
	(alts->endpoint[0].bSynchAddress & 0x7f)) {
      ERR("device %d interface %d altsetting %d invalid synch pipe",
	  dev->devnum, u->interface, fmt->altsetting);
      return -1;
    }
    u->syncpipe = usb_sndisocpipe(dev,alts->endpoint[1].bEndpointAddress&0xf);
    u->syncinterval = alts->endpoint[1].bRefresh;
  }
  if (d->srate < fmt->sratelo) d->srate = fmt->sratelo;
  if (d->srate > fmt->sratehi) d->srate = fmt->sratehi;

  if (usb_set_interface(dev,alts->bInterfaceNumber,fmt->altsetting) < 0) {
    ERR("usb_set_interface failed, device %d interface %d altsetting %d\n",
	dev->devnum,u->interface,fmt->altsetting);
    return -1;
  }
  if (fmt->sratelo == fmt->sratehi) return 0;
  ep = usb_pipeendpoint(u->datapipe) | (u->datapipe & USB_DIR_IN);

  return(enable_pitch_srate(dev,d,fmt,ep));
}

static int set_format_out(struct audiodev *as)
{
  struct usb_device *dev = as->state->usbdev;
  struct usb_config_descriptor *config = dev->actconfig;
  struct usb_interface_descriptor *alts;
  struct usb_interface *iface;
  struct iostream *u = &as->io[TO_DEV];
  struct dmabuf *d = &u->dma;
  struct audioformat *fmt;
  unsigned int ep;
  int fmtnr;

  if (u->interface < 0 || u->interface >= config->bNumInterfaces)
    return 0;
  iface = &config->interface[u->interface];

  fmtnr = find_format(as->fmtout, as->numfmtout, d->format, d->srate);
  if (fmtnr < 0) {
    ERR("set_format_out(): failed to find desired format/speed combination.");
    return -1;
  }

  fmt = as->fmtout + fmtnr;
  alts = &iface->altsetting[fmt->altsetting];
  u->format = fmt->format;
  u->datapipe = usb_sndisocpipe(dev,alts->endpoint[0].bEndpointAddress&0xf);
  u->syncpipe = u->syncinterval = 0;
  if ((alts->endpoint[0].bmAttributes & 0x0c) == 0x04) {
    if (alts->bNumEndpoints < 2 ||
	alts->endpoint[1].bmAttributes != 0x01 ||
	alts->endpoint[1].bSynchAddress != 0 ||
	alts->endpoint[1].bEndpointAddress !=
	(alts->endpoint[0].bSynchAddress | 0x80)) {
      ERR("device %d interface %d altsetting %d invalid synch pipe",
	  dev->devnum,u->interface,fmt->altsetting);
      return -1;
    }
    u->syncpipe = usb_rcvisocpipe(dev,alts->endpoint[1].bEndpointAddress&0xf);
    u->syncinterval = alts->endpoint[1].bRefresh;
  }
  if (d->srate < fmt->sratelo) d->srate = fmt->sratelo;
  if (d->srate > fmt->sratehi) d->srate = fmt->sratehi;
  if (usb_set_interface(dev,u->interface,fmt->altsetting) < 0) {
    ERR("usb_set_interface failed, device %d interface %d altsetting %d",
	dev->devnum,u->interface,fmt->altsetting);
    return -1;
  }
  if (fmt->sratelo == fmt->sratehi) return 0;
  ep = usb_pipeendpoint(u->datapipe) | (u->datapipe & USB_DIR_IN);

  return(enable_pitch_srate(dev,d,fmt,ep));
}

static int set_format(struct audiodev *s,
		      unsigned int fmode,
		      unsigned int fmt, /* OSS format plus stereo */
		      unsigned int srate)
{
  int ret1 = 0, ret2 = 0;

  if (!(fmode & (FMODE_READ | FMODE_WRITE)))
    return -EINVAL;

  if (fmode & FMODE_READ) {
    stop_stream(TO_HOST,s);
    s->io[TO_HOST].dma.ready = 0;
    if (fmt == AFMT_QUERY)
      fmt = s->io[TO_HOST].dma.format;
    else
      s->io[TO_HOST].dma.format = fmt;
    if (!srate)
      srate = s->io[TO_HOST].dma.srate;
    else
      s->io[TO_HOST].dma.srate = srate;
  }

  if (fmode & FMODE_WRITE) {
    stop_stream(TO_DEV,s);
    s->io[TO_DEV].dma.ready = 0;
    if (fmt == AFMT_QUERY)
      fmt = s->io[TO_DEV].dma.format;
    else
      s->io[TO_DEV].dma.format = fmt;
    if (!srate)
      srate = s->io[TO_DEV].dma.srate;
    else
      s->io[TO_DEV].dma.srate = srate;
  }

  if (fmode & FMODE_READ)
    ret1 = set_format_in(s);

  if (fmode & FMODE_WRITE)
    ret2 = set_format_out(s);

  return ret1 ? ret1 : ret2;
}

/*----------------------------------------------------------------------------*/
/*
 * should be called with open_sem hold, so that no new processes
 * look at the audio device to be destroyed
 */
void release(struct extigy *s)
{
  struct audiodev *as;
  struct mixerdev *ms;

  s->count--;
  if (s->count) {
    up(&open_sem);
    return;
  }
  up(&open_sem);
  wake_up(&open_wait);
  while (!list_empty(&s->audiolist)) {
    as = list_entry(s->audiolist.next, struct audiodev, list);
    list_del(&as->list);
    stop_stream(TO_HOST,as);
    stop_stream(TO_DEV,as);
    dmabuf_release(&as->io[TO_HOST].dma);
    dmabuf_release(&as->io[TO_DEV].dma);
    kfree(as);
  }
  while (!list_empty(&s->mixerlist)) {
    ms = list_entry(s->mixerlist.next, struct mixerdev, list);
    list_del(&ms->list);
    kfree(ms);
  }
  kfree(s);
}

/*----------------------------------------------------------------------------*/
extern inline int prog_dmabuf_in(struct audiodev *as)
{
  stop_stream(TO_HOST,as);
  return dmabuf_init(&as->io[TO_HOST].dma);
}

extern inline int prog_dmabuf_out(struct audiodev *as)
{
  stop_stream(TO_DEV,as);
  return dmabuf_init(&as->io[TO_DEV].dma);
}

/*----------------------------------------------------------------------------*/
static loff_t exaudio_llseek(struct file *file, loff_t offset, int origin)
{
  return -ESPIPE;
}

/*----------------------------------------------------------------------------*/
static int drain_out(struct audiodev *as, int nonblock)
{
  DECLARE_WAITQUEUE(wait, current);
  unsigned long flags;
  int count, tmo;

  if (as->io[TO_DEV].dma.mapped || !as->io[TO_DEV].dma.ready)
    return 0;
  start_stream(TO_DEV,as);
  add_wait_queue(&as->io[TO_DEV].dma.wait, &wait);
  for (;;) {
    __set_current_state(TASK_INTERRUPTIBLE);
    spin_lock_irqsave(&as->lock, flags);
    count = as->io[TO_DEV].dma.count;
    spin_unlock_irqrestore(&as->lock, flags);
    if (count <= 0)
      break;
    if (signal_pending(current))
      break;
    if (nonblock) {
      remove_wait_queue(&as->io[TO_DEV].dma.wait, &wait);
      set_current_state(TASK_RUNNING);
      return -EBUSY;
    }
    tmo = 3 * HZ * count / as->io[TO_DEV].dma.srate;
    tmo >>= AFMT_BYTESSHIFT(as->io[TO_DEV].dma.format);
    if (!schedule_timeout(tmo + 1)) {
      printk(KERN_DEBUG "exaudio: dma timed out??\n");
      break;
    }
  }
  remove_wait_queue(&as->io[TO_DEV].dma.wait, &wait);
  set_current_state(TASK_RUNNING);
  if (signal_pending(current))
    return -ERESTARTSYS;
  return 0;
}

/*----------------------------------------------------------------------------*/
static ssize_t exaudio_read(struct file *file,char *buffer,
			    size_t count,loff_t *ppos)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  DECLARE_WAITQUEUE(wait, current);
  ssize_t ret = 0;
  unsigned long flags;
  unsigned int ptr;
  int cnt, err;

  if (ppos != &file->f_pos)
    return -ESPIPE;
  if (as->io[TO_HOST].dma.mapped)
    return -ENXIO;
  if (!as->io[TO_HOST].dma.ready && (ret = prog_dmabuf_in(as)))
    return ret;
  if (!access_ok(VERIFY_WRITE, buffer, count))
    return -EFAULT;
  add_wait_queue(&as->io[TO_HOST].dma.wait, &wait);
  while (count > 0) {
    spin_lock_irqsave(&as->lock, flags);
    ptr = as->io[TO_HOST].dma.rdptr;
    cnt = as->io[TO_HOST].dma.count;
    /* set task state early to avoid wakeup races */
    if (cnt <= 0)
      __set_current_state(TASK_INTERRUPTIBLE);
    spin_unlock_irqrestore(&as->lock, flags);
    if (cnt > count)
      cnt = count;
    if (cnt <= 0) {
      if (start_stream(TO_HOST,as)) {
	if (!ret)
	  ret = -ENODEV;
	break;
      }
      if (file->f_flags & O_NONBLOCK) {
	if (!ret)
	  ret = -EAGAIN;
	break;
      }
      schedule();
      if (signal_pending(current)) {
	if (!ret)
	  ret = -ERESTARTSYS;
	break;
      }
      continue;
    }
    if ((err = dmabuf_copyout_user(&as->io[TO_HOST].dma, ptr, buffer, cnt))) {
      if (!ret)
	ret = err;
      break;
    }
    ptr += cnt;
    if (ptr >= as->io[TO_HOST].dma.dmasize)
      ptr -= as->io[TO_HOST].dma.dmasize;
    spin_lock_irqsave(&as->lock, flags);
    as->io[TO_HOST].dma.rdptr = ptr;
    as->io[TO_HOST].dma.count -= cnt;
    spin_unlock_irqrestore(&as->lock, flags);
    count -= cnt;
    buffer += cnt;
    ret += cnt;
  }
  __set_current_state(TASK_RUNNING);
  remove_wait_queue(&as->io[TO_HOST].dma.wait, &wait);
  return ret;
}

static ssize_t exaudio_write(struct file *file,const char *buffer,
			     size_t count,loff_t *ppos)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  DECLARE_WAITQUEUE(wait, current);
  ssize_t ret = 0;
  unsigned long flags;
  unsigned int ptr;
  unsigned int start_thr;
  int cnt, err;

  if (ppos != &file->f_pos)
    return -ESPIPE;
  if (as->io[TO_DEV].dma.mapped)
    return -ENXIO;
  if (!as->io[TO_DEV].dma.ready && (ret = prog_dmabuf_out(as)))
    return ret;
  if (!access_ok(VERIFY_READ, buffer, count))
    return -EFAULT;

  start_thr =
    (as->io[TO_DEV].dma.srate << AFMT_BYTESSHIFT(as->io[TO_DEV].dma.format)) /
    (1000 / (3 * DESCFRAMES));

  add_wait_queue(&as->io[TO_DEV].dma.wait, &wait);

  while (count > 0) {
    spin_lock_irqsave(&as->lock, flags);
    if (as->io[TO_DEV].dma.count < 0) {
      as->io[TO_DEV].dma.count = 0;
      as->io[TO_DEV].dma.rdptr = as->io[TO_DEV].dma.wrptr;
    }
    ptr = as->io[TO_DEV].dma.wrptr;
    cnt = as->io[TO_DEV].dma.dmasize - as->io[TO_DEV].dma.count;
    /* set task state early to avoid wakeup races */
    if (cnt <= 0) __set_current_state(TASK_INTERRUPTIBLE);
    spin_unlock_irqrestore(&as->lock,flags);
    if (cnt > count) cnt = count;
    if (cnt <= 0) {
      if (start_stream(TO_DEV,as)) {
	if (!ret) ret = -ENODEV;
	break;
      }
      if (file->f_flags & O_NONBLOCK) {
	if (!ret) ret = -EAGAIN;
	break;
      }
      schedule();
      if (signal_pending(current)) {
	if (!ret) ret = -ERESTARTSYS;
	break;
      }
      continue;
    }
    if ((err = dmabuf_copyin_user(&as->io[TO_DEV].dma, ptr, buffer, cnt))) {
      if (!ret) ret = err;
      break;
    }
    ptr += cnt;
    if (ptr >= as->io[TO_DEV].dma.dmasize) ptr -= as->io[TO_DEV].dma.dmasize;
    spin_lock_irqsave(&as->lock,flags);
    as->io[TO_DEV].dma.wrptr = ptr;
    as->io[TO_DEV].dma.count += cnt;
    spin_unlock_irqrestore(&as->lock,flags);
    count -= cnt;
    buffer += cnt;
    ret += cnt;
    if (as->io[TO_DEV].dma.count >= start_thr && start_stream(TO_DEV,as)) {
      if (!ret) ret = -ENODEV;
      break;
    }
  }
  __set_current_state(TASK_RUNNING);
  remove_wait_queue(&as->io[TO_DEV].dma.wait, &wait);
  return ret;
}

/* Called without the kernel lock - fine */
static unsigned int exaudio_poll(struct file *file,
				 struct poll_table_struct *wait)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  unsigned long flags;
  unsigned int mask = 0;

  if (file->f_mode & FMODE_WRITE) {
    if (!as->io[TO_DEV].dma.ready)
      prog_dmabuf_out(as);
    poll_wait(file, &as->io[TO_DEV].dma.wait, wait);
  }
  if (file->f_mode & FMODE_READ) {
    if (!as->io[TO_HOST].dma.ready)
      prog_dmabuf_in(as);
    poll_wait(file, &as->io[TO_HOST].dma.wait, wait);
  }
  spin_lock_irqsave(&as->lock, flags);
  if (file->f_mode & FMODE_READ) {
    if (as->io[TO_HOST].dma.count >= (signed) as->io[TO_HOST].dma.fragsize)
      mask |= POLLIN | POLLRDNORM;
  }
  if (file->f_mode & FMODE_WRITE) {
    if (as->io[TO_DEV].dma.mapped) {
      if (as->io[TO_DEV].dma.count >= (signed) as->io[TO_DEV].dma.fragsize)
	mask |= POLLOUT | POLLWRNORM;
    } else {
      if ((signed) as->io[TO_DEV].dma.dmasize >=
	  as->io[TO_DEV].dma.count + (signed) as->io[TO_DEV].dma.fragsize)
	mask |= POLLOUT | POLLWRNORM;
    }
  }
  spin_unlock_irqrestore(&as->lock, flags);
  return mask;
}

static int exaudio_mmap(struct file *file, struct vm_area_struct *vma)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  struct dmabuf *db;
  int ret = -EINVAL;

  lock_kernel();
  if (vma->vm_flags & VM_WRITE) {
    if ((ret = prog_dmabuf_out(as)) != 0)
      goto out;
    db = &as->io[TO_DEV].dma;
  } else if (vma->vm_flags & VM_READ) {
    if ((ret = prog_dmabuf_in(as)) != 0)
      goto out;
    db = &as->io[TO_HOST].dma;
  } else
    goto out;

  ret = -EINVAL;
  if (vma->vm_pgoff != 0) goto out;

  ret = dmabuf_mmap(db,vma);

out:
  unlock_kernel();
  return ret;
}

static int exaudio_ioctl(struct inode *inode,struct file *file,
			 unsigned int cmd,unsigned long arg)
{
  struct audiodev *as = (struct audiodev *)file->private_data;
  struct extigy *s = as->state;
  unsigned long flags;
  audio_buf_info abinfo;
  count_info cinfo;
  int val, val2, mapped=0, ret;

  if (!s->usbdev) return -EIO;

  if (((file->f_mode&FMODE_WRITE) && as->io[TO_DEV].dma.mapped) ||
      ((file->f_mode&FMODE_READ)  && as->io[TO_HOST].dma.mapped))
    mapped = 1;

  switch (cmd) {
  case OSS_GETVERSION:
    return put_user(SOUND_VERSION,(int*)arg);

  case SNDCTL_DSP_SYNC:
    if (file->f_mode & FMODE_WRITE)
      return drain_out(as,0/*file->f_flags & O_NONBLOCK */);
    return 0;

  case SNDCTL_DSP_SETDUPLEX:
    return 0;

  case SNDCTL_DSP_GETCAPS:
    return put_user(DSP_CAP_DUPLEX|DSP_CAP_REALTIME|DSP_CAP_TRIGGER|
		    DSP_CAP_MMAP|DSP_CAP_BATCH,(int*)arg);
    
  case SNDCTL_DSP_RESET:
    if (file->f_mode & FMODE_WRITE) {
      stop_stream(TO_DEV,as);
      as->io[TO_DEV].dma.rdptr = as->io[TO_DEV].dma.wrptr =
	as->io[TO_DEV].dma.count = as->io[TO_DEV].dma.total_bytes = 0;
    }
    if (file->f_mode & FMODE_READ) {
      stop_stream(TO_HOST,as);
      as->io[TO_HOST].dma.rdptr = as->io[TO_HOST].dma.wrptr =
	as->io[TO_HOST].dma.count = as->io[TO_HOST].dma.total_bytes = 0;
    }
    return 0;

  case SNDCTL_DSP_SPEED:
    if (get_user(val,(int*)arg)) return -EFAULT;
    if (val >= 0) {
      if (val < 4000) val = 4000;
      if (val > 100000) val = 100000;
      if (set_format(as,file->f_mode,AFMT_QUERY,val)) return -EIO;
    }
    return put_user((file->f_mode & FMODE_READ) ? 
		    as->io[TO_HOST].dma.srate :
		    as->io[TO_DEV].dma.srate, (int*)arg);

  case SNDCTL_DSP_STEREO:
    val2 = (file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.format :
      as->io[TO_DEV].dma.format;
    if (set_format(as,file->f_mode,val2|AFMT_STEREO,0)) return -EIO;
    return 0;

  case SNDCTL_DSP_CHANNELS:
    if (get_user(val,(int*)arg)) return -EFAULT;
    if (val != 0) {
      val2 = (file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.format : 
	as->io[TO_DEV].dma.format;
      if (val == 1) val2 &= ~AFMT_STEREO;
      else val2 |= AFMT_STEREO;
      if (set_format(as,file->f_mode,val2,0))
	return -EIO;
    }
    val2 = (file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.format :
      as->io[TO_DEV].dma.format;
    return put_user(AFMT_ISSTEREO(val2) ? 2 : 1, (int*)arg);

  case SNDCTL_DSP_GETFMTS: /* Returns a mask */
    return put_user(AFMT_S16_LE|AFMT_AC3,(int*)arg);

  case SNDCTL_DSP_SETFMT: /* Selects ONE fmt */
    if (get_user(val,(int*)arg)) return -EFAULT;

    if (0 == as->numfmtin) { /* its the AC3 */

      if (val == AFMT_QUERY)
	return put_user(AFMT_AC3,(int*)arg);

      if (set_format(as,file->f_mode,AFMT_S16_LE|AFMT_STEREO,0)) {
	err("cannot set_format()");
	return -EIO;
      }

      return put_user(val/*AFMT_AC3*/,(int*)arg);
    }

    if (val != AFMT_QUERY) {
      if (hweight32(val) != 1) {
	err("hweight");
	return -EINVAL;
      }
      if (!(val & (AFMT_S16_LE))) {
	return -EINVAL;
      }
      val2 = (file->f_mode & FMODE_READ) ? 
	as->io[TO_HOST].dma.format : as->io[TO_DEV].dma.format;
      val |= val2 & AFMT_STEREO;
      if (set_format(as,file->f_mode,val,0)) {
	err("cannot set_format()");
	return -EIO;
      }
    }

    val2 = (file->f_mode & FMODE_READ) ?
      as->io[TO_HOST].dma.format : as->io[TO_DEV].dma.format;

    return put_user(val2 & ~AFMT_STEREO,(int*)arg);

  case SNDCTL_DSP_POST:
    return 0;

  case SNDCTL_DSP_GETTRIGGER:
    val = 0;
    if (file->f_mode & FMODE_READ && as->io[TO_HOST].flags & FLG_RUNNING)
      val |= PCM_ENABLE_INPUT;
    if (file->f_mode & FMODE_WRITE && as->io[TO_DEV].flags & FLG_RUNNING)
      val |= PCM_ENABLE_OUTPUT;
    return put_user(val,(int*)arg);

  case SNDCTL_DSP_SETTRIGGER:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (file->f_mode & FMODE_READ) {
      if (val & PCM_ENABLE_INPUT) {
	if (!as->io[TO_HOST].dma.ready && (ret = prog_dmabuf_in(as)))
	  return ret;
	if (start_stream(TO_HOST,as))
	  return -ENODEV;
      } else
	stop_stream(TO_HOST,as);
    }
    if (file->f_mode & FMODE_WRITE) {
      if (val & PCM_ENABLE_OUTPUT) {
	if (!as->io[TO_DEV].dma.ready && (ret = prog_dmabuf_out(as)))
	  return ret;
	if (start_stream(TO_DEV,as))
	  return -ENODEV;
      } else
	stop_stream(TO_DEV,as);
    }
    return 0;

  case SNDCTL_DSP_GETOSPACE:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    if (!(as->io[TO_DEV].flags & FLG_RUNNING)
	&& (val = prog_dmabuf_out(as)) != 0)
      return val;
    spin_lock_irqsave(&as->lock, flags);
    abinfo.fragsize = as->io[TO_DEV].dma.fragsize;
    abinfo.bytes = as->io[TO_DEV].dma.dmasize - as->io[TO_DEV].dma.count;
    abinfo.fragstotal = as->io[TO_DEV].dma.numfrag;
    abinfo.fragments = abinfo.bytes >> as->io[TO_DEV].dma.fragshift;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

  case SNDCTL_DSP_GETISPACE:
    if (!(file->f_mode & FMODE_READ))
      return -EINVAL;
    if (!(as->io[TO_HOST].flags & FLG_RUNNING)
	&& (val = prog_dmabuf_in(as)) != 0)
      return val;
    spin_lock_irqsave(&as->lock, flags);
    abinfo.fragsize = as->io[TO_HOST].dma.fragsize;
    abinfo.bytes = as->io[TO_HOST].dma.count;
    abinfo.fragstotal = as->io[TO_HOST].dma.numfrag;
    abinfo.fragments = abinfo.bytes >> as->io[TO_HOST].dma.fragshift;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

  case SNDCTL_DSP_NONBLOCK:
    file->f_flags |= O_NONBLOCK;
    return 0;

  case SNDCTL_DSP_GETODELAY:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    val = as->io[TO_DEV].dma.count;
    spin_unlock_irqrestore(&as->lock, flags);
    return put_user(val, (int *) arg);

  case SNDCTL_DSP_GETIPTR:
    if (!(file->f_mode & FMODE_READ))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    cinfo.bytes = as->io[TO_HOST].dma.total_bytes;
    cinfo.blocks = as->io[TO_HOST].dma.count >> as->io[TO_HOST].dma.fragshift;
    cinfo.ptr = as->io[TO_HOST].dma.wrptr;
    if (as->io[TO_HOST].dma.mapped)
      as->io[TO_HOST].dma.count &= as->io[TO_HOST].dma.fragsize - 1;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &cinfo, sizeof(cinfo));

  case SNDCTL_DSP_GETOPTR:
    if (!(file->f_mode & FMODE_WRITE))
      return -EINVAL;
    spin_lock_irqsave(&as->lock, flags);
    cinfo.bytes = as->io[TO_DEV].dma.total_bytes;
    cinfo.blocks = as->io[TO_DEV].dma.count >> as->io[TO_DEV].dma.fragshift;
    cinfo.ptr = as->io[TO_DEV].dma.rdptr;
    if (as->io[TO_DEV].dma.mapped)
      as->io[TO_DEV].dma.count &= as->io[TO_DEV].dma.fragsize - 1;
    spin_unlock_irqrestore(&as->lock, flags);
    return copy_to_user((void *) arg, &cinfo, sizeof(cinfo));

  case SNDCTL_DSP_GETBLKSIZE:
    if (file->f_mode & FMODE_WRITE) {
      if ((val = prog_dmabuf_out(as)))
	return val;
      return put_user(as->io[TO_DEV].dma.fragsize, (int *) arg);
    }
    if ((val = prog_dmabuf_in(as)))
      return val;
    return put_user(as->io[TO_HOST].dma.fragsize, (int *) arg);

  case SNDCTL_DSP_SETFRAGMENT:
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (file->f_mode & FMODE_READ) {
      as->io[TO_HOST].dma.ossfragshift = val & 0xffff;
      as->io[TO_HOST].dma.ossmaxfrags = (val >> 16) & 0xffff;
      if (as->io[TO_HOST].dma.ossfragshift < 4)
	as->io[TO_HOST].dma.ossfragshift = 4;
      if (as->io[TO_HOST].dma.ossfragshift > 15)
	as->io[TO_HOST].dma.ossfragshift = 15;
      if (as->io[TO_HOST].dma.ossmaxfrags < 4)
	as->io[TO_HOST].dma.ossmaxfrags = 4;
    }
    if (file->f_mode & FMODE_WRITE) {
      as->io[TO_DEV].dma.ossfragshift = val & 0xffff;
      as->io[TO_DEV].dma.ossmaxfrags = (val >> 16) & 0xffff;
      if (as->io[TO_DEV].dma.ossfragshift < 4)
	as->io[TO_DEV].dma.ossfragshift = 4;
      if (as->io[TO_DEV].dma.ossfragshift > 15)
	as->io[TO_DEV].dma.ossfragshift = 15;
      if (as->io[TO_DEV].dma.ossmaxfrags < 4)
	as->io[TO_DEV].dma.ossmaxfrags = 4;
    }
    return 0;

  case SNDCTL_DSP_SUBDIVIDE:
    if ((file->f_mode & FMODE_READ && as->io[TO_HOST].dma.subdivision) ||
	(file->f_mode & FMODE_WRITE && as->io[TO_DEV].dma.subdivision))
      return -EINVAL;
    if (get_user(val, (int *) arg))
      return -EFAULT;
    if (val != 1 && val != 2 && val != 4)
      return -EINVAL;
    if (file->f_mode & FMODE_READ)
      as->io[TO_HOST].dma.subdivision = val;
    if (file->f_mode & FMODE_WRITE)
      as->io[TO_DEV].dma.subdivision = val;
    return 0;

  case SOUND_PCM_READ_RATE:
    return put_user((file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.
		    srate : as->io[TO_DEV].dma.srate, (int*)arg);

  case SOUND_PCM_READ_CHANNELS:
    val2 = (file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.format :
      as->io[TO_DEV].dma.format;
    return put_user(AFMT_ISSTEREO(val2) ? 2 : 1, (int*)arg);

  case SOUND_PCM_READ_BITS:
    val2 = (file->f_mode & FMODE_READ) ? as->io[TO_HOST].dma.format :
      as->io[TO_DEV].dma.format;
    return put_user(AFMT_IS16BIT(val2) ? 16 : 8, (int*)arg);

  case SOUND_PCM_WRITE_FILTER:
  case SNDCTL_DSP_SETSYNCRO:
  case SOUND_PCM_READ_FILTER:
    return -EINVAL;
  }
  return -ENOIOCTLCMD;
}

static int exaudio_open(struct inode *inode,struct file *file)
{
  int i, minor = MINOR(inode->i_rdev);
  DECLARE_WAITQUEUE(wait, current);
  struct list_head *devs, *adev;
  struct audiodev *as;
  struct extigy *exy;

  for (;;) {
    down(&open_sem);
    for (devs = alldevices.next; devs != &alldevices; devs = devs->next) {
      exy = list_entry(devs,struct extigy,devices);
      adev = exy->audiolist.next;
      as = list_entry(adev,struct audiodev,list);
      for (i = 0; i < N_MINORS; i++) {
	if ( !((exy->ossdev[i] ^ minor) & ~0xf) )
	  goto device_found;
      }
    }
    /* Requested device not found. */
    up(&open_sem);
    return(-ENODEV);
  device_found:
    if (!exy->usbdev) {
      up(&open_sem);
      return(-EIO);
    }

    /* If third dsp requested, use AC3 audiodev instead */
    if (2 == i) {
      adev = adev->next;
      as = list_entry(adev,struct audiodev,list);
    }

    /* Check which directions are unused. */
    if (exy->modes_in_use & file->f_mode) {
      ;      /* Sorry, go onto the wait queue. */
    } else {
      break; /* Streams are free to open. */
    }
    if (file->f_flags & O_NONBLOCK) {
      up(&open_sem);
      return(-EBUSY);
    }
    __set_current_state(TASK_INTERRUPTIBLE);
    add_wait_queue(&open_wait,&wait);
    up(&open_sem);
    schedule();
    __set_current_state(TASK_RUNNING);
    remove_wait_queue(&open_wait,&wait);
    if (signal_pending(current))
      return(-ERESTARTSYS);
  }

  if (file->f_mode & FMODE_READ)
    as->io[TO_HOST].dma.ossfragshift = as->io[TO_HOST].dma.ossmaxfrags =
      as->io[TO_HOST].dma.subdivision = 0;

  if (file->f_mode & FMODE_WRITE)
    as->io[TO_DEV].dma.ossfragshift = as->io[TO_DEV].dma.ossmaxfrags =
      as->io[TO_DEV].dma.subdivision = 0;

  if (set_format(as,file->f_mode,AFMT_S16_LE|AFMT_STEREO,48000)) {
    up(&open_sem);
    return(-EIO);
  }

  file->private_data = as;
  exy->modes_in_use |= (file->f_mode & (FMODE_READ|FMODE_WRITE));
  exy->count++;
  up(&open_sem);
  return 0;
}

static int exaudio_release(struct inode *inode, struct file *file)
{
  struct audiodev *as = (struct audiodev *) file->private_data;
  struct extigy *exy;
  struct usb_device *dev;
  struct usb_interface *iface;

  lock_kernel();
  exy = as->state;
  dev = exy->usbdev;
  if (file->f_mode & FMODE_WRITE)
    drain_out(as, file->f_flags & O_NONBLOCK);
  down(&open_sem);
  if (file->f_mode & FMODE_WRITE) {
    stop_stream(TO_DEV,as);
    if (dev && as->io[TO_DEV].interface >= 0) {
      iface = &dev->actconfig->interface[as->io[TO_DEV].interface];
      usb_set_interface(dev, iface->altsetting->bInterfaceNumber, 0);
    }
    dmabuf_release(&as->io[TO_DEV].dma);
    stop_stream(TO_DEV,as);
  }
  if (file->f_mode & FMODE_READ) {
    stop_stream(TO_HOST,as);
    if (dev && as->io[TO_HOST].interface >= 0) {
      iface = &dev->actconfig->interface[as->io[TO_HOST].interface];
      usb_set_interface(dev, iface->altsetting->bInterfaceNumber, 0);
    }
    dmabuf_release(&as->io[TO_HOST].dma);
    stop_stream(TO_HOST,as);
  }
  exy->modes_in_use &= (~file->f_mode) & (FMODE_READ|FMODE_WRITE);
  release(exy);
  wake_up(&open_wait);
  unlock_kernel();
  return 0;
}

static struct file_operations exaudio_fops = {
  owner:   THIS_MODULE,
  llseek:  exaudio_llseek,
  read:    exaudio_read,
  write:   exaudio_write,
  poll:    exaudio_poll,
  ioctl:   exaudio_ioctl,
  mmap:    exaudio_mmap,
  open:    exaudio_open,
  release: exaudio_release,
};

static struct usb_device_id exaudio_ids[] = {
  { USB_DEVICE(0x041E,0x3000),
    bInterfaceClass:    USB_CLASS_AUDIO,
    bInterfaceSubClass: 1
  },
  {} /* Terminating entry */
};

MODULE_DEVICE_TABLE(usb,exaudio_ids);

static struct usb_driver exaudio_driver = {
  name:        "exaudio",
  probe:       exaudio_probe,
  disconnect:  exaudio_disconnect,
  driver_list: LIST_HEAD_INIT(exaudio_driver.driver_list),
  id_table:    exaudio_ids,
};

static void init_audiodev(struct extigy *exy,struct audiodev *as,
			  int asifin,int asifout)
{
  init_waitqueue_head(&as->io[TO_HOST].dma.wait);
  init_waitqueue_head(&as->io[TO_DEV].dma.wait);
  spin_lock_init(&as->lock);
  spin_lock_init(&as->io[TO_HOST].durb[0].urb.lock);
  spin_lock_init(&as->io[TO_HOST].durb[1].urb.lock);
  spin_lock_init(&as->io[TO_HOST].surb[0].urb.lock);
  spin_lock_init(&as->io[TO_HOST].surb[1].urb.lock);
  spin_lock_init(&as->io[TO_DEV].durb[0].urb.lock);
  spin_lock_init(&as->io[TO_DEV].durb[1].urb.lock);
  spin_lock_init(&as->io[TO_DEV].surb[0].urb.lock);
  spin_lock_init(&as->io[TO_DEV].surb[1].urb.lock);
  as->state = exy;
  as->io[TO_HOST].interface = asifin;
  as->io[TO_DEV].interface = asifout;
}

#define PLAYBACK_IFNUM  1
#define RECORDING_IFNUM 2
#define AC3_IFNUM       3
#define MIDI_IFNUM      4

#ifdef ORIG_FIRMWARE
#define TOO_SHORT 794
#define BIGDESC   1701
#else
#define TOO_SHORT 483
#define BIGDESC   1709
#endif

static void *exaudio_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
  int err;
  struct audiodev *adev, *ac3;
  struct extigy *exy;
  struct usb_config_descriptor *config = dev->actconfig;
  unsigned int i;
#define N_RECORDING_FORMATS 5
  unsigned int rec_rates[N_RECORDING_FORMATS] = {
    8000, 16000, 24000, 32000, 48000
  };
  
  if (0x041E != dev->descriptor.idVendor ||
      0x3000 != dev->descriptor.idProduct) {
    err("not an extigy");
    return(NULL);
  }
  if (MIDI_IFNUM == ifnum) {
    INFO("for the MIDI interface, use the usb-midi driver");
  }
  if (ifnum != 0) {
    return(NULL);
  }
  if (usb_set_configuration(dev,config->bConfigurationValue) < 0) {
    err("set_configuration failed (ConfigValue 0x%x)",
	config->bConfigurationValue);
    return(NULL);
  }
  switch (config->wTotalLength) {
  case TOO_SHORT: /* Send message to force a reconnect with full interface. */
    err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),
			  0x10,0x43,0x0001,0x000a,NULL,0,HZ);
    if (err < 0) err("Error sending vendor msg: %d",err);
    err = usb_get_device_descriptor(dev);
    if (err < 0) {
      /* We expect this to fail. */
    }
    if (usb_set_configuration(dev,config->bConfigurationValue) < 0) {
      /* We expect this to fail. */
    }
    return(NULL);

  case BIGDESC: /* That's better, a full set of descriptors! */
    break;

  default:
    ERR("This extigy's layout is not known to me (firmware mismatch?)");
    ERR("wTotalLength=%u bNumInterfaces=%u",
	config->wTotalLength,config->bNumInterfaces);
    return(NULL);
  }

  exy = kmalloc(sizeof(struct extigy),GFP_KERNEL);
  if (!exy) return(NULL);
  memset(exy,0,sizeof(struct extigy));
  INIT_LIST_HEAD(&exy->audiolist);
  INIT_LIST_HEAD(&exy->mixerlist);
  exy->usbdev = dev;
  exy->count = 1;
  
  usb_driver_claim_interface(&exaudio_driver,
			     &config->interface[PLAYBACK_IFNUM],(void*)-1);
  usb_driver_claim_interface(&exaudio_driver,
			     &config->interface[RECORDING_IFNUM],(void*)-1);
  usb_driver_claim_interface(&exaudio_driver,
			     &config->interface[AC3_IFNUM],(void*)-1);

  /* Create the PCM playback/recording pair. */
  adev = kmalloc(sizeof(struct audiodev),GFP_KERNEL);
  if (!adev) {
    kfree(exy);
    return(NULL);
  }
  memset(adev,0,sizeof(struct audiodev));
  init_audiodev(exy,adev,RECORDING_IFNUM,PLAYBACK_IFNUM);
  adev->numfmtin = N_RECORDING_FORMATS;
  for (i = 0; i < N_RECORDING_FORMATS; i++) {
    adev->fmtin[i].format     = AFMT_STEREO | AFMT_S16_LE;
    adev->fmtin[i].sratelo    = rec_rates[i];
    adev->fmtin[i].sratehi    = rec_rates[i];
    adev->fmtin[i].altsetting = 1+i;
    adev->fmtin[i].attributes = 0;
  }
  adev->io[TO_HOST].flags = FLG_CONNECTED;
  adev->numfmtout = 1;
  adev->fmtout[0].format     = AFMT_STEREO | AFMT_S16_LE;
  adev->fmtout[0].sratelo    = 48000;
  adev->fmtout[0].sratehi    = 48000;
  adev->fmtout[0].altsetting = 1;
  adev->fmtout[0].attributes = 0;
  adev->io[TO_DEV].flags = FLG_CONNECTED;
  list_add_tail(&adev->list,&exy->audiolist);

  /* Create the AC3 playback/null pair. */
  ac3 = kmalloc(sizeof(struct audiodev),GFP_KERNEL);
  if (!ac3) {
    kfree(exy);
    return(NULL);
  }
  memset(ac3,0,sizeof(struct audiodev));
  init_audiodev(exy,ac3,-1,AC3_IFNUM);
  ac3->numfmtout = 1;
  ac3->fmtout[0].format     = AFMT_STEREO | AFMT_S16_LE;
  ac3->fmtout[0].sratelo    = 48000;
  ac3->fmtout[0].sratehi    = 48000;
  ac3->fmtout[0].altsetting = 1;
  ac3->fmtout[0].attributes = 0;
  ac3->io[TO_DEV].flags = FLG_CONNECTED;
  list_add_tail(&ac3->list,&exy->audiolist);

  exaudio_init_mixers(exy,ifnum);
  
  down(&open_sem);
  for (i = 0; i < N_MINORS; i++) {
    exy->ossdev[i] = register_sound_dsp(&exaudio_fops,-1);
    if (exy->ossdev[i] < 0) {
      err("cannot register dsp devices");
      kfree(adev);
      kfree(exy);
      return(NULL);
    }
    INFO("registered dsp (major 14 minor %d)",exy->ossdev[i]);
  }
  list_add_tail(&exy->devices,&alldevices);
  up(&open_sem);
  return(exy);
}

static void exaudio_disconnect(struct usb_device *dev,void *ptr)
{
  int i;
  struct audiodev *as;
  struct extigy *exy = (struct extigy*)ptr;
  struct list_head *list;
  struct mixerdev *ms;

  /* we get called with -1 for every audiostreaming interface registered */
  if (exy == (struct extigy*) -1) {
    dbg("exaudio_disconnect: called with -1");
    return;
  }
  if (!exy->usbdev) {
    dbg("exaudio_disconnect: already called for %p!",s);
    return;
  }
  down(&open_sem);
  list_del(&exy->devices);
  INIT_LIST_HEAD(&exy->devices);
  exy->usbdev = NULL;

  for (i = 0; i < N_MINORS; i++) {
    if (exy->ossdev[i] >= 0) unregister_sound_dsp(exy->ossdev[i]);
    exy->ossdev[i] = -1;
  }

  /* Deregister all audio and mixer devices,
   * so no new processes can open this device. */
  for (list = exy->audiolist.next; list != &exy->audiolist; list = list->next) {
    as = list_entry(list,struct audiodev,list);
    disc_stream(TO_HOST,as);
    disc_stream(TO_DEV,as);
    wake_up(&as->io[TO_HOST].dma.wait);
    wake_up(&as->io[TO_DEV].dma.wait);
  }
  for (list = exy->mixerlist.next; list != &exy->mixerlist; list = list->next) {
    ms = list_entry(list,struct mixerdev,list);
    if (ms->ossmixer >= 0)
      unregister_sound_mixer(ms->ossmixer);
    ms->ossmixer = -1;
  }
  release(exy);
  wake_up(&open_wait);
}

static int __init exaudio_init(void)
{
  usb_register(&exaudio_driver);
  INFO(DRIVER_VERSION);
  return(0);
}

static void __exit exaudio_cleanup(void)
{
  usb_deregister(&exaudio_driver);
}

module_init(exaudio_init);
module_exit(exaudio_cleanup);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
