/* $Id: dma.c,v 1.3 2003/05/13 17:36:57 cochran Exp $
 *
 *  OSS compatible ring buffer management. The ring buffer may be
 *  mmap'ed into an application address space.
 *
 *  I first used the rvmalloc stuff copied from bttv. Alan Cox did not
 *  like it, so we now use an array of pointers to a single page
 *  each. This saves us the kernel page table manipulations, but we
 *  have to do a page table alike mechanism (though only one
 *  indirection) in software.
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/wrapper.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/usb.h>

#include "exaudio.h"
#include "dma.h"

extern inline unsigned ld2(unsigned int x)
{
  unsigned r = 0;

  if (x >= 0x10000) {
    x >>= 16;
    r += 16;
  }
  if (x >= 0x100) {
    x >>= 8;
    r += 8;
  }
  if (x >= 0x10) {
    x >>= 4;
    r += 4;
  }
  if (x >= 4) {
    x >>= 2;
    r += 2;
  }
  if (x >= 2)
    r++;
  return r;
}

void dmabuf_copyin(struct dmabuf *db, const void *buffer, unsigned int size)
{
  unsigned int pgrem, rem;

  db->total_bytes += size;
  for (;;) {
    if (size <= 0)
      return;
    pgrem = ((~db->wrptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - db->wrptr;
    if (pgrem > rem)
      pgrem = rem;
    memcpy((db->sgbuf[db->wrptr >> PAGE_SHIFT]) +
	   (db->wrptr & (PAGE_SIZE - 1)), buffer, pgrem);
    size -= pgrem;
    (char *) buffer += pgrem;
    db->wrptr += pgrem;
    if (db->wrptr >= db->dmasize)
      db->wrptr = 0;
  }
}

int dmabuf_copyin_user(struct dmabuf *db, unsigned int ptr, const void *buffer,
		       unsigned int size)
{
  unsigned int pgrem, rem;

  if (!db->ready || db->mapped)
    return -EINVAL;
  for (;;) {
    if (size <= 0)
      return 0;
    pgrem = ((~ptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - ptr;
    if (pgrem > rem)
      pgrem = rem;
    if (copy_from_user
	((db->sgbuf[ptr >> PAGE_SHIFT]) + (ptr & (PAGE_SIZE - 1)), buffer,
	 pgrem))
      return -EFAULT;
    size -= pgrem;
    (char *) buffer += pgrem;
    ptr += pgrem;
    if (ptr >= db->dmasize)
      ptr = 0;
  }
}

void dmabuf_copyout(struct dmabuf *db, void *buffer, unsigned int size)
{
  unsigned int pgrem, rem;

  db->total_bytes += size;
  for (;;) {
    if (size <= 0)
      return;
    pgrem = ((~db->rdptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - db->rdptr;
    if (pgrem > rem)
      pgrem = rem;
    memcpy(buffer,
	   (db->sgbuf[db->rdptr >> PAGE_SHIFT]) +
	   (db->rdptr & (PAGE_SIZE - 1)), pgrem);
    size -= pgrem;
    (char *) buffer += pgrem;
    db->rdptr += pgrem;
    if (db->rdptr >= db->dmasize)
      db->rdptr = 0;
  }
}

int dmabuf_copyout_user(struct dmabuf *db, unsigned int ptr, void *buffer,
			unsigned int size)
{
  unsigned int pgrem, rem;

  if (!db->ready || db->mapped)
    return -EINVAL;
  for (;;) {
    if (size <= 0)
      return 0;
    pgrem = ((~ptr) & (PAGE_SIZE - 1)) + 1;
    if (pgrem > size)
      pgrem = size;
    rem = db->dmasize - ptr;
    if (pgrem > rem)
      pgrem = rem;
    if (copy_to_user
	(buffer, (db->sgbuf[ptr >> PAGE_SHIFT]) + (ptr & (PAGE_SIZE - 1)),
	 pgrem))
      return -EFAULT;
    size -= pgrem;
    (char *) buffer += pgrem;
    ptr += pgrem;
    if (ptr >= db->dmasize)
      ptr = 0;
  }
}

int dmabuf_init(struct dmabuf *db)
{
  unsigned int nr, bytepersec, bufs;
  void *p;

  /* initialize some fields */
  db->rdptr = db->wrptr = db->total_bytes = db->count = db->error = 0;
  /* calculate required buffer size */
  bytepersec = db->srate << AFMT_BYTESSHIFT(db->format);
  bufs = 1U << DMABUFSHIFT;
  if (db->ossfragshift) {
    if ((1000 << db->ossfragshift) < bytepersec)
      db->fragshift = ld2(bytepersec / 1000);
    else
      db->fragshift = db->ossfragshift;
  } else {
    db->fragshift =
      ld2(bytepersec / 100 / (db->subdivision ? db->subdivision : 1));
    if (db->fragshift < 3)
      db->fragshift = 3;
  }
  db->numfrag = bufs >> db->fragshift;
  while (db->numfrag < 4 && db->fragshift > 3) {
    db->fragshift--;
    db->numfrag = bufs >> db->fragshift;
  }
  db->fragsize = 1 << db->fragshift;
  if (db->ossmaxfrags >= 4 && db->ossmaxfrags < db->numfrag)
    db->numfrag = db->ossmaxfrags;
  db->dmasize = db->numfrag << db->fragshift;
  for (nr = 0; nr < NRSGBUF; nr++) {
    if (!db->sgbuf[nr]) {
      p = (void *) get_free_page(GFP_KERNEL);
      if (!p)
	return -ENOMEM;
      db->sgbuf[nr] = p;
      mem_map_reserve(virt_to_page(p));
    }
    memset(db->sgbuf[nr], AFMT_ISUNSIGNED(db->format) ? 0x80 : 0, PAGE_SIZE);
    if ((nr << PAGE_SHIFT) >= db->dmasize)
      break;
  }
  db->bufsize = nr << PAGE_SHIFT;
  db->ready = 1;
  dbg("dmabuf_init: bytepersec %d bufs %d ossfragshift %d ossmaxfrags %d "
      "fragshift %d fragsize %d numfrag %d dmasize %d bufsize %d fmt 0x%x",
      bytepersec, bufs, db->ossfragshift, db->ossmaxfrags,
      db->fragshift, db->fragsize, db->numfrag, db->dmasize,
      db->bufsize, db->format);
  return 0;
}

int dmabuf_mmap(struct dmabuf *db, struct vm_area_struct *vma)
{
  unsigned int nr;
  unsigned long start = vma->vm_start;
  unsigned long size = vma->vm_end - vma->vm_start;
  pgprot_t prot = vma->vm_page_prot;

  if (!db->ready || db->mapped || (start | size) & (PAGE_SIZE - 1)
      || size > db->bufsize)
    return -EINVAL;
  size >>= PAGE_SHIFT;
  for (nr = 0; nr < size; nr++)
    if (!db->sgbuf[nr])
      return -EINVAL;
  db->mapped = 1;

  for (nr = 0; nr < size; nr++) {
    if (remap_page_range(
#if defined(REDHAT_9) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,3)
			 vma, 
#endif
			 start,virt_to_phys(db->sgbuf[nr]),PAGE_SIZE,prot))
      return -EAGAIN;
    start += PAGE_SIZE;
  }

  return 0;
}

void dmabuf_release(struct dmabuf *db)
{
  unsigned int nr;
  void *p;

  for (nr = 0; nr < NRSGBUF; nr++) {
    if (!(p = db->sgbuf[nr]))
      continue;
    mem_map_unreserve(virt_to_page(p));
    free_page((unsigned long) p);
    db->sgbuf[nr] = NULL;
  }
  db->mapped = db->ready = 0;
}
