/* $Id: mixer.c,v 1.6 2003/06/24 02:24:57 cochran Exp $
 *
 *  Copyright (c) 2003 Richard Cochran (cochran@cs.umass.edu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 */
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <linux/vmalloc.h>

#include "exaudio.h"
#include "mixer.h"

/* Mixer defines, especially for exaudio */
#define TRUE  1
#define FALSE 0
#define RECORD   1
#define PLAYBACK 2

#define N_RECMIX_CH 4
#define REC_LINE_UNIT   8
#define REC_MIC_UNIT    10
#define REC_SRC_UNIT    11
#define REC_MASTER_UNIT 12

#define N_PLAYMIX_CH 11
#define PLAY_PCM_UNIT     2
#define PLAY_DIGIN_UNIT   6
#define PLAY_MASTER_UNIT  18
#define PLAY_CPAIR1_UNIT  18
#define PLAY_CPAIR2_UNIT  18
#define PLAY_CPAIR3_UNIT  18
#define PLAY_SPEAKER_UNIT 28
#define PLAY_DIGOUT1_UNIT 25
#define PLAY_DIGOUT2_UNIT 23
#define PLAY_DIGOUTS_UNIT 22
#define UPMIX_UNIT 17

#define EX1_UNIT   27
#define EX2_UNIT   21
#define FX3D_UNIT  15


/* private prototypes */
static void init_mixch(struct mixerdev* ms,struct mixerchannel* ch,
		       u16 osschan,u8 volctl,u8 mutectl,
		       u8 leftch,u8 rightch);
static int mixer_ioctl(struct inode *inode,struct file *file,
		       unsigned int cmd,unsigned long arg);
static int open_mixer(struct inode *inode,struct file *file);
static int query_vol_range(struct usb_device* dev,
			   unsigned char unitid,unsigned int ctrlif,
			   struct mixerchannel* ch);
static int release_mixer(struct inode *inode, struct file *file);
static int set_rec_src(struct mixerdev *ms, int srcmask);
static int wrmixer(struct mixerdev* ms,struct mixerchannel* ch,
		   unsigned value);

/* globals */
extern struct semaphore open_sem;
extern struct list_head alldevices;

static struct file_operations exaudio_mixer_fops = {
  owner:   THIS_MODULE,
  ioctl:   mixer_ioctl,
  open:    open_mixer,
  release: release_mixer,
};

static void SET(struct usb_device* dev,
		u8 req,u16 wValue,u16 wIndex,u16 d1,u16 d2,u16 len)
{
  int err;
  unsigned char* buf;
  if (len < 3) {
    buf = kmalloc(2,GFP_KERNEL);
    if (!buf) return;
    buf[0] = d1;
    buf[1] = d2;
    err = usb_control_msg(dev, usb_sndctrlpipe(dev,0), req,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  wValue, wIndex, buf, len, HZ);
    if (err < 0) err("cannot SET wIndex 0x%04hX : %d",wIndex,err);
    kfree(buf);
  }
}

/*----------------------------------------------------------------------------*/
void exaudio_init_mixers(struct extigy *exy,unsigned int ctrlif)
{
  struct mixerdev *ms;

  /* Make the RECORD mixer. */
  ms = kmalloc(sizeof(struct mixerdev) +
	       N_RECMIX_CH * sizeof(struct mixerchannel),GFP_KERNEL);
  if (!ms) {
    err("low memory");
    return;
  }
  memset(ms,0,sizeof(struct mixerdev));
  ms->state    = exy;
  ms->type     = RECORD;
  ms->selector = REC_SRC_UNIT;
  ms->iface    = ctrlif;
  ms->numch    = N_RECMIX_CH;
  init_mixch(ms,&ms->ch[0],SOUND_MIXER_LINE,REC_LINE_UNIT,REC_LINE_UNIT,1,2);
  init_mixch(ms,&ms->ch[1],SOUND_MIXER_MIC,REC_MIC_UNIT,REC_MIC_UNIT,0,0);
  init_mixch(ms,&ms->ch[2],SOUND_MIXER_IMIX,0,0,0,0);/*No voladj "What U Hear"*/
  init_mixch(ms,&ms->ch[3],SOUND_MIXER_VOLUME,
	     REC_MASTER_UNIT,REC_MASTER_UNIT,0,0);
  ms->ossmixer = register_sound_mixer(&exaudio_mixer_fops,-1);
  if (ms->ossmixer < 0) {
    err("cannot register mixer");
    kfree(ms);
    return;
  } else {
    INFO("registered mixer (major 14 minor %d)",ms->ossmixer);
  }
  list_add_tail(&ms->list,&exy->mixerlist);

  /* Make the PLAYBACK mixer. */
  ms = kmalloc(sizeof(struct mixerdev) +
	       N_PLAYMIX_CH * sizeof(struct mixerchannel),GFP_KERNEL);
  if (!ms) {
    err("low memory");
    return;
  }
  memset(ms,0,sizeof(struct mixerdev));
  ms->state    = exy;
  ms->type     = PLAYBACK;
  ms->selector = 0;
  ms->iface    = ctrlif;
  ms->numch    = N_PLAYMIX_CH;
  init_mixch(ms,&ms->ch[0],SOUND_MIXER_PCM,
	     PLAY_PCM_UNIT,PLAY_PCM_UNIT,0,0);
  init_mixch(ms,&ms->ch[1],SOUND_MIXER_ALTPCM,
	     PLAY_DIGIN_UNIT,PLAY_DIGIN_UNIT,0,0);
  init_mixch(ms,&ms->ch[2],SOUND_MIXER_VOLUME,
	     PLAY_MASTER_UNIT,PLAY_MASTER_UNIT,0,0);
  init_mixch(ms,&ms->ch[3],SOUND_MIXER_LINE1,PLAY_CPAIR1_UNIT,0,1,2);
  init_mixch(ms,&ms->ch[4],SOUND_MIXER_LINE2,PLAY_CPAIR2_UNIT,0,3,4);
  init_mixch(ms,&ms->ch[5],SOUND_MIXER_LINE3,PLAY_CPAIR3_UNIT,0,5,6);
  init_mixch(ms,&ms->ch[6],SOUND_MIXER_SPEAKER,0,PLAY_SPEAKER_UNIT,0,0);
  init_mixch(ms,&ms->ch[7],SOUND_MIXER_DIGITAL1,0,PLAY_DIGOUT1_UNIT,0,0);
  init_mixch(ms,&ms->ch[8],SOUND_MIXER_DIGITAL2,0,PLAY_DIGOUT2_UNIT,0,0);
  init_mixch(ms,&ms->ch[9],SOUND_MIXER_DIGITAL3,0,PLAY_DIGOUTS_UNIT,0,0);

  init_mixch(ms,&ms->ch[10],SOUND_MIXER_VIDEO,0,UPMIX_UNIT,0,0);

  ms->ossmixer = register_sound_mixer(&exaudio_mixer_fops,-1);
  if (ms->ossmixer < 0) {
    err("cannot register mixer");
    kfree(ms);
    return;
  } else {
    INFO("registered mixer (major 14 minor %d)",ms->ossmixer);
  }
  list_add_tail(&ms->list,&exy->mixerlist);

  SET(exy->usbdev,0x01,0x0000,0x1d00,0x01,0x0,1);//selector
  SET(exy->usbdev,0x01,0x0100,0x1100,0x01,0x0,1);//enable up MIX
  SET(exy->usbdev,0x01,0x0200,0x1100,0x05,0x0,1);//mode   up MIX
}

/*----------------------------------------------------------------------------*/
static int get_rec_src(struct mixerdev *ms)
{
  struct usb_device *dev = ms->state->usbdev;
  unsigned int mask = 0;
  unsigned char val;
  int err;

  switch (ms->type) {
  case RECORD:
    err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			  0,ms->iface|(ms->selector<<8),&val,1,HZ);
    if (err < 0) {
      err("selector read request DEV%u if %u unit %u failed",
	  dev->devnum,ms->iface,ms->selector&0xff);
      return(-EIO);
    } else {
      switch (val) {
      case 1: mask = 1<<SOUND_MIXER_LINE; break;
      case 2: mask = 1<<SOUND_MIXER_MIC;  break;
      case 3: mask = 1<<SOUND_MIXER_IMIX; break;
      default:;
      }
      return(mask);
    }
  case PLAYBACK: return(0);
  default:       return(-EINVAL);
  }
}

/*----------------------------------------------------------------------------*/
static void init_mixch(struct mixerdev* ms,struct mixerchannel* ch,
		       u16 osschan,u8 volctl,u8 mutectl,
		       u8 leftch,u8 rightch)
{
  struct usb_device* dev = ms->state->usbdev;

  ch->osschannel = osschan;
  ch->value      = 0;
  ch->volctl     = volctl;
  ch->mutectl    = mutectl;
  if (leftch) {
    ch->leftch  = leftch;
    ch->rightch = rightch;
  } else {
    ch->leftch  = 0;
    ch->rightch = 0;
  }
  ch->min        = 0;
  ch->max        = 0;
  ch->step       = 1;
  if (volctl) {
    if (!query_vol_range(dev,volctl,ms->iface,ch)) {
      err("cannot query VOL ranges on u%u",volctl);
    }
  }
}

/*----------------------------------------------------------------------------*/
static int mixer_ioctl(struct inode *inode,struct file *file,
		       unsigned int cmd,unsigned long arg)
{
  struct mixerdev *ms = (struct mixerdev*)file->private_data;
  int i, j, val;

  if (!ms->state->usbdev) return(-ENODEV);
  if (cmd == SOUND_MIXER_INFO) {
    mixer_info info;
    strncpy(info.id,"EXAUDIO",sizeof(info.id));
    if (RECORD == ms->type)
      strncpy(info.name,"EXAUDIO Recording Mixer",sizeof(info.name));
    else
      strncpy(info.name,"EXAUDIO Playback Mixer",sizeof(info.name));
    info.modify_counter = ms->modcnt;
    if (copy_to_user((void*)arg,&info,sizeof(info))) return(-EFAULT);
    return(0);
  }
  if (cmd == SOUND_OLD_MIXER_INFO) {
    _old_mixer_info info;
    strncpy(info.id,"EXAUDIO",sizeof(info.id));
    if (RECORD == ms->type)
      strncpy(info.name,"EXAUDIO Recording Mixer",sizeof(info.name));
    else
      strncpy(info.name,"EXAUDIO Playback Mixer",sizeof(info.name));
    if (copy_to_user((void*)arg,&info,sizeof(info))) return(-EFAULT);
    return(0);
  }
  if (cmd == OSS_GETVERSION) return(put_user(SOUND_VERSION,(int*)arg));
  if (_IOC_TYPE(cmd) != 'M' || _IOC_SIZE(cmd) != sizeof(int)) return(-EINVAL);

  if (_IOC_DIR(cmd) == _IOC_READ) {
    switch (_IOC_NR(cmd)) {

    case SOUND_MIXER_RECSRC: /* Set a bit for the current recording source. */
      val = get_rec_src(ms);
      dbg("mixer RECSRC request, reply %04X",val);
      if (val < 0) return(val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_DEVMASK: /* Set a bit for each supported device. */
      for (val = i = 0; i < ms->numch; i++) val |= 1 << ms->ch[i].osschannel;
      dbg("mixer DEVMASK request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_RECMASK: /* a bit for each supported recording source */
      if (RECORD == ms->type) {
	val =
	  1<<ms->ch[0].osschannel |
	  1<<ms->ch[1].osschannel |
	  1<<ms->ch[2].osschannel;
      } else {
	val = 0;
      }
      dbg("mixer RECMASK request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_STEREODEVS: /* Mixer channels supporting stereo */
      for (val = i = 0; i < ms->numch; i++)
	if (ms->ch[i].leftch) val |= 1 << ms->ch[i].osschannel;
      dbg("mixer STEREODEVS request, reply %04X",val);
      return(put_user(val,(int*)arg));

    case SOUND_MIXER_CAPS: /* We only allow one recording source at a time. */
      return(put_user(SOUND_CAP_EXCL_INPUT,(int*)arg));

    default: /* Read channel's current OSS value. */
      i = _IOC_NR(cmd);
      if (i >= SOUND_MIXER_NRDEVICES) return(-EINVAL);
      for (j = 0; j < ms->numch; j++)
	if (ms->ch[j].osschannel == i) {
	  return(put_user(ms->ch[j].value,(int*)arg));
	}
      return(-EINVAL);
    }
  }

  if (_IOC_DIR(cmd) != (_IOC_READ|_IOC_WRITE)) return(-EINVAL);
  ms->modcnt++;
  switch (_IOC_NR(cmd)) {
  case SOUND_MIXER_RECSRC: /* Arg contains a bit for each recording source */
    if (get_user(val,(int*)arg)) return(-EFAULT);
    return(set_rec_src(ms,val));

  default: /* Set the channel value. */
    i = _IOC_NR(cmd);
    if (i >= SOUND_MIXER_NRDEVICES) return(-EINVAL);
    for (j = 0; j < ms->numch && ms->ch[j].osschannel != i; j++)
      ;
    if (j >= ms->numch)             return(-EINVAL);
    if (get_user(val,(int*)arg))    return(-EFAULT);
    if (wrmixer(ms,&ms->ch[j],val)) return(-EIO);
    return(put_user(ms->ch[j].value,(int*)arg));
  }
}

/*----------------------------------------------------------------------------*/
static int open_mixer(struct inode *inode,struct file *file)
{
  int minor = MINOR(inode->i_rdev);
  struct list_head *devs, *mdevs;
  struct mixerdev *ms;
  struct extigy *exy;

  down(&open_sem);
  for (devs = alldevices.next; devs != &alldevices; devs = devs->next) {
    exy = list_entry(devs, struct extigy, devices);
    for (mdevs = exy->mixerlist.next; mdevs != &exy->mixerlist;
	 mdevs = mdevs->next) {
      ms = list_entry(mdevs, struct mixerdev, list);
      if (ms->ossmixer == minor)
	goto mixer_found;
    }
  }
  up(&open_sem);
  return -ENODEV;

mixer_found:
  if (!exy->usbdev) {
    up(&open_sem);
    return -EIO;
  }
  file->private_data = ms;
  exy->count++;
  up(&open_sem);
  return 0;
}

/*----------------------------------------------------------------------------*/
static int query_vol_range(struct usb_device* dev,
			   unsigned char unitid,unsigned int ctrlif,
			   struct mixerchannel* ch)
{ 
  int err, range;
  char v1=0, v2=0;
  short tmp;
  double step;
  unsigned char buf[2];

  ch->min = ch->max = ch->step = 0;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_MIN,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u min VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->min = tmp;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_MAX,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u max VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->max = tmp;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_RES,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|0,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u res VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  ch->step = tmp;

  range = ch->max - ch->min + 1;
  step = range/100.0;

  err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_CUR,
			USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			(VOLUME_CONTROL<<8)|ch->leftch,
			(unitid<<8)|ctrlif,
			buf,2,HZ);
  if (err < 0) return(FALSE);
  tmp = buf[0] | (buf[1]<<8);
  dbg("unit%u cur VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
  v1 = (tmp - ch->min) / step;

  if (ch->rightch) {
    err = usb_control_msg(dev,usb_rcvctrlpipe(dev,0),GET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			  (VOLUME_CONTROL<<8)|ch->rightch,
			  (unitid<<8)|ctrlif,
			  buf,2,HZ);
    if (err < 0) return(FALSE);
    tmp = buf[0] | (buf[1]<<8);
    dbg("unit%u cur VOL 0x%04X %d",unitid,tmp&0xFFFF,tmp);
    v2 = (tmp - ch->min) / step;
  }

  ch->value = v1 | (v2<<8);

  dbg("unit%d CUR OSS VOL: L %d R %d",unitid,v1,v2);

  return(TRUE);
}

/*----------------------------------------------------------------------------*/
static int release_mixer(struct inode *inode, struct file *file)
{
  struct mixerdev *ms = (struct mixerdev*)file->private_data;
  struct extigy *s;
  
  lock_kernel();
  s = ms->state;
  down(&open_sem);
  release(s);
  unlock_kernel();
  return 0;
}

/*----------------------------------------------------------------------------*/
static int set_rec_src(struct mixerdev *ms, int srcmask)
{
  struct usb_device *dev = ms->state->usbdev;
  unsigned char val = 1;
  int err;

  switch (ms->type) {
  case RECORD:
    if      (srcmask & (1<<SOUND_MIXER_LINE)) val = 1;
    else if (srcmask & (1<<SOUND_MIXER_MIC))  val = 2;
    else if (srcmask & (1<<SOUND_MIXER_IMIX)) val = 3;
    if (val) {
      err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			    USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			    0,ms->iface|(ms->selector<<8),&val,1,HZ);
      if (err < 0) {
	err("selector write request DEV%u IF%u unit%u failed",
	    dev->devnum,ms->iface,ms->selector);
	return(-EIO);
      }
      info("set recording selector to pin %u srcmask 0x%X",val,srcmask);
      return(0);
    } else {
      return(-EINVAL);
    }
  case PLAYBACK:
  default:
    return(-EINVAL);
  }
}

/*----------------------------------------------------------------------------*/
static int wrmixer(struct mixerdev* ms,struct mixerchannel* ch,
		   unsigned value)
{
  struct usb_device* dev = ms->state->usbdev;
  int err, range;
  char v1, v2;
  short usbval;
  unsigned char buf[2];
  double step;

  if (SOUND_MIXER_VIDEO == ch->osschannel) {
    v1 = (value & 0xFF)/10;
    if (v1 < 1) v1 = 1;
    else if (v1 > 7) v1 = 7;
    INFO("request=0x%X, setting UP DOWN MIX mode to %d",value,v1);
    SET(dev,0x01,0x0200,0x1100,v1,0x0,1); /* UP DOWN MIX mode */
    ch->value = v1 * 10; /* Remember OSS value. */
    return(0);
  }

  v1 = value & 0xFF;
  v2 = (value >> 8) & 0xFF;

  if (ch->mutectl) {
    int setmute = FALSE;
    if (v1 > 0 || v2 > 0) {
      if (0 == ch->value) { /* Unmute the control. */
	buf[0] = 0;
	setmute = TRUE;
      }
    } else {
      if (ch->value) { /* Mute the control */
	buf[0] = 1;
	setmute = TRUE;
      }
    }
    if (setmute) {
      err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			    USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			    (MUTE_CONTROL<<8)|ch->leftch,
			    (ch->mutectl<<8)|ms->iface,
			    buf,1,HZ);
      if (err < 0)
	err("cannot set mute on unit%u",ch->mutectl);
      if (ch->rightch) {
	err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			      USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			      (MUTE_CONTROL<<8)|ch->rightch,
			      (ch->mutectl<<8)|ms->iface,
			      buf,1,HZ);
	if (err < 0)
	  err("cannot set mute on unit%u",ch->mutectl);
      }
    }
  }

  if (ch->volctl) {
    range = ch->max - ch->min + 1;
    step = range/100.0;
    usbval = ch->min + (int)(v1*step);
    buf[0] = usbval & 0xFF;
    buf[1] = (usbval>>8) & 0xFF;
    err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  (VOLUME_CONTROL<<8)|ch->leftch,
			  (ch->volctl<<8)|ms->iface,
			  buf,2,HZ);
    if (err < 0)
      err("cannot set VOL on unit%u",ch->volctl);
    if (ch->rightch) {
      usbval = ch->min + (int)(v2*step);
      buf[0] = usbval & 0xFF;
      buf[1] = (usbval>>8) & 0xFF;
      err = usb_control_msg(dev,usb_sndctrlpipe(dev,0),SET_CUR,
			    USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			    (VOLUME_CONTROL<<8)|ch->rightch,
			    (ch->volctl<<8)|ms->iface,
			    buf,2,HZ);
      if (err < 0) err("exaudio: cannot set VOL on unit%u",ch->volctl);
    }
    ch->value = value;  /* Remember OSS value. */
  }

  return(0);
}
