/* $Id: remote.c,v 1.14 2003/06/24 02:24:57 cochran Exp $
 *
 *  Copyright (c) 2003 Richard Cochran (cochran@cs.umass.edu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 */
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#ifdef USE_LIRC
#include "../lirc/drivers/lirc.h"
#endif

#define SET_CUR    0x01   /* Audio Class specific Request Codes */
#define GET_CUR    0x81
#define SET_MIN    0x02
#define GET_MIN    0x82
#define SET_MAX    0x03
#define GET_MAX    0x83
#define SET_RES    0x04
#define GET_RES    0x84
#define SET_MEM    0x05
#define GET_MEM    0x85
#define GET_STAT   0xff

#define DRIVER_NAME "exaudio_remote"
MODULE_AUTHOR("Richard Cochran");
MODULE_DESCRIPTION("IR remote driver for extigy");
MODULE_LICENSE("GPL");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct devrequest {
  __u8 requesttype;
  __u8 request;
  __u16 value;
  __u16 index;
  __u16 length;
} __attribute__ ((packed)) devrequest;
#endif

/* globals */
static DECLARE_WAIT_QUEUE_HEAD(readwq);
static DECLARE_MUTEX(alldevmux);
static LIST_HEAD(alldevices);

static int irmajor;
static struct tq_struct remove_task;

struct ir_device {
  struct list_head devices; /* Points to next (!) extigy. */
  struct usb_device *dev;
  struct urb urb;
  unsigned char buf[8];
  int errors;
  u32 irval;
};

/* private prototypes */
static void ir_disconnect(struct usb_device *udev, void *context);
static int ir_open(struct inode *inode, struct file *filp);
static void *ir_probe(struct usb_device *dev,unsigned int ifnum,
			const struct usb_device_id *id);
static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos);
static int ir_release(struct inode *inode, struct file *filp);
static ssize_t ir_write(struct file *file,const char *buffer,
			size_t count,loff_t *ppos);
static void remove_ir_device(void* context);

static struct usb_device_id ir_id_table [] = {
  { USB_DEVICE(0x041E,0x3000),
    bInterfaceClass: USB_CLASS_AUDIO,
    bInterfaceSubClass: 1
  },
  {} /* Terminating entry */
};

static struct usb_driver ir_driver = {
  name:       DRIVER_NAME,
  disconnect: ir_disconnect,
  id_table:   ir_id_table,
  probe:      ir_probe,
};

static struct file_operations ir_fops = {
  owner:   THIS_MODULE,
  open:    ir_open,
  read:    ir_read,
  release: ir_release,
  write:   ir_write,
};

#ifdef USE_LIRC
#define LIRC_MAJOR 61

static int lirc_ioctl(struct inode *node,struct file *filep,unsigned int cmd,
		      unsigned long arg)
{
  int result;
  unsigned long features=LIRC_CAN_REC_CODE, mode;
  switch(cmd) {
  case LIRC_GET_FEATURES:
    result = put_user(features,(unsigned long*)arg);
    if (result) return(result); 
    break;
  case LIRC_GET_REC_MODE:
    result = put_user(LIRC_MODE_CODE,(unsigned long*)arg);
    if (result) return(result); 
    break;
  case LIRC_SET_REC_MODE:
    result = get_user(mode,(unsigned long*)arg);
    if (result) return(result);
    if (mode!=LIRC_MODE_CODE) return(-ENOSYS);
    break;
  default:
    return(-ENOIOCTLCMD);
  }
  return(0);
}

static int lirc_open(struct inode* node,struct file* filep)
{
  return(ir_open(node,filep));
}

static int lirc_release(struct inode* node,struct file* filep)
{
  return(0);
}

static struct file_operations lirc_fops = 
{
  owner:   THIS_MODULE,
  ioctl:   lirc_ioctl,
  open:    lirc_open,
  read:    ir_read,
  release: lirc_release,
};
#endif

/* module boiler plate */

int ir_init(void)
{
  info("$Revision: 1.14 $");
#ifdef USE_LIRC
  if (register_chrdev(LIRC_MAJOR,DRIVER_NAME,&lirc_fops) < 0) {
    err("register_chrdev() failed for lirc");
    return(-1);
  }
  info("registered lirc device, major %d",LIRC_MAJOR);
#endif
  irmajor = register_chrdev(0,DRIVER_NAME,&ir_fops);
  if (irmajor < 0) {
    err("cannot get a major number");
    return(-1);
  } else {
    info("registered /dev/ir (major %d)",irmajor);
  }
  return(usb_register(&ir_driver));
}

void ir_exit(void)
{
#ifdef USE_LIRC
  if (unregister_chrdev(LIRC_MAJOR,DRIVER_NAME))
    err("cannot unregister lirc device");
#endif
  if (unregister_chrdev(irmajor,DRIVER_NAME))
    err("cannot unregister ir device");
  usb_deregister(&ir_driver);
}

module_init(ir_init);
module_exit(ir_exit);

/* methods */

static int ctlmsg(struct usb_device* dev,u8 request,u8 requesttype,u16 value,
		  u16 index,u16 len,usb_complete_t complete,void* context)
{
  struct urb* urb;
  devrequest* dr;
  unsigned char* buf;
  urb = usb_alloc_urb(0);
  dr  = kmalloc(sizeof(devrequest),GFP_ATOMIC);
  buf = kmalloc(len,GFP_ATOMIC);
  if (!urb || !dr || !buf) {
    err("low memory");
    if (urb) kfree(urb);
    if (dr)  kfree(dr);
    if (buf) kfree(buf);
    return(-1);
  }
  dr->request     = request;
  dr->requesttype = requesttype;
  dr->value       = cpu_to_le16p(&value);
  dr->index       = cpu_to_le16p(&index);
  dr->length      = cpu_to_le16p(&len);
  usb_fill_control_urb(urb,dev,usb_rcvctrlpipe(dev,0),
		       (char*)dr,buf,len,complete,context);
  return(usb_submit_urb(urb));
}

static void fetch_ir_code(struct urb* u)
{
  struct ir_device* ir = (struct ir_device*)u->context;
  unsigned char* buf = u->transfer_buffer;
  if (u->status < 0) {
    err("control request with bad status=%d",u->status);
  } else {
    ir->irval = (buf[1]<<8) | buf[0];
  }
  kfree(u->setup_packet);
  kfree(u->transfer_buffer);
  usb_free_urb(u);
  wake_up_interruptible(&readwq);
}

static void handle_interrupt(struct urb* u)
{
  static unsigned int last = 0;
  struct ir_device* ir = (struct ir_device*)u->context;

  if (u->status < 0) {
    dbg("interrupt with bad status=%d",u->status);
    ir->errors++;
    remove_task.routine = remove_ir_device;
    remove_task.data = ir;
    schedule_task(&remove_task);
    wake_up_interruptible(&readwq);
    return;
  }

  /* Fix the toggle bit, perhaps. */
  dbg("IN BITs %08X last %d",ir->dev->toggle[0],last);
  if (last == usb_gettoggle(ir->dev,1,0)) {
    dbg("missed a toggle - fix it");
    usb_dotoggle(ir->dev,1,0);
  }
  last = usb_gettoggle(ir->dev,1,0);

  if (ir->buf[1]) return; /* A jack (un)plugged event. */

  /* Fetch the code from memory. */
  if (ctlmsg(ir->dev,GET_MEM,0xA1,0,0,2,fetch_ir_code,ir)) {
    err("cannot send control message to get IR code");
  }
}

static void ir_disconnect(struct usb_device *udev, void *context)
{
  /* This never gets called, since probe() returns null. */
  return;
}

static int ir_open(struct inode *inode, struct file *filp)
{
  int device_found=0;
  /*int minor=MINOR(inode->i_rdev);*/
  struct list_head *dev;
  struct ir_device* ir=NULL;

  if (down_interruptible(&alldevmux)) return(-ERESTARTSYS);
  for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
    ir = list_entry(dev,struct ir_device,devices);
    if (1 /*minor == ir->minor*/) {
      device_found = 1;
      break;
    }
  }
  if (!device_found) { /* Requested device not found. */
    up(&alldevmux);
    return(-ENODEV);
  }
  filp->private_data = ir;
  up(&alldevmux);
  return(0);
}

static void *ir_probe(struct usb_device *dev,unsigned int ifnum,
			 const struct usb_device_id *id)
{
  int err;
  unsigned int pipe;
  struct ir_device *ir;
  
  if (0x041E != dev->descriptor.idVendor ||
      0x3000 != dev->descriptor.idProduct) {
    err("not an extigy");
    return(NULL);
  }
  if (alldevices.next != &alldevices) {
    /* Allow only one device, for now. */
    return(NULL);
  }

  info("initializing driver on probe of IF%d",ifnum);
  MOD_INC_USE_COUNT;

  /* Allocate a device structure. */
  ir = kmalloc(sizeof(struct ir_device),GFP_KERNEL);
  if (!ir) {
    err("low memory");
    MOD_DEC_USE_COUNT;
    return(NULL);
  }
  memset(ir,0,sizeof(struct ir_device));
  ir->dev = dev;
  /* Setup interrupt. */
  pipe = usb_rcvintpipe(dev,1);
  FILL_INT_URB(&ir->urb,dev,pipe,ir->buf,8,handle_interrupt,ir,1);
  down(&alldevmux);
  err = usb_submit_urb(&ir->urb);
  if (err) {
    err("cannot submit interrupt urb: %d",err);
    kfree(ir);
    MOD_DEC_USE_COUNT;
    up(&alldevmux);
    return(NULL);
  }
  /* Add into alldevices. */
  list_add_tail(&ir->devices,&alldevices);
  up(&alldevmux);
  /* Don't bother to claim an interface. */
  return(NULL);
}

static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos)
{
  int device_found=0;
  struct ir_device* ir = (struct ir_device*)filp->private_data;
  struct list_head *dev;
  u8 code;
  u32 result;

  /* Go to sleep until an IR event occurs. */
  interruptible_sleep_on(&readwq);
  if (signal_pending(current)) return(-ERESTARTSYS);

  if (down_interruptible(&alldevmux)) return(-ERESTARTSYS);
  for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
    if (ir == list_entry(dev,struct ir_device,devices)) {
      device_found = 1;
      break;
    }
  }
  if (!device_found || ir->errors) { /* Device has disappeared. */
    up(&alldevmux);
    return(-ENODEV);
  }
  code = result = ir->irval;
  up(&alldevmux);

  switch (count) {
  case 1: if (copy_to_user(buf,&code,count))   return(-EFAULT); break;
  case 4: if (copy_to_user(buf,&result,count)) return(-EFAULT); break;
  default: return(-EINVAL);
  }
  *f_pos += count;
  return(count);
}

static int ir_release(struct inode *inode, struct file *file)
{
  return(0);
}

/* 
 *  ir_write, SET, and GET only exist to send control requests from user
 *  space, to try out new ideas quickly.
 */

static void SET(struct usb_device* dev,
		u8 req,u16 wValue,u16 wIndex,u16 d1,u16 d2,u16 len)
{
  int err;
  unsigned char* buf;
  if (len < 3) {
    buf = kmalloc(2,GFP_KERNEL);
    if (!buf) return;
    buf[0] = d1;
    buf[1] = d2;
    err = usb_control_msg(dev, usb_sndctrlpipe(dev,0), req,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  wValue, wIndex, buf, len, HZ);
    if (err < 0) err("cannot SET wIndex 0x%04hX : %d",wIndex,err);
    kfree(buf);
  }
}

static void GET(struct usb_device* dev,u8 req,u16 wValue,u16 wIndex,u16 len)
{
  int err;
  unsigned char buf[3] = {0};
  if (len < 4) {
    err = usb_control_msg(dev, usb_rcvctrlpipe(dev,0), req,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_IN,
			  wValue, wIndex, buf, len, HZ);
    if (err < 0) err("cannot GET wIndex 0x%04hX : %d",wIndex,err);
    else err("GOT 0x[%02hX][%02hX] from 0x%04hX",buf[1],buf[0],wIndex);
  }
}

static ssize_t ir_write(struct file *file,const char *buffer,
			size_t count,loff_t *ppos)
{
  char inbuf[256];
  int device_found=0, cnt;
  struct ir_device* ir = (struct ir_device*)file->private_data;
  struct list_head *dev;
  struct usb_device *usbdev;
  u16 req, val, idx, len;
  unsigned short buf[2] = {0};

  if (down_interruptible(&alldevmux)) return(-ERESTARTSYS);
  for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
    if (ir == list_entry(dev,struct ir_device,devices)) {
      device_found = 1;
      break;
    }
  }
  if (!device_found) { /* Device has disappeared. */
    up(&alldevmux);
    return(-ENODEV);
  }
  usbdev = ir->dev;
  up(&alldevmux);
  
  if (ppos != &file->f_pos) return(-ESPIPE);
  if (count > 256) count = 256;
  if (copy_from_user(inbuf,buffer,count)) return(-EFAULT);

  /* Req wValue wIndex Data */
  cnt = sscanf(inbuf,"%hd %hd %hd %hd %hd %hd",
	       &req,&val,&idx,&len,
	       &buf[0],&buf[1]);

  if (6 != cnt) {
    err("bad proc command, cnt=%d",cnt);
  } else {
    err("command: 0x%02hX 0x%04hX 0x%04hX %hu data 0x%04hX 0x%04hX",
	req,val,idx,len,buf[0],buf[1]);
    if (0x80 & req) {
      GET(usbdev,req,val,idx,len);
    } else {
      SET(usbdev,req,val,idx,buf[0],buf[1],len);
    }
  }
  return(count);
}

static void remove_ir_device(void* context)
{
  int err;
  struct ir_device *ir = (struct ir_device*)context;
  struct list_head *dev;
  dbg("remove_ir_device() called");
  down(&alldevmux);
  for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
    if (ir == list_entry(dev,struct ir_device,devices)) {
      dbg("remove_ir_device: list_del(), unlink, and free");
      list_del(dev);
      err = usb_unlink_urb(&ir->urb);
      if (err) err("remove_ir_device: cannot unlink data urb: %d",err);
      else dbg("remove_ir_device: unlinked data urb: %d",err);
      kfree(ir);
      MOD_DEC_USE_COUNT;
      break;
    }
  }
  up(&alldevmux);
}
