/*
 * Remote control driver for saa7134 based TV-cards
 * Key codes are obtained from the GPIO port
 *
 * By David Atkinson <px-gpio'AT'psykax.com>
 *
 * Heavily based on lirc_gpio.c (lirc driver for bttv-based tv cards)
 * (L) by Artur Lipowski <alipowski@interia.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 4)
#error "*******************************************************"
#error "Sorry, this driver needs kernel version 2.2.4 or higher"
#error "*******************************************************"
#endif

#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/wrapper.h>
#include <linux/errno.h>

#include "../../drivers/media/video/saa7134.h"
#include "lirc_dev.h"

/*
#if SAA7134_VERSION_CODE < KERNEL_VERSION(0,2,7)
#error "*******************************************************"
#error " Sorry, this driver needs saa7134 version 0.2.7 or     "
#error " higher.                                               "
#error "*******************************************************"
#endif
*/

static int debug = 0;
static int card = 0;
static int minor = -1;
static int board_id = SAA7134_BOARD_UNKNOWN;
static unsigned long gpio_mask = 0;
static unsigned long gpio_lock_mask = 0;
static unsigned long gpio_xor_mask = 0;
static unsigned int soft_gap = 0;
static unsigned char sample_rate = 10;

MODULE_PARM(debug,"i");
MODULE_PARM(card,"i");
MODULE_PARM(minor,"i");
MODULE_PARM(gpio_mask,"l");
MODULE_PARM(gpio_lock_mask,"l");
MODULE_PARM(gpio_xor_mask,"l");
MODULE_PARM(soft_gap,"i");
MODULE_PARM(sample_rate,"b");
MODULE_PARM(board_id,"i");

#undef dprintk
#define dprintk  if (debug) printk

struct rcv_info {
	int board_id;
	unsigned long gpio_mask;
	unsigned long gpio_lock_mask;
	unsigned long gpio_xor_mask;
	unsigned int soft_gap;
	unsigned char sample_rate;
	unsigned char code_length;
};


/* This can fix problems where holding a button down causes it to act as
   seperate keypresses. sample_rate should be set to between 1x and 2x the
   actual updating rate of the GPIO register. */
#define POLLING_GLITCH_FIX 0

static struct rcv_info rcv_infos[] = {
	/* Code length for fly2k/3k a guess based on fly98 */
	{ SAA7134_BOARD_FLYVIDEO2000,   0xEC00000, 0x0040000, 0x0040000, 0, 0, 42 },
	{ SAA7134_BOARD_FLYVIDEO3000,   0xEC00000, 0x0040000, 0x0040000, 0, 0, 42 },
	/* Cinergy 400 has some other input on pin 18 */
	{ SAA7134_BOARD_CINERGY400,          0x3F, 0x0010000, 0x0010000, 0, 10,  0 }
};

static unsigned char code_length = 0;
static unsigned char code_bytes = 1;

#define MAX_BYTES 8

#define SUCCESS 0
#define LOGHEAD "lirc_saa7134 (%d): "

/* how many bits GPIO value can be shifted right before processing
 * it is computed from the value of gpio_mask_parameter
 */
static unsigned char gpio_pre_shift = 0;

static inline int reverse(int data, int bits)
{
	int i;
	int c;

	for (c=0,i=0; i<bits; i++) {
		c |= (((data & (1<<i)) ? 1:0)) << (bits-1-i);
	}

	return c;
}

static int build_key(unsigned long gpio_val, unsigned char codes[MAX_BYTES])
{
	unsigned long mask = gpio_mask;
	unsigned char shift = 0;

	dprintk(LOGHEAD "gpio_val is %lx\n",card,(unsigned long) gpio_val);

	gpio_val ^= gpio_xor_mask;

	if (gpio_lock_mask && (gpio_val & gpio_lock_mask)) {
		return -EBUSY;
	}

	/* extract bits from "raw" GPIO value using gpio_mask */
	codes[0] = 0;
	gpio_val >>= gpio_pre_shift;
	while (mask) {
		if (mask & 1u) {
			codes[0] |= ((gpio_val & 1u) << shift);
			shift++;
		}
		mask >>= 1;
		gpio_val >>= 1;
	}

	dprintk(LOGHEAD "extracted code is %lx\n",card,(unsigned long) codes[0]);

	switch (board_id)
	{
	case SAA7134_BOARD_FLYVIDEO2000:
	case SAA7134_BOARD_FLYVIDEO3000:
		codes[0]=reverse(codes[0],5);
		/* a guess based on fly98 - needs confirming */
		codes[4]=codes[0]<<3;
		codes[5]=((~codes[4])&0xff);
		codes[0]=0x00;
		codes[1]=0x1A;
		codes[2]=0x1F;
		codes[3]=0x2F;
		break;
	default:
		break;
	}

	return SUCCESS;
}

static int get_key(void* data, unsigned char *key, int key_no)
{
	static unsigned long next_time = 0;
	static unsigned char codes[MAX_BYTES];
	unsigned long code = 0;
	unsigned char cur_codes[MAX_BYTES];

	if (key_no > 0)	{
		if (code_bytes < 2 || key_no >= code_bytes) {
			dprintk(LOGHEAD "something wrong in get_key\n",card);
			return -EBADRQC;
		}
		*key = codes[key_no];
		return SUCCESS;
	}

	if (saa7134_read_gpio(card, &code)) {
		dprintk(LOGHEAD "cannot read GPIO\n",card);
		return -EIO;
	}

	if (build_key(code, cur_codes)) {
		return -EFAULT;
	}

	if (soft_gap) {
		if (!memcmp(codes, cur_codes, code_bytes) &&
		    jiffies < next_time) {
			return -EAGAIN;
		}
		next_time = jiffies + soft_gap;
	}

#if POLLING_GLITCH_FIX
	/* Allow for the occasional missing code */
	if(memcmp(codes, cur_codes, code_bytes)) {
		memcpy(codes, cur_codes, code_bytes);
	}
#else
	memcpy(codes, cur_codes, code_bytes);
#endif

	*key = codes[0];

	return SUCCESS;
}

static int set_use_inc(void* data)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static void set_use_dec(void* data)
{
	MOD_DEC_USE_COUNT;
}

static wait_queue_head_t* get_queue(void* data)
{
	return saa7134_get_gpio_queue(card);
}

static struct lirc_plugin plugin = {
	name: "lirc_saa7134",
	get_key: get_key,
	get_queue: get_queue,
	set_use_inc: set_use_inc,
	set_use_dec: set_use_dec
};

/*
 *
 */
int gpio_remote_init(void)
{
	int ret;
	unsigned int mask;

	/* "normalize" gpio_mask
	 * this means shift it right until first bit is set
	 */
	while (!(gpio_mask & 1u)) {
		gpio_pre_shift++;
		gpio_mask >>= 1;
	}

	if (code_length) {
		plugin.code_length = code_length;
	} else {
		/* calculate scan code length in bits if needed */
		plugin.code_length = 1;
		mask = gpio_mask >> 1;
		while (mask) {
			if (mask & 1u) {
				plugin.code_length++;
			}
			mask >>= 1;
		}
	}

	code_bytes = (plugin.code_length/8) + (plugin.code_length%8 ? 1 : 0);
	if (MAX_BYTES < code_bytes) {
		printk (LOGHEAD "scan code too long (%d bytes)\n",
			minor, code_bytes);
		return -EBADRQC;
	}

	/* translate ms to jiffies */
	soft_gap = (soft_gap*HZ) / 1000;

	plugin.minor = minor;
	plugin.sample_rate = sample_rate;

	ret = lirc_register_plugin(&plugin);

	if (0 > ret) {
		printk (LOGHEAD "device registration failed with %d\n",
			minor, ret);
		return ret;
	}

	minor = ret;
	printk(LOGHEAD "driver registered\n", minor);

	return SUCCESS;
}

EXPORT_NO_SYMBOLS;

/* Dont try to use it as a static version !  */

#ifdef MODULE
MODULE_DESCRIPTION("Driver module for remote control (data from saa7134 GPIO port)");
MODULE_AUTHOR("Artur Lipowski");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/*
 *
 */
int init_module(void)
{
	int reserved, rcv_index;

	if (MAX_IRCTL_DEVICES < minor) {
		printk("lirc_saa7134: parameter minor (%d) must be less than %d!\n",
		       minor, MAX_IRCTL_DEVICES-1);
		return -EBADRQC;
	}

	request_module("saa7134");

	/* if gpio_mask not zero then use module parameters
	 * instead of autodetecting TV card
	 */
	if(gpio_mask) {
		if (sample_rate!=0 && (2 > sample_rate || 50 < sample_rate)) {
			printk(LOGHEAD "parameter sample_rate "
			       "must be beetween 2 and 50!\n", minor);
			return -EBADRQC;
		}

		if (sample_rate!=0 && soft_gap &&
		    ((2000/sample_rate) > soft_gap || 1000 < soft_gap)) {
			printk(LOGHEAD "parameter soft_gap "
			       "must be beetween %d and 1000!\n",
			       minor, 2000/sample_rate);
			return -EBADRQC;
		}
	} else {
		if(saa7134_get_cardinfo(card,&board_id,&reserved)==-1) {
			printk(LOGHEAD "could not get TV card information\n", minor);
		}

		printk(LOGHEAD "card saa7134[%u] type %u\n",minor,card,board_id);

		if (board_id == SAA7134_BOARD_UNKNOWN) {
			printk(LOGHEAD "cannot detect TV card %d\n",
				minor,card);
			return -EBADRQC;
		}

		for (rcv_index = 0;
		     rcv_index < sizeof(rcv_infos)/sizeof(struct rcv_info);
		     rcv_index++) {
			if (rcv_infos[rcv_index].board_id == board_id) {
				gpio_mask = rcv_infos[rcv_index].gpio_mask;
				gpio_lock_mask = rcv_infos[rcv_index].gpio_lock_mask;
				gpio_xor_mask = rcv_infos[rcv_index].gpio_xor_mask;
				soft_gap = rcv_infos[rcv_index].soft_gap;
				sample_rate = rcv_infos[rcv_index].sample_rate;
				code_length = rcv_infos[rcv_index].code_length;
				break;
			}
		}
		if (rcv_index == sizeof(rcv_infos)/sizeof(struct rcv_info)) {
			printk(LOGHEAD "TV board type %u not supported!\n", minor, board_id);
			return -EBADRQC;
		}
	}

	request_module("lirc_dev");

	return gpio_remote_init();
}

/*
 *
 */
void cleanup_module(void)
{
	int ret;

	ret = lirc_unregister_plugin(minor);

	if (0 > ret) {
		printk(LOGHEAD "error in lirc_unregister_minor: %d\n"
		       "Trying again...\n",
		       minor, ret);

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ);

		ret = lirc_unregister_plugin(minor);

		if (0 > ret) {
			printk(LOGHEAD "error in lirc_unregister_minor: %d!!!\n",
			       minor, ret);
			return;
		}
	}

	dprintk(LOGHEAD "module successfully unloaded\n", minor);
}
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
