/*
 * This module is an interface to the ZORAN ZR367x0 DVD decoder chip,
 * also known as i34.
 *
 * Copyright (c) 2000,2001 Merrick Johnson <inkling@nop.org>
 * Copyright (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modifications and extensions to original code from David Barth by:
 *	Merrick Johnson <inkling@nop.org> MAINTAINER, send queries here
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/tqueue.h>
#include <asm/io.h>
#include <linux/sched.h>
#ifdef CSS_SUPPORT
#include <linux/cdrom.h>
#endif

#include "mgacore.h"
#include "i34.h"
#include "mgadvd.h"
#include "mgacodec.h"
#include "mgabus.h"
#include "ks0127.h"


/* define this if you want to use the interrupt handler
   to send the bitstream to the decoder (default) */
#define USE_INTERRUPTS
/* define this to provide hardware CSS support
   (not functionnal at this time) */
#undef CSS_SUPPORT


#define BITSTREAM_BLOCK_SIZE	2048

//#ifndef ZR_BITSTREAM
#define ZR_BITSTREAM 4
//#endif



/****************************************************************************
 * LOWLEVEL DRIVER                                                          *
 ****************************************************************************/

/* main registers */
#define ZR_PARAM_ADDR	0x0
#define ZR_PARAM_DATA	0x1
#define ZR_ISR		0x2
#define ZR_IMR		0x2
#define ZR_STATUS0	0x3
#define ZR_HOSTCMD	0x3
#define ZR_STATUS1	0x4
#define ZR_STATUS2	0x5
#define ZR_DVP_DATA	0x6
#define ZR_DEVICE_ID	0x7
#define ZR_ADP_DATA	0x8
#define ZR_ADP_STATUS	0x8
#define ZR_SYS_CLOCK	0x9
#define ZR_NAV_DATA	0xA
#define ZR_NAV_ADDRESS	0xA
#define ZR_RESERVED1	0xB
#define ZR_SUBPICBUF	0xC
#define ZR_VIDEOBUF	0xD
#define ZR_AUDIOBUF	0xE
#define ZR_RESERVED2	0xF

/* indirect registers */
#define ZR_SYSCONFIG		0x00
#define ZR_VIDCONFIG		0x01
#define ZR_SDCONFIG		0x02
#define ZR_DVPGEN1		0x03
#define ZR_DVPGEN2		0x04
#define ZR_DVPGEN3		0x05
//
#define ZR_COLORV		0x08
#define ZR_COLORY		0x09
#define ZR_COLORU		0x0A
//
#define ZR_VIDOUT		0x10
#define ZR_HTOTAL		0x11
#define ZR_VTOTAL		0x12
#define ZR_HSYNCSIZE		0x13
#define ZR_VSYNCSIZE		0x14
#define ZR_CBHSTART		0x15
#define ZR_CBVSTART		0x16
#define ZR_CBHEND		0x17
#define ZR_CBVEND		0x18
#define ZR_ACTIVESTARTX		0x19
#define ZR_ACTIVESTARTY		0x1A
#define ZR_ACTIVEENDX		0x1B
#define ZR_ACTIVEENDY		0x1C
#define ZR_ACTIVESIZEX		0x1D
#define ZR_ACTIVESIZEY		0x1E
#define ZR_CAPTIONOFFSET	0x1F
#define ZR_STARTDISPLAY 	0x20
//
#define ZR_OSDCONTROL		0x28
#define ZR_OSDMEMSTART		0x29
#define ZR_OSDMEMSIZE		0x2A
#define ZR_OSDFIRSTLINE		0x2B
#define ZR_OSDLASTLINE		0x2C
#define ZR_OSDSWITCH		0x2D
//
#define ZR_SCLKVALUE		0x30
//
#define ZR_DISCTYPE		0x38
//
#define ZR_BITSTREAMSELECT	0x40
#define ZR_PLAYBACKMODE		0x41
#define ZR_AUDSID		0x42
#define ZR_VIDSID		0x43
#define ZR_SPSID		0x44
#define ZR_AUDPORTDELAY		0x45
#define ZR_VIDPORTDELAY		0x46
#define ZR_VIDTOLERANCE		0x47
#define ZR_VIDSYNCMODE		0x48
#define ZR_SPPALETTE		0x49
//
#define ZR_AUTOSCALING		0x50
#define	ZR_PANSCANBASEX		0x51
#define	ZR_PANSCANBASEY		0x52
#define	ZR_PANSCANSIZEX		0x53
#define	ZR_PANSCANSIZEY		0x54
#define	ZR_PICSIZEX		0x55
#define	ZR_PICSIZEY		0x56
#define ZR_IMAGESTARTX		0x57
#define ZR_IMAGESTARTY		0x58
#define ZR_IMAGESIZEX		0x59
#define ZR_IMAGESIZEY		0x5A
#define ZR_IMAGEENDX		0x5B
#define ZR_IMAGEENDY		0x5C
//
#define ZR_SCALERATIO		0x5E
#define ZR_SPSCALE		0x5F
//
#define ZR_AUTOPANSCAN		0x68
#define ZR_PANSCANOFFSETX	0x69
#define ZR_PANSCANOFFSETY	0x6A
#define ZR_BACKGROUNDSWITCH	0x6B
#define ZR_SPSWITCH		0x6C
#define ZR_CAPTIONWORD		0x6D
//
#define ZR_CAPTIONSWITCH	0x6F
//
#define ZR_HILIGHTBUTTON1	0x70
#define ZR_HILIGHTBUTTON2	0x71
#define ZR_HILIGHTCOLOR1	0x72
#define ZR_HILIGHTTIMING	0x73
#define ZR_HILIGHTSWITCH	0x74
#define ZR_HILIGHTCOLOR2	0x75
#define ZR_HILIGHTCOLOR3	0x76
//
#define ZR_OSDPALETTE0i		0x80
#define ZR_OSDPALETTE1i		0x90
//
#define ZR_DVPCMD		0xCE
#define ZR_DECCNTR		0xCF
//
#define ZR_DYNPANX		0xE0
#define ZR_DYNPANY		0xE1
//
#define ZR_INITFILE_1		0xF0
#define ZR_INITFILE_2		0xF1
#define ZR_INITFILE_3		0xF2
#define ZR_INITFILE_4		0xF3
//
#define ZR_MICROCODE		0xFD


/* host commands */
#define I34_CMD_START	0x0000
#define I34_CMD_PAUSE	0x0100
#define I34_CMD_STEP	0x0300
#define I34_CMD_SLOW	0x0500
#define I34_CMD_END	0x0700
#define I34_CMD_FAST	0x0800
#define I34_CMD_CONT	0x0900
#define I34_CMD_SEQEND	0x0E00

/* states */
#define ZR_CSTATE_RESET		0x00
#define ZR_CSTATE_INIT_PCLK	0x02
#define ZR_CSTATE_INIT_DISPLAY	0x03
#define ZR_CSTATE_PAUSE		0x10
#define ZR_CSTATE_NSPB		0x20
#define ZR_CSTATE_STEP		0x23
#define ZR_CSTATE_SLOW		0x25
#define ZR_CSTATE_IDLE		0x30
#define ZR_CSTATE_FLAG_NOACT	0x08


/* NEW ADP INTERFACE */

/* status bits */
#define ADPRDY		1
#define ADPEMPTY	2

/* ADP commands */
#define ADP_READ	0x00
#define ADP_NOP		0x80
#define ADP_VER		0x81
#define ADP_CFG		0x82

/* ADP audio microcode function setup */
#define ADP_PINK	0x83
#define ADP_LPCM	0x84
#define ADP_AC3		0x85
#define ADP_PCM		0x86
#define ADP_MPEG	0x87

/* ADP playback control */
#define ADP_UNMUTE	0x89
#define ADP_PLAY	0x8A
#define ADP_MUTE	0x8B
#define ADP_STOP	0x8C
#define ADP_STOPF	0x8D

/* ADP misc configuration */
#define ADP_STAT	0x8E
#define ADP_BOOT	0x90
#define ADP_INTRP	0x91
#define ADP_SETIO	0x92
#define ADP_POKE	0x93
#define ADP_PEEK	0x94
#define ADP_SPDIFCS	0x95
#define ADP_PARAM	0x96
#define ADP_PLLTAB	0x98
#define ADP_PLLCFG	0x99

/* ADP return codes */
#define ADP_NO_ERROR		0x80
#define ADP_OPCODE_ERROR	0x81
#define ADP_EXT_ERROR		0x82
#define ADP_NOT_READY_ERROR	0x83
#define ADP_RESERVED_ERROR	0x84
#define ADP_READY_FOR_NEW_CMD	0x85


/* ISR register bits */
#define ISR_NAVREADY	0x8000
#define ISR_DVPOBF	0x4000
#define ISR_VSYNC	0x2000
#define ISR_PLLpLOCKED	0x1000
#define ISR_PLLaLOCKED	0x0800
#define ISR_res1	0x0400
#define ISR_DVPSWIRQ	0x0200
#define ISR_ADPSWIRQ	0x0100
#define ISR_res2	0x0080
#define ISR_SCLKIRQ	0x0040
#define ISR_res3	0x0020
#define ISR_CODBUFIRQ	0x0010
#define ISR_res4	0x000F



/* SCLK is 90000 Hz, 1 ms = 90 SCLK
   PAL  1 frame = 40ms (3600 SCLKs)
   PAL  1 field = 20ms (1800 SCLKs)
   NTSC 1 frame = 33.366_ ms (3003 SCLKs)
   NTSC 1 field = 16.683_ ms (1501.5 SCLKs)
*/
#define MS_TO_SCLK(ms) (ms*90)

struct video_standard_s
{
	u16 VidFPS;
	u16 FieldSel;
	u16 HTotal;
	u16 VTotal;
	u16 HSyncSize;
	u16 VSyncSize;
	u16 ActiveStartX;
	u16 ActiveStartY;
	u16 ActiveEndX;
	u16 ActiveEndY;
	u16 ActiveSizeX;
	u16 ActiveSizeY;
	u16 VidConfig;
};

struct composite_blank_s
{
	u16 CBHStart;
	u16 CBVStart;
	u16 CBHEnd;
	u16 CBVEnd;
};

struct serial_port_config_s
{
	/* see p. 226 for a definition of the structure */
	u8 cpb;
	u8 cpa;
	u8 frb;
	u8 fra;
	u8 err;
	u8 isp;
	u8 osp;
	u8 spo;
	u8 inw;
	u8 outw;
	u8 spbs;
	u8 fmb;
	u8 fma;
};

struct mga_dev_s;


struct i34_dev_s
{
/* codec configuration section */
	/* WARNING: not all zoran 367x0 registers are 2 bytes long
	   but it's enough for those we're interested in reading back
	*/
	u32 shadow_register[0xff];
	u32 CodeBurstLen;
	u32 init_file_loaded;

	struct mga_codec_s *codec_interface;
	struct serial_port_config_s serial;
	u16 imr;


/* video configuration section */
	u32 dvp_microcode_loaded;
	u32 VideoStandard; 
	u32 SubPicturePalette[16];
	u32 BgSwitch;
	u32 AutoScaling;
	u32 AutoPanScan;
	u32 ScaleRatio;
	u32 VidEntry;
	u32 LastPic;
	u32 VidFloat;
	u32 SpSwitch;
	u32 SPScale;
	u32 CaptionSwitch;
	u32 video_tolerance;
	u32 video_delay;
	u32 video_sync_mode;
	u32 video_sid;
	u32 subpic_sid;
	u32 dvpgen2;
	struct video_standard_s *vs;
	struct composite_blank_s cbs;
	struct video_key ckey;

/* audio configuration section */
	u32 adp_ucode;
	u32 adp_microcode_loaded;
	u32 sr;
	u32 audio_gain;
	u32 amclk;
	u32 spbs;
	u32 audio_tolerance;
	u32 audio_delay;
	u32 audio_sync_mode;
	u32 audio_sid;

/* CSS configuration section */
#ifdef CSS_SUPPORT
	dvd_authinfo cssauth;
#endif

	spinlock_t lock;
	spinlock_t status_lock;

	struct mga_dev_s *mga;
	struct mgabus_client_s cbus;
	struct mgabus_client_vid_s vbus;
};


#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


static int debug = 0;
#define dprintk  if (debug) printk
#define dprintk2 if (debug>1) printk

/* microcode binary data stored in arrays */
static u8 dvp_dvd[] = {
#include "340edvd.inc"	/* DVD playback */
};
static u8 dvp_vcd[] = {
#include "340ev4p.inc"	/* VCD playback */
};
static u8 dvp_oth[] = {
#include "340eoth.inc"	/* MPEG audio and video playback */
};
static u8 dvp_scan[] = {
#include "340escan.inc" /* Fast Search (fast forward) support */
};
static u8 adp_ac3[] = {
#include "340eac3.inc"	/* Dolby 2.0, 2 channel audio, soundcard CD in */
};
static u8 adp_ac36[] = {
#include "340eac36.inc" /* Dolby 5.1, 6 channel audio, Marvel S/PDIF out */
};
static u8 adp_lpcm[] = {
#include "340elpcm.inc" /* DVD LPCM, 8 channel audio */
};
static u8 adp_mpeg[] = {
#include "340empeg.inc" /* MPEG-1,2 Layer II, 2 channel audio */
};
static u8 adp_cdda[] = {
#include "340ecdda.inc" /* CD Digital Audio, 2 and 4 channel audio */
};


/* checks from leon */
/* 1=reset, 2=init_pclk, 4=init_display, 8=idle, 16=playback */
static u8 state_to_stage[] = {
	1, 0, 2, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	16,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	16,0, 0, 16,0,16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
	
static u8 write_stage_perm[] =
{
	2, 0,	2, 0,	2, 0,	12,16,	12,16,	12,16,	0, 0,	0, 0,	//0x00
	2, 31,	2, 31,	2, 31,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	2, 12,	2, 12,	2, 12,	2, 12,	2, 12,	2, 31,	2, 31,	2, 31,	//0x10
	2, 31,	2, 12,	2, 12,	2, 12,	2, 12,	2, 12,	2, 12,	2, 12,
	
	2, 12,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0x20
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	0, 0,	0, 0,
	
	16,16,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0x30
	12, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	12, 0,	12,0,	12,16,	12,16,	12,16,	12,0,	12,0,	12,16,	//0x40
	12,16,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	//0x50
	12,16,	12,16,	12,16,	12,16,	12,16,	0, 0,	12,16,	12,16,
	
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0x60
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	0, 0,	12,16,
	
	16,16,	16,16,	16,16,	16,16,	16,16,	16,16,	16,16,	0, 0,	//0x70
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	//0x80
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,
	
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	//0x90
	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,	12,16,
	
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xa0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xb0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xc0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	16,16,	16,16,
	
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xd0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	28,28,	28,28,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xe0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,
	
	12,0,	12,16,	14,0,	12,0,	0, 0,	0, 0,	0, 0,	0, 0,	//0xf0
	0, 0,	0, 0,	0, 0,	0, 0,	0, 0,	12,12,	0, 0,	0, 0,
};

/* access zoran indirect registers with c-state verification
   table from pages 80 to 84 of 36710 spec (whew) just under 100 registers :>
   -inkling
*/

/* remapped cstate values to bit fields for each cstate */
#define pclk 1	/* init_pclk */
#define idsp 2	/* init_display */
#define idle 4	/* idle */
#define nspb 8	/* non-stop playback */

/* these are not used in array below, but perhaps should be,
    i.e. are they considered to be "during playback" ? -ink */
/* maybe consolidate these into one playback indication (nspb)
   so the extra bits could be used for readable/writable and others,
   but not sure how pause fits into the playback category.
   -inkling */
#define paus 16
#define step 32
#define slow 64

/* may need to make this for idle or init_display or playback modes */
#define any 128
/* special case, register can be changed at any time.
   TESTME: or is it "any time after pclk" ? */

/* zoran index registers are 00-FF, less than a hundred are actually used.
   allocate global array to index register access against allowed cstate,
   then populate the array with the allowed cstate flags for each index value.
   this method allows for fast lookup, but needs to be initialized before use.

   this should actually be stored in the shadow array prior to use of
   it making shadow 2d instead of 1d but would keep everything together;
   should be an easy change

   NOTE: initfile3 is only one that has to be written once only after reset,
	 but sysconfig vidconfig and sdconfig want to be done only after pclk;
	 rest seem to be pretty flexible but need read/write ability tagged

   -inkling
*/
static u8 zir_cs[255];

/* call this once to set up the array */
static void
i34_init_register_cstate(void)
{
	int i;
	
	for (i=0;i<=255;i++)
	{
		/* load with zeros for reserved register check */
		zir_cs[ i ] = 0;
	}

	zir_cs[ ZR_SYSCONFIG ] =	pclk;
	zir_cs[ ZR_VIDCONFIG ] =	pclk;
	zir_cs[ ZR_SDCONFIG ] =		pclk;

	zir_cs[ ZR_DVPGEN1 ] =		idsp | idle | nspb;
	zir_cs[ ZR_DVPGEN2 ] =		idsp | idle | nspb;
	zir_cs[ ZR_DVPGEN3 ] =		idsp | idle | nspb;

	zir_cs[ ZR_COLORV ] =		pclk | any;
	zir_cs[ ZR_COLORY ] =		pclk | any;
	zir_cs[ ZR_COLORU ] =		pclk | any;

	zir_cs[ ZR_VIDOUT ] =		pclk | idsp | idle;
	zir_cs[ ZR_HTOTAL ] =		pclk | idsp | idle;
	zir_cs[ ZR_VTOTAL ] =		pclk | idsp | idle;
	zir_cs[ ZR_HSYNCSIZE ] =	pclk | idsp | idle;
	zir_cs[ ZR_VSYNCSIZE ] =	pclk | idsp | idle;

	zir_cs[ ZR_CBHSTART ] =		pclk | any;
	zir_cs[ ZR_CBVSTART ] =		pclk | any;
	zir_cs[ ZR_CBHEND ] =		pclk | any;
	zir_cs[ ZR_CBVEND ] =		pclk | any;

	zir_cs[ ZR_ACTIVESTARTX ] =	pclk | idsp | idle;
	zir_cs[ ZR_ACTIVESTARTY ] =	pclk | idsp | idle;
	zir_cs[ ZR_ACTIVEENDX ] =	pclk | idsp | idle;
	zir_cs[ ZR_ACTIVEENDY ] =	pclk | idsp | idle;
	zir_cs[ ZR_ACTIVESIZEX ] =	pclk | idsp | idle;
	zir_cs[ ZR_ACTIVESIZEY ] =	pclk | idsp | idle;
	zir_cs[ ZR_CAPTIONOFFSET ] =	pclk | idsp | idle;
	zir_cs[ ZR_STARTDISPLAY ] =	pclk | idsp | idle;

	zir_cs[ ZR_OSDCONTROL ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDMEMSTART ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDMEMSIZE ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDFIRSTLINE ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDLASTLINE ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDSWITCH ] =	idsp | idle | nspb;

	zir_cs[ ZR_SCLKVALUE ] =	nspb;

	zir_cs[ ZR_DISCTYPE ] =		idsp | idle;

	zir_cs[ ZR_BITSTREAMSELECT ] =	idsp | idle;
	zir_cs[ ZR_PLAYBACKMODE ] =	idsp | idle;
	zir_cs[ ZR_AUDSID ] =		idsp | idle | nspb;
	zir_cs[ ZR_VIDSID ] =		idsp | idle | nspb;
	zir_cs[ ZR_SPSID ] =		idsp | idle | nspb;
	zir_cs[ ZR_AUDPORTDELAY ] =	idsp | idle;
	zir_cs[ ZR_VIDPORTDELAY ] =	idsp | idle;
	zir_cs[ ZR_VIDTOLERANCE ] =	idsp | idle | nspb;
	zir_cs[ ZR_VIDSYNCMODE ] =	idsp | idle | nspb;
	zir_cs[ ZR_SPPALETTE ] =	idsp | idle | nspb;

	zir_cs[ ZR_AUTOSCALING ] = 	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANBASEX ] =	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANBASEY ] =	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANSIZEX ] =	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANSIZEY ] =	idsp | idle | nspb;
	zir_cs[ ZR_PICSIZEX ] =		idsp | idle | nspb;
	zir_cs[ ZR_PICSIZEY ] =		idsp | idle | nspb;
	zir_cs[ ZR_IMAGESTARTX ] =	idsp | idle | nspb;
	zir_cs[ ZR_IMAGESTARTY ] =	idsp | idle | nspb;
	zir_cs[ ZR_IMAGESIZEX ] =	idsp | idle | nspb;
	zir_cs[ ZR_IMAGESIZEY ] =	idsp | idle | nspb;
	zir_cs[ ZR_IMAGEENDX ] =	idsp | idle | nspb;
	zir_cs[ ZR_IMAGEENDY ] =	idsp | idle | nspb;

	zir_cs[ ZR_SCALERATIO ] =	idsp | idle | nspb;
	zir_cs[ ZR_SPSCALE ] =		idsp | idle | nspb;

	zir_cs[ ZR_AUTOPANSCAN ] =	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANOFFSETX ] =	idsp | idle | nspb;
	zir_cs[ ZR_PANSCANOFFSETY ] =	idsp | idle | nspb;
	zir_cs[ ZR_BACKGROUNDSWITCH ] =	idsp | idle | nspb;
	zir_cs[ ZR_SPSWITCH ] =		idsp | idle | nspb;
	zir_cs[ ZR_CAPTIONWORD ] =	idsp | idle | nspb;

	zir_cs[ ZR_CAPTIONSWITCH ] =	idsp | idle | nspb;

	zir_cs[ ZR_HILIGHTBUTTON1 ] =	nspb;
	zir_cs[ ZR_HILIGHTBUTTON2 ] =	nspb;
	zir_cs[ ZR_HILIGHTCOLOR1 ] =	nspb;
	zir_cs[ ZR_HILIGHTTIMING ] =	nspb;
	zir_cs[ ZR_HILIGHTSWITCH ] =	nspb;
	zir_cs[ ZR_HILIGHTCOLOR2 ] =	nspb;
	zir_cs[ ZR_HILIGHTCOLOR3 ] =	nspb;

	zir_cs[ ZR_OSDPALETTE0i ] =	idsp | idle | nspb;
	zir_cs[ ZR_OSDPALETTE1i ] =	idsp | idle | nspb;

	zir_cs[ ZR_DVPCMD ] =		nspb;
	zir_cs[ ZR_DECCNTR ] =		nspb;

	zir_cs[ ZR_DYNPANX ] =		idsp | idle | nspb;
	zir_cs[ ZR_DYNPANY ] =		idsp | idle | nspb;

	zir_cs[ ZR_INITFILE_1 ] =	idsp | idle;
	zir_cs[ ZR_INITFILE_2 ] =	idsp | idle | nspb;

	/* init file 3 is only odd one: only once after reset is write ok */
	zir_cs[ ZR_INITFILE_3 ] =	idsp | pclk | idle;
	zir_cs[ ZR_INITFILE_4 ] =	idsp | idle;

	zir_cs[ ZR_MICROCODE ] =	idsp | idle;
}



/* prototypes */
int i34_enable_interrupts(struct i34_dev_s *i34, u16 zirq);
int i34_disable_interrupts(struct i34_dev_s *i34, u16 zirq);




#ifdef NEVER
static void
rgb2yuv2( int r, int g, int b, u8 *yuv)
{
	long iy, icb, icr;

	iy  =  16 + (( (16863 * r) + (32850 * g) +  (6351 * b)) >> 16);
	icr = 128 + (( (28689 * r) - (24090 * g) -  (4599 * b)) >> 16);
	icb = 128 + (( (-9636 * r) - (19053 * g) + (28689 * b)) >> 16);

	yuv[0] = iy;
	yuv[1] = icr;
	yuv[2] = icb;
}
#endif


/* Red Green Blue to Luminance DeltaChromaRed DeltaChromaBlue
   scales to 0-219 studio range 4:4:4 format, 8-bit Y +16, Cr & Cb +/- 110
   (220 cubed vs 256 cubed) 10M colors vs 16M colors
   formula should be for studio 219
 -inkling
*/
static unsigned long
RGB2YCrCb(long rgb24)
{
	int r, g, b;
	long iy, icb, icr;

	r = (rgb24 & 0xFF0000) >> 16;
	g = (rgb24 & 0xFF00) >> 8;
	b = (rgb24 & 0xFF);

	iy  =  16 + (( (16863 * r) + (32850 * g) +  (6351 * b)) >> 16);
	icr = 128 + (( (28689 * r) - (24090 * g) -  (4599 * b)) >> 16);
	icb = 128 + (( (-9636 * r) - (19053 * g) + (28689 * b)) >> 16);

	return (iy&255)<<16 | (icr&255)<< 8 | (icb&255);
}


/* initialize default sub-picture palette in RGB then convert to YCrCb */
static void
i34_init_subpicture_palette(struct i34_dev_s *i34)
{
	int i;

	dprintk("i34: init_subpicture_palette\n");

	/* load subpicture (subtitle) palette with RGB values
	   bg is background (all black except 15), fg is foreground WRGBCMYK
	   [0] is white on black normal,  [15] is black on white inverse
	*/
	i34->SubPicturePalette[0]  = 0x000000; /* bg black	*/
	i34->SubPicturePalette[1]  = 0xFFFFFF; /* fg white	*/
	i34->SubPicturePalette[2]  = 0x000000; /* bg black	*/
	i34->SubPicturePalette[3]  = 0xFF0000; /* fg red	*/
	i34->SubPicturePalette[4]  = 0x000000; /* bg black	*/
	i34->SubPicturePalette[5]  = 0x00FF00; /* fg green	*/
	i34->SubPicturePalette[6]  = 0x000000; /* bg black	*/
	i34->SubPicturePalette[7]  = 0x0000FF; /* fg blue	*/
	i34->SubPicturePalette[8]  = 0x000000; /* bg black	*/
	i34->SubPicturePalette[9]  = 0x00FFFF; /* fg cyan	*/
	i34->SubPicturePalette[10] = 0x000000; /* bg black	*/
	i34->SubPicturePalette[11] = 0xFF00FF; /* fg magenta	*/
	i34->SubPicturePalette[12] = 0x000000; /* bg black	*/
	i34->SubPicturePalette[13] = 0xFFFF00; /* fg yellow	*/
	i34->SubPicturePalette[14] = 0x000000; /* fg black	*/
	i34->SubPicturePalette[15] = 0xFFFFFF; /* bg white	*/

	/* convert palette to YCrCb and swap 16 bit words (swapw not swapb)
	   NOTE: skip the swap step for big endian
	   these color values may be overridden by ioctl data from user DVD
	   play, but should not be overwritten in case needed later
	   word swap is done when palette is actually loaded
	   heh just occurred to me it might swap back if load_subpalette
	   called twice
	 -inkling
	*/

	/* convert from easier to read RGB24 to YCrCb219 */
	for (i=0;i<16;i++)
	{
	    i34->SubPicturePalette[i] = RGB2YCrCb(i34->SubPicturePalette[i]);
	    // dprintk2("i34: palette[%1X] = %08X \n",i,j);
	}
	
}


static int
i34_exec_commands(struct mga_codec_s *codec, u8 and, u8 or)
{
	/*
	 * FIXME: horrible hack to make mgadvd work !!!!
	 * without this, the commands hanged...
	   lost acknowledge values in write/read combo's
	 */
	/* still needed? to set which zoran to use in marvel/rrg setup? */
	mgacodec_push_readw(codec, 0x7);		/* ZR_DEVICE_ID */

	return mgacodec_exec_commands(codec, and, or);
}


static u16
i34_read(struct i34_dev_s *i34, u8 reg)
{
	mgacodec_push_readw(i34->codec_interface, reg);
	if (i34_exec_commands(i34->codec_interface, ~0, 0) != 0) {
		dprintk( "i34_exec_commands error\n" );
	}
	return mgacodec_get_word(i34->codec_interface);
}


static int
i34_state(struct i34_dev_s *i34, int *s)
{
	*s = (i34_read(i34, ZR_STATUS0) >> 10);
	return 0;
}


/* check index register access against current cstate
   returns:
     0 if ok,
     required c-state flags if not ok (with dprintk),
     0xFF if reserved register attempted (with dprintk)
 -inkling
*/
static int
i34_check_register_cstate(struct i34_dev_s *i34, u8 index_register)
{
	unsigned char cstate_flag;
	unsigned int cstate;
	unsigned int index;
	
	/* get current cstate */
	i34_state(i34, &cstate);
	
	/* dont care about the active decoded image flag */
	cstate &= 0x37;

	/* range limit index_register so no out of bounds on array */
	index = index_register & 0xFF;

	/* translate current cstate into bit flags */
	switch(cstate)
	{
		case ZR_CSTATE_INIT_PCLK:
			cstate_flag = pclk;
			break;

		case ZR_CSTATE_INIT_DISPLAY:
			cstate_flag = idsp;
			break;

		/* playback modes, should all be equivalent? */
		case ZR_CSTATE_PAUSE:
		case ZR_CSTATE_STEP:
		case ZR_CSTATE_SLOW:
		case ZR_CSTATE_NSPB:
			cstate_flag = nspb;
			break;

		case ZR_CSTATE_IDLE:
			cstate_flag = idle;
			break;

		default:
			dprintk("i34: bad check_reg_cs %02X ireg %02X\n",
				cstate, index);
			/* FIXME: error or needs more checking? */
			return 0xFF;
			break;
	}

	/* return error if reserved/undefined register access */
	if ( zir_cs[ index ] == 0)
	{
		dprintk("i34: check_reg_cs reserved ireg %02X\n", index);
		return 0xFF;
	}

	/* check for any flag and return ok if set */
	/* FIXME: might need additional check for 'after init_pclk' */
	if ( zir_cs[ index ] & any )
		return 0;

	/* check cstate_flag against allowed for register access */
	if ( zir_cs[ index ] & cstate_flag )
		return 0;

	/* otherwise log error and return the flags needed */
	dprintk("i34: ireg %02X cstate %02X, needs cstate(s): ",
		 index, cstate );
	if ( zir_cs[ index ] & pclk ) dprintk("pclk ");
	if ( zir_cs[ index ] & idsp ) dprintk("idsp ");
	if ( zir_cs[ index ] & idle ) dprintk("idle ");
	if ( zir_cs[ index ] & nspb ) dprintk("nspb ");
	dprintk("\n");

	/* return flags that needed to be set for this to work
	 *  zero if it worked (above). FIXME: may need only non-zero
	 */
	return zir_cs[ index ];
}		


static void
i34_check_state_configured( struct i34_dev_s *i34 )
{
	int cstate;
	int stage;
	int i;
	u8 p;
	
	i34_state(i34, &cstate);
	
	stage = state_to_stage[ cstate & 0x37 ];
	
	for (i=0; i<256; i++) {
		if (i34->shadow_register[ i ] == -1) {
			/* unwritten param; check if should have been
			  written before leaving cstate */
	
			p = write_stage_perm[ i*2 ];
			if ((p & stage)==0)
				continue; /* didn't have to be written */
			
			/* check if could also be written in later state */
			if ((p & ~stage) > stage)
				continue;
			dprintk2( "unconfigured reg %02X in %02x\n", i, cstate );
		}
	}
}


static int
i34_check_state_perm( struct i34_dev_s *i34, u8 addr, int un_inited)
{
	int cstate;
	int stage;
	
	return 0; /* this code is somewhat redundant now -ink */

	i34_state(i34, &cstate);
	
	stage = state_to_stage[ cstate & 0x37 ];
	if (write_stage_perm[ addr*2 + (un_inited?0:1) ] & stage)
		return 0;
	dprintk( "i34: check_state_perm writing param 0x%02X in wrong state 0x%02X\n", addr, cstate );
	return 1;
}
	

static int
i34_check_stage_eq( struct i34_dev_s *i34, int eq_stage )
{
	int cstate;
	int stage;
	
	i34_state(i34, &cstate);
	stage = state_to_stage[ cstate & 0x37 ];
	if ( stage & eq_stage )
		return 0;
	dprintk( "i34_check_stage_eq: wrong state %i\n", cstate );
	return 1;
}


static int
i34_write(struct i34_dev_s *i34, u8 reg, u16 data)
{
	mgacodec_push_writew(i34->codec_interface, reg, data);
	if (i34_exec_commands(i34->codec_interface, ~0, 0) != 0) {
		dprintk( "i34_exec_commands error\n" );
		return 1;
	}
	return 0;
}


#ifdef NEVER
static u16
i34_read_parameter(struct i34_dev_s *i34, u8 addr)
{
	mgacodec_begin_commands(i34->codec_interface);
	/* see table 32 p. 79 for the protocol to access the indirect address space
	   addr. in the l.s. bits, m.s. bits must be 0
	*/
	mgacodec_push_writew(i34->codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	mgacodec_push_readw(i34->codec_interface, ZR_PARAM_DATA);
	i34_exec_commands(i34->codec_interface, ~0, 0);
	/* we need shadow registers, because most of them are write only...
	 especially when dealing with and/or operations
	*/
	i34->shadow_register[addr] = mgacodec_get_word(i34->codec_interface);
	
	mgacodec_end_commands(i34->codec_interface);
	return i34->shadow_register[addr];
}
#endif


static int
i34_write_parameter(struct i34_dev_s *i34, u8 addr, u16 data)
{
	///////
	i34_check_state_perm(i34, addr, i34->shadow_register[addr] == -1);
	///////

	
	if ( i34_check_register_cstate(i34, addr) != 0) {
	dprintk("i34: writing to ireg %02X in wrong cstate?\n", addr);
	}


	i34->shadow_register[addr] = data;
	
	mgacodec_push_writew(i34->codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	mgacodec_push_writew(i34->codec_interface, ZR_PARAM_DATA, data);
	if (i34_exec_commands(i34->codec_interface, ~0, 0) != 0) {
		dprintk( "i34_exec_commands error\n" );
		return 1;
	}
	return 0;
}

/* change some bits of a parameter register.
   this could also be named i34_parameter_and_or...
*/
static int
i34_change_parameter(struct i34_dev_s *i34, u16 addr, u16 or, u16 and)
{
	u16 value;
	if ( i34->shadow_register[addr] == -1 ) {
		dprintk( KERN_ERR "i34: changing unwritten parameter 0x%02X\n", addr );
		i34->shadow_register[addr] = 0;
	}
	value = (i34->shadow_register[addr] & ~and) | or;
	return i34_write_parameter(i34, addr, value);
}


#ifdef NEVER	
static int
i34_write_seq(struct i34_dev_s *i34, u8 addr, u16* buffer, u32 size_in_words)
{
	u32 i;

	if (buffer == NULL) return 1;

	mgacodec_begin_commands(i34->codec_interface);

	for (i=0; i<size_in_words; i++) {
		mgacodec_push_writew(i34->codec_interface, addr, *buffer++);
		i34_exec_commands(i34->codec_interface, ~0, 0);
	}
	
	mgacodec_end_commands(i34->codec_interface);
	return 0;
}
#endif


static int
i34_write_parameter_seq(struct i34_dev_s *i34, u8 addr, u16* buffer, u32 size_in_words)
{
	u32 i;


	if (buffer == NULL) return 1;


	if ( i34_check_register_cstate(i34, addr) != 0 ) {
	dprintk("i34: writing seq to ireg %02X in wrong cstate?\n", addr);
	}


	/* multiple read/writes without modifying the Parameter Address Register */
	mgacodec_push_writew(i34->codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	
	for (i=0; i<size_in_words; i++) {
		mgacodec_push_writew(i34->codec_interface, ZR_PARAM_DATA, *buffer++);
		if (i34_exec_commands(i34->codec_interface, ~0, 0) != 0) {
			dprintk( "i34_exec_commands error\n" );
		}
	}
	return 0;

}


/* read the System Clock */
static u32
i34_read_sclk(struct i34_dev_s *i34)
{	
	u16 a, b;
	unsigned long flags;

	spin_lock_irqsave( &i34->lock, flags );

	a = i34_read(i34, ZR_SYS_CLOCK);
	b = i34_read(i34, ZR_SYS_CLOCK);

	spin_unlock_irqrestore( &i34->lock, flags );

	return (a<<16 | b);
}


/* wait 'delay' microseconds */	/*fixme: microsec. resolution is too optimistic */
static void
i34_wait(int delay)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout( ((delay * HZ)+500000)/1000000 );
}


static void
i34_reset_struct(struct i34_dev_s *i34)
{
	int i;

	/* defaults for parameter registers */
	for (i=0; i<sizeof(i34->shadow_register)/sizeof(i34->shadow_register[0]); i++)
		i34->shadow_register[i] = -1;

	i34->CodeBurstLen = 64;
	i34->init_file_loaded = 0;
	
	i34->dvp_microcode_loaded = DVP_MICROCODE_NONE;
	
	i34->video_sid = 0;
	i34->subpic_sid = 0;
	
	i34->adp_microcode_loaded = ADP_MICROCODE_NONE;
	
	i34->audio_sid = 0;

	i34->imr = 0xFFFF; /* no interrupts by default */

	i34_init_subpicture_palette(i34); /* set up palette to defaults */
	i34_init_register_cstate(); /* set up cstate vs indirect registers */


}



/*	see 36710 spec p245-246, section 13.2 for timing information
	on how to properly do reset and standby

	miscctl bit 5:	0-> ZR36700.RESET asserted
	miscctl bit 6:	0-> ZR36700.STNDBY asserted
	miscctl bit 7:	1-> ZR36700.HCS asserted
	
	GCLK/VDD/RESET must be have been stable for 30ms (given)
	HCS asserted during any change to RESET or STANDBY

	RESET:
		starts 16 GCLK after RESET asserted
		PLL reset 128 GCLK after RESET assert
		ends 16 GCLKS after RESET de-assert
	
	STNDBY within RESET:
		assert STNDBY after RESET asserted for >160GCLK
		de-assert STNDBY >32GCLK before RESET de-assert

	STNDBY+RESET:
		tri-state and enter low power mode
		no timing information given on this mode,
		so assuming it's the same as STNDBY within RESET
*/

#define RRG_MISCCTL_MASK 0x1F

static void
i34_powerdown(struct i34_dev_s *i34)
{
	struct mga_codec_s *codec = i34->codec_interface;

/* NOBUZZ leaves zoran powered and outputs enabled, no floating input buzz */
/* this shouldn't hurt reset routines; powerup does powerdown anyway */
// #define NOBUZZ
#ifdef NOBUZZ
	return;
#endif
	dprintk("i34: powerdown\n");

	/* NOTE: keep HCS asserted, (bit 7=1) throughout this process!
	   no chip select and it'll ignore everything else you tell it.
	*/

	/* FIXME: need to set/clear appropriate Marvel/RRG bits here
	   for and+or as well, so you don't trigger it by accident, 
	   and then use RRG_MISCCTL_MASK defined above to not change
	   them once they've been set/reset. it's probably way too much
	   overkill to do Marvel/RRG disable/mask this way, but wtf.  :>
	  -inkling
	*/

	/* assert HCS, deassert RESET, deassert STNDBY */
	mgacodec_and_or_miscctl(codec,
				 0x00 | RRG_MISCCTL_MASK,
				 0xE0 | 0x00 ); /* OR the RRG disable here */
	udelay(10);

	/* assert RESET (bit 5=0) and wait >160 GCLK */
	mgacodec_and_or_miscctl(codec,
				 0xC0 | RRG_MISCCTL_MASK,
				 0x00 & ~RRG_MISCCTL_MASK);
	udelay(10);

	/* assert STNDBY (bit 6=0), and wait a bit */
	mgacodec_and_or_miscctl(codec,
				 0x80 | RRG_MISCCTL_MASK,
				 0x00 & ~RRG_MISCCTL_MASK);
	udelay(10);

	/* deassert HCS after entering low power mode (about 116 milliWatt) */
	mgacodec_and_or_miscctl(codec,
				 0x7F | RRG_MISCCTL_MASK,
				 0x00 & ~RRG_MISCCTL_MASK);
	udelay(10);

}


static void
i34_powerup(struct i34_dev_s *i34)
{
	struct mga_codec_s *codec = i34->codec_interface;


#ifdef NEVER
	/* program the MISC<2> and DDC<2> lines
	   DDC<2> is connected to HRDY according to the G100 specs	   
	*/
	mgacodec_set_misc2(codec, 1);
	mgacodec_set_ddc2(codec, 1);
#endif

	dprintk("i34: powerup\n");

	/* assert HCS, RESET & STNDBY to enter tri-state/low power mode
	   this is a known state and the best place to start from.
	   duplicating part of powerdown code here instead of
	   trying to slam it into powerdown w/o proper timings
	*/

	/* FIXME: need to set/clear appropriate Marvel/RRG bits here
	   for and+or as well, so you don't trigger it by accident, 
	   and then use RRG_MISCCTL_MASK defined above to not change
	   them once they've been set/reset. it's probably way too much
	   overkill to do Marvel/RRG disable/mask this way, but wtf.  :>
	  -inkling
	*/

	/* assert HCS, deassert RESET, deassert STNDBY to power up,
	   tri-state/disable marvel/rrg (not done yet),
	   then wait for >30ms, see p248, 14.1 before asserting RESET
	*/
	mgacodec_and_or_miscctl(codec,
				 0x00 | RRG_MISCCTL_MASK,
				 0xE0 | 0x00); /* OR the RRG disable here */
	i34_wait(40*1000);

	/* assert RESET (bit 5=0) and wait >160 GCLK (5.93us) */
	mgacodec_and_or_miscctl(codec,
				 0xC0 | RRG_MISCCTL_MASK,
				 0x00 & ~RRG_MISCCTL_MASK);
	udelay(10);

	/* assert STNDBY (bit 6=0), stay tri-stated for a while */
	mgacodec_and_or_miscctl(codec,
				 0x80 | RRG_MISCCTL_MASK,
				 0x00 & ~RRG_MISCCTL_MASK);
	udelay(20);

	/* deassert STNDBY (bit 6=1) and wait >32 GCLK (1.19usec) */
	mgacodec_and_or_miscctl(codec,
				 0x80 | RRG_MISCCTL_MASK,
				 0x40 & ~RRG_MISCCTL_MASK);
	udelay(3);

	/* deassert RESET (bit 5=1) and wait >16 GCLK (0.593usec) */
	mgacodec_and_or_miscctl(codec,
				 0xC0 | RRG_MISCCTL_MASK,
				 0x20 & ~RRG_MISCCTL_MASK);
	udelay(1);

	/* wait for PLL's to stabilize. maybe could be adjusted shorter.
	   do it here and now to shorten i34_wait_for_pll loops. */

	i34_wait(40*1000);

	/* HCS stays asserted after full power up (about 1.1 Watt) */

//	mga_test_lines(i34->mga, 1);
}


static void
i34_codec_callback(struct mga_dev_s *mga, int state)
{
	if (state == MGACODEC_STOP)
		mgacodec_stall_transfer(mga->codec);
}


static int
i34_cbus_set_dir( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data )
{
	struct i34_dev_s *i34 = data;
	struct mga_codec_s *codec = i34->codec_interface;

	dprintk( "i34_cbus_set_dir %i\n", dir );
	
	switch ( dir ) {
		case BUSDIR_SOURCE:
			mgacodec_callback(codec, i34_codec_callback);
			break;
		
		case BUSDIR_TRISTATE:
			mgacodec_callback(codec, NULL);
			i34_powerdown(i34);
			break;
		
		case BUSDIR_SINK:
		default:
			return 1;
	}
	return 0;
}


static enum bus_direction
i34_cbus_get_dir( struct mgabus_s *bus, void *data )
{
	struct i34_dev_s *i34 = data;

	/* fake the return */
	if ( i34->mga->codec_bus->source == &i34->cbus ) {
		dprintk( "i34_cbus_get_dir=%i\n", BUSDIR_SOURCE );
		return BUSDIR_SOURCE;
	}
	dprintk( "i34_cbus_get_dir=%i\n", BUSDIR_TRISTATE );
	return BUSDIR_TRISTATE;
}



static int
i34_vbus_set_dir( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data )
{
	struct i34_dev_s *i34 = data;

	dprintk( "i34_vbus_set_dir %i\n", dir );

	switch ( dir ) {
		case BUSDIR_SOURCE:
			/* nop */
			break;
		
		case BUSDIR_SINK:						
			if ( addr != BUS_ADDR_BCAST )
				return 1;
			/* fallthrough */
		case BUSDIR_TRISTATE:
			i34_powerdown(i34);
			break;
		
		default:
			return 1;
	}
	return 0;
}


static enum bus_direction
i34_vbus_get_dir( struct mgabus_s *bus, void *data )
{
	struct i34_dev_s *i34 = data;

	/* fake the return */
	if ( i34->mga->vip_bus->source == &i34->vbus.super ) {
		dprintk( "i34_vbus_get_dir=%i\n", BUSDIR_SOURCE );
		return BUSDIR_SOURCE;
	}
	dprintk( "i34_vbus_get_dir=%i\n", BUSDIR_TRISTATE );
	return BUSDIR_TRISTATE;
}


			


static int
i34_vbus_get_format( struct mgabus_s *bus, int input,
			struct mgabus_video_input_s *params, void *data )
{
	struct i34_dev_s *i34 = (struct i34_dev_s*)data;

dprintk( "i34 vbus_get_format %i (%p) -> %p\n", input, i34, params );

	switch (i34->VideoStandard) {
			
		case VIDEO_MODE_PAL:
			params->norm = KS_STD_PAL;
			break;

		case VIDEO_MODE_NTSC:
		default:
			params->norm = KS_STD_NTSC;
			break;
	}

	params->brightness = 1;
	params->hue = 2;
	params->contrast = 3;
	params->color = 4;

	params->crop.x = 0;
	params->crop.y = 0;
	if (i34->vs != NULL) {
		params->crop.width = i34->vs->ActiveSizeX;
		params->crop.height = i34->vs->ActiveSizeY * 2;

		params->vblcnt0 = i34->vs->VSyncSize; //fixme: check on G400??
		params->vblcnt1 = i34->vs->VSyncSize; //
		params->vbicnt0 = 0x00;
		params->vbicnt1 = 0x00;
	}
	else {
		params->crop.width = 0;
		params->crop.height = 0;
		params->vblcnt0 = 0;
		params->vblcnt1 = 0;
		params->vbicnt0 = 0x00;
		params->vbicnt1 = 0x00;
	}
	
	return 0;
}



static int
i34_wait_for_pll(struct i34_dev_s *i34)
{
	u16 i, isr;

	for (i=0; i<5000; i++) {
		isr = i34_read(i34, ZR_ISR);
		if ((isr & ( ISR_PLLpLOCKED | ISR_PLLaLOCKED ))
			== ( ISR_PLLpLOCKED | ISR_PLLaLOCKED ))
		{
			/* PLLp & PLLa are locked */
			dprintk("PLL's locked, status 0x%04X loops %d\n", isr, i);
			
			/* what does it do? --lvs */
			/*
			i34_write_parameter(i34, ZR_INITFILE_4, 0xA0);
			i34_write_parameter(i34, ZR_INITFILE_4, 0x88);
			*/
			return 0;
		}
		i34_wait(100);
	}
	dprintk("PLL's NOT locked, status 0x%04X\n", isr);
	return 1;
}


static int
i34_reset_internal(struct i34_dev_s *i34, int flag)
{
	dprintk("i34: Reset\n");

	/* i34 will also take over videobus when enabled; make
	sure video bus is free for us */
	dprintk("i34: setting video source\n");
	if ( mgabus_set_source( i34->mga->vip_bus, &i34->vbus.super, BUS_ADDR_BCAST ) ) {
		dprintk( "i34: failed to set video source\n" );
		return 1;
	}
	
	dprintk("i34: setting codec source\n");
	if ( mgabus_set_source( i34->mga->codec_bus, &i34->cbus, 1 ) ) {
		dprintk( "i34: failed to set codec source\n" );
		return 1;
	}

	/* reset codec interface to correct protocol */
	mgacodec_init(i34->codec_interface, MGACODEC_MODE_VMI_A, ZR_BITSTREAM,
		(i34->mga->options & MGAVIDEO_HAS_RRG) != 0 ? 0x07 : 0 );

	i34_powerup(i34);

//	i34_enable_interrupts(i34, 0); /* load irqs from list of currently enabled */
	
	i34_reset_struct(i34);

	/* seems to need some delay here to settle down */
	i34_wait(500);
	
	/* accesses to zr need to be on word basis; it might be/get in a confused
	state when accessing it for the very first time (ie., miss the first byte
	and get stuck between bytes in every following access). bitstream reg can
	handle byte accesses, so write something to it
	--lvs (that's at least what i think happens; fixes bad detects anyway) */
	i34_write(i34, ZR_BITSTREAM, 0);

	if (flag) return 0;

	if (i34_wait_for_pll(i34) != 0)
		return 1;
		
	return 0;
}


static int
adp_read(struct i34_dev_s *i34, u8 *data)
{
	u16 status0;

	/* see table 48 p. 115 */
	status0 = i34_read(i34, ZR_STATUS0);
	if (!(status0 & ADPRDY))
	{
		dprintk( "adp_read wait\n" );
		i34_wait(200);
		status0 = i34_read(i34, ZR_STATUS0);
		if (!(status0 & ADPRDY))
		{
			dprintk("i34: Unable to read ADP register (!ADPRDY)\n");
			return 1;
		}
	}

	*data = (i34_read(i34, ZR_ADP_DATA) & 0xFF);
	// dprintk( "adp_read %02X\n", *data );
	return 0;
}


static int
adp_write(struct i34_dev_s *i34, u8 data)
{
	u16 status0;
	int i;
	
	for (i=0; i<4; i++ ) {
		status0 = i34_read(i34, ZR_STATUS0);
		if (status0 & ADPEMPTY) {
			i34_write(i34, ZR_ADP_DATA, (data & 0x00FF));
//dprintk( "adp_written %02X\n", data );
			return 0;
		}
dprintk( "adp_write wait\n" );
		i34_wait(200);
	}
	dprintk("i34: Unable to write ADP register (!ADPEMPTY)\n");
	return 1;
}


/* clean the pending bytes waiting in the ADP Data Register
   used to start from a clean command/read pipe, because of
   the offset between commands sent and response byte
   (off by 2)
*/
static int
adp_clean(struct i34_dev_s *i34)
{
	u16 status0;
	u8 istatus1;
	int i;

	/* empty ADP buffered answers */
	
	for (i=0; i<128; i++) {
		status0 = i34_read(i34, ZR_STATUS0);
		if (status0 & ADPRDY) {
			adp_read(i34, &istatus1);
			dprintk("adp clean %02X\n", istatus1 );
		} else {
			i34_wait( 200 );
			status0 = i34_read(i34, ZR_STATUS0);
			if (!(status0 & ADPRDY)) {
				break;
			}
		}
	}
	
	return 0;
}


static int
adp_handle_error(struct i34_dev_s *i34, u8 opcode_or_ext, u8 error_code, u8 awaited)
{
	/* no need to go further if the adp returned what we were waiting for */
	if (error_code == awaited) return 0;
	if ( (opcode_or_ext==ADP_PARAM) && (error_code<awaited) ) return 0;
	dprintk( "adp: awaited x%02X but got x%02X\n", awaited, error_code );
	
	switch(error_code) {
	case ADP_NO_ERROR:
		return 0;
		break;

	case ADP_OPCODE_ERROR:
		dprintk("i34: ADP opcode error (opcode 0x%02X)\n", opcode_or_ext);
		break;

	case ADP_EXT_ERROR:
		dprintk("i34: ADP EXT param error (EXT 0x%02X)\n", opcode_or_ext);
		break;

	case ADP_NOT_READY_ERROR:
		dprintk("i34: ADP not ready for new command (opcode 0x%02X)\n", opcode_or_ext);
		break;

	case ADP_RESERVED_ERROR:
		dprintk("i34: ADP reserved error (opcode 0x%02X)\n", opcode_or_ext);
		break;

	case ADP_READY_FOR_NEW_CMD:
		/* FIXME: is it really an error ? */
		dprintk("i34: ADP ready for new command (opcode 0x%02X)\n", opcode_or_ext);
		break;

	default:
		dprintk("i34: ADP count mismatch, 0x%02X instead of 0x%02X after opcode 0x%02X)\n",
			error_code, awaited, opcode_or_ext);
	}
	adp_clean(i34);
	return 1;
}


static int
adp_simple_command(struct i34_dev_s *i34, u8 opcode)
{
	u8 dummy, istatus;

	/* same as below but we don't need to check for a parameter count 
	   during the write sequence: just make sure that we get ADP_NO_ERROR
	   at the end
	*/

	adp_write(i34, opcode);
	adp_read(i34, &dummy);
	adp_write(i34, ADP_READ);
	adp_read(i34, &istatus);
//dprintk( "adp simple %02x %02x\n", opcode, istatus );
	if (adp_handle_error(i34, opcode, istatus, ADP_NO_ERROR) != 0)
		return 1;

	return 0;
}


static int
adp_exec_command(struct i34_dev_s *i34, u8 opcode, u8 *buffer, int param_len, int resp_len)
{
	u8 count, istatus;
	int index = 0;

	/* see p. 115 & 204 */	
	adp_write(i34, opcode);		/* send opcode */
	adp_read(i34, &count);
//dprintk( "adp %02x, w %02x %02x %02x (%02x)\n", count, opcode, param_len, resp_len,
//	opcode==ADP_PARAM?buffer[0]:0xff
//);
	while (index < param_len) {
		/* check if number of awaited parameters matches the param_len */
		if (adp_handle_error(i34, opcode, count, param_len-index+1) != 0)
			return 1;
		adp_write(i34, buffer[index]);
		adp_read(i34, &count);		/* count of bytes left to send (+1) */
		index++;
	}
	/* just before reading the response bytes the count MUST be 1 */
	if (adp_handle_error(i34, opcode, count, 1) != 0)
		return 1;
	index = 0;
	while (index < resp_len) {
		adp_write(i34, ADP_READ);
		adp_read(i34, &buffer[index++]);
	}
//	adp_write(i34, ADP_READ);
	adp_write(i34, ADP_NOP);
	adp_read(i34, &istatus);		/* reading ISTATUS */

	if (adp_handle_error(i34, opcode, istatus, ADP_NO_ERROR) != 0)
		return 1;

	return 0;
}


/* configure the direction of the GPAI/O pins (see p. 231)
  iocmask: which bits to modify
  iocbits: new value for bits (0=input, 1=output)
*/
static int
adp_config_gpaio(struct i34_dev_s *i34, u8 iocmask, u8 iocbits)
{
	u8 adp_buffer[4];

	adp_buffer[0] = iocmask & 0x3f;
	adp_buffer[1] = iocbits & 0x3f;
	adp_buffer[2] = 0;
	adp_buffer[3] = 0;
	return adp_exec_command(i34, ADP_SETIO, adp_buffer, 4, 2);
}

/* set the signal level of the GPAI/O pins (see p. 231)
  iomask: which bits to modify
  iobits: new value for bits (0=low, 1=high)
*/
static int
adp_set_gpaio(struct i34_dev_s *i34, u8 iomask, u8 iobits)
{
	u8 adp_buffer[4];

	adp_buffer[0] = 0;
	adp_buffer[1] = 0;
	adp_buffer[2] = iomask & 0x3f;
	adp_buffer[3] = iobits & 0x3f;
	return adp_exec_command(i34, ADP_SETIO, adp_buffer, 4, 2);
}


static int
adp_rom_version(struct i34_dev_s *i34)
{
	u8 istatus, dummy;
	u8 adp_buffer[4];

	adp_write(i34, ADP_VER);
	adp_read(i34, &dummy);
	adp_write(i34, ADP_READ);
	adp_read(i34, &adp_buffer[0]);
	adp_write(i34, ADP_READ);
	adp_read(i34, &adp_buffer[1]);
	adp_write(i34, ADP_READ);
	adp_read(i34, &adp_buffer[2]);
	adp_write(i34, ADP_READ);
	adp_read(i34, &adp_buffer[3]);
	adp_write(i34, ADP_READ);
	adp_read(i34, &istatus);
	if (istatus != ADP_NO_ERROR) {
		dprintk("i34: adp_rom_version() error 0x%02X\n", istatus);
		return 1;
	}
	dprintk("i34: ADP ROM Version: 0x%02X%02X%02X%02X\n",
		adp_buffer[0],
		adp_buffer[1],
		adp_buffer[2],
		adp_buffer[3]);
	return 0;
}


/*
	The math used to compute a PERFECT pll_tab is the follow method:
	givens:
		27000000 is GCLK1
		256 comes from the setting of pin 6 of the AD1857
		fs(x) is the set of supported sample rates
		fs(x)		= { 48000, 44100, 32000, 96000 }
	math:
		Greatest Common Divisors of 27000000 and ( 256 * fs(x) ):
		  gcdfs(x)	= { 24000, 14400, 8000, 24000 }

		divide 27000000 by gcds(fs(x)) to get the AUDD
		  audd(x)	= { 1125, 1875, 3375, 1125 }

		divide (fs(x) * 256) by gcds(fs(x)) to get AUDM.
		  audm(x)	= { 512, 784, 1024, 1024 }
	proof:
		no rounding errors

	it's not worth an algorithm, so it's a table.
 -inkling
*/

/* SR bits on p229 */
int sr_tab[] = { 48000,
		 44100,
		 32000,
		 96000
};

/* pll_tab has audm & audd in same order as sr_tab, see above
   audd=3375 is right at the allowed limit for audd */
int pll_tab[] = {
			512, 1125,	/* 48000 */
			784, 1875,	/* 44100 */
			1024, 3375,	/* 32000 */
			1024, 1125	/* 96000 */
};

static int
adp_plltab(struct i34_dev_s *i34)
{
	u32 audm, audd, dspm, dspd, hz, sri;
	u8 adp_buffer[6];

	dprintk("i34: ADP_PLLTAB\n");
	
	i34_check_stage_eq(i34, 2);

	/* see p. 36 & 228
	   default values after reset:
	   DSPM = 3
	   DSPD = 1
	*/
	dspm = 3;
	dspd = 1;

	for (sri=0; sri<4; sri++) {
		hz = sr_tab[sri];
		audm = pll_tab[sri*2];
		audd = pll_tab[(sri*2)+1];
	
		adp_buffer[0] = (sri << 5) | ((audd & 0x1F00) >> 8);
		adp_buffer[1] = (audd & 0x00FF);
		adp_buffer[2] = (audm & 0x1F00) >> 8;
		adp_buffer[3] = (audm & 0x00FF);
		adp_buffer[4] = dspd & 0x3f;
		adp_buffer[5] = dspm;

		// dprintk("i34:  %dHz * %d == 27M * %4d/%d\n", hz, 256, audm, audd);

		adp_exec_command(i34, ADP_PLLTAB, adp_buffer, 6, 0);
	}
	/* relock aPLL */
	return i34_wait_for_pll(i34);
}

/* p 229, PLLCFG command must be done before CFG command, but there is no
   restriction on doing it again after the CFG command, as long as you don't
   reset PCLK PLL more than once after reset. this code doesn't reset PCLK PLL,
   since we allow the 36700 to use the default settings for that, i.e.
   the DSPM/DSPD in PLLTAB are the same as the power up defaults, see p 230.
*/
static int
adp_pllcfg(struct i34_dev_s *i34, u8 sr)
{
	int pa, pd;
	u8 adp_buffer[1];
	
	/* check stage not necessary,
	   sample rate can be changed on the fly
	   -ink */
	// i34_check_stage_eq(i34, 2|16);
	
	dprintk("i34: ADP_PLLCFG(%d)\n", sr);

	adp_buffer[0] = (sr & 7)<<5;
	/* relock AMCLK unless SR=111b (p. 38) */
	/* SR=111b doesn't seem to work, at least not with mpeg-1 av stream
	   either a bug or feature missing from 36700 or bad programming :>
	   -ink */
	// if (sr != 7)
		adp_buffer[0] |= 2;

	if (adp_exec_command(i34, ADP_PLLCFG, adp_buffer, 1, 1) != 0) {
		dprintk("i34: error during ADP_PLLCFG\n");
		/* don't exit now: try to get a stable clock first !!! */
		i34_wait(50);
		i34_wait_for_pll(i34);
		return 1;
	}
	pa = adp_buffer[0] & 0x02;
	pd = adp_buffer[0] & 0x01;
	dprintk("i34: AMCLK %s; PCLK %s\n",
		pa ? "locked" : "not locked",
		pd ? "locked" : "not locked");
	
	i34_wait(50);
	i34_wait_for_pll(i34);

	return 0;
}


static int
adp_config_serial(struct i34_dev_s *i34, struct serial_port_config_s *config)
{
	u8 adp_buffer[8];

	dprintk("i34: ADP_CFG\n");
	i34_check_stage_eq(i34, 2);

	adp_buffer[0] = 0xc0;
	adp_buffer[1] = 0x18;
	adp_buffer[2] = (config->fra & 7) | (config->frb & 7)<<3 | (config->cpa & 1)<<6 | (config->cpb & 1)<<7;
	adp_buffer[3] = (config->osp & 1) | (config->isp & 1)<<1 | (config->err & 1)<<4;
	adp_buffer[4] = (config->outw & 3)<<1 | (config->inw & 3)<<3 | (config->spo & 1)<<5;
	adp_buffer[5] = 0;

	/* we assume AMCLK = 256 (see table 29 p. 73) */
	/*
	switch (i34->amclk) {
	case 256:
		if (config->frb == 0)
			config->spbs = 4;
		else 
			config->spbs = 2;
		break;
	case 384:
		if (config->frb == 0)
			config->spbs = 6;
		else if (config->frb == 1)
			config->spbs = 3;
		else
			config->spbs = 4;
		break;
	default:
		dprintk("i34: invalid AMCLK\n");
		return 0;
	}
*/
	config->spbs = i34->spbs;

	adp_buffer[6] = config->spbs;
	adp_buffer[7] = (config->fma & 7) | (config->fmb & 7)<<3;

	if (adp_exec_command(i34, ADP_CFG, adp_buffer, 8, 0) != 0)
		return 1;

	return 0;
}


static int
adp_config_spdif(struct i34_dev_s *i34)
{
	u8 adp_buffer[4];

	dprintk("i34: ADP_SPDIFCS\n");
	i34_check_stage_eq(i34, 2);

	switch (i34->sr) {
	case 0:
		adp_buffer[0] = 2;		/* 48KHz (see p. 230) */
		break;
	case 1:
		adp_buffer[0] = 0;		/* 44.1KHz (see p. 230) */
		break;
	case 2:
		adp_buffer[0] = 3;		/* 32KHz (see p. 230) */
		break;
	default:
		dprintk("i34: invalid SPDIF sample rate\n");
		return 1;
	}
	adp_buffer[1] = 0;		/* IEC 958 category mode (???) */
	adp_buffer[2] = 0;		/* PCM encoded data */
	adp_buffer[3] = 0;
	return adp_exec_command(i34, ADP_SPDIFCS, adp_buffer, 4, 0);
}


static int
adp_audio_tolerance(struct i34_dev_s *i34, u16 tolerance)
{
	u8 adp_buffer[3];

	dprintk("i34: ADP tolerance\n");
	i34_check_stage_eq(i34, 12|16);

	adp_buffer[0] = 0x02;
	adp_buffer[1] = tolerance >> 8;
	adp_buffer[2] = tolerance;
	return adp_exec_command(i34, ADP_PARAM, adp_buffer, 3, 0);
}

/* 0 no audio sync
   1 audio syncs to sclk (clock master)
   2 sclk syncs to audio timestamps (audio master)
   3 on next frame, audio goes clock master, then audio master
*/
static int
adp_audio_sync(struct i34_dev_s *i34, u8 sync)
{
	u8 adp_buffer[3];

	dprintk("i34: ADP sync %d\n", sync);
	i34_check_stage_eq(i34, 12|16);

	adp_buffer[0] = 0x03;
	adp_buffer[1] = 0;
	adp_buffer[2] = sync & 3;
	return adp_exec_command(i34, ADP_PARAM, adp_buffer, 3, 0);	
}


static int
adp_set_function(struct i34_dev_s *i34)
{
	int function_to_call;
	u8 adp_buffer[8];

	switch (i34->adp_microcode_loaded)
	{
	  case ADP_MICROCODE_AC36:
	  case ADP_MICROCODE_AC3:

/* turned off 3d by default again. need to make this user settable because
   it doesn't always make it sound better. it may be raising the noise floor
   causing hiss to be audible as low level crumbling sound in the background
   when the foreground sound level drops. it may also be an artifact caused by
   playing AC3 6 channel thru 2 channel microcode, still testing this, 
   but it's possible it's a limitation of AC3 itself and nothing can be done
   beyond adjusting the dynamic range to filter it out. lots of possibilities
   to test in ADP_CFG for this so may take some time.
   -inkling
*/

/* see p. 206 */
/*
 summary of AC3 (DVD AC3 audio) function bits current settings:
 no ProLogic, Compression mode is line mode (no effect?), no Dual Mono Mode
 no Surround Filter, no Auto Balance, no Bass redirection, no SubWoofer,
  2/0 Output speaker configuration with downmix
  (can try w/ or w/o downmix or other spkr cfgs, tests indicate no effect?)
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 High & Low Dynamic Range set to maximum range for loud and soft sounds
  (experiment with this value. max seems to make all sounds less distorted) 
 Karaoke disabled, no Repeat Count (repeats sound like bad synth jazz)
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    COMP    DMM			*/
		adp_buffer[0] = (0<<6) |(1<<2) |0;
		/*              SF      AB      BCFG    SW      OCFG	*/
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY (16 enables 3D)		*/
		adp_buffer[2] = (0<<5) |00;
		/*              HDYNRNG (00-7F range)			*/
		adp_buffer[3] = 0x7F;
		/*              LDYNRNG (00-7F range)			*/
		adp_buffer[4] = 0x7F;
		/*              KAR     RPC				*/
		adp_buffer[5] = (0<<3) |0;
		/*              PCMSF (ms byte 00-7F range)		*/
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)		*/
		adp_buffer[7] = 0xFF;

                function_to_call = ADP_AC3;
                break;

	  case ADP_MICROCODE_MPEG:
/* see p. 213 */
/*
 summary of MPEG (DVD MPEG audio) function bits current settings:
 similar to AC3 function but without: COMP, HDYNRNG, LDYNRNG, KAR, RPC
 no ProLogic, no Dual Mono Mode
 no Surround Filter, no Auto Balance, no Bass redirection, no SubWoofer,
  2/0 Output speaker configuration without downmix
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    DMM				*/                               
		adp_buffer[0] = (0<<6) |0;
		/*              SF      AB      BCFG    SW      OCFG	*/
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY (16 enables 3D)		*/
		adp_buffer[2] = (0<<5) |00;
		/*              not used set to 0			*/
		adp_buffer[3] = 0x00;
		/*              not used set to 0			*/
		adp_buffer[4] = 0x00;
		/*              not used set to 0			*/
		adp_buffer[5] = 0x00;
		/*              PCMSF (ms byte 00-7F range)		*/
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)		*/
		adp_buffer[7] = 0xFF;

		function_to_call = ADP_MPEG;
		break;


	  case ADP_MICROCODE_LPCM:
/* see p. 211 */
/*
 summary of LPCM (DVD PCM audio) function bits current settings:
 similar to AC3 function but without: COMP, DMM, HDYNRNG, LDYNRNG, KAR, RPC
 added functions are DE, DRC(similar to H&LDYNRNG) and CMX
 no ProLogic, no De-Emphasis (dont want it fiddling with GPAIO0 bit?)
 no Surround Filter (Noise Reduction and Low Pass(?), experiment with these)
  no Auto Balance,  no Bass redirection, no SubWoofer,
  2/0 Output Configuration without downmix
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 Dynamic Range Control to max (experiment with this)
 Channel Mixing uses no default OCFG=2 downmix table
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    DE				*/
		adp_buffer[0] = (0<<6) |(0<<2);
		/*              SF      AB      BCFG    SW      OCFG	*/
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY (16 enables 3D)		*/
		adp_buffer[2] = (0<<5) |00;
		/*              DRC					*/
		adp_buffer[3] = 0x00;
		/*              not used, set to 0			*/
		adp_buffer[4] = 0x00;
		/*              CMX					*/
		adp_buffer[5] = (0<<7) |0;
		/*              PCMSF (ms byte 00-7F range)		*/
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)		*/
		adp_buffer[7] = 0xFF;

		function_to_call = ADP_LPCM;
		break;

	  default:
		dprintk("i34: invalid ADP microcode selected for function set\n");
		return 1;
	}

	return adp_exec_command(i34, function_to_call, adp_buffer, 8, 0);
}


static int
adp_audio_gain(struct i34_dev_s *i34, u32 gain)
{
	u8 adp_buffer[4];

	/* see p. 221 */
	/* means PARAM EXT = ADP Command - Volume Control */
	adp_buffer[0] = 0x00;
	adp_buffer[1] = (gain>>16) & 0x07;
	adp_buffer[2] = (gain>>8)  & 0xFF;
	adp_buffer[3] = (gain)     & 0xFF;
	return adp_exec_command(i34, ADP_PARAM, adp_buffer, 4, 0);
}

/* return adp decoder status via pointer, dont go past 16 bytes */
static int
adp_decoder_status(struct i34_dev_s *i34, u8 *adp_buffer)
{
	int i;
	u8 count, istatus;
//	u8 adp_buffer[256];

	/* see p. 235 */
	adp_write(i34, ADP_STAT);
	adp_read(i34, &count);		/* count of bytes left to read */

	for (i=0; i<count; i++)
	{
		adp_write(i34, ADP_READ);
		/* FIXME: replace with MAXADPDECODELIMIT */
		if (i < 16) adp_read(i34, &adp_buffer[i]);
	}
	adp_write(i34, ADP_READ);
	adp_read(i34, &istatus);		/* reading ISTATUS */
	if (istatus != ADP_NO_ERROR)
	{
		dprintk("i34: adp_decoder_status error %02X\n", istatus);
		return 1;
	}
	return 0;
}






/* CSS sub-system */
#ifdef CSS_SUPPORT

int css_read_key(struct i34_dev_s *i34, u8 *buffer, int length)
{
	int i;

	for (i=0; i<length; i++) {
		buffer[i] = i34_read_parameter(i34, 0xA1);
	}
	return 1;
}

int css_write_key(struct i34_dev_s *i34, u8 *buffer, int length)
{
	int i;

	for (i=0; i<length; i++) {
		i34_write_parameter(i34, 0xA2, buffer[i]);
	}
	dprintk("i34: css status 0x%04X\n", i34_read_parameter(i34, 0xA1));
	return 1;
}

int css_wait(struct i34_dev_s *i34, int a, int b, int *c)
{
	int i;

	for (i=0; i<1000; i++) {
		*c = i34_read_parameter(i34, 0xA1);
		i34_wait(10);
		if ((*c & a) == b)
			return 1;
	}
	return 0;
}

int css_select(struct i34_dev_s *i34, int a)
{
	if ((i34_read_parameter(i34, 0xA1) & 0xC0) != 0x80)
		return 0;
	/* i34_change_parameter(0xA0... */
	return 1;
}

int css_write_disk_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

int css_write_title_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

int css_read_chal_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

int css_write_chal_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

int css_read_resp_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

int css_write_resp_key(struct i34_dev_s *i34, dvd_authinfo auth)
{
	return 1;
}

#endif


#define HCREADY (1<<15)

static int
i34_host_command(struct i34_dev_s *i34, u16 cmd)
{
	int i;
	u16 status1;

	for (i=0; i<HZ/10; i++) {
		/* p. 100 & 147 ZR HB:
		   check that the host is ready to receive a new command */
		status1 = i34_read(i34, ZR_STATUS1);
		if (status1 & HCREADY) {
			i34_write(i34, ZR_HOSTCMD, cmd);
			return 0;
		} else {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout( 1 );
		}
	}
	dprintk("i34: host not ready for command (STATUS1=%04X)\n",
		status1);
	return 1;
}


#define CSTATE(c,s) ((c & ~ZR_CSTATE_FLAG_NOACT) == s)

static int
i34_wait_cstate(struct i34_dev_s *i34, int target, int timeout)
{
	int cstate = 0;
	int i;
	
	for (i=0; i<timeout; i++) {
		if (i34_state(i34, &cstate) != 0)
			return 1;
		if ( CSTATE(cstate, target) )
			return 0;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout( 1 );
	}
	dprintk("i34: wait_cstate 0x%02X, not 0x%02X\n", cstate, target);
	return 1;
}


/*

   NOTE: you don't want to send commands more frequently than the
   i34_wait below, about 1 field NTSC, so if you want to send more
   than ~15 commands/sec you will need to adjust this code or your code.
   this may be reason for step corrupting when repeat rate is too high
   from holding step key down and letting repeat kick in to > 15 keys/sec
   -inkling
*/

static int
i34_send_command(struct i34_dev_s *i34, u16 command, u8 parameter)
{
	int cmd, cstate1, cstate2, res;
	 
	res = 0;
	
	cmd = (command & 0xFF00) | parameter;

	i34_state(i34, &cstate1);

	switch(command)
	{
		case I34_CMD_START:
		/* p150 says must not send START while already in playback
		    mode (only IDLE or INIT_DISPLAY allowed for START) */

		/* p150 also says it will discard any host commands that are
		   not allowed to be executed from the current state of
		   operation (except START/END), so this simplifies things
		  -inkling
		*/
			if ( (cstate1 != ZR_CSTATE_INIT_DISPLAY)
				&& (cstate1 != ZR_CSTATE_IDLE) )
			{
				dprintk("i34: wrong c-state 0x%02X for cmd START\n",
					 cstate1);
				return 1;
			}

			res = i34_host_command(i34, cmd);
			
			if (i34_wait_cstate( i34, ZR_CSTATE_NSPB, 200 ) != 0)
			{
				dprintk( "not playing after START\n" );
				return 1;
			}
			
			break;


		case I34_CMD_PAUSE:
			res = i34_host_command(i34, cmd);

			if (i34_wait_cstate( i34, ZR_CSTATE_PAUSE, 200 ) != 0)
			{
				dprintk( "not paused after PAUSE\n" );
				return 1;
			}
			break;
			

		/* spec state diagram indicates single step only works when
		   already in pause mode, but text description is different
		   from state diagram. pause/slow/step interaction testing
		   shows these valid state changes:
		    play-pause-step
		    play-slow-step
		    play-pause-step-slow
		    play-pause-slow-step
		    play-step does nothing, matches diagram but not text
		   more testing might be needed on this
		  -inkling
		 */
		case I34_CMD_STEP:
			res = i34_host_command(i34, cmd);
			break;
		

		/* this is a nicely working trick mode
		   NOTE: audio is enabled, can use tell if sync is offset
		   -inkling
		*/
		case I34_CMD_SLOW:
			res = i34_host_command(i34, cmd);
			break;
			

		/* spec state diagram doesn't show any way to leave fast.
		   paused fast only continues back with fast. p154 says
		   use end_playback. -ink */
		case I34_CMD_FAST:
			res = i34_host_command(i34, cmd);
			break;


		case I34_CMD_CONT:
			res = i34_host_command(i34, cmd);
			break;


		case I34_CMD_END:
		/* it takes a while to bring the card down to the idle state */
			res = i34_host_command(i34, cmd);
			if (i34_wait_cstate( i34, ZR_CSTATE_IDLE, 200 ) != 0)
			{
				dprintk( "not stopped after END\n" );
				return 1;
			}
			break;


		case I34_CMD_SEQEND:
		/* it takes a while to bring the card down to the idle state */
			res = i34_host_command(i34, cmd);
			if (i34_wait_cstate( i34, ZR_CSTATE_IDLE, 10000 ) != 0)
			{
				dprintk( "not stopped after SEQEND\n" );
				return 1;
			}
			break;

		default:
			dprintk( "i34: unknown host command 0x%02X\n", command );
			return 1;
			break;
	}

	/* needed? */
	// i34_wait(17*1000); /* 17ms = about 1 field of NTSC */

	i34_state(i34, &cstate2);

	dprintk("i34: hostcmd %04X cstate %02X to %02X res %d\n",
		 cmd, cstate1, cstate2, res);
	return 0;
}



static int
i34_color_key(struct i34_dev_s *i34, struct video_key *ck)
{
	i34_write_parameter(i34, ZR_COLORY, ck->key[0]);
	i34_write_parameter(i34, ZR_COLORU, ck->key[1]);
	i34_write_parameter(i34, ZR_COLORV, ck->key[2]);
	return 0;
}


static int
i34_composite_blank_sync(struct i34_dev_s *i34, struct composite_blank_s *cb)
{
	int cstate;

	if (i34_state(i34, &cstate) != 0) return 1;

	/* if init_pclk, init_display or idle */
	if (	(cstate == ZR_CSTATE_INIT_DISPLAY) ||
		(cstate == ZR_CSTATE_IDLE) ||
		(cstate == ZR_CSTATE_INIT_PCLK)) {
		i34_write_parameter(i34, ZR_CBHSTART, cb->CBHStart);
		i34_write_parameter(i34, ZR_CBVSTART, cb->CBVStart);
		i34_write_parameter(i34, ZR_CBHEND, cb->CBHEnd);
		i34_write_parameter(i34, ZR_CBVEND, cb->CBVEnd);
		return 0;
	} else {
		dprintk("i34: wrong c-state (0x%02X) for composite blank sync cfg\n", cstate);
		return 1;
	}
}


/* see 36710 spec p 88:
   ntsc: (givens: htotal 858, active sizex 720, vtotal 525, active sizey 240)
	htotal is 858
	active sizex is 720 (must be multiple of 8)
	hsyncsize is a minimum of 8
	active startx is a minimum of 40, must be even (odd messes up colors)
	active startx = (htotal - active sizex)/2
	  (trying to center it may or may not be a good idea)
	  ( may need to take into account hsyncsize, i.e.
	   active startx = (htotal - hsyncsize - activesizex)/2 )
		(858-720)/2 = 138/2 = 69 (68)
		(858-8-720)/2 = 130/2 = 65 (64)
	active endx is active startx plus active sizex

	vtotal is number of lines for BOTH fields
	active sizey is number of visible lines in ONE field (480/2) = 240
	vsyncsize is a minimum of 1
	active starty is minimum of 4:
	   use (525 - vsyncsize - 480)/2 = 22
	active endy is active starty + field size (22+240)

	fieldsel:
	0 top field is output during field I, bottom field on field II
	1 bottom field is output during field I, top field on field II

	vidfps:
	0 output frame rate is 29.97 (30/1.001) fps - ntsc
	1 output frame rate is 25 fps - pal

	vidconfig: (using 0x4006) see p 85
	0x4006 = pixlim | video8 | sync8
	some of the other bits may need tweaking

	these settings work for mpeg-1 a/v playback but break vob.
	actually don't notice any diff between these and the older settings
	that are commented out below. vob playback broken, mpeg-1 ok.
	vob playback works correctly with the pal1 settings, but breaks mpeg-1
	on *some* mpeg-1 vids (ntsc ones).
 -inkling
*/

static struct video_standard_s ntsc1 = {
	HTotal:858,
	HSyncSize:8,
	ActiveStartX:64,
	ActiveEndX:784,
	ActiveSizeX:720,
	
	VTotal:525,
	VSyncSize:1,
	ActiveStartY:22,
	ActiveEndY:22+240,
	ActiveSizeY:240,
	
	VidFPS:0,
	FieldSel:1,
	VidConfig:0x4006
};

#ifdef NEVER
static struct video_standard_s ntsc1 = {

	HTotal:858,
	HSyncSize:64,
	ActiveStartX:856-720,
	ActiveEndX:856,
	ActiveSizeX:720,

	VTotal:525,
	VSyncSize:6,
	ActiveStartY:14,
	ActiveEndY:14+240,
	ActiveSizeY:240,

	VidFPS:0,
	FieldSel:0,
	VidConfig:0x4006
};
#endif

static struct video_standard_s ntsc2 = {
	0, 0, 858, 525,	64, 6, 136, 16, 856, 352, 720, 240, 0x4006
};

static struct video_standard_s pal1 = {
	VidFPS:1, FieldSel:0,
	
	HSyncSize:64, VSyncSize:6,
	
	HTotal:864,
	ActiveStartX:856-720,
	ActiveEndX:856,
	ActiveSizeX:720,
	
	VTotal:625,
	ActiveStartY:21,
	ActiveEndY:21+288,
	ActiveSizeY:288,
	
	VidConfig:0x4006
};


static struct video_standard_s pal2 = {
	1, 0, 864, 625, 64, 6, 142, 16, 862, 310, 720, 288, 0x4006
};


static int
i34_video_standard_internal(struct i34_dev_s *i34, int standard)
{
	int cstate;

	dprintk("i34: vstd_int\n");

	switch(standard) {

	case VIDEO_MODE_NTSC:
		dprintk("i34: video standard NTSC\n");
		i34->vs = &ntsc1;
		break;
	case I34_NTSC_ALT:
		dprintk("i34: video standard NTSC_ALT\n");
		i34->vs = &ntsc2;
		break;
	case VIDEO_MODE_PAL:
		dprintk("i34: video standard PAL\n");
		i34->vs = &pal1;
		break;
	case I34_PAL_ALT:
		dprintk("i34: video standard PAL_ALT\n");
		i34->vs = &pal2;
		break;
	default:
		dprintk("i34: video standard NTSC\n");
		i34->vs = &ntsc1;
		break;
	}

	if (i34_state(i34, &cstate) != 0)
		return 1;

/* new checking in write_param now, so it should be ok to
   write all these flat out and return ok. nothing
   happens until the startdisplay at the end
 -ink */
	/* if init_pclk, init_display or idle */
//	if (	(cstate == ZR_CSTATE_INIT_DISPLAY) ||
//		(cstate == ZR_CSTATE_IDLE) ||
//		(cstate == ZR_CSTATE_INIT_PCLK)) {

		i34_write_parameter(i34, ZR_VIDOUT,
			((i34->vs->FieldSel & 1)<<1) | (i34->vs->VidFPS & 1) );
		i34_write_parameter(i34, ZR_HTOTAL, i34->vs->HTotal);
		i34_write_parameter(i34, ZR_VTOTAL, i34->vs->VTotal);
		i34_write_parameter(i34, ZR_HSYNCSIZE, i34->vs->HSyncSize);
		i34_write_parameter(i34, ZR_VSYNCSIZE, i34->vs->VSyncSize);
		i34_write_parameter(i34, ZR_ACTIVESTARTX, i34->vs->ActiveStartX);
		i34_write_parameter(i34, ZR_ACTIVESTARTY, i34->vs->ActiveStartY);
		i34_write_parameter(i34, ZR_ACTIVEENDX, i34->vs->ActiveEndX);
		i34_write_parameter(i34, ZR_ACTIVEENDY, i34->vs->ActiveEndY);
		i34_write_parameter(i34, ZR_ACTIVESIZEX, i34->vs->ActiveSizeX);
		i34_write_parameter(i34, ZR_ACTIVESIZEY, i34->vs->ActiveSizeY);


//		}

		i34_composite_blank_sync(i34, &i34->cbs);
		i34_color_key(i34, &i34->ckey);
		i34_write_parameter(i34, ZR_CAPTIONOFFSET, 21);
		
//		if (cstate == ZR_CSTATE_INIT_PCLK)
//			i34_check_state_configured(i34);

		/* per p. 80 VidConfig only at init_pclk cstate */
		if (cstate == ZR_CSTATE_INIT_PCLK) {
		    i34_write_parameter(i34, ZR_VIDCONFIG, i34->vs->VidConfig);
		}

		i34_write_parameter(i34, ZR_STARTDISPLAY, 1);

		i34_write_parameter(i34, ZR_BACKGROUNDSWITCH, 1);

		return 0;

//	} else {
//		dprintk("i34: incorrect state for video configuration (c-state=%#2.2x)\n", cstate);
//		return 1;
//	}
}


static int
i34_adp_boot_write(struct i34_dev_s *i34, u8 *ucode, int size)
{
	int i;
	u8 count;

	for ( i=0; i<size; i++ ) {
		if (adp_write(i34, ucode[i]) != 0)
			return 1;
		if (adp_read(i34, &count) != 0)
			return 1;
		if ( count != ((size-i)&0xff) ) {
			dprintk( "adp boot: got %02x, not %02x (at %p[%i of %i])\n", count, (size-i)&0xff, ucode, i, size );
			return 1;
		}
	}
	return 0;
}


static int
i34_check_adp_boot(struct i34_dev_s *i34, u8 *ucode, int size)
{
	u32 *hdr;
	u8 *p;
	u32 target, n, m;
	u32 v1, v2, v3;
	int i;
	u8 opcode;
	u8 count;
	
	struct {
		u8 opcode; u8 plen;
	} opcodes[] = {
		{ 0x00, 0x00 },
		{ 0x91, 0x04 },
	};
	
//	dprintk( "adp checked boot %p, %i\n", ucode, size );

	dprintk("i34: ADP_BOOT\n");
	i34_check_stage_eq(i34, 12);

	p = ucode;
restart:	
	if (adp_write(i34, ADP_BOOT) != 0)
		return 1;
	if (adp_read(i34, &count) != 0)
		return 1;
	if ( count != 5 ) {
		dprintk( "adp: BOOT %02x\n", count );
	}
	while ( p-ucode < size ) {
		hdr = (u32*)p;
		
		target =	__be32_to_cpu( hdr[ 0 ] ) & 0x000FFFFF;
		n =		__be32_to_cpu( hdr[ 1 ] ) & 0x000FFFFF;
		m =		__be32_to_cpu( hdr[ 2 ] ) & 0x000FFFFF;
		
//		dprintk( "adp chk %5x %5x %5x\n", target, n, m );
		
		if (i34_adp_boot_write(i34, p, 4) != 0) return 1;
		if (i34_adp_boot_write(i34, p+4, 4) != 0) return 1;
		if (i34_adp_boot_write(i34, p+8, 4) != 0) return 1;
		
		p += 3*sizeof(u32);
		
		if ( n > 0 ) {
			if (i34_adp_boot_write(i34, p, n * sizeof(u32)) != 0) return 1;
		
			p += n * sizeof(u32);
		}
		else {
			hdr = (u32*)p;
			
			v1 =	__be32_to_cpu( hdr[ 0 ] );
			v2 =	__be32_to_cpu( hdr[ 1 ] );
			v3 =	__be32_to_cpu( hdr[ 2 ] );
			
//			dprintk( "adp chk v %08x %08x %08x\n", v1, v2, v3 );

			if (i34_adp_boot_write(i34, p, 4) != 0) return 1;
			if (i34_adp_boot_write(i34, p+4, 4) != 0) return 1;
			if (i34_adp_boot_write(i34, p+8, 4) != 0) ;//return 0;
			
			p += 3*sizeof(u32);
			
			while ( p-ucode < size ) {
			
				opcode = *p++;
			
				if ( opcode == ADP_BOOT ) {
					goto restart;
				}
				
				if (adp_write(i34, opcode) != 0) return 1;
				if (adp_read(i34, &count) != 0) return 1;
			
				for (i=sizeof(opcodes)/sizeof(opcodes[0]); i-->0 ; ) {
					if ( opcodes[i].opcode == opcode ) {
						if ( (count != opcodes[i].plen+1) &&
							(count != ADP_NO_ERROR) &&
							(count != ADP_READY_FOR_NEW_CMD)
							) {
							dprintk( "adp: opcode %02x start %02x\n",
								opcode, count );
							return 1;
						}
					
						if (i34_adp_boot_write(i34, p, opcodes[i].plen) != 0)
							return 1;
						p += opcodes[i].plen;
						break;
					}
				}
				if (i==-1) {
					dprintk( "unknown opcode %02x\n", opcode );
				}
			}
		}
	}
	if ( p-ucode != size ) {
		dprintk( "adp boot size mismatch\n" );
	}
	return 0;
}


static int
i34_load_adp(struct i34_dev_s *i34, u8 *ucode, int size)
{
	u8 opcode_error;
	
	/* in case we fail */
	i34->adp_microcode_loaded = ADP_MICROCODE_NONE;

	dprintk("i34: ADP microcode version %d.%d\n", ucode[0], ucode[1]);
	
	if (i34_check_adp_boot(i34, ucode+2, size-2) != 0)
		return 1;

	adp_write(i34, ADP_READ);
	adp_read(i34, &opcode_error);
	if (opcode_error != ADP_NO_ERROR) {
		dprintk("i34: error loading ADP microcode (%02x)\n", opcode_error);
		return 1;
	}

	dprintk("i34: ADP loaded (%d bytes)\n", size);
	return 0;
}


static int
i34_adp_boot_internal(struct i34_dev_s *i34, int adp_microcode)
{
	int adp_retcode, size;
	u8 *ucode;

	if (i34->adp_microcode_loaded == adp_microcode) {
		dprintk("i34: ADP microcode not changed\n");
		return 0;
	}

	switch (adp_microcode)
	{
		case ADP_MICROCODE_AC3:
			ucode = adp_ac3;
			size  = sizeof(adp_ac3);
			dprintk("i34: ADP microcode AC3\n");
			break;
		
		case ADP_MICROCODE_AC36:
			ucode = adp_ac36;
			size  = sizeof(adp_ac36);
			dprintk("i34: ADP microcode AC36\n");
			break;

		case ADP_MICROCODE_LPCM:
			ucode = adp_lpcm;
			size  = sizeof(adp_lpcm);
			dprintk("i34: ADP microcode LPCM\n");
			break;

		case ADP_MICROCODE_MPEG:
			ucode = adp_mpeg;
			size  = sizeof(adp_mpeg);
			dprintk("i34: ADP microcode MPEG\n");
			break;

		case ADP_MICROCODE_PCM:
			ucode = adp_cdda;
			size  = sizeof(adp_cdda);
			dprintk("i34: ADP microcode PCM\n");
			break;

	default:
		dprintk("i34: invalid ADP microcode selection (%d)\n",
			 adp_microcode);
		return 1;
	}

	adp_retcode = i34_load_adp(i34, ucode, size);
	if (adp_retcode == 0) i34->adp_microcode_loaded = adp_microcode;

	return adp_retcode;
}

	
static int
i34_load_dvp(struct i34_dev_s *i34, u8 *ucode, int size)
{
	u8 buffer[60], temp;
	int i, count, pad, limit;

	/* in case we fail */
	i34->dvp_microcode_loaded = DVP_MICROCODE_NONE;

	dprintk("i34: DVP microcode version %d.%d\n", ucode[0], ucode[1]);
	ucode += 2;
	limit = size;
	limit -= 2;
	/* prepare a chunk of microcode
	   swap bytes, because it's DVP microcode
	*/
	while (limit > 0)
	{
		for (i=0; (i<sizeof(buffer)) && (limit>0); i++, limit--)
		buffer[i] = ucode[i];
		count = i;
		for (i=1; i<count; i+=2)
		{
			temp = buffer[i-1];
			buffer[i-1] = buffer[i];
			buffer[i]   = temp;
		}
		ucode += count;
		if (i34_write_parameter_seq(i34,
						ZR_MICROCODE,
						(u16*)buffer,
						count>>1) != 0)
		{
			dprintk("i34: error downloading DVP microcode\n");
			return 1;
		}
	}

	/* zero-pad microcode */
	count = 0;
	pad = 0;
	if (size < 16384)
	{
		pad = count = (16384 - size);
		count >>= 1;
		dprintk("i34: padding %d bytes %d words\n", pad, count);
		for (i=0; i<count; i++) {
			i34_write_parameter(i34, ZR_MICROCODE, 0);
		}
	}

	dprintk("i34: DVP loaded (%d bytes, %d pad)\n", size, pad); 
	return 0;
}


static int
i34_dvp_boot_internal(struct i34_dev_s *i34, int dvp_microcode)
{
	int dvp_retcode, size;
	u8 *ucode;

	if (i34->dvp_microcode_loaded == dvp_microcode) {
		dprintk("i34: DVP microcode not changed\n");
		return 0;
	}
	switch(dvp_microcode)
	{
		case DVP_MICROCODE_DVD:
			ucode = dvp_dvd;
			size = sizeof(dvp_dvd);
			dprintk("i34: DVP microcode DVD\n");
			break;

		case DVP_MICROCODE_VCD:
			ucode = dvp_vcd;
			size = sizeof(dvp_vcd);
			dprintk("i34: DVP microcode VCD\n");
			break;

		case DVP_MICROCODE_MPEG:
			ucode = dvp_oth;
			size = sizeof(dvp_oth);
			dprintk("i34: DVP microcode MPEG\n");
			break;

		case DVP_MICROCODE_SCAN:
			ucode = dvp_scan;
			size = sizeof(dvp_scan);
			dprintk("i34: DVP microcode SCAN\n");
			break;

		default:
			dprintk("i34: invalid DVP microcode selection (%d)\n",
				dvp_microcode);
			return 1;
	}
	dvp_retcode = i34_load_dvp(i34, ucode, size);
	if (dvp_retcode == 0) i34->dvp_microcode_loaded = dvp_microcode;
	return dvp_retcode;
}


static u8 init_file[] = {
#include "Zi34slt1.inc"
};


static int
i34_load_init_file(struct i34_dev_s *i34)
{
	int i, count, size;
	u8 *codePtr;
	u8 buffer[64];

	if (i34->init_file_loaded) {
		dprintk("i34: init_file already loaded\n");
		return 0;
	}	
	size = sizeof(init_file) - 2;
	codePtr = &init_file[2];
	count = size;

	while(count>0) {
		for (i=0; (i+1<64) & (count>0); i+=2) {
			buffer[i+1] = codePtr[i];
			buffer[i  ] = codePtr[i+1];
			count -= 2;
		}
		codePtr += i;
		if ( i34_write_parameter_seq(i34, ZR_INITFILE_3, (u16*)buffer, i>>1) ) {
			dprintk("i34: error while loading init_file\n");
			return 1;
		}
	}
	dprintk("i34: SLOT MACHINE init_file loaded (%d bytes)\n", size);
	i34->init_file_loaded = 1;
	return 0;
}




/* Implement the protocol to read/write to/from the DVP data buffer
   (see p.106 and following). These functions always process 64 bytes
   at a time (32 words)
*/
#define DVPIBE	0x0080
#define DVPVAL	0x0100
#define DVPOBF	0x4000

#ifdef NEVER
static int
i34_read_dvp_data(struct i34_dev_s *i34, u16 *data)
{
	u16 status1;
	u16 isr;
	int i;

	/* FIXME: this should go into the interrupt routine */
	isr = i34_read(i34, ZR_ISR);
	if (isr & DVPOBF) {
		for (i=0; i<32; i++)
			data[i] = i34_read(i34, ZR_DVP_DATA);
		status1 = i34_read(i34, ZR_STATUS1);
		if (status1 & DVPVAL) {
			dprintk("i34: data read from DVP has been corrupted\n");
			return 1;
		}
		return 0;
	}
	return 1;
}


static int
i34_write_dvp_data(struct i34_dev_s *i34, u16 *data)
{
	u16 status1;

	status1 = i34_read(i34, ZR_STATUS1);
	if (status1 & DVPIBE) {
		return 0;
	}
	dprintk("i34: DVP not ready to receive data\n");
	return 1;
}
#endif


static int
i34_read_nav_data(struct i34_dev_s *i34, int segment, u16 *data, int count)
{
	u16 status;
	u16 addr;
	int i, c;
	int nav_i;
	
	addr = segment * 0x0400;
	i34_write(i34, ZR_NAV_ADDRESS, addr);
	
	nav_i = 0;
	while (nav_i < count) {
		for (i=0; i<100; i++) {
			status = i34_read(i34, ZR_STATUS0);
			if (status & 0x80)
				/* DRAMBF is set */
				break;
			udelay(3);
		}
		if (i==100) {
			printk( "no NAV data\n" );
			return 1;
		}
		c = (count-nav_i) < 32 ? (count-nav_i) : 32;
		for (i=0; i<c; i++)
			data[ nav_i++ ] = i34_read(i34, ZR_NAV_DATA);
	}
	if (i34->dvpgen2 & 1) {
		/* must write 0x00 to DecCntr or else
		   bitstream will stall when 7 segs loaded */
		i34_write(i34, ZR_DECCNTR, 0x00);
	}
	return 0;
}


/* Prepares the display
   while state = init_pclk (see p. 80)
*/
static int
i34_init_video_port(struct i34_dev_s *i34)
{
	int cstate;

	/* Video configuration */
	//i34_init_subpicture_palette(i34);

	i34->VideoStandard = VIDEO_MODE_PAL;

	i34->cbs.CBHStart = 0x0A;
	i34->cbs.CBVStart = 0;
	i34->cbs.CBHEnd = 0;
	i34->cbs.CBVEnd = 0;
	i34->ckey.key[0] = 0x10;
	i34->ckey.key[1] = 0x80;
	i34->ckey.key[2] = 0x80;
//	rgb2yuv2( 0x40, 0x00, 0x80, i34->ckey.key ); // test: dark-ish purple

	i34->AutoScaling = 3;
	i34->AutoPanScan = 1;
	i34->ScaleRatio = (i34->AutoScaling & 3) ? 0x0000 : 0x1111;
	i34->SPScale = 0;

	i34->BgSwitch = 1;
	i34->VidEntry = 0;
	i34->LastPic = 0x200;
	i34->VidFloat = 0; /* 0x40; */

	i34->SpSwitch = 5;
	i34->CaptionSwitch = 0;

	/* first video stream ID encountered  (see p. 128) */
//	i34->video_sid = 0x0000;

	i34->video_tolerance = MS_TO_SCLK(17); /* 1 field NTSC */
	i34->video_delay = 0;

	/* video synchronize to SCLK (p. 142) */
	i34->video_sync_mode = 2;
	
	/* DVPGEN2 bits from spec:
	   (DVPGEN2 bits have different functions depending on DVP microcode)
	   bit 7: (DVD) Error Concealment, see p140
		=0 at macroblock, slice & picture level,
		   may cause more freezes but less green blocks
	******** you wouldn't think 'concealing picture errors' stalls less
		=1 at macroblock & slice level,
		   may cause less freezes, but more green blocks
	   NOTE: apparently 36710 spec has it backwards, or 36700 is different:
	    setting bit 7 causes the freezing problem in NTSC VOB playback.
	    now playing NTSC VOBs in PAL mode stutters, but that doesn't matter
	    because player should parse bitstream and set NTSC or PAL mode.
	PAL mpg played in PAL mode works, NTSC played in NTSC mode works
	mixing them not such a good idea, working on auto selection in
	mzplay to handle this for you so you don't have to worry about it.	   

	   bit 0: (DVD) [name of bit unknown] see p118
		=0 overwrite NAV segments with new NV_PCKs when NAV buffer full
		=1 stop bitstream transfer when NAV segment buffer full
	-inkling
	*/
	i34->dvpgen2 = 0x00;

	dprintk("i34: i34_init_video_port\n");

	i34_video_standard_internal(i34, i34->VideoStandard);

	if (i34_state(i34, &cstate) != 0)
		return 1;

	/* we should now be in state = init_display */
	if (cstate != 0x03) {
		dprintk("i34: warning device not in 'init_display' state after init_video_port\n");
		return 1;
	}
	return 0;
}


int
i34_load_subpicture_palette(struct i34_dev_s *i34, u32 *palette)
{
	int z;
	u32 j;

	dprintk("i34: load_subpicture_palette\n");

	/* fix the palette word order and display new order */
	for ( z = 0; z<16; z++)
	{
		/* FIXME: is this swap step needed for big endian? */
		/*   x86: swap just the words, not byte order */
		j = palette[z];
		j = (j & 0xffff)<< 16 | (j & 0xffff0000)>>16;
		palette[z] = j;
		dprintk2("i34: sppal: %02X %08X\n", z, palette[z] );
	}
	/* this is rather odd looking. goal is to move 16 words of 32-bit
	   via a 16-bit bus, so we actually move 32 of 16-bit half-words
	*/
	return i34_write_parameter_seq(i34, ZR_SPPALETTE, (u16*)palette, 32);
}


static int
i34_config_audio_stage1(struct i34_dev_s *i34)
{
	dprintk("i34: i34_config_audio_stage1\n");

	/* Audio Configuration */

	/* allow slight bit of drift (+/-1 field NTSC) before skipping audio.
	   theory is to not be too strict on audio sync to prevent audible
	   audio frame drops, but in practice seems not to matter */
	i34->audio_tolerance = MS_TO_SCLK(17); /* 1 field NTSC */
	i34->audio_tolerance = 0;

	/* no audio reconstruction delay. doesn't seem to be needed anyway,
	   but may find a use for it to be set by the player code one day. */
	i34->audio_delay = 0;

	/* p141-142 & 221:
	   DVP automatically initializes SCLK to first SCR in bitstream
	   then continues in clock master mode. this is fine for bitstreams
	   that have locked audio and video clocks, but actual practice
	   seems to favor letting the audio clock be the master, so
	   start in audio master mode (2) to let audio timestamps set SCLK

	   p145, can switch between 0 and 1 audio sync modes during playback
	   when in clock master mode (sync on/off allowed here)
	*/
	i34->audio_sync_mode = 2;	/* fiddle with this -ink */

	i34->audio_gain = 0x7FFFF;	/* max gain */

	/* This configures the serial interface between the decoder
	   and the external DAC.
	   see p. 226 ZR HB
	   see p. 10 AD1857 HB
	   
   	   pin 3 on 1857 determines left-justified or I2S - is it programmable?
	   or pulled up/down?  It looks like it is in i2s mode, according to
	   the correctly working set of params, so base all on figure 10. */

	/* when aout/alrclock are sampled: 0=falling, 1=rising edge of abclk
	   p.10 1857: rising edge (unless there are inverters in the path..) */
	i34->serial.cpa = i34->serial.cpb = 1;
	
	/* error output pin disabled */
	i34->serial.err = 0;
	
	/* S/PDIF acts as a S/PDIF transmiter (0=4th chan, 1=spdif) */
	i34->serial.spo = 1;			
	
	/* 32 bits/64 BCLK cycles frame size (0=16, 1=32, 7=24) */
	i34->serial.fra = i34->serial.frb  = 1;
		
	/* 20 bits input/output word (0=20, 1=18, 2=16, 3=24) */
	i34->serial.inw = i34->serial.outw = 0;	
	
	/* osp needs to be 0. see also fig. 10 --lvs */
	/* 0=(left=low LRCLK), 1=(left=high LRCLK) */
	i34->serial.isp = i34->serial.osp = 0;	
	
	/* 0=(left-justified data (sample at l/r transition),
	   1=(left-justified data + 1 bit) (I2S)
	   7=(right justified)
	*/
	i34->serial.fmb = i34->serial.fma = 1;

	adp_clean(i34);
	adp_rom_version(i34);

	/* external DAC is hardwired to 256xFs (pin 6 to GND) */
	i34->amclk = 256;

	/* see table p.73:
		256,64 (64 cycles: clocks 2x20bits per L+R sample) */
	i34->spbs = 2;

    	i34->adp_microcode_loaded = ADP_MICROCODE_NONE;

	/* p249: "the audio port can only be configured once after RESET.
	   if this port needs to be re-initialized, then another RESET must be
	   executed before new values are loaded."
	*/
	   
	/* FIXME: MPEG microcode supports more than one sample rate. */
	/* FIXME: need to specify the sample rate before calling audio stage1.
	   it's defaulting to 48000Hz first time thru and requires two tries.
	 -inkling
	*/

	adp_clean(i34);

	/* p114: table 47:
	   check if it's valid to send an adp command at a certain cstate
	   (not doing this currently)
	*/

	/* p229: PLLTAB first, then PLLCFG, then CFG */
	adp_plltab(i34);

	/* sample rate specified by user */
	adp_pllcfg(i34, i34->sr);

	/* p227: CFG command must be followed by an AC3 or 'something' */
	adp_config_serial(i34, &i34->serial);

	/* 'something' is setting up the sony/phillips digi if */
    	adp_config_spdif(i34);
	
	return 0;
}


static int
i34_config_audio_stage2(struct i34_dev_s *i34)
{
	dprintk("i34: i34_config_audio_stage2\n");


	/* set the 2 GPAIO lines in output mode to drive the external DAC and multiplexer
	   Thanks to Gernot Ziegler <gz@lysator.liu.se> for this information :
	   GPA I/O 0 leads to Pin 10, Philips HT4052 (acting as select pin S0 ?)
	   GPA I/O 1 leads to Pin 2, AD 1857 (!PD/!RST, make HI for normal operation, says the manual) 
	*/
	adp_config_gpaio(i34, 1, 1);	/* set line0=output */
	adp_config_gpaio(i34, 2, 2);	/* set line1=output */
	adp_set_gpaio(i34,    2, 0);	/* reset/powerdown the DAC, assert PD/RST */
	udelay(1);			/* keep low at least 4MCLK (355ns) */

	adp_config_gpaio(i34, 1, 1);	/* again? */
	adp_set_gpaio(i34,    1, 0);	/* select S0 on the analog demultiplexer */

	adp_config_gpaio(i34, 2, 2);	/* again? */
	adp_set_gpaio(i34,    2, 2);	/* deassert PD/RST; this starts the reset.
					boot up the external DAC */
	i34_wait(27*1000);		/* DAC synchronization takes 30000MCLK (27ms at 11MHz) */

	adp_set_function(i34);

	i34_write_parameter(i34, ZR_AUDSID, i34->audio_sid);

	adp_audio_tolerance(i34, i34->audio_tolerance);
	adp_audio_sync(i34, i34->audio_sync_mode);
	
	i34_write_parameter(i34, ZR_AUDPORTDELAY, i34->audio_delay);

	adp_simple_command(i34, ADP_PLAY);
	adp_simple_command(i34, ADP_UNMUTE);
	adp_audio_gain(i34, i34->audio_gain);

	return 0;
}


static int
i34_configure_display(struct i34_dev_s *i34)
{
	/* Scaling & Panning (p. 158) */
	i34_write_parameter(i34, ZR_AUTOPANSCAN, i34->AutoPanScan);
	i34_write_parameter(i34, ZR_PANSCANBASEX, 0);
	i34_write_parameter(i34, ZR_PANSCANBASEY, 0);
	i34_write_parameter(i34, ZR_PANSCANOFFSETX, 0);
	i34_write_parameter(i34, ZR_PANSCANOFFSETY, 0);
	i34_write_parameter(i34, ZR_AUTOSCALING, i34->AutoScaling);
	i34_write_parameter(i34, ZR_SCALERATIO, i34->ScaleRatio);
	i34_write_parameter(i34, ZR_SPSCALE, i34->SPScale);

	i34_write_parameter(i34, ZR_IMAGESTARTX, i34->vs->ActiveStartX);
	i34_write_parameter(i34, ZR_IMAGESTARTY, i34->vs->ActiveStartY);
	i34_write_parameter(i34, ZR_IMAGEENDX, i34->vs->ActiveEndX);
	i34_write_parameter(i34, ZR_IMAGEENDY, i34->vs->ActiveEndY);
	i34_write_parameter(i34, ZR_IMAGESIZEX, i34->vs->ActiveSizeX);
	i34_write_parameter(i34, ZR_IMAGESIZEY, i34->vs->ActiveSizeY);
	
	if (i34->AutoPanScan == 0) {
		// panscanoffsetx, y
	}
	i34_write_parameter(i34, ZR_DVPGEN3, 0);
	i34_write_parameter(i34, ZR_DISCTYPE, 0);

	/* p90: VidEntry: set for seq hdr, clear for I-pic GOP or seq hdr:
	   clear if you want to start anywhere in a stream or play
	   streams that have some amount of corruption at the beginning 
	   that causes it to stall when VidEntry is set */
//	i34_change_parameter(i34, ZR_BITSTREAMSELECT, i34->VidEntry, 1);

	/* p90: Black: set for image area to background color when no decoded
	   image.  prevents trash on screen when no active video, i.e.
	   c-state 0x28 NSPB w/o active video. */
	i34_write_parameter(i34, ZR_PLAYBACKMODE,
		(i34->LastPic?0x200:0)|
		(i34->VidFloat?0x40:0)|
		0x10/*black*/ );

	i34_load_subpicture_palette(i34, i34->SubPicturePalette);
	i34_write_parameter(i34, ZR_SPSWITCH, i34->SpSwitch & 0x0007);
	i34_write_parameter(i34, ZR_CAPTIONSWITCH, i34->CaptionSwitch);
	i34_write_parameter(i34, ZR_OSDSWITCH, 0);

	i34_write_parameter(i34, ZR_DVPGEN2, i34->dvpgen2);

	i34_write_parameter(i34, ZR_VIDSID, i34->video_sid);
	i34_write_parameter(i34, ZR_SPSID, 0x00ff); /* no subpics */

	i34_write_parameter(i34, ZR_STARTDISPLAY, 1);	/* validate the new parameters (still usefull ?) */
	
	i34_write_parameter(i34, ZR_VIDTOLERANCE, i34->video_tolerance);
	i34_write_parameter(i34, ZR_VIDSYNCMODE, i34->video_sync_mode);
	i34_write_parameter(i34, ZR_VIDPORTDELAY, i34->video_delay);

	return 0;
}

/* set the sample rate based on the adp microcode loaded,
   so that plltab/cfg can use it *once*. might need error return
*/

static void
i34_set_sample_rate(struct i34_dev_s *i34, int adp_microcode)
{

	i34->adp_ucode = adp_microcode & 7;

	switch (i34->adp_ucode)
	{
		case ADP_MICROCODE_AC3:
		case ADP_MICROCODE_AC36:
			i34->sr = 0; /* 48000 Hz */
			break;
		// case ADP_MICROCODE_CDDA:
		case ADP_MICROCODE_PCM:
		case ADP_MICROCODE_MPEG:
			i34->sr = 1; /* 44100 Hz */
			break;

		/* 32000 Hz would go here. no idea which adp ucode uses it */
		/* user app has to parse mpeg-1 stream to determine
		   the sample rate and set it accordingly, see below
		   -inkling */

		case ADP_MICROCODE_LPCM:
			i34->sr = 0;
		/* 3 = 96000 Hz, but crackles on DVD-Video with LPCM substream,
		   however 0=48000Hz works on DVD-Video with LPCM substream.
		   3 might only be for DVD-Audio discs, but i'm only guessing
		   will have to locate a DVD-Audio LPCM disc for testing
		  -inkling */
			break;
		default:
			dprintk("i34: sample rate not set (%d)\n",
				adp_microcode);
			return;
	}
	dprintk("i34: changing sample rate to %d Hz\n", sr_tab[i34->sr]);

	/* careful inspection of zoran spec indicates you can change
	   the sample rate on the fly and relock amclk as many times as
	   necessary after reset, but pclk can only be locked once after reset
	   since adp_pllcfg only changes amclk, it should be ok.
	   now can add a IOCS_SAMPLE_RATE function for user app
	   -inkling
	*/
	adp_pllcfg(i34, i34->sr);
	
}

static int
i34_select_bitstream_internal(struct i34_dev_s *i34, int bs)
{
	u16 cbs;

	cbs = bs & 0x1F;	/* sanity limit bitstream value */
	
	// dprintk("i34: select_bitstream %02X\n",cbs);
	switch (cbs)
	{
		/* only allow the following bitstream type values -ink */
		case BITSTREAM_DVD_VOB:
		case BITSTREAM_VCD:
		case BITSTREAM_CDDA_PCM:
		case BITSTREAM_VCD_AUX:
		case BITSTREAM_DVD_NAV:
		case BITSTREAM_MPEG12_PS:
		case BITSTREAM_MPEG12_V_ES:
		case BITSTREAM_MPEG12_AV_PES:
		case BITSTREAM_MPEG12_A_ES:
		case BITSTREAM_AC3_A_ES:
		case BITSTREAM_LPCM_A_ES_DVD:
			break;

		default:
			dprintk("i34: invalid bitstream selection (0x%02X)\n",
				cbs);
			return 1;
	}

	i34_write_parameter(i34, ZR_BITSTREAMSELECT,
		(cbs << 11) | ( (i34->VidEntry != 0)  ? 1 : 0));
	dprintk("i34: codec bitstream 0x%02X selected (0x%04X)\n",cbs,cbs<<11);

	return 0;
}



/* 0 bit sets interrupt enabled */
int
i34_enable_interrupts(struct i34_dev_s *i34, u16 zirq)
{
	i34->imr &= ~zirq;
	i34_write( i34, ZR_IMR, i34->imr );
	dprintk("i34: enable irq %04X %04X\n", zirq, i34->imr );
	return 0;
}
/* 1 bit sets interrupt disabled */
int
i34_disable_interrupts(struct i34_dev_s *i34, u16 zirq)
{
	i34->imr |= zirq;
	i34_write(i34, ZR_IMR, i34->imr );
	dprintk("i34: disable zoran irqs %04X %04X\n", zirq, i34->imr );
	return 0;
}

/* main 36700 interrupt handler, using mga vinvsync sleep not ext_irq
   called after mga vinvsync_sleep wakes caller
   it examines the zoran isr, status and data out buffers, stores in status
   caller copies to userland and goes back to vsync_sleep
 */
static void
i34_vsync_isr(struct i34_dev_s *i34, struct mgadvd_status_s *st)
{
	u16 isreg;
	u16 status;
	u16 status0, status1, status2;
	int i;

	if (i34 == NULL)
	{
		printk("i34: isr *i34 NULL\n");
		return;
	}

	/* needed? */
	spin_lock( &i34->status_lock);

	/* first get non-ISR status */
	isreg = i34_read(i34, ZR_ISR);
	st->isreg = isreg;
	st->sclk    = i34_read_sclk(i34);
	st->status0 = status0 = i34_read(i34, ZR_STATUS0);
	st->status1 = status1 = i34_read(i34, ZR_STATUS1);
	st->status2 = status2 = i34_read(i34, ZR_STATUS2);
	
	// dprintk2( "i34: isr %04X ", isreg);
	// dprintk2( " status0 %04X", status0 >> 10);
	// dprintk2( " status1 %04X", status1 & 0x7c07 );
	// dprintk2( " status2 %04X\n", status2 );

	/* report demultiplexed but undecoded bytes in buffer status */
	st->vidbufs = i34_read(i34, ZR_VIDEOBUF );
	st->audbufs = i34_read(i34, ZR_AUDIOBUF );
	st->spbufs = i34_read(i34, ZR_SUBPICBUF );
	
	/* it's actually between *32 and *32+31 */
	// dprintk2( " pending: vid %d", st->vidbufs * 32 + 31);
	// dprintk2( " aud %d", st->audbufs * 32 + 31 );
	// dprintk2( " sub %d\n", st->spbufs * 32 + 31);
		

	/* get same as i34_status for now */
	/* need to test each section added after here */
	return;

	/* see if ADP decoder status has anything */
	/* read more spec! */
	if ( adp_decoder_status(i34, st->adp_status_reply) != 0)
	{
		printk("i34: adp_decoder_status !READY\n");
		return ; //-EBUSY;
	}

	/* now get ISR status */
	/* FIXME: needs more items in status structure to fill in -ink */

	/* NAV data ready */
	if (isreg & ISR_NAVREADY) {
		u16 navbuf[1024];
		u8 navseg;

		navseg = ((status1&1)<<2) + ((status1>>5)&3);
		dprintk( "i34: isr_navready\n" );
		
		/* FIXME: sizeof(navbuf) */
		if (i34_read_nav_data(i34, navseg, navbuf, 1024) != 0) {
			dprintk( "error reading nav data\n" );
		} else {
			st->nav_seg = navseg;
			memcpy(st->nav_buf, navbuf, 1024);
		}
	}

	/* keep the DVPO data */	
	if (isreg & ISR_DVPOBF) {
		/* dvp outbuf full */
		u16 dvpbuf[64];
		u16 valid;
		
		/* spec recommends reading every vsync to prevent overruns */
		for (i=0; i<64; i++) /* p107: 64x16 bit words */
			dvpbuf[i] = i34_read(i34, ZR_DVP_DATA);
		status = i34_read(i34, ZR_STATUS1);

		/* check if not overwritten, DVPVAL = 1 means it's invalid :> */
		valid = (status & DVPVAL) ? 0 : 1;
		if (valid)
		{
			memcpy(st->dvp_buf, dvpbuf, 128);
			st->dvp_valid = 1;
		}

		dprintk( "i34: isr_dvpobf: dvp data 0x%04X (valid %i)\n", dvpbuf[0], valid );
			
		/*  FIXME: update st-> with all of these */
		if ( valid ) {
			u16 dvp_type =	(dvpbuf[0]>>14)  & 0x03;
			u16 nwords =	(dvpbuf[0]>> 7)  & 0x7f;
			u16 type_dep =	(dvpbuf[0]    )  & 0x7f;
			u32 global_delta;

			dprintk2( "dvpo type %i, nwords %i, tdi 0x%02X\n",
				dvp_type, nwords, type_dep );
			
			switch (dvp_type) {
			case 0x00:
				dprintk2( "top_field_flag_of_gop %i\n",
					(dvpbuf[0]>>6) & 1);
				dprintk2( "number_of_displayed_field_gop %i\n",
					dvpbuf[0] & 0x3f);
				break;
				
			case 0x01:
				switch (type_dep) {
				case 0x00:
					dprintk2( "sequence header params\n" );
					break;
				case 0x01:
					dprintk2( "PAUSE_DRIVE req\n" );
					break;
				case 0x02:
					dprintk2( "RESUME_DRIVE req\n" );
					break;
				case 0x05:
					if (nwords < 2) {
						dprintk( "i34: not enough type data\n" );
					}
					global_delta = (((u32)dvpbuf[1])<<16) + dvpbuf[2];
					dprintk( "GLOBAL_DELTA = %u\n", global_delta );
					break;
				default:
					dprintk2( "other seq\n" );
					break;
				}
			case 0x02:
				dprintk2( "picture header params\n" );
				break;
			case 0x03:
			default:
				dprintk2( "no segment\n" );
				break;
			}
		}
	} 
	
	/* this is what brought us here, dont need to know already know */
	if (isreg & ISR_VSYNC) {
		// dprintk2( "i34: isr_vsync\n" );
	}
	if (isreg & ISR_DVPSWIRQ) {
		dprintk( "i34: isr_dvp sw irq\n" );
	}
	if (isreg & ISR_ADPSWIRQ) {
		dprintk( "i34: isr_adp sw irq\n" );
	}
	if (isreg & ISR_SCLKIRQ) {
		dprintk( "i34: isr_sclk irq\n" );
		st->sclkirq = 1;

	}

	/* buffer under/overflow interrupt */
	if (isreg & ISR_CODBUFIRQ) {
		dprintk( "i34: isr_CODBUFIRQ\n" );
		
		status = status0;
		if (status & 0x40) {
			dprintk2( "vid code buf empty\n" );
		}
		if (status & 0x20) {
			dprintk2( "vid code buf full\n" );
		}
		if (status & 0x10) {
			dprintk2( "aud code buf empty\n" );
		}
		if (status & 0x08) {
			dprintk2( "aud code buf full\n" );
		}
		if (status & 0x04) {
			dprintk2( "subpic code buf full\n" );
		}
	}
	if (isreg & 0x04af) {
		dprintk( "i34: reserved ISR bit\n" );
	}
	
	spin_unlock( &i34->status_lock);
}
	



/*********************************************************************
	public api
*********************************************************************/

struct i34_dev_s*
i34_attach(struct mga_dev_s *mga, int *errcode)
{
	struct i34_dev_s *i34;
	u32 id;

	dprintk("i34: i34_attach\n");
	
	*errcode = 0;


	i34 = kmalloc(sizeof(*i34), GFP_KERNEL);
	if (i34 == NULL) {
		*errcode = -ENOMEM;
		return NULL;
	}

	memset(i34, 0, sizeof(*i34));
	i34->mga = mga;
	i34->codec_interface = mga->codec;

	i34->imr = 0xFFFF;

	
	spin_lock_init( &i34->lock );
	spin_lock_init( &i34->status_lock );

	i34_reset_struct(i34);


	mgabus_init_client( &i34->cbus, BUSCLIENT_BASE,
		"i34 host interface", 0x34, i34 );
	i34->cbus.set_dir = i34_cbus_set_dir;
	i34->cbus.get_dir = i34_cbus_get_dir;
	mgabus_attach( mga->codec_bus, &i34->cbus );
	

	mgabus_init_client( &i34->vbus, BUSCLIENT_VIDEO,
		"i34 video out", 0x34, i34 );
	i34->vbus.super.set_dir = i34_vbus_set_dir;
	i34->vbus.super.get_dir = i34_vbus_get_dir;
	i34->vbus.get_format = i34_vbus_get_format;
	i34->vbus.nr_inputs = 1;
	i34->vbus.active_input = 0;
	mgabus_attach( mga->vip_bus, &i34->vbus.super );


	if (i34_reset_internal(i34, 1) != 0) {
		id = -1;
	} else {
		id = i34_read(i34, ZR_DEVICE_ID);
	}
	
	/* put 36700 in low power mode until it's needed */
	i34_powerdown(i34);
	
	/* we should obtain: 0x34 m.s. bits, rev in l.s. bits p. 112 ZR HB */
	/* warning: this test is arch. dependent ! */
#ifndef __LITTLE_ENDIAN
#error "This driver may not work with big endian architectures:"
#error "contact dbarth@besancon.net or leonvs@iae.nl if you want to help"
#else
	if (((id>>8) & 0xFF) == 0x34) {
		dprintk("i34: Found ZR367x0 chip ID 0x%04X\n", id );
	} else {
		dprintk("i34: no chipset found (codec response: 0x%04X)\n", id);
		mgabus_detach(mga->vip_bus, &i34->vbus.super);
		mgabus_detach(mga->codec_bus, &i34->cbus);
		kfree( i34 );
		*errcode = -ENODEV;
		return NULL;
	}
#endif

/* disable all 36700 irqs (no mga ext_irq, using mga vinvsync instead) */
	i34_disable_interrupts(i34, 0xFFFF);


/* set mga vinvsync irq to wake up a user thread calling mgacodec_vsync_sleep */
//	mga_enable_irq( mga, MGA_IRQ_VINVSYNC,
//		(mga_interrupt_handler_t)mgacodec_vsync_irq);
//	dprintk("i34: enabled MGA VINVSYNC IRQ handler\n");
	return i34;
}


void
i34_deattach(struct i34_dev_s *i34)
{
	if ( i34 != NULL ) {
		mgabus_detach(i34->mga->vip_bus, &i34->vbus.super);
		mgabus_detach(i34->mga->codec_bus, &i34->cbus);
	
		i34_disable_interrupts( i34, 0xFFFF);

		/* put the 36700 into low power mode */
		i34_powerdown(i34);

		kfree(i34);
	}
}


int
i34_video_standard(struct i34_dev_s *i34, int standard)
{
	i34->VideoStandard = standard;
	if (i34_video_standard_internal( i34, standard ) != 0)
		return -EINVAL;
	return 0;
}


int
i34_change_video_sid(struct i34_dev_s *i34, int sid)
{
	dprintk("i34: i34_change_video_SID 0x%02X\n", sid);

	i34->video_sid = sid;
	i34_write_parameter(i34, ZR_VIDSID, i34->video_sid);
	return 0;
}


int
i34_change_subpic_sid(struct i34_dev_s *i34, int sid)
{
	dprintk("i34: i34_change_subpic_SID 0x%02X\n", sid);

	i34->subpic_sid = sid;
	i34_write_parameter(i34, ZR_SPSID, i34->subpic_sid);
	return 0;
}

int
i34_change_audio_sid(struct i34_dev_s *i34, int sid)
{
	dprintk("i34: i34_change_audio_SID 0x%02X\n", sid);

	adp_simple_command(i34, ADP_STOPF);
	/* see p. 258 */
	i34_write_parameter(i34, ZR_AUDSID, 0x00FF);
	i34_wait(5000);
	/* FIXME: if audioStandard like AC3 or PCM changes
		ADP_CFG
		ADP_SPDIFCS
		ADP_audioStandardCFG
	*/
	adp_audio_sync(i34, i34->audio_sync_mode);
	adp_simple_command(i34, ADP_PLAY);
	i34->audio_sid = sid;
	i34_write_parameter(i34, ZR_AUDSID, sid);
	adp_simple_command(i34, ADP_UNMUTE);
	return 0;
}


int
i34_audio_sync(struct i34_dev_s *i34, int sync)
{
	dprintk("i34: audio_sync %d\n",sync);
	i34->audio_sync_mode = sync & 3;
	if (adp_audio_sync(i34, i34->audio_sync_mode) != 0)
		return -EINVAL;
	return 0;
}


int
i34_video_sync(struct i34_dev_s *i34, int sync)
{
	dprintk("i34: video_sync %d\n",sync);
	i34->video_sync_mode = sync & 3;
	i34_write_parameter(i34, ZR_VIDSYNCMODE, i34->video_sync_mode);
	return 0;
}


int
i34_dvp_boot(struct i34_dev_s *i34, int dvp_microcode)
{
	if (i34_dvp_boot_internal( i34, dvp_microcode ) != 0)
		return -EINVAL;
	return 0;
}


int
i34_adp_boot(struct i34_dev_s *i34, int adp_microcode)
{
	/* set sample rate based on microcode type -ink */
	i34_set_sample_rate(i34, adp_microcode);
	
	if (i34_adp_boot_internal( i34, adp_microcode ) != 0)
		return -EINVAL;
	
	if (i34_config_audio_stage2(i34) != 0)
		return -ENXIO;
	
	return 0;
}


int
i34_select_bitstream(struct i34_dev_s *i34, int bs)
{
	if (i34_select_bitstream_internal( i34, bs ) != 0)
		return -EINVAL;
	return 0;
}

/* start playback */

int
i34_start(struct i34_dev_s *i34)
{
	dprintk("i34: START NSPB\n");
	// i34_check_state_configured( i34 );

	/* wonder what happens when you issue two adp_play commands
	   with no adp_stop between? -ink */

	adp_simple_command(i34, ADP_PLAY);

	if (i34_send_command(i34, I34_CMD_START, 0) != 0)
		return -EINVAL;

	return 0;
}

/* pause playback */
int
i34_pause(struct i34_dev_s *i34)
{
	dprintk("i34: PAUSE\n");
	if (i34_send_command(i34, I34_CMD_PAUSE, 0) != 0)
		return -EINVAL;
	return 0;
}


/* continue playback from pause, step, slow.
   fast needs different approach */
int
i34_continue(struct i34_dev_s *i34)
{
	dprintk("i34: CONTINUE\n");
	if (i34_send_command(i34, I34_CMD_CONT, 0) != 0)
		return -EINVAL;
	return 0;
}


/* single step frames */
int
i34_step(struct i34_dev_s *i34)
{
	dprintk("i34: STEP\n");
	if (i34_send_command(i34, I34_CMD_STEP, 0) != 0)
		return -EINVAL;
	return 0;
}

/* slow motion */
int
i34_slowmo(struct i34_dev_s *i34, int sss)
{
	/* SSS is bottom 3 bits of parameter, waits SSS-1 frame periods,
	   2 is speediest slow-mo mode. so it's all offset from there,
	   with 1 being the slowest slowmo. -ink
	 */
	dprintk("i34: SLOMO %d\n",sss);

	if (i34_send_command(i34, I34_CMD_SLOW, (sss & 7)) != 0)
		return -EINVAL;
	return 0;
}


/* fast forward */
int
i34_fastfwd(struct i34_dev_s *i34)
{
	/* only 1-4 allowed for fast forward parameter.
	   it doesn't seem to work for MPEG, VCD or DVD microcodes.
	   p154 section 6.12 may give the reason why. some microcodes
	   don't support fast_search. have to use the SCAN microcode.
	   (was wondering what that microcode was for)

	   FIXME: follow the steps in sections 6.12 and 6.13.
		I&P mode doesn't require change to userland player,
		I-only mode would require changes to userland player.
	  -inkling

	stop playback, load scan microcode, clear some dvpgen bits,
	start playback, some words on limiting bitstream feed to I type??
	might need some more host/player work parsing stream
	*/

	dprintk("i34: FFWD\n");

	if (i34_send_command(i34, I34_CMD_FAST, 1) != 0)
		return -EINVAL;
	return 0;
}


/* i34_stop flag:
   0 continue display of last frame and stop without trashing codec buffers
     this is used for non-seamless resync in DVD/VOB playback
   1 display background, stop and reset codec buf ptrs (full-stop)
 */
int
i34_stop(struct i34_dev_s *i34, int flag)
{
	unsigned int cstate = 0;
	unsigned int csloop = 0;

	dprintk("i34: STOP %d\n", flag);
	/* stop video processing */
	if (i34_send_command(i34, I34_CMD_END, flag) != 0)
		return -EINVAL;
	/* stop audio processing */
	adp_simple_command(i34, ADP_STOPF);

	/* need to wait for cstate to go idle? */
	/* get current cstate */
	do
	{
	    i34_state(i34, &cstate);
	    csloop++;
	    /* dont care about the active decoded image flag */
	    cstate &= 0x37;
	    if ( cstate == ZR_CSTATE_IDLE ) break;
	} while (csloop < 100);

	/* end w/o flush finishes in 1 loop, need to test how many for flush
	   or force a possibly long loop here that needs to give back
           time to the scheduler
	 */
	dprintk("i34: stop took %u loops\n", csloop);

	/* if displaying background (full-stop) reset codec buffer pointers */
	if (flag != 0)
	{
		dprintk("i34: codecbufptr reset\n");
		mgacodec_stop_transfer(i34->codec_interface);
	}

	return 0;
}


/* this is of dubious value, sync/flush method has changed now */
#ifdef NEVER
/* flag non zero to issue stopf on adp too */
int
i34_stop_flush(struct i34_dev_s *i34, int flag)
{
	int i;

	if ( mgacodec_get_transfer(i34->codec_interface) )
	{
		/* wait for 'empty' codec buffer */
		for (i=0; i<HZ; i++)
		{
			if (mgacodec_is_stalled( i34->codec_interface ) )
				break;
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout( 1 );
		}
	}

	/* the above is almost enough but not quite. probably should
	   check for some 36700 status bits too.  may need the waits
	   and checks after seqend and another after stopf, too.
	  -inkling
	*/

	if (i34_send_command(i34, I34_CMD_SEQEND, 0) != 0)
		return -EINVAL;

	if (flag != 0) adp_simple_command(i34, ADP_STOPF);

	return 0;
}
#endif

/* flag non zero for non seamless resync, otherwise seamless */
int
i34_resync(struct i34_dev_s *i34, int flag)
{
	dprintk("i34: RESYNC %d\n",flag);

	if (flag != 0)
	{
	/* non-seamless resync:
		1) pause playback to stop g200 transfer near frame boundary
		2) end playback
		3) start playback
	*/
		/* stop on frame boundary */
		//i34_pause(i34);
		/* keep last frame up, dont reset codebuf, stop/start */
		i34_stop(i34, 0);
		i34_audio_sync(i34, 3);
		i34_start(i34);
		return 0;
	} else
	{
	/* seamless resync:
		1) adp sync off
		2) vsyncmode/vlock off
		3) vsyncmode/vlock on to sync SCLK to next SCR in bitstream
		4) adp syncs to SCLK then switches to audio master mode
		not sure if 1 and 4 are needed -ink

	  FIXME: this breaks a later non-seamless resync and halts playback

		trying video only for seamless, audio master
	*/

//		i34_audio_sync(i34, 0);
//		i34_video_sync(i34, 2);
//		i34_video_sync(i34, 3);
//		i34_audio_sync(i34, 3);
	}
	return 0;
}

/* this is quick status read. it reads a subset of total status
   it doesn't get anything from isr or output buffers adp dvp nav
   caller (mga_dvd_ioctl) copies this minimal status to userland

   I34_IOCGSTATUS
 */
int
i34_status(struct i34_dev_s *i34, struct mgadvd_status_s *st)
{
	/* clear immediate status structure, needed? */
	// memset(st, 0, sizeof(mgadvd_status_s));

	st->sclk    = i34_read_sclk(i34);

	st->status0 = i34_read(i34, ZR_STATUS0);
	st->status1 = i34_read(i34, ZR_STATUS1);
	st->status2 = i34_read(i34, ZR_STATUS2);

	st->vidbufs = i34_read(i34, ZR_VIDEOBUF);
	st->audbufs = i34_read(i34, ZR_AUDIOBUF);
	st->spbufs  = i34_read(i34, ZR_SUBPICBUF);

	return 0;
}

/* user creates a thread to loop on this via ioctl I34_IOCGVSYNCSTATUS
   it waits for mgacodec_vsync_sleep to return, then reads all the status
   it can and stores it into caller supplied mgadvd_status_s structure.
   caller (mga_dvd_ioctl) copies that to userland

   I34_IOCGVSYNCSTATUS
 */
int
i34_vsync_status(struct i34_dev_s *i34, struct mgadvd_status_s *st)
{
	mgacodec_vsync_sleep(i34->codec_interface);
	i34_vsync_isr( i34, st );
	return 0;
}


/* it looks like a player key to me.
 * we shouldnt depend on hardware keys -ink
 */
int
i34_player_key(struct i34_dev_s *i34)
{
	/* CSS voodoo?
	i34_change_parameter(i34, 0xA0, 0x80, 0x80);
	i34_change_parameter(i34, 0xA0, 0x00, 0x80);
	i34_change_parameter(i34, 0xA0, 0x0F, 0x0F);
	i34_change_parameter(i34, 0xA4, 0x00, 0x40);
	i34_change_parameter(i34, 0xA4, 0x01, 0x07);
	*/
	return 0;
}

int
i34_reset(struct i34_dev_s *i34)
{
	int cstate;

	i34_reset_struct(i34);

	/* According to Annex C p. 248 of the ZR36710 Handbook */

	/* Stage 1 : Reset */
	dprintk("i34: stage 1\n");
	if (i34_reset_internal(i34, 0) != 0)
		return -EBUSY;

	i34_state(i34, &cstate);
	if ( cstate != ZR_CSTATE_INIT_PCLK ) {
		dprintk( "i34: not in init_pclk state after reset\n" );
	}

	/* State = init_pclk */

	/* NumDRAMDev = 0; CodeSource = 1; Decryption bypass */
	i34_write_parameter(i34, ZR_SYSCONFIG, (0<<13) | (1<<12) | (i34->CodeBurstLen<<4) | (0<<3));

	/* can't depend on this hardware player key staying valid -ink */
	/* i34_player_key(i34) */

	i34_write_parameter(i34, ZR_SDCONFIG, 0);

	/* no interrupts */
	i34_disable_interrupts(i34, 0xFFFF);

	if (i34_config_audio_stage1(i34) != 0)
		return -EBUSY;

	if (i34_init_video_port(i34) != 0)
		return -EBUSY;

	if (i34_load_init_file(i34) != 0)
		return -EBUSY;
		
	i34_configure_display(i34);

	mgabus_notify( i34->mga->vip_bus, BUSCODE_FORMAT );
	
	return 0;
}


////////////////////////////////





int
i34_open(struct mga_dev_s *mga)
{
	struct i34_dev_s *i34 = mga->i34;
	
	if (i34 == NULL)
		return -ENXIO;
	
	if (mga_card_lock(mga)!=0)
		return -EBUSY;

	if (i34_reset(i34) != 0) {
		i34_powerdown(i34);
		mga_card_unlock(mga);
		return -EBUSY;
	}
	
	MOD_INC_USE_COUNT;
	return 0;
}


void
i34_close(struct mga_dev_s *mga)
{

	struct i34_dev_s *i34 = mga->i34;
	
	/* "release" the DAC : without this your CD player will stay muted... */
	adp_config_gpaio(i34, 3, 0);
	
	/* put the 36700 into low power mode */
	i34_powerdown(i34);

	mga_card_unlock(mga);
	MOD_DEC_USE_COUNT;
}


#ifdef MODULE
int init_module(void)
#else
int init_i34(void) //??
#endif
{
	struct mga_dev_s *mga;
	int		 i;
	int		err;
	int	active = 0;

	printk(KERN_INFO "Matrox G200 DVD Add-on driver\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {

		err = 0;	
		/* detect dvd module */	
		mga->i34 = i34_attach( mga, &err );
		if (err && (mga->i34 != NULL)) {
			i34_deattach(mga->i34);
			mga->i34 = NULL;
		}
		if (mga->i34 == NULL)
			mga_release_card(mga);
		else
			active++;
	}

	if(active == 0) {
		return -ENXIO;
	}
	return 0;
}


#ifdef MODULE
void cleanup_module(void)
{
	int i = 0;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {

		if ( mga->i34 != NULL ) {
			i34_deattach(mga->i34);
			mga->i34 = NULL;
			mga_release_card(mga);
		}
		mga_release_card(mga);
	}
}
#endif

