/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200,G400 and Rainbow Runner-G series
 *
 * This module is an interface to the KS0127 video decoder chip.
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Gerard v.d. Horst
 *	Leon van Stuivenberg <leonvs@iae.nl>
 *	Gernot Ziegler <gz@lysator.liu.se>
 * 
 * Version History:
 * V1.0 Ryan Drake	   Initial version by Ryan Drake
 * V1.1 Gerard v.d. Horst  Added some debugoutput, reset the video-standard
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include "i2cdef.h"
#include "ks0127.h"

#define dprintk     if (debug) printk

#define KS0127_DEVNAME  "ks0127"

/* i2c identification */
#define I2C_KS0127_ADDON   0xD8
#define I2C_KS0127_ONBOARD 0xDA

#define KS_TYPE_UNKNOWN	0
#define KS_TYPE_0122S	1
#define KS_TYPE_0127	2
#define KS_TYPE_0127B	3

/* ks0127 control registers */
#define KS_STAT     0x00
#define KS_CMDA     0x01
#define KS_CMDB     0x02
#define KS_CMDC     0x03
#define KS_CMDD     0x04
#define KS_HAVB     0x05
#define KS_HAVE     0x06
#define KS_HS1B     0x07
#define KS_HS1E     0x08
#define KS_HS2B     0x09
#define KS_HS2E     0x0a
#define KS_AGC      0x0b
#define KS_HXTRA    0x0c
#define KS_CDEM     0x0d
#define KS_PORTAB   0x0e
#define KS_LUMA     0x0f
#define KS_CON      0x10
#define KS_BRT      0x11
#define KS_CHROMA   0x12
#define KS_CHROMB   0x13
#define KS_DEMOD    0x14
#define KS_SAT      0x15
#define KS_HUE      0x16
#define KS_VERTIA   0x17
#define KS_VERTIB   0x18
#define KS_VERTIC   0x19
#define KS_HSCLL    0x1a
#define KS_HSCLH    0x1b
#define KS_VSCLL    0x1c
#define KS_VSCLH    0x1d
#define KS_OFMTA    0x1e
#define KS_OFMTB    0x1f
#define KS_VBICTL   0x20
#define KS_CCDAT2   0x21
#define KS_CCDAT1   0x22
#define KS_VBIL30   0x23
#define KS_VBIL74   0x24
#define KS_VBIL118  0x25
#define KS_VBIL1512 0x26
#define KS_TTFRAM   0x27
#define KS_TESTA    0x28
#define KS_UVOFFH   0x29
#define KS_UVOFFL   0x2a
#define KS_UGAIN    0x2b
#define KS_VGAIN    0x2c
#define KS_VAVB     0x2d
#define KS_VAVE     0x2e
#define KS_CTRACK   0x2f
#define KS_POLCTL   0x30
#define KS_REFCOD   0x31
#define KS_INVALY   0x32
#define KS_INVALU   0x33
#define KS_INVALV   0x34
#define KS_UNUSEY   0x35
#define KS_UNUSEU   0x36
#define KS_UNUSEV   0x37
#define KS_USRSAV   0x38
#define KS_USREAV   0x39
#define KS_SHS1A    0x3a
#define KS_SHS1B    0x3b
#define KS_SHS1C    0x3c
#define KS_CMDE     0x3d
#define KS_VSDEL    0x3e
#define KS_CMDF     0x3f
#define KS_GAMMA0   0x40
#define KS_GAMMA1   0x41
#define KS_GAMMA2   0x42
#define KS_GAMMA3   0x43
#define KS_GAMMA4   0x44
#define KS_GAMMA5   0x45
#define KS_GAMMA6   0x46
#define KS_GAMMA7   0x47
#define KS_GAMMA8   0x48
#define KS_GAMMA9   0x49
#define KS_GAMMA10  0x4a
#define KS_GAMMA11  0x4b
#define KS_GAMMA12  0x4c
#define KS_GAMMA13  0x4d
#define KS_GAMMA14  0x4e
#define KS_GAMMA15  0x4f
#define KS_GAMMA16  0x50
#define KS_GAMMA17  0x51
#define KS_GAMMA18  0x52
#define KS_GAMMA19  0x53
#define KS_GAMMA20  0x54
#define KS_GAMMA21  0x55
#define KS_GAMMA22  0x56
#define KS_GAMMA23  0x57
#define KS_GAMMA24  0x58
#define KS_GAMMA25  0x59
#define KS_GAMMA26  0x5a
#define KS_GAMMA27  0x5b
#define KS_GAMMA28  0x5c
#define KS_GAMMA29  0x5d
#define KS_GAMMA30  0x5e
#define KS_GAMMA31  0x5f
#define KS_GAMMAD0  0x60
#define KS_GAMMAD1  0x61
#define KS_GAMMAD2  0x62
#define KS_GAMMAD3  0x63
#define KS_GAMMAD4  0x64
#define KS_GAMMAD5  0x65
#define KS_GAMMAD6  0x66
#define KS_GAMMAD7  0x67
#define KS_GAMMAD8  0x68
#define KS_GAMMAD9  0x69
#define KS_GAMMAD10 0x6a
#define KS_GAMMAD11 0x6b
#define KS_GAMMAD12 0x6c
#define KS_GAMMAD13 0x6d
#define KS_GAMMAD14 0x6e
#define KS_GAMMAD15 0x6f
#define KS_GAMMAD16 0x70
#define KS_GAMMAD17 0x71
#define KS_GAMMAD18 0x72
#define KS_GAMMAD19 0x73
#define KS_GAMMAD20 0x74
#define KS_GAMMAD21 0x75
#define KS_GAMMAD22 0x76
#define KS_GAMMAD23 0x77
#define KS_GAMMAD24 0x78
#define KS_GAMMAD25 0x79
#define KS_GAMMAD26 0x7a
#define KS_GAMMAD27 0x7b
#define KS_GAMMAD28 0x7c
#define KS_GAMMAD29 0x7d
#define KS_GAMMAD30 0x7e
#define KS_GAMMAD31 0x7f


/****************************************************************************
* mga_dev : represents one ks0127 chip.
****************************************************************************/

struct adjust {
	int	contrast;
	int	bright;
	int	hue;
	int	ugain;
	int	vgain;
};

struct ks0127 {
	struct i2c_client *client;
	unsigned char	addr;
	int		format_width;
	int		format_height;
	int		cap_width;
	int		cap_height;
	int		norm;
	int		ks_type;
	u8 		regs[256];
};


static int	ks0127_proc_read(char *buf, char **start, off_t offset, int len, int *eof, void *data);
static int	ks0127_proc_write(struct file *file, const char *buf, unsigned long count, void *data);


static int debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

static u8 reg_defaults[ 64 ];


/* We need to manually read because of a bug in the KS0127 chip.
 *
 * An explanation from kayork@mail.utexas.edu:
 *
 * During I2C reads, the KS0127 only samples for a stop condition
 * during the place where the acknoledge bit should be. Any standard
 * I2C implementation (correctly) throws in another clock transition
 * at the 9th bit, and the KS0127 will not recognize the stop condition
 * and will continue to clock out data.
 *
 * So we have to do the read ourself.  Big deal.
           workaround in i2c-algo-ks
 */


static u8 ks0127_read( struct ks0127* ks, u8 reg )
{
	char val = 0;

	if ( i2c_master_send(ks->client, &reg, 1 ) != 1 ) {
		dprintk( "ks0127_write error\n" );
	}
	if ( i2c_master_recv(ks->client, &val, 1 ) != 1 ) {
		dprintk( "ks0127_write error\n" );
	}
	return val;
}


static void ks0127_write( struct ks0127* ks, u8 reg, u8 val )
{
	char msg[] = { reg, val };
	
	if ( i2c_master_send(ks->client, msg, sizeof(msg) ) != sizeof(msg) ) {
		dprintk( "ks0127_write error\n" );
	}
	ks->regs[ reg ] = val;
}


/* generic bit-twiddling */
static void ks0127_and_or( struct ks0127* ks, u8 reg, u8 and_v, u8 or_v )
{
        u8 val = ks->regs[ reg ];
        val = (val & and_v) | or_v;
        ks0127_write( ks, reg, val );
}



/****************************************************************************
* ks0127 private api
****************************************************************************/
static void ks0127_reset( struct ks0127* ks )
{
	int i;
	u8* table = reg_defaults;
	
	ks->ks_type = KS_TYPE_UNKNOWN;

	table[KS_CMDA]     = 0x2c;  /* VSE=0, CCIR 601, autodetect standard */
	table[KS_CMDB]     = 0x12;  /* VALIGN=0, AGC control and input */
	table[KS_CMDC]     = 0x00;  /* Test options */
	table[KS_CMDD]     = 0x01;  /* clock & input select, write 1 to PORTA */
	table[KS_HAVB]     = 0x00;  /* HAV Start Control */
	table[KS_HAVE]     = 0x00;  /* HAV End Control */
	table[KS_HS1B]     = 0x10;  /* HS1 Start Control */
	table[KS_HS1E]     = 0x00;  /* HS1 End Control */
	table[KS_HS2B]     = 0x00;  /* HS2 Start Control */
	table[KS_HS2E]     = 0x00;  /* HS2 End Control */
	table[KS_AGC]      = 0x53;  /* Manual setting for AGC */
	table[KS_HXTRA]    = 0x00;  /* Extra Bits for HAV and HS1/2 */
	table[KS_CDEM]     = 0x00;  /* Chroma Demodulation Control */ 
	table[KS_PORTAB]   = 0x0f;  /* port B is input, port A output GPPORT */
	table[KS_LUMA]     = 0x01;  /* Luma control */
	table[KS_CON]      = 0x00;  /* Contrast Control */
	table[KS_BRT]      = 0x00;  /* Brightness Control */
	table[KS_CHROMA]   = 0x2a;  /* Chroma control A */ 
	table[KS_CHROMB]   = 0x90;  /* Chroma control B */
	table[KS_DEMOD]    = 0x00;  /* Chroma Demodulation Control & Status */
	table[KS_SAT]      = 0x00;  /* Color Saturation Control*/
	table[KS_HUE]      = 0x00;  /* Hue Control */
	table[KS_VERTIA]   = 0x00;  /* Vertical Processing Control A */
	table[KS_VERTIB]   = 0x12;  /* Vertical Processing Control B, luma 1 line delayed */
	table[KS_VERTIC]   = 0x0b;  /* Vertical Processing Control C */
	table[KS_HSCLL]    = 0x00;  /* Horizontal Scaling Ratio Low */
	table[KS_HSCLH]    = 0x00;  /* Horizontal Scaling Ratio High */
	table[KS_VSCLL]    = 0x00;  /* Vertical Scaling Ratio Low */
	table[KS_VSCLH]    = 0x00;  /* Vertical Scaling Ratio High */
	table[KS_OFMTA]    = 0x33;  /* 8 bit YCbCr 4:2:2 output */
	table[KS_OFMTB]    = 0x00;  /* Output Control B */ 
	table[KS_VBICTL]   = 0x5d;  /* VBI Decoder Control; 4bit fmt: avoid Y overflow */
	table[KS_CCDAT2]   = 0x00;  /* Read Only register */
	table[KS_CCDAT1]   = 0x00;  /* Read Only register */
	table[KS_VBIL30]   = 0xa8;  /* VBI data decoding options */
	table[KS_VBIL74]   = 0xaa;  /* VBI data decoding options */
	table[KS_VBIL118]  = 0x2a;  /* VBI data decoding options */
	table[KS_VBIL1512] = 0x00;  /* VBI data decoding options */
	table[KS_TTFRAM]   = 0x00;  /* Teletext frame alignment pattern */
	table[KS_TESTA]    = 0x00;  /* test register, shouldn't be written */
	table[KS_UVOFFH]   = 0x00;  /* UV Offset Adjustment High */
	table[KS_UVOFFL]   = 0x00;  /* UV Offset Adjustment Low */
	table[KS_UGAIN]    = 0x00;  /* U Component Gain Adjustment */
	table[KS_VGAIN]    = 0x00;  /* V Component Gain Adjustment */
	table[KS_VAVB]     = 0x07;  /* VAV Begin */    
	table[KS_VAVE]     = 0x00;  /* VAV End */
	table[KS_CTRACK]   = 0x00;  /* Chroma Tracking Control */
	table[KS_POLCTL]   = 0x41; //0x01;  /* Timing Signal Polarity Control */
	table[KS_REFCOD]   = 0x80; /*0xa4;*/  /* Reference Code Insertion Control */
	table[KS_INVALY]   = 0x10;  /* Invalid Y Code */
	table[KS_INVALU]   = 0x80;  /* Invalid U Code */
	table[KS_INVALV]   = 0x80;  /* Invalid V Code */
	table[KS_UNUSEY]   = 0x10;  /* Unused Y Code */
	table[KS_UNUSEU]   = 0x80;  /* Unused U Code */
	table[KS_UNUSEV]   = 0x80;  /* Unused V Code */
	table[KS_USRSAV]   = 0x00;  /* reserved */
	table[KS_USREAV]   = 0x00;  /* reserved */
	table[KS_SHS1A]    = 0x00;  /* User Defined SHS1 A */ 
	table[KS_SHS1B]    = 0x80;  /* User Defined SHS1 B, ALT656=1 on 0127B */
	table[KS_SHS1C]    = 0x00;  /* User Defined SHS1 C */
	table[KS_CMDE]     = 0x00;  /* Command Register E */
	table[KS_VSDEL]    = 0x00;  /* VS Delay Control */ 
	table[KS_CMDF]     = 0x02;  /* Command Register F, update -immediately- (there might come no vsync)*/

	dprintk( "ks0127: reset\n" );
	udelay(1000);

	/* initialize all registers to known values (except STAT, 0x21, 0x22, TEST and 0x38,0x39 ) */

	for(i = 1; i < 33; i++ )
		ks0127_write( ks, i, table[i] );

	for(i = 35; i < 40; i++)
		ks0127_write( ks, i, table[i] );

	for(i = 41; i < 56; i++)
		ks0127_write( ks, i, table[i] );

	for(i = 58; i < 64; i++)
		ks0127_write( ks, i, table[i] ); 


	if ((ks0127_read(ks, KS_STAT) & 0x80) == 0) {
		ks->ks_type = KS_TYPE_0122S;
		dprintk("ks0127: ks0122s Found\n");
		return;
	}

	switch(ks0127_read(ks, KS_CMDE) & 0x0f) {

	case 0:
		ks->ks_type = KS_TYPE_0127;
		dprintk("ks0127: ks0127 found\n");
		break;

	case 9:
		ks->ks_type = KS_TYPE_0127B;
		dprintk("ks0127: ks0127B Revision A found\n");
		break;
		
	default:
		dprintk("ks0127: unknown revision\n");
		break;
	}
}



void ks0127_getcrop(struct ks0127 *ks, int *xstart, int *xend, int *ystart, int *yend)
{
	*xstart = ((ks0127_read(ks, KS_HXTRA) & 0xe0) << 3) + 
		ks0127_read(ks, KS_HAVB);

	*xend = ((ks0127_read(ks, KS_HXTRA) & 0x1c) << 6) + 
		ks0127_read(ks, KS_HAVE);

	if ((*xstart) & 0x400) 
		*xstart |= ~0x3ff;

	if ((*xend) & 0x400) 
		*xend |= ~0x3ff;

	*ystart = ks0127_read(ks, KS_VAVB) >> 2;
	*yend = ks0127_read(ks, KS_VAVE);
	dprintk("ks0127: ystart=%d yend=%d\n", *ystart, *yend);

	*ystart = (*ystart - 2) * 2;
	*yend = (*yend - 2) * 2;
}


static int ks0127_command(struct i2c_client *client, unsigned int cmd, void *arg )
{
	struct ks0127 *ks = (struct ks0127*)client->data;

	int		*iarg = (int*)arg;
	int		size;
	int		xstart;
	int		ystart;
	int		xend;
	int		yend;
	int		hav;
	int		vav, h1, h2, i;

	struct mga_procread *prd;
	struct mga_procwrite *pwr;


	if( !ks )
		return -ENODEV;

	switch (cmd) {

	case KS0127_RESET:
		dprintk( "ks0127: command KS0127_RESET\n" );
		ks0127_reset( ks );
		break;

	case KS0127_SET_INPUT:
		switch( *iarg ) {
		case KS_INPUT_COMPOSITE:
			dprintk( "ks0127: command KS0127_SET_INPUT: INPUT_COMPOSITE\n" );
			ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); /* autodetect 50/60 Hz */
			ks0127_and_or( ks, KS_CMDA,   ~0x40, 0x00 ); /* VSE=0 */
			ks0127_and_or( ks, KS_CMDB,   0xb0, 0x02 ); /* set input line = AY2, VALIGN=0 */
			ks0127_and_or( ks, KS_CMDC,   0x70, 0x0a ); /* non-freerunning mode */
			ks0127_and_or( ks, KS_CMDD,   0x03, 0x00 ); /* analog input */
			ks0127_and_or( ks, KS_CTRACK, 0xcf, 0x00 ); /* enable chroma demodulation */
	//			ks0127_and_or( ks, KS_LUMA,   0x00, (reg_defaults[KS_LUMA]&0xf7)|0x04 ); /* chroma trap */
	//			ks0127_and_or( ks, KS_VERTIA, 0x08, 0xb3 ); /* manual luma comb [1 0 0], LPF/comb disabled */
	//			ks0127_and_or( ks, KS_VERTIC, 0x0f, 0xd0 ); /* manual chroma comb [0 1 0] */

			ks0127_and_or( ks, KS_LUMA,   0x00, (reg_defaults[KS_LUMA])|0x0c ); /* chroma trap, HYBWR=1 */
			ks0127_and_or( ks, KS_VERTIA, 0x08, 0x81 ); /* scaler fullbw, luma comb off */
			ks0127_and_or( ks, KS_VERTIC, 0x0f, 0x90 ); /* manual chroma comb .25 .5 .25 */

			ks0127_and_or( ks, KS_CHROMB, 0x0f, 0x90 ); /* chroma path delay */
					
			ks0127_write( ks, KS_UGAIN, reg_defaults[KS_UGAIN] );
			ks0127_write( ks, KS_VGAIN, reg_defaults[KS_VGAIN] );
			ks0127_write( ks, KS_UVOFFH, reg_defaults[KS_UVOFFH] );
			ks0127_write( ks, KS_UVOFFL, reg_defaults[KS_UVOFFL] );
		break;

		case KS_INPUT_SVIDEO:
		dprintk( "ks0127: command KS0127_SET_INPUT: INPUT_SVIDEO\n" );
			ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); /* autodetect 50/60 Hz */
			ks0127_and_or( ks, KS_CMDA,   ~0x40, 0x00 ); /* VSE=0 */
			ks0127_and_or( ks, KS_CMDB,   0xb0, 0x08 ); /* set input line = AY0-AC0, VALIGN=0 */
			ks0127_and_or( ks, KS_CMDC,   0x70, 0x0a ); /* non-freerunning mode */
			ks0127_and_or( ks, KS_CMDD,   0x03, 0x00 ); /* analog input */
			ks0127_and_or( ks, KS_CTRACK, 0xcf, 0x00 ); /* enable chroma demodulation */
			ks0127_and_or( ks, KS_LUMA,   0x00, reg_defaults[KS_LUMA] );
			ks0127_and_or( ks, KS_VERTIA, 0x08, (reg_defaults[KS_VERTIA]&0xf0)|0x01 ); /* disable luma comb */
			ks0127_and_or( ks, KS_VERTIC, 0x0f, reg_defaults[KS_VERTIC]&0xf0 );
		
			ks0127_and_or( ks, KS_CHROMB, 0x0f, reg_defaults[KS_CHROMB]&0xf0 );
		
			ks0127_write( ks, KS_UGAIN, reg_defaults[KS_UGAIN] );
			ks0127_write( ks, KS_VGAIN, reg_defaults[KS_VGAIN] );
			ks0127_write( ks, KS_UVOFFH, reg_defaults[KS_UVOFFH] );
			ks0127_write( ks, KS_UVOFFL, reg_defaults[KS_UVOFFL] );
		break;

		case KS_INPUT_TUNER:
		dprintk( "ks0127: command KS0127_SET_INPUT: INPUT_TUNER\n" );
			ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); /* autodetect 50/60 Hz */
			ks0127_and_or( ks, KS_CMDA,   ~0x40, 0x00 ); /* VSE=0 */
			ks0127_and_or( ks, KS_CMDB,   0xb0, 0x01 ); /* set input line = AY1, VALIGN=0 */
			ks0127_and_or( ks, KS_CMDC,   0x70, 0x0a ); /* non-freerunning mode */
			ks0127_and_or( ks, KS_CMDD,   0x03, 0x00 ); /* analog input */
			ks0127_and_or( ks, KS_CTRACK, 0xcf, 0x00 ); /* enable chroma demodulation */
	//			ks0127_and_or( ks, KS_LUMA,   0x00, (reg_defaults[KS_LUMA]&0xf7)|0x04 ); /* chroma trap */
	//			ks0127_and_or( ks, KS_VERTIA, 0x08, 0xb3 ); /* manual luma comb [1 0 0], LPF/comb disabled */
	//			ks0127_and_or( ks, KS_VERTIC, 0x0f, 0xd0 ); /* manual chroma comb [0 1 0] */

			ks0127_and_or( ks, KS_LUMA,   0x00, (reg_defaults[KS_LUMA])|0x0c ); /* chroma trap, HYBWR=1 */
			ks0127_and_or( ks, KS_VERTIA, 0x08, 0x81 ); /* scaler fullbw, luma comb off */
			ks0127_and_or( ks, KS_VERTIC, 0x0f, 0x90 ); /* manual chroma comb .25 .5 .25 */
								
			ks0127_and_or( ks, KS_CHROMB, 0x0f, reg_defaults[KS_CHROMB]&0xf0 );
		
			ks0127_write( ks, KS_UGAIN, reg_defaults[KS_UGAIN] );
			ks0127_write( ks, KS_VGAIN, reg_defaults[KS_VGAIN] );
			ks0127_write( ks, KS_UVOFFH, reg_defaults[KS_UVOFFH] );
			ks0127_write( ks, KS_UVOFFL, reg_defaults[KS_UVOFFL] );
		break;

		case KS_INPUT_YUV656:
		dprintk( "ks0127: command KS0127_SET_INPUT: INPUT_YUV656\n" );
			if (ks->norm == KS_STD_NTSC_M || ks->norm == KS_STD_PAL_M) {
				ks0127_and_or( ks, KS_CMDA,   0xfc, 0x03 ); /* force 60 Hz */
			} else {
				ks0127_and_or( ks, KS_CMDA,   0xfc, 0x02 ); /* force 50 Hz */
			}

			ks0127_and_or( ks, KS_CMDA,   0xff, 0x40 ); /* VSE=1 */
			ks0127_and_or( ks, KS_CMDB,   0xb0, 0x48 ); /* set input line = AY0-AC0, VALIGN=1 */
			ks0127_and_or( ks, KS_CMDC,   0x70, 0x87 ); /* freerunning mode, TSTGEN = 1 TSTGFR=11 TSTGPH=0 TSTGPK=0  VMEM=1*/
			ks0127_and_or( ks, KS_CMDD,   0x03, 0x08); /* digital input, SYNDIR = 0 INPSL=01 CLKDIR=0 EAV=0 */
			ks0127_and_or( ks, KS_CTRACK, 0xcf, 0x30 ); /* disable chroma demodulation */
			ks0127_and_or( ks, KS_LUMA,   0x00, 0x71 ); /* HYPK =01 CTRAP = 0 HYBWR=0 PED=1 RGBH=1 UNIT=1 */
	//			ks0127_and_or( ks, KS_VERTIA, 0x08, 0xb0 ); /* manual luma comb [1 0 0] */
			ks0127_and_or( ks, KS_VERTIC, 0x0f, reg_defaults[KS_VERTIC]&0xf0 );

			ks0127_and_or( ks, KS_VERTIA, 0x08, 0x81 ); /* scaler fullbw, luma comb off */
								
			ks0127_and_or( ks, KS_CHROMB, 0x0f, reg_defaults[KS_CHROMB]&0xf0 );
		
			ks0127_and_or( ks, KS_CON, 0x00, 0x00);
			ks0127_and_or( ks, KS_BRT, 0x00, 32);	/* spec: 34 */
			ks0127_and_or( ks, KS_SAT, 0x00, 0xe8);	/* spec: 229 (e5) */
			ks0127_and_or( ks, KS_HUE, 0x00, 0);

			ks0127_and_or( ks, KS_UGAIN, 0x00, 238);
			ks0127_and_or( ks, KS_VGAIN, 0x00, 0x00);

			ks0127_and_or( ks, KS_UVOFFH, 0x00, 0x4f); /*UOFF:0x30, VOFF:0x30, TSTCGN=1 */
			ks0127_and_or( ks, KS_UVOFFL, 0x00, 0x00);

		break;
		}
	
		/* hack: CDMLPF sometimes spontaneously switches on; force back off */
		ks0127_write( ks, KS_DEMOD, reg_defaults[KS_DEMOD] );
	
		break;

        case KS0127_SET_OUTPUT:
		switch( *iarg ) {
		case KS_OUTPUT_YUV656E:
			dprintk( "ks0127: command KS0127_SET_OUTPUT: OUTPUT_YUV656E\n" );
#ifdef	NEVER
	//                    ks0127_and_or( ks, KS_OFMTA, 0xc0, 0x33 );
			ks0127_and_or( ks, KS_OFMTA, 0xf0, 0x03 );
#endif
			break;

		case KS_OUTPUT_EXV:
			dprintk( "ks0127: command KS0127_SET_OUTPUT: OUTPUT_EXV\n" );
	//                    ks0127_and_or( ks, KS_OFMTA, 0xc0, 0x39 );
			ks0127_and_or( ks, KS_OFMTA, 0xf0, 0x09 );
			break;
		}
		break;

	case KS0127_SET_STANDARD:
		/* Set to automatic SECAM/Fsc mode */
		ks0127_and_or( ks, KS_DEMOD, 0xf0, 0x00 );

		ks->norm = *iarg;
		switch( *iarg )
		{
		/* this is untested !! It just detects PAL_N/NTSC_M (no special frequencies)
		And you have to set the standard a second time afterwards */
		case KS_STD_AUTO:
			dprintk( "ks0127: command KS0127_SET_STANDARD: AUTO\n" ); 

			/* The chip determines the format based on the current field rate */
			ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); 
			ks0127_and_or( ks, KS_CHROMA, 0x9f, 0x20 ); //0x40 );
			/* This is wrong for PAL ! As I said, you need to set the standard once again !! */
			ks->format_height = 240;
			ks->format_width = 704;
			break;

		case KS_STD_NTSC_M:
			dprintk( "ks0127: command KS0127_SET_STANDARD: NTSC_M\n" ); 
		//                    ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); 
			ks0127_and_or( ks, KS_CHROMA, 0x9f, 0x20 ); //0x00 );
			ks->format_height = 240;
			ks->format_width = 704;
			break;

		case KS_STD_NTSC_N:
			dprintk( "ks0127: command KS0127_SET_STANDARD: NTSC_N\n" );
		//                    ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); 
			ks0127_and_or( ks, KS_CHROMA, 0x9f, 0x40 ); //0x00 );
			ks->format_height = 240;
			ks->format_width = 704;
			break;

		case KS_STD_PAL_N:
			dprintk( "ks0127: command KS0127_SET_STANDARD: PAL_N\n" );
		//                    ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); 
			ks0127_and_or( ks, KS_CHROMA, 0x9f, 0x20 ); //0x60 );
			ks->format_height = 290;
			ks->format_width = 704;
			break;

		case KS_STD_PAL_M:
			dprintk( "ks0127: command KS0127_SET_STANDARD: PAL_M\n" );
		//                    ks0127_and_or( ks, KS_CMDA,   0xfc, 0x00 ); 
			ks0127_and_or( ks, KS_CHROMA, 0x9f, 0x40 ); //0x60 );
			ks->format_height = 290;
			ks->format_width = 704;
			break;

		case KS_STD_SECAM:
			dprintk( "ks0127: command KS0127_SET_STANDARD: SECAM\n" );
			ks->format_height = 290;
			ks->format_width = 704;

			/* set to secam autodetection */
			ks0127_and_or( ks, KS_CHROMA, 0xdf, 0x20 );
			ks0127_and_or( ks, KS_DEMOD, 0xf0, 0x00 );
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ/10+1);

			/* did it autodetect? */
			if( ks0127_read( ks, KS_DEMOD ) & 0x40 )
				break;

			/* force to secam mode */
			ks0127_and_or( ks, KS_DEMOD, 0xf0, 0x0f );
			break;
		}
		break;

        case KS0127_SET_BRIGHTNESS:
		ks0127_write( ks, KS_BRT, (signed char)*iarg );
		break;

        case KS0127_SET_CONTRAST:
		ks0127_write( ks, KS_CON, (signed char)*iarg );
		break;

        case KS0127_SET_HUE:
		ks0127_write( ks, KS_HUE, (signed char)*iarg );
		break;

        case KS0127_SET_SATURATION:
		ks0127_write( ks, KS_SAT, (signed char)*iarg );
		if (*iarg == 0) 
			ks0127_and_or( ks, KS_UVOFFH, 0xb0, 0x00);
		else 
			ks0127_and_or( ks, KS_UVOFFL, 0x00, 0x40);
		break;

        case KS0127_SET_AGC_MODE:
		switch( *iarg ) {
		case KS_AGC_NORMAL:
			ks0127_and_or( ks, KS_CMDB, 0x4f, 0x00 );
			break;

		case KS_AGC_MODE1:
			ks0127_and_or( ks, KS_CMDB, 0x4f, 0x80 );
			break;

		case KS_AGC_MANUAL:
			ks0127_and_or( ks, KS_CMDB, 0xcf, 0x10 );
			break;

		case KS_AGC_AUTO:
			ks0127_and_or( ks, KS_CMDB, 0xdf, 0x20 );
			break;
		}
		break;

        case KS0127_SET_AGC:
		ks0127_write( ks, KS_AGC, *iarg );
		break;
                
        case KS0127_SET_CHROMA_MODE:
		switch( *iarg ) {
		case KS_CHROMA_AUTO:
			ks0127_and_or( ks, KS_CHROMA, 0xfc, 0x00 );
			break;

		case KS_CHROMA_ON:
			ks0127_and_or( ks, KS_CHROMA, 0xfc, 0x02 );
			break;

		case KS_CHROMA_OFF:
			ks0127_and_or( ks, KS_CHROMA, 0xfc, 0x03 );
			break;
		}
		break;

        case KS0127_SET_PIXCLK_MODE:
		switch( *iarg ) {
		case KS_PIXCLK_CCIR601:
			ks0127_and_or( ks, KS_CMDA, 0xfb, 0x04 );
			break;
		case KS_PIXCLK_SQUARE:
			ks0127_and_or( ks, KS_CMDA, 0xfb, 0x00 );
			break;
		}
		break;

        case KS0127_SET_GAMMA_MODE:
		switch( *iarg ) {
		case KS_GAMMA_OFF:
			ks0127_and_or( ks, KS_OFMTA, 0x3f, 0x00 );
			break;
		case KS_GAMMA_Y:
			ks0127_and_or( ks, KS_OFMTA, 0x3f, 0x40 );
			break;
		case KS_GAMMA_UV:
			ks0127_and_or( ks, KS_OFMTA, 0x3f, 0x80 );
			break;
		case KS_GAMMA_YUV:
			ks0127_and_or( ks, KS_OFMTA, 0x3f, 0xc0 );
			break;
		}
		break;

        case KS0127_SET_UGAIN:
                ks0127_write( ks, KS_UGAIN, *iarg );
                break;

        case KS0127_SET_VGAIN:
                ks0127_write( ks, KS_VGAIN, *iarg );
                break;

        case KS0127_SET_INVALY:
                ks0127_write( ks, KS_INVALY, *iarg );
                break;

        case KS0127_SET_INVALU:
                ks0127_write( ks, KS_INVALU, *iarg );
                break;

        case KS0127_SET_INVALV:
                ks0127_write( ks, KS_INVALV, *iarg );
                break;

        case KS0127_SET_UNUSEY:
                ks0127_write( ks, KS_UNUSEY, *iarg );
                break;

        case KS0127_SET_UNUSEU:
		ks0127_write( ks, KS_UNUSEU, *iarg );
		break;

        case KS0127_SET_UNUSEV:
		ks0127_write( ks, KS_UNUSEV, *iarg );
		break;

        case KS0127_SET_VSALIGN_MODE:
		switch( *iarg ) {
		case KS_VSALIGN0:
			ks0127_and_or( ks, KS_CMDB, 0xbf, 0x00 );
			break;
		case KS_VSALIGN1:
			ks0127_and_or( ks, KS_CMDB, 0xbf, 0x40 );
			break;
		}
		break;

        case KS0127_SET_OUTPUT_MODE:
		switch( *iarg ) {
		case KS_OUTPUT_NORMAL:
	//                    ks0127_and_or( ks, KS_CDEM,  0x7f, 0x00 );
			ks0127_and_or( ks, KS_OFMTA, 0xcf, 0x30 );//0x00 );
			break;

		case KS_OUTPUT_HIZ:
	//                    ks0127_and_or( ks, KS_CDEM,  0x7f, 0x80 );
			ks0127_and_or( ks, KS_OFMTA, 0xcf, 0x00 );
			break;

		case KS_OUTPUT_SYNC_HIZ:
	//                    ks0127_and_or( ks, KS_CDEM,  0x7f, 0x80 );
			ks0127_and_or( ks, KS_OFMTA, 0xcf, 0x10 );
			break;

		case KS_OUTPUT_SYNC_CLK_HIZ:
	//                    ks0127_and_or( ks, KS_CDEM,  0x7f, 0x80 );
			ks0127_and_or( ks, KS_OFMTA, 0xcf, 0x20 );
			break;
		}
		break;

	case KS0127_SET_WIDTH:
		ks0127_getcrop( ks, &xstart, &xend, &ystart, &yend);
		dprintk("ks0127: xstart=%d xend=%d ystart=%d, yend=%d\n",
			xstart, xend, ystart, yend);

		hav = 720 - (xstart + xend);
		vav = yend - ystart;
		dprintk("ks0127: hav = %d vav=%d\n", hav, vav);

		size = (*iarg * 0x7fff) / hav;
		ks->cap_width = *iarg;
		dprintk("ks0127: width size = %d program = 0x%x\n", *iarg, size);
		ks0127_and_or( ks, KS_HSCLL, 0x01, size << 1);
		ks0127_write( ks, KS_HSCLH, size >> 7);


   /* STILL needed ? (gz/000219) */
		/*
		 * This little hack can for samsung note was
		 * scaler fix -- turn off fifi for at least 1 line
		 */
		for(i=0; i < 10; i++) 
			ks0127_and_or( ks, 0x0d, ~0x03, 0);
		ks0127_and_or( ks, 0x0d, ~0x03, 1);

		

		if (ks->cap_width <= 384) {
			ks0127_and_or( ks, KS_VERTIA, ~0x08, 0x00);
		} else {

			ks0127_and_or( ks, KS_VERTIA, ~0x08, 0x08);
		}

		break;

	case KS0127_SET_HEIGHT:
	  
		ks0127_getcrop( ks, &xstart, &xend, &ystart, &yend);
		dprintk("ks0127: xstart=%d xend=%d ystart=%d, yend=%d\n",
			xstart, xend, ystart, yend);

		hav = 720 - (xstart + xend);
		vav = yend - ystart;
		dprintk("ks0127: hav = %d vav=%d\n", hav, vav);

		size = (*iarg * 0x3fff) / vav;
		ks->cap_height = *iarg;
		dprintk("ks0127: !!! height size = %d program = 0x%x\n", *iarg, size);

		h1 = ks0127_read( ks, KS_HSCLL);
		h2 = ks0127_read( ks, KS_HSCLH);

		/* Reset the vertical scaling - reason ? */
		ks0127_and_or( ks, KS_VSCLL, 0x03, 0xfc);
		ks0127_write( ks, KS_VSCLH, 0xff);

		/* Also reset the horizontal scaling, but keep the comb bit 0 - reason ? */
		ks0127_and_or( ks, KS_HSCLL, 0x01, 0xfe);
		ks0127_write( ks, KS_HSCLH, 0);

		/* set the vertical scaling */
		ks0127_and_or( ks, KS_VSCLL, 0x03, size << 2);
		ks0127_write( ks, KS_VSCLH, size >> 6);

		/* Disable the vertical scaling */
		ks0127_and_or( ks, KS_VERTIB, ~0x06, 0x12);

		/* Keep the comb adjustments, and move in the old horizontal scaling values */
		ks0127_and_or( ks, KS_HSCLL, 0x01, h1);
		ks0127_write( ks, KS_HSCLH, h2);

#if 0
		/* this block was deactivated before */
		if (ks->cap_height == ks->format_height) {
			ks0127_and_or( ks, KS_VERTIB, ~0x06, 0x12);
			dprintk("ks0127: Vertical scaling is disabled. Video is 1-line delayed.\n");
		} else {
			ks0127_and_or( ks, KS_VERTIB, ~0x06, 0x00);
			dprintk("ks0127: Vertical scaling is enabled.\n");
		}

		if (ks->cap_height <= ks->format_height/2) {
			ks0127_and_or( ks, KS_VERTIC, ~0x04, 0x04);
			dprintk("ks0127: Reduced luma vertical bandwidth.\n");
		} else {
			ks0127_and_or( ks, KS_VERTIC, ~0x04, 0x00);
			dprintk("ks0127: Full luma vertical bandwidth.\n");
		}
#endif

		break;

	case KS0127_SET_HSCALE:
		dprintk( "ks: set hscale %x\n", *iarg );
		ks0127_and_or( ks, KS_HSCLL, 0x01, (*iarg & 0x7F) << 1);
		ks0127_write( ks, KS_HSCLH, (*iarg & 0x7F80) >> 7);
		break;

	case KS0127_GET_STATUS:
		*iarg = ks0127_read( ks, KS_STAT );
		break;
		
	case KS0127_READ_PROC:
		prd = (struct mga_procread*)arg;
		return ks0127_proc_read(prd->buf, prd->start, prd->offset, prd->len, prd->eof, prd->data);

	case KS0127_WRITE_PROC:
		pwr = (struct mga_procwrite*)arg;
		return ks0127_proc_write(pwr->file, pwr->buf, pwr->count, pwr->data);

	default:
		return -EINVAL;
	}

	return 0;
}



static int
ks0127_proc_read(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
	int	i;
	u8	reg[64];
	struct ks0127	*ks0127;

	static  char *hfsel[4] = {"Very Fast", "Fast", "VCR", "TV" };
	static  char *insel[16] = { 
		"AY0 Composite", "AY1 Composite","AY2 Composite",
		"AC0 Composite", "AC1 Composite","AC2 Composite",
		"AY0 Svideo",    "AY1 Svideo",   "AY2 SVIDEO",
		"UNKNOWN-09",    "UNKNOWN-0a",   "UNKNOWN-0b",
		"UNKNOWN-0c",    "UNKNOWN-0d",   "UNKNOWN-0e",
		"Cb Cr"
	};
	static char *tstgfr[] = {
		"h Stop Freq Tracking", "h Freq tracks input",
		"h freq ifmt and pxel", "h freq ifmt and pxel"
	};

	len = 0;
	ks0127 = (struct ks0127 *)data;

	len += sprintf(buf, "ks0127 interface\n");

	for(i=0; i < 64; i++) {
		if ((i & 0x0f) == 0) {
			len += sprintf(buf + len, "\n%02x: ", i);
		}
		reg[i] = ks0127_read(ks0127, i);
		len += sprintf(buf + len, " %02x", reg[i]);
	}

	len += sprintf(buf + len, "\n\n");


	len += sprintf(buf + len, "clock:  %s\n", 
		(reg[0] & 1) ? "Locked" : "Not lock");
	len += sprintf(buf + len, "hclock: %s\n", 
		(reg[0] & 2) ? "Steady state track" : "Initial Tracking");
	len += sprintf(buf + len, "cdet: %s\n",
		(reg[0] & 4) ? "Color signal" : "No Color signal");
	len += sprintf(buf + len, "paldet: %s\n", 
		(reg[0] & 8) ? "PAL" : "NTSC");
	len += sprintf(buf + len, "ffredet: %s\n",
		(reg[0] & 16) ? "60 HZ" : "50 HZ");
	len += sprintf(buf + len, "novid: %s\n", 
		(reg[0] & 32) ? "No sync" : "sync");
	len += sprintf(buf + len, "vbiflg: %s\n", 
		(reg[0] & 64) ? "Blanking region" : "Active region");
	len += sprintf(buf + len, "chipid: %s\n\n",
		(reg[0] & 128) ? "KS0127" : "KS0122");

	len += sprintf(buf + len, "ifmt: %s\n", 
		(reg[1] & 0x02) ? ((reg[1] & 0x01) ? "60 HZ" : "50 HZ") : "Not used");
	len += sprintf(buf + len, "pixsel: %s\n",
		(reg[1] & 0x04) ? "CCIR 601" : "Square");
	len += sprintf(buf + len, "hfsel: %s\n", hfsel[(reg[1]>>4)&2]);
	len += sprintf(buf + len, "vse:   %s\n", 
		(reg[1] & 0x40) ? "9/9.5" : "10/10.5");

	len += sprintf(buf + len, "insel:    %s\n", insel[reg[2] & 0x0f]);
	len += sprintf(buf + len, "agcfrz:   %s\n", 
		(reg[2] & 0x10) ? "AGC running" : "AGC frozzen");
	len += sprintf(buf + len, "agcovf:   %s\n", 
		(reg[2] & 0x20) ? "AGC tracks" : "AGC overflow reduced");
	len += sprintf(buf + len, "valign:   %s\n", 
		(reg[2] & 0x40) ? "1" : "0");
	len += sprintf(buf + len, "AGCGN:   %s\n\n", 
		(reg[2] & 0x80) ? "sync tip" : "normal");

	len += sprintf(buf + len, "TSTGEN: %s\n", 
		(reg[3] & 0x01) ? "Manaul freq Tracking" : "Auto freq Tracking");
	len += sprintf(buf + len, "TSTGFR: %s\n",
		(reg[3] & 0x01) ? tstgfr[(reg[3] & 0x06) >> 1] : "Not used");
	len += sprintf(buf + len, "TSTGPH: %s\n",
		(reg[3] & 0x01) ? (reg[3] & 0x08) ? "Phase tracks input" : 
		       "No phase tracking" : "Not used");
	len += sprintf(buf + len, "TSTGPK: %s\n",
		(reg[3] & 0x40) ? (reg[3] & 0x20) ? "AGC Clamps Sync tip" :
		       "AGC Clamps back Porch" : "Not used");
	len += sprintf(buf + len, "VMEM: %s\n\n",
		 (reg[3] & 0x80) ? "V sync free running" : "V sync track input");

	len += sprintf(buf + len, "con(10):      %02x\n", reg[0x10]);
	len += sprintf(buf + len, "brt(11):      %02x\n", reg[0x11]);
	len += sprintf(buf + len, "sat(15):      %02x\n", reg[0x15]);
	len += sprintf(buf + len, "hue(16):      %02x\n", reg[0x16]);
	len += sprintf(buf + len, "ugain(2b):    %02x\n", reg[0x2b]);
	len += sprintf(buf + len, "vgain(2c):    %02x\n", reg[0x2c]);
	len += sprintf(buf + len, "invaild:      y=%02x u=%02x v=%02x\n", reg[0x32], reg[0x33], reg[0x34]);
	len += sprintf(buf + len, "inactive:     y=%02x u=%02x v=%02x\n", reg[0x35], reg[0x36], reg[0x37]);
	
	len += sprintf(buf + len, "vavb:   %d\n", (reg[0x2d] >> 2));
	len += sprintf(buf + len, "vave:   %d\n", reg[0x2e]);
	len += sprintf(buf + len, "vaved0: %d\n", (reg[0x2d] & 1));
	len += sprintf(buf + len, "vavod0: %d\n", (reg[0x2d] >> 1) & 1);
	
#ifdef NEVER
	if ( 1==1 ) {
		int xs, xe, ys, ye;
		unsigned hscl, vscl;
		
		ks0127_getcrop( ks0127, &xs, &xe, &ys, &ye );
		len += sprintf(buf + len, "crop: %i,%i  %i,%i\n", xs, ys, xe, ye );
		
		hscl = ((int)reg[ KS_HSCLH ] << 7) | (reg[ KS_HSCLL ] >> 1);
		if ( hscl == 0 )
			hscl = 0x8000;
		vscl = ((int)reg[ KS_VSCLH ] << 6) | (reg[ KS_VSCLL ] >> 2);
		if ( vscl == 0 )
			vscl = 0x4000;
			
		len += sprintf(buf + len, "hscl %i/32768  vscl %i/16384\n", hscl, vscl );
	}
#endif
	return len;
}

static int
ascii_to_hex(char *ptr)
{
	int	value;

	value = 0;
	for(;;) {
		if (*ptr >= '0' && *ptr <= '9') {
			value = value * 16 + *ptr - '0';
			ptr++;
		} else if (*ptr >= 'a' && *ptr <= 'f') {
			value = value * 16 + *ptr - 'a' + 10;
			ptr++;
		} else if (*ptr >= 'A' && *ptr <= 'F') {
			value = value * 16 + *ptr - 'A' + 10;
			ptr++;
		} else {
			return value;
		}
	}
}

static int
ks0127_proc_write(struct file *file, const char *buf, unsigned long count, 
void *data)
{
	char	*ptr;
	u8	reg;
	u8	val;
	struct ks0127* ks0127;

	ptr = (char *)buf;
	ks0127 = (struct ks0127 *)data;

	reg = ascii_to_hex(&ptr[0]);
	val = ascii_to_hex(&ptr[3]);
	if (ptr[2] != '=' || ptr[5] != '\n') {
		return count;
	}

	ks0127_write(ks0127, reg, val);
	printk("ks0127 writing reg %02x value %02x\n", reg, val);

	return count;
}



static int ks0127_attach(struct i2c_adapter *adapter);
static int ks0127_detach(struct i2c_client *client);
static int ks0127_command(struct i2c_client *client, unsigned int cmd, void *arg );


/* Addresses to scan */
static unsigned short normal_i2c[] =		{ I2C_CLIENT_END };
static unsigned short normal_i2c_range[] =	{ I2C_KS0127_ADDON>>1, I2C_KS0127_ONBOARD>>1, I2C_CLIENT_END };
static unsigned short probe[2]        = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe_range[2]  = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[2]       = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore_range[2] = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short force[2]        = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static struct i2c_client_address_data addr_data = {
        normal_i2c, normal_i2c_range, 
        probe, probe_range, 
        ignore, ignore_range, 
        force
};


static struct i2c_driver i2c_driver_ks0127 = {
	owner:		THIS_MODULE,
	name:		KS0127_DEVNAME,
	id:		I2C_DRIVERID_KS0127,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	ks0127_attach,
	detach_client:	ks0127_detach,
	command:	ks0127_command,
};


static struct i2c_client ks0127_client_tmpl =
{
	name:		"(nil)",
	id:		-1,
	flags:		I2C_CLIENT_ALLOW_USE,
	addr:		0,
	adapter:	NULL,
	driver:		&i2c_driver_ks0127,
	data:		NULL,
	usage_count:	0
};


static int ks0127_found_proc(struct i2c_adapter *adapter,
                                     int addr, unsigned short flags,int kind)
{
	struct ks0127 *ks;
	struct i2c_client *client;

	client = kmalloc(sizeof(*client), GFP_KERNEL);
	if(client == NULL)
		return -ENOMEM;
	memcpy( client, &ks0127_client_tmpl, sizeof(*client) );


	ks = kmalloc( sizeof(*ks), GFP_KERNEL );
	if( ks == NULL ) {
		kfree( client );
		return -ENOMEM;
	}
	memset( ks, 0, sizeof( *ks ) );

	client->data = ks;
        client->adapter = adapter;
        client->addr = addr;
	sprintf(client->name, "%s-%02x", KS0127_DEVNAME, adapter->id);

	ks->client = client;
	ks->addr = addr;
	ks->ks_type = KS_TYPE_UNKNOWN;
	
	/* power up */
	ks0127_write( ks, KS_CMDA, 0x2c );
	mdelay(10);
 	
	/* reset the device */
	ks0127_reset( ks );
	printk(KERN_INFO "ks0127: attach: %s video decoder\n",
	       ks->addr==I2C_KS0127_ADDON ? "addon" : "on-board");

	/* tristate */
//	mdelay(150);
//	ks0127_and_or( ks, KS_CDEM,  0x7f, 0x80 );
//	ks0127_and_or( ks, KS_OFMTA, 0xcf, 0x20 );

	i2c_attach_client(client);
	MOD_INC_USE_COUNT;
	
	return 0;
}


static int ks0127_attach(struct i2c_adapter *adapter)
{
	if (	((adapter->id & I2C_ALGO_MASK) == I2C_ALGO_BIT) ||
		((adapter->id & I2C_ALGO_MASK) == I2C_ALGO_EXP) )
                return i2c_probe(adapter, &addr_data, ks0127_found_proc);
	return 0;
}


static int ks0127_detach(struct i2c_client *client)
{
	struct ks0127 *ks = (struct ks0127*)client->data;
	
	ks0127_write( ks, KS_OFMTA, 0x20 ); /*tristate*/
	ks0127_write( ks, KS_CMDA, 0x2c | 0x80 ); /* power down */

        i2c_detach_client(client);
        kfree(ks);
        kfree(client);
	
	dprintk( "ks0127: detach\n" );
	
        MOD_DEC_USE_COUNT;
        return 0;
}



/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module(void)
#else
int ks0127_init(void)
#endif
{
	i2c_add_driver(&i2c_driver_ks0127);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_del_driver(&i2c_driver_ks0127);
}
#endif

