/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the MAVEN tv-out chip.
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

/* The maven setup comes from the matroxfb driver by Petr Vandrovec. */
/* It just needed some parameter changes. */

/*
 2000 July, 9th :
	Yeah ! Video passes through the marvel and back to the TV !
	(works for PAL for the moment).
	Why does maven_read not work ?
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include "i2cdef.h"
#include "maven.h"

#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


#define MAVEN_DEVNAME  "maven"


/* i2c identification */
#define I2C_MAVEN 0x36


static int
maven_proc_read(char *buf, char **start, off_t offset, int len, int *eof, void *unused);

static int
maven_proc_write(struct file *file, const char *buf, unsigned long count,void *data);


/* maven control registers */
/* TODO */


/****************************************************************************
* mga_dev : represents one maven chip.
****************************************************************************/
struct maven_s {
	struct i2c_client *client;
	unsigned char	addr;
	int		mode;
	int             cable;
};

struct mavenregs {
	u_int8_t regs[256];
	int	 vlines;
	int	 xtal;
	int	 fv;

	u_int16_t htotal;
	u_int16_t hcorr;
};

struct timing {
	int	hdisplay;
	int	lmargin;
	int	htotal;
	int	vdisplay;
	int	vtotal;
};

static	struct	timing ntsc_timing = {
	700, 48, 858,
	480, 525
};

static struct timing pal_timing = {
	700, 48, 864,
	575, 625
};


/****************************************************************************
* raw register access : these routines directly interact with the maven's
*                       registers via the i2c
****************************************************************************/

static u8 maven_read(struct maven_s *mv, u8 reg)
{
	char val = 0;

	struct i2c_msg msg[] = {
		{ mv->addr, I2C_M_REV_DIR_ADDR,		1, &reg },	// write reg but via readaddr
		{ mv->addr, I2C_M_RD | I2C_M_NOSTART,	1, &val }	// read
	};

	if ( i2c_transfer( mv->client->adapter, msg, 2 ) != 2 ) {
		dprintk( "maven_read error\n" );
	}
	
	return val;
}


static void maven_write(struct maven_s *mv, u8 reg, u8 val)
{
	char msg[] = { reg, val };
	
	if ( i2c_master_send(mv->client, msg, sizeof(msg) ) != sizeof(msg) ) {
		dprintk( "maven_write error\n" );
	}
}


static void maven_write_pair(struct maven_s *mv, u8 reg, int val)
{
	char msg1[] = { reg, val & 0xFF };
	char msg2[] = { reg+1, val >> 8 };
	
	struct i2c_msg msg[] = {
		{ mv->addr, 0,	sizeof(msg1), msg1 },
		{ mv->addr, 0,	sizeof(msg2), msg2 }
	};

	if ( i2c_transfer( mv->client->adapter, msg, 2 ) != 2 ) {
		dprintk( "maven_write_pair error\n" );
	}
}


/****************************************************************************
* maven private api
****************************************************************************/
static void maven_reset( struct maven_s* mv )
{
	/* Mmmh ? Make something to output no signal. */
	maven_write (mv, 0xBF, 0x08);
}

static void maven_init_TVdata(struct mavenregs* data, int mode) {

	static struct mavenregs palregs = { {
		0x2A, 0x09, 0x8A, 0xCB,	/* 00: chroma subcarrier */
		0x00,
		0x00,	/* ? not written */
		0x09,	/* 06 */
		0x00,	/* ? not written */
		0x7E,	/* 08 */
		0x40,	/* 09 */
		0x9C,	/* 0A */
		0x30,	/* 0B */
		0x21,	/* 0C */
		0x00,	/* ? not written */
		0x3F, 0x03, /* 0E-0F */
		0x3F, 0x03, /* 10-11 */
		0x1A,	/* 12 */
		0x2A,	/* 13 */
		0x1C, 0x3D, 0x14, /* 14-16 */
		0x9C, 0x01, /* 17-18 */
		0x00,	/* 19 */
		0xFE,	/* 1A */
		0x7E,	/* 1B */
		0x60,	/* 1C */
		0x05,	/* 1D */
		0xC7, 0x03, /* 1E-1F */
		0x8A,	/* 20 (sat)*/
		0x07,	/* 21 */
		0x8A,	/* 22 */
		0x00,	/* 23 */
		0x00,	/* 24 */
		0x00,	/* 25 (hue) */
		0x08,	/* 26 */
		0x04,	/* 27 */
		0x00,	/* 28 */
		0x1A,	/* 29 */
		0x55, 0x01, /* 2A-2B */
		0x26,	/* 2C */
		0x07, 0x7E, /* 2D-2E */
		0x02, 0x54, /* 2F-30 */
		0xB0, 0x00, /* 31-32 */
		0x14,	/* 33 */
		0x49,	/* 34 */
		0x10,	/* 35 written multiple times */
		0x00,	/* 36 not written */
		0xA3,	/* 37 */
		0xC8,	/* 38 */
		0x22,	/* 39 */
		0x02,	/* 3A */
		0x22,	/* 3B */
		0x3F, 0x03, /* 3C-3D */
		0x00,	/* 3E written multiple times */
		0x20,	/* 3F not written */
	}, 0, 625, 50 };

	static struct mavenregs ntscregs = { {
		0x21, 0xF0, 0x7C, 0x1F,	/* 00: chroma subcarrier */
		0x00,
		0x00,	/* ? not written */
		0x00,	/* modified by code (F9 written...) */
		0x00,	/* ? not written */
		0x7E,	/* 08 */
		0x43,	/* 09 */
		0x7E,	/* 0A */
		0x3D,	/* 0B */
		0x00,	/* 0C */
		0x00,	/* ? not written */
		0x41, 0x00, /* 0E-0F */
		0x3C, 0x00, /* 10-11 */
		0x17,	/* 12 */
		0x21,	/* 13 */
		0x1B, 0x1B, 0x24, /* 14-16 */
		0x83, 0x01, /* 17-18 */
		0x00,	/* 19 */
		0x0F,	/* 1A */
		0x0F,	/* 1B */
		0x60,	/* 1C */
		0x05,	/* 1D */
		0x89, 0x02, /* 1E-1F */
		0x5F,	/* 20 */
		0x04,	/* 21 */
		0x5F,	/* 22 */
		0x01,	/* 23 */
		0x02,	/* 24 */
		0x00,	/* 25 */
		0x0A,	/* 26 */
		0x05,	/* 27 */
		0x00,	/* 28 */
		0x10,	/* 29 */
		0xFF, 0x03, /* 2A-2B */
		0x24,	/* 2C */
		0x0F, 0x78, /* 2D-2E */
		0x00, 0x00, /* 2F-30 */
		0xB2, 0x04, /* 31-32 */
		0x14,	/* 33 */
		0x02,	/* 34 */
		0x00,	/* 35 written multiple times */
		0x00,	/* 36 not written */
		0xA3,	/* 37 */
		0xC8,	/* 38 */
		0x15,	/* 39 */
		0x05,	/* 3A */
		0x3B,	/* 3B */
		0x3C, 0x00, /* 3C-3D */
		0x00,	/* 3E written multiple times */
		0x00,	/* never written */
	}, 1, 525, 60 };

	if (mode == MAVEN_OUTPUT_PAL) {
		dprintk("setting maven to pal\n");
		*data = palregs;
	} else {
		*data = ntscregs;
		dprintk("setting maven to ntsc\n");
	}

	/* gamma correction registers */
	data->regs[0x83] = 0x78;
	data->regs[0x84] = 0x99;
	data->regs[0x85] = 0xAF;
	data->regs[0x86] = 0xF8;
	data->regs[0x87] = 0xD4;
	data->regs[0x88] = 0xE5;
	data->regs[0x89] = 0xA5;
	data->regs[0x8A] = 0x5A;
	data->regs[0x8B] = 0xB4;

	return;
}

static inline int maven_compute_timming(struct mavenregs* m, int mode, int cable)
{
	struct	timing	*active_timing;

	maven_init_TVdata(m, mode);

	if (mode == MAVEN_OUTPUT_PAL) {
		active_timing = &pal_timing;
	} else {
		active_timing = &ntsc_timing;
	}

	/* horizontal decimator 0x80 = 1:1 0x40 2:1 */
 	m->regs[0x90] = 122;	/* < 0x40 || > 0x80 is bad... 0x80 is questionable */

	/* Vertical Decimator ((OutLines * Htotal * 32768)/Vtotal) */
	m->regs[0x91] = 69;
	m->regs[0x92] = 127;


	/* Antialiasing mode */
	m->regs[0x93] = 0xA2;

	/* TV mode B3, Monitor mode B2 */
	m->regs[0x94] = 0xb3;
	
	
	/* Last Line length (Monitor  Htotal, TV Htotal - 2) */
	m->regs[0x96] = active_timing->htotal & 0xff;
	m->regs[0x97] = active_timing->htotal >> 8;

	/* unknown */
	m->regs[0x98] = 0x00; 
	m->regs[0x99] = 0x00;

	/* left margin width (Monitor horizonal sync length, 
	 * TV htotal - hsync
	 */
	m->regs[0x9A] = active_timing->lmargin & 0xff;
	m->regs[0x9B] = active_timing->lmargin >> 8;

	/* who knows */
	m->regs[0x9C] = 3;
	m->regs[0x9D] = 0;

	/* hend */
	m->regs[0x9e] = (active_timing->hdisplay - 1) & 0xff;
	m->regs[0x9f] = (active_timing->hdisplay - 1) >> 8;

	/* htotal */
	m->regs[0xa0] = active_timing->htotal & 0xff;
	m->regs[0xa1] = active_timing->htotal >> 8;

	/* vblanking end */
	m->regs[0xA2] = 32 & 0xff;	/* stop vblanking */
	m->regs[0xA3] = 32 >> 8;

	/* something end... [A6]+1..[A8] */
	m->regs[0xA4] = 31 & 0xff;
	m->regs[0xA5] = 31 >> 8;

	/* something start... 0..[A4]-1 */
	m->regs[0xA6] = 11 & 0xff;
	m->regs[0xA7] = 11 >> 8;

	/* vertical line count - 1 */
	m->regs[0xA8] = active_timing->vtotal & 0xff;
	m->regs[0xA9] = active_timing->vtotal >> 8;

	/* horizontal vidrst pos */
	m->regs[0xAA] = 700 & 0xff;		/* 0 <= hcrt <= htotal - 2 */
	m->regs[0xAB] = 700 >> 8;
	/* vertical vidrst pos */
	m->regs[0xAC] = 0 & 0xff;
	m->regs[0xAD] = 0 >> 8;

	/* moves picture up/down and so on... */
	m->regs[0xAE] = 0x00; /* Fix this... 0..VTotal */
	m->regs[0xAF] = 0x00;


	m->regs[0xBE] = 46;
	if (cable == MAVEN_CABLE_COMPOSITE)
		m->regs[0xB0] = 0x81;
	else                                 /* SVIDEO */
		m->regs[0xB0] = 0x80;
	m->regs[0xC2] = 0;


	m->regs[0x80] = 0x03;
	m->regs[0x81] = 0x03;
	m->regs[0x82] = 0x20;
	
	return 0;
}

#define LR(x) maven_write(c, (x), m->regs[(x)])
#define LRP(x) maven_write_pair(c, (x), m->regs[(x)] | (m->regs[(x)+1] << 8))
static void maven_init_TV(struct maven_s* c, const struct mavenregs* m) {

	maven_write(c, 0x3E, 0x01);
	maven_read(c, 0x82);	/* fetch oscillator state? */
	maven_write(c, 0x8C, 0x00);
	maven_read(c, 0x94);	/* get 0x82 */
	maven_write(c, 0x94, 0x82); /* 0xA2 */

	maven_write_pair(c, 0x8E, 0x1EFF);
	maven_write(c, 0xC6, 0x01);

	/* removed code... */

	maven_read(c, 0x06);
	maven_write(c, 0x06, 0x09);/*f9*/	/* or read |= 0xF0 ? */

	/* real code begins here? */
	/* chroma subcarrier */
	LR(0x00); LR(0x01); LR(0x02); LR(0x03);

	LR(0x04);

	LR(0x2C);
	LR(0x08);
	LR(0x0A);
	LR(0x09);
	LR(0x29);
	LRP(0x31);
	LRP(0x17);
	LR(0x0B);
	LR(0x0C);
	maven_write(c, 0x35, 0x10); /* ... */

	LRP(0x10);

	LRP(0x0E);
	LRP(0x1E);

	LR(0x20);	/* saturation #1 */
	LR(0x22);	/* saturation #2 */
	LR(0x25);	/* hue */
	LR(0x34);
	LR(0x33);
	LR(0x19);
	LR(0x12);
	LR(0x3B);
	LR(0x13);
	LR(0x39);
	LR(0x1D);
	LR(0x3A);
	LR(0x24);
	LR(0x14);
	LR(0x15);
	LR(0x16);
	LRP(0x2D);
	LRP(0x2F);
	LR(0x1A);
	LR(0x1B);
	LR(0x1C);
	LR(0x23);
	LR(0x26);
	LR(0x28);
	LR(0x27);
	LR(0x21);
	LRP(0x2A);
	maven_write(c, 0x35, 0x1D);	/* ... */

	LRP(0x3C);
	LR(0x37);
	LR(0x38);
	maven_write(c, 0xB3, 0x01);

	maven_read(c, 0xB0);	/* read 0x80 */
	maven_write(c, 0xB0, 0x80);	/*0x08*/ /* ugh... */
	maven_read(c, 0xB9);	/* read 0x7C */
	maven_write(c, 0xB9, 0x78);
	maven_read(c, 0xBF);	/* read 0x00 */
	maven_write(c, 0xBF, 0x02);
	maven_read(c, 0x94);	/* read 0x82 */
	maven_write(c, 0x94, 0x82); /*0xB3);*/ /* to test : 0x82 */

	LR(0x80); /* 04 1A 91 or 05 21 91 */
	LR(0x81);
	LR(0x82);

	maven_write(c, 0x8C, 0x00); /* 0x20 */
	maven_read(c, 0x8D);
	maven_write(c, 0x8D, 0x07); /* 0x10 */

	LR(0x90); /* 4D 50 52 or 4E 05 45 */
	LR(0x91);
	LR(0x92);

	LRP(0x9A); /* 0049 or 004F */
	LRP(0x9C); /* 0004 or 0004 */
	LRP(0x9E); /* 0458 or 045E */
	LRP(0xA0); /* 05DA or 051B */
	LRP(0xA2); /* 00CC or 00CF */
	LRP(0xA4); /* 007D or 007F */
	LRP(0xA6); /* 007C or 007E */
	LRP(0xA8); /* 03CB or 03CE */
	LRP(0x98); /* 0000 or 0000 */
	LRP(0xAE); /* 0044 or 003A */
	LRP(0x96); /* 05DA or 051B */
	LRP(0xAA); /* 04BC or 046A */
	LRP(0xAC); /* 004D or 004E */

	LR(0xBE);
	LR(0xC2);

	maven_read(c, 0x8D);
	maven_write(c, 0x8D, 0x07);

	LR(0x20);	/* saturation #1 */
	LR(0x22);	/* saturation #2 */
	LR(0x93);	/* whoops */
	LR(0x20);	/* oh, saturation #1 again */
	LR(0x22);	/* oh, saturation #2 again */
	LR(0x25);	/* hue */
	LRP(0x0E);
	LRP(0x1E);
	LRP(0x0E);	/* problems with memory? */
	LRP(0x1E);	/* yes, matrox must have problems in memory area... */

	/* load gamma correction stuff */
	LR(0x83);
	LR(0x84);
	LR(0x85);
	LR(0x86);
	LR(0x87);
	LR(0x88);
	LR(0x89);
	LR(0x8A);
	LR(0x8B);

	LR(0x33);
	LR(0x19);
	LR(0x12);
	LR(0x3B);
	LR(0x13);
	LR(0x39);
	LR(0x1D);
	LR(0x3A);
	LR(0x24);
	LR(0x14);
	LR(0x15);
	LR(0x16);
	LRP(0x2D);
	LRP(0x2F);
	LR(0x1A);
	LR(0x1B);
	LR(0x1C);
	LR(0x23);
	LR(0x26);
	LR(0x28);
	LR(0x27);
	LR(0x21);
	LRP(0x2A);
	maven_write(c, 0x35, 0x1D);	/* ... */
	LRP(0x3C);
	LR(0x37);
	LR(0x38);

	maven_read(c, 0xB0);
	LR(0xB0);	/* output mode */
	LR(0x90);
	LR(0xBE);
	LR(0xC2);

	LRP(0x9A);
	LRP(0xA2);
	LRP(0x9E);
	LRP(0xA6);
	LRP(0xAA);
	LRP(0xAC);
	maven_write(c, 0x3E, 0x00);
	maven_write(c, 0x95, 0x20);
}

/****************************************************************************
* maven i2c driver interface
****************************************************************************/


static int maven_ioctl(struct maven_s* mv, unsigned int cmd, void* arg )
{
	struct mavenregs m;
	int *iarg = (int*)arg;
	struct mga_procread *prd;
	struct mga_procwrite *pwr;

	if(!mv)
		return -ENODEV;

	switch(cmd) {
	case MAVEN_SET_STD:
		/* reset the device */
		maven_reset(mv);
      
		mv->mode = *iarg;
		maven_compute_timming(&m, mv->mode, mv->cable);
		maven_init_TV(mv, &m);
		break;
		
	case MAVEN_READ_PROC:
		prd = (struct mga_procread*)arg;
		return maven_proc_read(prd->buf, prd->start, prd->offset, prd->len, prd->eof, prd->data);
		break;

	case MAVEN_WRITE_PROC:
		pwr = (struct mga_procwrite*)arg;
		return maven_proc_write(pwr->file, pwr->buf, pwr->count, pwr->data);
		break;

	case MAVEN_SET_CABLE:
		if (mv->cable != *iarg) {
			mv->cable = *iarg;
			maven_compute_timming(&m, mv->mode, mv->cable);
			maven_init_TV(mv, &m);
		}
		break;                                                                              
	}
	return 0;
}



static int maven_attach(struct i2c_adapter *adapter);
static int maven_detach(struct i2c_client *client);
static int maven_command(struct i2c_client *client, unsigned int cmd, void *arg );


/* Addresses to scan */
static unsigned short normal_i2c[] =		{ I2C_MAVEN>>1 };
static unsigned short normal_i2c_range[] =	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe[2]        = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe_range[2]  = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[2]       = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore_range[2] = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short force[2]        = 	{ I2C_CLIENT_END, I2C_CLIENT_END };
static struct i2c_client_address_data addr_data = {
        normal_i2c, normal_i2c_range, 
        probe, probe_range, 
        ignore, ignore_range, 
        force
};


static struct i2c_driver i2c_driver_maven = {
	owner:		THIS_MODULE,
	name:		MAVEN_DEVNAME,
	id:		I2C_DRIVERID_MGATVO,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	maven_attach,
	detach_client:	maven_detach,
	command:	maven_command,
};


static struct i2c_client maven_client_tmpl =
{
	name:		"(nil)",
	id:		-1,
	flags:		I2C_CLIENT_ALLOW_USE,
	addr:		0,
	adapter:	NULL,
	driver:	&i2c_driver_maven,
	data:		NULL,
	usage_count:	0
};


static int maven_found_proc(struct i2c_adapter *adapter,
                                     int addr, unsigned short flags,int kind)
{
	struct maven_s* mv;
	struct i2c_client *client;
	
	client = kmalloc(sizeof(*client), GFP_KERNEL);
	if(client == NULL)
		return -ENOMEM;
	memcpy( client, &maven_client_tmpl, sizeof(*client) );
	

	mv = kmalloc(sizeof(*mv), GFP_KERNEL);
	if(mv == NULL) {
		kfree( client );
		return -ENOMEM;
	}
	memset(mv, 0, sizeof(*mv));
	
	client->data = mv;
	client->adapter = adapter;
	client->addr = addr;
	strcpy( client->name, MAVEN_DEVNAME );
	
	mv->client = client;
	mv->addr = addr;
	mv->cable = MAVEN_CABLE_COMPOSITE;

	/* reset the device */
	maven_reset(mv);

	printk(KERN_INFO "maven: attach: tv-out\n" );

	i2c_attach_client(client);
	MOD_INC_USE_COUNT;
	
	return 0;
}


static int maven_attach(struct i2c_adapter *adapter)
{
	if (	((adapter->id & I2C_ALGO_MASK) == I2C_ALGO_BIT) ||
		((adapter->id & I2C_ALGO_MASK) == I2C_ALGO_EXP) )
                return i2c_probe(adapter, &addr_data, maven_found_proc);
	return 0;
}


static int maven_detach(struct i2c_client *client)
{
	struct maven_s *mv = (struct maven_s*)client->data;
	
	maven_reset(mv);

	i2c_detach_client(client);
	kfree(mv);
	kfree(client);
	
	dprintk( "maven: detach\n" );
	
	MOD_DEC_USE_COUNT;
	return 0;
}


static int maven_command(struct i2c_client *client, unsigned int cmd, void *arg )
{
	struct maven_s *mv = (struct maven_s*)client->data;
	
	return maven_ioctl( mv, cmd, arg );
}
       

/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module(void)
#else
int maven_init(void)
#endif
{
	i2c_add_driver(&i2c_driver_maven);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_del_driver(&i2c_driver_maven);
}
#endif



static int
ascii_to_hex(char *ptr)
{
	int	value;

	value = 0;
	for(;;) {
		if (*ptr >= '0' && *ptr <= '9') {
			value = value * 16 + *ptr - '0';
			ptr++;
		} else if (*ptr >= 'a' && *ptr <= 'f') {
			value = value * 16 + *ptr - 'a' + 10;
			ptr++;
		} else if (*ptr >= 'A' && *ptr <= 'F') {
			value = value * 16 + *ptr - 'A' + 10;
			ptr++;
		} else {
			return value;
		}
	}
}

int
maven_proc_read(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
	struct maven_s * maven = data;
	
	/* Dump everything */
	int i;
	unsigned char v;
	len = 0;
	for (i = 0; i < 256; i++) {
		if ((i & 15) == 0)
			len += sprintf (buf+len, "0x%02x  ",i);
		v = maven_read(maven, i);
		len += sprintf(buf+len, "%02x ", v);
		if ((i & 15) == 15)
			len += sprintf (buf+len, "\n");
	}
	return len;
}

int
maven_proc_write(struct file *file, const char *buf, unsigned long count, void *data)
{
	struct maven_s * maven = data;

	char	*ptr;
	u8	reg;
	u8	val;

	ptr = (char *)buf;

	if (strncmp(ptr, "output composite", 16) == 0) {
		int cable = MAVEN_CABLE_COMPOSITE;
		maven_ioctl(maven, MAVEN_SET_CABLE, &cable);
		printk("maven : set cable composite\n");
	} else if (strncmp(ptr, "output svideo", 13) == 0) {
		int cable = MAVEN_CABLE_SVIDEO;
		maven_ioctl(maven, MAVEN_SET_CABLE, &cable);
		printk("maven : set cable svideo\n");
	} else {
		reg = ascii_to_hex(&ptr[0]);
		val = ascii_to_hex(&ptr[3]);
		if (ptr[2] != '=' || ptr[5] != '\n') {
			return count;
		}
		maven_write(maven, reg, val);
		printk("maven writing reg %02x value %02x\n", reg, val);
	}
	return count;
}

