/*
 * back-end scaler
 *
 * Copyright (C) 1999,2000 marvel project @ sourceforge
 * (based on code from Benoit Mathieu and Aaron Holtzman)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Merrick Johnson <inkling@nop.org>
 *			bes vertical adjust for interlace
 * 			colorkey modifications
 *	Leon van Stuivenberg <leonvs@iae.nl> (SMP safeness)
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/videodev.h>	/* for VIDEO_PALETTE_... */
#include <linux/proc_fs.h>
#include "mgacore.h"
#include "mgabes.h"

EXPORT_SYMBOL(mgabes_set_source);
EXPORT_SYMBOL(mgabes_set_window);
EXPORT_SYMBOL(mgabes_set_buffers);
EXPORT_SYMBOL(mgabes_set_frame);
EXPORT_SYMBOL(mgabes_quick_set_frame);
EXPORT_SYMBOL(mgabes_set_mode);
EXPORT_SYMBOL(mgabes_set_keymode);
EXPORT_SYMBOL(mgabes_set_colorkey);
EXPORT_SYMBOL(mgabes_activate);
EXPORT_SYMBOL(mgabes_desactivate);
EXPORT_SYMBOL(mgabes_pitch_rounding);

int mgabes_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data);

int mgabes_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data);

MODULE_PARM(bes_brightness, "i");
MODULE_PARM(bes_contrast, "i");
MODULE_PARM(bes_debug, "i");

int	bes_brightness = 0x00;
int	bes_contrast = 0x80;
int	bes_debug = 0;


#define dprintk if (bes_debug) printk


/* Backend scaler registers */
#define BESA1ORG        0x3d00
#define BESA2ORG        0x3d04
#define BESB1ORG        0x3d08
#define BESB2ORG        0x3d0c
#define BESA1CORG       0x3d10
#define BESA2CORG       0x3d14
#define BESB1CORG       0x3d18
#define BESB2CORG       0x3d1c
#define BESCTL          0x3d20
#define BESPITCH        0x3d24
#define BESHCOORD       0x3d28
#define BESVCOORD       0x3d2c
#define BESHISCAL       0x3d30
#define BESVISCAL       0x3d34
#define BESHSRCST       0x3d38
#define BESHSRCEND      0x3d3c
#define BESLUMACTL	0x3d40	/* G400 */
#define BESV1WGHT       0x3d48
#define BESV2WGHT       0x3d4c
#define BESHSRCLST      0x3d50
#define BESV1SRCLST     0x3d54
#define BESV2SRCLST     0x3d58
#define BESA1C3ORG	0x3d60	/* G400 */
#define BESA2C3ORG	0x3d64	/* G400 */
#define BESB1C3ORG	0x3d68	/* G400 */
#define BESB2C3ORG	0x3d6c	/* G400 */
#define BESGLOBCTL      0x3dc0
#define BESSTATUS       0x3dc4

#define VCOUNT		0x1e20

#define XMULCTRL      0x19
#define BPP_8         0x00
#define BPP_15        0x01
#define BPP_16        0x02
#define BPP_24        0x03
#define BPP_32_DIR    0x04
#define BPP_32_PAL    0x07

/*
 * These registers closely match the names of the hardware spec :
 * (check p. 3-59 (99) and following in the G400 manual)
 */
struct bes_registers_s
{
	u32 besctl;
	u32 besglobctl;
	u32 beslumactl;
	u32 bespitch;
	u32 besa1c3org;
	u32 besa1corg;
	u32 besa1org;
	u32 besa2c3org;
	u32 besa2corg;
	u32 besa2org;
	u32 besb1c3org;
	u32 besb1corg;
	u32 besb1org;
	u32 besb2c3org;
	u32 besb2corg;
	u32 besb2org;
	u32 beshcoord;
	u32 beshiscal;
	u32 beshsrcst;
	u32 beshsrcend;
	u32 beshsrclst;
	u32 besvcoord;
	u32 besviscal;
	u32 besv1srclst;
	u32 besv1wght;
	u32 besv2srclst;
	u32 besv2wght;

	u32 topfield_offset;
	u32 bottomfield_offset;
};

struct mga_bes_s
{
	/*
	 * mode (interlaced, double buffered) and options like
	 * accelerated 2x zoom, hfilter, ...
	 * BES_DEINTERLACE mode specs : when on, image is DEINTERLACED,
	 * and fields are shown separately.
	 * ON should be used for TV programs.
	 * OFF should be used for DVD films and for yuv software rendering.
	 */
	u32 mode;

	/* source video (for source image cropping) */
	u32 srcx;
	u32 srcy;
	u32 srcheight, srcwidth;


	/* overlay video window */
	int x, y; /* Might be negative */
	u32 dstwidth, dstheight;
	int palette;
	
	/* Destination window cropping */
	u32 offset_left, offset_right;
	u32 offset_top, offset_bottom;

	/* buffers to display */
	u32 bufwidth;
	u32 bufheight;
	u32 buf_a, buf_b;

	u32 desktop_offsetx;	/* X, Y offset of the displayed area */
	u32 desktop_offsety;	/* of the virtual desktop */

	/* frame to display */
	int frame;

	u32 chromakey;
	int overlay_mode;

	/* registers */
	struct bes_registers_s regs;

	/* host card */
	struct mga_dev_s *mga;

	spinlock_t lock;
};

/**
 *	mgabes_init - 	initialize the BES
 *	@mga:	mga instance
 *
 *	This function initializes the BackEnd Scaler, responsible for
 *	displaying a scaled video overlay.
 */
void mgabes_attach(struct mga_dev_s *mga)
{
	struct mga_bes_s *bes;

	if (!mga) return;

	mga->bes = bes = kmalloc(sizeof(*bes), GFP_KERNEL);

	memset(bes, 0, sizeof(*bes));
	spin_lock_init(&bes->lock);
	bes->mga = mga;


	bes->regs.besctl = 0;
	mgabes_desactivate(bes);

	/* default settings */
	bes->srcx = 0;
	bes->srcy = 0;
	bes->srcwidth = 720;
	bes->srcheight = 576;
	bes->x = 0;
	bes->y = 0;
	bes->dstwidth = 720;
	bes->dstheight = 576;
	bes->chromakey = 0;
	bes->overlay_mode = 0;
	bes->palette = VIDEO_PALETTE_YUV422;
	bes->buf_a = 0;
	bes->buf_b = 0;
	bes->bufwidth = 720;
	bes->bufheight = 576;
	bes->frame = BES_FRAME_B;
	bes->mode = 0;
	
	mga_create_proc_entry(bes->mga, "bes", mgabes_read_proc, 
		mgabes_write_proc, bes);

}


void mgabes_deattach(struct mga_bes_s *bes)
{
	struct mga_dev_s	*mga;

	mga = bes->mga;

	mga_remove_proc_entry(mga, "bes");

	kfree(bes);
	mga->bes = NULL;
}


static unsigned int
internal_pitch_rounding(struct mga_bes_s *bes, int width)
{
	unsigned int pitch;

	/*
	 * see p. 4-116 (566) G400 spec
	 * and p. 4-99  (477) G200 spec
	 */
	if ((bes->palette == VIDEO_PALETTE_YUV420) ||
	    (bes->palette == VIDEO_PALETTE_YUV420P)) {
			pitch = (width + 7) & ~7;
	} else { /* palette == VIDEO_PALETTE_YUV422 */
		if (bes->mga->mga_type == MGA_G400)
			pitch = (width + 7) & ~7;
		else
			pitch = (width + 3) & ~3;
	}
	return pitch;
}

/**
 *	mgabes_pitch_rounding - 	return the pitch for a given width
 *	@bes:		bes instance
 *	@width:		width of the video frame
 *
 *	This function returns the pitch value depending on the video
 *	format, the video buffer width and the card type.
 *
 *	FIXME:	This could be a macro/inline function usable from
 *	userland.
 */
unsigned int mgabes_pitch_rounding(struct mga_bes_s *bes, int width)
{
	unsigned int rv;
	unsigned long flags;

	spin_lock_irqsave( &bes->lock, flags );

	rv = internal_pitch_rounding( bes, width );

	spin_unlock_irqrestore( &bes->lock, flags );
	return rv;	
}



static void compute_horizontal_scaling(struct mga_bes_s *bes)
{
	u32 intrep, ifactor, ifactorbetter, condition, roundoff,
		acczoom; 
	u32 cropped_width;
	struct bes_registers_s *regs = &bes->regs;

	/* see section 4.10.2 p. 4-118 (566) in the G400 spec */ 
	regs->beshsrclst = (bes->bufwidth - 1) << 16;

	/* 4.10.2.1 : Horizontal Inverse Scaling Factor */
	cropped_width = bes->srcwidth;
	if (bes->mode & BES_HFILTER)
		intrep = (bes->dstwidth == cropped_width) ? 0 : 1;
	else
		intrep = (bes->dstwidth >= cropped_width) ? 0 : 1;
	ifactor = ((cropped_width - intrep) << 14) / (bes->dstwidth - intrep);
	ifactorbetter = ((cropped_width - intrep) << 20) / (bes->dstwidth - intrep) + intrep;
	condition = ( ((ifactorbetter * (bes->dstwidth-1)) >> 20) > ((ifactor * (bes->dstwidth-1)) >> 14) );
	if (condition)
		roundoff = (bes->mode & BES_HFILTER) ? 0 : 1;
	else
		roundoff = 0;
	acczoom = (bes->mode & BES_ACC2XZOOM) ? 2 : 1;
	regs->beshiscal = (acczoom * ifactor) + roundoff;
	if (regs->beshiscal >= (32 << 14))
		regs->beshiscal = (32 << 14) - 1;
	if (regs->beshiscal < ((bes->srcx + bes->srcwidth) << 14) / 16384)
		regs->beshiscal = ((bes->srcx + bes->srcwidth) << 14) / 16384;

	/* Horizontal source positioning */
	regs->beshsrcst = (bes->srcx << 14) + bes->offset_left * (ifactor + roundoff);
	regs->beshsrcend = regs->beshsrcst + ((bes->dstwidth - bes->offset_left - bes->offset_right - 1) / acczoom)
	             * (acczoom * ifactor + roundoff);

	/* adjust some registers */
	regs->beshsrcend <<= 2;
	regs->beshsrcst <<= 2;
	regs->beshiscal <<= 2;
}

static void
compute_vertical_source_positionning(struct mga_bes_s *bes, int field, u32 sc)
{
	struct bes_registers_s *regs = &bes->regs;
	u32 croptop, cropped_height;
	u32 bes420pl, dataformat, format420, bes3plane;
	int weight;
	u32 vsrclst, vsrcstp, val, absval, tmp;
	u32 base_address, origin_address;
	u32 chroma_plane_base_address, chroma_plane_origin_address;
	u32 chroma_3plane_base_address, chroma_3plane_origin_address;
	u32 modified_pitch;

	croptop = bes->srcy;
	cropped_height = bes->srcheight;

	if (bes->mode & BES_DEINTERLACE) {
		croptop = croptop / 2;
		cropped_height = cropped_height / 2;
		modified_pitch = regs->bespitch * 2;
	} else {
		modified_pitch = regs->bespitch;
	}


	/*
	 * Vertical source positioning
	 * (see p. 4-124 (572) in the G400 spec)
	 *
	 */
	switch (bes->palette) {
	case VIDEO_PALETTE_YUV420:
		bes420pl = 1;
		bes3plane = 0;
		dataformat = 1;
		format420 = 1;
		break;
	case VIDEO_PALETTE_YUV420P:
		bes420pl = 1;
		bes3plane = 1;
		dataformat = 1;
		format420 = 2;
		break;
	case VIDEO_PALETTE_YUV422:
	default:
		bes420pl = 0;
		bes3plane = 0;
		dataformat = 2;
		format420 = 1;
		break;
	}

	if (regs->besviscal >= (1 << 14)) {
		/* If downscaling */
		val = croptop + ((bes->offset_top * regs->besviscal + sc) >> 14);
		origin_address = val * modified_pitch * dataformat;

		chroma_plane_origin_address = (val / (2 * format420)) * modified_pitch;

		chroma_3plane_origin_address = (val / 4) * modified_pitch;

/*?l		if (bes->mode & BES_DEINTERLACE) {
			if ((field == BES_FIELD_A1) || (field == BES_FIELD_B1)) {
				tmp = croptop + ((bes->offset_top * regs->besviscal) >> 14);
				vsrclst = croptop + cropped_height - 1 - tmp;
			} else {
				tmp = croptop - ((bes->offset_top * regs->besviscal + sc) >> 14);
				vsrclst = croptop + cropped_height - 1 - tmp;
			}
		} else {
			tmp = croptop + ((bes->offset_top * regs->besviscal) >> 14);
			vsrclst = croptop + cropped_height - 1 - tmp;
		}		
		vsrcstp = (tmp & (1 << 14)) ? 1 : 0;
*/		vsrcstp = val & 1;
		vsrclst = croptop + cropped_height - 1 - (val & 0x03ff);
		tmp = val;
		

		if ((bes->mode & BES_DEINTERLACE) &&
		    ((field == BES_FIELD_A2) || (field == BES_FIELD_B2))) {

			if (bes->offset_top == 0) {
//?l				weight = -((1<<13) + regs->besviscal);
				weight = -((1<<13) + (regs->besviscal&0x3fff) );
			} else {
				weight = (bes->offset_top * regs->besviscal - 
					  (1 << 13));
			}
		} else {
			weight = (bes->offset_top * regs->besviscal);
		}
	} else {
		absval = croptop + ((abs((signed int)bes->offset_top - (sc>>14)) * regs->besviscal) >> 14);
		
		origin_address = absval * modified_pitch * dataformat;
		chroma_plane_origin_address = (absval / (2 * format420)) * modified_pitch;
		chroma_3plane_origin_address = (absval / 4) * modified_pitch;

		if ((bes->mode & BES_DEINTERLACE)
		    && ((field == BES_FIELD_A2) || (field == BES_FIELD_B2)))
			if ((bes->offset_top * regs->besviscal) >= (1<<13)) {
				weight = -(1<<13) + bes->offset_top;
			} else {
				weight = -((1<<13) + bes->offset_top * regs->besviscal);
			}
		else {
			weight = bes->offset_top * regs->besviscal;
		}

		if (bes->mode & BES_DEINTERLACE)
			vsrclst = bes->srcheight / 2 - 1;
		else
			vsrclst = bes->srcheight - 1;

//vsrclst = croptop + cropped_height - 1 - (absval & 0x03ff);
									
//		vsrcstp = (((croptop << 14) + absval) & (1 << 14)) ? 1 : 0;
		vsrcstp = absval & 1;
	}

	/* FIXME: add mirror parameters if mirroring is supported */

	tmp = ((weight & 0x3FFF) << 2) | ((weight>>15) & 0x10000);

	switch (field) {
	case BES_FIELD_A1:
	case BES_FIELD_B1:
		regs->besctl &= ~(1<<6); /* clear besv1srcstp */
		regs->besctl |= (vsrcstp&1)  << 6; /* set besv1srcstp */
		regs->besv1wght = tmp;
		regs->besv1srclst = vsrclst;
		break;

	case BES_FIELD_A2:
	case BES_FIELD_B2:
		regs->besv2wght = tmp;
		regs->besv2srclst = vsrclst;
		regs->besctl &= ~(1<<7); /* clear besv2srcstp */
		/* shouldn't v2srcstp be inverse of field 1? -ink */
		regs->besctl |= (~vsrcstp&1)  << 7;
		break;
	default:
		printk( KERN_ERR "mgabes: wrong field\n" );
		break;
	}
	
	if (bes3plane)
		regs->besglobctl |= 0x20; /* bes3plane */
	if (bes420pl)
		regs->besctl |= 0x20000; /* bes420pl */
		
	chroma_plane_base_address = regs->bespitch * dataformat * bes->bufheight; //bes->srcheight;
	chroma_3plane_base_address = chroma_plane_base_address +
				//	(regs->bespitch * bes->srcheight) / 4;
					(regs->bespitch * bes->bufheight) / 4;

	switch (field) {
	default:
	case BES_FIELD_A1:
		base_address = bes->buf_a;
		regs->besa1org = base_address + origin_address;
		regs->topfield_offset = origin_address;
		regs->besa1corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besa1c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_A2:
		base_address = bes->buf_a;
		regs->besa2org = base_address + origin_address +
			regs->bespitch * dataformat;
		regs->bottomfield_offset =  origin_address +
			regs->bespitch * dataformat;
		regs->besa2corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besa2c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_B1:
		base_address = bes->buf_b;
		regs->besb1org = base_address + origin_address;
		regs->topfield_offset = origin_address;
		regs->besb1corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besb1c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_B2:
		base_address = bes->buf_b;
		regs->besb2org = base_address + origin_address +
			regs->bespitch * dataformat;
		regs->bottomfield_offset =  origin_address +
			regs->bespitch * dataformat;
		regs->besb2corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besb2c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	}
}

static void
compute_vertical_scaling(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs = &bes->regs;
	u32 cropped_height;

	u32 intrep, ifactor, ifactorbetter, condition, roundoff;
	u32 subpixelcomp, h;

	cropped_height = bes->srcheight;
	h = bes->dstheight;
	
	if (bes->mode & BES_DEINTERLACE) {
		cropped_height /= 2;
	}
	
	/* Inverse Vertical scaling factor */
	if (bes->mode & BES_VFILTER)
		intrep = (bes->dstheight == cropped_height) ? 0 : 1;
	else
		intrep = (bes->dstheight >= cropped_height) ? 0 : 1;
	ifactor = ( ((cropped_height - intrep) << 14) / (h - intrep) ) & 0x0007ffff;
	ifactorbetter = ( ((cropped_height - intrep) << 20) / (h - intrep) & 0x01ffffff ) + intrep;
	condition = ( (((long long)ifactorbetter * (h-1)) >> 20) > ((ifactor * (h-1)) >> 14) );
	if (condition)
		roundoff = (bes->mode & BES_VFILTER) ? 0 : 1;
	else
		roundoff = 0;
		
	regs->besviscal = ifactor + roundoff;
	if (regs->besviscal >= (32 << 14))
		regs->besviscal = (32 << 14) - 1;
	if (regs->besviscal < ((bes->srcy + bes->srcheight) << 14) / 16384 )	/* or (croptop+cropheight)? */
		regs->besviscal = ((bes->srcy + bes->srcheight) << 14) / 16384;

	/* Vertical subpixel compensation */
	if (regs->besviscal >= (1 << 14)) /* Downscaling */
		subpixelcomp = (regs->besviscal - (1 << 13)) & 0x7ffff;
	else                        /* Upscaling */
		subpixelcomp = ( ((h - intrep) << 14)
		                / ((cropped_height - intrep) * 2) ) & 0x7ffff;

	if ((bes->mode & BES_DEINTERLACE) == 0 )
		subpixelcomp = 0;

	compute_vertical_source_positionning(bes, BES_FIELD_A1, 0);
	compute_vertical_source_positionning(bes, BES_FIELD_A2, subpixelcomp);
	compute_vertical_source_positionning(bes, BES_FIELD_B1, 0);
	compute_vertical_source_positionning(bes, BES_FIELD_B2, subpixelcomp);
	
	regs->besviscal <<= 2;
}


static void mgabes_write_regs(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs = &bes->regs;
	struct mga_dev_s *mga = bes->mga;
	u32 vcnt;

	/* Make sure internal registers don't get updated until we're done */
	vcnt = mga_read(mga, VCOUNT);

	if (vcnt == 0) vcnt = 480; /* should = vmaxlines? */
	/* is this a possibly invalid assumption about screen size?
	   should it not instead know how many lines there are?
	 -ink */

	mga_write(mga, BESGLOBCTL, (vcnt-1) << 16);
	
	mga_write(mga, BESCTL, regs->besctl);
	if (bes->mode & BES_DEINTERLACE)
		mga_write(mga, BESPITCH, regs->bespitch * 2); 
	else
		mga_write(mga, BESPITCH, regs->bespitch); 
	
	mga_write(mga, BESA1ORG, regs->besa1org);
	mga_write(mga, BESA1CORG, regs->besa1corg);
	mga_write(mga, BESA2ORG, regs->besa2org);
	mga_write(mga, BESA2CORG, regs->besa2corg);
	mga_write(mga, BESB1ORG, regs->besb1org);
	mga_write(mga, BESB1CORG, regs->besb1corg);
	mga_write(mga, BESB2ORG, regs->besb2org);
	mga_write(mga, BESB2CORG, regs->besb2corg);

	if(bes->mga->mga_type == MGA_G400) {
		mga_write(mga, BESLUMACTL, regs->beslumactl); 
		mga_write(mga, BESA1C3ORG, regs->besa1c3org);
		mga_write(mga, BESA2C3ORG, regs->besa2c3org);
		mga_write(mga, BESB1C3ORG, regs->besb1c3org);
		mga_write(mga, BESB2C3ORG, regs->besb2c3org);
	}
	mga_write(mga, BESHCOORD, regs->beshcoord);
	mga_write(mga, BESHISCAL, regs->beshiscal);
	mga_write(mga, BESHSRCST,  regs->beshsrcst);
	mga_write(mga, BESHSRCEND, regs->beshsrcend);
	mga_write(mga, BESHSRCLST, regs->beshsrclst);
	
	mga_write(mga, BESVCOORD, regs->besvcoord);
	mga_write(mga, BESVISCAL, regs->besviscal);
	mga_write(mga, BESV1SRCLST, regs->besv1srclst);
	mga_write(mga, BESV1WGHT, regs->besv1wght);
	mga_write(mga, BESV2SRCLST, regs->besv2srclst);
	mga_write(mga, BESV2WGHT, regs->besv2wght);
	
	/* update the registers somewhere between 1 and 2 frames from now. */
	mga_write(mga, BESGLOBCTL, regs->besglobctl);
}

/* return 1 if interlaced, 0 if not */
static int
mga_is_interlaced( struct mga_dev_s *mga )
{
	unsigned long flags;
	int interlace;
	
	spin_lock_irqsave( &mga->reg_lock, flags );

	/*   have to point to the CRTC extension register we want to read,
	   then read it. bit 7 is the "in interlaced mode" flag -ink */
	mga_writeb(mga, CRTCEXTADD, 0x00); /* CRTCEXT0 is reg we want */
	interlace = (mga_readb(mga, CRTCEXTREG) & 0x80) != 0;
	
	spin_unlock_irqrestore( &mga->reg_lock, flags );
	
	return interlace;
}


/* see p. 4-96 in the G200 spec */
static void mgabes_program(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs;
	u32 bes_was_enabled;
	int		width;
	int		height;
	int		depth;

	mga_display_size(bes->mga, &width, &height, &depth);
	
	if (mga_is_interlaced(bes->mga)) {
		height /= 2;
	}

	regs = &bes->regs;

	/* Compute destination cropping (window partially visible) */
	bes->offset_left = (bes->x < 0) ? -(bes->x) : 0;
	bes->offset_right = (bes->x + bes->dstwidth > width) ?
			bes->x + bes->dstwidth - width : 0;
	bes->offset_top = (bes->y < 0) ? -(bes->y) : 0;
	bes->offset_bottom = (bes->y + bes->dstheight > height) ?
			bes->y + bes->dstheight - height : 0;

	/* window coordinates */
	regs->beshcoord = (bes->x + bes->dstwidth - bes->offset_right - 1)
			+ ((bes->x + bes->offset_left) << 16);
	regs->besvcoord = (bes->y + bes->dstheight - bes->offset_bottom - 1)
			+ ((bes->y + bes->offset_top) << 16);

	regs->bespitch = internal_pitch_rounding(bes, bes->bufwidth);

	/* save this value, as we will overwrite it in the next subroutine call */
	bes_was_enabled = (bes->regs.besctl & 1);

	bes->regs.besctl = 0;
	bes->regs.besglobctl = 0;
	compute_horizontal_scaling(bes);
	compute_vertical_scaling(bes);

	/* 
	 * see p. 3-52 G200 specs
	 */
	bes->regs.besctl |= bes_was_enabled
		| ((bes->mode & BES_HFILTER) ? (1<<10) : 0)
		| ((bes->mode & BES_VFILTER) ? (1<<11) : 0)
		| ((bes->mode & BES_HFILTER) ? (1<<16) : 0)
		| ((bes->mode & BES_DITHER) ? (1<<18) : 0)
		| ((bes->mode & BES_BW) ? (1<<20) : 0)
		| ((bes->frame & BES_FRAME_AUTO) ? (1<<24) : 0);
	
	/* if not auto field select, load the bes field to display */
	if (!(bes->frame & BES_FRAME_AUTO))
		regs->besctl |= ((bes->frame & 3) << 25);

	regs->besglobctl |= ((bes->mode & BES_ACC2XZOOM) ? 1 : 0)
		| ((bes->mode & BES_HFILTER) ? 2 : 0);

	if (bes->mga->mga_type == MGA_G400) {
		/* enable brightness/contrast control */
		regs->besglobctl |= 0x00000080;
		/* set brightness/contrast */
		regs->beslumactl = ((bes_brightness & 0xff) << 16) | (bes_contrast & 0xff);
	}
	
	mgabes_write_regs(bes);
}

/**
 *	mgabes_activate -	activate the BES
 *	@bes:	bes instance
 *
 *	This function actives the BackEnd Scaler, so that the video
 *	overlay is displayed on screen.
 */
void mgabes_activate(struct mga_bes_s* bes)
{
	unsigned long flags;

	if (!bes) return;

	spin_lock_irqsave( &bes->lock, flags );

        /* Activate scaler */
	bes->regs.besctl |= 1;
       	mgabes_write_regs(bes);

	spin_unlock_irqrestore( &bes->lock, flags );
}


/**
 *	mgabes_disable -	disables the BES
 *	@bes:	bes instance
 *
 *	This function stops the BackEnd Scaler, disabling the
 *	display of overlaid video on screen.
 */
void mgabes_desactivate(struct mga_bes_s* bes)
{
	unsigned long flags;

	if (!bes) return;

	spin_lock_irqsave( &bes->lock, flags );

	bes->regs.besctl &= ~1;
	/*
	* do not reprogram the whole register set,
	* just stop the BES, because we're not sure
	* that the parameters are still valid and won't
	* crash the computation routines.
	*/
	mga_write(bes->mga, BESCTL, bes->regs.besctl);

	spin_unlock_irqrestore( &bes->lock, flags );
}

/**
 *	mgabes_set_window -	change displayed video window
 *	@bes:	bes instance
 *	@x:	x coordinate of the overlay window
 *	@y:	y coordinate of the overlay window
 *	@width:	width of the overlay window
 *	@height: height of the overlay window
 *
 *	This function moves and changes the size of the overlay window.
 *
 *	Notes: The BES must be activated for this function to have a
 *	*visible* effect... However, the BES parameters are immediatly
 *	re-programmed.
 */
void mgabes_set_window(struct mga_bes_s *bes, int x, int y, int width, int height)
{
	unsigned long flags;
	struct mga_dev_s *mga = bes->mga;
	int hh = height;
	int hy = y;
	int interlace = 0;

	if (!bes) return;

	/* check for interlace mode, if so, halve height and y offset -ink */
	interlace = mga_is_interlaced(mga);
	
	if (interlace) {
		hh = height >> 1;
		hy = y >> 1;
 	}

	spin_lock_irqsave( &bes->lock, flags );

	/* sanity limits, should it return EINVAL if out of range? -ink */
	if ((width >= 24) && (hh >= 18)) {
		bes->x = x;
		bes->y = hy;
		bes->dstwidth = width;
		bes->dstheight = hh;
		mgabes_program(bes);
	}
	spin_unlock_irqrestore( &bes->lock, flags );
}

/*
 *	@base_a:	base address of video buffer A
 *	@base_b:	base address of video buffer B
 *	@dataformat:	format of the video data
 */
void mgabes_set_buffers(struct mga_bes_s *bes, u32 base_a, u32 base_b, int width, int height, int dataformat)
{
	unsigned long flags;
	spin_lock_irqsave( &bes->lock, flags );

	bes->buf_a = base_a;
	bes->buf_b = base_b;
	bes->bufwidth = width;
	bes->bufheight = height;
	bes->palette = dataformat;
	mgabes_program(bes);

	spin_unlock_irqrestore( &bes->lock, flags );
}


/**
 *	mgabes_set_source -	set the source video buffer
 *	@bes:		bes instance
 *	@width:		width of a source video frame
 *	@height:	height of a source video frame

 *	This function attaches the BES to one or two source video
 *	buffers, with the specified dimensions and data format.
 *	
 *	The 2nd buffer can be NULL. In this case, the Automatic Field
 *	Select mode shouldn't be used.
 *
 *	Formats:	%VIDEO_PALETTE_YUV420,
 *	%VIDEO_PALETTE_YUV420P, and %VIDEO_PALETTE_YUV422
 *
 *	Note:	This function will only have an effect after
 *	mgabes_enable() is called.
 *
 *	See also:	%VIDEO_PALETTE_ and struct &video_buffer in
 *	<videodev.h> 
 */
void mgabes_set_source(struct mga_bes_s *bes, int x, int y, int width, int height)
{
	unsigned long flags;
	spin_lock_irqsave( &bes->lock, flags );

	bes->srcx = x;
	bes->srcy = y;
	bes->srcwidth = width;
	bes->srcheight = height;
	mgabes_program(bes);

	spin_unlock_irqrestore( &bes->lock, flags );
}

static void
internal_set_keymode(struct mga_bes_s *bes, int mode)
{
	struct	mga_dev_s *mga;

	mga = bes->mga;
  
	if (mode) {
		bes->overlay_mode = 1;
		mga_indirect_writeb(mga, XKEYOPMODE, 1);
		dprintk("mgabes: color keying enabled\n");
	} else {
		bes->overlay_mode = 0;
		mga_indirect_writeb(mga, XKEYOPMODE, 0);
		dprintk("mgabes: color keying disabled\n");
	}
}

/**	mgabes_set_keymode - changes overlay keying mode
 *
 *	@bes:		bes instance
 *	mode:		0 - overlay forced, 1 colorkey
 *
 *	This function immediatly update the keyopmode used for
 *	the overlay window.
 */
void
mgabes_set_keymode(struct mga_bes_s *bes, int mode)
{
	unsigned long flags;

	spin_lock_irqsave( &bes->lock, flags );
	spin_lock( &bes->mga->reg_lock );

	internal_set_keymode( bes, mode );

	spin_unlock( &bes->mga->reg_lock );
	spin_unlock_irqrestore( &bes->lock, flags );
}


/**
 *	mgabes_set_colorkey -	change overlay colorkey
 *	@bes:		bes instance
 *	@chromakey:	new chromakey
 *	
 *	This function immediatly updates the chromakey used for
 *	the overlay window.
 */
void mgabes_set_colorkey(struct mga_bes_s *bes, u32 chromakey)
{
	struct mga_dev_s *mga;
	int r, g, b;
	unsigned long flags;
	unsigned long colorkey;


	r = g = b = 0;
	mga = bes->mga;
	
	spin_lock_irqsave( &bes->lock, flags );
	spin_lock( &mga->reg_lock );

	colorkey = chromakey>>8; /* what's using bottom 8 bits? -ink */

	dprintk("mgabes: chromakey %06lX\n",colorkey);


/* chromakey RGB value is unpacked, so using X pixel value directly only
 * works for 24 and 32 bpp. it's user job to convert X/any colors.
 * all we do here is limit RGB values to what current mode supports
 -ink */

	switch (mga_indirect_readb(mga, XMULCTRL)) {

	case BPP_8:
		/* Need to look up the color index, just using
		   color 0 for now. */
	/* PALRDADD? -ink */
		break;

	case BPP_15:
		/* 555 RGB bits */
		r = ( colorkey >> 16) & 0x1F;
		g = ( colorkey >> 8) & 0x1F;
		b = ( colorkey ) & 0x1F;
		break;

	case BPP_16:
		r = ( colorkey >> 16) & 0x1F;
		g = ( colorkey >> 8) & 0x3F;
		b = ( colorkey ) & 0x1F;
		break;

	case BPP_24:
	case BPP_32_DIR:
	case BPP_32_PAL:
		r = ( colorkey >> 16) & 0xFF;
		g = ( colorkey >> 8) & 0xFF;
		b = ( colorkey ) & 0xFF;

	default:
		break;
	}

	internal_set_keymode(bes, 1);

	/* not using ranged colors currently */
	mga_indirect_writeb(mga, XCOLMSK0RED, 0xff);
	mga_indirect_writeb(mga, XCOLMSK0GREEN, 0xff);
	mga_indirect_writeb(mga, XCOLMSK0BLUE, 0xff);

	mga_indirect_writeb(mga, XCOLKEY0RED, r);
	mga_indirect_writeb(mga, XCOLKEY0GREEN, g);
	mga_indirect_writeb(mga, XCOLKEY0BLUE, b);

	dprintk("mgabes: colorkey R%02X G%02X B%02X\n", r, g, b );

	spin_unlock( &mga->reg_lock );

	spin_unlock_irqrestore( &bes->lock, flags );
}

/**
 *	mgabes_set_frame -	change the displayed frame
 *	@bes:	bes instance
 *	@frame:	frame to display
 *
 *	mga_set_frame() changes the frame displayed in the overlay
 *	window. The 'frame' parameter can be %BES_FRAME_A or
 *	%BES_FRAME_B to toggle between one of the 2 buffers. The
 *	%BES_FRAME_AUTO parameter should be used in conjunction with
 *	the video grabber to automagically preview the captured video
 *	stream.
 *
 *	Notes:	This function reprograms the BES immediatly. You
 *	should make sure that valid parameters have been entered
 *	previously (like video source buffers or video window).
 *
 *	See also:	4.10.5.2 Automatic Field Select p. 4-111 of
 *	the G200 spec.
 */
void mgabes_set_frame(struct mga_bes_s *bes, int frame)
{
	unsigned long flags;
	spin_lock_irqsave( &bes->lock, flags );

	bes->frame = frame;
	mgabes_program(bes);

	spin_unlock_irqrestore( &bes->lock, flags );
}

/* This is a faster version of it, to use when only the frame changes. */
/* Other parameters are kept the same, and not even reprogrammed, so */
/* one should call mgabes_set_frame once befor using the quick version. */

void mgabes_quick_set_frame(struct mga_bes_s *bes, int frame, u32 buffer)
{
	unsigned long flags;
	spin_lock_irqsave( &bes->lock, flags );

	if (!(bes->mode & BES_DEINTERLACE)) {
		/* Interlaced preview : ignore field 1 requests if last field is 2 */
		/* ignore field 2 requests if last field is 1 */
		/* And always dispaly field1 (no subpixel compensation) */
		if ( ( (bes->mode & BES_WAITFIELD2) != 0 && (frame & 1) == 0 )
		   ||( (bes->mode & BES_WAITFIELD2) == 0 && (frame & 1) != 0 ) ) {
			spin_unlock_irqrestore( &bes->lock, flags );
			return;
		}
		frame = frame & 2;
	}			

	bes->frame = frame & 3;
	bes->regs.besctl &= ~(3 << 25);
	bes->regs.besctl |= (bes->frame << 25);
	mga_write(bes->mga, BESCTL, bes->regs.besctl);

	/* Note from Benoit : is the following code necessary ? */
	/* we always put the same values is the registers ... */

	/* (gz): & 0x01 can't be right, if we want to check 4 values !! */
	/* (bm): right, also above ... but I think the code here should */
	/*       be removed : it a job for the setup routine to set these */
	/*       registers. No reason the value change aferwards. */
#ifdef NEVER
	switch(frame & 0x03) {
	case 0:
		bes->regs.besa1org = buffer + bes->regs.topfield_offset;
		mga_write(bes->mga, BESA1ORG, bes->regs.besa1org);
		mga_write(bes->mga, BESB1ORG, bes->regs.besa1org);
		break;

	case 1:
		bes->regs.besa2org = buffer + bes->regs.bottomfield_offset;
		mga_write(bes->mga, BESA2ORG, bes->regs.besa2org);
		mga_write(bes->mga, BESB2ORG, bes->regs.besa2org);
		break;


	case 2:
		bes->regs.besb1org = buffer + bes->regs.topfield_offset;
		mga_write(bes->mga, BESB1ORG, bes->regs.besb1org);
		break;

	case 3:
		bes->regs.besb2org = buffer + bes->regs.bottomfield_offset;
		mga_write(bes->mga, BESB2ORG, bes->regs.besb2org);
		break;
	}
#endif
	spin_unlock_irqrestore( &bes->lock, flags );
}

/**
 *	mgabes_set_mode -	change BES mode/options
 *	@bes:	bes instance
 *	@mode:	new mode
 *
 *	mga_bes_set_mode() changes the operating mode and or options
 *	of the BES.
 *
 *	Modes:
 *	BES_HFILTER	horizontal filtering
 *
 *	BES_VFILTER	vertical filterting
 *
 *	BES_ACC2XZOOM	2X Zoom (for high MCLKs)
 *
 *	BES_DITHER	dither enabled
 *
 *	BES_BW		black & white mode
 *
 *	BES_DEINTERLACE	source video data is interlaced
 *
 *	Note:	This function will only have an effect after
 *	mgabes_enable() is called.
 *
 */
void mgabes_set_mode(struct mga_bes_s *bes, int mode)
{
	unsigned long flags;
	spin_lock_irqsave( &bes->lock, flags );

	bes->mode = mode;

	spin_unlock_irqrestore( &bes->lock, flags );
}


int mgabes_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data)
{
	struct mga_bes_s *bes = (struct mga_bes_s*)data;
	char *s = page;
	int len;
	unsigned long flags;
	char *bfmt;
	spin_lock_irqsave( &bes->lock, flags );

	/* Build the message (it is less than 4k, no buffer overrun) */
	s = s + sprintf(s, "bes %s\n", (bes->regs.besctl & 1) ?
			"enabled" : "disabled");
	s = s + sprintf(s, "buffers: 0x%x 0x%x %d x %d\n",
		bes->buf_a, bes->buf_b, bes->bufwidth, bes->bufheight);
	s = s + sprintf(s, "source %d,%d %d x %d\n",
		bes->srcx, bes->srcy, bes->srcwidth, bes->srcheight);
	s = s + sprintf(s, "window %d,%d  %d x %d\n",
		bes->x, bes->y, bes->dstwidth, bes->dstheight);
	s = s + sprintf(s, "color_key_rgb %d %d %d\n",
		(bes->chromakey >> 24) & 0xff, (bes->chromakey >> 16) & 0xff, 
		(bes->chromakey >> 8) & 0xff);
	if (bes->overlay_mode)
		s = s + sprintf(s, "overlay 1 (use colorkey)\n");
	else
		s = s + sprintf(s, "overlay 0 (don't use colorkey)\n");


	s = s + sprintf(s, "horiz_filter %s\n",
		(bes->mode & BES_HFILTER) ? "1" : "0");
	s = s + sprintf(s, "vfilter_filter %s\n",
		(bes->mode & BES_VFILTER) ? "1" : "0");
	s = s + sprintf(s, "acc2xzoom %s\n",
		(bes->mode & BES_ACC2XZOOM) ? "1" : "0");
	s = s + sprintf(s, "dither %s\n",
		(bes->mode & BES_DITHER) ? "1" : "0");
	s = s + sprintf(s, "bw %s\n",
		(bes->mode & BES_BW) ? "1" : "0");
	s = s + sprintf(s, "deinterlaced_preview %s\n",
		(bes->mode & BES_DEINTERLACE) ? "1" : "0");

	s = s + sprintf(s, "globctl 0x%8X\n", bes->regs.besglobctl );

	s = s + sprintf(s, "ctl 0x%8X\n", bes->regs.besctl );


	switch (bes->palette) {
	case VIDEO_PALETTE_YUV422:
		bfmt = "VIDEO_PALETTE_YUV422";
		break;
	case VIDEO_PALETTE_YUV420:
		bfmt = "VIDEO_PALETTE_YUV420";
		break;
	case VIDEO_PALETTE_YUV420P:
		bfmt = "VIDEO_PALETTE_YUV420P";
		break;
	default:
		bfmt = "INVALID!";
		break;
	}
	s = s + sprintf(s, "format %s\n", bfmt );
			

	s = s + sprintf(s, "\n\nbeshiscal: %08x whole %d fract %d\n", 
			bes->regs.beshiscal, bes->regs.beshiscal >> 16,
			(bes->regs.beshiscal & 0xffff) >> 2);
	s = s + sprintf(s, "beshcoord: %08x  left=%d right=%d\n",
			bes->regs.beshcoord, bes->regs.beshcoord >> 16,
			bes->regs.beshcoord & 0xffff);
	s = s + sprintf(s, "beshsrcst: %08x whole %d fract %d\n", 
			bes->regs.beshsrcst, bes->regs.beshsrcst >> 16,
			(bes->regs.beshsrcst & 0xFFFF) >> 2);
	s = s + sprintf(s, "beshsrcend: %08x whole %d fract %d\n", 
			bes->regs.beshsrcend, bes->regs.beshsrcend >> 16,
			(bes->regs.beshsrcend & 0xFFFF) >> 2);
	s = s + sprintf(s, "beshsrclst: %08x %d",
			bes->regs.beshsrclst, bes->regs.beshsrclst);

	s = s + sprintf(s, "\n\nbesviscal: %08x whole %d fract %d\n", 
			bes->regs.besviscal, bes->regs.besviscal >> 16,
			(bes->regs.besviscal & 0xffff) >> 2);
	s = s + sprintf(s, "besvcoord: %08x  top=%d bottom=%d\n",
			bes->regs.besvcoord, bes->regs.besvcoord >> 16,
			bes->regs.besvcoord & 0xffff);
	s = s + sprintf(s, "besv1srclst: %08x %d\n", 
			bes->regs.besv1srclst, bes->regs.besv1srclst);
	s = s + sprintf(s, "besv1wght: %08x sign %d weight %d\n", 
			bes->regs.besv1wght, bes->regs.besv1wght >> 16,
			(bes->regs.besv1wght & 0xFFFF) >> 2);
	s = s + sprintf(s, "besv2srclst: %08x %d\n", 
			bes->regs.besv2srclst, bes->regs.besv2srclst);
	s = s + sprintf(s, "besv2wght: %08x sign %d weight %d\n", 
			bes->regs.besv2wght, bes->regs.besv2wght >> 16,
			(bes->regs.besv2wght & 0xFFFF) >> 2);

	spin_unlock_irqrestore( &bes->lock, flags );

	len = s - page;

	return len;
}


#define READ_ERROR { printk ("bes_write_proc : invalid data\n"); return count; }
/* /proc routines */
#define ERR_FORMAT -1
#define ERR_OVERFLOW -2

int proc_read_uint(const char *s, unsigned int *val)
{
	unsigned int tmp = 0;
	int to_add;
	int err = 0;
	int stop = 0;
	int readhex = 0;
	int i = 0, i0;
	/* skip leading whitespaces */
	while (s[i] == ' ' || s[i] == '\t')
		i++;
	i0 = i;
	if (s[i] == '0' && s[i+1] == 'x') {
		i += 2;
		readhex = 1;
	}
	do {
		to_add = -1;
		if (s[i] == ' ' || s[i] == '\n' || s[i] == '\t' || s[i] == 0) {
			if (i == i0) err = ERR_FORMAT;
			stop = 1;
		} else if (s[i] >= '0' && s[i] <= '9') {
			to_add = s[i] - '0';
		} else if (s[i] >= 'a' && s[i] <= 'f') {
			if (!readhex)
				err = ERR_FORMAT;
			else
				to_add = s[i] - 'a' + 10;
		} else if (s[i] >= 'A' && s[i] <= 'F') {
			if (!readhex)
				err = ERR_FORMAT;
			else
				to_add = s[i] - 'A' + 10;
		} else {
			err = ERR_FORMAT;
		}
		if (to_add>=0) {
			if (readhex) {
				if (tmp & 0xF0000000)
					err = ERR_OVERFLOW;
				else
					tmp = (tmp << 4) + to_add;
			} else {
				if (tmp > 0x19999999 || (tmp == 0x19999999 && to_add > 5))
					err = ERR_OVERFLOW;
				else
					tmp = tmp * 10 + to_add;
			}
		}
		i++;
	} while (!stop && !err);
	if (!err) {
		*val = tmp;
		err = i;
	}
	return err;
}



int proc_read_int(const char *s, int *val)
{
	unsigned int tmp;
	int err, i = 0;

	/* skip leading whitespaces */
	while (s[i] == ' ' || s[i] == '\t')
		i++;
	
	if (s[i] == '-') {
		i++;
		err = proc_read_uint(s+i, &tmp);
		if (err<0) return err;
		if (tmp & 0x80000000) return ERR_OVERFLOW;
		*val = -tmp;
		return err + i;
	} else {
		err = proc_read_uint(s+i, &tmp);
		if (err<0) return err;
		if (tmp & 0x80000000) return ERR_OVERFLOW;
		*val = tmp;
		return err + i;
	}
	return err;
}

#define DO 1
int mgabes_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data)
{
	struct mga_bes_s *bes = (struct mga_bes_s *)data;
	int err, s1, s2, s3;
	unsigned int u1, u2, setflag = 0;
	char *flagname = 0;

	if (!strncmp(buf,"window",6)) {
		buf += 6;  if ((err = proc_read_int(buf, &s1)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_int(buf, &s2)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_uint(buf, &u1)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_uint(buf, &u2)) < 0) READ_ERROR
		mgabes_set_window(bes, s1, s2, u1, u2);
	} else if (!strncmp(buf,"source",6)) {
		buf += 6;  if ((err = proc_read_int(buf, &s1)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_int(buf, &s2)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_uint(buf, &u1)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_uint(buf, &u2)) < 0) READ_ERROR
		mgabes_set_source (bes, s1, s2, u1, u2);
	} else if (!strncmp(buf,"color_key_rgb",13)) {
		buf += 13; if ((err = proc_read_int(buf, &s1)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_int(buf, &s2)) < 0) READ_ERROR
		buf += err;if ((err = proc_read_int(buf, &s3)) < 0) READ_ERROR
		mgabes_set_colorkey(bes, (s1 << 24) | (s2 << 16) | (s3 << 8));
	} else if (!strncmp(buf,"overlay",7)) {
		buf += 7;  if ((err = proc_read_uint(buf, &u1)) < 0) READ_ERROR
		mgabes_set_keymode(bes, u1);
	} else if (!strncmp(buf,"PREVIEW_ACTIVE",14)) {
		buf += 14; if ((err = proc_read_uint(buf, &u1)) < 0) READ_ERROR
#ifdef	NEVER
		mgavideo_preview_enable (mga, u1);
#endif
	} else if (!strncmp(buf,"deinterlace_preview",18)) {
		buf += 19; setflag = BES_DEINTERLACE; flagname = "deinterlace_preview";
	} else if (!strncmp(buf,"horiz_filter",12)) {
		buf += 12; setflag = BES_HFILTER; flagname = "horiz_filter";
	} else if (!strncmp(buf,"vert_filter",11)) {
		buf += 11; setflag = BES_VFILTER; flagname = "vert_filter";
	} else if (!strncmp(buf,"ACC2X_ZOOM",10)) {
#ifdef	DO
		buf += 10; setflag = BES_ACC2XZOOM; flagname = "ACC2X_ZOOM";
#endif
	} else if (!strncmp(buf,"black_white",11)) {
		buf += 11; setflag = BES_BW; flagname = "black_white";
	} else if (!strncmp(buf,"blank",5)) {
		buf += 5; setflag = BES_BLANK; flagname = "blank";
	} else if (!strncmp(buf,"dither",6)) {
		buf += 6; setflag = BES_DITHER; flagname = "dither";
	} else if (!strncmp(buf,"horiz_mirror",12)) {
		buf += 12; setflag = BES_MIRROR; flagname = "horiz_mirror";
	} 

	if (setflag) {
		if ((err = proc_read_uint(buf, &u1)) < 0) READ_ERROR
		u1 = (u1 ? setflag : 0);
		bes->mode = (u1 == setflag) ? (bes->mode | setflag) : (bes->mode & ~setflag);
		/* Reprogram the bes ! */
		mgabes_set_frame(bes, bes->frame);
	}
	
	return count;		
}

#undef READ_ERROR

