/*
 * An abstraction of the various busses on the card
 *
 * Copyright (C) 2000  Leon van Stuivenberg <leonvs@iae.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/proc_fs.h>
#include "mgabus.h"
#include "mgacore.h"
#include "ks0127.h"

MODULE_PARM(mgabus_debug, "i");

int	mgabus_debug;

#define dprintk if (mgabus_debug) printk

#define MAX_BUSSES	8


struct mgabus_admin_s
{
	struct mga_dev_s *mga;
       	struct proc_dir_entry *proc_directory;
	
	struct mgabus_s*	busses[MAX_BUSSES];
};



EXPORT_SYMBOL(mgabus_new);
EXPORT_SYMBOL(mgabus_free);
EXPORT_SYMBOL(mgabus_attach);
EXPORT_SYMBOL(mgabus_detach);
EXPORT_SYMBOL(mgabus_set_master);
EXPORT_SYMBOL(mgabus_set_source);
EXPORT_SYMBOL(mgabus_init_client);
EXPORT_SYMBOL(mgabus_get_input);
EXPORT_SYMBOL(mgabus_get_format);
EXPORT_SYMBOL(mgabus_get_dir);
EXPORT_SYMBOL(mgabus_set_source_addr);
EXPORT_SYMBOL(mgabus_get_source_addr);
EXPORT_SYMBOL(mgabus_notify);
EXPORT_SYMBOL(mgabus_add_observer);
EXPORT_SYMBOL(mgabus_remove_observer);



struct id_to_sym_s
{
	int id;
	const char *sym;
};

static const char*
id_to_ascii( struct id_to_sym_s *lut, int size, int id )
{
	int i;
	for ( i=0; i<size/sizeof(*lut); i++ ) {
		if ( lut[ i ].id == id )
			return lut[ i ].sym;
	}
	return "None";
}


int bus_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data)
{
	struct mgabus_s *bus = (struct mgabus_s*)data;
	char *s = page;
	int i, j;
	int len;

	static struct id_to_sym_s print_norm[] = {
		{ KS_STD_NTSC, "NTSC" },
		{ KS_STD_PAL, "PAL" },
		{ KS_STD_NTSC_M, "NTSC 50Hz" },
		{ KS_STD_PAL_N, "PAL 60Hz" },
		{ KS_STD_SECAM, "SECAM" }
	};


	
	for ( i=0; i<BUS_MAXOBSERVERS; i++ )
	if (bus->observers[ i ] != NULL) {
		s = s + sprintf( s, "observer: %p(%p)\n",
			bus->observers[ i ],
			bus->observer_data[ i ] );
	}
	s = s + sprintf( s, "\n" );
	

	for ( i=0; i<BUS_MAXCLIENTS; i++ )
	if (bus->clients[ i ] != NULL ) {
		struct mgabus_client_s *cl = bus->clients[ i ];
		
		s = s + sprintf(s, "client at 0x%04x, type %i: %s",
			cl->addr, cl->type, cl->name );
		if ( cl == bus->master )
			s = s + sprintf(s, " <MASTER>" );
		if ( cl == bus->source ) {
			s = s + sprintf(s, " <SOURCE>" );
			
			s = s + sprintf( s, " DEST=" );
			if ( bus->dest_addr == BUS_ADDR_NIL )
				s = s + sprintf( s, "(nil)" );
			else if ( bus->dest_addr == BUS_ADDR_BCAST )
				s = s + sprintf( s, "broadcast" );
			else
				s = s + sprintf( s, "%04x", bus->dest_addr );
		}
			
		s = s + sprintf(s, "\n" );
		
		if ( cl->type == BUSCLIENT_VIDEO ) {
		
			struct mgabus_client_vid_s *vid = (struct mgabus_client_vid_s*)cl;
			
			s = s + sprintf(s, "\tnumber inputs: %i\n", vid->nr_inputs );
			s = s + sprintf(s, "\tactive input: %i\n", vid->active_input );
			
			if ( cl == bus->source ) {
				struct mgabus_video_input_s fmt;
			
				for (j=0; j<vid->nr_inputs; j++) {
				
					s = s + sprintf(s, "\tinput %i:\n", j );

					if ( mgabus_get_format( bus, j, &fmt ) != 0 ) {
						s = s + sprintf(s, "\t\t(no format info)\n" );
						continue;
					}

					s = s + sprintf(s, "\t\tstandard: %s\n",
						id_to_ascii( print_norm, sizeof(print_norm), fmt.norm ) );
						
					s = s + sprintf(s , "\t\tBrightness: %d Hue: %d Contrast: %d Saturation: %d\n",
						fmt.brightness,	
						fmt.hue,	
						fmt.contrast,	
						fmt.color);
						
					s = s + sprintf(s, "\t\tCropping: %d,%d %dx%d\n",
						fmt.crop.x,
						fmt.crop.y,
						fmt.crop.width,
						fmt.crop.height);
				}
			}
			s = s +sprintf(s, "\n");
		}
	}

	len = s - page;

	return len;
}


int bus_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data)
{
	struct mgabus_s *bus = (struct mgabus_s*)data;

	(void)bus;

	return count;		
}




static struct proc_dir_entry*
bus_create_proc_entry(struct mgabus_admin_s *admin, char *name,
	read_proc_t *rp, write_proc_t *wp, void *data)
{
	struct proc_dir_entry	*entry;

	if ((entry = create_proc_entry(name, S_IFREG | S_IRUGO | S_IWUSR,
		admin->proc_directory)) == NULL) {
		return NULL;
	}

	entry->data = data;
	entry->read_proc = rp;
	entry->write_proc = wp;

	return entry;
}


static void
bus_remove_proc_entry(struct mgabus_admin_s *admin, char *name)
{
	if (admin->proc_directory) {
		remove_proc_entry(name, admin->proc_directory);
	}
}




static int
nop_set_dir( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data )
{
	return 0;
}

static int
nop_set_ctl( struct mgabus_s *bus, enum bus_control ctl, void *data )
{
	return 0;
}

enum bus_direction
nop_get_dir( struct mgabus_s *bus, void *data )
{
	return BUSDIR_SOURCE;
}

void
mgabus_init_client( void *client, enum busclient_type type, const char *name, int addr, void *data )
{
	struct mgabus_client_s *base = (struct mgabus_client_s*)client;
	
	base->type = type;
	strcpy( base->name, name );
	base->addr = addr;
	base->data = data;
	base->usage = 0;
	base->set_dir = nop_set_dir;
	base->get_dir = nop_get_dir;
	base->set_ctl = nop_set_ctl;
	
	if ( type == BUSCLIENT_VIDEO ) {
		struct mgabus_client_vid_s *vid = (struct mgabus_client_vid_s*)client;
		
		vid->nr_inputs = 0;
		vid->active_input = -1;
		vid->get_format = NULL;
	}
}


struct mgabus_s*
mgabus_new( struct mgabus_admin_s *admin, size_t size, const char *name )
{
	struct mgabus_s *bus;
	int i;
	
	bus = kmalloc(size, GFP_KERNEL);
	if ( bus == NULL ) {
		return NULL;
	}
	memset(bus, 0, size);

	bus->admin = admin;
	strcpy( bus->name, name );
	bus->n_clients = 0;
	bus->source = NULL;
	bus->master = NULL;
	bus->dest_addr = BUS_ADDR_NIL;
	
	for (i=0; i<MAX_BUSSES; i++) {
		if ( admin->busses[ i ] == NULL ) {
			admin->busses[ i ] = bus;
			break;
		}
	}

	bus_create_proc_entry(admin, bus->name, bus_read_proc, 
		bus_write_proc, bus);
			
	return bus;
}


int
mgabus_free( struct mgabus_s *bus )
{
	struct mgabus_admin_s *admin = bus->admin;
	int i;
	
	for (i=0; i<MAX_BUSSES; i++) {
		if ( admin->busses[ i ] == bus ) {
			admin->busses[ i ] = NULL;
			break;
		}
	}
	
	bus_remove_proc_entry(admin, bus->name);

	if ( bus != NULL )
		kfree( bus );
	return 0;
}


int
mgabus_attach( struct mgabus_s *bus, struct mgabus_client_s *client )
{
	int i;
	
	for ( i=0; i<BUS_MAXCLIENTS; i++ ) {
		if (bus->clients[ i ] == NULL ) {
			bus->clients[ i ] = client;
			bus->n_clients++;
			return 0;
		}
	}
	return 1;
}


int
mgabus_detach( struct mgabus_s *bus, struct mgabus_client_s *client )
{
	int i;
	
	for ( i=0; i<BUS_MAXCLIENTS; i++ ) {
		if (bus->clients[ i ] == client ) {
			bus->clients[ i ] = NULL;
			bus->n_clients--;
			if (bus->master == client)
				bus->master = NULL;
			if (bus->source == client)
				bus->source = NULL;
			return 0;
		}
	}
	return 1;
}


int
mgabus_set_master( struct mgabus_s *bus, struct mgabus_client_s *new_master )
{
	int i;
	
	for ( i=0; i<BUS_MAXCLIENTS; i++ ) {
		if ( (bus->clients[ i ] != NULL) &&
			(bus->clients[ i ] != new_master) ) {
			if (bus->clients[ i ]->set_ctl != NULL) {
				if ( bus->clients[ i ]->set_ctl( bus, BUSCTL_SLAVE, bus->clients[ i ]->data ) ) {
					dprintk( "failed to slave %i, %s\n",
						i,
						bus->clients[ i ]->name );
					return 1;
				}
			}
		}
	}
	bus->master = NULL;
	for ( i=0; i<BUS_MAXCLIENTS; i++ ) {
		if ( (bus->clients[ i ] != NULL) &&
			(bus->clients[ i ] == new_master) ) {
			if (bus->clients[ i ]->set_ctl != NULL) {
			}
			
		}
	}
//todo

return -1;	

}


static int
mgabus_set_source_internal( struct mgabus_s *bus, struct mgabus_client_s *new_source, int dest_addr )
{
	int i;
	int new_src;

	new_src = -1;
	for ( i=0; i<BUS_MAXCLIENTS; i++ ) {
		if (bus->clients[ i ] != NULL) {
			if (bus->clients[ i ] != new_source) {
			
				enum bus_direction new_dir =
					(dest_addr == BUS_ADDR_BCAST) ||
					(bus->clients[ i ]->addr == dest_addr)
					? BUSDIR_SINK :	BUSDIR_TRISTATE;
				
				if ( bus->clients[ i ]->get_dir( bus, bus->clients[ i ]->data )
					!= new_dir ) {
			
					if ( bus->clients[ i ]->set_dir( bus,
						new_dir,
						dest_addr, bus->clients[ i ]->data ) ) {
					
						dprintk( "source switch fails: sink %s\n", bus->clients[ i ]->name );
						return 1;
					}
				}

				if (bus->clients[ i ] == bus->source)
					bus->source = NULL;
			}
			else {
				new_src = i;
			}
		}
	}
	
	if ( (new_src != -1) && (new_source != bus->source) ) {
	
		enum bus_direction new_dir = BUSDIR_SOURCE;
	
		if ( bus->clients[ new_src ]->get_dir( bus, bus->clients[ new_src ]->data )
			!= new_dir ) {

			if ( bus->clients[ new_src ]->set_dir( bus,
				new_dir,
				dest_addr,
				bus->clients[ new_src ]->data ) ) {
				dprintk( "source switch fails: source %s\n", bus->clients[ new_src ]->name );
				return 1;
			}
		}
	}
	bus->source = new_source;
	bus->dest_addr = dest_addr;
	return 0;
}


static int
mgabus_validate(struct mgabus_s *bus)
{
	int i;
	int rv;
		
	for ( i=0; i<BUS_MAXCLIENTS; i++ )
	if (bus->clients[ i ] != NULL ) {
		struct mgabus_client_s *cl = bus->clients[ i ];

//dprintk( "validate client %s\n", cl->name );		

		if ( cl->get_dir != NULL ) {
			enum bus_direction dir;
			dir = cl->get_dir( bus, cl->data );
			
			if ( cl == bus->source ) {
				if (dir != BUSDIR_SOURCE) {
					rv = cl->set_dir( bus, BUSDIR_SOURCE,
						bus->dest_addr, cl->data );
					if (rv) {
						bus->source = NULL;
					}
				}
			}
			else {
				enum bus_direction ndir = (bus->dest_addr == BUS_ADDR_BCAST)
					|| (cl->addr == bus->dest_addr)
					? BUSDIR_SINK :	BUSDIR_TRISTATE;
				if (dir != ndir) {
					rv = cl->set_dir( bus, ndir,
						bus->dest_addr, cl->data );
					if (rv) {
						printk( "mgabus: client %s refused dir %i\n", cl->name, ndir );
						return 1;
					}
				}
			}
		}
	}
	return 0;
}


static int
mgabus_admin_validate(struct mgabus_admin_s *admin)
{
	int i;
	int rv;

	for (i=0; i<MAX_BUSSES; i++)
	if ( admin->busses[ i ] != NULL ) {
//dprintk( "validate bus %s\n", admin->busses[ i ]->name );	
		rv = mgabus_validate( admin->busses[ i ] );
		if ( rv )
			return rv;
	}
	return 0;
}


int
mgabus_set_source( struct mgabus_s *bus, struct mgabus_client_s *new_source, int dest_addr )
{
	int rv;
	
	rv = mgabus_set_source_internal( bus, new_source, dest_addr );
	if ( rv ) {
		return rv;
	}
	rv = mgabus_admin_validate( bus->admin );
	return rv;
}


int
mgabus_set_source_addr( struct mgabus_s *bus, int src_addr, int dest_addr )
{
	int i;
		
	for ( i=0; i<BUS_MAXCLIENTS; i++ )
	if (bus->clients[ i ] != NULL ) {
		struct mgabus_client_s *cl = bus->clients[ i ];
		if ( cl->addr == src_addr ) {
			return mgabus_set_source( bus, cl, dest_addr );
		}
	}
	return 1;
}


int
mgabus_get_source_addr( struct mgabus_s *bus )
{
	if ( bus->source != NULL )
		return bus->source->addr;
	return -1;
}


int
mgabus_get_input( struct mgabus_s *bus )
{
	if (bus->source != NULL) {
		if (bus->source->type == BUSCLIENT_VIDEO) {
			struct mgabus_client_vid_s *src = ((struct mgabus_client_vid_s*)bus->source);
			
			if ( (0 <= src->active_input) && (src->active_input < src->nr_inputs) )
				return src->active_input;
		}
	}
	dprintk( "mgabus_get_input: no input\n" );
	return -1;
}


int
mgabus_get_format( struct mgabus_s *bus, int input, struct mgabus_video_input_s *params )
{
	if ((input != -1) && (bus->source != NULL)) {
		if (bus->source->type == BUSCLIENT_VIDEO) {
			struct mgabus_client_vid_s *src = 
				(struct mgabus_client_vid_s*)bus->source;
			if (src != NULL)
				return src->get_format(
					bus, input, params, bus->source->data );
		}
	}
	dprintk( "mgabus_get_format: no source/input\n" );
	return -1;
}


void
mgabus_notify(struct mgabus_s *bus, enum bus_notify_code code )
{
	int i;
		
	for ( i=0; i<BUS_MAXOBSERVERS; i++ )
	if (bus->observers[ i ] != NULL) {
		bus->observers[ i ]( bus, code, bus->observer_data[ i ] );
	}
	
	/*
	for ( i=0; i<BUS_MAXCLIENTS; i++ )
	if (bus->clients[ i ] != NULL ) {
		struct mgabus_client_s *cl = bus->clients[ i ];
		if (cl->type == BUSCLIENT_VIDEO) {
			struct mgabus_client_vid_s *vcl = (struct mgabus_client_vid_s*)cl;
			if (vcl->notify != NULL)
				vcl->notify( bus, code, cl->data );
		}
	}
	*/
}


int
mgabus_add_observer(struct mgabus_s *bus, bus_observer_t obs, void *data)
{
	int i;

	for ( i=0; i<BUS_MAXOBSERVERS; i++ )
	if (bus->observers[ i ] == NULL) {
		bus->observers[ i ] = obs;
		bus->observer_data[ i ] = data;
		return 0;
	}
	return 1;
}


void
mgabus_remove_observer(struct mgabus_s *bus, bus_observer_t obs)
{
	int i;

	for ( i=0; i<BUS_MAXOBSERVERS; i++ )
	if (bus->observers[ i ] == obs) {
		bus->observers[ i ] = NULL;
		bus->observer_data[ i ] = NULL;
	}
}


enum bus_direction
mgabus_get_dir( struct mgabus_s *bus, int addr )
{
	int i;
		
	for ( i=0; i<BUS_MAXCLIENTS; i++ )
	if (bus->clients[ i ] != NULL ) {
		struct mgabus_client_s *cl = bus->clients[ i ];
		
		if ( cl->addr == addr ) {
			if ( cl->get_dir != NULL )
				return cl->get_dir( bus, cl->data );
			break;
		}
	}
	return (enum bus_direction)-1;
}



void
mgabus_admin_attach(struct mga_dev_s *mga)
{
	struct mgabus_admin_s *admin;

	mga->bus_admin = admin = kmalloc(sizeof(*admin), GFP_KERNEL);
	if (admin == NULL)
		return;

	memset(admin, 0, sizeof(*admin));

	admin->mga = mga;

	admin->proc_directory = create_proc_entry( "bus", S_IFDIR, mga->proc_directory);

//no!	MOD_INC_USE_COUNT;
}


void
mgabus_admin_deattach(struct mgabus_admin_s *admin)
{
	if (admin->proc_directory) {
		remove_proc_entry( "bus", admin->mga->proc_directory);
		admin->proc_directory = NULL;
	}

	admin->mga->bus_admin = NULL;
	kfree(admin);

//no!	MOD_DEC_USE_COUNT;
}

