#ifndef __MGABUS_H
#define __MGABUS_H


#define BUS_MAXCLIENTS	4	/* max nr of attached clients to a bus */
#define BUS_ADDR_BCAST	-1	/* broadcast address */
#define BUS_ADDR_NIL	-2	/* no address */

#define BUS_MAXOBSERVERS	8

struct mga_dev_s;

struct mgabus_s;
struct mgabus_client_s;
struct mgabus_client_vid_s;
struct mgabus_admin_s;


enum bus_direction
{
	BUSDIR_SOURCE,		/* client is sending data onto the bus */
	BUSDIR_SINK,		/* client is receiving data from the bus */
	BUSDIR_TRISTATE		/* client is tri-stated */
};

enum bus_control
{
	BUSCTL_MASTER,		/* client is clock master */
	BUSCTL_SLAVE		/* client is clock slave */
};

enum busclient_type
{
	BUSCLIENT_BASE,		/* client is of class 'base' */
	BUSCLIENT_VIDEO		/* client is of subclass 'video' */
};


/** A base client that attaches to a bus.
 */	
struct mgabus_client_s {
	enum busclient_type	type;		/* class type */
	char			name[32];	/* user-friendly name for /proc */
	void			*data;		/* data passed back to client in methods */
	int			addr;		/* bus address of client */
	int			usage;		/* (not used) */
	
	/** set tranceiver state of client
	 *  @param dir	new state
	 *  @param addr	destination address of requested connection
	 *  @return <>0 on error
	 */
	int			(*set_dir)( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data );
	
	/** get tranceiver state of client
	 */
	enum bus_direction	(*get_dir)( struct mgabus_s *bus, void *data );
	
	/** (not used)
	 */
	int			(*set_ctl)( struct mgabus_s *bus, enum bus_control, void *data );
	
};


struct mgabus_video_area_s {
	u16	x;
	u16	y;
	u16	width;
	u16	height;
};


/** Parameters of the signal on a video bus
 */
struct mgabus_video_input_s {
	int	norm;		/* video standard (pal,ntsc,secam) */
				//fixme: using KS_xx constants for the moment..
	
        short	brightness;	/* -128..127 */
	short	hue;		/* -128..127 */
	short	contrast;	/* -128..127 */
	short	color;		/* -128..127 */
	
	u16	h_total;
	u16	v_total;
	struct mgabus_video_area_s active;	/* active area */
	struct mgabus_video_area_s crop;	/* used active area */
	
	u8	vbicnt0;	/* VBI line count for field 0 */
	u8	vblcnt0;	/* vertical blanking line count for field 0 */
	u8	vbicnt1;	/* VBI line count for field 1 */
	u8	vblcnt1;	/* vertical blanking line count for field 1 */
};


/** A client that attaches to a video bus.
 */	
struct mgabus_client_vid_s {
	struct mgabus_client_s super;	/* super class */
	
	int	nr_inputs;		/* nr of inputs the client has available when
					   it's sending */
	int	active_input;		/* currently active input */
	
	/** get video format of specified input
	 *  @param input	requested input
	 *  @param params	structure to be filled in by client
	 */
	int	(*get_format)( struct mgabus_s *bus, int input,
			struct mgabus_video_input_s *params, void *data );
};


enum bus_notify_code
{
	BUSCODE_FORMAT,
	BUSCODE_INPUT
};

/** signals the observer about a change of (video) bus configuration
 */
typedef void	(*bus_observer_t)( struct mgabus_s *bus, enum bus_notify_code code, void *data );
	

/** A bus (codec/video/..)
 */
struct mgabus_s {
	struct mgabus_admin_s	*admin;
	
	char			name[32];	/* user-friendly name for /proc */
	
	struct mgabus_client_s	*clients[ BUS_MAXCLIENTS ]; /* list of clients */
	int			n_clients;
	
	struct mgabus_client_s	*source;	/* the client that's currently sending */
	struct mgabus_client_s	*master;	/* the clock master */
	
	int			dest_addr;	/* destination of data sent by source */
	
	bus_observer_t		observers[ BUS_MAXOBSERVERS ];
	void*			observer_data[ BUS_MAXOBSERVERS ];
};


/** initializes bus administration for specified card
 */
void mgabus_admin_attach(struct mga_dev_s *mga);

/** cleans up bus administration
 */
void mgabus_admin_deattach(struct mgabus_admin_s *admin);


/** initializes bus client structure
 */
void mgabus_init_client( void *client, enum busclient_type type,
		const char *name, int addr, void *data );


/** allocates and initializes a new bus structure
 */
struct mgabus_s* mgabus_new( struct mgabus_admin_s *admin, size_t size, const char *name );

/** cleans up a bus structure
 */
int mgabus_free( struct mgabus_s *bus );

/** attaches a client to the specified bus
 */
int mgabus_attach( struct mgabus_s *bus, struct mgabus_client_s *client );

/** detaches a client from the specfied bus
 */
int mgabus_detach( struct mgabus_s *bus, struct mgabus_client_s *client );

/**
 */
int mgabus_set_master( struct mgabus_s *bus, struct mgabus_client_s *new_master );

/** requests a new connection from the specified source
 *  to the destination address.
 *  @return <>0 if failed
 */
int mgabus_set_source( struct mgabus_s *bus, struct mgabus_client_s *new_source, int dest_addr );

/** requests a new connection from the specified (by address) source
 *  to the destination address.
 *  @return <>0 if failed
 */
int mgabus_set_source_addr( struct mgabus_s *bus, int src_addr, int dest_addr );

int mgabus_get_source_addr( struct mgabus_s *bus );

/** returns the currently active input of the current source
 *  (use only for video busses)
*/
int mgabus_get_input( struct mgabus_s *bus );

/** returns the format of the video on the bus
 *  (use only for video busses)
 */
int mgabus_get_format( struct mgabus_s *bus, int input, struct mgabus_video_input_s *params );

/** returns the state of the specified client
 *  (use only for video busses)
 */
enum bus_direction mgabus_get_dir( struct mgabus_s *bus, int addr );

/**
 */
void mgabus_notify(struct mgabus_s *bus, enum bus_notify_code code );

int mgabus_add_observer(struct mgabus_s *bus, bus_observer_t obs, void *data);
void mgabus_remove_observer(struct mgabus_s *bus, bus_observer_t obs);


/* bus overview

   (bus)			(addr)			(inputs)

  codec bus			
     |
     |--[mga codec interface]	@0x0001
     |
     |--[marvel/rrg zr36060]	@0x0033
     |
     |--[dvd module zr367xx]	@0x0034
     |
     x


  VIP bus
     |
     |--[mga video-in port]	@0x0001
     |
     |--[G400 CRTC2-656 port]	@0x0002
     |
     |--[KS0127]		@0x0127			(comp,SVHS,TV,zr)
     |
     |--[dvd module zr367xx]	@0x0034			(dvd)
     |
     x


  VDO bus
     |
     |--[mga VDOUT 11:0]	@0x0001			(vip,crt2-656,crt1-24,crt2-32)
     |
     |--[maven/MGA TVO]		@0x0004 (KP004)
     |
     x


  EXV bus				(not yet)
     |
     |--[KS0127]		@?			(comp,SVHS,TV)
     |
     |--[marvel/rrg zr36060]	@?			(zr)
     |
     x
     
*/

#endif

