/*
 * Various clock related functions
 *
 * Copyright (C) 2000  Leon van Stuivenberg <leonvs@iae.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/proc_fs.h>
#include "mgacore.h"
#include "mgaclk.h"


/* Memory clock speed register */
#define MCTLWTST 0x1c08


#define MAXPLL		310000
#define MAXPLL_G400	310000
#define MAXPIXCLK_G200	250000
#define MAXPIXCLK_G400	270000
#define MAXMCLK_G200	125000
#define MAXMCLK_G400	166667
#define MAXGCLK		 90000
#define MAXWCLK		100000
#define MAXCODCLK	 90000

#define PCICLK		 33000
#define AGPCLK		 66000
#define VDOCLK		 27000	//?

//14318.18
#define PIXFREF		 27050.5
#define SYSFREF		 27050.5

#define MHZ(f)		(((f)+500)/1000)
#define KHZ(f)		(f)
#define KLOG		KERN_ERR


EXPORT_SYMBOL(mga_get_clk);
EXPORT_SYMBOL(mga_mod_clkdiv_g400);
EXPORT_SYMBOL(mga_get_bios2);
EXPORT_SYMBOL(mga_check_mctlwtst);

MODULE_PARM(mgaclk_debug, "i");
MODULE_PARM(oclock_mem,"i");

static int mgaclk_debug = 0;
static int oclock_mem = 0; /* insmod parameter */



#define dprintk if (mgaclk_debug) printk



struct mctlwtst_s
{
	u8	casltncy;	/* 2 - 5 */
	u8	rrddelay;	/* 1 - 3 */
	u8	rcddelay;	/* 2 - 4 */
	u8	rasmin;		/* 3 - 10 */
	
	u8	rpdelay;	/* 2 - 5 */
	u8	wrdelay;	/* 1 - 2 */
	u8	rddelay;	/* 0 - 1 */
	u8	smrdelay;	/* 1 - 2 */
	
	u8	bwcdelay;	/* 1 - 2 */
	u8	bpldelay;	/* 1 - 6 */
};


static struct mctlwtst_s default_mctlwtst_sdram[] =
{
	{ 3, 1, 2, 6,	2, 1, 0, 1,	1, 1 }, /* template for custom wtst */
	{ 3, 1, 2, 6,	2, 1, 0, 1,	1, 1 }, /* from MGATweak */
	{ 3, 1, 2, 5,	2, 1, 0, 1,	1, 1 },	/* bm / MGATweak */
	{ 2, 1, 2, 3,	2, 1, 0, 1,	1, 1 },	/* MGATweak / max */
};
//	{ 3, 2, 3, 6,	3, 1, 0, 2,	1, 1 },	/* lvs, NEC 4516161 spec p.43, 125MHz(-80) */


static struct mctlwtst_s default_mctlwtst_sgram[] =
{
	{ 3, 1, 2, 6,	2, 1, 0, 1,	1, 1 }, /* template for custom wtst */
	{ 3, 1, 2, 6,	2, 1, 0, 1,	1, 1 }, /* from MGATweak */
	{ 3, 1, 2, 5,	2, 1, 0, 1,	1, 1 }	/* from MGATweak */
};




static unsigned
calc_pll( struct mga_dev_s *mga, int fref_kHz8, int m, int n, int p )
{
	unsigned Fo;
	unsigned pllmax;

	Fo = (fref_kHz8 * (n+1)) / ((m+1) * (p+1));
	Fo >>= 8;

	pllmax = mga->mga_type == MGA_G400 ? MAXPLL_G400 : MAXPLL;

	if ( Fo > pllmax ) {
		printk( KLOG "overclocked pll %i, %i\n", Fo, pllmax );
	}

	return Fo;
}

/* pre: have reg_lock spinlock
 */
static unsigned
get_pll( struct mga_dev_s *mga, int fref, u8 mnpbase )
{
	int m, n, p, s;

	m = mga_indirect_readb(mga, mnpbase + 0) & 0x1f;
	n = mga_indirect_readb(mga, mnpbase + 1) & 0x7f;
	p = mga_indirect_readb(mga, mnpbase + 2) & 0x07;
	s = (mga_indirect_readb(mga, mnpbase + 2) >> 3) & 0x03;

	return calc_pll( mga, fref, m, n, p );
}


static unsigned
calc_required_dcyc( int divider, unsigned clk_kHz )
{
	unsigned dcyc;
	
	if ( (divider == 2) || (divider == 4) ) {
		dcyc = ((4000000/clk_kHz)-9)/2;	/* (clk_in_ns-2.25)/0.5 */
		if (dcyc < 0) dcyc = 0;
		if (dcyc > 15) dcyc = 15;
	}
	else
		dcyc = 0;
	return dcyc;
}


static unsigned
calc_clk400( const char* msg, u8 divider, u8 clksel, unsigned syspll, unsigned testclk, unsigned dcyc )
{
	struct {
		u8 mul;
		u8 div;
	} pllmuldiv[] = { {1,3}, {2,5}, {4,9}, {1,2}, {2,3}, {1,1}, 
		{1,1}/*res*/, {1,1}/*res*/ };

	unsigned clk;
	unsigned dcycreq;

	switch ( clksel ) {
		case 0x00:
			clk = PCICLK;
			break;
		case 0x01:
			clk = syspll * pllmuldiv[ divider ].mul / pllmuldiv[ divider ].div;
			break;
		case 0x02:
			clk = testclk;
			break;
		case 0x03:
		default:
			clk = AGPCLK * pllmuldiv[ divider ].mul / pllmuldiv[ divider ].div;
			break;
	}
	if ( (divider == 2) || (divider == 4) ) {
		dcycreq = calc_required_dcyc( divider, clk );

		if ( dcyc != dcycreq ) {
			dprintk( "%s dcyc (%i) is set for %i..%i MHz, not for %i MHz\n", msg,
				dcyc,
				4000/((dcyc+1)*2+9),
				4000/(dcyc*2+9),
				clk/1000 );
		}
	}
	return clk;
}


static unsigned
calc_clk200( u8 divider, u8 nodiv, u8 sysclksl, unsigned syspll )
{
	if ( (sysclksl != 0x01) || nodiv)
		return syspll;
	else if ( divider )
		return syspll * 2 / 3;
	else
		return syspll / 2;
}



int
mga_get_clk( struct mga_dev_s *mga, int clk_id, int *ret_clk )
{
	struct {
		u8 mul;
		u8 div;
	} codmuldiv[] = { {1,1}, {1,2}, {2,3}, {1,3} };

	u32 option;
	u32 option2;
	u32 option3;

	unsigned pixpll;
	unsigned syspll;
	unsigned pix_fref;
	unsigned sys_fref;

	unsigned testclk;
	unsigned codclk;
	unsigned pixclk;
	unsigned memclk;
	unsigned warpclk;
	unsigned gfxclk;
	unsigned cmax;
	u8 divider;
	u8 mnpbase;
	u8 ctrl;
	u8 misc;
	
	unsigned long flags;
	

	testclk = 1;

	option = 0;
	option2 = 0;
	option3 = 0;

	if ( pci_read_config_dword( mga->dev, 0x40, &option ) != PCIBIOS_SUCCESSFUL ) {
		printk( KLOG "mgaclk: error reading CS x40\n" );
		return 1;
	}
	if ( pci_read_config_dword( mga->dev, 0x50, &option2 ) != PCIBIOS_SUCCESSFUL ) {
		printk( KLOG "mgaclk: error reading CS x50\n" );
		return 1;
	}
	if (mga->mga_type == MGA_G400) {
		if ( pci_read_config_dword( mga->dev, 0x54, &option3 ) != PCIBIOS_SUCCESSFUL ) {
			printk( KLOG "mgaclk: error reading CS x54\n" );
			return 1;
		}
	}
	
	dprintk( "clk opt x%08x x%08x x%08x\n", option, option2, option3 );

	pix_fref = (int)(PIXFREF * 256);
	sys_fref = (int)(SYSFREF * 256);
	
	spin_lock_irqsave( &mga->reg_lock, flags );

	syspll = get_pll( mga, sys_fref, 0x2c );

	misc = mga_readb( mga, CRTMISC );
	switch ((misc >> 2) & 0x03) {
		case 0x00:
			mnpbase = 0x44;
			break;
		case 0x01:
			mnpbase = 0x48;
			break;
		default:
			mnpbase = 0x4c;
			break;
	}
	pixpll = get_pll( mga, pix_fref, mnpbase );

	if (option & 0x40) { /* pllsel */
		/* pix and sys (p1/p2) reversed */
		unsigned tmp = pixpll;
		pixpll = syspll;
		syspll = tmp;
	}

	ctrl = mga_indirect_readb(mga, XPIXCLKCTRL);
	switch (ctrl & 0x03) { /* pixclksl */
		case 0x00:
			//pci
			pixclk = PCICLK;
			break;
		case 0x01:
			//pixpll
			pixclk = pixpll;
			break;
		case 0x02:
			//vdoclk
			pixclk = VDOCLK;
			break;
		case 0x03:
		default:
			pixclk = syspll;
			break;
	}
	
	spin_unlock_irqrestore( &mga->reg_lock, flags );

	if (mga->mga_type == MGA_G400) {
		divider = ( option2 >> 26 ) & 0x03; /* coddiv, reset: 10b */
		switch ((option2 >> 24) & 0x03) { /* codclksl, reset: 11b */
			case 0x00:
				//pci
				codclk = PCICLK;
				break;
			case 0x01:
				codclk = syspll * codmuldiv[ divider ].mul / codmuldiv[ divider ].div;
				break;
			case 0x02:
				codclk = testclk;
				break;
			case 0x03:
			default:
				//agpdll
				codclk = AGPCLK * codmuldiv[ divider ].mul / codmuldiv[ divider ].div;
				break;
		}
	
		gfxclk = calc_clk400( "gfxclk",
			(option3 >> 3) & 0x07,	/* gclkdiv */
			((option3 ) & 0x03),	/* gclksel */
			syspll, testclk,
			(option3 >> 6) & 0x0f	/* gclkdcyc */
			);

		memclk = calc_clk400( "memclk",
			(option3 >> 13) & 0x07,	/* mcldiv */
			(option3 >> 10) & 0x03, /* mclksel */
			syspll, testclk,
			(option3 >> 16) & 0x0f	/* mclkdcyc */
			);
			
		warpclk = calc_clk400( "warpclk",
			(option3 >> 23) & 0x07,	/* wclkdiv */
			(option3 >> 20) & 0x03, /* wclksel */
			syspll, testclk,
			(option3 >> 26) & 0x0f	/* wclkdcyc */
			);
		
		if ( !(pixclk < warpclk) ) {
			dprintk( "PCLK < WCLK constraint failed, %i %i\n", MHZ(pixclk), MHZ(warpclk) );
		}
		if ( !(gfxclk < 2*memclk) ) {
			dprintk( "GCLK < 2 X MCLK constraint failed, %i %i\n", MHZ(gfxclk), MHZ(2*memclk) );
		}
		if ( !(pixclk < memclk) ) {
			dprintk( "PCLK < MCLK constraint failed, %i %i\n", MHZ(pixclk), MHZ(memclk) );
		}
		if ( memclk < PCICLK ) {
			dprintk( "MCLK below PCICLK, %i %i\n", memclk, MHZ(PCICLK) );
		}
		if ( (memclk < codclk) || (memclk < gfxclk) || (memclk < warpclk) ) {
			dprintk( "MCLK too low, M%i C%i G%i W%i\n", MHZ(memclk), MHZ(codclk), MHZ(gfxclk), MHZ(warpclk) );
		}
	}
	else { //G200

		codclk = 0;
		
		gfxclk = calc_clk200(
			(option >> 3) & 1,	/* gclkdiv */
			(option2 >> 14) & 1,	/* nogclkdiv */
			option & 3,		/* sysclksl */
			syspll );

		memclk = calc_clk200(
			(option >> 4) & 1,	/* mclkdiv */
			(option2 >> 15) & 1,	/* nomclkdiv */
			option & 3,		/* sysclksl */
			syspll );

		warpclk = calc_clk200(
			(option2 >> 17) & 1,	/* wclkdiv */
			(option2 >> 16) & 1,	/* nowclkdiv */
			option & 3,		/* sysclksl */
			syspll );

	 	if ( warpclk < PCICLK ) {
			dprintk( "WCLK below PCICLK, %i %i\n", MHZ(warpclk), MHZ(PCICLK) );
		}
	}


	cmax = mga->mga_type == MGA_G400 ? MAXPIXCLK_G400 : MAXPIXCLK_G200;
	if ( pixclk > cmax ) {
		dprintk( "overclocked pixclk %i > %i\n", MHZ(pixclk), MHZ(cmax) );
	}

	cmax = mga->mga_type == MGA_G400 ? MAXMCLK_G400 : MAXMCLK_G200;
	if ( memclk > cmax ) {
		dprintk( "overclocked memclk %i > %i\n", MHZ(memclk), MHZ(cmax) );
	}

	if ( codclk > MAXCODCLK ) {
		dprintk( "overclocked codclk %i > %i\n", MHZ(codclk), MHZ(MAXCODCLK) );
	}
	
	if ( gfxclk > MAXGCLK ) {
		dprintk( "overclocked gfxclk %i > %i\n", MHZ(gfxclk), MHZ(MAXGCLK) );
	}

	if ( warpclk > MAXWCLK ) {
		dprintk( "overclocked warpclk %i > %i\n", MHZ(warpclk), MHZ(MAXWCLK) );
	}

	dprintk( "clk syspll %6i pixpll %6i pixel %6i\n", KHZ(syspll), KHZ(pixpll), KHZ(pixclk) );
	dprintk( "clk grafix %6i memory %6i warp  %6i\n", KHZ(gfxclk), KHZ(memclk), KHZ(warpclk) );
	dprintk( "clk codec  %6i\n", KHZ(codclk) );

	if ( ret_clk != NULL ) {
		switch ( clk_id ) {
			case CLK_PIXEL:
				*ret_clk = pixclk;
				break;
			case CLK_MEMORY:
				*ret_clk = memclk;
				break;
			case CLK_WARP:
				*ret_clk = warpclk;
				break;
			case CLK_GFXENG:
				*ret_clk = gfxclk;
				break;
			case CLK_CODEC:
				*ret_clk = codclk;
				break;
			default:
				return 1;
		}
	}
	return 0;
}





int
mga_mod_clkdiv_g400( struct mga_dev_s *mga, int clk_id, int div_delta )
{
	int div;
	unsigned dcyc;
	int clk;
	u32 option;
	u32 option3;
	u32 new_option3;

	if (mga->mga_type != MGA_G400)
		return 1;
	
	clk = 0;
	if ( mga_get_clk( mga, clk_id, &clk ) )
		return 1;

	option = 0;
	option3 = 0;

	if ( pci_read_config_dword( mga->dev, 0x40, &option ) != PCIBIOS_SUCCESSFUL ) {
		printk( "mgaclk: error reading CS x40\n" );
		return 1;
	}
	if ( pci_read_config_dword( mga->dev, 0x54, &option3 ) != PCIBIOS_SUCCESSFUL ) {
		printk( "mgaclk: error reading CS x54\n" );
		return 1;
	}
				
	switch ( clk_id ) {
		case CLK_GFXENG:
			div = (option3 >> 3) & 0x07;
			break;
		case CLK_MEMORY:	
			div = (option3 >> 13) & 0x07;
			break;
		case CLK_WARP:
			div = (option3 >> 23) & 0x07;
			break;
		default:
			div = 0;
			break;
	}
	
	div -= div_delta;
	if (div < 0) div = 0;
	if (div > 5) div = 5;
	
	dcyc = calc_required_dcyc( div, clk );
	
	switch ( clk_id ) {
		case CLK_GFXENG:
			new_option3 = (option3 & ~(0x7f<<3))  | (div<<3)  | (dcyc<<6);
			break;
		case CLK_MEMORY:
			new_option3 = (option3 & ~(0x7f<<13)) | (div<<13) | (dcyc<<16);
			break;
		case CLK_WARP:
			new_option3 = (option3 & ~(0x7f<<23)) | (div<<23) | (dcyc<<26);
			break;
		default:
			new_option3 = option3;
			break;
	}

	/* sysclkdis=1 */
	if ( pci_write_config_dword( mga->dev, 0x40, option | 0x04 ) != PCIBIOS_SUCCESSFUL ) {
		return 1;
	}
	pci_write_config_dword( mga->dev, 0x54, new_option3 );
	/* sysclkdis=0 */
	pci_write_config_dword( mga->dev, 0x40, option );
	
	return 0;
}




static int
mga_validate_bios( MGABios2Info* pbios )
{
	u8 chksum;
	u8* ppins;
	int i;
	
	if ( pbios->PinID != 0x412e ) {
		printk( "no pins info\n" );
		return 1;
	}
	
	/* check that the pins info is correct */
	if ( pbios->StructLen != 0x40 ) {
		printk( "bios struct len not %i but %i\n", 0x40, pbios->StructLen );
		return 1;
	}
	/* check that the chksum is correct */
	chksum = 0;
	ppins = (u8*)&pbios->PinID;
	for (i=0; i < pbios->StructLen; i++) {
		chksum += *ppins;
		ppins++;
	}
	if ( chksum ) {
		printk( "bios chksum error\n" );
		return 1;
	}

	if ( pbios->StructRev == 0 ) {
		printk( "bios invalid revision\n" );
		return 1;
	}
	return 0;
}

//#include <linux/delay.h>

int
mga_get_bios2(struct mga_dev_s *mga, MGABios2Info *biosinfo)
{
	u32 option;
	u32 rombase;
	u8* bios;
	u8 tmp[ 64 ];
	u32 offset;
	int rv;
//	int i;

	mga_writeb(mga, 0x3e40/*CODECCTL*/, 0); /* CODEC multiplexed on seeprom interface;
							needs to be disabled */
			
	option = 0;
	if ( pci_read_config_dword( mga->dev, 0x40, &option ) != PCIBIOS_SUCCESSFUL ) {
		return 1;
	}
	if ( (option & (1<<30)) == 0 ) {
		/* BIOS enable is off */
		printk( "biosen is off\n" );
		return 1;
	}
	
	rombase = 0;
	if ( pci_read_config_dword( mga->dev, 0x30, &rombase ) != PCIBIOS_SUCCESSFUL ) {
		printk( "bios i/o err\n" );
		return 1;
	}
	if ( (rombase & 1) == 0 ) {
		/* ROM enable is off */
		dprintk( "romen is off\n" );

		if ( pci_write_config_dword( mga->dev, 0x30, rombase | 1 ) != PCIBIOS_SUCCESSFUL ) {
			printk( "bios i/o err\n" );
			return 1;
		}
		
		if ( pci_read_config_dword( mga->dev, 0x30, &rombase ) != PCIBIOS_SUCCESSFUL ) {
			printk( "bios i/o err\n" );
			return 1;
		}
		if ( (rombase & 1) == 0 ) {
			printk( "romen still off\n" );
			return 1;
		}
	}
	rombase &= 0xffff0000;
	dprintk( "rombase %x\n", rombase );
	
	bios = (u8*)ioremap( rombase, 0x8000 );
	dprintk( "bios = %p\n", bios );

	memcpy_fromio( tmp, bios + 0, sizeof( tmp ) );
/*	for ( i=0; i<sizeof( tmp )/4; i++ ) {
		((u32*)tmp)[ i ] = readl( ((u32*)bios) + i );
		udelay(10);
	}
	for ( i=0; i<4; i++ )
		printk( "%02x ", (int)tmp[ i ] );
	for ( i=45; i<45+6; i++ )
		printk( "%c", tmp[ i ] );
	printk( "\n" );
*/	
	
	
	rv = 1;
        if ( tmp[ 0 ] != 0x55
                || tmp[ 1 ] != 0xaa
                || memcmp( tmp+45, "MATROX", 6 ) ) {
		printk( "Video BIOS info block not detected\n");
        }
	else {
		/* Get the info block offset */
		offset = readw( bios + 0x7ffc );
		dprintk( "offset = 0x%x\n", offset );
		if ( offset < 0x8000 ) {
			memcpy_fromio( biosinfo, bios + offset, sizeof(*biosinfo) );
			rv = mga_validate_bios( biosinfo );
		}
	}

	/* restore romen setting if we changed it */
	if ( (rombase & 1) == 0 ) {
		if ( pci_write_config_dword( mga->dev, 0x30, rombase ) != PCIBIOS_SUCCESSFUL ) {
			printk( "bios i/o err\n" );
		}
	}
	
	vfree( bios );
	
	return rv;
}


void
mga_check_mctlwtst(struct mga_dev_s *mga)
{
	if ( oclock_mem && (mga->mga_type == MGA_G200) ) {
	
		u32 option;	
		
		option = 0;
		if ( pci_read_config_dword( mga->dev, 0x40, &option ) != PCIBIOS_SUCCESSFUL ) {
			printk( KLOG "mgavid: read error CS 40\n" );
		} else {
			struct mctlwtst_s *ctl;
			int n;
			u32 wtst;
			
			if ( option & 0x00004000 ) {
				ctl = default_mctlwtst_sgram;
				n = sizeof(default_mctlwtst_sgram)/sizeof(*ctl);
			}
			else {
				ctl = default_mctlwtst_sdram;
				n = sizeof(default_mctlwtst_sdram)/sizeof(*ctl);
			}
			if ( (oclock_mem > 0) && (oclock_mem <= n) ) {
				struct mctlwtst_s custom_wtst;
			
				ctl += oclock_mem-1;
				
				if ( oclock_mem == 1 ) {
					/* adapt waitstates according to bios info */
					custom_wtst = *ctl;
					ctl = &custom_wtst;
					
					ctl->casltncy = 2 + ((mga->bios_mctlwtst>>0) & 1);
					ctl->rcddelay = 2 + ((mga->bios_mctlwtst>>1) & 1);
					ctl->rasmin   = 4 + ((mga->bios_mctlwtst>>2) & 3);
				}
			
				wtst = 	((ctl->casltncy-2)<<0) |
					((ctl->rrddelay-1)<<4) |
					((ctl->rcddelay-2)<<7) |
					((ctl->rasmin-3)<<10) |
					((ctl->rpdelay-2)<<14) |
					((ctl->wrdelay-1)<<18) |
					((ctl->rddelay)<<21) |
					((ctl->smrdelay-1)<<23) |
					((ctl->bwcdelay-1)<<26) |
					((ctl->bpldelay-1)<<29)
					;
					
				if ( (wtst & 0x12532248) == 0 ) {
					dprintk( "setting MCTLWTST 0x%08x\n", wtst );
					mga_write(mga, MCTLWTST, wtst );
				}
				else {
					printk( KLOG "invalid MCTLWTST 0x%08x\n", wtst );
				}
			} else {
				printk( KLOG "oclock_mem (%i) out of range [%i,%i]\n",
					oclock_mem, 1, n );
			}
		}	
	}
}

