/*
 * This module provides an interface to the core video capabilities
 * of the G100/G200/G400 chipset.
 *
 * (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Publi License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *      Mike Bernson <mike@mlb.org>
 *	Merrick Johnson <inkling@nop.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/videodev.h>
#include <linux/proc_fs.h>
#include <linux/wrapper.h>
#include "mgacore.h"
#include "mgabes.h"
#include "mgaclk.h"
#include "mgacodec.h"
#include "mgabus.h"
#include "mgai2c.h"

#include "ks0127.h"
#include "maven.h"
#include "audiochip.h"



MODULE_AUTHOR("David Barth <dbarth@besancon.net>");
MODULE_DESCRIPTION("Core Matrox module");

MODULE_PARM(mgacore_debug, "i");
MODULE_PARM(vdo_enable, "i");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

int	mgacore_debug = 0;
int	vdo_enable = 1;

#define dprintk if (mgacore_debug) printk

EXPORT_SYMBOL(mga_display_size);
EXPORT_SYMBOL(mga_get_card);
EXPORT_SYMBOL(mga_release_card);
EXPORT_SYMBOL(mga_alloc_buffer);
EXPORT_SYMBOL(mga_free_buffer);
EXPORT_SYMBOL(mga_enable_irq);
EXPORT_SYMBOL(mga_disable_irq);
EXPORT_SYMBOL(mga_create_proc_entry);
EXPORT_SYMBOL(mga_remove_proc_entry);
EXPORT_SYMBOL(mga_rvmalloc);
EXPORT_SYMBOL(mga_rvfree);
EXPORT_SYMBOL(mga_check_irqs);
EXPORT_SYMBOL(mga_display_pitch);
EXPORT_SYMBOL(mga_test_lines);
EXPORT_SYMBOL(mga_enable_core_irq);
EXPORT_SYMBOL(mga_disable_core_irq);
EXPORT_SYMBOL(mga_card_lock);
EXPORT_SYMBOL(mga_card_unlock);
EXPORT_SYMBOL(mga_i2c_control);
EXPORT_SYMBOL(mga_set_defaults);
//EXPORT_SYMBOL(mga_save_settings);
//EXPORT_SYMBOL(mga_restore_settings);
EXPORT_SYMBOL(mga_set_audioinput);


#define MAX_CARDS	4

#ifndef PCI_DEVICE_ID_MATROX_G400
#define PCI_DEVICE_ID_MATROX_G400 0x0525
#endif

#define BPP_8         0x00
#define BPP_15        0x01
#define BPP_16        0x02
#define BPP_24        0x03
#define BPP_32_DIR    0x04
#define BPP_32_PAL    0x07


#ifndef	MAP_NR
#define	MAP_NR(x)	virt_to_page(x)
#endif

static struct mga_dev_s card[MAX_CARDS];


static struct mga_pci_id_s
{
	unsigned short dev_id;
	int mga_type;
	const char *name;
} mga_pci_ids[] = {
	{ PCI_DEVICE_ID_MATROX_G400,		MGA_G400,	"MGA G400" },
	{ PCI_DEVICE_ID_MATROX_G200_AGP,	MGA_G200,	"MGA G200 AGP" },
	{ PCI_DEVICE_ID_MATROX_G200_PCI,	MGA_G200,	"MGA G200 PCI" },
	
	/* are any of these supportable by the driver?
	{ PCI_DEVICE_ID_MATROX_MIL,		MGA_G200,	"MGA Millenium" },
	{ PCI_DEVICE_ID_MATROX_MIL_2,		MGA_G200,	"MGA Millenium2" },
	{ PCI_DEVICE_ID_MATROX_MIL_2_AGP,	MGA_G200,	"MGA Millenium2 AGP" },
	
	{ PCI_DEVICE_ID_MATROX_MYS,		MGA_G200,	"MGA Mystique" },
	*/
};


/* /proc/mgavideo
 directory */
static struct proc_dir_entry *mga_procdir;

/* * PRIVATE FUNCTIOM *
 *	mga_interrupt_handler -	main interrupt handler
 *	@mga:		mga instance
 *	@dev_id:	
 *	@regs:
 *
 *	This function handles interrupt requests. It will call
 *	registered sub-handlers for each specific interrupt.
 *	
 *	See also:
 *	mga_request_irq()
 *
 *	mga_release_irq()
 *
 *	Notes:
 *	mga_interrupt_hw() is responsible for clearing the
 *	interrupt. Sub-handlers should not worry about it.
 */
static void mga_interrupt_handler(int irq, void *dev_id,
				  struct pt_regs *regs)
{
	u32 status, vstatus;
	struct mga_dev_s *mga = dev_id;
	int i, j, k;

	if (mga == NULL) {
		printk("mgacore: mga interrupt handler says *mga is NULL!\n");
		return;
	}

	spin_lock(&mga->irq_lock);

	for (i=0; i<8; i++) {

		/* read status register */
		vstatus = mga_read(mga, VSTATUS);
		status  = mga_read(mga, STATUS);

		if ( ((vstatus & mga->enabled_irq) == 0) && ((status & mga->enabled_core_irq) == 0) ) {
			/* from other dev on shared irq line? */
			spin_unlock(&mga->irq_lock);
			return;
		}
		mga->status = vstatus;
		
		if (i!=0) {
			dprintk( "mgacore: isr iteration %i, %x, %x\n", i, vstatus, status );
		}

		/* handle VIN irqs */
		for (j=0; j<MGA_MAX_VIRQ; j++) {
			if ((vstatus & (1<<j)) & (mga->enabled_irq & (1<<j))) {
			
				mga_write(mga, VICLEAR, (1<<j));

				for ( k=0; k<MGA_MAX_ISR; k++ ) {
					if (mga->irq_handler[j][k])
						(*mga->irq_handler[j][k])(mga, regs);
				}
			}
		}
				

		if (vstatus & (1<<MGA_IRQ_DCMPEOI)) {
			printk("dmcpi interrupt seen\n");
		}
		

		/* handle core irqs */
		for (j=0; j<MGA_MAX_IRQ; j++) {
			if ((status & (1<<j)) & (mga->enabled_core_irq & (1<<j))) {
				dprintk( "handle core irq 0x%x\n", 1<<j );
				for ( k=0; k<MGA_MAX_ISR; k++ ) {
					if (mga->core_irq_handler[j][k])
						(*mga->core_irq_handler[j][k])(mga, regs);
				}
				/* clear it, except when reserved bit */
				mga_write(mga, ICLEAR, (1<<j) & 0x0fa5 );
			}
		}
		
		if (status & (1<<MGA_CORE_IRQ_EXTERNAL)) {
			/* for now, disable external ints, since they
			won't shut up properly yet. (both i33 and i34 would
			need exec_commands to do that, but that doesn't
			work from within isr's yet */
			mga_write(mga, IEN,
				(mga->enabled_core_irq & ~(1<<MGA_CORE_IRQ_EXTERNAL))
				| mga->saved_core_irq );
			
			printk( "external int\n" );
		}
		
	}
	      
	spin_unlock(&mga->irq_lock);
}

/**
 *	mga_request_irq - install an interrupt handler for PCI IRQ
 *	@mga:		mga instance
 *	@handler:	interrupt handler function
 *
 *	This function requests the PCI IRQ used by the card from
 *	the kernel and installs an interrupt handler to manage
 *	all of the various IRQ events that may occur.
 *
 *	All card interrupts are disabled BEFORE the request so that
 *	no spurious interrupts will occur before any of the
 *	sub-handlers have been initialized.
 *	
 *	If another handler is already installed, this function will
 *	NOT overwrite it. The other handler must be released before.
 *
 *	IRQs:
 *	%MGA_IRQ_VINVSYNC
 *
 *	%MGA_IRQ_DCMPEOI
 *
 *	%MGA_IRQ_DCMPEOI
 *
 *	%MGA_IRQ_EXTERNAL
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_release_irq()
 *			mga_enable_irq()
 *			mga_disable_irq()
 */
static int 
mga_request_irq(struct mga_dev_s *mga)
{
	unsigned long flags;
	u32 vien_chk;

	spin_lock_irqsave(&mga->irq_lock, flags);

	if (!mga->irq_inuse)
	{
		/* see G200 spec, Table 4-8 on p 4-115
		   clear and disable the 4 MGA video IRQ sources. this may
		   be overkill but better too much caution than not enough.
		   -inkling */
		mga_write(mga, VIEN, 0);
		mga_write(mga, VICLEAR, MGA_IRQ_ALLMASK);

		/* all MGA Video interrupts should be disabled now. we
		   need to check them and if they are not disabled,
		   return an error for total abort of process
		   -inkling */
		
		vien_chk = mga_read(mga, VIEN) & MGA_IRQ_ALLMASK; /* mask reserved bits */

		if (vien_chk)
		{
			spin_unlock_irqrestore(&mga->irq_lock, flags);
			printk("mgacore: PCI IRQ request could not clear VIEN (0x0%1X)\n",
				vien_chk);
			return -ENODEV;
		}

		mga->saved_core_irq = mga_read(mga, IEN) & 0x0fe5;
		mga->enabled_core_irq = 0;

		if (request_irq(mga->dev->irq, mga_interrupt_handler,
				SA_INTERRUPT | SA_SHIRQ, "mgacore", mga) < 0)
		{
			spin_unlock_irqrestore(&mga->irq_lock, flags);

			printk("mgacore: denied PCI IRQ %d\n",  mga->dev->irq);
			return -ENODEV;
		}
		mga->irq_inuse = 1; /* set the IRQ In Use counter to 1 */
		printk("mgacore: PCI IRQ %d handler registered\n",
			mga->dev->irq);
	} else {
		printk("mgacore: PCI IRQ %d handler is already registered\n",
			mga->dev->irq);
	}
	spin_unlock_irqrestore(&mga->irq_lock, flags);

	return 0;
}

/**
 *	mga_release_irq - release a PCI IRQ interrupt handler
 *	@mga:	mga instance
 *
 *	This function releases the PCI IRQ handler used by the card
 *	after disabling all enabled interrupt sources on the card.
 *	Any associated sub-handlers are indicated upon release.
 *
 *	See also:
 *			mga_request_irq()
 *			mga_enable_irq()
 *			mga_disable_irq()
 *
 *	FIXME: all the sub-handler pointers should also be set to NULL.
 * 		it's not really a crucial step, but more a neatness thing
 *		-inkling
 */
static int 
mga_release_irq(struct mga_dev_s *mga)
{
	unsigned long flags;
	u32 vien_chk;

	spin_lock_irqsave(&mga->irq_lock, flags);

	if (mga->irq_inuse)
	{
		mgacore_printk("mgacore: releasing PCI IRQ %d handler\n",
			mga->dev->irq);

		if (mga->irq_inuse > 1)
		{
			printk("mgacore: %d IRQ sub-handlers will be affected!\n",
				mga->irq_inuse - 1);
		}
		/* see G200 spec, Table 4-8 on p 4-115
		   clear and disable the 4 MGA video IRQ sources. this may
		   be overkill but better too much caution than not enough.
		   -inkling */
		mga_write(mga, VIEN, 0);
		mga_write(mga, VICLEAR, MGA_IRQ_ALLMASK);

		/* all MGA Video interrupts should be disabled now. now we
		   need to check them and if they are not disabled,
		   return an error
		   -inkling */
		
		vien_chk = mga_read(mga, VIEN) & MGA_IRQ_ALLMASK; /* mask reserved bits */

		if (vien_chk)
		{
			spin_unlock_irqrestore(&mga->irq_lock, flags);
			printk("mgacore: PCI IRQ %d release could not clear VIEN (0x0%1X)\n",
				mga->dev->irq, vien_chk);
			return -ENODEV;
		}
		
		/* restore original IEN value */
		mga_write(mga, IEN, mga->saved_core_irq );


		free_irq(mga->dev->irq, mga);
		mgacore_printk("mgacore: PCI IRQ %d released\n", mga->dev->irq);
	} else {
		printk("mgacore: no registered PCI IRQ handler to release!\n");
	}			

	spin_unlock_irqrestore(&mga->irq_lock, flags);

	return 0;
}

/**
 *	mga_enable_irq - enable an interrupt source to use a sub-handler
 *	@mga:		mga instance
 *	@irq:		IRQ to enable
 *
 *	This function enables the hardware to generate interrupts from
 *	the specified source.
 *
 *	This will NOT allow you to replace a sub-handler.  You must disable
 *	it first if you want to replace it with a new sub-handler.
 *
 *	FIXME: We may want to modify this to allow sub-handler replacement.
 *	-inkling	
 *
 *	IRQs:
 *	%MGA_IRQ_VINVSYNC
 *
 *	%MGA_IRQ_BUFLVL
 *
 *	%MGA_IRQ_DCMPEOI
 *
 *	%MGA_IRQ_EXTERNAL
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_disable_irq()
 *			mga_request_irq()
 *			mga_release_irq()
 */
int mga_enable_irq(struct mga_dev_s *mga, int irq, mga_interrupt_handler_t sub_handler)
{
	unsigned long flags;
 	u32 vien_chk;
	int	i;
	
	if (irq < 0 || irq >= MGA_MAX_VIRQ) {
		dprintk( "invalid irq %i\n", irq );
		return 1;
	}

	spin_lock_irqsave(&mga->irq_lock, flags);

 
	/* next we assign the sub-handler pointer and increment in_use count.
	   no change and no error (except a klog msg) if a sub-handler
	   is already registered and try to re-enable the MGA Video IRQ.
	   -inkling */
	for(i=0; i < MGA_MAX_ISR; i++) {
		if (mga->irq_handler[irq][i] == 0)
			break;
	}

	if (i < MGA_MAX_ISR) {
		mga->irq_handler[irq][i] = sub_handler;
		mga->enabled_irq |= (1<<irq) & MGA_IRQ_ALLMASK;
		mga->irq_inuse++;
	} else {
		printk("mgacore: sub-handler for VIRQ 0x%1X already exists!\n",
			irq);
	}

	/* last step is to clear again any pending video irq (cautious)
	   and enable the interrupt to use the handler
	   -inkling */
	if ( i == 0 )
		mga_write(mga, VICLEAR, (1<<irq));
	mga_write(mga, VIEN, mga->enabled_irq );

	/* check to see if it actually did enable the interrupt.
	   return failure code if it did NOT (re)enable the interrupt.
	   -inkling */
	vien_chk = mga_read(mga, VIEN);

	if ( vien_chk != mga->enabled_irq ) {
		dprintk("mgacore: VIRQs %x enabled, but wanted %x\n", vien_chk, mga->enabled_irq );
	}

	spin_unlock_irqrestore(&mga->irq_lock, flags);

	return 0;
}


void
mga_check_irqs( struct mga_dev_s *mga )
{
	unsigned long flags;
 	u32 vien_chk;
	u32 vien;

	spin_lock_irqsave(&mga->irq_lock, flags);

	vien = mga->enabled_irq;
	mga_write(mga, VIEN, vien );
	
	vien_chk = mga_read(mga, VIEN) & MGA_IRQ_ALLMASK;

	spin_unlock_irqrestore(&mga->irq_lock, flags);
	
	if ( vien_chk != vien ) {
		dprintk("mgacore: check_irqs %x enabled, wanted %x\n", vien_chk, vien );
	}
}


/**
 *	mga_disable_irq - disable an interrupt source
 *	@mga:	mga instance
 *	@irq:	IRQ to disable
 *
 *	This function disables the generation of hardware interrupts from
 *	the specified source.
 *
 *	The associated interrupt handler IS uninstalled by this
 *	function.
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_enable_irq()
 *			mga_request_irq()
 *			mga_release_irq()
 */
int mga_disable_irq(struct mga_dev_s *mga, int irq, mga_interrupt_handler_t sub_handler)
{
	unsigned long flags;
	int i;
	u32 vien;
	
	if (irq < 0 || irq >= MGA_MAX_VIRQ) {
		dprintk( "invalid irq %i\n", irq );
		return 1;
	}

	spin_lock_irqsave(&mga->irq_lock, flags);

	for(i=0; i < MGA_MAX_ISR; i++) {
		if (mga->irq_handler[irq][i] == sub_handler)
			break;
	}

	if (i < MGA_MAX_ISR) {
		mga->irq_handler[irq][i] = NULL;
		mga->irq_inuse--;
		for(i=0; i < MGA_MAX_ISR; i++) {
			if (mga->irq_handler[irq][i])
				break;
		}
		if (i == MGA_MAX_ISR) {
			mga->enabled_irq &= ~(1<<irq);
			mga_write(mga, VIEN, mga->enabled_irq );
			mga_write(mga, VICLEAR, (1<<irq));
			vien = mga_read(mga, VIEN);
			if (vien & (1<<irq)) {
				spin_unlock_irqrestore(&mga->irq_lock, flags);
				printk("mgacore: disable VIRQ is unable to clear VIRQ %d\n",
					irq);
				return -ENODEV;
			}		
		}
	} else {
		printk("mgacore: no sub-handler for VIRQ 0x%1X to remove!\n",
			irq);
	}

	spin_unlock_irqrestore(&mga->irq_lock, flags);
	return 0;
}



int mga_enable_core_irq(struct mga_dev_s *mga, int irq, mga_interrupt_handler_t sub_handler)
{
	unsigned long flags;
	int	i;
	
	if ( irq<0 || irq >= MGA_MAX_IRQ ) {
		dprintk( "invalid irq %i\n", irq );
		return 1;
	}

	spin_lock_irqsave(&mga->irq_lock, flags);

	for(i=0; i < MGA_MAX_ISR; i++) {
		if (mga->core_irq_handler[irq][i] == 0)
			break;
	}

	if (i < MGA_MAX_ISR) {
		mga->core_irq_handler[irq][i] = sub_handler;
		mga->enabled_core_irq |= 1 << irq;
		mga->irq_inuse++;
	} else {
		printk("mgacore: sub-handler for IRQ 0x%1X already exists!\n",
			irq);
	}

	if ( i == 0 )
		mga_write(mga, ICLEAR, (1<<irq) & 0x0fa5 );
	mga_write(mga, IEN, mga->enabled_core_irq | mga->saved_core_irq );

	spin_unlock_irqrestore(&mga->irq_lock, flags);

	return 0;
}


int mga_disable_core_irq(struct mga_dev_s *mga, int irq, mga_interrupt_handler_t sub_handler)
{
	unsigned long flags;
	int i;
	
	if ( irq<0 || irq >= MGA_MAX_IRQ ) {
		dprintk( "invalid irq %i\n", irq );
		return 1;
	}

	spin_lock_irqsave(&mga->irq_lock, flags);

	for(i=0; i < MGA_MAX_ISR; i++) {
		if (mga->core_irq_handler[irq][i] == sub_handler)
			break;
	}

	if (i < MGA_MAX_ISR) {
		mga->core_irq_handler[irq][i] = NULL;
		mga->irq_inuse--;
		for(i=0; i < MGA_MAX_ISR; i++) {
			if (mga->core_irq_handler[irq][i])
				break;
		}
		if (i == MGA_MAX_ISR) {
			mga->enabled_core_irq &= ~(1<<irq);
			mga_write(mga, IEN, mga->enabled_core_irq | mga->saved_core_irq);
			mga_write(mga, ICLEAR, (1<<irq) & 0x0fa5 );
		}
	} else {
		printk("mgacore: no sub-handler for IRQ 0x%1X to remove!\n",
			irq);
	}

	spin_unlock_irqrestore(&mga->irq_lock, flags);
	return 0;
}



/* * PRIVATE FUNCTION
 *      mga_test_lines - print I/O lines status
 *      @mga:     mga instance
 *      @details: if non null print more details
 *
 *      Debugging function printing the state of the general I/O lines.
 */
void mga_test_lines(struct mga_dev_s *mga, int details)
{
	u8 ddcoe, miscoe;
	u8 ddcdata, miscdata;
	u8 xgenioctrl, xgeniodata;

	xgenioctrl = mga_indirect_readb(mga, XGENIOCTRL);
	ddcoe  = xgenioctrl & 0x0F;
	miscoe = (xgenioctrl & 0x70)>>4;
	xgeniodata = mga_indirect_readb(mga, XGENIODATA);
	ddcdata  = xgeniodata & 0x0F;
	miscdata = (xgeniodata & 0x70)>>4;

	if (details) {
		printk("mgacore: I/O lines summary\n");

		printk("mgacore: DDC<0> %s %d\n", (ddcoe & 1) ? "->" : "<-", ddcdata & 1);
		printk("mgacore: DDC<1> %s %d\n", (ddcoe & 2) ? "->" : "<-", (ddcdata & 2)>>1);
		printk("mgacore: DDC<2> %s %d\n", (ddcoe & 4) ? "->" : "<-", (ddcdata & 4)>>2);
		printk("mgacore: DDC<3> %s %d\n", (ddcoe & 8) ? "->" : "<-", (ddcdata & 8)>>3);

		printk("mgacore: MISC<0> %s %d\n", (miscoe & 1) ? "->" : "<-", miscdata & 1);
		printk("mgacore: MISC<1> %s %d\n", (miscoe & 2) ? "->" : "<-", (miscdata & 2)>>1);
		printk("mgacore: MISC<2> %s %d\n", (miscoe & 4) ? "->" : "<-", (miscdata & 4)>>2);
	} else {
		printk("mgacore: DDC: %d%d%d%d MISC: %d%d%d\n",
		       ddcdata & 1, (ddcdata & 2)>>1, (ddcdata & 4)>>2, (ddcdata & 8)>>3,
		       miscdata & 1, (miscdata & 2)>>1, (miscdata & 4)>>2);
	}
}



/**
 *	mga_display_pitch - returns bytes per line
 *	@mga:	mga instance
*/
int
mga_display_pitch(struct mga_dev_s *mga)
{
	int tmp;
	int offs;
	int pitch;
	int inter;
	unsigned long flags;

	spin_lock_irqsave( &mga->reg_lock, flags );

	mga_writeb(mga, CRTCADD, 0x13);		/* offset */
	offs = mga_readb( mga, CRTCREG );	/* offset 7:0 */

	mga_writeb(mga, CRTCEXTADD, 0x00);	/* addr gen extensions */
	tmp = mga_readb( mga, CRTCEXTREG);
	offs |= (tmp&0x30)<<4;			/* offset 9:8 */

	mga_writeb(mga, CRTCADD, 0x14);		/*underline loc */
	tmp = mga_readb( mga, CRTCREG );
	if ( tmp & 0x40 ) {			/* dword<6>*/
		pitch = offs * 32;		/*fixme, probably wrong*/
	}
	else {
		mga_writeb(mga, CRTCADD, 0x17);	/* crtc mode control */
		tmp = mga_readb( mga, CRTCREG );
		if (tmp & 0x40) {		/* wbmode<6> */
			pitch = offs * 8;
		} else {
			pitch = offs * 16;	/*fixme, probably wrong*/
		}
	}

	mga_writeb(mga, CRTCEXTADD, 0x00);
        tmp  = mga_readb(mga, CRTCEXTREG);
	inter = (tmp & 0x80) != 0;
	if (!inter) /* not interlaced */
		pitch *= 2;

	spin_unlock_irqrestore( &mga->reg_lock, flags );
	
	dprintk("display pitch %d (offs %d, i%d)\n", pitch, offs, inter);
	
	return pitch;
}


void
mga_display_size(struct mga_dev_s *mga, int *width, int *height, int *depth)
{
	int	vtotal;
	int	htotal;
	int	dtotal;
	int	tmp;
	unsigned long flags;

	spin_lock_irqsave( &mga->reg_lock, flags );

	/* Read HDisplay form hardware (width >> 3) -1 */
	mga_writeb(mga, CRTCADD, 0x1);
	htotal  = mga_readb(mga, CRTCREG) + 1;

	mga_writeb(mga, CRTCEXTADD, 0x01);
        tmp  = mga_readb(mga, CRTCEXTREG);
	htotal += (tmp & 0x02) ? 0x100 : 0;
	htotal <<= 3;

	mga_writeb(mga, CRTCADD, 0x12);
        vtotal  = mga_readb(mga, CRTCREG) + 1;
	mga_writeb(mga, CRTCADD, 0x07);
        tmp  = mga_readb(mga, CRTCREG);

	vtotal += (tmp & 0x02) ? 256 : 0;
	vtotal += (tmp & 0x40) ? 512 : 0;
	mga_writeb(mga, CRTCEXTADD, 0x02);
        tmp  = mga_readb(mga, CRTCEXTREG);
	vtotal += (tmp & 0x04) ? 1024 : 0;

	mga_writeb(mga, CRTCEXTADD, 0x00);
        tmp  = mga_readb(mga, CRTCEXTREG);
	if (tmp & 0x80) /* interlaced */
		vtotal *= 2;

	switch (mga_indirect_readb(mga, XMULCTRL)) {
	case BPP_8:
		dtotal = 8;
		break;
	case BPP_15:
		dtotal = 15;
		break;
	case BPP_16:
		dtotal = 16;
		break;
	case BPP_24:
		dtotal = 24;
		break;
	case BPP_32_DIR:
	case BPP_32_PAL:
		dtotal = 32;
		break;
	default:
		dtotal = 0;
	}
	spin_unlock_irqrestore( &mga->reg_lock, flags );

	dprintk("display %d x %d depth %d\n", htotal, vtotal, dtotal);

	*width = htotal;
	*height = vtotal;
	*depth = dtotal;
}


#define MMALLOC_MAGIC1	((u32)0xfeedcafe)
#define MMALLOC_MAGIC2	((u32)0xdeafbabe)
#define MMA_PUT_U32( addr, u ) writel((u), (u8*)mga->fb + (addr))
#define MMA_GET_U32( addr, u ) *(u) = readl((u8*)mga->fb + (addr))

/**
 *	mga_alloc_buffer - 	allocate a buffer
 *	@mga:	mga instance
 *	@size:	size of the buffer to allocate
 *
 *	This is a really simple memory manager for the framebuffer.
 *	It justs decrements a pointer from the end of the framebuffer
 *	and DOES NOT really manage an array of allocated blocks.
 *	
 *	An allocated block must be deallocated in the SAME ORDER (with
 *	respect to other modules requesting memory) for this to work.
 *
 *	Return value:	returns a pointer to the allocated buffer
 *
 *	See also:	mga_free_buffer()
 *
 *	Notes:	The memory address is expressed as a value *inside*
 *	the framebuffer address space. For example a buffer at address
 *	0x0 represents the beginning of the video framebuffer.
 *
 *
 *	Allocation is now slightly safer. Blocklayout:
 *	{
 *		u32 MAGIC1
 *		u32 requested_size
 *		u32 ptr_to_end_of_block
 *		u8  aligned_user_block[ size ]
 *		u8  alignment_filler[n]
 *		u32 MAGIC2
 *	}
 *		--lvs
 */
u32 mga_alloc_buffer(struct mga_dev_s *mga, unsigned long size, unsigned long alignment )
{
	unsigned long bsize = size + alignment + 4*sizeof(u32);
	u32 rv;
	u32 saved_limit;
	unsigned long flags;

	spin_lock_irqsave( &mga->malloc_lock, flags );

	if (mga->malloc_limit < bsize) {
		spin_unlock_irqrestore( &mga->malloc_lock, flags );
		return 0;
	}

	mga->malloc_limit -= sizeof(u32);
	MMA_PUT_U32( mga->malloc_limit, MMALLOC_MAGIC2 );

	saved_limit = mga->malloc_limit;
	
	mga->malloc_limit -= size;
	mga->malloc_limit &= -alignment;
	rv = mga->malloc_limit;
	
	mga->malloc_limit -= sizeof(u32);
	MMA_PUT_U32( mga->malloc_limit, saved_limit );
	
	mga->malloc_limit -= sizeof(u32);
	MMA_PUT_U32( mga->malloc_limit, size );
	
	mga->malloc_limit -= sizeof(u32);
	MMA_PUT_U32( mga->malloc_limit, MMALLOC_MAGIC1 );

	memset_io((u8*)mga->fb + rv, 0xeb, size );
	
	mgacore_printk("mgacore: alloc %ld (%#lx) at %08x\n", size, size, rv);

	spin_unlock_irqrestore( &mga->malloc_lock, flags );

	return rv;
}

/**
 *	mga_free_buffer - 	deallocate a buffer
 *	@mga:	mga instance
 *	@size:	size of the block to allocate
 *
 *	Increments the allocation counter, to simulate memory
 *	deallocation.
 *
 *	An allocated block must be deallocated in the SAME ORDER (with
 *	respect to other modules requesting memory) for this to work.
 *
 *	See also:	mga_alloc_buffer()
 */
void mga_free_buffer(struct mga_dev_s *mga, unsigned long size)
{
	u32 u;
	unsigned long flags;

	spin_lock_irqsave( &mga->malloc_lock, flags );
	
	MMA_GET_U32( mga->malloc_limit, &u );
	if ( u != MMALLOC_MAGIC1 ) {
		dprintk( "mgacore: block corruption at begin\n" );
		spin_unlock_irqrestore( &mga->malloc_lock, flags );
		return;
	}
	mga->malloc_limit += sizeof(u32);

	MMA_GET_U32( mga->malloc_limit, &u );
	if ( u != size ) {
		dprintk( "mgacore: not freeing own block\n" );
		// not fatal; alloc/free order mixed up
	}
	mga->malloc_limit += sizeof(u32);
	
	MMA_GET_U32( mga->malloc_limit, &u );
	mga->malloc_limit = u;
	
	MMA_GET_U32( mga->malloc_limit, &u );
	if ( u != MMALLOC_MAGIC2 ) {
		dprintk( "mgacore: block corruption at end\n" );
		//maybe we get lucky
	}
	mga->malloc_limit += sizeof(u32);

	printk("mgacore: free %ld (%#lx)\n", size, size);

	spin_unlock_irqrestore( &mga->malloc_lock, flags );
}


static void
mga_malloc_init( struct mga_dev_s *mga )
{
	spin_lock_init(&mga->malloc_lock);

	mga->malloc_limit = mga->memory_size;

	mga->malloc_limit -= sizeof(u32);
	MMA_PUT_U32( mga->malloc_limit, MMALLOC_MAGIC2 );
}


static void
mga_malloc_check( struct mga_dev_s *mga )
{
	u32 u;
	u32 limit;
	unsigned long flags;

	spin_lock_irqsave( &mga->malloc_lock, flags );

	limit = mga->malloc_limit;

	while ( 1==1 ) {	
		MMA_GET_U32( limit, &u );
		limit += sizeof(u32);
		if ( u == MMALLOC_MAGIC2 )
			break;
		if ( u == MMALLOC_MAGIC1 ) {
			u32 size;
			
			MMA_GET_U32( limit, &size );
			limit += sizeof(u32);
		
			dprintk( "mgacore: leaked block 0x%x\n", size );

	dprintk( "mgacore: old limit: %x\n", limit );
			MMA_GET_U32( limit, &limit );
			
	dprintk( "mgacore: new limit: %x\n", limit );
			
			MMA_GET_U32( limit, &u );
			limit += sizeof(u32);
			
			if ( u != MMALLOC_MAGIC2 ) {
				dprintk( "mgacore: block corruption at end\n" );
				break;
			}
		}
		else {
			dprintk( "mgacore: malloc corruption\n" );
			break;
		}
	}

	spin_unlock_irqrestore( &mga->malloc_lock, flags );
}


/* * PRIVATE FUNCTION *
 *	mga_guess_memory_size -	detect frambuffer size
 *	@mga:   mga device structure
 *
 *	This function queries the graphic card to detect the size of
 *	the framebuffer.
 *
 *	Return value:
 *	Memory size
 *
 *	Notes:
 * 	This memory sizer is from the Matrox framebuffer driver
 *	developed by Petr Vandrovec <vandrove@vc.cvut.cz>,
 *	et. al. Thanks for porting the code Steve. 
 */
static unsigned int mga_guess_memory_size(struct mga_dev_s *mga)
{
	unsigned int offs;
        unsigned char store;
        unsigned char tmp;
	unsigned int maxSize = 0x2000000; /* 32 meg is the largest available */
	unsigned char crtcext3; /* save contents of mga reg */
	unsigned int detected;

#if LINUX_VERSION_CODE >= 0x020300
	mga->ctrl = ioremap_nocache(mga->dev->resource[1].start, 0x4000);
	maxSize = mga->dev->resource[0].end - mga->dev->resource[0].start + 1;
	mga->fb = ioremap(mga->dev->resource[0].start, maxSize);
#else
	mga->ctrl = ioremap_nocache(mga->dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK,0x4000);
	mga->fb = ioremap(mga->dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK, maxSize);
#endif


        maxSize &= ~0x1FFFFF;   /* must be X*2MB (really it must be 2 or X*4MB) */
        /* at least 2MB */
        if (maxSize < 0x0200000) return 0;
        if (maxSize > 0x2000000) maxSize = 0x2000000;

	writeb(0x03, mga->ctrl + CRTCEXTADD);
	crtcext3 = readb(mga->ctrl + CRTCEXTREG);               /* save the reg */
	writeb((crtcext3 | 0x80), mga->ctrl + CRTCEXTREG);      /* turn on power graphics mode */
	store = readb(mga->fb + 0x1234);

        for (offs = 0x100000; offs < maxSize; offs += 0x200000) {
                tmp = readb(mga->fb + offs);

		writeb(0x02, mga->fb + offs);
		writeb(0x00, mga->ctrl + CACHEFLUSH);

                if (readb(mga->fb + offs) != 0x02)
                        break;

                writeb(readb(mga->fb + offs) - 0x02, mga->fb + offs);

                if (readb(mga->fb + offs))
                        break;

		writeb(tmp, mga->fb + offs);
	}

        writeb(store, mga->fb + 0x1234);
	writeb(0x03, mga->ctrl + CRTCEXTADD);
	writeb(crtcext3, mga->ctrl + CRTCEXTREG);       /* restore prior reg value */

	detected = offs - 0x100000;
	printk("mgacore: %d MB detected\n", detected / 0x100000);

	iounmap(mga->ctrl);
	iounmap(mga->fb);

        return detected;
}
               
struct proc_dir_entry	*
mga_create_proc_entry(struct mga_dev_s *mga, char *name,
	read_proc_t *rp, write_proc_t *wp, void *data)
{
	struct proc_dir_entry	*entry;

	if ((entry = create_proc_entry(name, S_IFREG | S_IRUGO | S_IWUSR,
		mga->proc_directory)) == NULL) {
		return NULL;
	}

	entry->data = data;
	entry->read_proc = rp;
	entry->write_proc = wp;

	return entry;
}

void mga_remove_proc_entry(struct mga_dev_s *mga, char *name)
{
	if (mga_procdir) {
		if (mga->proc_directory) {
			remove_proc_entry(name, mga->proc_directory);
		}
	}
}

void *mga_rvmalloc(signed long size, unsigned long page_table[])
{
	void	*mem;
	unsigned long adr;
	unsigned long page;

#if LINUX_VERSION_CODE >= 0x020300
	mem = vmalloc_32(size);
#else
	mem = vmalloc(size);
#endif
	if (mem == 0)
		return mem;

	memset(mem, 0, size);	/* clear the ram out, no junk to user space */
	adr = (unsigned long)mem;
	while(size > 0) {
		page = kvirt_to_pa(adr);
		if (page_table) {
			*page_table++ = page;
		}
		mem_map_reserve(MAP_NR(__va(page)));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}



void mga_rvfree(void *mem, signed long size, unsigned long page_table[])
{
	unsigned long adr;
	unsigned long page;

	if (mem == 0)
		return ;


	adr = (unsigned long)mem;
	while(size > 0) {
		if (page_table) {
			page = *page_table++;
		} else {
			page = kvirt_to_pa(adr);
		}
		mem_map_unreserve(MAP_NR(__va(page)));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}


int
mga_i2c_control( struct mga_dev_s *mga, int driver_id, unsigned int cmd, void *arg )
{
	struct i2c_client *cl;

//	cl = i2c_get_client( driver_id, mga->i2c->id, NULL );	kernel modules like tuner.o don't allow client access
	switch (driver_id) {	
	case I2C_DRIVERID_MSP3400:
		cl = mga->msp_client;
		break;
	case I2C_DRIVERID_TUNER:
		cl = mga->tuner_client;
		break;
	case I2C_DRIVERID_KS0127:
		cl = mga->ks_client;
		break;
	case I2C_DRIVERID_MGATVO:
		cl = mga->maven_client;
		break;
	default:
		cl = NULL;
		break;
	}
	if (cl == NULL) {
		dprintk( "mga_i2c_ioctl: no driver 0x%x\n", driver_id );
		return -1;
	}
	if ((cl->driver == NULL) || (cl->driver->command == NULL)) {
		dprintk( "mga_i2c_ioctl: no ioctl for driver 0x%x\n", driver_id );
		return -1;
	}
	return cl->driver->command(cl, cmd, arg);
}


void
mga_set_audioinput(struct mga_dev_s *mga, int channel)
{
	if (mga->options & MGAVIDEO_HAS_AUDIO) {
		mga_i2c_control(mga,
			I2C_DRIVERID_MSP3400, 
			AUDC_SET_INPUT, 
			&channel);
	}
	mga->msp_input = channel;
}


int
mga_set_defaults(struct mga_dev_s *mga)
{
	mga_set_audioinput(mga, 4);
	return 0;
}


/**********************************************/


#define OUTSEL_MONITOR		0
#define OUTSEL_RESERVED1	1
#define OUTSEL_BYPASS		2
#define OUTSEL_RESERVED2	3
#define OUTSEL_2NDRGB24		4	//G400 only
#define OUTSEL_RESERVED3	5	//G400 only
#define OUTSEL_2ND656		6	//G400 only

#define FUNCSEL_RESERVED	0
#define FUNCSEL_MAFC		1
#define FUNCSEL_PANELLINK	2
#define FUNCSEL_DISABLE		3

static void
set_mafc_sel(struct mga_dev_s *mga, int outsel, int funcsel)
{
	mga_indirect_and_or_mt(mga, XMISCCTRL, 0x19, ((outsel&3)<<5) | ((funcsel&3)<<1) );
}


#define VDOINPUT_VIP	0
#define VDOINPUT_CRTC	1
#define VDOINPUT_CRTC2	2	//G400 only
#define VDOINPUT_COUNT	3



static int
vdo_set_dir( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data )
{
	return dir == BUSDIR_SOURCE ? 0 : 1;
}


static enum bus_direction
vdo_get_dir( struct mgabus_s *bus, void *data )
{
	return BUSDIR_SOURCE;
}


static int
vdo_get_format( struct mgabus_s *bus, int input,
			struct mgabus_video_input_s *params, void *data )
{
	struct mga_dev_s *mga = (struct mga_dev_s*)data;

	if (input == VDOINPUT_VIP) {
		return mgabus_get_format( mga->vip_bus, mgabus_get_input(mga->vip_bus), params );
	}
	else if (input == VDOINPUT_CRTC) {
		/* fixme: make up a format when crtc is source */
		return -1;
	}
	else if (input == VDOINPUT_CRTC2) {
		/* fixme: make up a format when crtc2 is source */
		return -1;
	}
	return 1;
}




/* attempts to switch automatically between different VDO inputs
   based on what's available on the VIP bus */
static void
update_vdo( struct mga_dev_s *mga )
{
	int src = mgabus_get_source_addr( mga->vip_bus );
//	int input = mgabus_get_input( mga->vip_bus );

	/* fixme: need cleaner test to decide whether to output vip or crtc */
	
	if (	((src == 0x0127) /*&& (input == KS_INPUT_YUV656)*/) ||
		(src == 0x0002) ||
		(src == 0x0034) ) {
		
		set_mafc_sel(mga, OUTSEL_BYPASS, FUNCSEL_MAFC); /* clock VIN directly to VDO */
		mga->vdo_port->active_input = VDOINPUT_VIP;
		return;
		
	} else if (/*want CRTC on vdo output*/1) {
		int dotclk_kHz;
		if (mga_get_clk(mga, CLK_PIXEL, &dotclk_kHz)==0) {
			/* MAFC vdo max dot: 65MHz */
			if (dotclk_kHz < 65*1000) {
				set_mafc_sel(mga, OUTSEL_MONITOR, FUNCSEL_MAFC);
				mga->vdo_port->active_input = VDOINPUT_CRTC;
				return;
			}
		}
	} else if (/*want CRTC2 on vdo output*/0) {
		if (1) {
			/*want 656*/
			set_mafc_sel(mga, OUTSEL_2ND656, FUNCSEL_MAFC);
		}
		else {
			/*want rgb*/
			set_mafc_sel(mga, OUTSEL_2NDRGB24, FUNCSEL_MAFC);
		}
		mga->vdo_port->active_input = VDOINPUT_CRTC2;
		return;
	}
	set_mafc_sel(mga, OUTSEL_BYPASS, FUNCSEL_DISABLE);
	mga->vdo_port->active_input = -1;
}


static void
vip_notify( struct mgabus_s *bus, enum bus_notify_code code, void *data )
{
	struct mga_dev_s *mga = (struct mga_dev_s*)data;

	update_vdo(mga);
	
	if (code == BUSCODE_INPUT) {
		mgabus_notify(mga->vdo_bus, BUSCODE_INPUT);
	}
	else if (code == BUSCODE_FORMAT) {
		if (mga->vdo_port->active_input == VDOINPUT_VIP)
			mgabus_notify(mga->vdo_bus, BUSCODE_FORMAT);
	}
}



static int
tvo_set_dir( struct mgabus_s *bus, enum bus_direction dir, int addr, void *data )
{
	return dir == BUSDIR_SINK ? 0 : 1;
}


static enum bus_direction
tvo_get_dir( struct mgabus_s *bus, void *data )
{
	return BUSDIR_SINK;
}


static void
tvo_notify( struct mgabus_s *bus, enum bus_notify_code code, void *data )
{
	struct mga_dev_s *mga = (struct mga_dev_s*)data;
	struct mgabus_video_input_s params;
	int std;

	if (mgabus_get_format( mga->vdo_bus, mgabus_get_input(mga->vdo_bus), &params )) {
		/* fixme: disable maven? */
		return;
	}

	switch(params.norm) {

	case KS_STD_PAL:
		std = MAVEN_OUTPUT_PAL;
		break;

	case KS_STD_NTSC:
		std = MAVEN_OUTPUT_NTSC;
		break;

	case KS_STD_SECAM:
		std = MAVEN_OUTPUT_PAL;
		break;

	default:
		std = MAVEN_OUTPUT_NTSC;
	}

	if ( mga->maven_std != std ) {
		mga->maven_std = std;
		
		mga_i2c_control(mga, I2C_DRIVERID_VIDEOENCODER, 
			MAVEN_SET_STD, &std);
	}
}



static int 
mga_init_device(struct mga_dev_s *mga, struct pci_dev *dev, int type,int card)
{
	int	i;
	struct mgabus_s *vbus;

	memset(mga, 0, sizeof(struct mga_dev_s));
  
	mga->dev = dev;
	mga->card_num = card;
	mga->mga_type = type;
	mga->memory_size = mga_guess_memory_size(mga);

	if (mga->memory_size < 1024 * 1024)
		return 1;

#if LINUX_VERSION_CODE >= 0x020300
	mga->ctrl = ioremap_nocache(dev->resource[1].start, 0x4000);
	mga->fb = ioremap(dev->resource[0].start, mga->memory_size);
#else
	mga->ctrl = ioremap_nocache(dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK, 0x4000);
	mga->fb = ioremap(dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK, mga->memory_size);
#endif
	/* need to start alloc memory at end of display card. The memory
	 * size to X11 need to made smaller by amount used by drivers
	 */
	mga_malloc_init(mga);

        /* Create a directory entry for each card  */
	sprintf(mga->name, "mga%d", card);
        mga->proc_directory = create_proc_entry(mga->name, S_IFDIR, mga_procdir);

	spin_lock_init(&mga->irq_lock);
	spin_lock_init(&mga->reg_lock);

	mga->maven_std = -1;
        
	mga_request_irq(mga);
	
	mgabus_admin_attach(mga);

	mgabes_attach(mga);
	mgacodec_attach(mga);

	mga_display_size(mga, &i, &i, &i);
	
	mga_display_pitch(mga);
	
	mga_get_clk( mga, -1, NULL );
	
	mga->bios_mctlwtst = 0xff;
#ifdef NEVER
	if ( mga->mga_type == MGA_G200 ) {
		MGABios2Info biosinfo;
		if ( mga_get_bios2( mga, &biosinfo ) == 0 ) {
			printk( "MCTLWTST %02x\n", biosinfo.MCTLWTST );
			mga->bios_mctlwtst = biosinfo.MCTLWTST;
		}
	}
#endif

	vbus = mgabus_new( mga->bus_admin, sizeof(*vbus), "vip" );
	if ( vbus == NULL ) {
		return 1;
	}
	mga->vip_bus = vbus;
	
	vbus = mgabus_new( mga->bus_admin, sizeof(*vbus), "vdo" );
	if ( vbus == NULL ) {
		return 1;
	}
	mga->vdo_bus = vbus;
	
	/* detect any devices setting on i2c bus */
	mgai2c_init(mga);

	mga->vdo_port = kmalloc(sizeof(*mga->vdo_port), GFP_KERNEL);
	if (mga->vdo_port == NULL)
		return 1; //fixme
	mgabus_init_client( mga->vdo_port, BUSCLIENT_VIDEO, "mga video out", 0x0001, mga );
	mga->vdo_port->super.set_dir = vdo_set_dir;
	mga->vdo_port->super.get_dir = vdo_get_dir;
	mga->vdo_port->get_format = vdo_get_format;
	mga->vdo_port->nr_inputs = VDOINPUT_COUNT;
	mgabus_attach( mga->vdo_bus, &mga->vdo_port->super );

	if (mga->options & MGAVIDEO_HAS_ENCODER) {
		mga->tvo_port = kmalloc(sizeof(*mga->tvo_port), GFP_KERNEL);
		if (mga->tvo_port == NULL)
			return 1; //fixme
		mgabus_init_client( mga->tvo_port, BUSCLIENT_VIDEO, "maven video encoder", 0x0004, mga );
		mga->tvo_port->super.set_dir = tvo_set_dir;
		mga->tvo_port->super.get_dir = tvo_get_dir;
		mgabus_attach( mga->vdo_bus, &mga->tvo_port->super );
		mgabus_add_observer( mga->vdo_bus, tvo_notify, mga );
	}

	/* configure vdo bus */
	mgabus_set_source_addr( mga->vdo_bus, 0x0001, BUS_ADDR_BCAST );

	if (vdo_enable) {
		mgabus_add_observer( mga->vip_bus, vip_notify, mga );
//		update_vdo(mga);
	}
	
	mga_set_defaults(mga);

	return 0;
}


static void
mga_deinit_device( struct mga_dev_s *mga )
{
	if (mga->tvo_port != NULL) {
		mgabus_remove_observer(mga->vdo_bus, tvo_notify);
		mgabus_detach(mga->vdo_bus, &mga->tvo_port->super);
		kfree(mga->tvo_port);
	}

	if (mga->vdo_port != NULL) {
		mgabus_remove_observer(mga->vip_bus, vip_notify);
		mgabus_detach(mga->vdo_bus, &mga->vdo_port->super);
		kfree(mga->vdo_port);
	}

	mgai2c_cleanup(mga);

	if ( mga->vdo_bus != NULL )
		mgabus_free( mga->vdo_bus );
	mga->vdo_bus = NULL;

	if ( mga->vip_bus != NULL )
		mgabus_free( mga->vip_bus );
	mga->vip_bus = NULL;

	if ( mga->codec )
		mgacodec_deattach( mga->codec );
		
	if (mga->bus_admin)
		mgabus_admin_deattach(mga->bus_admin);

	if (mga->irq_inuse)
		mga_release_irq(mga);
		
	if ( mga->ctrl )
		iounmap(mga->ctrl);
		
	if (mga->fb) {
		mga_malloc_check(mga);
		iounmap(mga->fb);
	}

	
	if (mga->ctrl) {
		mgabes_deattach(mga->bes);
		mga_release_card(mga);
	}
	
	if (mga_procdir) {
		if (mga->proc_directory) {
			remove_proc_entry(mga->name, mga_procdir);
		}
	}
}


/* * PRIVATE FUNCTION *
 *	mga_detect_card -	detect a Matrox card
 *
 *	This function tries to detect a Matrox card (G series)
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	Notes:	only detects G200 and G400 at this time. It should be
 *	easy to add other models to the list.
 */
static int mga_detect_cards(void)
{
	struct pci_dev *dev;
	int cidx = 0;
	int i;

        /* Create a directory entry in /proc  (0 is root of proc_fs) */
        mga_procdir = create_proc_entry("mgavideo", S_IFDIR, 0);

	for (i=0; i<sizeof(mga_pci_ids)/sizeof(mga_pci_ids[0]); i++) {
	
		dev = NULL;
		while ( (cidx<MAX_CARDS) &&
			((dev = pci_find_device(PCI_VENDOR_ID_MATROX, mga_pci_ids[ i ].dev_id, dev)) != NULL)
			) {
			printk("mgacore: Found %s\n", mga_pci_ids[ i ].name );
			if (mga_init_device(&card[cidx++], dev, mga_pci_ids[ i ].mga_type, cidx)) {
				memset(&card[--cidx], 0, sizeof(card[cidx]));
			}
		}
	}

	if (cidx == 0) {
		remove_proc_entry("mgavideo", 0);
		printk("mgacore: No supported cards found\n");
		return -EIO;   
	}
	
	return 0;
}

/**
 *	mga_get_card -	return a device structure
 *
 *	This function returns a pointer to a &mga_dev_s device.
 */
struct mga_dev_s *mga_get_card(int num)
{
	if (num >= MAX_CARDS)
		return NULL;

	if (card[num].fb && card[num].ctrl) {
		card[num].in_use++;
		MOD_INC_USE_COUNT;
		return &card[num];
	}

	return NULL;
}

/**
 *	mga_release_card -	reset the usage count of the card
 *	@mga:	mga instance
 *
 *	This function resets the usage counter of the device.
 */
void mga_release_card(struct mga_dev_s *mga)
{
	if (mga) {
		mga->in_use--;
		MOD_DEC_USE_COUNT;
	}
}


int mga_card_lock(struct mga_dev_s *mga)
{
	return test_and_set_bit(0, &mga->card_lock);
}


void mga_card_unlock(struct mga_dev_s *mga)
{
	mga->card_lock = 0;
}


int init_module(void)
{
	printk("Matrox core interface v0.1 (c) David Barth\n");

	return mga_detect_cards();
}

void cleanup_module(void)
{
	int i;

	for (i=0; i<MAX_CARDS; i++) {
		if (card[i].fb && card[i].ctrl) {
			printk("cleanup for card %d\n", i);
			mga_deinit_device(&card[i]);
		}
	}

	remove_proc_entry("mgavideo", 0);
}

