/*
 * This module is an interface to the ZORAN ZR367x0 DVD decoder chip,
 * also known as I34.
 *
 * Copyright (c) 2000-2001 Merrick Johnson <inkling@nop.org>
 * Copyright (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Merrick Johnson <inkling@nop.org> MAINTAINER, send queries here
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/videodev.h>
#include <linux/proc_fs.h>
#include "mgacore.h"
#include "mgacodec.h"
#include "mgacap.h"
#include "i34.h"
#include "mgadvd.h"
#include "audiochip.h"


#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */
static int video_nr = -1;

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif
#if LINUX_VERSION_CODE >= 0x020405
MODULE_PARM(video_nr,"i");
#endif

MODULE_AUTHOR("David Barth <dbarth@besancon.net>");
MODULE_DESCRIPTION("Driver for the Matrox DVD module (ZR36700)");


#define MGA_MAX_WIDTH 720
#define MGA_MIN_WIDTH 8
#define MGA_MAX_HEIGHT 576
#define MGA_MIN_HEIGHT 8 



struct mgadvd_v4l1 {
	struct video_device	video_dev;
	struct mga_dev_s	*mga;
	int			id;
	int			user;
        int			mgacaps;

	/* Video for Linux parameters */
	struct video_picture picture;	/* Current picture params */
	
	int write_mode;
};


/****************************************************************************
 * V4L1 INTERFACE                                                           *
 ****************************************************************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
unsigned int mgadvd_poll(struct video_device *dev, struct file *file, struct poll_table_struct *wait)
#else
unsigned int mgadvd_poll(struct file *file, struct poll_table_struct *wait)
#endif
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;
	
//	dprintk("mgadvd poll\n");
	
	/* if (!&host_mga.transfer) return POLLOUT | POLLWRNORM; */
	/* if (mgacodec_buffer_free(&host_mga) > 0)
	   return mask = POLLOUT | POLLWRNORM;
	*/
	/* poll_wait(file, &host_mga.codec_inq, wait); */
	/* if (mgacodec_buffer_free(&host_mga) > 0)
		mask = POLLOUT | POLLWRNORM;
	*/
	
	return mgacodec_poll(mv->mga->codec, file, wait, 0, 2048 );
}


static int mga_open(struct video_device *dev, int flags)
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;
	int rv;
	
	if ((rv=i34_open(mv->mga)) != 0) {
		return rv;
	}

	rv = mgacap_open( mv->mga->cap_v4l1, flags );
	if (rv != 0) {
		i34_close(mv->mga);
		return rv;
	}

	/* enable audio-out on marvel box */
	mga_set_audioinput(mga, 4);

	dprintk("mgadvd: opening device\n");
	mv->user++;


	MOD_INC_USE_COUNT;
	return 0;
}


static void mga_close(struct video_device *dev)
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;
	
	dprintk("mgadvd: closing device\n");
	
	mgacap_close( mv->mga->cap_v4l1 );

	i34_close(mv->mga);
		
	mv->user--;

	MOD_DEC_USE_COUNT;
}


static long mga__read(struct video_device *dev, char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}


static long mga__write(struct video_device *dev, const char *buf, unsigned long count, int nonblock)
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;

	unsigned long todo = count;
	unsigned int  chunk;
	int ret;
	
//	dprintk("mgacap write %p %lu %i\n", buf, count, nonblock );
	
	if (	(mv->write_mode == VID_WRITE_MPEG_AUD) ||
		(mv->write_mode == VID_WRITE_MPEG_VID) ) {
		
		while (todo > 0) {
			chunk = (todo > CODEC_BUFFER_SIZE/4) ? CODEC_BUFFER_SIZE/4 : todo;
			if ((ret = mgacodec_send_bitstream(mv->mga->codec, buf, chunk)) < 0) {
				return (long)ret;
			}
			buf  += chunk;
			todo -= chunk;
			
		}
		return count;
	}
	else
		return -EINVAL;
}


static int
mga_dvd_ioctl(struct video_device *dev, unsigned int cmd, void *arg, int *handled)
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;
	struct mga_dev_s *mga = mga = mv->mga;
	int ret;

	if ( mv->mga->i34 == NULL ) {
		*handled = 0;
		return 0;
	}
	
	*handled = 1;

	switch (cmd) {

	case VIDIOCSPLAYMODE:
	{
		struct video_play_mode pmode;
		
		dprintk("mgadvd ioctl VIDIOCSPLAYMODE\n");
		
		if (copy_from_user((void *)&pmode, arg,
				   sizeof(struct video_play_mode)))
			return -EFAULT;
		switch (pmode.mode)
		{
		case VID_PLAY_VID_OUT_MODE:
			if ( (pmode.p1 != VIDEO_MODE_NTSC)
			  && (pmode.p1 != VIDEO_MODE_PAL) )
				return -EINVAL;
			if ( i34_video_standard(mga->i34, pmode.p1) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_NORMAL:
			if ( i34_start(mga->i34) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_PAUSE:
			if ( i34_pause(mga->i34) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_RESET:
			if (i34_reset(mv->mga->i34) != 0)
				return -EBUSY;
			break;

		case VID_PLAY_IMMEDIATE_NORMAL:
			if ( i34_continue(mga->i34) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_MASTER_MODE:
			//VID_PLAY_MASTER_NONE
			//VID_PLAY_MASTER_VIDEO
			//VID_PLAY_MASTER_AUDIO
			
			break;
			
		case VID_PLAY_END_MARK:
			if ( i34_stop(mga->i34, 0) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_SINGLE_FRAME:
			if ( i34_step(mga->i34) != 0 )
				return -EBUSY;
			break;

		
		case VID_PLAY_SLOW_MOTION:
			if ( i34_slowmo(mga->i34, pmode.p1) != 0 )
				return -EBUSY;
			break;

		case VID_PLAY_FAST_FORWARD:
			if ( i34_fastfwd(mga->i34) != 0 )
				return -EBUSY;
			break;
		
		case VID_PLAY_FREEZE_FRAME: // pause?
		case VID_PLAY_STILL_MODE:// pause?
		
		case VID_PLAY_GENLOCK:
		case VID_PLAY_SWITCH_CHANNELS:
		case VID_PLAY_ACTIVE_SCANLINES:

		default:
			return -EINVAL;
		}
		return 0;
	}

	case VIDIOCSMICROCODE:
	{
		struct video_code ucode;
		
		dprintk("mgadvd ioctl VIDIOCSMICROCODE\n");

		if ( copy_from_user((void *) &ucode, arg, sizeof(ucode)) != 0 )
			return -EFAULT;
		if (ucode.datasize > 16384) /* Zoran ucode is small... */
			return -EINVAL;
		if ( (ucode.data = vmalloc(ucode.datasize)) == NULL )
			return -ENOMEM;
		if ( copy_from_user((void *) ucode.data, ucode.data,
				    ucode.datasize) != 0 ) {
			vfree(ucode.data);
			return -EFAULT;
		}

		/* return i34_install_microcode(ucode); */
		
		vfree(ucode.data);////i34 is owner?
		return -EINVAL;
	}

	/* then the private ones */

	case I34_IOCSVIDEOSID:
		dprintk("mgadvd ioctl I34_IOCSVIDEOSID\n");
	
		if ( i34_change_video_sid(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSAUDIOSID:
		dprintk("mgadvd ioctl I34_IOCSAUDIOSID\n");
	
		if ( i34_change_audio_sid(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSSUBPICSID:
		dprintk("mgadvd ioctl I34_IOCSSUBPICSID\n");
	
		if ( i34_change_subpic_sid(mga->i34, (int)arg) != 0 )
               		return -EINVAL;
		break;

	case I34_IOCSADPMICROCODE:
		dprintk("mgadvd ioctl I34_IOCSADPMICROCODE\n");
	
		if ( i34_adp_boot(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSDVPMICROCODE:
		dprintk("mgadvd ioctl I34_IOCSDVPMICROCODE\n");
	
		if ( i34_dvp_boot(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSSTREAMTYPE:
		dprintk("mgadvd ioctl I34_IOCSSTREAMTYPE\n");
	
		if ( i34_select_bitstream(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSVIDEOSYNC:
		dprintk("mgadvd ioctl I34_IOCSVIDEOSYNC\n");
	
		if ( i34_video_sync(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSAUDIOSYNC:
		dprintk("mgadvd ioctl I34_IOCSAUDIOSYNC\n");
	
		if ( i34_audio_sync(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCRESYNC:
		dprintk("mgadvd ioctl I34_IOCSRESYNC\n");
		if ( i34_resync(mga->i34, (int)arg) != 0 )
			return -EINVAL;
		break;

	case I34_IOCRESET:
		dprintk("mgadvd ioctl I34_IOCRESET\n");
		if ( i34_reset(mga->i34) != 0 )
			return -EINVAL;
		break;

	case I34_IOCSSPPALETTE:
	{
		u32 sppal[16];

		dprintk("mgadvd ioctl I34_IOCSSPPALETTE\n");

		if ( copy_from_user( &sppal, arg, sizeof(sppal) ) )
		{
			return -EFAULT;
		}

		i34_load_subpicture_palette(mga->i34, sppal);

		break;
	}

/* two different kinds of status reads
 */
/* this is a quick status read, gets minimal status but no buffered data */
	case I34_IOCGSTATUS:
	{
		struct mgadvd_status_s status;
		
		dprintk("mgadvd ioctl I34_IOCGSTATUS\n");

		if ( (ret = i34_status(mga->i34, &status)) != 0 )
			return ret;
		if (copy_to_user((struct mgadvd_status_s *)arg, &status,
		     sizeof(struct mgadvd_status_s)))
			return -EFAULT;
		break;
	}

/* player dispatches thread for this status read. this sleeps until vinsync,
   then reads status + buffered data and copies it to userland
 */
	case I34_IOCGVSYNCSTATUS:
	{
		struct mgadvd_status_s status;

		i34_vsync_status(mga->i34, &status);
		ret = copy_to_user(	(struct mgadvd_status_s *)arg,
					&status,
					sizeof(struct mgadvd_status_s) );

		if (ret != 0) return -EFAULT;
		break;
	}


#ifdef CSS_SUPPORT
	case I34_IOCS_DISK_KEY:
		dprintk("mgadvd ioctl I34_IOCS_DISK_KEY\n");
		if ( copy_from_user(&mga->i34.cssauth, (dvd_authinfo *)arg,
				    sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		if ( css_write_disk_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		break;

	case I34_IOCS_TITLE_KEY:
		dprintk("mgadvd ioctl I34_IOCS_TITLE_KEY\n");
		if ( copy_from_user(&mga->i34.cssauth, (dvd_authinfo *)arg,
				     sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		if ( css_write_title_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		break;

	case I34_IOCG_CHAL_KEY:
		dprintk("mgadvd ioctl I34_IOCG_CHAL_KEY\n");
		if ( css_read_chal_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		if ( copy_to_user((dvd_authinfo *)arg, &mga->i34.cssauth,
				    sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		break;

	case I34_IOCS_CHAL_KEY:
		dprintk("mgadvd ioctl I34_IOCS_CHAL_KEY\n");
		if ( copy_from_user(&mga->i34.cssauth, (dvd_authinfo *)arg,
				     sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		if ( css_write_chal_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		break;

	case I34_IOCG_RESP_KEY:
		dprintk("mgadvd ioctl I34_IOCG_RESP_KEY\n");
		if ( css_read_resp_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		if ( copy_to_user((dvd_authinfo *)arg, &mga->i34.cssauth,
				   sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		break;

	case I34_IOCS_RESP_KEY:
		dprintk("mgadvd ioctl I34_IOCS_RESP_KEY\n");
		if ( copy_from_user(&mga->i34.cssauth, (dvd_authinfo *)arg,
				     sizeof(dvd_authinfo)) != 0 )
			return -EFAULT;
		if ( css_write_resp_key(mga->i34.cssauth) != 0 )
			return -EINVAL;
		break;
#endif

	default:
		*handled = 0;
		break;
	}
	return 0;
}


static int mga_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct mgadvd_v4l1* mv = (struct mgadvd_v4l1*)dev;
	int rv, handled;

	switch (cmd) {

	case VIDIOCGCAP:
		{
			struct video_capability b;
			int			width;
			int			height;
			int			depth;

			dprintk("mga_v4l1 ioctl VIDIOCGCAP\n");
			strncpy(b.name, mv->video_dev.name, sizeof(b.name));
			
			/* bes functionality */
			b.type =  VID_TYPE_SCALES
				| VID_TYPE_OVERLAY
				| VID_TYPE_CHROMAKEY;

			b.channels = 0;
			b.audios = 0;
			
			b.type |= VID_TYPE_CAPTURE
				| VID_TYPE_MPEG_DECODER
				| VID_TYPE_FRAMERAM;

			/* theoretically we could also flag VID_TYPE_SUBCAPTURE
			but this is not even implemented in the BTTV driver */
			
			mga_display_size(mv->mga, &width, &height, &depth);
			
			b.maxwidth = (width < 720) ? 720 : width;
			b.maxheight = (height < 576) ? 576 : height;
			b.minwidth = MGA_MIN_WIDTH;
			b.minheight = MGA_MIN_HEIGHT;

			if ( copy_to_user(arg, &b, sizeof(b)) != 0 ) {
				return -EFAULT;
			}
			return 0;
		}


	case VIDIOCGPICT:
		{
			struct video_picture p = mv->picture;

			dprintk("mga_v4l1 ioctl VIDIOCGPICT\n");
			p.depth = mgacap_get_depth(mv->mga->cap_v4l1);
			switch (p.depth) {
			case 8:
				p.palette = VIDEO_PALETTE_GREY;
				break;

			case 15:
				p.palette = VIDEO_PALETTE_RGB555;
				break;

			case 16:
				p.palette = VIDEO_PALETTE_RGB565;
				break;

			case 24:
				p.palette = VIDEO_PALETTE_RGB24;
				break;

			case 32:
				p.palette = VIDEO_PALETTE_RGB32;
				break;
			}
			
			/*
			vp.brightness = 0x8000;
			vp.hue = 0x8000;
			vp.colour = 0x8000;
			vp.whiteness = 0x8000;
			vp.depth = 8;
			vp.palette = VIDEO_PALETTE_YUV420;
			*/

			if (copy_to_user(arg, &p, sizeof(p)) != 0 ) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCGUNIT:
		{
			struct video_unit vu;

			dprintk("mga_v4l1 ioctl VIDIOCGUNIT\n");
			vu.video = mv->video_dev.minor;
			vu.vbi = VIDEO_NO_UNIT;
			vu.radio = VIDEO_NO_UNIT;
			vu.audio = VIDEO_NO_UNIT;
			vu.teletext = VIDEO_NO_UNIT;
			
			mgacap_get_unit( mv->mga->cap_v4l1, &vu );
			if ( copy_to_user(arg, &vu, sizeof(vu)) != 0 )
				return -EFAULT;
			return 0;
		}

//	case VIDIOCMCAPTURE:
//		{
//			mgavideo_grab_needed(mv->mga->video, 1);

	default:
		rv = mga_dvd_ioctl( dev, cmd, arg, &handled );
		if ( handled != 0 ) {
			return rv;
		}

		rv = mgacap_ioctl( mv->mga->cap_v4l1, cmd, arg, &handled );
		if ( handled != 0 ) {
			return rv;
		}
		return -ENOIOCTLCMD;
	}
	return 0;
}



static int mga_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct mgadvd_v4l1 *mv = (struct mgadvd_v4l1*)dev;

	return mgacap_mmap( mv->mga->cap_v4l1, adr, size );
}



static int mga_init_done(struct video_device *dev)
{
	return 0;
}


static struct video_device mga_template =
{
	name:		MGADVD_ID_STRING,
	type:		VID_TYPE_CAPTURE | VID_TYPE_OVERLAY | VID_TYPE_CLIPPING
			| VID_TYPE_FRAMERAM | VID_TYPE_SCALES
			| VID_TYPE_SUBCAPTURE |	VID_TYPE_MPEG_DECODER,
	hardware:	VID_HARDWARE_ZR36700,
	open:		mga_open,
	close:		mga_close,
	read:		mga__read,
	write:		mga__write,
	poll:		mgadvd_poll,
	ioctl:		mga_ioctl,
	mmap:		mga_mmap,
	initialize:	mga_init_done,
	0,
};


static void mga_init(struct mgadvd_v4l1 *mv, struct mga_dev_s *mga)
{
	mv->mga = mga;
	mv->user = 0;

	mga->dvd_v4l1 = mv;

	mv->picture.colour = 32768;
	mv->picture.brightness = 32768;
	mv->picture.hue = 32768;
	mv->picture.contrast = 32768;
	mv->picture.whiteness = 0;
	mv->picture.depth = 0;
	mv->picture.palette = 0;

	mv->write_mode = VID_WRITE_MPEG_VID;
}


#ifdef MODULE
EXPORT_NO_SYMBOLS;
int init_module(void)
#else
int init_dvd(void) //??
#endif
{
	struct mgadvd_v4l1	 *mv;
	struct mga_dev_s *mga;
	int		i;
	int		rv;
	int		active = 0;

	printk(KERN_INFO "Matrox DVD Driver (V4L1)\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {
		if ( mga->i34 == NULL ) {
			/* no dvd addon on this card */
			mga_release_card(mga);
			continue;
		}
		if ((mv = kmalloc(sizeof(*mv), GFP_KERNEL)) == NULL) {
			rv = -ENOMEM;
			goto skip;
		}
		memset( mv, 0, sizeof(*mv) );

		memcpy(&mv->video_dev, &mga_template, sizeof(mga_template));
		sprintf(mv->video_dev.name, "mgadvd-%u", i);

#if LINUX_VERSION_CODE >= 0x020405
		if (video_register_device(&mv->video_dev, VFL_TYPE_GRABBER, video_nr) < 0) {
#else
		if (video_register_device(&mv->video_dev, VFL_TYPE_GRABBER) < 0) {
#endif
			rv = -EIO;
			goto skip;
		}

		mga_init(mv, mga);
		active++;
		continue;
		
	skip:
		if (mv != NULL) {
			kfree(mv);
		}
		mga_release_card(mga);
		if (rv)
			return rv;
	}

	if(active == 0) {
		return -ENXIO;
	}
	return 0;
}


#ifdef MODULE
void cleanup_module(void)
{
	int i = 0;
	struct mgadvd_v4l1* mv;
	struct mga_dev_s *mga;

	for (i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		mv = mga->dvd_v4l1;

		if (mv != NULL) {
			video_unregister_device(&mv->video_dev);

			mga_release_card(mv->mga);
			mga->dvd_v4l1 = NULL;
			kfree(mv);
		}
		mga_release_card(mv->mga);
	}
}
#endif

