/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the G100 and G200 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 *	Gernot Ziegler <gz@lysator.liu.se>
 */

/* Changes: 000219 - gz: .) Changed the picture parameter range 
                            (like brightness) to signed char.
                         .) Implemented VIDIOCGAUDIO.
			 .) Changed input channel order (to fix xawtv).
			    (_incompatible_ with the current buz_tools !!!!)
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>
#include <linux/proc_fs.h>

#if LINUX_VERSION_CODE >= 0x020300
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif

#include <linux/videodev.h>

#include <linux/version.h>
#include <asm/uaccess.h>

#include "aiee.h"
#include "mgacore.h"
#include "mgavideo.h"
#include "i33.h"
#include "zr36060.h"
#include "mgacap.h"
#include "ks0127.h"
#include "audiochip.h"
#include "videodev_mjpeg.h"
#include <linux/video_decoder.h>
#include <linux/video_encoder.h>

#include "mgadvd.h" /* for MJPEG_DECODER etc.. */


#define MIN(a,b) (((a)>(b))?(b):(a))
#define MAX(a,b) (((a)>(b))?(a):(b))

#define MGA_MAX_WIDTH 720
#define MGA_MIN_WIDTH 8
#define MGA_MAX_HEIGHT 580
#define MGA_MIN_HEIGHT 8 

#define MGA_NAME "mga_v4l1cap"
#define V4L_NAME      "V4L-mgavideo"   /* name of the driver */

#define V4L_ERR       KERN_ERR     V4L_NAME
#define V4L_DEBUG     KERN_INFO    V4L_NAME
#define V4L_INFO      KERN_INFO    V4L_NAME
#define V4L_WARNING   KERN_WARNING V4L_NAME


#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */
static int video_nr = -1;


#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif
#if LINUX_VERSION_CODE >= 0x020405
MODULE_PARM(video_nr,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


struct mgajpg_v4l1 {
	struct video_device	video_dev;
	struct mga_dev_s	*mga;
	int			id;
	int			initialized;
	int			user;
        int			mgacaps;

	/* Video for Linux parameters */
	struct video_picture picture;	/* Current picture params */

        int map_mjpeg_buffers;
};



static int mga_open(struct video_device *dev, int flags)
{
	struct mgajpg_v4l1* mv = (struct mgajpg_v4l1*)dev;
	int rv;
	
	if (i33_lock(mv->mga)) {
		return -EBUSY;
	}
	mgavideo_lock(mv->mga);

	rv = mgacap_open( mv->mga->cap_v4l1, flags );
	if (rv) {
		mgavideo_unlock(mv->mga);
		i33_unlock(mv->mga);
		return rv;
	}
	
	dprintk("mgacap: opening device\n");
	mv->user++;

	MOD_INC_USE_COUNT;
	return 0;
}


static void mga_close(struct video_device *dev)
{
	struct mgajpg_v4l1* mv = (struct mgajpg_v4l1*)dev;
	enum	buz_codec_mode mode;
	
	dprintk("mgacap: closing device\n");
	
	mgacap_close( mv->mga->cap_v4l1 );

	if (mv->mga->options & MGAVIDEO_HAS_ZR36060) {
	
		mode = BUZ_MODE_IDLE;

		zr36060_ioctl(mv->mga->zr, ZR36060_SET_MODE, &mode);

		zr36060_ioctl(mv->mga->zr, ZR36060_CANBUFS, NULL);
	}

	mga_set_defaults(mv->mga);

	mgavideo_unlock(mv->mga);
	i33_unlock(mv->mga);
		
	mv->user--;

	MOD_DEC_USE_COUNT;
}


static long mga__read(struct video_device *dev, char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}


static long mga__write(struct video_device *dev, const char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}


static int
mga_mjpeg_ioctl(struct video_device *dev, unsigned int cmd, void *arg, int *handled)
{
	struct mgajpg_v4l1* mv = (struct mgajpg_v4l1*)dev;
	struct mga_video_s *video;

	video = mv->mga->video;
	
	if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0) {
		*handled = 0;
		return 0;
	}
	
	*handled = 1;

	switch (cmd) {

	case MJPIOC_G_PARAMS:
	  {
		struct mjpeg_params	params;
		int			res;


		dprintk(V4L_DEBUG "-%u: ioctl MJPIOC_G_PARAMS\n", mv->id);
		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;


		res = zr36060_ioctl(mv->mga->zr, ZR36060_GET_PARAMS, &params);
		if (res)
			return res;

		dprintk("input = %d norm = %d fields_per_buf=%d x=%d y=%d width=%d height=%d quality=%d\n",
			params.input, params.norm, params.field_per_buff, params.img_x, params.img_y,
			params.img_width, params.img_height, params.quality);

		if (copy_to_user(arg, &params, sizeof(params))) 
			return -EFAULT;

		return 0;
	  }

	case MJPIOC_S_PARAMS:
	  {
		int	res;
		struct mjpeg_params params;

		if (copy_from_user(&params, arg, sizeof(params)))
			return -EFAULT;

		dprintk(V4L_DEBUG "-%u: ioctl MJPIOC_S_PARAMS\n", mv->id);

		dprintk("input = %d norm = %d fields_per_buf=%d x=%d y=%d width=%d height=%d quality=%d\n",
			params.input, params.norm, params.field_per_buff, params.img_x, params.img_y,
			params.img_width, params.img_height, params.quality);

		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;

		/* Check the params first before overwriting our internal values */
		res = zr36060_ioctl(mv->mga->zr, ZR36060_SET_PARAMS, &params);
		if (res)
			return res;

		if (copy_to_user(arg, &params, sizeof(params)))
			return -EFAULT;
		return 0;
	  }

        case MJPIOC_REQBUFS:
	  {
		int	res;
		struct mjpeg_requestbuffers br;

		if (copy_from_user(&br, arg, sizeof(br)))
			return -EFAULT;

		dprintk(V4L_DEBUG "-%u: ioctl MJPIOC_REQBUFS count = %lu size=%lu\n",
                            mv->id, br.count,br.size);

		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;

		res = zr36060_ioctl(mv->mga->zr, ZR36060_REQBUFS, &br);

		/* The next mmap will map the mjpeg buffer */
		mv->map_mjpeg_buffers = 1;

		if (copy_to_user(arg, &br, sizeof(br)))
			return -EFAULT;

		return 0;
	  }

	case MJPIOC_QBUF_CAPT:
	  {
	    int nb;
	    
	    if (copy_from_user((void *)&nb, (void *)arg, sizeof(nb))) {
	      return -EFAULT;
	    }
	    dprintk(V4L_DEBUG "-%u: ioctl MJPIOC_QBUF_CAPT %d\n",mv->id, nb);

		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;


	    return zr36060_ioctl(mv->mga->zr, ZR36060_QBUF_CAPT, &nb);
	  }

	case MJPIOC_QBUF_PLAY:
	  {
	    int nb;
	    
	    if (copy_from_user((void *)&nb, (void *)arg, sizeof(nb))) 
	      return -EFAULT;
         
	    dprintk(V4L_DEBUG "-%u: ioctl MJPIOC_QBUF_PLAY %d\n",mv->id, nb);
		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;

	    return zr36060_ioctl(mv->mga->zr, ZR36060_QBUF_PLAY, &nb);
	  }

	case MJPIOC_SYNC:
	  {
		struct mjpeg_sync bs;
		int res;
	    

		dprintk(V4L_DEBUG "-%u: buz ioctl MJPIOC_SYNC\n", mv->id);
		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;

		res =  zr36060_ioctl(mv->mga->zr, ZR36060_SYNC, &bs);
		dprintk("frame=%ld, length=%ld, seq=%ld, res =0x%x\n", 
			bs.frame, bs.length, bs.seq, res);

		if (!res && copy_to_user(arg, &bs, sizeof(bs))) 
			return -EFAULT;
	    
		return res;
	  }


	case MJPIOC_G_STATUS:
	  {
	    struct mjpeg_status bs;
	    dprintk(V4L_DEBUG "-%u: buz ioctl BUZIOC_G_STATUS\n", mv->id);
		if ((mv->mga->options & MGAVIDEO_HAS_ZR36060) == 0)
			return -EINVAL;

	    if (copy_from_user(&bs, arg, sizeof(bs))) 
	      return -EFAULT;

#ifdef	NEVER
	    mgavideo_get_channelstatus(mv->mga, bs.input, &bs.signal, &bs.norm, &bs.color);
#endif
	    
	    if (copy_to_user(arg, &bs, sizeof(bs))) 
	      return -EFAULT;

	    return 0;
	  }

	default:
		*handled = 0;
		break;
	}
	return 0;
}


static int mga_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct mgajpg_v4l1* mv = (struct mgajpg_v4l1*)dev;
	struct mga_video_s *video;
	int rv, handled;

	video = mv->mga->video;

	switch (cmd) {

	case VIDIOCGCAP:
		{
			struct video_capability b;
			int			width;
			int			height;
			int			depth;

			dprintk("mga_v4l1 ioctl VIDIOCGCAP\n");
			strncpy(b.name, mv->video_dev.name, sizeof(b.name));
			
			/* bes functionality */
			b.type = VID_TYPE_SCALES | VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY;
			b.channels = 0;
			b.audios = 0;
			
			b.type |= VID_TYPE_CAPTURE
				/* | VID_TYPE_FRAMERAM */
				;
			if (mv->mga->options & MGAVIDEO_HAS_ZR36060) {
				b.type |= VID_TYPE_MJPEG_DECODER | VID_TYPE_MJPEG_ENCODER;
			}
			/* theoretically we could also flag VID_TYPE_SUBCAPTURE
			but this is not even implemented in the BTTV driver */
			
			if ( mv->mga->options & MGAVIDEO_HAS_DECODER ) {
				b.channels += 2;
			}
			if (mv->mga->options & MGAVIDEO_HAS_TUNER) {
				b.type |= VID_TYPE_TUNER;
				b.channels += 1;
			}
			if (mv->mga->options & MGAVIDEO_HAS_AUDIO) {
				b.audios += 1;
			}
			
			mga_display_size(mv->mga, &width, &height, &depth);
			
			b.maxwidth = width < 720 ? 720 : width;
			b.maxheight = height < 588 ? 588 : height;
			b.minwidth = MGA_MIN_WIDTH;
			b.minheight = MGA_MIN_HEIGHT;

			if (copy_to_user(arg, &b, sizeof(b))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCGCHAN:
		{
			struct video_channel v;

			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}
			dprintk("mga_v4l1 ioctl VIDIOCGCHAN for channel %d\n", v.channel);
			switch (v.channel) {
			case 0:
				strcpy(v.name, "Composite");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;
			case 1:
				strcpy(v.name, "S-Video");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;

			case 2:
				strcpy(v.name, "Television");
				v.tuners = 1;
				v.flags = VIDEO_VC_TUNER;
				v.type = VIDEO_TYPE_TV;
				break;

			default:
				return -EINVAL;
			}
			v.norm = mgavideo_get_norm(video, v.channel);

			if(mv->mgacaps & MGAVIDEO_HAS_AUDIO) {
				v.flags |= VIDEO_VC_AUDIO;
			}

			if (copy_to_user(arg, &v, sizeof(v))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSCHAN:
		{
			struct video_channel v;
			int input, norm, status = 0;


			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}

			dprintk("mga_v4l1 ioctl VIDIOCSCHAN: channel=%d, norm=%d\n", 
			       v.channel, v.norm);
			switch (v.channel) {
			case 0:
				input = KS_INPUT_COMPOSITE;
				break;
			case 1:
				input = KS_INPUT_SVIDEO;
				break;

			case 2:
				input = KS_INPUT_TUNER;
				break;

			default:
				return -EINVAL;
			}

			switch (v.norm) {
			case VIDEO_MODE_PAL:
				norm = KS_STD_PAL;
				break;
			case VIDEO_MODE_NTSC:
				norm = KS_STD_NTSC;
				break;
			case VIDEO_MODE_SECAM:
				norm = KS_STD_SECAM;
				break;
			case VIDEO_MODE_AUTO:
				norm = KS_STD_AUTO;
				break;
			default:
				return -EINVAL;
			}

			mgavideo_set_active(video, input, norm);
			if (mv->mga->options & MGAVIDEO_HAS_ZR36060) {
				zr36060_ioctl(mv->mga->zr, ZR36060_SET_INPUT, &v.channel);
				zr36060_ioctl(mv->mga->zr, ZR36060_SET_NORM, &v.norm);
			}
			
			if (norm == KS_STD_AUTO) {
				printk("Autodetection started !");
				mgavideo_decoder(video, KS0127_GET_STATUS, &status);
				mdelay(10);
				norm = (status >> 4) ? KS_STD_NTSC : KS_STD_PAL;

				mgavideo_set_active(video, input, norm);
				dprintk("VIDIOCSCHAN: Autodetected and set %s norm\n", 
					(norm == KS_STD_PAL) ? "PAL" : "NTSC");
			}
			return 0;
		}

	case VIDIOCGTUNER:
		{
			struct video_tuner v;
			int		   tmp;

			if(copy_from_user(&v, arg, sizeof(v)) != 0)
				return -EFAULT;
			if(v.tuner != 0)
				return -EINVAL;
			strcpy(v.name, "Television");
			v.rangelow = 0;
			v.rangehigh = 0xFFFFFFFF;
			v.flags = VIDEO_TUNER_PAL |
			          VIDEO_TUNER_NTSC |
				  VIDEO_TUNER_SECAM |
			          VIDEO_TUNER_NORM;

			switch(mgavideo_get_norm(video, 2)) {
			case KS_STD_PAL:
				v.mode = VIDEO_MODE_PAL;
				break;
			case KS_STD_NTSC:
				v.mode = VIDEO_MODE_NTSC;
				break;
			case KS_STD_SECAM:
				v.mode = VIDEO_MODE_SECAM;
				break;
			default:
				v.mode = VIDEO_MODE_NTSC;
				break;
			}

			mgavideo_decoder(video, KS0127_GET_STATUS, &tmp);
			dprintk("GTUNER ks0127 status = %x clock %d cdet = %d paldet=%d ffrdet = %d video det = %d\n", 
				tmp, tmp & 1, (tmp >> 2) & 1, (tmp >>3) & 1, (tmp >> 4) ? 60 : 50, (tmp >> 5) & 1);
			v.signal = ((tmp >> 5) & 1) ? 0 : 1;
			if(copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}

	case VIDIOCSTUNER:
		{
			struct video_tuner v;
			int		   norm;
			dprintk("mga_v4l1 ioctl VIDIOCSTUNER v.tuner=%d\n", v.tuner);
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.tuner == 0)
				return -EINVAL;

			switch (v.mode) {
			case VIDEO_MODE_PAL:
				norm = KS_STD_PAL;
				break;
			case VIDEO_MODE_NTSC:
				norm = KS_STD_NTSC;
				break;
			case VIDEO_MODE_SECAM:
				norm = KS_STD_SECAM;
				break;
			default:
				return -EOPNOTSUPP;
			}

			mgavideo_set_norm(video, 2, norm);
			return 0;
		}

	case VIDIOCGPICT:
		{
			struct video_picture p = mv->picture;

			dprintk("mga_v4l1 ioctl VIDIOCGPICT\n");
			p.depth = mgacap_get_depth(mv->mga->cap_v4l1);
			switch (p.depth) {
			case 8:
				p.palette = VIDEO_PALETTE_GREY;
				break;

			case 15:
				p.palette = VIDEO_PALETTE_RGB555;
				break;

			case 16:
				p.palette = VIDEO_PALETTE_RGB565;
				break;

			case 24:
				p.palette = VIDEO_PALETTE_RGB24;
				break;

			case 32:
				p.palette = VIDEO_PALETTE_RGB32;
				break;
			}

			if (copy_to_user(arg, &p, sizeof(p))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSPICT:
		{
		        int brightness, hue, contrast, colour;
			struct video_picture p;

			if (copy_from_user(&p, arg, sizeof(p))) {
				return -EFAULT;
			}
			dprintk("mgacap ioctl VIDIOCSPICT bri=%d hue=%d col=%d con=%d dep=%d pal=%d\n",
			       p.brightness, p.hue, p.colour, p.contrast, p.depth, p.palette);

			/* We want -127 to 128 (signed char), but we get 0-65535 */  
			brightness = (p.brightness >> 8) - 128;
			hue = (p.hue >> 8) - 128;
			contrast = (p.contrast >> 8) - 128;
			colour = (p.colour >> 8) - 128;
			mgavideo_set_color(video, brightness, hue, colour, contrast);

			mv->picture = p;
			return 0;
		}


	case VIDIOCGFREQ:
		{
			unsigned long freq;

			freq = mgavideo_get_freq(video);
			if(copy_to_user(arg, &freq, sizeof(freq)))
				return -EFAULT;
			dprintk("mga_v4l1 ioctl VIDIOCGFREQ: freq=0x%lx\n", freq);
			return 0;
		}

	case VIDIOCSFREQ:
		{
			unsigned long	freq;
			int	norm;

			norm = mgavideo_get_norm(video, 2);
			if(copy_from_user(&freq, arg, sizeof(freq)))
				return -EFAULT;

			dprintk("mga_v4l1 ioctl VIDIOCSFREQ: freq=0x%lx\n", freq);
			mgavideo_set_freq(video, freq);
			return 0;
		}


	case VIDIOCGAUDIO:
        	{
		  struct video_audio v; 
		  dprintk("mga_v4l1 ioctl VIDIOCGAUDIO:\n");
		  /*return -EINVAL;*/ 
		  /* v = mv->mga->audio_dev; Ok, let's try without, but I think I have to create one */
		  v.flags&=~(VIDEO_AUDIO_MUTE|VIDEO_AUDIO_MUTABLE);
		  v.flags|=VIDEO_AUDIO_MUTABLE;
		  strcpy(v.name,"Television");
		  
		  if (mv->mga->options & MGAVIDEO_HAS_AUDIO) {
		    __u16 temp_voll,temp_volr;
		    v.flags|=VIDEO_AUDIO_VOLUME |
		      VIDEO_AUDIO_BASS |
		      VIDEO_AUDIO_TREBLE;
		    mgavideo_audio(video, AUDC_GET_VOLUME_LEFT, &temp_voll);
		    mgavideo_audio(video, AUDC_GET_VOLUME_RIGHT, &temp_volr);
		    v.volume=MAX(temp_voll,temp_volr);
		    v.balance=(32768*MIN(temp_voll,temp_volr))/
		      (v.volume ? v.volume : 1);
		    
		    v.balance=(temp_voll<temp_volr)?
		      (65535-v.balance):v.balance;
		    mgavideo_audio(video, AUDC_GET_BASS,&(v.bass));
		    mgavideo_audio(video, AUDC_GET_TREBLE,&(v.treble));
		    mgavideo_audio(video, AUDC_GET_STEREO,&(v.mode));
		  }
		  else v.mode = VIDEO_SOUND_MONO; /* We would have no sound, how do I write that ? */
		  if(copy_to_user(arg,&v,sizeof(v)))
		    return -EFAULT;
		  return 0;
        	}


	case VIDIOCSAUDIO:
        	{
			struct video_audio v;
			dprintk("mga_v4l1 ioctl VIDIOCSAUDIO: vol=%d, bal=%d !\n",
				v.volume, v.balance);
			if(copy_from_user(&v,arg, sizeof(v)))
				return -EFAULT;

			if (mv->mga->options & MGAVIDEO_HAS_AUDIO)
			  { if (!(v.flags&VIDEO_AUDIO_MUTE)) 
			    {
			      __u16 temp_vol;
			      temp_vol=(2*MIN(65536-v.balance,32768) *
					v.volume)/65536;
			      mgavideo_audio(video, AUDC_SET_VOLUME_LEFT, &temp_vol);	
			      temp_vol=(2*MIN(v.balance,32768) *
					v.volume)/65536;
			      mgavideo_audio(video, AUDC_SET_VOLUME_RIGHT, &temp_vol);
			      /* bass and treble are meaningless, but we set them anyways */
			      mgavideo_audio(video, AUDC_SET_BASS,&(v.bass));
			      mgavideo_audio(video, AUDC_SET_TREBLE,&(v.treble));
			      
			      mgavideo_audio(video, AUDC_SET_STEREO,&(v.mode));
			    }
			  else
			    mgavideo_audio(video, AUDC_SWITCH_MUTE, 0);	
			  }
			return 0;
        	}


	case VIDIOCMCAPTURE:
		rv = mgacap_ioctl( mv->mga->cap_v4l1, cmd, arg, &handled );
		if ( mgacap_get_pending( mv->mga->cap_v4l1 ) == 1 ) {
			mgavideo_grab_needed(mv->mga->video, 1);
		}
		return rv;
		
	case VIDIOCSYNC:
		rv = mgacap_ioctl( mv->mga->cap_v4l1, cmd, arg, &handled );
		if ( mgacap_get_pending( mv->mga->cap_v4l1 ) == 0 ) {
			mgavideo_grab_needed(mv->mga->video, 0);
		}
		return rv;
			

	case VIDIOCGUNIT:
		{
			struct video_unit vu;

			dprintk("mga_v4l1 ioctl VIDIOCGUNIT\n");
			vu.video = mv->video_dev.minor;
			vu.vbi = VIDEO_NO_UNIT;
			vu.radio = VIDEO_NO_UNIT;
			vu.audio = VIDEO_NO_UNIT;
			vu.teletext = VIDEO_NO_UNIT;
			
			mgacap_get_unit( mv->mga->cap_v4l1, &vu );
			if (copy_to_user(arg, &vu, sizeof(vu)))
				return -EFAULT;
			return 0;
		}

	case VIDIOCGMBUF:
		{
			dprintk( "mga_v4l1 ioctl VIDIOCGMBUF\n");
		
			/* mmap v4l buffers next */
			mv->map_mjpeg_buffers = 0;
			
			return mgacap_ioctl( mv->mga->cap_v4l1, cmd, arg, &handled );
		}


	default:
		rv = mga_mjpeg_ioctl( dev, cmd, arg, &handled );
		if ( handled ) {
			return rv;
		}
	
		rv = mgacap_ioctl( mv->mga->cap_v4l1, cmd, arg, &handled );
		if ( handled ) {
			return rv;
		}
		return -ENOIOCTLCMD;
	}
	return 0;
}



static int mga_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct mgajpg_v4l1 *mv = (struct mgajpg_v4l1*)dev;
	int		i;

	if (mv->map_mjpeg_buffers) {
		if ((i = zr36060_mmap_bufs(mv->mga->zr, (unsigned long)adr, size)))
			return i;
	} else  {
		return mgacap_mmap( mv->mga->cap_v4l1, adr, size );
	}
	return 0;
}



static int mga_init_done(struct video_device *dev)
{
	return 0;
}

static struct video_device mga_template =
{
	name:		MGA_NAME,
	type:		VID_TYPE_CAPTURE | VID_TYPE_OVERLAY | VID_TYPE_CLIPPING |
				VID_TYPE_FRAMERAM | VID_TYPE_SCALES | VID_TYPE_SUBCAPTURE,
	hardware:	VID_HARDWARE_BT848,
	open:		mga_open,
	close:		mga_close,
	read:		mga__read,
	write:		mga__write,
	poll:		NULL,
	ioctl:		mga_ioctl,
	mmap:		mga_mmap,
	initialize:	mga_init_done,
	0,
};



static void mga_init(struct mgajpg_v4l1 *mv, struct mga_dev_s *mga)
{
	mv->mga = mga;
	mv->user = 0;

	mga->jpg_v4l1 = mv;

	mv->map_mjpeg_buffers = 1;

	mv->picture.colour = 32768;
	mv->picture.brightness = 32768;
	mv->picture.hue = 32768;
	mv->picture.contrast = 32768;
	mv->picture.whiteness = 0;
	mv->picture.depth = 0;
	mv->picture.palette = 0;

	mv->initialized = 1;
}


#ifdef MODULE
int init_module(void)
#else
int init_mga_cards(struct video_init *unused)
#endif
{
	struct mgajpg_v4l1	 *mv;
	struct mga_dev_s *mga;
	int		 i;

	printk(KERN_INFO "Matrox Marvel and Rainbow Runner G Capture Driver (V4L1)\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {
		if ((mv = kmalloc(sizeof(*mv), GFP_KERNEL)) == NULL) {
			return -ENOMEM;
		}
		memset( mv, 0, sizeof(*mv) );

		memcpy(&mv->video_dev, &mga_template, sizeof(mga_template));
		sprintf(mv->video_dev.name, "mgacap-%u", i);
		
		mga_init(mv, mga);
		
		if (mv->mga->options & MGAVIDEO_HAS_ZR36060) {
			mv->video_dev.type |= VID_TYPE_MJPEG_DECODER | VID_TYPE_MJPEG_ENCODER;
		}

#if LINUX_VERSION_CODE >= 0x020405
		if (video_register_device(&mv->video_dev, VFL_TYPE_GRABBER, video_nr) < 0) {
#else
		if (video_register_device(&mv->video_dev, VFL_TYPE_GRABBER) < 0) {
#endif
                        mga_release_card(mga);
			return -EIO;
		}
	}

	if(i == 0) {
		return -ENXIO;
	}
	return 0;
}


#ifdef MODULE
void cleanup_module(void)
{
	int i = 0;
	struct mgajpg_v4l1* mv;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		mv = mga->jpg_v4l1;

		if(mv->initialized) {
		
			video_unregister_device(&mv->video_dev);

			/* need to release it 2 times */
			mga_release_card(mv->mga);
			mga_release_card(mv->mga);

			mga->jpg_v4l1 = NULL;
			kfree(mv);
		}
	}
}
#endif

