/*
 * mgayuv.c
 *
 * New version of mga_vid using mgacore
 *
 * Based on code Copyright (C) 1999 Aaron Holtzman
 *
 * This software has been released under the terms of the GNU Public
 * license. See http://www.gnu.org/copyleft/gpl.html for details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/videodev.h>
#include <linux/proc_fs.h>

#ifndef VIDEO_WINDOW_CHROMAKEY
#define VIDEO_WINDOW_CHROMAKEY	16
#endif /* VIDEO_WINDOW_CHROMAKEY */

#include "mgacore.h"
#include "mgabes.h"

#define BES_MAXSRCW 768
#define BES_MAXSRCH 576


#define MGAYUV_ID_STRING "Matrox G200/G400 YUV"


struct mgayuv_s
{
	struct video_device	video_dev;
	struct mga_dev_s	*mga;
};

#if LINUX_VERSION_CODE >= 0x020405
static int video_nr;
#endif

static int norm = VIDEO_MODE_PAL;
struct video_window vw = { 0, 0, 720, 480 };
struct video_buffer vbufalloc;
struct video_buffer vbuf = { 0, width:720, height:480};
struct video_picture vp = {
	brightness: 0x8000,
	hue: 0x8000,
	colour: 0x8000,
	whiteness: 0x8000,
	depth: 8,
	palette: VIDEO_PALETTE_YUV420
};


static u32 frame_size = 0;

static long mgayuv_read(struct video_device *v, char *buf, unsigned long count, 
	int noblock)
{
	return -EINVAL;
}

static long mgayuv_write(struct video_device *v, const char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}


static void program_bes(struct mga_dev_s *mga)
{
	mgabes_set_buffers(mga->bes, (u32)vbuf.base, (u32)vbuf.base/*0*/ /* no frame B */,
		vbuf.width, vbuf.height,
		vp.palette);

	vbuf.bytesperline = mgabes_pitch_rounding(mga->bes, vbuf.width) * 1; /* for a line of Y */

	mgabes_set_source(mga->bes, 0, 0, vbuf.width, vbuf.height);
	mgabes_set_frame(mga->bes, BES_FRAME_A);
	mgabes_set_window(mga->bes, vw.x, vw.y, vw.width, vw.height);
}


static int attach_buffer(struct mga_dev_s *mga)
{
	if (vbuf.base)
		mga_free_buffer(mga, frame_size);

	/* allocate a buffer for a video frame */
	
	vbufalloc.width = mgabes_pitch_rounding(mga->bes, BES_MAXSRCW);
	vbufalloc.height = BES_MAXSRCH;
	vbufalloc.bytesperline = vbufalloc.width * 2; /* YUV, 16bpp*/
	frame_size = vbufalloc.bytesperline * vbufalloc.height;
	
	vbuf.base = (void*) mga_alloc_buffer(mga, frame_size, 4096);
	
	program_bes( mga );

	return 0;
}


static int mgayuv_mmap(struct video_device *dev, const char *addr, unsigned long size)
{
	struct mgayuv_s *myuv = (struct mgayuv_s*)dev;
	struct mga_dev_s *mga = myuv->mga;
        unsigned long phys_addr;

	if (!mga) return -ENODEV;

#if LINUX_VERSION_CODE >= 0x020300
	phys_addr = mga->dev->resource[0].start + (u32)vbuf.base;
#else
	phys_addr = (mga->dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK) + (u32)vbuf.base;
#endif

	if (remap_page_range((unsigned long)addr, phys_addr, size, PAGE_SHARED)) {
		printk("mgayuv: error mapping video buffer\n");
		return -EAGAIN;
	}

	return 0;
}

static int mgayuv_open(struct video_device *dev, int mode)
{
	struct mgayuv_s *myuv = (struct mgayuv_s*)dev;
	struct mga_dev_s *mga = myuv->mga;

	if (!mga) return -ENODEV;

	mgabes_set_mode(mga->bes, BES_HFILTER | BES_VFILTER | BES_ACC2XZOOM | BES_DITHER);
	attach_buffer(mga);

	MOD_INC_USE_COUNT;
	return 0;
}

static void mgayuv_close(struct video_device *dev)
{
	struct mgayuv_s *myuv = (struct mgayuv_s*)dev;
	struct mga_dev_s *mga = myuv->mga;

	if (!mga) return;

	mgabes_desactivate(mga->bes);

	if (vbuf.base)
		mga_free_buffer(mga, frame_size);
	vbuf.base = 0;

	MOD_DEC_USE_COUNT;
	return;
}

static int mgayuv_init_done(struct video_device *dev)
{
	return 0;
}

static int mgayuv_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct mgayuv_s *myuv = (struct mgayuv_s*)dev;
	struct mga_dev_s *mga = myuv->mga;
	struct mga_bes_s *bes;

	if (!mga) return -ENODEV;

	bes = mga->bes;

	switch(cmd)
	{
		case VIDIOCGCAP:
		{
			struct video_capability b;
			strcpy(b.name, MGAYUV_ID_STRING);
			b.type = VID_TYPE_SCALES | VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY;
			b.channels = 1;
			b.audios = 0;
			b.maxwidth = 1024;	/* FIXME: check the specs */
			b.maxheight = 768;
			b.minwidth = 32;
			b.minheight = 16;
			if(copy_to_user(arg,&b,sizeof(b)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCGFBUF:
		{
			struct video_buffer fake_fb_info;
			/*
			 * In all this driver code we use the 'base'
			 * field as the relative address of the video
			 * buffer inside the framebuffer (for
			 * convenience), whereas the standard says
			 * it's the physical address.
			 */
			fake_fb_info.width = vbuf.width;
			fake_fb_info.height = vbuf.height;
			fake_fb_info.depth = vbuf.depth;
			fake_fb_info.bytesperline = vbuf.bytesperline;

#if LINUX_VERSION_CODE >= 0x020300
			fake_fb_info.base += mga->dev->resource[0].start;
#else
			fake_fb_info.base += (mga->dev->base_address[0] &
					      PCI_BASE_ADDRESS_MEM_MASK);
#endif
			if(copy_to_user(arg, &fake_fb_info, 
					sizeof(struct video_buffer)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCSFBUF:
		{
			return -EINVAL;
		}
                case VIDIOCGCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.channel != 0)
				return -EINVAL;
			v.flags = 0;
			v.tuners = 0;
			v.type = VIDEO_TYPE_CAMERA;
			v.norm = VIDEO_MODE_AUTO;
			strcpy(v.name, "Video Overlay");
			if(copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCSCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.channel != 0)
				return -EINVAL;
			if(v.norm == VIDEO_MODE_AUTO)
				return -EINVAL;

			/*
			 * change overlay buffer dimension
			 */
			if ((v.norm == VIDEO_MODE_PAL) ||
			    (v.norm == VIDEO_MODE_SECAM)) {
				vbuf.width = 768;   /* 720 gives trouble */
				vbuf.height = 576;
			} else { /* VIDEO_MODE_NTSC */
				vbuf.width = 640;
				vbuf.height = 480;
			}
			program_bes( mga );
			
			norm = v.norm;
			return 0;
		}

		case VIDIOCSWIN:
		{
			if(copy_from_user(&vw, arg, sizeof(vw)))
				return -EFAULT;
			if(vw.x <0 || vw.y <0 || vw.width < 32 
				|| vw.height < 16)
				return -EINVAL;
			if (vw.flags == VIDEO_WINDOW_CHROMAKEY)
				mgabes_set_colorkey(bes, vw.chromakey);
			else
				mgabes_set_keymode(bes, 0);
			mgabes_set_window(bes, vw.x, vw.y, vw.width, vw.height);

			return 0;
				
		}
		case VIDIOCGPICT:
		{
			if(copy_to_user(arg, &vp, sizeof(vp)))
				return -EFAULT;
			return 0;
		}
		
		case VIDIOCGWIN:
		{
			if(copy_to_user(arg, &vw, sizeof(struct video_window)))
				return -EFAULT;
			return 0;
		}
		
		case VIDIOCSPICT:
		{
			struct video_picture nvp;
			if(copy_from_user(&nvp, arg, sizeof(nvp)))
				return -EFAULT;
				
			vp.palette = nvp.palette;
			program_bes( mga );
			return 0;
		}
		
		case VIDIOCCAPTURE:
		{
			if (arg)
				mgabes_activate(bes);
			else
				mgabes_desactivate(bes);

			return 0;
		}
		default:
			return -ENOIOCTLCMD;
	}
}

static struct video_device mgayuv_template =
{
	name:		MGAYUV_ID_STRING,
	type:		VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY | VID_TYPE_SCALES,
	hardware:	VID_HARDWARE_BT848,		/* This is a lie for now */
	open:		mgayuv_open,
	close:		mgayuv_close,
	read:		mgayuv_read,
	write:		mgayuv_write,
	poll:		NULL,
	ioctl:		mgayuv_ioctl,
	mmap:		mgayuv_mmap,
	initialize:	mgayuv_init_done,
	0,
};




/* 
 * Main Initialization Function 
 */
static int mgayuv_initialize(void)
{
	int			i;
	struct	mga_dev_s	*mga;
	
	struct mgayuv_s *myuv;

	printk(MGAYUV_ID_STRING"\n");

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		if ((myuv = kmalloc(sizeof(*myuv), GFP_KERNEL)) == NULL) {
			return -ENOMEM;
		}
		memset( myuv, 0, sizeof(*myuv) );

		memcpy(&myuv->video_dev, &mgayuv_template, sizeof(mgayuv_template));
		sprintf(myuv->video_dev.name, "mgayuv-%u", i);
		myuv->mga = mga;
		
		mga->yuv_v4l1 = myuv;

#if LINUX_VERSION_CODE >= 0x020405
		if (video_register_device(&myuv->video_dev, VFL_TYPE_GRABBER, video_nr) < 0) {
#else
		if (video_register_device(&myuv->video_dev, VFL_TYPE_GRABBER) < 0) {
#endif

			printk("mgayuv: unable to register.\n");
			kfree( myuv );
			mga_release_card(mga);
			return -EINVAL;
		}
	}

	return 0;
}

int init_module(void)
{
	return mgayuv_initialize();
}

void cleanup_module(void)
{
	int i = 0;
	struct mgayuv_s* myuv;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		myuv = mga->yuv_v4l1;
		
		video_unregister_device(&myuv->video_dev);

		/* need to release it 2 times */
		mga_release_card(myuv->mga);
		mga_release_card(myuv->mga);

		mga->yuv_v4l1 = NULL;
		kfree(myuv);
	}
}

