/* Start of file */

/* {{{ [fold] Comments */
/*
 * qce-ga, linux V4L driver for the Quickcam Express and Dexxa Quickcam
 *
 * memory.c - contains all needed memory management functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* }}} */
/* {{{ [fold] Includes */
#include <linux/config.h>
#include <linux/version.h>

#ifdef CONFIG_SMP
#define __SMP__
#endif

#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include "qc-memory.h"

#include <linux/wrapper.h>
#include <asm/io.h>
#include <linux/mm.h>		/* Required on Alpha, from Bob McElrath <mcelrath@draal.physics.wisc.edu> */
#include <asm/pgtable.h>	/* Required on Alpha */
#include <linux/vmalloc.h>	/* Required on Alpha */
#include <linux/pagemap.h>	/* pmd_offset requires this on SuSE supplied kernels */
/* }}} */
/* {{{ [fold] Compatibility wrappers */
#ifndef HAVE_VMA
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,5,3) || defined(pte_offset_map)
#define HAVE_VMA 1
#else
#define HAVE_VMA 0
#endif
#endif

#if HAVE_VMA
#define qc_remap_page_range(vma, start, addr, size, prot) remap_page_range(vma, start, addr, size, prot)
#else
#define qc_remap_page_range(vma, start, addr, size, prot) remap_page_range(start, addr, size, prot)
#endif

#if HAVE_VMA
#define pte_offset(pmd,adr) pte_offset_map(pmd,adr)	/* Emulation for a kernel using the new rmap-vm */
#endif							/* Fix by Michele Balistreri <brain87@gmx.net> */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#define virt_to_page(page)	(MAP_NR(page))
#define page_address(x)		(x | PAGE_OFFSET)
#endif
/* }}} */

#if 0
/* {{{ [fold] uvirt_to_bus(unsigned long adr) */
static inline unsigned long uvirt_to_bus(unsigned long adr)
{
	unsigned long kva, ret;

	kva = uvirt_to_kva(pgd_offset(current->mm, adr), adr);
	ret = virt_to_bus((void *) kva);

	return ret;
}
/* }}} */
/* {{{ [fold] kvirt_to_bus(unsigned long adr) */
static inline unsigned long kvirt_to_bus(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = virt_to_bus((void *) kva);

	return ret;
}
/* }}} */
#endif
/* {{{ [fold] uvirt_to_kva(pgd_t *pgd, unsigned long adr) */
/*
 * Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
	unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;

	if (!pgd_none(*pgd)) {
		pmd = pmd_offset(pgd, adr);
		if (!pmd_none(*pmd)) {
			ptep = pte_offset(pmd, adr);
			pte = *ptep;
			if (pte_present(pte)) {
				ret = (unsigned long) page_address(pte_page(pte));
				ret |= (adr & (PAGE_SIZE - 1));
			}
		}
	}
	return ret;
}
/* }}} */
/* {{{ [fold] kvirt_to_pa(unsigned long adr) */
/*
 * Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
static unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);

	return ret;
}
/* }}} */
/* {{{ [fold] qc_mm_rvmalloc(unsigned long size) */
void *qc_mm_rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr, page;

	size = PAGE_ALIGN(size);		/* Align to next page boundary */
	mem = vmalloc(size);
	if (!mem) return NULL;
	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		page = kvirt_to_pa(adr);
		mem_map_reserve(virt_to_page(__va(page)));
		adr += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	return mem;
}
/* }}} */
/* {{{ [fold] qc_mm_rvfree(void *mem, unsigned long size) */
void qc_mm_rvfree(void *mem, unsigned long size)
{
	unsigned long adr, page;

	if (!mem) return;
	size = PAGE_ALIGN(size);		/* Round it upward to the next multiple of PAGE_SIZE */
	adr = (unsigned long) mem;
	while (size > 0) {
		page = kvirt_to_pa(adr);
		mem_map_unreserve(virt_to_page(__va(page)));
		adr += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	vfree(mem);
}
/* }}} */
/* {{{ [fold] qc_mm_remap(vma, src, src_size, adr, adr_size) */
int qc_mm_remap(struct vm_area_struct *vma, void *src, unsigned long src_size, const void *dst, unsigned long dst_size)
{
	unsigned long start = (unsigned long)dst;
	unsigned long pos = (unsigned long)src;
	unsigned long page;

	if (dst_size > PAGE_ALIGN(src_size)) return -EINVAL;
	while (dst_size > 0) {
		page = kvirt_to_pa(pos);
		if (qc_remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)<0) return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (dst_size > PAGE_SIZE)
			dst_size -= PAGE_SIZE;
		else
			dst_size = 0;
	}
	return 0;
}
/* }}} */

/* End of file */
