/*
 *
 *    Connectix USB QuickCam VC Video Camera driver
 *
 *    (C) Copyright 2001 De Marchi Daniele
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of the
 *    License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Jan 2001 This driver develop started on the linux 
 *            kernel 2.4.0.
 *
 */

#include "qcamvc.h"
#include <linux/parport.h>
#include <linux/delay.h>
#include <linux/smp_lock.h>

static struct qcamvc *cam_list;

struct pp_cam_entry{
  struct parport *port;
  struct pardevice *pdev;
  
  wait_queue_head_t wq_stream;

  int open_count;
};


/* Magic numbers for defining port-device mappings */
#define QCAMVC_PARPORT_UNSPEC -4
#define QCAMVC_PARPORT_AUTO -3
#define QCAMVC_PARPORT_OFF -2
#define QCAMVC_PARPORT_NONE -1

#ifdef MODULE
static int parport_nr[PARPORT_MAX] = {[0 ... PARPORT_MAX - 1] = QCAMVC_PARPORT_UNSPEC};
static char *parport[PARPORT_MAX] = {NULL,};
MODULE_AUTHOR("De Marchi Daniele <demarchidaniele@libero.it>");
MODULE_DESCRIPTION("Connectix QuickCam VC(PP) Driver");
MODULE_PARM(parport, "1-" __MODULE_STRING(PARPORT_MAX) "s");
MODULE_PARM_DESC(parport, "'auto' or a list of parallel port numbers. Just like lp.");
#else
static int parport_nr[PARPORT_MAX] __initdata =
	{[0 ... PARPORT_MAX - 1] = QCAMVC_PARPORT_UNSPEC};
static int parport_ptr = 0;
#endif

/*****************************************************************************/
/*                                                                           */
/* Lowlevel driver for Connectix QuickCam VC(PP)                             */
/*                                                                           */
/*****************************************************************************/
static int           qcamvc_open(void *privdata);
static int           qcamvc_close(void *privdata);
static size_t        qcamvc_get_reg(void *privdata, unsigned char reg, void *buffer, size_t len);
static int           qcamvc_set_reg(void *privdata, unsigned char reg, void *buffer, size_t len);
static size_t        qcamvc_read(void *privdata, void *buffer, size_t len);
static int           qcamvc_write(void *privdata, void *buffer, size_t len);
static int           qcamvc_stream_start(void *privdata);
static int           qcamvc_stream_stop(void *privdata);
static size_t        qcamvc_stream_read(void *privdata, void *buffer, size_t len);

static struct qcamvc_camera_ops qcamvc_pp_ops = {
  qcamvc_open,
  qcamvc_close,
  qcamvc_write,
  qcamvc_read,
  qcamvc_set_reg,
  qcamvc_get_reg,
  qcamvc_stream_start,
  qcamvc_stream_stop,
  qcamvc_stream_read
};

static int qcamvc_open(void *privdata)
{
  struct pp_cam_entry *cam = privdata;
  size_t ret=0;

  parport_claim_or_block(cam->pdev);
  parport_enable_irq(cam->port);
  mdelay(10);

  MOD_INC_USE_COUNT;
  return ret;
}


static int qcamvc_close(void *privdata)
{
  struct pp_cam_entry *cam = privdata;
  size_t ret=0;

  parport_release(cam->pdev);
  parport_disable_irq(cam->port);
  mdelay(10);

  MOD_DEC_USE_COUNT;
  return ret;
}

static int qcamvc_stream_start(void *privdata)
{
  int ret = 0;
  return ret;
}

static int qcamvc_stream_stop(void *privdata)
{
  int ret = 0;
  return ret;
}

static size_t qcamvc_stream_read(void *privdata, void *buffer, size_t len)
{
  struct pp_cam_entry *cam = privdata;
  struct parport *parport = cam->port;
  size_t ret=0;
  unsigned char a = 0;
  unsigned long counter=0;
  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }
  parport_ieee1284_ecp_write_data (parport, (void *) &a, 1, 0);
  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }

  for (counter=0;counter<len;counter++)
    {
      if (current->need_resched) schedule();
      parport_ieee1284_ecp_read_data(parport, buffer + counter, 1, 0);
    }
  return ret=len;
}

static size_t qcamvc_read(void *privdata, void *buffer, size_t len)
{
  struct pp_cam_entry *cam = privdata;
  struct parport *parport = cam->port;
  size_t ret=0;
  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }
  parport_ieee1284_ecp_read_data (parport, buffer, len, 0);

  return ret = len;
}

static int qcamvc_write(void *privdata, void *buffer, size_t len)
{
  struct pp_cam_entry *cam = privdata;
  struct parport *parport = cam->port;  
  int ret=0;
  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }
  parport_ieee1284_ecp_write_data (parport, buffer, len, 0);

  return ret = len;
}

static int qcamvc_set_reg(void *privdata, unsigned char reg, void *buffer, size_t len)
{
  struct pp_cam_entry *cam = privdata;
  struct parport *parport = cam->port;
  int ret=0;

  if(parport_negotiate(cam->port, IEEE1284_ADDR | IEEE1284_MODE_ECP) != 0)
    {
    printk(KERN_INFO "QuickCamVC:WCN1 ieee1284 failed negotiation\n");
    return -1;
    }
  parport_ieee1284_ecp_write_addr(parport, &reg, 1, 0);

  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }
  parport_ieee1284_ecp_write_data (parport, buffer, len, 0);
  return ret = len;
}

static size_t qcamvc_get_reg(void *privdata, unsigned char reg, void *buffer, size_t len)
{
  struct pp_cam_entry *cam = privdata;
  struct parport *parport = cam->port;

  if(parport_negotiate(cam->port, IEEE1284_ADDR | IEEE1284_MODE_ECP) != 0)
    {
    printk(KERN_INFO "QuickCamVC:WCN1 ieee1284 failed negotiation\n");
    return -1;
    }
  parport_ieee1284_ecp_write_addr(parport, &reg, 1, 0);

  if(parport_negotiate(parport, IEEE1284_DATA | IEEE1284_MODE_ECP) != 0)
    {
      printk(KERN_INFO "QuickCamVC: ieee1284 failed negotiation\n");
      return -1;
    }
  parport_ieee1284_ecp_read_data (parport, buffer, len, 0);
  return len;
}

static int qcamvc_pp_close(void *privdata)
{
	struct pp_cam_entry *cam = privdata;
	if (--cam->open_count == 0) {
		parport_release(cam->pdev);
	}
	return 0;
}

static int qcamvc_register(struct parport *port) 
{
  struct pardevice *pdev = NULL;
  struct pp_cam_entry *cam;
  struct qcamvc *qcamvc;
  
  if (!(port->modes & PARPORT_MODE_ECP) &&
      !(port->modes & PARPORT_MODE_TRISTATE)) {
    LOG("port is not ECP capable\n");
    return -ENXIO;
  }
  cam = kmalloc(sizeof(struct pp_cam_entry), GFP_KERNEL);
  if (cam == NULL) {
    LOG("failed to allocate camera structure\n");
    return -ENOMEM;
  }
  memset(cam,0,sizeof(struct pp_cam_entry));
  
  pdev = parport_register_device(port, "Connectix QuickCam VC", NULL, NULL,
				 NULL, 0, cam);
  
  if (!pdev) {
    LOG("failed to parport_register_device\n");
    kfree(cam);
    return -ENXIO;
  }
  
  cam->pdev = pdev;
  cam->port = port;
  init_waitqueue_head(&cam->wq_stream);

  parport_claim_or_block(cam->pdev);
  if((qcamvc = qcamvc_register_camera(&qcamvc_pp_ops, cam,"PP QuickCam VC")) == NULL)
    {
      LOG("failed to qcamvc_register_camera\n");
      parport_release(cam->pdev);
      parport_unregister_device(pdev);
      kfree(cam);
      return -ENXIO;
    }
  
  parport_release(cam->pdev);
  ADD_TO_LIST(cam_list, qcamvc);
  
  return 0;
}

static void qcamvc_pp_detach(struct parport *port)
{
  struct qcamvc *qcamvc = NULL;
  
  for(qcamvc = cam_list; qcamvc != NULL; qcamvc = qcamvc->next) {
    struct pp_cam_entry *cam = (struct pp_cam_entry *) qcamvc->lowlevel_data;
    if ((cam != NULL) && (cam->port->number == port->number)) {
      
      if(cam->open_count > 0) 
	{
	  qcamvc_pp_close(cam);
	}
      qcamvc_unregister_camera(qcamvc);
      parport_unregister_device(cam->pdev);

      REMOVE_FROM_LIST(qcamvc);     
      kfree(cam);
      qcamvc->lowlevel_data = NULL;
      break;
    }
  }
}

static void qcamvc_pp_attach(struct parport *port)
{
  unsigned int i;
  
  switch (parport_nr[0])
    {
    case QCAMVC_PARPORT_UNSPEC:
    case QCAMVC_PARPORT_AUTO:
      if (port->probe_info[0].class != PARPORT_CLASS_MEDIA ||
	  strncmp(port->probe_info[0].mfr, "Connectix", 9) != 0 ||
	  strncmp(port->probe_info[0].model, "QuickCam VC", 11) != 0)
	{
	  return;
	}
      qcamvc_register(port);     
      break;
      
    default:
      for (i = 0; i < PARPORT_MAX; ++i) {
	if (port->number == parport_nr[i]) {
	  qcamvc_register(port);
	  break;
	}
      }
      break;
    }
}

static struct parport_driver qcamvc_pp_driver = {
	"Connectix QuickCam VC",
	qcamvc_pp_attach,
	qcamvc_pp_detach,
	NULL
};

int qcamvc_pp_init(void)
{
  if(parport_nr[0] == QCAMVC_PARPORT_OFF) {
    printk("  disabled\n");
    return 0;
  }
  
  if (parport_register_driver (&qcamvc_pp_driver)) {
    printk("unable to register with parport\n");
    return -EIO;
  }
  
  cam_list = NULL;
  return 0;
}

#ifdef MODULE
int qcamvc_pp_init_module (void)
{
  printk("QuickCam VC(PP): (C) 2001 by De Marchi Daniele, <demarchidaniele@libero.it>\n");
  printk("QuickCam VC(PP): lowlevel driver version %s registered.\n",QCAMVCVERSION);

  if (parport[0]) {
    /* The user gave some parameters.  Let's see what they were. */
    if (!strncmp(parport[0], "auto", 4)) {
      parport_nr[0] = QCAMVC_PARPORT_AUTO;
    } else {
      int n;
      for (n = 0; n < PARPORT_MAX && parport[n]; n++) {
	if (!strncmp(parport[n], "none", 4)) {
	  parport_nr[n] = QCAMVC_PARPORT_NONE;
	} else {
	  char *ep;
	  unsigned long r = simple_strtoul(parport[n], &ep, 0);
	  if (ep != parport[n]) {
	    parport_nr[n] = r;
	  } else {
	    printk("bad port specifier `%s'\n", parport[n]);
	    return -ENODEV;
	  }
	}
      }
    }
  }
#if defined(CONFIG_KMOD) && defined(CONFIG_PNP_PARPORT_MODULE)
  if(parport_enumerate() && !parport_enumerate()->probe_info.model) {
    request_module("parport_probe");
  }
#endif
  
  return qcamvc_pp_init();
}


void qcamvc_pp_exit_module(void)
{
  parport_unregister_driver (&qcamvc_pp_driver);
}

module_init(qcamvc_pp_init_module);
module_exit(qcamvc_pp_exit_module);

#else /*MODULE*/

static int __init qcamvc_pp_setup(char *str)
{
	if (!strncmp(str, "parport", 7)) {
		int n = simple_strtoul(str + 7, NULL, 10);
		if (parport_ptr < PARPORT_MAX) {
			parport_nr[parport_ptr++] = n;
		} else {
			LOG("too many ports, %s ignored.\n", str);
		}
	} else if (!strcmp(str, "auto")) {
		parport_nr[0] = QCAMVC_PARPORT_AUTO;
	} else if (!strcmp(str, "none")) {
		parport_nr[parport_ptr++] = QCAMVC_PARPORT_NONE;
	}

	return 0;
}

__setup("qcamvc_pp=", qcamvc_pp_setup);

#endif /* !MODULE */
