/*-*- linux-c -*-
 *  linux/drivers/video/i810_accel.c -- Hardware Acceleration
 *
 *      Copyright (C) 2001 Antonino Daplas
 *      All Rights Reserved      
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,14)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#include <linux/tty.h>
#include "i810_regs.h"
#include "i810_common.h"
#include "i810_accel.h"

/* BLT Engine Routines */
static inline void i810_report_error(void)
{
	printk("IIR     : 0x%04x\n"
	       "EIR     : 0x%04x\n"
	       "PGTBL_ER: 0x%04x\n"
	       "IPEIR   : 0x%04x\n"
	       "IPEHR   : 0x%04x\n",
	       i810_readw(IIR),
	       i810_readb(EIR),
	       i810_readl(PGTBL_ER),
	       i810_readl(IPEIR), 
	       i810_readl(IPEHR));
}

/**
 * wait_for_space - check ring buffer free space
 * @space: amount of ringbuffer space needed in bytes
 *
 * DESCRIPTION:
 * The function waits until a free space from the ringbuffer
 * is available 
 */	
inline int wait_for_space(u32 space)
{
	u32 head, count = WAIT_COUNT, tail;

	tail = i810_accel->cur_tail;
	while (count--) {
		head = i810_readl(IRING + 4) & RBUFFER_HEAD_MASK;	
		if ((tail == head) || 
		    (tail > head && (RINGBUFFER_SIZE - tail + head) >= space) || 
		    (tail < head && (head - tail) >= space)) {
			return 0;	
		}
	}
	printk("ringbuffer lockup!!!\n");
	i810_report_error(); 
	i810_accel->lockup = 1;
	i810_set_iface_lockup();
	return 1;
}



/** 
 * wait_for_engine_idle - waits for all hardware engines to finish
 *
 * DESCRIPTION:
 * This waits for lring(0), iring(1), and batch(3), etc to finish
 */
inline int wait_for_engine_idle(void)
{
	int count = WAIT_COUNT;

	while((i810_readw(INSTDONE) & 0x7B) != 0x7B && count--); 
	if (count) return 0;

	printk("accel engine lockup!!!\n");
	printk("INSTDONE: 0x%04x\n", i810_readl(INSTDONE));
	i810_report_error(); 
	i810_accel->lockup = 1;
	i810_set_iface_lockup();
	return 1;
}

/* begin_iring - prepares the ringbuffer 
 * @space: length of sequence in dwords
 *
 * DESCRIPTION:
 * Checks/waits for sufficent space in ringbuffer of size
 * space.  Returns the tail of the buffer
 */ 
inline u32 begin_iring(u32 space)
{
	if (i810_accel->sync) wait_for_engine_idle();
	return wait_for_space(space);
}

/**
 * end_iring - advances the tail, which begins execution
 *
 * DESCRIPTION:
 * This advances the tail of the ringbuffer, effectively
 * beginning the execution of the instruction sequence.
 */
inline void end_iring(void)
{
	u32 tail;

	tail = i810_readl(IRING) & ~RBUFFER_TAIL_MASK;
	i810_writel(IRING, tail | i810_accel->cur_tail);
	if (i810_accel->sync) wait_for_engine_idle();
}

/**
 * flush_gfx - flushes graphics pipeline
 *
 * DESCRIPTION:
 * Flushes the graphics pipeline.  This is done via
 * sending instruction packets to the ringbuffer
 */
inline void flush_gfx(void)
{
	if (begin_iring(8 + IRING_PAD)) return;
	PUT_RING(PARSER | FLUSH);
	PUT_RING(NOP);
	end_iring();
}

inline void arb_on_off(u32 mode)
{
	if (begin_iring(8 + IRING_PAD)) return;
	PUT_RING(PARSER | ARB_ON_OFF | mode);
	PUT_RING(NOP);
	end_iring();
}

/**
 * emit_instruction - process instruction packets
 * @dsize: length of instruction packets in dwords
 * @pointer: pointer to physical address of buffer 
 * @trusted: whether the source of the instruction came
 *           from a trusted process or not (root)
 *
 * DESCRIPTION:
 * This function initiates the ringbuffer, copies all packets
 * to the ringbuffer, and initiates processing of the packets.
 * This function is reserved for non-kernel clients doing
 * DMA's in userland
 */
inline void emit_instruction (u32 dsize, u32 pointer, u32 trusted)
{
	if (begin_iring(16 + IRING_PAD)) return;
	PUT_RING(PARSER | BATCH_BUFFER | 1);
	PUT_RING(pointer | trusted);
	PUT_RING(pointer + (dsize << 2) - 4);
	PUT_RING(NOP);	 
	end_iring();
}

/**
 * mono_pat_blit - monochromatic pattern BLIT
 * @dpitch: no of pixel per line of destination 
 * @dheight:  height in pixels of the pattern
 * @dwidth: width in pixels of the pattern
 * @where: where to place the pattern
 * @fg: foreground color
 * @bg: background color
 * @rop: graphics raster operation
 * @patt_1: first part of an 8x8 pixel pattern
 * @patt_2: second part of an 8x8 pixel pattern
 * @blit_bpp: pixel format which can be different from the 
 *            buffers pixel format
 *
 * DESCRIPTION:
 * Immediate monochromatic pattern BLIT function.  The pattern
 * is directly written to the ringbuffer.
 */

inline void mono_pat_blit(int dpitch, int dheight, int dwidth, int dest, 
			  int fg, int bg, int rop, int patt_1, int patt_2, 
			  int blit_bpp)
{
	if (begin_iring(32 + IRING_PAD)) return;
	PUT_RING(BLIT | MONO_PAT_BLIT | 6);
	PUT_RING(rop << 16 | dpitch | DYN_COLOR_EN | blit_bpp);
	PUT_RING(dheight << 16 | dwidth);
	PUT_RING(dest);
	PUT_RING(bg);
	PUT_RING(fg);
	PUT_RING(patt_1);
	PUT_RING(patt_2);
	end_iring();
}

/**
 * source_copy_blit - BLIT transfer operation
 * @dwidth: width of rectangular graphics data
 * @dheight: height of rectangular graphics data
 * @dpitch: pixels per line of destination buffer
 * @xdir: direction of copy (left to right or right to left)
 * @spitch: pixels per line of source buffer
 * @from: source address
 * @where: destination address
 * @rop: raster operation
 * @blit_bpp: pixel format which can be different from the 
 *            buffers pixel format
 *
 * DESCRIPTION:
 * This is a BLIT operation typically used when doing
 * a 'Copy and Paste'
 */
inline void source_copy_blit(int dwidth, int dheight, int dpitch, int xdir, 
			     int spitch, int src, int dest, int rop, int blit_bpp)
{
	if (begin_iring(24 + IRING_PAD)) return;
	PUT_RING(BLIT | SOURCE_COPY_BLIT | 4);
	PUT_RING(xdir | rop << 16 | dpitch | DYN_COLOR_EN | blit_bpp);
	PUT_RING(dheight << 16 | dwidth);
	PUT_RING(dest);
	PUT_RING(spitch);
	PUT_RING(src);
	end_iring();
}	

/**
 * color_blit - solid color BLIT operation
 * @width: width of destination
 * @height: height of destination
 * @pitch: pixels per line of the buffer
 * @where: destination
 * @rop: raster operation
 * @what: color to transfer
 * @blit_bpp: pixel format which can be different from the 
 *            buffers pixel format
 *
 * DESCRIPTION:
 * A BLIT operation which can be used for  color fill/rectangular fill
 */
inline void color_blit(int width, int height, int pitch, 
		       int dest, int rop, int what, int blit_bpp)
{
	if (begin_iring(24 + IRING_PAD)) return;
	PUT_RING(BLIT | COLOR_BLT | 3);
	PUT_RING(rop << 16 | pitch | SOLIDPATTERN | DYN_COLOR_EN | blit_bpp);
	PUT_RING(height << 16 | width);
	PUT_RING(dest);
	PUT_RING(what);
	PUT_RING(NOP);
	end_iring();
}

/* i810fb_load_front - initializes frontbuffer
 * @offset: offset from start of framebuffer memory
 * @pitch: pitch of buffer
 * @async: asynchronous or synchronous flips
 *
 * DESCRIPTION:
 * similar to load_start, but uses ringbuffer to transfer the instructions.
 * This can be useful if asynchronous flips are required.
 */
int i810fb_load_front(u32 offset, u32 pitch, u32 async)
{
	u32 header;
	
	if (i810_accel->lockup || not_safe()) return -ENODEV;
	header = PARSER | FRONT_BUFFER | async << 6;
	if (!async)
		header |= (pitch >> 3) << 8;
	flush_gfx();
	if (begin_iring(8 + IRING_PAD)) return -ENODEV;
	PUT_RING(header);
	PUT_RING((i810_accel->fb_offset << 12) + offset);
	end_iring();
	return 0;
}

/* i810fb_load_back - initializes backbuffer
 * @pitch_bits: pitch of rendering surface
 *
 * DESCRIPTION:
 * Required if asynchronous buffer flips are done.  
 */
void i810fb_load_back(u32 pitch_bits)
{
	if (i810_accel->lockup || not_safe()) return;
	flush_gfx();
	if (begin_iring(8 + IRING_PAD)) return;
	PUT_RING(PARSER | DEST_BUFFER);
	PUT_RING(i810_accel->fb_offset << 12 | pitch_bits);
	end_iring();
}

/**
 * i810fb_iring_enable - enables/disables the ringbuffer
 * @mode: enable or disable
 *
 * DESCRIPTION:
 * Enables or disables the ringbuffer, effectively enabling or
 * disabling the instruction/acceleration engine.
 */
inline void i810fb_iring_enable(u32 mode)
{
	u32 tmp;
	tmp = i810_readl(IRING + 12);
	if (mode == OFF) 
		tmp &= ~1;
	else 
		tmp |= 1;
	wait_for_engine_idle();
	flush_cache();
	i810_writel(IRING + 12, tmp);
}       

/**
 * i810fb_init_ringbuffer - initialize the ringbuffer
 *
 * DESCRIPTION:
 * Initializes the ringbuffer by telling the device the
 * size and location of the ringbuffer.  It also sets 
 * the head and tail pointers = 0
 */
void i810fb_init_ringbuffer(void)
{
	u32 tmp1, tmp2;
	
	wait_for_engine_idle();
	i810fb_iring_enable(OFF);
	i810_writel(IRING, 0);
	i810_writel(IRING + 4, 0);
	i810_accel->cur_tail = 0;

	tmp2 = i810_readl(IRING + 8) & ~RBUFFER_START_MASK; 
	tmp1 = i810_accel->iring_start_phys;
	i810_writel(IRING + 8, tmp2 | tmp1);

	tmp1 = i810_readl(IRING + 12);
	tmp1 &= ~RBUFFER_SIZE_MASK;
	tmp2 = (RINGBUFFER_SIZE - I810_PAGESIZE) & RBUFFER_SIZE_MASK;
	i810_writel(IRING + 12, tmp1 | tmp2);
	i810fb_iring_enable(ON);
}

/**
 * i810fb_restore_ringbuffer - restores saved ringbuffers
 * @iring: pointer to the ringbuffe structure
 */
void i810fb_restore_ringbuffer(struct ringbuffer *iring)
{
	u32 tmp1, tmp2;
	wait_for_engine_idle();
	i810fb_iring_enable(OFF);
	
	i810_writel(IRING, 0);
	i810_writel(IRING + 4, 0);
	
	tmp1 = i810_readl(IRING + 8);
	tmp1 &= ~RBUFFER_START_MASK;
	tmp2 = iring->start;
	tmp2 &= RBUFFER_START_MASK;
	i810_writel(IRING + 8, tmp1 | tmp2);

	tmp1 = i810_readl(IRING + 12);
	tmp1 &= ~RBUFFER_SIZE_MASK;
	tmp2 = iring->size;
	tmp2 &= RBUFFER_SIZE_MASK;
	i810_writel(IRING + 12, tmp1 | tmp2);
	
	tmp1 = iring->size;
       	i810fb_iring_enable(tmp1 & 1);
}

void i810fb_unbind_accel_mem(void)
{
	if (i810_accel->i810_iring_memory->is_bound)
		agp_unbind_memory(i810_accel->i810_iring_memory);
}

int i810fb_bind_accel_mem(void)
{
	if (!i810_accel->i810_iring_memory->is_bound) {
		if (agp_bind_memory(i810_accel->i810_iring_memory, 
				    i810_accel->iring_offset)) {
			printk("i180fb: can't rebind command buffer memory\n");
			return -EBUSY;
		}
	}
	return 0;
}

u32 __devinit i810fb_alloc_iring_mem(u32 fb_offset, u32 aper_size)
{
	if (fb_offset < RINGBUFFER_SIZE >> 12)
		i810_accel->iring_offset = ((aper_size - RINGBUFFER_SIZE) >> 12) - 1;
	else
		i810_accel->iring_offset = fb_offset - (RINGBUFFER_SIZE >> 12);
	if (!(i810_accel->i810_iring_memory = 
	      agp_allocate_memory(RINGBUFFER_SIZE >> 12, 
				  AGP_NORMAL_MEMORY))) {
		printk("i810fb_alloc_iringmem:  cannot allocate ringbuffer memory\n");
		return 0;
	}	
	if (agp_bind_memory(i810_accel->i810_iring_memory, 
			    i810_accel->iring_offset)) {
		printk("i810fb_alloc_iringmem: cannot bind ringbuffer memory\n");
		return 0;
	}	
	return i810_accel->iring_offset;
}

void __devinit i810fb_fix_accel_pointer(u32 fb_base_phys, u32 fb_base_virtual)
{
	i810_accel->iring_start_phys = fb_base_phys + 
		(i810_accel->iring_offset << 12);
	i810_accel->iring_start_virtual = fb_base_virtual + 
		(i810_accel->iring_offset << 12);
}

u32 __devinit i810fb_init_accel(u32 fb_offset, u32 aper_size, u32 sync)
{
	if (!(i810_accel = kmalloc(sizeof(struct accel_data), GFP_KERNEL)))
		return -ENOMEM;
	memset(i810_accel, 0, sizeof(struct accel_data));
	i810_accel->fb_offset = fb_offset;
	i810_accel->sync = sync;
	return i810fb_alloc_iring_mem(fb_offset, aper_size);
}

void i810fb_accel_cleanup(void)
{
	if (i810_accel->i810_iring_memory) {
		i810fb_iring_enable(OFF);
		agp_free_memory(i810_accel->i810_iring_memory);
	}
	if (i810_accel)
		kfree(i810_accel);
}
