/*-*- linux-c -*-
 *  linux/drivers/video/i810_fbcon_accel.c -- Accelereted Console Drawing Functions
 *
 *      Copyright (C) 2001 Antonino Daplas
 *      All Rights Reserved      
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/console.h>
#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>

#include "i810_regs.h"
#include "i810_common.h"
#include "i810_fbcon_accel.h"

/* 
 * Helper functions
 */

inline int i810_accel8_getbg(struct display *p, int c)
{
	return attr_bgcol(p, c);
}

inline int i810_accel8_getfg(struct display *p, int c)
{
	return attr_fgcol(p, c);
}

inline int i810_accel16_getbg(struct display *p, int c)
{
	return (int) ((u16 *) p->dispsw_data)[attr_bgcol(p, c)];
}

inline int i810_accel16_getfg(struct display *p, int c)
{
	return (int) ((u16 *) p->dispsw_data)[attr_fgcol(p, c)];
}

inline int i810_accel24_getbg(struct display *p, int c)
{
	return ((int *) p->dispsw_data)[attr_bgcol(p, c)];
}

inline int i810_accel24_getfg(struct display *p, int c)
{
	return ((int *) p->dispsw_data)[attr_fgcol(p, c)];
}
void i810_accel_setup(struct display *p)
{
	p->next_line = p->line_length ? p->line_length : 
		get_line_length(i810_orient->vxres, p->var.bits_per_pixel);
	p->next_plane = 0;
}

/*
 * Start of actual drawing functions
 */
void i810_accel_bmove(struct display *p, int sy, int sx, 
		      int dy, int dx, int height, int width)
{
	int pitch, xdir, src, dest, mult, blit_bpp = 0;
	
	if (i810_accel->lockup || not_safe())
		return;

	dy *= fontheight(p);
	sy *= fontheight(p);
	height *= fontheight(p);
 	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	switch (mult) {
	case 1:
		blit_bpp = BPP8;
		break;
	case 2:
		blit_bpp = BPP16;
		break;
	case 3:
		blit_bpp = BPP24;
		break;
	}

	sx *= fontwidth(p) * mult;
	dx *= fontwidth(p) * mult;
	width *= fontwidth(p) * mult;			

	if (dx <= sx) 
		xdir = INCREMENT;
	else {
		xdir = DECREMENT;
		sx += width - 1;
		dx += width - 1;
	}
	if (dy <= sy) 
		pitch = p->next_line;
	else {
		pitch = (-(p->next_line)) & 0xFFFF; 
		sy += height - 1;
		dy += height - 1;
	}
	src = (i810_accel->fb_offset << 12) + (sy * p->next_line) + sx; 
	dest = (i810_accel->fb_offset << 12) + (dy * p->next_line) + dx;
	source_copy_blit(width, height, pitch, xdir, 
			 pitch, src, dest, PAT_COPY_ROP, blit_bpp);
}

void i810_accel_putc(struct vc_data *conp, struct display *p, 
		     int c, int yy, int xx)
{
	u8 *cdat;
	int dheight, dwidth, dpitch, fg = 0, bg = 0;
	int dest, fontwidth, mult, blit_bpp = 8;
	int i, j, cols, rows, pattern = 0;

	if (i810_accel->lockup || not_safe())
		return;

	/* 8x8 pattern blit */
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	dheight = 8;              
	dwidth = mult << 3;
	dpitch = p->next_line;
	switch (mult) {
	case 1:
		fg = (int) attr_fgcol(p,c);
		bg = (int) attr_bgcol(p,c);
		blit_bpp = BPP8;
		break;
	case 2:
		bg = (int) ((u16 *)p->dispsw_data)[attr_bgcol(p, c)];
		fg = (int) ((u16 *)p->dispsw_data)[attr_fgcol(p, c)];
		blit_bpp = BPP16;
		break;
	case 3:
		fg = ((int *) p->dispsw_data)[attr_fgcol(p, c)];
		bg = ((int *) p->dispsw_data)[attr_bgcol(p, c)];
		blit_bpp = BPP24;
	}
	rows = fontheight(p) >> 3;
	if(fontwidth(p) <= 8) {
		fontwidth = 8;
		cdat = p->fontdata + ((c & p->charmask) * fontheight(p));
		cols = 1;
	}
	else {
		fontwidth = fontwidth(p);
		cdat = p->fontdata + 
			((c & p->charmask) * (fontwidth >> 3) * fontheight(p));
		cols = fontwidth >> 3;
	}
	dest = (i810_accel->fb_offset << 12) + 
		(yy * fontheight(p) * p->next_line) + 
		xx * (fontwidth * mult);

	for (i = cols; i--; ) {
		for (j = 0; j < rows; j++) {
			pattern = j << (cols + 2);
			mono_pat_blit(dpitch, dheight, dwidth, 
				      dest + (j * (dpitch << 3)),
				      fg, bg, COLOR_COPY_ROP, 
				      *(u32 *)(cdat + pattern), 
				      *(u32 *)(cdat + pattern + (cols << 2)),
				      blit_bpp);
		}
		cdat += 4;
		dest += dwidth;
	}
}


void i810_accel_putcs(struct vc_data *conp, struct display *p,
			const unsigned short *s, int count, int yy, int xx)
{
	u8 *cdat, *cdat0;
	int chars, dpitch, dheight, dwidth, dest, next;
	int mult, fontwidth, blit_bpp = 0, cols, rows;
	int c, dest0, i, j;
	int (*get_bg)(struct display *p, int c) = NULL;
	int (*get_fg)(struct display *p, int c) = NULL;
	
	if (i810_accel->lockup || not_safe())
		return;

	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	switch (mult) {
	case 1:
		get_bg = i810_accel8_getbg;
		get_fg = i810_accel8_getfg;
		blit_bpp = BPP8;
		break;
	case 2:
		get_bg = i810_accel16_getbg;
		get_fg = i810_accel16_getfg;
		blit_bpp = BPP16;
		break;
	case 3:	
		get_bg = i810_accel24_getbg;
		get_fg = i810_accel24_getfg;
		blit_bpp = BPP24;
	}
	
	dwidth = mult << 3;
	dheight = 8;
	dpitch = p->next_line;
	rows = fontheight(p) >> 3;

	if(fontwidth(p) <= 8) {
		cols = 1;
		fontwidth = 8;
	}
	else {
		cols = fontwidth(p) >> 3;
		fontwidth = fontwidth(p);
	}

	dest = (i810_accel->fb_offset << 12) + 
		(yy * fontheight(p) * p->next_line) + 
		(xx * (fontwidth * mult));
	chars = p->next_line/fontwidth;
	next = 1 << (cols + 2);
	while(count--) {
		c = scr_readw(s++);
		cdat = p->fontdata + ((c & p->charmask)*cols*fontheight(p));
		for (i = cols; i--; ) {
			dest0 = dest;
			cdat0 = cdat;
			for(j = rows; j--; ) {
				mono_pat_blit(dpitch, dheight, dwidth, dest0, 
					      get_fg(p, c), get_bg(p, c), 
					      COLOR_COPY_ROP, 
					      *((u32 *) (cdat0)),
					      *((u32 *) (cdat0 + (cols << 2))),
					      blit_bpp);
				dest0 += dpitch << 3;      	      
				cdat0 += next;
			}
			dest += dwidth;
			cdat += 4;
		}
		if(++xx > chars) {
			xx = 0;
			yy++;
			dest = (i810_accel->fb_offset << 12) + 
				(yy * fontheight(p) * p->next_line);
		}
		
	}	 
}

void i810_accel_revc(struct display *p, int xx, int yy)
{
	int width, height, dest, pitch, mult, blit_bpp = 0;

	if (i810_accel->lockup || not_safe())
		return;

	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	switch (mult) {
	case 1:
		blit_bpp = BPP8;
		break;
	case 2:
		blit_bpp = BPP16;
		break;
	case 3:
		blit_bpp = BPP24;
		break;
	}
	pitch = p->next_line;
	if (fontwidth(p) <= 8)
		width = 8*mult;
	else
		width = fontwidth(p)*mult;
	height = fontheight(p);
	dest = (i810_accel->fb_offset << 12) + 
		(yy * height * p->next_line) + 
		(xx * width);
	color_blit(width, height, pitch, dest, INVERT_ROP, 0x0F, blit_bpp);
}

void i810_accel_clear(struct vc_data *conp, struct display *p, 
		      int sy, int sx, int height, int width)
{
	int mult, blit_bpp = 0;
	u32 bgx = 0, dest = 0;
	
	if (i810_accel->lockup || not_safe())
		return;

	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	height *= fontheight(p);
	sy *= fontheight(p);

	switch(mult) {
	case 1:
		bgx = (u32) attr_bgcol_ec(p, conp);
		blit_bpp = BPP8;
		break;
	case 2:
		bgx = (int) ((u16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
		blit_bpp = BPP16;
		break;
	case 3:
		bgx = ((int *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
		blit_bpp = BPP24;
	}	         	         

	width *= fontwidth(p) * mult;
	sx *= fontwidth(p) * mult;
	dest = (i810_accel->fb_offset << 12) + 
		(sy * p->next_line) + sx;
	color_blit(width, height, p->next_line, dest, 
		   COLOR_COPY_ROP, bgx, blit_bpp);
}

void i810_accel_clear_margins(struct vc_data *conp, struct display *p,
			       int bottom_only)
{
	int bytes = p->next_line, blit_bpp = 0;
	u32 bgx = 0, mult;
	unsigned int right_start;
	unsigned int bottom_start;
	unsigned int right_width, bottom_height;

	if (i810_accel->lockup || not_safe())
		return;

	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;

	switch(mult) {
	case 1:
		bgx = (u32) attr_bgcol_ec(p, conp);
		blit_bpp = BPP8;
		break;
	case 2:
		bgx = (int) ((u16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
		blit_bpp = BPP16;
		break;
	case 3:
		blit_bpp = BPP24;
		bgx = ((int *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	}	         	         

	right_width = p->var.xres % fontwidth(p);
	right_start = p->var.xres - right_width;
	if (!bottom_only && right_width) 
		color_blit(right_width*mult, p->var.yres_virtual, bytes, 
			   (i810_accel->fb_offset << 12) + 
			   ((right_start+p->var.xoffset)*mult), 
			   COLOR_COPY_ROP, bgx, blit_bpp); 
	bottom_height = p->var.yres % fontheight(p);
	if (bottom_height) {
		bottom_start = p->var.yres - bottom_height;
		color_blit(right_start*mult, bottom_height, bytes, 
			   (i810_accel->fb_offset << 12) + 
			   (p->var.yoffset+bottom_start)*bytes, 
			   COLOR_COPY_ROP, bgx, blit_bpp);
	}
}

struct display_switch i810_accel_upright = {
	i810_accel_setup,
	i810_accel_bmove,
	i810_accel_clear,
	i810_accel_putc,
	i810_accel_putcs,
	i810_accel_revc,
	NULL,
	NULL,
	i810_accel_clear_margins,
	FONTWIDTH(8) | FONTWIDTH(16)
};

