/*-*- linux-c -*-
 *  linux/drivers/video/i810_main.c -- Intel 810 frame buffer device
 *
 *      Copyright (C) 2001 Antonino Daplas
 *      All Rights Reserved      
 *
 *      Contributors:
 *         Michael Vogt <mvogt@acm.org> - added support for Intel 815 chipsets
 *                                        and enabling the power-on state of 
 *                                        external VGA connectors for 
 *                                        secondary displays
 *
 *	The code framework is a modification of vfb.c by Geert Uytterhoeven.
 *      DotClock and PLL calculations are partly based on i810_driver.c 
 *              in xfree86 v4.0.3 by Precision Insight.
 *      Watermark calculation and tables are based on i810_wmark.c 
 *              in xfre86 v4.0.3 by Precision Insight.  Slight modifications 
 *              only to allow for integer operations instead of floating point.
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,14)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#include <linux/vmalloc.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/resource.h>
#include <linux/selection.h>
#include <linux/console.h>
#include <linux/vt_kern.h>
#include <linux/unistd.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/div64.h>

#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif 

#include <asm/page.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>

#include "i810_regs.h"
#include "i810_common.h"
#include "i810_main.h"

/* 
 * IOCTL operations
 */

/**
 * i810fb_release_gart - release GART (pagetable) for use by others
 *
 * DESCRIPTION:
 * If the graphics device needs to be acquired by another app, it can
 * request the i810fb driver (via ioctl I810FB_IOC_CLAIMGART).  All
 * memory currently bound will be unbound and the backend released.
 */

static void i810fb_release_gart(void)
{
	i810_info->gart_is_claimed = 1;
	i810_info->in_context = 0;
	wait_for_engine_idle();
	if (i810_info->i810_gtt.i810_fb_memory->is_bound)
		agp_unbind_memory(i810_info->i810_gtt.i810_fb_memory);
	i810fb_unbind_accel_mem();
	if (hwcur && i810_info->i810_gtt.i810_cursor_memory->is_bound) 
		agp_unbind_memory(i810_info->i810_gtt.i810_cursor_memory);
	i810fb_unbind_iface_mem();
	i810fb_restore_regs();   
	agp_backend_release();
}

/**
 * i810fb_bind_all - bind all unbound memory
 *
 * DESCRIPTION:
 * This attempts to bind all the drivers gtt memory, and all gtt memory
 * requested by clients.  Failure to bind the driver's memory will exit 
 * with an -EBUSY. 
 */
int i810fb_bind_all(void)
{
	int err = 0;

	if (!i810_info->i810_gtt.i810_fb_memory->is_bound) {
		if (agp_bind_memory(i810_info->i810_gtt.i810_fb_memory, 
				    i810_info->fb_offset)) { 
                        /* if we reach this, somethng bad just happened */
			printk("i810fb: cannot rebind framebuffer memory\n");
			return -EBUSY;
		}	
	}
	if ((err = i810fb_bind_accel_mem()))
		return err;

	if (hwcur && !i810_info->i810_gtt.i810_cursor_memory->is_bound) {
		if (agp_bind_memory(i810_info->i810_gtt.i810_cursor_memory, 
				    i810_info->cursor_offset)) {
			printk("i810fb: can't rebind cursor memory\n");
			return -EBUSY;
		}
	}
	return ((err = (i810fb_bind_iface_mem()))) ? err : 0;
}
/**
 * i810fb_reacquire_gart - reacquire the GART for use by the framebuffer
 *
 * DESCRIPTION:
 * An application which is currently holding the graphics device can tell
 * the framebuffer driver (via IOCTL I810FB_IOC_RELEASEGART) that it can
 * have control of the graphics device.  This should be done if the app 
 * needs to switch to the console or is exiting. 
 */
int i810fb_reacquire_gart(void)
{
	wait_for_engine_idle();
	i810fb_save_regs();   

	/* We'll do what X is doing, disregard the -EBUSY return 
	   of agp_backend_acquire */
	agp_backend_acquire();
	if (i810fb_bind_all()) 
		agp_backend_release();
	else 
		i810_info->gart_is_claimed = 0;
	return 0;
}


static struct fb_ops i810fb_ops = {
    THIS_MODULE, 
    i810fb_open, 
    i810fb_release, 
    i810fb_get_fix, 
    i810fb_get_var, 
    i810fb_set_var, 
    i810fb_get_cmap, 
    i810fb_set_cmap, 
    i810fb_pan_display, 
    i810fb_ioctl, 
    i810fb_mmap
};



    /*
     *  Open/Release the frame buffer device
     */

static int i810fb_open(struct fb_info *info, int user)
{
	if (i810_info->gart_is_claimed)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	return(0);                              
}
        
static int i810fb_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return(0);                                                    
}


    /*
     *  Get the Fixed Part of the Display
     */

static int i810fb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info)
{
	struct fb_var_screeninfo *var;
	if (con == -1)
		var = &i810fb_default;
	else
		var = &fb_display[con].var;
	i810fb_encode_fix(fix, var);
	return 0;
}

    /*
     *  Get the User Defined Part of the Display
     */

static int i810fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
	if (con == -1)
		*var = i810fb_default;
	else
		*var = fb_display[con].var;
	set_color_bitfields(var);
	return 0;
}

    /*
     *  Round off values to capability of hardware
     */
static  void i810fb_round_off(struct fb_var_screeninfo *var)
{
	/*
	 *  Presently supports only these configurations 
	 */
	round_off_xres(var);
	if (var->xres < 8)
		var->xres = 8;
	if (var->xres >= 1600) 
		var->xres = 1600;
	var->xres = (var->xres + 7) & ~0x07;
	if (var->xres_virtual < var->xres) 
		var->xres_virtual = var->xres;

	round_off_yres(var);
	if (var->yres < 8)
		var->yres = 8;
	if (var->yres >= 1600)
		var->yres = 1600;
	var->yres = (var->yres + 7) & ~0x07;
	if (var->yres_virtual < var->yres) 
		var->yres_virtual = var->yres;

	if (var->bits_per_pixel < 16)
		var->bits_per_pixel = 8;
	else if (var->bits_per_pixel < 24) 
		var->bits_per_pixel = 16;
	else if (var->bits_per_pixel < 32)
		var->bits_per_pixel = 24;
	else 
		var->bits_per_pixel = 32;

	if ((accel && var->accel_flags) && 
	     (var->bits_per_pixel < 24 ||
	      (!render && var->bits_per_pixel < 32))) {
		switch (rotate) {
		case NO_ROTATION:
		case ROTATE_180:
			i810_orient->xres_var = var->xres;
			i810_orient->yres_var = var->yres;
			i810_orient->vxres_var = var->xres_virtual;
			i810_orient->vyres_var = var->yres_virtual;
			break;
		case ROTATE_RIGHT:
		case ROTATE_LEFT:
			i810_orient->xres_var = var->yres;
			i810_orient->yres_var = var->xres;
			i810_orient->vxres_var = var->yres_virtual;
			i810_orient->vyres_var = var->xres_virtual;
		}
		i810_orient->rotate_var = rotate;
	}
	else {
		i810_orient->xres_var = var->xres;
		i810_orient->yres_var = var->yres;
		i810_orient->vxres_var = var->xres_virtual;
		i810_orient->vyres_var = var->yres_virtual;
		i810_orient->rotate_var = NO_ROTATION;
	}
	
}	
    /*
     * check if parameters passed to var are valid
     */
static int i810fb_check_params(struct fb_var_screeninfo *var, 
			       struct fb_info *info)
{
	int line_length, vidmem;

	/*
	 *  Memory limit
	 */
	line_length = get_line_length(i810_orient->vxres_var, 
				      var->bits_per_pixel);
	vidmem = line_length*i810_orient->vyres_var;
	if (vidmem > i810_info->fb_size || !vidmem) {
		printk("i810fb: required video memory, %dMB, is out of range\n",
		       vidmem/1000000);
		return -ENOMEM;
	}
	switch (i810_orient->rotate_var) {
	case NO_ROTATION:
	case ROTATE_180:
		var->xres_virtual = line_length/(var->bits_per_pixel >> 3);
		i810_orient->vxres_var = var->xres_virtual;
		break;
	case ROTATE_RIGHT:
	case ROTATE_LEFT:
		var->yres_virtual = line_length/(var->bits_per_pixel >> 3);
		i810_orient->vxres_var = var->yres_virtual;
	}

			
	/*
	 * Monitor limit
	 */
	return (i810fb_get_timings(info, i810_timings, i810_orient->xres_var,
				   i810_orient->yres_var)) ? -EINVAL : 0; 
}	
     /*
      * Update parameters from var once they pass the initial check
      */
static void i810fb_set_params(struct fb_var_screeninfo *var)
{
		
	i810_orient->xres = i810_orient->xres_var;
	i810_orient->yres = i810_orient->yres_var;
	i810_orient->vxres = i810_orient->vxres_var;
	i810_orient->vyres = i810_orient->vyres_var;
	i810_orient->rotate = i810_orient->rotate_var;

	if (render) {
		switch (i810_orient->vxres * (var->bits_per_pixel >> 3)) {
		case 0 ... 512:
			i810_info->i810_gtt.tile_pitch = 2 << 4;
			i810_info->i810_gtt.pitch_bits = 0;
			break;
		case 513 ... 1024:
			i810_info->i810_gtt.tile_pitch = 3 << 4 ;
			i810_info->i810_gtt.pitch_bits = 1;
			break;
		case 1025 ... 2048:
			i810_info->i810_gtt.tile_pitch = 4 << 4;
			i810_info->i810_gtt.pitch_bits = 2;
			break;
		case 2049 ... 4096:
			i810_info->i810_gtt.tile_pitch = 5 << 4;
			i810_info->i810_gtt.pitch_bits = 3;
			break;
		}
	}
	set_color_bitfields(var);
}

     /*
      *  Set the User Defined Part of the Display
      */

static int i810fb_set_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info)
{
	int err, activate = var->activate;
	struct display *display;
	struct fb_var_screeninfo vartmp;
	if (con >= 0)
		display = &fb_display[con];
	else
	    display = &i810_info->disp;	/* used during initialization */
	
	/*
	 *  FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
	 *  as FB_VMODE_SMOOTH_XPAN is only used internally
	 */

	if (var->vmode & FB_VMODE_CONUPDATE) {
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = display->var.xoffset;
		var->yoffset = display->var.yoffset;
	}
	vartmp = display->var;
	i810fb_round_off(var);
	if ((err = i810fb_check_params(var, info))) 
		return err;
	i810fb_set_params(var);
	if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		display->var = *var;
		info->var = *var;
		if (vartmp.xres != var->xres || 
		    vartmp.yres != var->yres ||
		    vartmp.xres_virtual != var->xres_virtual || 
		    vartmp.yres_virtual != var->yres_virtual ||
		    vartmp.bits_per_pixel != var->bits_per_pixel || 
		    vartmp.accel_flags != var->accel_flags) {
			struct fb_fix_screeninfo fix;
			
			i810fb_encode_fix(&fix, var);
			display->screen_base = 
				(char *) i810_info->fb_start_virtual;
			display->visual = fix.visual;
			display->type = fix.type;
			display->type_aux = fix.type_aux;
			display->ypanstep = fix.ypanstep;
			display->ywrapstep = fix.ywrapstep;
			/* Choose the most efficient scrolling method */
			if (var->yres_virtual > var->yres) 
				display->scrollmode = SCROLL_YNOMOVE;
			else
				display->scrollmode = SCROLL_YREDRAW;
			display->line_length = fix.line_length;
			display->can_soft_blank = 1;
			display->inverse = 0;

			switch (var->bits_per_pixel) {
			case 8:
				set_var8(display, var);
				break;
			case 16:
				set_var16(display, var);
				break;
			case 24:
				set_var24(display, var);
				break;
			case 32:
				set_var32(display, var);
				break;
			default:
				display->dispsw = &fbcon_dummy;
				break;
			}
			if (i810_info->fb_info.changevar)
				(*i810_info->fb_info.changevar)(con);
		}
		if (con < currcon)
			return 0;
/*		if (vartmp.bits_per_pixel != var->bits_per_pixel) { */
			if ((err = fb_alloc_cmap(&display->cmap, 0, 0))) 
				return err;
			do_install_cmap(con, info);
/*		} */
		i810_calc_dclk(&i810_info->mode_params, i810_timings->pixelclock);
		i810fb_fill_vga_registers(info, display);

		i810fb_load_regs(display);
		i810fb_update_display(display, con, info);
		if (hwcur) {
			i810fb_init_cursor();
			i810fb_set_cursor(display);
		}

	}
	return 0;
}

    /*
     *  Pan or Wrap the Display
     *
     *  This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
     */

static int i810fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
	if (var->vmode & FB_VMODE_YWRAP) {
		if (var->yoffset < 0 ||
		    var->yoffset >= fb_display[con].var.yres_virtual ||
		    var->xoffset)
			return -EINVAL;
	} else {
		if (var->xoffset+fb_display[con].var.xres >
		    fb_display[con].var.xres_virtual ||
		    var->yoffset+fb_display[con].var.yres >
		    fb_display[con].var.yres_virtual)
			return -EINVAL;
	}
	fb_display[con].var.xoffset = var->xoffset;
	fb_display[con].var.yoffset = var->yoffset;
	i810fbcon_updatevar(con, info);
	if (var->vmode & FB_VMODE_YWRAP)
		fb_display[con].var.vmode |= FB_VMODE_YWRAP;
	else
		fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;
	return 0;
}

    /*
     *  Get the Colormap
     */
 
static int i810fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
	if (con == currcon) /* current console? */
		return fb_get_cmap(cmap, kspc, i810fb_getcolreg, info);
	else if (fb_display[con].cmap.len) /* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else 
		fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel),
			     cmap, kspc ? 0 : 2);
	return 0;
}

    /*
     *  Set the Colormap
     */

static int i810fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
	int err;
	
	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		if ((err = fb_alloc_cmap(&fb_display[con].cmap,
					 1<<fb_display[con].var.bits_per_pixel,
					 0)))
			return err;
	}
	if (con == currcon)			/* current console? */
		return fb_set_cmap(cmap, kspc, i810fb_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}

    /*
     *  i810 Frame Buffer Specific ioctls
     */

static int i810fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		     u_long arg, int con, struct fb_info *info)
{
	agp_mem_user agp_mem;
	i810_command command;
	int err = 0;
	
	switch(cmd) {
	case I810FB_IOC_COMMAND:
		if (copy_from_user(&command, (void *) arg, sizeof(command))) 
			return -EFAULT;
		return i810fb_process_command(&command);
	case I810FB_IOC_REQUESTAGPMEM:
		if (copy_from_user(&agp_mem, (void *) arg, sizeof(agp_mem)))
			return -EFAULT;
		if ((err = i810fb_allocate_agpmemory(&agp_mem)))
			return err;
		return (copy_to_user((void *) arg, &agp_mem, 
				     sizeof(agp_mem))) ? -EFAULT : 0;
	case I810FB_IOC_RELEASEAGPMEM:
		if (copy_from_user(&agp_mem, (void *) arg, sizeof(agp_mem)))
			return -EFAULT;
		return i810fb_free_agpmemory(&agp_mem);
	case I810FB_IOC_ACQUIREFB:
		if (i810_info->gart_is_claimed && 
		    !i810_info->gart_countdown_active)
			return -EINVAL;
		if (-1 == (err = i810fb_acquire_fb()))
			return -EINVAL;
		put_user(err, (int *) arg);
		return 0;
	case I810FB_IOC_AREYOUTHERE:
		return 0;
	case I810FB_IOC_CLAIMGART:
		i810fb_release_gart();
		return 0;
	case I810FB_IOC_RELEASEGART:
		return 	i810fb_reacquire_gart();
	default:
		return -EINVAL;
	}
}

/**
 * i810fb_mmap - mmap framebuffer, mmio and off-screen surface
 * @info: pointer to fb_info
 * @file: file descriptor
 * @vma: virtual memory area structure
 *
 * DESCRIPTION:
 * This is a specialized mmap service for the i810fb.  Aside from memory 
 * mapping the framebuffer and MMIO space, it also allows mapping of 
 * off-screen surfaces for use as DMA/FIFO buffers which have been previously 
 * allocated via I810FB_IOC_REQUESTAGPMEM ioctl. The format to map the 
 * off-screen surface is fix->smem_len + fix->mmio_len + surface offset.  
 * The "surface offset" is returned by the previous ioctl call.
 */
static int i810fb_mmap(struct fb_info *info, struct file *file, 
		       struct vm_area_struct *vma)
{
	u32 off, start;
	u32 len;

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;
	off = vma->vm_pgoff << PAGE_SHIFT;
	start =i810_info->fb_start_phys;
	len = PAGE_ALIGN((start & ~PAGE_MASK) + i810_info->fb_size);
	if (off >= len && off < len + MMIO_SIZE) {
		/* memory mapped io */
		off -= len;
		if (info->var.accel_flags)
			return -EINVAL;
		start = i810_info->mmio_start_phys;
		len = PAGE_ALIGN((start & ~PAGE_MASK) + MMIO_SIZE);
	}
	else if (off >= len + MMIO_SIZE && off < len + MMIO_SIZE + i810_info->aper_size) { 
		/* Client off-screen memory */
		off -= len + MMIO_SIZE;
		if (info->var.accel_flags)
			return -EINVAL;
		if (!(len = i810fb_check_agp_mmap(off >> 12, file->f_owner.uid)))
			return -EINVAL;
		start = i810_info->fb_base_phys + off;
		len += PAGE_ALIGN(start & ~PAGE_MASK);
		off = 0;
	}
	else if (off >= len + MMIO_SIZE + i810_info->aper_size) {
		/* sarea */
		off -= len + MMIO_SIZE + i810_info->aper_size;
		if (info->var.accel_flags)
			return -EINVAL;
		if (!(len = i810fb_check_sarea(off >> 12, file->f_owner.uid)))
			return -ENODEV;
		start = i810fb_get_sarea_start();
		len += PAGE_ALIGN(start & ~PAGE_MASK);
		off = 0;
	}
	start &= PAGE_MASK;
	if ((vma->vm_end - vma->vm_start + off) > len) 
		return -EINVAL;
	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;
#if defined(__i386__) || defined(__x86_64__)
 	if (boot_cpu_data.x86 > 3)
		pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
#endif 
	return (io_remap_page_range(vma->vm_start, off, 
				vma->vm_end - vma->vm_start, 
				vma->vm_page_prot)) ?
		-EAGAIN : 0;
}


int __init i810fb_setup(char *options)
{
	char *this_opt;
	
	fontname[0] = '\0';
	i810fb_initialized = 1;
	if (!options || !*options)
		return 0;
	
	for (this_opt = strtok(options, ":"); this_opt;
	     this_opt = strtok(NULL, ":")) {
		if (!strncmp(this_opt, "font=", 5))
			strcpy(fontname, this_opt+5);
		else if (!strncmp(this_opt, "mtrr", 4))
			mtrr = 1;
		else if (!strncmp(this_opt, "accel", 5))
			accel = 1;
		else if (!strncmp(this_opt, "ext_vga", 7))
			ext_vga = 1;
		else if (!strncmp(this_opt, "hwcur", 5))
			hwcur = 1;
		else if (!strncmp(this_opt, "sync_on_pan", 11))
			sync_on_pan = 1;
		else if (!strncmp(this_opt, "sync", 4))
			sync = 1;
		else if (!strncmp(this_opt, "vram=", 5))
			vram = (simple_strtoul(this_opt+5, NULL, 0));
		else if (!strncmp(this_opt, "xres=", 5))
			xres = simple_strtoul(this_opt+5, NULL, 0);
		else if (!strncmp(this_opt, "yres=", 5))
			yres = simple_strtoul(this_opt+5, NULL, 0);
		else if (!strncmp(this_opt, "vyres=", 6))
			vyres = simple_strtoul(this_opt+6, NULL, 0);
		else if (!strncmp(this_opt, "bpp=", 4))
			bpp = simple_strtoul(this_opt+4, NULL, 0);
		else if (!strncmp(this_opt, "hsync1=", 7))
			hsync1 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "hsync2=", 7))
			hsync2 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "vsync1=", 7))
			vsync1 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "vsync2=", 7))
			vsync2 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "render", 6))
			render = 1;
		else if (!strncmp(this_opt, "rotate=", 7))
			rotate = simple_strtoul(this_opt+7, NULL, 0);
	}
	return 0;
}


/* Internal  routines */
/* 
 * Helper inline functions
 */
inline u8 i810_readb(u32 where)
{
        return readb(i810_info->mmio_start_virtual + where);
}

inline u16 i810_readw(u32 where)
{
	return readw(i810_info->mmio_start_virtual + where);
}

inline u32 i810_readl(u32 where)
{
	return readl(i810_info->mmio_start_virtual + where);
}

inline void i810_writeb(u32 where, u8 val)
{
	writeb(val, i810_info->mmio_start_virtual + where);
}

inline void i810_writew(u32 where, u16 val)
{
	writew(val, i810_info->mmio_start_virtual + where);
}

inline void i810_writel(u32 where, u32 val)
{
	writel(val, i810_info->mmio_start_virtual + where);
}

static inline u8 is_cga(void)
{	
	return i810_readb(MSR_READ) & 1;
}

static inline void i810_wait_for_scan_start(void)
{
	u32 count = WAIT_COUNT;

	while((i810_readw(DISP_SL) & 0xFFF) && count--);
}

static inline void vsync_is_active(u32 port)
{
	u32 count = WAIT_COUNT;

	while ((i810_readb(port) & 0x08) && count--);
}

static inline void vsync_is_inactive(u32 port)
{
	u32 count = WAIT_COUNT;

	while (!(i810_readb(port) & 0x08) && count--);
}

static inline void i810_wait_for_hsync(void)
{
	u32 count = WAIT_COUNT;

	while((i810_readw(DISP_SL) & 0xFFF) < 5 && count--);
}
	
static inline void i810_wait_for_vsync(int count)
{
	u32 port;	

	if (!count) return;

	if (is_cga) port = ST01_CGA;
	else port =  ST01_MDA;
		
	while(count--) {
		 vsync_is_active(port);
		 vsync_is_inactive(port);
	}
}       

/* Internal Routines */

/**
 * i810fb_screen_off - turns off/on display
 * @mode: on or off
 *
 * DESCRIPTION:
 * Blanks/unblanks the display
 */
static void i810fb_screen_off(u8 mode)
{
	u8 val;

	i810_writeb(SR_INDEX, SR01);
	val = i810_readb(SR_DATA);
	if (mode == OFF) 
		val |= SCR_OFF;
	else
		val &= ~SCR_OFF;

	i810_wait_for_scan_start(); 
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, val);
}


/**
 * i810fb_dram_off - turns off/on dram refresh
 * @mode: on or off
 *
 * DESCRIPTION:
 * Turns off DRAM refresh.  Must be off for only 2 vsyncs
 * before data becomes corrupt
 */
static void i810fb_dram_off(u8 mode)
{
	u8 val;

	val = i810_readb(DRAMCH);
	if (mode == OFF)
		val &= DRAM_OFF;
	else {
		val &= DRAM_OFF;
		val |= DRAM_ON;
	}
	i810_writeb(DRAMCH, val);
}

/**
 * i810fb_protect_regs - allows rw/ro mode of certain VGA registers
 * @mode: protect/unprotect
 *
 * DESCRIPTION:
 * The IBM VGA standard allows protection of certain VGA registers.  
 * This will  protect or unprotect them. 
 */
static void i810fb_protect_regs(int mode)
{
	u32 dataport;
	u32 indexport;
	u8 reg;

	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR11);
	reg = i810_readb(dataport);
	if (mode == OFF)
		reg &= ~0x80;
	else 
		reg |= 0x80;
 		
	i810_writeb(indexport, CR11);
	i810_writeb(dataport, reg);
}

/**
 * i810fb_get_mem_freq - get RAM BUS frequency
 *
 * DESCRIPTION:
 * Determines if RAM bus frequency is 100 or 133 MHz. Writes the result
 * to info->mem_freq
 */
static void __devinit i810fb_get_mem_freq(void)
{
	u8 reg;
	pci_read_config_byte(i810_info->i810_gtt.i810_gtt_info.device, 
			     0x50, &reg);
	reg &= FREQ_MASK;
	if (reg)
		i810_info->mem_freq = 133;
	else
		i810_info->mem_freq = 100;
}

/* Best to avoid floating point calculations in the kernel ...	*/
/**
 * i810_calc_dclk - calculates the P, M, and N values or a pixelclock value
 * @params: pointer to structure of video registervalues
 * @freq: pixclock to calculate
 *
 * DESCRIPTION:
 * Based on the formula Freq_actual = (4*M*Freq_ref)/(N^P)
 * Repeatedly computes the Freq until the actual Freq is equal to
 * the target Freq or until the loop count is zero.  In the latter
 * case, the actual frequency nearest the target will be used.
 * M, N, P registers are write to the params structure.
 */
static void i810_calc_dclk(struct mode_registers *params, u32 freq)
{
	u32 m_reg, n_reg, p_divisor, n_target_max;
	u32 m_target, n_target, p_target, val_min, n_best, m_best;
	u32 f_out, ref_freq, target_freq, diff = 0, target_val;

	n_best = m_best = m_target = f_out = 0;
	ref_freq = 24000;
	/* Make target frequency divisible by 125000 */
	target_freq =  ((((freq + 999)/1000) + 124)/125) * 125;
	n_target_max = 30;
	target_val = 0xFFFF;
	val_min = 0xFFFF;

        /* Let's get a P divisor such that 
	 * (16 * ref_freq)/(target_freq * p)  < 1 
	 */
	p_divisor = 1;
	p_target = 0;
	while(((32 * ref_freq)/(target_freq * p_divisor)) && p_divisor < 32) {
		p_divisor <<= 1;
		++p_target;
	}

	n_reg = m_reg = n_target = 3;	
	while ((target_val) && (n_target < n_target_max)) {
		f_out = (4 * m_reg * ref_freq)/(n_reg * p_divisor);
		m_target = m_reg;
		n_target = n_reg;
		if (f_out <= target_freq) {
			++m_reg;
			diff = target_freq - f_out;
		}
		else {
			++n_reg;
			diff = f_out - target_freq;
		}
		if (diff > 10000) 
			target_val = 0xFFFF;
		else 
			target_val = (diff*10000)/target_freq;
		if (val_min > target_val) {
			val_min = target_val;
			n_best = n_target;
			m_best = m_target;
		}		 

	} 
	params->M = (m_best - 2) & 0x3FF;
	params->N = (n_best - 2) & 0x3FF;
	params->P = (p_target << 4);
	params->pixclock = f_out;
}

/**
 * i810fb_get_vblank - get vertical blank time
 * @hfreq: horizontal freq
 *
 * DESCRIPTION:
 * vblank = front porch + sync + back porch 
 *    where: front porch = 1;
 *           sync + back porch = flyback * hfreq
 *                               ---------------
 *                                  1000000
 *           and flyback is set to 550
 */
static u32 i810fb_get_vblank(u32 hfreq)
{
	u32 vblank;

	vblank = (hfreq * FLYBACK)/1000; 
	vblank = (vblank + 500)/1000;
	return (vblank + V_FRONTPORCH);
}

/** 
 * i810fb_get_hblank - get horizontal blank time
 * @hfreq: horizontal freq
 * @xres: horizontal resolution in pixels
 *
 * DESCRIPTION:
 * duty cycle = C - (M/Hfreq)
 * where: C = ((offset - scale factor) * blank_scale)
 *            -------------------------------------- + scale factor
 *                        256 
 *        M = blank_scale * gradient
 */
static u32 i810fb_get_hblank(u32 hfreq, u32 xres)
{
	u32 c_val, m_val, duty_cycle, hblank;

	c_val = (((H_OFFSET - H_SCALEFACTOR) * H_BLANKSCALE)/256 + H_SCALEFACTOR) * 1000;
	m_val = (H_BLANKSCALE * H_GRADIENT)/256;
	m_val = (m_val * 1000000)/hfreq;
	duty_cycle = c_val - m_val;
	hblank = (xres * duty_cycle)/(100000 - duty_cycle);
	hblank = (hblank + 4) & ~7;
	return (hblank);
}

/**
 * i810fb_estimate_hfreq - estimate hsync
 * @vfreq: vertical refresh rate
 * @yres: vertical resolution
 *
 * DESCRIPTION:
 * Based on:
 *
 * (yres + front_port) * vfreq * 1000000
 * -------------------------------------
 * (1000000 - (vfreq * FLYBACK)
 * 
 */

static u32 i810fb_estimate_hfreq(u32 vfreq, u32 yres)
{
	u64 hfreq;
	u32 divisor;
	
	divisor = 1000000 - (vfreq * FLYBACK);
	
	hfreq = (u64) (yres + V_FRONTPORCH) *  
		(u64) (vfreq)  * 1000000;
	do_div(hfreq, divisor);
	
	return ((u32) hfreq);
}


/**
 * i810fb_get_timings - calculate video timings 
 * @info: pointer to fb_info structure
 * @timings: pointer to video_timings structure
 * @xres: horizontal resolution
 * @yres: vertical resolution
 * 
 * DESCRIPTION:
 * Calculates necessary timing information based on 
 * monitor specifications.  This will use the 
 * VESA general timing foruula
 */                      
static int i810fb_get_timings(struct fb_info *info, 
			      struct video_timings *timings,
			      int xres, int yres)
{
	u32 htotal = 0, vtotal, hfreq, vfreq, hblank, vblank, dclk;

	memset(timings, 0, sizeof(struct video_timings));
	hfreq = info->monspecs.hfmax;
	vblank = i810fb_get_vblank(hfreq);
	vtotal = yres + vblank;
	vfreq = hfreq/vtotal;
	if (vfreq > info->monspecs.vfmax) { 
		vfreq = info->monspecs.vfmax;
		hfreq = i810fb_estimate_hfreq(vfreq, yres);
		vblank = i810fb_get_vblank(hfreq);
		vtotal = yres + vblank;
	}	
	hblank = i810fb_get_hblank(hfreq, xres);
	htotal = xres + hblank;
	dclk = htotal * hfreq;
	while(dclk > MAX_PIXELCLOCK         &&  
	      hfreq >= info->monspecs.hfmin &&
	      vfreq >= info->monspecs.vfmin)   {
		hfreq -= 1000;
		vblank = i810fb_get_vblank(hfreq);
		vtotal = yres + vblank;
		vfreq = hfreq/vtotal;
		hblank = i810fb_get_hblank(hfreq, xres);
		htotal = xres + hblank;
		dclk = hfreq * htotal;
	} 
	if (vfreq < info->monspecs.vfmin) {
		printk("i810fb: required vertical refresh, %dHz, "
		       "for %dx%d is out of range\n",
		       vfreq, xres, yres);
		return -1;
	}
	if (hfreq < info->monspecs.hfmin) {
		printk("i810fb: required horizontal sync frequency, %dKHz, "
		       "for %dx%d is out of range\n", hfreq/1000, xres, yres);
		return -1;
	}
	if (dclk < MIN_PIXELCLOCK) {
		printk("i810fb: required pixelclock, %dMHz, for %dx%d"
		       " is out of range\n", dclk/1000000, xres, yres);
		return -1;
	}
	timings->htotal = htotal;
	timings->vtotal = vtotal;
	timings->vfreq = vfreq;
	timings->hfreq = hfreq;
	timings->hblank = hblank;
	timings->vblank = vblank;
	timings->pixelclock = dclk;
	return 0;
}


/**
 * i810fb_load_pll
 * @disp: pointer to display structure
 *
 * DESCRIPTION:
 * Loads the P, M, and N registers.  
 * USES:
 * info->mode_params
 */
static void i810fb_load_pll(struct display *disp)
{
 	u32 tmp1, tmp2;
	
	tmp1 = i810_info->mode_params.M | i810_info->mode_params.N << 16;
	tmp2 = i810_readl(DCLK_2D);
	tmp2 &= ~MN_MASK;
	i810_writel(DCLK_2D, tmp1 | tmp2);
	
	tmp1 = i810_info->mode_params.P;
	tmp2 = i810_readl(DCLK_0DS);
	tmp2 &= ~(P_OR << 16);
	i810_writel(DCLK_0DS, (tmp1 << 16) | tmp2);

	i810_writeb(MSR_WRITE, 0xC8);

}

/**
 * i810fb_load_vga - load standard VGA registers
 *
 * DESCRIPTION:
 * Load values to VGA registers
 */
static void i810fb_load_vga(void)
{	
	u32 indexport, dataport;
      
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR00);
	i810_writeb(dataport, i810_info->mode_params.cr00);
	i810_writeb(indexport, CR01);
	i810_writeb(dataport, i810_info->mode_params.cr01);
	i810_writeb(indexport, CR02);
	i810_writeb(dataport, i810_info->mode_params.cr02);
	i810_writeb(indexport, CR03);
	i810_writeb(dataport, i810_info->mode_params.cr03);
	i810_writeb(indexport, CR04);
	i810_writeb(dataport, i810_info->mode_params.cr04);
	i810_writeb(indexport, CR05);
	i810_writeb(dataport, i810_info->mode_params.cr05);
	i810_writeb(indexport, CR06);
	i810_writeb(dataport, i810_info->mode_params.cr06);
	i810_writeb(indexport, CR09);
	i810_writeb(dataport, i810_info->mode_params.cr09);
	i810_writeb(indexport, CR10);
	i810_writeb(dataport, i810_info->mode_params.cr10);
	i810_writeb(indexport, CR11);
	i810_writeb(dataport, i810_info->mode_params.cr11);
	i810_writeb(indexport, CR12);
	i810_writeb(dataport, i810_info->mode_params.cr12);
	i810_writeb(indexport, CR15);
	i810_writeb(dataport, i810_info->mode_params.cr15);
	i810_writeb(indexport, CR16);
	i810_writeb(dataport, i810_info->mode_params.cr16);
}


/**
 * i810fb_load_vgax - load extended VGA registers
 *
 * DESCRIPTION:
 * Load values to extended VGA registers
 */

static void i810fb_load_vgax(void)
{
	u32 indexport, dataport;

	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR30);
	i810_writeb(dataport, i810_info->mode_params.cr30);
	i810_writeb(indexport, CR31);
	i810_writeb(dataport, i810_info->mode_params.cr31);
	i810_writeb(indexport, CR32);
	i810_writeb(dataport, i810_info->mode_params.cr32);
	i810_writeb(indexport, CR33);
	i810_writeb(dataport, i810_info->mode_params.cr33);
	i810_writeb(indexport, CR35);
	i810_writeb(dataport, i810_info->mode_params.cr35);
	i810_writeb(indexport, CR39);
	i810_writeb(dataport, i810_info->mode_params.cr39);
}

/**
 * i810fb_load_2d - load grahics registers
 * @wm: watermark
 *
 * DESCRIPTION:
 * Load values to graphics registers
 */

static void i810fb_load_2d(u32 wm)
{
	u32 tmp;

  	i810_writel(FW_BLC, wm); 
	tmp = i810_readl(PIXCONF);
	tmp |= 1;
	i810_writel(PIXCONF, tmp);
	i810fb_write_sarea_reg(); 

}	

/**
 * i810fb_hires - enables high resolution mode
 */
static void i810fb_hires(void)
{
	u32 indexport, dataport;
	u8 val;
	
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}


	i810_writeb(indexport, CR80);
	val = i810_readb(dataport);
	i810_writeb(indexport, CR80);
	i810_writeb(dataport, val | 1);
}

/**
 * i810fb_load_fbstart - loads pointer of the framebuffer
 * @offset: where to start display from the start of framebuffer
 */
static void i810fb_load_fbstart(int offset)
{
	u32 addr, indexport, dataport, vidmem;
	u8 val;

	vidmem = i810_info->fb_start_phys + offset;
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	
	addr = ((vidmem + 2) >> 2) & 0xFF;
	i810_writeb(indexport, CR0D);
	i810_writeb(dataport, (u8) addr);
	
	addr = (vidmem >> 10) & 0xFF;
	i810_writeb(indexport, CR0C);
	i810_writeb(dataport, (u8) addr);
	
	addr = (vidmem >> 18) & 0x3F;
	i810_writeb(indexport, CR40);
	val = i810_readb(dataport) & ~0x3F;
	i810_writeb(indexport, CR40);
	i810_writeb(dataport, (u8) addr | val);
	
	addr = vidmem >> 24;
	i810_writeb(indexport, CR42);
	i810_writeb(dataport, (u8) addr);
	
	i810_writeb(indexport, CR40);
	val = i810_readb(dataport) | 0x80;
	i810_writeb(indexport, CR40);
	i810_writeb(dataport, val);
}
	
/**
 * i810fb_load_pitch - loads the characters per line of the display
 * @var: pointer to fb_var_screeninfo
 *
 * DESCRIPTION:
 * Loads the characters per line
 */	
static void i810fb_load_pitch(struct fb_var_screeninfo *var)
{

	u32 tmp, indexport, dataport;
	u32 line_length = get_line_length(i810_orient->vxres, 
					  var->bits_per_pixel) >> 3;
	u8 val;
			
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}

	i810_writeb(SR_INDEX, SR01);
	val = i810_readb(SR_DATA);
	val &= 0xE0;
	val |= 0x01;
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, val);

	tmp = line_length & 0xFF;
	i810_writeb(indexport, CR13);
	i810_writeb(dataport, (u8) tmp);
	
	tmp = line_length >> 8;
	i810_writeb(indexport, CR41);
	val = i810_readb(dataport) & ~0x0F;
	i810_writeb(indexport, CR41);
	i810_writeb(dataport, (u8) tmp | val);
}

/**
 * i810fb_load_color - loads the color depth of the display
 * @var: pointer to fb_var_screeninfo
 *
 * DESCRIPTION:
 * Loads the color depth of the display and the graphics engine
 */
static void i810fb_load_color(struct fb_var_screeninfo *var)
{
	u32 reg1;
	u16 reg2;
	reg1 = i810_readl(PIXCONF) & ~0xF0000;
	reg2 = i810_readw(BLTCNTL) & ~0x30;
	switch(var->bits_per_pixel) {
	case 8:
		reg1 |= 0x20000;
		break;
	case 16: 
		reg1 |= 0x50000;
		reg2 |= 0x10;
		break;
	case 24:
		reg1 |= 0x60000;
		reg2 |= 0x20;
		break;
	case 32:
		reg1 |= 0x70000;
		reg2 |= 0x20;
		break;
	}
	reg1 |= 0x8000;  
	i810_writel(PIXCONF, reg1);
	i810_writew(BLTCNTL, reg2);
}

static void i810fb_load_fence(void)
{
	u32 fence, fence_bit;

	fence_bit = i810_info->i810_gtt.fence_size;
	fence = i810_info->fb_offset << 12;
	fence |= TILEWALK_X | i810_info->i810_gtt.fence_size << 8 |
                 i810_info->i810_gtt.tile_pitch | 1;
        i810_writel(FENCE, fence);
}         


/**
 * i810fb_load_regs - loads all registers for the mode
 * @disp: pointer to display structure
 * 
 * DESCRIPTION:
 * Loads registers
 */
 
static void i810fb_load_regs(struct display *disp)
{
	i810fb_screen_off(OFF);
	i810fb_protect_regs(OFF);
	i810fb_dram_off(OFF);
	i810fb_load_pll(disp);
	i810fb_load_vga();
	i810fb_load_vgax();
	i810fb_dram_off(ON);	
	i810fb_load_2d(i810fb_get_watermark(&disp->var));
	i810fb_hires();
	i810fb_screen_off(ON);
	i810fb_protect_regs(ON);
	if (render) i810fb_load_fence();
}


/**
 * i810fb_update_display - initialize display
 * @disp: pointer to display structure
 */
static void i810fb_update_display(struct display *disp, int con, 
			    struct fb_info *info)
{
	i810fb_load_color(&disp->var);
	i810fb_load_pitch(&disp->var); 
	if (!i810_info->gart_is_claimed)
		i810fb_init_ringbuffer();
	i810fbcon_updatevar(con, info);	
}

/*
 * Hardware Cursor Routines
 */

/**
 * i810fb_load_cursor_image - create cursor bitmap
 * @p: pointer to display structure
 * 
 * DESCRIPTION:
 * Creates the cursor bitmap using the 2bpp, 2plane
 * with transparency format of the i810.  The first plane
 * is the actual monochrome bitmap, and the second plane is 
 * the transparency bit. This particular cursor is a 
 * rectangular block, a fontwidth wide and 2 scanlines high, 
 * starting just below the fontbase.  
 */
static void i810fb_load_cursor_image(struct display *p)
{
 	int i, j, h, w;
 	u8 *addr;
 	
	h = fontheight(p);
	w = fontwidth(p);
	(u32)addr = i810_info->cursor_start_virtual;
	for (i = 64; i--; ) {
		for (j = 0; j < 8; j++) {
			addr[j] = 0xFF;               /* transparent - yes */
			addr[j+8] = 0x00;             /* use background  0x00*/
		}	
		addr +=16;
	}
	
	switch (i810_orient->rotate) {
	case NO_ROTATION:
		(u32)addr = i810_info->cursor_start_virtual + ((h-1) << 4);	
		for (i = 0; i < 2; i++) {
			for (j = 0; j < (w >> 3); j++) {
				addr[(j+0) << 3] = 0x00;        /* transparent - no */
				addr[(j+1) << 3] = 0xFF;        /* use foreground */
				addr +=16;
			}
		}
		break;
	case ROTATE_RIGHT:
		(u32) addr = i810_info->cursor_start_virtual;
		for (i = 0; i < w; i++) {
			addr[0] = (~(3 << 6)) & 0xFF;
			addr[8] = 3 << 6;
			addr += 16;
		}	
		break;
	case ROTATE_180:
		(u32)addr = i810_info->cursor_start_virtual;	
		for (i = 0; i < 2; i++) {
			for (j = 0; j < (w >> 3); j++) {
				addr[(j+0) << 3] = 0x00;        /* transparent - no */
				addr[(j+1) << 3] = 0xFF;        /* use foreground */
				addr +=16;
			}
		}
		break;
	case ROTATE_LEFT:
		(u32) addr = i810_info->cursor_start_virtual + ((h-8) >> 3);
		for (i = 0; i < w; i++) {
			addr[0] = ~3;
			addr[8] = 3;
			addr += 16;
		}	
		break;
	}
	flush_cache();
}	 			

/**
 * i810fb_set_cursor_color - set the cursor CLUT
 * @p: pointer to display structure
 *
 * DESCRIPTION:
 * The i810 has two DACS, the standard 256-index CLUT and the
 * alternate 8-index CLUT. The alternate CLUT is where the
 * cursor gets the color information.
 */
static void i810fb_set_cursor_color(struct display *p)
{
	u32 fgr, fgg, fgb, fgt;
	u32 bgr, bgg, bgb, bgt;
	u8 temp;
	
	i810fb_getcolreg(p->bgcol, &bgr, &bgg, &bgb, &bgt, p->fb_info);
	i810fb_getcolreg(p->fgcol, &fgr, &fgg, &fgb, &fgt, p->fb_info);

	temp = i810_readb(PIXCONF1);
	i810_writeb(PIXCONF1, temp | EXTENDED_PALETTE);

	i810_writeb(DACMASK, 0xFF); 
        i810_writeb(CLUT_INDEX_WRITE, 0x04);	

	i810_writeb(CLUT_DATA, (u8) bgr);
	i810_writeb(CLUT_DATA, (u8) bgg);
	i810_writeb(CLUT_DATA, (u8) bgb);

	i810_writeb(CLUT_DATA, (u8) fgr);
	i810_writeb(CLUT_DATA, (u8) fgg);
	i810_writeb(CLUT_DATA, (u8) fgb);

	temp = i810_readb(PIXCONF1);
	i810_writeb(PIXCONF1, temp & ~EXTENDED_PALETTE);
	
}		 

static void i810fb_set_cursor(struct display *disp)
{
	i810fb_set_cursor_color(disp); 
	i810fb_load_cursor_image(disp);
}	
	
/**
 * i810fb_enable_cursor - show or hide the hardware cursor
 * @mode: show (1) or hide (0)
 *
 * Description:
 * Shows or hides the hardware cursor
 */
void i810fb_enable_cursor(int mode)
{
	u32 temp;
	
	temp = i810_readl(PIXCONF);
	if (mode == ON)
		temp |= CURSOR_ENABLE_MASK;
	else
		temp &= ~CURSOR_ENABLE_MASK;
	i810_writel(PIXCONF, temp);
}
			
/**
 * i810_cursor_timer_handler - cursor timer handler
 * @data: arbitrary data which points to cursor structure
 *
 * DESCRIPTION:
 * The cursor timer handler.  This handles the blinking
 * of the cursor
 */
static void i810_cursor_timer_handler(unsigned long data)
{ 
	struct cursor_data *cur;
	(u32) cur = data;
	if (cur->cursor_enable && !cur->blink_count) {
		if (cur->cursor_show) {
			i810fb_enable_cursor(OFF);
			cur->cursor_show = 0; 
		}
		else {
			i810fb_enable_cursor(ON);
			cur->cursor_show = 1;
		}
		cur->blink_count = cur->blink_rate;
	}	
	--cur->blink_count;
	cur->timer->expires = jiffies + (HZ/100);
	add_timer(cur->timer);
}				


/**
 * i810fb_init_cursor - initializes the cursor
 *
 * DESCRIPTION:
 * Initializes the cursor registers
 */
static void i810fb_init_cursor(void)
{
	i810fb_enable_cursor(OFF);
	i810_writel(CURBASE, i810_info->cursor_start_phys);
	i810_writew(CURCNTR, COORD_ACTIVE | CURSOR_MODE_64_TRANS);
}	




static int i810fbcon_switch(int con, struct fb_info *info)
{
	currcon = con;
	i810fb_set_var(&fb_display[con].var, con, info);
    	return 0;
}

    /*
     *  Update the `var' structure (called by fbcon.c)
     */
static int i810fbcon_updatevar(int con, struct fb_info *info)
{
  	struct display *p = &fb_display[con];
	int yoffset, xoffset, depth, total = 0;
	
	yoffset = p->var.yoffset; 
	xoffset = p->var.xoffset;
	depth = p->var.bits_per_pixel >> 3;

	switch (i810_orient->rotate) {
	case NO_ROTATION:
		if (p->var.xres + xoffset <= p->var.xres_virtual)
			total += xoffset * depth;
		if (p->var.yres + yoffset <= p->var.yres_virtual)
			total += yoffset * p->next_line;
		break;
	case ROTATE_RIGHT:
		if (p->var.xres + xoffset <= p->var.xres_virtual)
			total += xoffset * p->next_line;
		if (p->var.yres + yoffset <= p->var.yres_virtual)
			total += (i810_orient->vxres - p->var.yres - 
				  yoffset) * depth;
		break;
	case ROTATE_180:
		if (p->var.xres + xoffset <= p->var.xres_virtual)
			total += (i810_orient->vxres - 
				  p->var.xres - xoffset) * depth;
		if (p->var.yres + yoffset <= p->var.yres_virtual)
			total += (i810_orient->vyres - 
				  p->var.yres - yoffset) * p->next_line;
		break;
	case ROTATE_LEFT:
		if (p->var.xres + xoffset <= p->var.xres_virtual)
			total += (i810_orient->vyres - p->var.xres - 
				  xoffset) * p->next_line;
		if (p->var.yres + yoffset <= p->var.yres_virtual)
			total += yoffset * depth;
	}

	if (!i810_info->gart_is_claimed && render) {
		if (!sync_on_pan)
			i810fb_load_back(i810_info->i810_gtt.pitch_bits);
		if (i810fb_load_front(total, p->next_line, (sync_on_pan) ? 0 : 1))
			i810fb_load_fbstart(total);
	}
	else
		i810fb_load_fbstart(total);
	return 0;
}

    /*
     *  Blank the display.
     */
static void i810fbcon_blank(int blank, struct fb_info *info)
{
	int mode = 0;
	
	switch(blank) {
	case VESA_NO_BLANKING:
		mode = POWERON;
		break;
	case VESA_VSYNC_SUSPEND:
		mode = STANDBY;
		break;
	case VESA_HSYNC_SUSPEND:
		mode = SUSPEND;
		break;
	case VESA_POWERDOWN:
		mode = POWERDOWN;
	}
	i810_writel(HVSYNC, mode);
	
}

u32 get_line_length(int xres_virtual, int bpp)
{
	u32 length;
	
	length = xres_virtual*bpp;
	length = (length+31)&-32;
	length >>= 3;
	
	if (render) {
		switch (length) {
		case 0 ... 512:
			length = 512;
			break;
		case 513 ... 1024:
			length = 1024;
			break;
		case 1025 ... 2048:
			length = 2048;
			break;
		case 2049 ... 4096:
			length = 4096;
			break;
		default:
			length = 0;
		}
	}
	return(length);
}

static void i810fb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var)
{
    	memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    	strcpy(fix->id, i810fb_name);
    	fix->smem_start = i810_info->fb_start_phys;
    	fix->smem_len = i810_info->fb_size;
    	fix->type = FB_TYPE_PACKED_PIXELS;
    	fix->type_aux = 0;
    	switch (var->bits_per_pixel) {

	case 8:
	    	fix->visual = FB_VISUAL_PSEUDOCOLOR;
	    	break;
	case 16:
	case 24:
	case 32:
	    	fix->visual = FB_VISUAL_TRUECOLOR;
	    	break;
    }
    	fix->ywrapstep = 0;
	fix->xpanstep = 1;
	fix->ypanstep = 1;
    	fix->line_length = get_line_length(i810_orient->vxres, 
					   var->bits_per_pixel);
	fix->mmio_start = i810_info->mmio_start_phys;
	fix->mmio_len = MMIO_SIZE;
	fix->accel = FB_ACCEL_I810;
		
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
	switch (var->bits_per_pixel) {
	case 8:       
		var->red.offset = 0;
		var->red.length = 8;
		var->green.offset = 0;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 16:	/* RGB 565 */
		var->red.offset = 11;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 0;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 24:	/* RGB 888 */
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 32:	/* RGB 888 */
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 24;
		var->transp.length = 8;
		break;
	}
	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
}


    /*
     *  Read a single color register and split it into
     *  colors/transparent. Return != 0 for invalid regno.
     */

static int i810fb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
	if (regno > 255) return 1;
 	i810_writeb(CLUT_INDEX_READ, (u8) regno);
 	*red = (u32) i810_readb(CLUT_DATA);
 	*green = (u32) i810_readb(CLUT_DATA);
 	*blue = (u32) i810_readb(CLUT_DATA);
    	*transp = 0;
    	return 0;
}


    /*
     *  Set a single color register. The values supplied are already
     *  rounded down to the hardware's capabilities (according to the
     *  entries in the var structure). Return != 0 for invalid regno.
     */

static int i810fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
    	if (regno > 255) return 1;
	i810_writeb(CLUT_INDEX_WRITE, (u8) regno);
	i810_writeb(CLUT_DATA, (u8) red);
	i810_writeb(CLUT_DATA, (u8) green);
	i810_writeb(CLUT_DATA, (u8) blue); 	
	if (regno < 16) {
		switch (info->var.bits_per_pixel) {
		case 16:		/* RGB 565 */
			((u16 *)(palette))[regno] = (red & 0xf800) |
				((green & 0xfc00) >> 5) |
				((blue & 0xf800) >> 11);
			break;
		case 24:		/* RGB 888 */
		case 32:		/* RGBA 8888 */
			((u32 *)(palette))[regno] = ((red & 0xff00) << 8) |
				(green & 0xff00) |
				((blue & 0xff00) >> 8);
			break;
		}
	}
      	return 0;
}


static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon) return;
	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, i810fb_setcolreg, info);
	else
		fb_set_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), 1,
			    i810fb_setcolreg, info);
}


/**
 * i810fb_create_cursor - creates the cursor structure
 *
 * DESCRIPTION:
 * Creates the hardware cursor structure and starts the 
 * cursor timer
 *
 * RETURNS: nonzero if success
 */
static int __devinit i810fb_create_cursor(void) {
	i810_info->cursor.timer = kmalloc(sizeof(struct timer_list), 
					  GFP_KERNEL); 
	if (!i810_info->cursor.timer) return 0;
	init_timer(i810_info->cursor.timer);
	i810_info->cursor.blink_rate = 40;
	i810_info->cursor.blink_count = i810_info->cursor.blink_rate; 
	i810_info->cursor.cursor_show = 0;
	i810_info->cursor.cursor_enable = 0;
	i810_info->cursor.timer->data = (u32) &i810_info->cursor;
	i810_info->cursor.timer->expires = jiffies + (HZ/100);
	i810_info->cursor.timer->function = i810_cursor_timer_handler;
	return 1;
}

/**
 * i810fb_init_monspecs
 * @fb_info: pointer to device specific info structure
 *
 * DESCRIPTION:
 * Sets the the user monitor's horizontal and vertical
 * frequency limits
 */

static void __devinit i810fb_init_monspecs(struct fb_info *fb_info)
{
	if (!hsync1)
		hsync1 = HFMIN;
	if (!hsync2) 
		hsync2 = HFMAX;
	fb_info->monspecs.hfmax = hsync2 * 1000;
	fb_info->monspecs.hfmin = hsync1 * 1000;
	if (hsync2 < hsync1) 
		fb_info->monspecs.hfmin = hsync2 * 1000;

	if (!vsync2)
		vsync2 = VFMAX;
	if (!vsync1) 
		vsync1 = VFMIN;
	fb_info->monspecs.vfmax = vsync2;
	fb_info->monspecs.vfmin = vsync1;		
	if (vsync2 < vsync1) 
		fb_info->monspecs.vfmin = vsync2;
}

/**
 * i810fb_init_defaults - initializes default values to use
 */
static void __devinit i810fb_init_defaults(void)
{
	i810fb_default.xres = xres;
	i810fb_default.yres = yres;
	i810fb_default.yres_virtual = vyres;
	i810fb_default.bits_per_pixel = bpp;
	i810fb_init_monspecs(&i810_info->fb_info);
	if (!has_mtrr()) mtrr = 0;
	if (!has_rotate()) rotate = 0;
	if (!has_accel()) accel = 0;
	if (accel) i810fb_default.accel_flags = 1;
	if (rotate > 3) rotate = NO_ROTATION;
	if (hwcur) {
		fix_cursor_accel();
		fix_cursor_rotate();
		i810_noaccel8.cursor = i810_cursor;
		i810_noaccel16.cursor = i810_cursor;
		i810_noaccel24.cursor = i810_cursor;
		i810_noaccel32.cursor = i810_cursor;
		fbcon_dummy.cursor = i810_cursor;
	}
}


/**
 * i810fb_save_vgax - save extended register states
 */
static void i810fb_save_vgax(void)
{
	u32 indexport, dataport;
	u8 i;

	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
	
	for (i = 0; i < 4; i++) {
		i810_writeb(indexport, CR30 + i);
		*(&(i810_info->hw_state.cr30) + i) = i810_readb(dataport);
	}
	i810_writeb(indexport, CR35);
	i810_info->hw_state.cr35 = i810_readb(dataport);
	i810_writeb(indexport, CR39);
	i810_info->hw_state.cr39 = i810_readb(dataport);
	i810_writeb(indexport, CR41);
	i810_info->hw_state.cr41 = i810_readb(dataport);
	i810_writeb(indexport, CR70);
	i810_info->hw_state.cr70 = i810_readb(dataport);	
	i810_info->hw_state.msr = i810_readb(MSR_READ);
	i810_writeb(indexport, CR80);
	i810_info->hw_state.cr80 = i810_readb(dataport);
	i810_writeb(SR_INDEX, SR01);
	i810_info->hw_state.sr01 = i810_readb(SR_DATA);
}

/**
 * i810fb_save_vgax - save standard register states
 */
static void i810fb_save_vga(void)
{
	u32 indexport, dataport;
	u8 i;
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}	
	for (i = 0; i < 10; i++) {
		i810_writeb(indexport, CR00 + i);
		*((&i810_info->hw_state.cr00) + i) = i810_readb(dataport);
	}
	for (i = 0; i < 8; i++) {
		i810_writeb(indexport, CR10 + i);
		*((&i810_info->hw_state.cr10) + i) = i810_readb(dataport);
	}
}

/**
 * i810fb_restore_pll - restores saved PLL register
 */
static void i810fb_restore_pll(void)
{
	u32 tmp1, tmp2;
	
	tmp1 = i810_info->hw_state.dclk_2d;
	tmp2 = i810_readl(DCLK_2D);
	tmp1 &= ~MN_MASK;
	tmp2 &= MN_MASK;
	i810_writel(DCLK_2D, tmp1 | tmp2);

	tmp1 = i810_info->hw_state.dclk_1d;
	tmp2 = i810_readl(DCLK_1D);
	tmp1 &= ~MN_MASK;
	tmp2 &= MN_MASK;
	i810_writel(DCLK_1D, tmp1 | tmp2);

	i810_writel(DCLK_0DS, i810_info->hw_state.dclk_0ds);


}

/**
 * i810fb_restore_dac - restores saved DAC register
 */
static void i810fb_restore_dac(void)
{
	u32 tmp1, tmp2;

	tmp1 = i810_info->hw_state.pixconf;
	tmp2 = i810_readl(PIXCONF);
	tmp1 &= DAC_BIT;
	tmp2 &= ~DAC_BIT;
	i810_writel(PIXCONF, tmp1 | tmp2);
}

/**
 * i810fb_restore_vgax - restores saved extended VGA registers
 */
static void i810fb_restore_vgax(void)
{
	u32 indexport, dataport;
	u8 i, j;
	
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
			
	for (i = 0; i < 4; i++) {
		i810_writeb(indexport, CR30+i);
		i810_writeb(dataport, *(&(i810_info->hw_state.cr30) + i));
	}
	i810_writeb(indexport, CR35);
	i810_writeb(dataport, i810_info->hw_state.cr35);
	i810_writeb(indexport, CR39);
	i810_writeb(dataport, i810_info->hw_state.cr39);
	i810_writeb(indexport, CR41);
	i810_writeb(dataport, i810_info->hw_state.cr39);

	/*restore interlace*/
	i810_writeb(indexport, CR70);
	i = i810_info->hw_state.cr70;
	i &= INTERLACE_BIT;
	j = i810_readb(dataport);
	i810_writeb(indexport, CR70);
	i810_writeb(dataport, j | i);

	i810_writeb(indexport, CR80);
	i810_writeb(dataport, i810_info->hw_state.cr80);
	i810_writeb(MSR_WRITE, i810_info->hw_state.msr);
	i810_writeb(SR_INDEX, SR01);
	i = (i810_info->hw_state.sr01) & ~0xE0 ;
	j = i810_readb(SR_DATA) & 0xE0;
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, i | j);
}

/**
 * i810fb_restore_vga - restores saved standard VGA registers
 */
static void i810fb_restore_vga(void)
{
	u32 indexport, dataport;
	u8 i;
	
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
	for (i = 0; i < 10; i++) {
		i810_writeb(indexport, CR00 + i);
		i810_writeb(dataport, *((&i810_info->hw_state.cr00) + i));
	}
	for (i = 0; i < 8; i++) {
		i810_writeb(indexport, CR10 + i);
		i810_writeb(dataport, *((&i810_info->hw_state.cr10) + i));
	}
}

/**
 * i810fb_restore_addr_map - restores saved address registers
 */
static void i810fb_restore_addr_map(void)
{
	u8 tmp;
	i810_writeb(GR_INDEX, GR10);
	tmp = i810_readb(GR_DATA);
	tmp &= ADDR_MAP_MASK;
	tmp |= i810_info->hw_state.gr10;
	i810_writeb(GR_INDEX, GR10);
	i810_writeb(GR_DATA, tmp);
}

/**
 * i810fb_restore_2D - restores saved graphics registers
 */
static void i810fb_restore_2d(void)
{
	u32 tmp_long;
	u16 tmp_word;

	tmp_word = i810_readw(BLTCNTL);
	tmp_word &= ~(3 << 4); 
	tmp_word |= i810_info->hw_state.bltcntl;
	i810_writew(BLTCNTL, tmp_word);
       
	i810_wait_for_hsync(); 
	i810fb_dram_off(OFF);
	i810_writel(PIXCONF, i810_info->hw_state.pixconf);
	i810fb_dram_off(ON);

	tmp_word = i810_readw(HWSTAM);
	tmp_word &= 3 << 13;
	tmp_word |= i810_info->hw_state.hwstam;
	i810_writew(HWSTAM, tmp_word);

	tmp_long = i810_readl(FW_BLC);
	tmp_long &= FW_BLC_MASK;
	tmp_long |= i810_info->hw_state.fw_blc;
	i810_writel(FW_BLC, tmp_long);

	i810_writel(HWS_PGA, i810_info->hw_state.hws_pga); 
	if (render)
		i810_writel(FENCE, i810_info->hw_state.fence0);
}
	
/**
 * i810fb_save_2d - save graphics register states
 */
static void i810fb_save_2d(void)
{
	i810_info->hw_state.iring_state.head = i810_readl(IRING);
	i810_info->hw_state.iring_state.tail = i810_readl(IRING + 4);
	i810_info->hw_state.iring_state.start = i810_readl(IRING + 8);
	i810_info->hw_state.iring_state.size = i810_readl(IRING + 12);
	i810_info->hw_state.dclk_2d = i810_readl(DCLK_2D);
	i810_info->hw_state.dclk_1d = i810_readl(DCLK_1D);
	i810_info->hw_state.dclk_0ds = i810_readl(DCLK_0DS);
	i810_info->hw_state.pixconf = i810_readl(PIXCONF);
	i810_info->hw_state.fw_blc = i810_readl(FW_BLC);
	i810_info->hw_state.bltcntl = i810_readw(BLTCNTL);
	i810_info->hw_state.hwstam = i810_readw(HWSTAM); 
	i810_info->hw_state.hws_pga = i810_readl(HWS_PGA); 
	if (render)
		i810_info->hw_state.fence0 = i810_readl(FENCE);
}
	     

/**
 * i810fb_save_regs - saves the register state
 * 
 * DESCRIPTION:
 * Saves the ALL the registers' state
 */

static void i810fb_save_regs(void)
{
	i810fb_save_vga();
	i810fb_save_vgax();
	i810fb_save_2d();
}

	
/**
 * i810b_init_device - initialize device
 */
static void __devinit i810fb_init_device(void)
{
	i810fb_save_regs();
	if (hwcur) { 
		i810fb_init_cursor();
		add_timer(i810_info->cursor.timer);
	}
	/* mvo: enable external vga-connector (for laptops) */
	if(ext_vga) {
		i810_writel(HVSYNC, 0);
		i810_writel(PWR_CLKC, 3);
	}
  	i810fb_get_mem_freq();
}

/**
 * i810fb_restore_regs - loads the saved register state
 * 
 * DESCRIPTION:
 * Restores ALL of the registers' save state
 */
static void i810fb_restore_regs(void)
{
	i810fb_screen_off(OFF);
	i810fb_protect_regs(OFF);
	i810fb_dram_off(OFF);
	i810fb_restore_pll();
	i810fb_restore_dac();
	i810fb_restore_vga();
	i810fb_restore_vgax();
	i810fb_restore_addr_map();
	i810fb_dram_off(ON);
	i810fb_restore_2d();
	i810fb_restore_ringbuffer(&i810_info->hw_state.iring_state); 
	i810fb_screen_off(ON);
	i810fb_protect_regs(ON);
}

static int __devinit i810fb_alloc_fbmem(void)
{


	/* align surface for memory tiling*/
	if (render) {
		int i = 0;
		render = vram << 10;
		while (i < arraysize(i810_fence) &&
		       render > i810_fence[i]) 
			i++; 
		if (i >= arraysize(i810_fence))
			i--;
		i810_info->i810_gtt.fence_size = i;
		render = i810_fence[i];
		i810_info->fb_size = render << 10;
		if (i810_info->fb_size >= 16 << 20)
			i810_info->fb_offset = i810_info->aper_size - 
				i810_info->fb_size - 4096;
		else
			i810_info->fb_offset = i810_info->aper_size -
				(16 << 20) - 4096;
		i810_info->fb_offset &= ~(((512 << i) * 1024) - 1);
	}
	else {	
		i810_info->fb_size = vram << 20;
		if (vram >= 16)
			i810_info->fb_offset = i810_info->aper_size - 
				i810_info->fb_size - 4096;
		else
			i810_info->fb_offset = i810_info->aper_size -
				(16 << 20) - 4096;
		i810_info->fb_offset &= ~((16 << 20) - 1); 
	}
	i810_info->fb_offset >>= 12;

	if (!(i810_info->i810_gtt.i810_fb_memory = 
	      agp_allocate_memory(i810_info->fb_size >> 12, 
				  AGP_NORMAL_MEMORY))) {
		printk("i810fb_alloc_fbmem: can't allocate framebuffer memory\n");
		return -ENOMEM;
	}
	if (agp_bind_memory(i810_info->i810_gtt.i810_fb_memory, 
			    i810_info->fb_offset)) {
		printk("i810fb_alloc_fbmem: can't bind framebuffer memory\n");
		return -EBUSY;
	}	
	return 0;
}	


static int __devinit i810fb_alloc_cursormem(u32 iring_offset)
{
	if (iring_offset < CURSOR_SIZE >> 12)
		i810_info->cursor_offset = ((i810_info->aper_size - CURSOR_SIZE) >> 12) - 1;
	else
		i810_info->cursor_offset = iring_offset - 
			(CURSOR_SIZE >> 12);
	if (!(i810_info->i810_gtt.i810_cursor_memory = 
	      agp_allocate_memory(CURSOR_SIZE >> 12, AGP_PHYSICAL_MEMORY))) {
		printk("i810fb_alloc_cursormem:  can't allocate" 
		       "cursor memory\n");
		return -ENOMEM;
	}
	if (agp_bind_memory(i810_info->i810_gtt.i810_cursor_memory, 
			    i810_info->cursor_offset)) {
		printk("i810fb_alloc_cursormem: cannot bind cursor memory bound\n");
		return -EBUSY;
	}	
	return 0;
}


static int __devinit i810fb_init_agp(void)
{
	int err;
	u32 iring_offset;
	u32 sarea_offset;

	if ((err = agp_backend_acquire())) {
		printk("i810fb_init_agp: agpgart is busy\n");
		return err;
	}	
	if ((err = i810fb_alloc_fbmem())) return err;
	if (!(iring_offset = i810fb_init_accel(i810_info->fb_offset, 
					       i810_info->aper_size,
					       sync                 )))
		return -ENOMEM;
	if (hwcur) {
		if ((err = i810fb_alloc_cursormem(iring_offset))) return err;
		sarea_offset = i810_info->cursor_offset - (SAREA_SIZE >> 12);
	}	
	else
		sarea_offset = iring_offset - (SAREA_SIZE >> 12);

	/* any failures are not critical */
	if ((i810fb_init_iface(sarea_offset))) return 0;
	return 0;
}

static int __devinit i810fb_fix_pointers(void)
{
      	i810_info->fb_start_phys = i810_info->fb_base_phys + 
		(i810_info->fb_offset << 12);
	i810_info->fb_start_virtual = i810_info->fb_base_virtual + 
		(i810_info->fb_offset << 12);
	i810fb_fix_accel_pointer(i810_info->fb_base_phys, i810_info->fb_base_virtual);
	if (hwcur) {
		i810_info->cursor_start_phys = 
			i810_info->i810_gtt.i810_cursor_memory->physical;
		i810_info->cursor_start_virtual = 
			i810_info->fb_base_virtual + 
			(i810_info->cursor_offset << 12);
		if (!i810fb_create_cursor()) {
			i810fb_release_resource();
			return -ENOMEM;
		}
	}		
	i810fb_fix_iface_pointers(i810_info->fb_size, i810_info->aper_size, 
				  i810_info->fb_base_phys, i810_info->fb_base_virtual);
	return 0;
}

    /*
     *  Initialisation
     */

static int __devinit i810fb_init_pci (struct pci_dev *dev, 
				      const struct pci_device_id *entry)
{
	struct resource *res;
	int err;

	if (!i810fb_initialized)
		return -EINVAL;
	if(!(i810_info = kmalloc(sizeof(struct i810_fbinfo), GFP_KERNEL)))
		return -ENOMEM;
	memset(i810_info, 0, sizeof(struct i810_fbinfo));
	if(!(i810_orient = kmalloc(sizeof(struct orientation), GFP_KERNEL)))
		return -ENOMEM;
	memset(i810_orient, 0, sizeof(struct orientation));
	if(!(i810_timings = kmalloc(sizeof(struct video_timings), GFP_KERNEL)))
		return -ENOMEM;
	memset(i810_timings, 0, sizeof(struct video_timings));
	if ((err = pci_enable_device(dev))) { 
		printk("i810fb_init: cannot enable device\n");
		return err;		
	}
	if ((err = load_agpgart())) {
		printk("i810fb_init: cannot initialize agpgart\n");
		return err;
	}

	agp_copy_info(&i810_info->i810_gtt.i810_gtt_info);
	if (!(i810_info->i810_gtt.i810_gtt_info.aper_size)) {
		printk("i810fb_init: device is disabled\n");
		return -ENOMEM;
	}
	res = &dev->resource[0];
	i810_info->fb_base_phys = i810_info->i810_gtt.i810_gtt_info.aper_base;
	i810_info->aper_size = 
		(i810_info->i810_gtt.i810_gtt_info.aper_size) << 20;

	if (!request_mem_region(i810_info->fb_base_phys, 
				i810_info->aper_size, 
				i810_pci_list[entry->driver_data])) {
		printk("i810fb_init: cannot request framebuffer region\n");
		kfree(i810_info);
		return -ENODEV;
	}
	res = &dev->resource[1];
	i810_info->mmio_start_phys = res->start;
	if (!request_mem_region(i810_info->mmio_start_phys, 
				MMIO_SIZE, 
				i810_pci_list[entry->driver_data])) {
		printk("i810fb_init: cannot request mmio region\n");
		release_mem_region(i810_info->fb_base_phys, i810_info->aper_size);
		kfree(i810_info);
		return -ENODEV;
	}

	if ((err = i810fb_init_agp())) {
		i810fb_release_resource();
		return err;
	}
	i810_info->fb_base_virtual = 
		(u32) ioremap_nocache(i810_info->fb_base_phys, 
				      i810_info->aper_size);
        i810_info->mmio_start_virtual = 
		(u32) ioremap_nocache(i810_info->mmio_start_phys, MMIO_SIZE);

	if ((err = i810fb_fix_pointers())) {
		printk("i810fb_init: cannot fix pointers, no memory\n");
		return err;
	}
	
	if (mtrr) set_mtrr();
	i810fb_init_device();        
	i810fb_init_defaults();
	strcpy(i810_info->fb_info.modename, i810fb_name);
	strcpy(i810_info->fb_info.fontname, fontname);
	i810_info->fb_info.changevar = NULL;
	i810_info->fb_info.node = -1;
	i810_info->fb_info.fbops = &i810fb_ops;
	i810_info->fb_info.disp = &i810_info->disp;
	i810_info->fb_info.switch_con = &i810fbcon_switch;
	i810_info->fb_info.updatevar = &i810fbcon_updatevar;
	i810_info->fb_info.blank = &i810fbcon_blank;
	i810_info->fb_info.flags = FBINFO_FLAG_DEFAULT;

	if((err = i810fb_set_var(&i810fb_default, -1, &i810_info->fb_info))) {
		i810fb_release_resource();
		printk("i810fb_init: cannot set display video mode\n");
		return err;
	}

   	if((err = register_framebuffer(&i810_info->fb_info))) {
    		i810fb_release_resource(); 
		printk("i810fb_init: cannot register framebuffer device\n");
    		return err;  
    	}   
	i810fb_initialized = 0;
      	printk("fb%d: %s v%d.%d.%d%s, Tony Daplas\n"
      	       "     Video RAM      : %dK\n" 
	       "     Mode           : %dx%d-%dbpp\n"
	       "     Acceleration   : %sabled\n"
	       "     MTRR           : %sabled\n"
	       "     External VGA   : %sabled\n"
	       "     Hardware cursor: %sabled\n" 
	       "     Video Timings  : %s\n"	
	       "     Rotation Code  : %sbuilt\n" 
	       "     Interface      : %sabled\n"
	       "     Memory Tiling  : %sabled\n",
	       GET_FB_IDX(i810_info->fb_info.node), 
	       i810_pci_list[entry->driver_data],
	       VERSION_MAJOR, VERSION_MINOR, VERSION_TEENIE, BRANCH_VERSION,
	       (int) i810_info->fb_size>>10, i810fb_default.xres, 
	       i810fb_default.yres, i810fb_default.bits_per_pixel,
	       (accel) ? "en" : "dis", 
	       (i810_info->mtrr_is_set) ? "en" : "dis", 
	       (ext_vga) ? "en" : "dis", (hwcur) ? "en" : "dis",
	       (is_std()) ? "Intel" : "VESA GTF (US)", (has_rotate()) ? "" : "not ",
               (has_iface()) ? "en" : "dis", 
	       (render) ? "en" : "dis");
	
	return 0;
}

static void i810fb_release_resource(void)
{

	unset_mtrr();
	i810fb_iface_cleanup();
	if (hwcur) {
		if (i810_info->cursor.timer) {
			del_timer(i810_info->cursor.timer);
			kfree(i810_info->cursor.timer);
		}
		if (i810_info->i810_gtt.i810_cursor_memory) 
			agp_free_memory(i810_info->i810_gtt.i810_cursor_memory);
	}	
	if (i810_info->i810_gtt.i810_fb_memory) 
		agp_free_memory(i810_info->i810_gtt.i810_fb_memory);
	i810fb_accel_cleanup();
	agp_backend_release();
	if (i810_info->mmio_start_virtual) 
		iounmap((void *) i810_info->mmio_start_virtual);
	if (i810_info->fb_base_virtual) 
		iounmap((void *) i810_info->fb_base_virtual);
	release_mem_region(i810_info->fb_base_phys, i810_info->aper_size);
	release_mem_region(i810_info->mmio_start_phys, MMIO_SIZE);
	pci_disable_device(i810_info->i810_gtt.i810_gtt_info.device); 
	if (i810_orient)
		kfree(i810_orient);
	if (i810_timings)
		kfree(i810_timings);
	if (i810_info) 
		kfree(i810_info);
}

static void __devexit i810fb_remove_pci(struct pci_dev *dev)
{
	i810fb_restore_regs();
	unregister_framebuffer(&i810_info->fb_info);  
	i810fb_release_resource();
	printk("cleanup_module:  unloaded i810 framebuffer device\n");
}                                                	


/* Modularization */

int __init i810fb_init(void)
{
	int err;

	i810fb_initialized = 1;
	if ((err = pci_module_init(&i810fb_driver)))
		return err;
#ifdef MODULE
	i810fb_set_var(&i810fb_default, -1, &i810_info->fb_info);
#endif
	return 0;
}

#ifdef MODULE

MODULE_PARM(vram, "i");
MODULE_PARM_DESC(vram, "System RAM to allocate to framebuffer in MiB" 
		 " (default=4)");
MODULE_PARM(bpp, "i");
MODULE_PARM_DESC(bpp, "Color depth for display in bits per pixel"
		 " (default = 8)");
MODULE_PARM(xres, "i");
MODULE_PARM_DESC(xres, "Hozizontal resolution in pixels (default = 640)");
MODULE_PARM(yres, "i");
MODULE_PARM_DESC(yres, "Vertical resolution in scanlines (default = 480)");
MODULE_PARM(vyres, "i");
MODULE_PARM_DESC(vyres, "Virtual vertical resolution in scanlines"
		 " (default = 480)");
MODULE_PARM(hsync1, "i");
MODULE_PARM_DESC(hsync1, "Mimimum horizontal frequency of monitor in KHz"
		 " (default = 31)");
MODULE_PARM(hsync2, "i");
MODULE_PARM_DESC(hsync2, "Maximum horizontal frequency of monitor in KHz"
		 " (default = 31)");
MODULE_PARM(vsync1, "i");
MODULE_PARM_DESC(vsync1, "Minimum vertical frequency of monitor in Hz"
		 " (default = 50)");
MODULE_PARM(vsync2, "i");
MODULE_PARM_DESC(vsync2, "Maximum vertical frequency of monitor in Hz" 
		 " (default = 60)");
MODULE_PARM(accel, "i");
MODULE_PARM_DESC(accel, "Use Acceleration (BLIT) engine (default = 0)");
MODULE_PARM(mtrr, "i");
MODULE_PARM_DESC(mtrr, "Use MTRR (default = 0)");
MODULE_PARM(ext_vga, "i");
MODULE_PARM_DESC(ext_vga, "Enable external VGA connector (default = 0)");
MODULE_PARM(hwcur, "i");
MODULE_PARM_DESC(hwcur, "use hardware cursor (default = 0)");
MODULE_PARM(render, "i");
MODULE_PARM_DESC(render, "if true, use Memory tiling and alignment"
		 " (default = not set)");
MODULE_PARM(sync_on_pan, "i");
MODULE_PARM_DESC(sync_on_pan, "wait for vsync before panning the display"
		 " (default = 0)");
MODULE_PARM(sync, "i");
MODULE_PARM_DESC(sync, "wait for accel engine to finish drawing"
		 " (default = 0)");
MODULE_PARM(rotate, "i");
MODULE_PARM_DESC(rotate, "rotate display: 0 - normal; 1 - rotate right; 2 - "
		 "upside down; 3 - rotate left (default = 0)");


MODULE_AUTHOR("Tony A. Daplas");
MODULE_DESCRIPTION("Framebuffer device for the Intel 810/815 and"
		   " compatible cards");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
MODULE_LICENSE("GPL"); 
#endif

static void __exit i810fb_exit(void)
{
	pci_unregister_driver(&i810fb_driver);
}
module_init(i810fb_init);
module_exit(i810fb_exit);

#endif

