#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <linux/string.h>
#include <linux/file.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/grinternal.h>

extern ssize_t write_grsec_handler(struct file * file, const char * buf,
				   size_t count, loff_t *ppos);
extern int gr_acl_is_enabled(void);

static DECLARE_WAIT_QUEUE_HEAD(learn_wait);
static DECLARE_WAIT_QUEUE_HEAD(input_wait);
static atomic_t learn_buffer_count = ATOMIC_INIT(0);
static int gr_learn_attached;

#define LEARN_BUFFER_SLOTS 256
#define LEARN_BUFFER_SIZE 16384

static spinlock_t learn_buffer_lock[LEARN_BUFFER_SLOTS] = { [0 ... (LEARN_BUFFER_SLOTS - 1)] = SPIN_LOCK_UNLOCKED };
static char *learn_buffer[LEARN_BUFFER_SLOTS];
static int learn_buffer_len[LEARN_BUFFER_SLOTS];

static ssize_t
read_learn(struct file *file, char * buf, size_t count, loff_t * ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	ssize_t retval = 0;
	char *tmp;
	unsigned int len;
	int i;

	add_wait_queue(&learn_wait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	do {
		if (atomic_read(&learn_buffer_count) > 1)
			break;

		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			goto out;
		}
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto out;
		}

		schedule();
	} while (1);


	for (i = 0; i < LEARN_BUFFER_SLOTS; i++) {
		spin_lock(&learn_buffer_lock[i]);
		len = learn_buffer_len[i];
		tmp = learn_buffer[i];
		if (!len || !tmp) {
			spin_unlock(&learn_buffer_lock[i]);
			continue;
		}
		learn_buffer[i] = NULL;
		learn_buffer_len[i] = 0;
		spin_unlock(&learn_buffer_lock[i]);

		if (count < ((i * LEARN_BUFFER_SIZE) + len)) {
			retval = -EINVAL;
			vfree(tmp);
			goto out;
		}
		if (copy_to_user(buf + (i * LEARN_BUFFER_SIZE), tmp, len)) {
			retval = -EFAULT;
			vfree(tmp);
			goto out;
		}

		retval += len;
		vfree(tmp);
		atomic_dec(&learn_buffer_count);
		atomic_dec(&learn_buffer_count);
	}

	wake_up(&input_wait);
out:
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&learn_wait, &wait);
	return retval;
}

static unsigned int
poll_learn(struct file * file, poll_table * wait)
{
	poll_wait(file, &learn_wait, wait);

	if (atomic_read(&learn_buffer_count) > 1)
		return (POLLIN | POLLRDNORM);

	return 0;
}

void
gr_clear_learn_entries(void)
{
	int i;

	atomic_set(&learn_buffer_count, 0);
	wake_up(&input_wait);
	
	for (i = 0; i < LEARN_BUFFER_SLOTS; i++) {
		if (learn_buffer_len[i]) {
			vfree(learn_buffer[i]);
			learn_buffer[i] = NULL;
			learn_buffer_len[i] = 0;
		}
	}

	return;
}

void
gr_add_learn_entry(const char *fmt, ...)
{
	DECLARE_WAITQUEUE(wait, current);
	va_list args;
	char *tmpbuf;
	char *buf;
	int i;
	unsigned int len;

	if (!gr_learn_attached)
		return;

	tmpbuf = vmalloc(LEARN_BUFFER_SIZE);

	if (tmpbuf == NULL)
		return;

	va_start(args, fmt);
	len = vsnprintf(tmpbuf, LEARN_BUFFER_SIZE, fmt, args);
	va_end(args);

	if (len < 0)
		len = LEARN_BUFFER_SIZE - 1;

	buf = vmalloc(len + 1);

	if (buf == NULL) {
		vfree(tmpbuf);
		return;
	}

	memcpy(buf, tmpbuf, len);
	buf[len] = '\0';
	vfree(tmpbuf);

	add_wait_queue(&input_wait, &wait);

	atomic_inc(&learn_buffer_count);
	while (atomic_read(&learn_buffer_count) > ((2 * (LEARN_BUFFER_SLOTS - 1)) + 1)) {
		/* don't sleep under the BKL */
		if (unlikely(current->lock_depth >= 0)) {
			remove_wait_queue(&input_wait, &wait);
			atomic_dec(&learn_buffer_count);
			vfree(buf);
			return;
		}
		sleep_on(&input_wait);
	}

	if (!gr_acl_is_enabled()) {
		remove_wait_queue(&input_wait, &wait);
		atomic_dec(&learn_buffer_count);
		vfree(buf);
		return;
	}

	for (i = 0; i < LEARN_BUFFER_SLOTS; i++) {
		spin_lock(&learn_buffer_lock[i]);

		if (learn_buffer_len[i]) {
			spin_unlock(&learn_buffer_lock[i]);
			continue;
		}

		learn_buffer[i] = buf;

		learn_buffer_len[i] = len + 1;

		atomic_inc(&learn_buffer_count);
		spin_unlock(&learn_buffer_lock[i]);
		break;
	}

	remove_wait_queue(&input_wait, &wait);
	wake_up_interruptible(&learn_wait);

	return;
}

static int
open_learn(struct inode *inode, struct file *file)
{
	if (file->f_mode & FMODE_READ && gr_learn_attached)
		return -EBUSY;
	else if (file->f_mode & FMODE_READ)
		gr_learn_attached = 1;

	return 0;
}

static int
close_learn(struct inode *inode, struct file *file)
{
	if (file->f_mode & FMODE_READ)
		gr_learn_attached = 0;

	return 0;
}
		
struct file_operations grsec_fops = {
	read:		read_learn,
	write:		write_grsec_handler,
	open:		open_learn,
	release:	close_learn,
	poll:		poll_learn,
};
