
/*
 * Copyright (c) 1996, 1997, 2000-2002 Michael Shalayeff.
 * 
 * Version 1.89, last modified 19-Sep-99
 *    
 * Copyright Theodore Ts'o, 1994, 1995, 1996, 1997, 1998, 1999.
 * All rights reserved.
 *
 * Copyright 1998 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 * Theo de Raadt <deraadt@openbsd.org> came up with the idea of using
 * such a mathematical system to generate more random (yet non-repeating)
 * ids to solve the resolver/named problem.  But Niels designed the
 * actual system based on the constraints.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/smp_lock.h>
#include <linux/random.h>

#define RU_OUT 180
#define RU_MAX 30000
#define RU_GEN 2
#define RU_N 32749
#define RU_AGEN 7
#define RU_M 31104
#define PFAC_N 3
const static __u16 pfacts[PFAC_N] = { 2, 3, 2729 };

static __u16 ru_x;
static __u16 ru_seed, ru_seed2;
static __u16 ru_a, ru_b;
static __u16 ru_g;
static __u16 ru_counter = 0;
static __u16 ru_msb = 0;
static unsigned long ru_reseed = 0;
static __u32 tmp;

static struct arc4_stream {
	__u8 s[256];
	unsigned int cnt;
	__u8 i;
	__u8 j;
} arc4random_state[NR_CPUS];

#define TCP_RNDISS_ROUNDS	15
#define TCP_RNDISS_OUT		7200
#define TCP_RNDISS_MAX		30000

static __u8 tcp_rndiss_sbox[NR_CPUS][128];
static __u16 tcp_rndiss_msb[NR_CPUS];
static __u16 tcp_rndiss_cnt[NR_CPUS];
static unsigned long tcp_rndiss_reseed[NR_CPUS];

static unsigned long arc4random_reseed[NR_CPUS];

static int arc4random_initialized[NR_CPUS];

static __u16 pmod(__u16, __u16, __u16);
static void ip_initid(void);
__u16 ip_randomid(void);

static __inline__ __u8 arc4_getbyte(void);
static __inline__ void arc4_stir(void);
static void arc4maybeinit(void);
__u32 arc4random(void);

static __inline__ __u8
arc4_getbyte(void)
{
	register __u8 si, sj;
	int cpu = smp_processor_id();

	arc4random_state[cpu].cnt++;
	arc4random_state[cpu].i++;
	si = arc4random_state[cpu].s[arc4random_state[cpu].i];
	arc4random_state[cpu].j += si;
	sj = arc4random_state[cpu].s[arc4random_state[cpu].j];
	arc4random_state[cpu].s[arc4random_state[cpu].i] = sj;
	arc4random_state[cpu].s[arc4random_state[cpu].j] = si;
	return arc4random_state[cpu].s[(si + sj) & 0xff];
}

static __inline__ void
arc4_stir(void)
{
	__u8 buf[256];
	register __u8 si;
	register int n;
	int len = sizeof (buf);
	int cpu = smp_processor_id();
	unsigned long flags;

	do_gettimeofday((struct timeval *) buf);
	get_random_bytes(buf + sizeof (struct timeval),
			 len - sizeof (struct timeval));

	save_flags(flags);
	cli();

	arc4random_state[cpu].i--;
	for (n = 0; n < 256; n++) {
		arc4random_state[cpu].i++;
		si = arc4random_state[cpu].s[arc4random_state[cpu].i];
		arc4random_state[cpu].j += si + buf[n % len];
		arc4random_state[cpu].s[arc4random_state[cpu].i] =
		    arc4random_state[cpu].s[arc4random_state[cpu].j];
		arc4random_state[cpu].s[arc4random_state[cpu].j] = si;
	}
	arc4random_state[cpu].j = arc4random_state[cpu].i;
	arc4random_state[cpu].cnt = 0;

	restore_flags(flags);

	for (n = 0; n < 256 * 4; n++)
		arc4_getbyte();

	return;
}

void
arc4_init(void)
{
	int i, n;

	for (n = 0; n < NR_CPUS; n++) {
		for (i = 0; i < 256; i++)
			arc4random_state[n].s[i] = i;
		arc4random_initialized[n] = 0;
	}
}

static void
arc4maybeinit(void)
{
	int cpu = smp_processor_id();

	if (arc4random_initialized[cpu] &&
	    time_after(xtime.tv_sec, arc4random_reseed[cpu]))
		arc4random_initialized[cpu] = 0;

	if (!arc4random_initialized[cpu]) {
		arc4random_initialized[cpu]++;
		arc4_stir();
		arc4random_reseed[cpu] = xtime.tv_sec + (10 * 60);
	}
}

__u32
arc4random(void)
{
	arc4maybeinit();
	return ((arc4_getbyte() << 24) | (arc4_getbyte() << 16)
		| (arc4_getbyte() << 8) | arc4_getbyte());
}

static __u16
pmod(__u16 gen, __u16 exp, __u16 mod)
{
	__u16 s, t, u;

	s = 1;
	t = gen;
	u = exp;

	while (u) {
		if (u & 1)
			s = (s * t) % mod;
		u >>= 1;
		t = (t * t) % mod;
	}
	return (s);
}

static void
ip_initid(void)
{
	__u16 j, i;
	int noprime = 1;

	ru_x = ((tmp = arc4random()) & 0xFFFF) % RU_M;

	ru_seed = (tmp >> 16) & 0x7FFF;
	ru_seed2 = arc4random() & 0x7FFF;

	ru_b = ((tmp = arc4random()) & 0xfffe) | 1;
	ru_a = pmod(RU_AGEN, (tmp >> 16) & 0xfffe, RU_M);
	while (ru_b % 3 == 0)
		ru_b += 2;

	j = (tmp = arc4random()) % RU_N;
	tmp = tmp >> 16;

	while (noprime) {
		for (i = 0; i < PFAC_N; i++)
			if (j % pfacts[i] == 0)
				break;

		if (i >= PFAC_N)
			noprime = 0;
		else
			j = (j + 1) % RU_N;
	}

	ru_g = pmod(RU_GEN, j, RU_N);
	ru_counter = 0;

	ru_reseed = xtime.tv_sec + RU_OUT;
	ru_msb = ru_msb == 0x8000 ? 0 : 0x8000;
}

__u16
ip_randomid(void)
{
	int i, n;

	if (ru_counter >= RU_MAX || time_after(xtime.tv_sec, ru_reseed))
		ip_initid();

	if (!tmp)
		tmp = arc4random();

	n = tmp & 0x3;
	tmp = tmp >> 2;
	if (ru_counter + n >= RU_MAX)
		ip_initid();
	for (i = 0; i <= n; i++)
		ru_x = (ru_a * ru_x + ru_b) % RU_M;
	ru_counter += i;

	return ((ru_seed ^ pmod(ru_g, ru_seed2 ^ ru_x, RU_N)) | ru_msb);
}

__u16
tcp_rndiss_encrypt(__u16 val)
{
	__u16 sum = 0, i;
	int cpu = smp_processor_id();

	for (i = 0; i < TCP_RNDISS_ROUNDS; i++) {
		sum += 0x79b9;
		val ^= ((__u16) tcp_rndiss_sbox[cpu][(val ^ sum) ^ 0x7f]) << 7;
		val = ((val & 0xff) << 7) | (val >> 8);
	}

	return val;
}

static void
tcp_rndiss_init(void)
{
	int cpu = smp_processor_id();

	get_random_bytes(tcp_rndiss_sbox[cpu], sizeof (tcp_rndiss_sbox));
	tcp_rndiss_reseed[cpu] = xtime.tv_sec + TCP_RNDISS_OUT;
	tcp_rndiss_msb[cpu] = tcp_rndiss_msb[cpu] == 0x8000 ? 0 : 0x8000;
	tcp_rndiss_cnt[cpu] = 0;
}

__u32
ip_randomisn(void)
{
	int cpu = smp_processor_id();

	if (tcp_rndiss_cnt[cpu] >= TCP_RNDISS_MAX ||
	    time_after(xtime.tv_sec, tcp_rndiss_reseed[cpu]))
		tcp_rndiss_init();

	return (((tcp_rndiss_encrypt(tcp_rndiss_cnt[cpu]++) |
		  tcp_rndiss_msb[cpu]) << 16) | (arc4random() & 0x7fff));
}
