/*
 * Freeze_and_free.c
 *
 * Freeze_and_free contains the routines software suspend uses to freeze other
 * processes during the suspend cycle and to (if necessary) free up memory in
 * accordance with limitations on the image size.
 *
 * Ideally, the image saved to disk would be an atomic copy of the entire 
 * contents of all RAM and related hardware state. One of the first 
 * prerequisites for getting our approximation of this is stopping the activity
 * of other processes. We can't stop all other processes, however, since some 
 * are needed in doing the I/O to save the image. Freeze_and_free.c contains 
 * the routines that control suspension and resuming of these processes.
 * 
 * Under high I/O load, we need to be careful about the order in which we
 * freeze processes. If we freeze processes in the wrong order, we could
 * deadlock others. The freeze_order array this specifies the order in which
 * critical processes are frozen. All others are suspended after these have
 * entered the refrigerator.
 *
 * Another complicating factor is that freeing memory requires the processes
 * to not be frozen, but at the end of freeing memory, they need to be frozen
 * so that we can be sure we actually have eaten enough memory. This is why
 * freezing and freeing are in the one file. The freezer is not called from
 * the main logic, but indirectly, via the code for eating memory. The eat
 * memory logic is iterative, first freezing processes and checking the stats,
 * then (if necessary) unfreezing them and eating more memory until it looks 
 * like the criteria are met (at which point processes are frozen & stats
 * checked again).
 */

#define SWSUSP_FREEZER_C

#include "common.h"
#include <linux/notifier.h>

extern int pagedir1_size, pagedir2_size;
extern int pageset2_sizelow;
extern unsigned long orig_mem_free;
extern void markpagesforpageset2(void);
extern int expected_size1, expected_size2; 

/* Pagedir.c */
extern pageset_sizes_t count_data_pages(int whichpagedir);
extern int calcpagedirsize(int size);

/* ------------------------------------------------------------------------ */

/* Processes which are never put in the freezer */
#define IS_IGNORED(p) \
 ((p->flags & PF_IOTHREAD) || (p == current) || (p->state == TASK_ZOMBIE))

/* Processes which need to be frozen before others in this particular order */
static char * freeze_order[] = { "kupdated", "kjournald", "kswapd" };
static const int freeze_order_num = 3;

static int processestofreeze = 0;
static int processesfrozen = 0;
static int freezing = 0;

/**
 * refrigerator - idle routine for frozen processes
 * @flag: unsigned long, non zero if signals to be flushed.
 *
 * A routine for kernel threads which should not do work during suspend
 * to enter and spin in until the process is finished.
 *
 */

void refrigerator(unsigned long flag)
{
	long save;
	/* You need correct to work with real-time processes.
	   OTOH, this way one process may see (via /proc/) some other
	   process in stopped state (and thereby discovered we were
	   suspended. We probably do not care). 
	 */
	processesfrozen++;
	printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s refrigerated.\n", current->comm);
	
	if (flag)
		flush_signals(current); /* We have signaled a kernel thread, which isn't normal behaviour
					   and that may lead to 100%CPU sucking because those threads
					   just don't manage signals. */

	/* 
	 * We don't look at suspend_task here anymore, because
	 * it isn't set when first threads are signalled.
	 */

	if (freezing) {
		save = current->state;
		current->state = TASK_STOPPED;
		current->flags |= PF_FROZEN;
		while (current->flags & PF_FROZEN)
			schedule();
		current->state = save;
	} else
		printk(KERN_DEBUG "%s entered refrigerator when suspend not in progress!\n", current->comm);
	current->flags &= ~PF_FREEZE;
	processesfrozen--;
}

/*
 * freeze_processes - Freeze processes prior to saving an image of memory to disk
 * 
 * Return value: 0 = success, else # of processes that we failed to stop 
 */
static int freeze_processes(void)
{
	int todo, threadcount;
	unsigned long start_time, iorequestflags;
	struct task_struct *p;
	int index = 0;
	
	/* 
	 * Only use our count of processes yet to be frozen
	 * when no processes have been frozen yet (first pass).
	 */

	freezing = 1;

	if (!processesfrozen)
		processestofreeze = 0;

	start_time = jiffies;
	printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s:\n",
		freeze_order[0]);

	swsusp_spin_lock_irqsave(&io_request_lock, iorequestflags);

	do {
		todo = 0;
		threadcount = 0;
		read_lock(&tasklist_lock);
		for_each_task(p) {
			unsigned long flags;
			int getsattention = ((index == freeze_order_num) || 
					(!strcmp(p->comm, freeze_order[index])));
			if (IS_IGNORED(p) || (p->flags & PF_FROZEN) ||
					(!(p->flags & PF_REFRIGERATE)))
				continue;
			if (getsattention)
				threadcount++;
			todo++;
			if (!getsattention)
				continue;
			if (!(p->flags & PF_FREEZE))
				printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "  pid %d\n", p->pid);
			/* FIXME: smp problem here: we may not access other process' flags
			   without locking */
			p->flags |= PF_FREEZE;
			swsusp_spin_lock_irqsave(&p->sigmask_lock, flags);
			signal_wake_up(p);
			swsusp_spin_unlock_irqrestore(&p->sigmask_lock, flags);
		}
		read_unlock(&tasklist_lock);
		if (!threadcount && index < freeze_order_num) {
			printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "done.\n");
			index++;
			if (index < freeze_order_num) {
				printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s:\n",
					freeze_order[index]);
			} else {
				printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "Remaining threads:");
			}
		}
		if (processestofreeze == 0)
			processestofreeze = todo;
		yield();
		if (time_after(jiffies, start_time + TIMEOUT)) {
			SET_RESULT_STATE(SUSPEND_ABORTED);
			SET_RESULT_STATE(SUSPEND_FREEZING_FAILED);
			printk(KERN_ERR name_suspend "\nStopping tasks failed.\n");
			show_state();
			printk(KERN_ERR "Tasks that refused to be refrigerated were:\n");
			for_each_task(p) {
				if (IS_IGNORED(p) || (p->flags & PF_FROZEN) ||
					(!(p->flags & PF_FREEZE)))
					continue;
				/* 
				 * We still let the process enter the refrigerator when it's 
				 * ready. It will just flush signals and exit immediately 
				 * because suspend_task isn't set.
				 */
				printk(" - %s\n", p->comm);
			}
			freezing = 0;
			swsusp_spin_unlock_irqrestore(&io_request_lock, iorequestflags);
			return todo;
		}
		if (processestofreeze)
			update_status(processestofreeze - processesfrozen , processestofreeze, NULL);
	} while(processestofreeze > processesfrozen);

	suspend_task = current->pid;
	freezing = 0;

	swsusp_spin_unlock_irqrestore(&io_request_lock, iorequestflags);

	check_shift_keys(1, "Freezing processes completed. ");
	
	return 0;
}

void thaw_processes(void)
{
	struct task_struct *p;
	printnolog(SUSPEND_FREEZER, SUSPEND_LOW, 1, name_resume "Thawing tasks\n");
	
	suspend_task = 0;
	
	for_each_task(p) {
		if (IS_IGNORED(p))
			continue;
		if (p->flags & PF_FROZEN) {
			p->flags &= ~PF_FROZEN;
			wake_up_process(p);
		}
	}
}

/*
 * We need to eat memory until we can:
 * 1. Perform the save without changing anything (RAM_NEEDED < max_mapnr)
 * 2. Fit it all in available swap (swapinfo.freeswap >= SWAP_NEEDED)
 * 3. Reload the pagedir and pageset1 to places that don't collide with their
 *    final destinations, not knowing to what extent the resumed kernel will
 *    overlap with the one loaded at boot time. I think the resumed kernel should overlap
 *    completely, but I don't want to rely on this as it is an unproven assumption. We
 *    therefore assume there will be no overlap at all (worse case).
 * 4. Meet the user's requested limit (if any) on the size of the image.
 *    The limit is in MB, so pages/256 (assuming 4K pages).
 *
 *    I have seen simply waiting for the user to press SHIFT increase the
 *    number of pages to save by one (logs?), so tests are <, not <= to allow
 *    one extra page. (Final test in save_image doesn't use EATEN_ENOUGH_MEMORY)
 *
 *    RAM_TO_RESUME includes +8 to allow for extra pages allocated during read_primary_suspend_image.
 *    pageset1_size includes space for both pagedirs.
 */

static int amount_needed(int eaten)
{
	return (MAX(MAX(MAX(RAM_TO_SUSPEND - nr_free_pages() - eaten, RAM_TO_RESUME - max_mapnr), 
			       SWAP_NEEDED -  swapinfo.freeswap), 
			   ((image_size_limit) ? (SWAP_NEEDED - (image_size_limit << 8)) : 0)));
}

#define EATEN_ENOUGH_MEMORY() (amount_needed(amount_eaten) < 1)

/* 
 * We are assuming here that pageset1_size always excludes pagedir size.
 */
void display_stats(unsigned int eaten)
{ 
	printnolog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE, 0, "Free:%d(%d). Sets:%d,%d (%d). PD:(%d,%d). Swap:%d/%lu. RAM s:%d; r:%lu. Limits:%lu,%d\n", 
			nr_free_pages() + eaten, 
			nr_free_pages(),
			pageset1_size, 
			pageset2_size,
			pageset2_sizelow,
			pagedir1_size,
			pagedir2_size,
			SWAP_NEEDED,
			swapinfo.freeswap,
			RAM_TO_SUSPEND,
			RAM_TO_RESUME,
			max_mapnr,
			image_size_limit << 8); 
}

/*
 * Eaten is the number of pages which have been eaten.
 * Pagedirincluded is the number of pages which have been allocated for the pagedir.
 */
void recalculate_stats(int amount_eaten) 
{ 
	pageset_sizes_t result;
	markpagesforpageset2();  /* Need to call this before getting pageset1_size! */
	result = count_data_pages(0);	/* Only counts because of zero parameter */
	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, "Recalculate stats: PD2: %d->%d ", result.size2, calcpagedirsize(result.size2));
	pageset2_sizelow = result.size2low;
	pageset2_size = result.size2;
	pagedir2_size = calcpagedirsize(pageset2_size);
	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, "PD1: %d->%d ", result.size1, calcpagedirsize(result.size1));
	pageset1_size = result.size1 + pagedir2_size;
	pagedir1_size = calcpagedirsize(pageset1_size + calcpagedirsize(pageset1_size));
	pageset1_size += pagedir1_size;
	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, "Pageset1 Total %d + %d + %d = %d. Check: %d.\n",
			result.size1,
			pagedir1_size,
			pagedir2_size,
			pageset1_size,
			calcpagedirsize(pageset1_size));
	si_swapinfo_devsonly(&swapinfo);
}

struct eaten_memory_t
{
	void * prev;
	int order; /* Order of _this_ allocation */
};

struct eaten_memory_t *eaten_memory = NULL;

static int grab_free_memory(void)
{
	int order, k, amount_eaten = 0;
	struct eaten_memory_t *prev = eaten_memory;

	/*
	 * First, quickly eat all memory that's already free.
	 */
	
	for (order = MAX_ORDER - 1; order > -1; order--) {
		eaten_memory = (struct eaten_memory_t *) __get_free_pages(__GFP_FAST, order);
		while (eaten_memory) {
			struct page * page = virt_to_page(eaten_memory);
			eaten_memory->prev = prev;
			eaten_memory->order = order;
			prev = eaten_memory;
			amount_eaten += (1 << order);
			for (k=0; k < (1 << order); k++)
				SetPageNosave(page + k);
			eaten_memory = (struct eaten_memory_t *) __get_free_pages(__GFP_FAST, order);
		}
	}

	eaten_memory = prev;
	return amount_eaten;
}

void free_grabbed_memory(void)
{
	struct eaten_memory_t *prev = NULL;
	int j, num_freed = 0;

	/* Free all eaten pages immediately */
	while(eaten_memory) {
		struct page * page = virt_to_page(eaten_memory);
		prev = eaten_memory->prev;
		for (j=0; j < (1 << (eaten_memory->order)); j++) {
			ClearPageNosave(page + j);
			num_freed++;
		}
		free_pages((unsigned long) eaten_memory, eaten_memory->order);
		eaten_memory = prev;
	}
	printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE,"Freed %d pages.\n", num_freed);
}

int attempt_to_freeze(void)
{
	int lastfreezeresult, freezeattemptsleft = 3;

	/* Stop processes before checking again */
	do {
		prepare_status(1, 1, "Freezing processes: Attempt %d", 4 - freezeattemptsleft);
		lastfreezeresult = freeze_processes();
		freezeattemptsleft--;
	} while (freezeattemptsleft && lastfreezeresult);

	return lastfreezeresult;
}

int eat_memory_and_freeze_processes(void)
{
	int orig_memory_still_to_eat, last_amount_needed = 0, times_criteria_met = 0;
	int amount_eaten = 0, arefrozen = 0;

	printlog(SUSPEND_EAT_MEMORY, SUSPEND_LOW, name_suspend "Eating pages\n");

	recalculate_stats(0);
	display_stats(0);

	prepare_status(0, 1, "Freeing memory...");

	orig_memory_still_to_eat = amount_needed(0);
	last_amount_needed = orig_memory_still_to_eat;

	/*
	 * Note that if we have enough swap and enough free memory, we may exit 
	 * without eating anything. We give up when the last 10 iterations ate
	 * no extra pages because we're not going to get much more anyway, but
	 * the few pages we get will take a lot of time.
	 *
	 * We freeze processes before beginning, and then unfreeze them if we
	 * need to eat memory until we think we have enough. If our attempts
	 * to freeze fail, we give up and abort.
	 */

	if (attempt_to_freeze())
		return 1;
	arefrozen = 1;
	
	printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE, "Frozen processes at entry to eating memory.\n");

	while ((!EATEN_ENOUGH_MEMORY()) && (!SUSPEND_ABORTED) && (times_criteria_met < 10) && (swapinfo.freeswap)) {
		if (orig_memory_still_to_eat)
			update_status(orig_memory_still_to_eat - amount_needed(amount_eaten), orig_memory_still_to_eat, "Image size %d ", MB(SWAP_NEEDED));
		//printk("Amount needed:%d, Orig:%d, Percent: %d\n", amount_needed(), orig_memory_still_to_eat, (orig_memory_still_to_eat - amount_needed()) * 100 / orig_memory_still_to_eat);
		
		if (arefrozen) {
			printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE, "Waking for memory grab...\n");
			thaw_processes();
			arefrozen = 0;
		}

		if (((last_amount_needed - amount_needed(amount_eaten)) == 0) && (amount_eaten > 0))
			times_criteria_met++;
		else
			times_criteria_met = 0;
		last_amount_needed = amount_needed(amount_eaten);
		try_to_free_pages_swsusp(GFP_KSWAPD, amount_needed(amount_eaten));
		amount_eaten += grab_free_memory();
		recalculate_stats(amount_eaten);
		display_stats(amount_eaten);

		check_shift_keys(0, NULL);

		if (EATEN_ENOUGH_MEMORY())  { /* Make sure we really have eaten enough before we leave the loop */
			do_suspend_sync();
			
			if (times_criteria_met == 10) 
				if (attempt_to_freeze())
					return 1;

			recalculate_stats(amount_eaten);
		}
	}

	printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE, "\n");
	
	if (!arefrozen)
		if (attempt_to_freeze())
			return 1;
	
	update_status(100, 100, NULL);
	if ((amount_needed(amount_eaten) > 0))
		printlog(SUSPEND_EAT_MEMORY, SUSPEND_LOW, "\n--- UNABLE TO GET ANY MORE PAGES ---\n");
	
	check_shift_keys(1, "Memory eating completed. ");

	return 0;
}

/*
 * The following code is based on the reboot notifier code in kernel/sys.c.
 * It is used to notify drivers when a suspend cycle finishes, so that
 * timers which have been stopped can be restarted.
 */


/*
 *	Notifier list for kernel code which wants to be called
 *	at resume. This is used to restart timers which were
 *	stopped during suspend.
 */

static struct notifier_block *resume_notifier_list;
rwlock_t swsusp_notifier_lock = RW_LOCK_UNLOCKED;

/**
 *	register_resume_notifier - Register function to be called at resume time
 *	@nb: Info about notifier function to be called
 *
 *	Registers a function with the list of functions
 *	to be called at resume time.
 *
 *	Currently always returns zero, as notifier_chain_register
 *	always returns zero.
 */
 
int register_resume_notifier(struct notifier_block * nb)
{
	return notifier_chain_register(&resume_notifier_list, nb);
}

/**
 *	unregister_resume_notifier - Unregister previously registered resume notifier
 *	@nb: Hook to be unregistered
 *
 *	Unregisters a previously registered resume
 *	notifier function.
 *
 *	Returns zero on success, or %-ENOENT on failure.
 */
 
int unregister_resume_notifier(struct notifier_block * nb)
{
	return notifier_chain_unregister(&resume_notifier_list, nb);
}

inline int notify_resume(void)
{
	return notifier_call_chain(&resume_notifier_list, 0, NULL);
}

