/*
 * linux/drivers/swsusp/io.c
 *
 * This file contains IO routines for swsusp.
 *
 */

#define SWSUSP_IO_C

#include "common.h"
extern void reset_io_stats(void);
extern void check_io_stats(void);
extern int write_swap_page(swp_entry_t posn, struct page * page, int asyncio);
extern int get_phys_params(struct pbe * pbe);
extern void swsusp_rw_swap_page(int rw, struct pbe * pbe);
extern unsigned long swapforimage;
extern unsigned long base_mem_free;
extern int expected_size1, expected_size2; 
extern int swsusp_act_used;
extern int swsusp_lvl_used;
extern int swsusp_dbg_used;
extern void warmup_collision_cache(void);
extern int relocate_pagedir(void);
extern int check_pagedir(void);
extern struct pagedir __nosavedata pagedir_nosave;
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
extern unsigned long pagechecksum(struct page * page);
unsigned long checkpagecopy = 0;
unsigned long checkpage = 0;
#endif
extern int allocatemap(unsigned long ** pagemap, int setnosave);
extern unsigned long * pagedirmap;

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION

extern int zlib_compress(unsigned char *data_in, unsigned char *cpage_out, __u32 *sourcelen, __u32 *dstlen);
extern void zlib_decompress(unsigned char *data_in, unsigned char *cpage_out, __u32 srclen, __u32 destlen);
static void *compression_workspace;

/*
 * prepare_zlib_deflate_memory()
 *
 * Description: Allocate space for zlib's deflate code.
 * Returns: -ENOMEM on failure, 0 on success.
 *
 */

static int prepare_zlib_deflate_memory(void)
{
	compression_workspace = vmalloc(zlib_deflate_workspacesize());
	if (!compression_workspace) {
		printk(KERN_WARNING "Failed to allocate %d bytes for deflate workspace\n", zlib_deflate_workspacesize());
		return -ENOMEM;
	}
	return 0;
}

/*
 * prepare_zlib_inflate_memory()
 *
 * Description: Allocate space for zlib inflate code.
 * Returns: -ENOMEM on failure, 0 on success.
 *
 */

static int prepare_zlib_inflate_memory(void)
{
	compression_workspace = vmalloc(zlib_inflate_workspacesize());
	if (!compression_workspace) {
		printk(KERN_WARNING "Failed to allocate %d bytes for inflate workspace\n", zlib_inflate_workspacesize());
		return -ENOMEM;
	}
	return 0;
}

/*
 * release_zlib_memory()
 *
 * Description: Release space for zlib [de|in]flate code.
 * Returns: Nothing
 *
 */

static void release_zlib_memory(void)
{
	if (compression_workspace)
		vfree(compression_workspace);
}
#endif // CONFIG_SOFTWARE_SUSPEND_COMPRESSION

/*
 * write_pageset()
 *
 * Description: Write a pageset to disk.
 * @pagedir: Pointer to the pagedir to be saved.
 * @whichtowrite: Controls what debugging output is printed.
 * Returns: Zero on success or -1 on failure.
 *
 */

int write_pageset(struct pagedir * pagedir, int whichtowrite)
{
	int nextupdate = 0, size, ret = 0, i;
	int start_time, end_time;
	long result = 0;
	struct pbe * pbe;
	unsigned int origfree = nr_free_pages();
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	z_stream strm;
	int ret2 = 0;
	struct pbe * swap_pbe = NULL;
	unsigned long compression_buffer = 0;
#endif

#define abort_and_exit(f, a...) do { \
			abort_suspend(f, ## a); \
			result = -1; \
			goto write_pageset_free_buffers; \
	} while(0)

	reset_io_stats();
	PRINTFREEMEM("at start of write pageset");

	size = pagedir->pageset_size;
	if (!size)
		return 0;

	if (whichtowrite == 1) {
		prepare_status(1, 1, "Writing suspend image (1)...");
	} else {
		prepare_status(1, 1, "Writing suspend image (2)...");
	}	
	
	start_time = jiffies;

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		compression_buffer = get_zeroed_page(GFP_ATOMIC);

		if (!compression_buffer) {
			printk(KERN_ERR name_suspend "Failed to allocate memory for compression buffer.\n");
			return -1;
		}
	
		if (prepare_zlib_deflate_memory()) {
			printk(KERN_ERR name_suspend "Failed to allocate memory for zlib to work in.\n");
			free_pages(compression_buffer, 0);
			return -1;
		}
	
		strm.total_in = 0;
		strm.total_out = 0;
		strm.workspace = compression_workspace;
		strm.next_out = (char *) compression_buffer;
		strm.avail_out = PAGE_SIZE;

		if (Z_OK != zlib_deflateInit(&strm, Z_BEST_SPEED)) {
			printk(KERN_ERR name_suspend "Failed to initialise zlib.\n");
			release_zlib_memory();
			free_pages(compression_buffer, 0);
			return -1;
		}
		swap_pbe = pagedir->data;
	}
#endif

	pbe = pagedir->data;

	for (i=0; i<size; i++) {
		if (!(i&0x1FF)) 
			printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_LOW, ".");
		if ((i >= nextupdate) || (!(i%(1 << (20 - PAGE_SHIFT)))))
			nextupdate = update_status(i, size, " %d/%d MB ", MB(i+1), MB(size));
		printnolog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, 1, name_suspend "Writing pageset: %d/%d ", i + 1, size);
	
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
		if ((unsigned long) page_address(pbe->origaddress) == checkpage) {
			checkpagecopy = (unsigned long) page_address(pbe->address);
			printk(KERN_ERR "Page checksum for page %lx is %lx. Writing to %lx.\n",
				checkpage,
				pagechecksum(pbe->origaddress),
				pbe->swap_address.val);
		}
#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
		if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
			strm.next_in = kmap_nonblock(pbe->address);
			if (!strm.next_in) {
				zlib_deflateEnd(&strm);
				abort_and_exit("Failed to kmap a page.\n");
			}
			strm.avail_in = PAGE_SIZE;

			while ((strm.avail_in) || (((i+1) == size) && (ret != Z_STREAM_END))) {
				if ((i+1) == size) {
					ret = zlib_deflate(&strm, Z_FINISH);
					if (ret == Z_STREAM_END)
						zlib_deflateEnd(&strm);
				} else
					ret = zlib_deflate(&strm, Z_PARTIAL_FLUSH);
				if ((ret != Z_OK) && (ret != Z_STREAM_END)) {
					zlib_deflateEnd(&strm);
					abort_and_exit("Failed to finish compressing data. Result %d received.\n", ret);
				}
				if ((!strm.avail_out) || (ret == Z_STREAM_END)) {
					/* Submit previous buffer for writing */
					ret2 = write_swap_page(swap_pbe->swap_address, virt_to_page(compression_buffer), 1);
					if (ret2) {
						zlib_deflateEnd(&strm);
						abort_and_exit("Failed writing page to swap. Error number was %d.\n", ret);
					}

					/* Get next buffer */
					swap_pbe = get_next_pbe(swap_pbe);
					/* 
					 * Since it is a copy of the compression buffer that is written 
					 * asynchronously, we can immediately reuse our buffer.
					 */
					strm.next_out = (char *) compression_buffer;
					strm.avail_out = PAGE_SIZE;
				}
			}
		
			kunmap(pbe->address);
		} else {
#endif
			if (whichtowrite == 1) {
				printnolog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, 0, "!%d! [1]%lx -> [2]%lx -> [sw]%lx.", 
					i,
					(unsigned long) pbe->origaddress, 
					(unsigned long) pbe->address, 
					pbe->swap_address.val);
			} else {
				printnolog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, 0, "!%d! %s[2]%lx -> [sw]%lx",
					i,
					PageHighMem(pbe->origaddress) ? "[HighMem]" : "",
					(unsigned long) pbe->address,
					pbe->swap_address.val);
			}

			ret = write_swap_page(pbe->swap_address, pbe->address, 1);
			if (ret)
				abort_and_exit("Failed writing page to swap. Error number was %d.\n", ret);
		
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
		}
#endif
		check_shift_keys(0, NULL);
		if (TEST_RESULT_STATE(SUSPEND_ABORTED))
			abort_and_exit("Aborting as requested.\n");

		pbe = get_next_pbe(pbe);
	}

	update_status(1, 1, NULL);

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		if (ret != Z_STREAM_END)
			printk(KERN_ERR "Final zlib return value wasn't stream_end!\n");
	}
#endif
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "|\n");
write_pageset_free_buffers:
	finish_all_io();
	
	end_time = jiffies;
	if ((end_time - start_time) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)))
		printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_MEDIUM, "Time to write data: %d jiffies. Pages written per jiffy: %d.\n", end_time - start_time, size / (end_time - start_time));

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		release_zlib_memory();
		free_pages(compression_buffer, 0);
	}
#endif
	PRINTFREEMEM("at end of write pageset");
	check_io_stats();
	if (nr_free_pages() != origfree) {
		abort_suspend("Number of free pages at start and end of write pageset don't match! (%d != %d)\n", origfree, nr_free_pages());
	}
	return result;
}

/*
 * read_pageset()
 *
 * Description: Read a pageset from disk.
 * @pagedir: Pointer to the pagedir to be saved.
 * @whichtowrite: Controls what debugging output is printed.
 * @overwrittenpagesonly: Whether to read the whole pageset or only part.
 * Returns: Zero on success or -1 on failure.
 *
 */

int read_pageset(struct pagedir * pagedir, int whichtoread, int overwrittenpagesonly)
{
	int nextupdate = 0, size, result = 0;
	int start_time, end_time, finish_at = 0;
	long i;
	char * dest_address = NULL;
	struct pbe * pbe;
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	struct pbe * checkpagepbe = NULL;
#endif
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	int ret = 0, ret2 = 0;
	z_stream strm;
	struct pbe * swap_pbe = NULL;
	unsigned long compression_buffer = 0;
	struct pbe readpbe;
#endif

#undef abort_and_exit
#define abort_and_exit(f, a...) do { \
			abort_suspend(f, ## a); \
			result = -1; \
			goto read_pageset_free_buffers; \
	} while(0)

	reset_io_stats();
	PRINTFREEMEM("at start of read pageset");

	size = pagedir->pageset_size;

	if (whichtoread == 1) {
		prepare_status(1, 1, "Reading suspend image (1)...");
	} else {
		prepare_status(1, 1, "Reading suspend image (2)...");
	}	
	beepOK;
	
	start_time=jiffies;

	if (overwrittenpagesonly && (whichtoread == 2))
		finish_at = MIN(pageset1_size, pageset2_size);
	else
		finish_at = pagedir->pageset_size;

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		compression_buffer = get_zeroed_page(GFP_ATOMIC);

		if (!compression_buffer) {
			printk(KERN_ERR name_suspend "Failed to allocate memory for compression buffer.\n");
			return -1;
		}
	
		if (prepare_zlib_inflate_memory()) {
			printk(KERN_ERR name_suspend "Failed to allocate memory to write the pageset.\n");
			free_pages(compression_buffer, 0);
			return -1;
		}
	
		strm.total_in = 0;
		strm.total_out = 0;
		strm.workspace = compression_workspace;
		strm.avail_in = 0;

		if (Z_OK != zlib_inflateInit(&strm)) {
			printk(KERN_ERR name_suspend "Failed to initialise zlib.\n");
			free_pages(compression_buffer, 0);
			goto read_pageset_free_buffers;
		}

		swap_pbe = pagedir->data;
	}
#endif
	pbe = pagedir->data;

	for (i=0; i< finish_at; i++) {
		if (!(i&0x1FF)) 
			printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_LOW, ".");
		if ((i >= nextupdate) || (!(i%(1 << (20 - PAGE_SHIFT)))))
			nextupdate = update_status(i, finish_at, " %d/%d MB ", MB(i+1), MB(finish_at));
		printnolog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, 1, name_suspend "Reading pageset: %d/%d ", i + 1, finish_at);

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
		if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
			dest_address = kmap_nonblock(pbe->address);
			if (!dest_address)
				abort_and_exit("Failed to kmap a page!");

			strm.next_out = dest_address;
			strm.avail_out = PAGE_SIZE;

			while ((strm.avail_out) && (ret != Z_STREAM_END) && (ret != Z_BUF_ERROR)) {
				if (!strm.avail_in) {
					/* Get next page of the compressed image */
					readpbe.swap_address = swap_pbe->swap_address;
					readpbe.swap_bdev = swap_pbe->swap_bdev;
					readpbe.block = swap_pbe->block;
					readpbe.address = virt_to_page(compression_buffer);
					ret2 = bdev_read_swap(&readpbe, 0);
					if (ret2)
						abort_and_exit("Failed reading page to swap. Error number was %d.\n", ret);
					strm.next_in = (char *) compression_buffer;
					strm.avail_in = PAGE_SIZE;
					swap_pbe = get_next_pbe(swap_pbe);
				}
				ret = zlib_inflate(&strm, Z_PARTIAL_FLUSH);
				if ( ret == Z_BUF_ERROR && strm.avail_in == 0 )
					continue;
				if (ret == Z_STREAM_END)
					zlib_inflateEnd(&strm);
				check_shift_keys(0, NULL);
				if ((ret != Z_OK) && (ret != Z_STREAM_END)) {
					zlib_inflateEnd(&strm);
					abort_and_exit("Failed to decompress data. Result %d received. Message is %s\n", ret, strm.msg);
				}
			}
			kunmap(pbe->address);
		} else {
#endif
			if (bdev_read_swap(pbe, 1)) {
				result = -EIO;
				abort_and_exit("Failed to read page %lx from swap to %lx!", pbe->swap_address.val, dest_address);
			}
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
		}
#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
		if ((unsigned long) page_address(pbe->origaddress) == checkpage)
			checkpagepbe = pbe;
#endif

		pbe = get_next_pbe(pbe);

		check_shift_keys(0, NULL);
	}

	update_status(1, 1, NULL);

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		/* If only reading part, still need to clean up */
		if ((i > finish_at) && (ret != Z_STREAM_END))
			zlib_inflateEnd(&strm);
	
		if (ret != Z_STREAM_END)
			printk(KERN_ERR "Final zlib return value wasn't stream_end!\n");
	}
#endif
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_LOW, "|\n");

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	if (checkpagepbe)
		printk(KERN_ERR "Page checksum for page %lx is %lx. For data loaded from %lx (to %p) is %lx.\n",
			checkpage,
			pagechecksum(checkpagepbe->origaddress),
			checkpagepbe->swap_address.val,
			checkpagepbe->address,
			pagechecksum(checkpagepbe->address));
#endif

read_pageset_free_buffers:
	finish_all_io();
	
	end_time=jiffies;
	if ((end_time - start_time) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)))
		printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_MEDIUM, "Time to read data: %d jiffies. Pages read per jiffy: %d.\n", end_time - start_time, finish_at / (end_time - start_time));

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	if (!TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION)) {
		release_zlib_memory();
		free_pages(compression_buffer, 0);
	}
#endif
	PRINTFREEMEM("at end of read pageset");
	check_io_stats();
	return result;
}

/*
 * write_pagedir_and_suspend_header()
 *
 * Description: Write the pagedir and suspend header to disk.
 * Returns: Zero on success or -1 on failure.
 *
 */

int  write_pagedir_and_suspend_header(void)
{
	int i, j, len, *ptr, nextupdate = 0, ret, arrayorder;
	swp_entry_t entry, prev = { 0 };
	int total = pagedir1.pageset_size+pagedir2.pageset_size+pagedir1.pagedir_size+2;
	int progress = total-1;
	//struct pbe pbearray[pagedir1.pagedir_size+1];
	struct pbe * pbearray;
	struct pbe pbeheader;
	struct pbe * thispagedirpage = pagedir1.data;
	union p_diskpage diskpage;

#undef abort_and_exit
#define abort_and_exit(f, a...) do { \
			abort_suspend(f, ## a); \
			goto write_pagedir_abort; \
	} while(0)
	
	arrayorder = get_bitmask_order(SUSPEND_PD_PAGES(pagedir1.pagedir_size + 1));
	pbearray = (struct pbe *) __get_free_pages(__GFP_FAST, arrayorder);
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_LOW, name_suspend "Writing pagedir (%d pages)\n", pagedir1.pagedir_size);

	for (i=0; i<(pagedir1.pagedir_size); i++) {
		printnolog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, 1, "%d/%d = %p", i+1, pagedir1.pagedir_size, thispagedirpage);
		pbearray[i].address = virt_to_page(thispagedirpage);
		printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_LOW, "."); 
		if (i >= nextupdate)
			nextupdate = update_status(progress + i, total, NULL);
		pbearray[i].swap_address = entry = get_swap_page(1);
		if (!entry.val)
			abort_and_exit("Not enough swapspace when writing pgdir\n");		
		get_phys_params(&pbearray[i]);
		check_shift_keys(0, NULL);
		if (TEST_RESULT_STATE(SUSPEND_ABORTED))
			abort_and_exit("Aborting as requested.\n");
		diskpage.address = (unsigned long) thispagedirpage;
		diskpage.pointer->link.next = prev;
		
		ret = write_swap_page(pbearray[i].swap_address, pbearray[i].address, 1);
		if (ret)
			abort_and_exit("Failed writing page to swap. Error number was %d.\n", ret);
		
		prev = entry;
		thispagedirpage = get_next_pagedir_page(thispagedirpage);
	}

	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "H");
#ifdef SOFTWARE_SUSPEND_MTRR
	ptr = mtrr_suspend();
	if(ptr)
		len = *ptr;
	else
#endif
	{
		len = 4;	/* we need to put a zero */
		ptr = NULL;
	}
	if (len + sizeof(struct suspend_header) > PAGE_SIZE-sizeof(swp_entry_t))
		abort_and_exit("Size of suspend header too big!\n");
	if (sizeof(union diskpage) != PAGE_SIZE)
		abort_and_exit("Size of a disk page is not PAGE_SIZE!\n");

	diskpage.address = get_free_page(GFP_ATOMIC);
	pbeheader.address = virt_to_page(diskpage.address);
	pbeheader.swap_address = get_swap_page(1);
	get_phys_params(&pbeheader);
	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto write_pagedir_abort;
	
	if (fill_suspend_header(&(diskpage.pointer->sh)))
		abort_and_exit("Out of memory while writing header!\n");
	if(ptr)
		memcpy(diskpage.ptr+sizeof(struct suspend_header), ptr, len);
	else {
		len = 0;
		ptr = &len;
		memcpy(diskpage.ptr+sizeof(struct suspend_header), ptr, sizeof(int));
	}	  
	flush_dcache_page(virt_to_page(pbeheader.address));
	diskpage.pointer->link.next = prev;

	swsusp_rw_swap_page(WRITE, &pbeheader);
	prev = pbeheader.swap_address;

	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "S");
	update_status(total - 1, total, NULL);
	mark_swapfiles(prev, MARK_SWAP_SUSPEND);
	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto write_pagedir_abort;
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "|\n");
	update_status(total, total, NULL);

	MDELAY(1000);
	finish_all_io();
	free_pages((unsigned long) diskpage.address, 0);
	free_pages((unsigned long) pbearray, arrayorder);
	return 0;

write_pagedir_abort:
	swap_free(entry);
	swap_free(prev);
	printk(KERN_ERR "Attempting to free %d pagedir entries in swap.\n", i);
	thispagedirpage = pagedir1.data;
	for (j=i-1; j >= 0; j--) {
		swap_free(((union diskpage *) thispagedirpage)->link.next);
		thispagedirpage = get_next_pagedir_page(thispagedirpage);
	}
	if (pbeheader.swap_address.val)
		swap_free(pbeheader.swap_address);
	finish_all_io();
	free_pages((unsigned long) diskpage.address, 0);
	free_pages((unsigned long) pbearray, arrayorder);
	return -1;
}

/*
 * Sanity check if this image makes sense with this kernel/swap context
 * I really don't think that it's foolproof but more than nothing..
 */

/*
 * write_pagedir_and_suspend_header()
 *
 * Description: Write the pagedir and suspend header to disk.
 * Returns: Zero on success or -1 on failure.
 *
 */


static int sanity_check_failed(char *reason)
{
	beepERR;
	printk(KERN_EMERG name_swsusp "%s\n",reason);
	printk(KERN_EMERG "BIG FAT WARNING!!\n");
	printk(KERN_EMERG "If you continue booting with this kernel, be sure to mkswap your swap\n");
	printk(KERN_EMERG "partition. Otherwise, you may severely damage your filesystem on next\n");
	printk(KERN_EMERG "reboot with the kernel that wrote the resume image.\n");
	printk(KERN_EMERG "Press SHIFT to reboot or CONTROL to continue booting with this kernel\n");
	while (!(shift_state & ((1 << KG_SHIFT) | (1 << KG_CTRL)))) 
		yield(); 
	if (shift_state & (1 << KG_SHIFT))
		machine_restart(NULL);
	/* Wait for release */
	while (shift_state) 
		yield(); 
	printk(KERN_EMERG "ARE YOU SURE: Unless you really know what you're doing just reboot on the correct kernel or use the noresume boot option.");
	printk(KERN_EMERG "Press SHIFT to reboot or CONTROL to continue booting with this kernel\n");
	while (!(shift_state & ((1 << KG_SHIFT) | (1 << KG_CTRL)))) 
		yield(); 
	if (shift_state & (1 << KG_SHIFT))
		machine_restart(NULL);
	/* Wait for release */
	while (shift_state) 
		yield(); 
	return -EPERM;
}

static int sanity_check(struct suspend_header *sh)
{
	if (sh->version_code != LINUX_VERSION_CODE)
		return sanity_check_failed("Incorrect kernel version");
	if (sh->num_physpages != num_physpages)
		return sanity_check_failed("Incorrect memory size");
	if (strncmp(sh->machine, system_utsname.machine, 65))
		return sanity_check_failed("Incorrect machine type");
	if (strncmp(sh->version, system_utsname.version, 65))
		return sanity_check_failed("Incorrect version");
	if (sh->num_cpus != smp_num_cpus)
		return sanity_check_failed("Incorrect number of cpus");
	if (sh->page_size != PAGE_SIZE)
		return sanity_check_failed("Incorrect PAGE_SIZE");
	return 0;
}

static int __read_primary_suspend_image(union p_diskpage diskpage, int noresume)
{				/* returned values:
				   -EINVAL in case of swapspace or if noresume is set and swap signature
				   wasn't correctly fixed (non fatal error),
				   0 if suspended image was correctly read or if noresume is set
				   and swap signature was correctly fixed,
				   other non zero values are errors. */
	swp_entry_t next;
	int i, len, *ptr;
	kdev_t header_block_device = 0;
	struct pbe * thispbepage;
	struct pbe * nextpbepage = NULL;

	if (bdev_read_page(resume_device, 0, virt_to_page(diskpage.ptr))) return -EIO;

	if ((!memcmp("SWAP-SPACE",diskpage.pointer->swh.magic.magic,10)) ||
	    (!memcmp("SWAPSPACE2",diskpage.pointer->swh.magic.magic,10))) {
		printk(KERN_ERR name_resume "This is normal swap space\n" );
		return -EINVAL;	/* non fatal error */
	}
	/* Put bdev of suspend header in last byte of swap header (unsigned short) */
	{
		kdev_t * header_ptr = (kdev_t *) (&(diskpage.pointer->swh.magic.magic[2]));
		header_block_device = *header_ptr;
	}

	next = diskpage.pointer->link.next;

	if (!memcmp("1R",diskpage.pointer->swh.magic.magic,2))
		memcpy(diskpage.pointer->swh.magic.magic,"SWAP-SPACE",10);
	else if (!memcmp("2R",diskpage.pointer->swh.magic.magic,2))
		memcpy(diskpage.pointer->swh.magic.magic,"SWAPSPACE2",10);
	else {
		printk("%sUnable to find suspended-data signature (%.10s - misspelled?\n", 
			name_resume, diskpage.pointer->swh.magic.magic);
		return -EINVAL;
		/* We want to abort_suspend even with noresume -- we certainly don't want to add
		   our signature into your ext2 filesystem ;-) */
	}
	flush_dcache_page(virt_to_page(diskpage.pointer));
	if (noresume)
		return(noresume_fix_signature(diskpage));
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, name_resume "Signature found, resuming\n");
	MDELAY(1000);
	
	/* 
	 * If the header is not on the resume_device, get the resume device first.
	 */
	if (header_block_device != resume_device) {
		int blksize = 0;
		if (!blksize_size[MAJOR(header_block_device)]) {
			printk(name_resume "%x: Blocksize not known?", header_block_device);
		} else blksize = blksize_size[MAJOR(header_block_device)][MINOR(header_block_device)];
		if (!blksize) {
			printk(name_resume "%x: Blocksize not set?", header_block_device);
			blksize = PAGE_SIZE;
		}
		set_blocksize(header_block_device, PAGE_SIZE);
	}

	if (bdev_read_page(header_block_device, SWP_OFFSET(next), virt_to_page(diskpage.ptr))) return -EIO;
	if (sanity_check(&(diskpage.pointer->sh))) /* Is this the same machine? */
		return -EPERM;
#ifdef SOFTWARE_SUSPEND_MTRR
	ptr = ((int *)(diskpage.ptr+sizeof(struct suspend_header)));
	len = *ptr;
	if(len)
		mtrr_resume(ptr);
#endif
	next = diskpage.pointer->link.next;

	base_mem_free = diskpage.pointer->sh.base_mem_free;
	swapforimage = diskpage.pointer->sh.swapforimage;
	pagedir_nosave.pageset_size = diskpage.pointer->sh.pagedir.pageset_size;
	pagedir_nosave.pagedir_size = diskpage.pointer->sh.pagedir.pagedir_size;
	expected_size1 = diskpage.pointer->sh.expected_size1;
	expected_size2 = diskpage.pointer->sh.expected_size2;
	swsusp_result = diskpage.pointer->sh.param0;
	if (!swsusp_act_used)
		swsusp_action = diskpage.pointer->sh.param1;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	if (!swsusp_dbg_used)
		swsusp_debug_state = diskpage.pointer->sh.param2;
#endif
	if (!swsusp_lvl_used)
		console_loglevel = diskpage.pointer->sh.param3;
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	checkpage = diskpage.pointer->sh.param4;
#endif
#if defined(CONFIG_FBCON_SPLASHSCREEN)
	fb_splash_progress[0] = diskpage.pointer->sh.progress0;
	fb_splash_progress[1] = diskpage.pointer->sh.progress1;
	fb_splash_progress[2] = diskpage.pointer->sh.progress2;
	fb_splash_progress[3] = diskpage.pointer->sh.progress3;
	fb_splash_progress[4] = diskpage.pointer->sh.progress4;
#endif
	suspend_task = current->pid;

	/*
	 * We don't need to save swsusp_state[4] because it's not used during resume
	 * and will be restored with the image anyway
	 */
	
	now_resuming = 1;
	if (prepare_suspend_console())
		printk(name_resume "Can't allocate a console... proceeding\n");

	/* 
	 * Swap hasn't been turned on yet, so we can safely store our kdevs in the original array.
	 * We're going to be reloading the same values anyway, so nothing will be broken by this.
	 */

	for (i = 0; i < MAX_SWAPFILES; i++) {
		kdev_t thisdevice = diskpage.pointer->sh.kdevs[i];
		int blksize = 0;

		printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 1, "Swap device %d is %x.", i, diskpage.pointer->sh.kdevs[i]);

		swap_info[i].swap_device = thisdevice;
		
		if (!thisdevice)
			continue;

		if (thisdevice == resume_device) {
			printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 0, "Resume root device %x", thisdevice);
			continue;
		}

		printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 0, "Resume secondary device %x", thisdevice);
		if (!blksize_size[MAJOR(thisdevice)]) {
			printk(name_resume "%x: Blocksize not known?", thisdevice);
		} else blksize = blksize_size[MAJOR(thisdevice)][MINOR(thisdevice)];
		if (!blksize) {
			printk(name_resume "%x: Blocksize not set?", thisdevice);
			blksize = PAGE_SIZE;
		}
		set_blocksize(thisdevice, PAGE_SIZE);
	}

	printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 1, "");

	check_shift_keys(1, "About to read pagedir. ");

	/* We get pages in reverse order of saving. Since this is the only place
	 * we would _really_ need the reverse part doubly linked, we just
	 * allocate the pages as we go. */
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,name_resume "Reading pagedir of %d pages, %lx", pagedir_nosave.pagedir_size, next.val);

	for (i=pagedir_nosave.pagedir_size-1; i>=0; i--) {
		thispbepage = (struct pbe *)get_zeroed_page(GFP_ATOMIC);
		if ((!thispbepage) || (!next.val)) {
			int j;
			thispbepage = nextpbepage;
			for (j = i; j < pagedir_nosave.pagedir_size - 1; j++) {
				nextpbepage = get_next_pagedir_page(thispbepage);
				free_page((unsigned long) thispbepage);
				thispbepage = nextpbepage;
			}
			if (!thispbepage)
				abort_suspend("Unable to allocate a pagedir.\n");
			else
				abort_suspend("next.val NULL in __read_primary_suspend_image!\n");
			//swsusp_spin_unlock_irq_restore(&suspend_pagedir_lock, pagedirflags);
			return -ENOMEM;
		}
		printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 0, "[sw]%lx -> [pd%d]%p.\n", next.val, i, thispbepage);
		diskpage.pointer = (union diskpage *) thispbepage;
		{
			kdev_t bdev = swap_info[SWP_TYPE(next)].swap_device;
			int pos = SWP_OFFSET(next); 
			if (bdev_read_page(bdev, pos, virt_to_page(diskpage.ptr))) return -EIO;
		}
		next = diskpage.pointer->link.next;
		if (!i)
			pagedir_nosave.data = thispbepage;
		set_next_pagedir_page(thispbepage, nextpbepage);
		nextpbepage = thispbepage;
	}
	if (next.val) {
		abort_suspend("next.val not NULL in __read_primary_suspend_image!\n");
		return -EPERM;
	}

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "Pagedir_nosave.data is %p\n", pagedir_nosave.data);
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "Pagedir_nosave.pageset_size is %d\n", pagedir_nosave.pageset_size);

 	warmup_collision_cache();
 
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "Attempting to relocate pagedir with %d pages free.\n", nr_free_pages());
	if (allocatemap(&pagedirmap, 1))
		return -ENOMEM;
	
	if (relocate_pagedir())
		return -ENOMEM;
	if (check_pagedir())
		return -ENOMEM;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,name_resume "Reading primary pageset data (%d pages). Pagedir at %lx\n", pagedir_nosave.pageset_size, (unsigned long) pagedir_nosave.data);
	
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\n");

	check_shift_keys(1, "About to read pageset 1. ");

	if (read_pageset(&pagedir_nosave, 1, 0))
		return -EPERM;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\n|\n");
	
	update_status(100, 100, NULL);
	
	check_shift_keys(1, "About to restore original kernel. ");

	return 0;
}

int read_primary_suspend_image(char * specialfile, int noresume)
{
	union p_diskpage cur;
	unsigned long scratch_page = 0;
	int error, blksize = 0;

	beepOK;
	resume_device = name_to_kdev_t(specialfile);
	scratch_page = get_zeroed_page(GFP_ATOMIC);
	if (scratch_page) {
		cur.address = scratch_page;
		printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, name_resume "Resuming from device %x\n", resume_device);
		
		if (!blksize_size[MAJOR(resume_device)]) {
			printk(name_resume "Blocksize not known?\n");
		} else blksize = blksize_size[MAJOR(resume_device)][MINOR(resume_device)];
		if (!blksize) {
			printk(name_resume "Blocksize not set?\n");
			blksize = PAGE_SIZE;
		}
		set_blocksize(resume_device, PAGE_SIZE);
		error = __read_primary_suspend_image(cur, noresume);
		free_page(scratch_page);
	} else error = -ENOMEM;

	switch (error) {
		case 0:
			beepOK;
		case -EINVAL:	/* non fatal error */
			set_blocksize(resume_device, blksize);
			software_suspend_enabled = 1;
			MDELAY(1000);
			return(error);
			break;
		case -EIO:
			printk(KERN_CRIT name_resume "I/O error\n");
			beepERR;
			break;
		case -ENOENT:
			printk(KERN_CRIT name_resume "%s: No such file or directory\n", specialfile);
			break;
		case -EPERM:
			printk(KERN_CRIT name_resume "Sanity check error\n");
			beepERR;
			break;
		default:
			printk(KERN_CRIT name_resume "Error %d resuming\n", error);
			beepERR;
			break;
	}
	__read_primary_suspend_image(cur, 1); /* Try to fix the swap signature */
	abort_suspend("Error %d in read_primary_suspend_image",error);
	return(error);
}

int read_secondary_pagedir(int overwrittenpagesonly)
{
	int result = 0;

	if (!pageset2_size)
		return 0;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "Beginning of read_secondary_pagedir: "); 

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,name_resume "Reading secondary pageset data (%d pages). Pagedir at %lx.\n", pageset2_size, (unsigned long) pagedir2.data);

	result = read_pageset(&pagedir2, 2, overwrittenpagesonly);

	update_status(100, 100, NULL);
	check_shift_keys(1, "Pagedir 2 read. ");

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\n");
	return result;
}

