/*
 * linux/drivers/swsusp/io.c
 *
 * This file contains IO routines for swsusp.
 *
 */

#define SWSUSP_LOWLEVEL_IO_C

#include "common.h"

extern int max_async_ios;

/* Exported variables */
struct page * last_suspend_cache_page;	/* Pointer to the last cache page */

#define BUFFER_BUSY_BITS ((1<<BH_Dirty) | (1<<BH_Lock))
#define buffer_busy(bh)	(atomic_read(&(bh)->b_count) | \
			((bh)->b_state & BUFFER_BUSY_BITS))

#define BREAD_BUFFER_HEAD(dev, pos) \
	bh = bread(dev, pos, PAGE_SIZE); \
	if (!bh || (!bh->b_data)) { \
		return -EIO; \
	}

#define BFREE(bh) \
	if (!buffer_busy(bh)) \
		try_to_free_buffers(bh->b_page, __GFP_FAST); \
	free_suspend_cache_page(last_suspend_cache_page); \
	last_suspend_cache_page = NULL;

#define PANIC(f, a...) 	panic(f, ## a) 

struct io_info {
	struct buffer_head * bh;
	struct page * cache_page;
	struct page * page;
	char flags;
	pte_t * pte;
	int pageprot;
	struct page * data_page;
	swp_entry_t swap_address;
	kdev_t dev;
	int zones[PAGE_SIZE/512];
	int zones_used;
	int block_size;
	struct io_info * next;
};
static struct io_info * io_info = NULL;
static int infopages = 0;
static int outstanding_io = 0;
static int pages_allocated = 0;
static int pages_freed = 0;

#define IO_IN_USE 0
#define IO_WRITING 1
#define IO_RESTORE_PAGE_PROT 2
#define IO_AWAITING_READ 3
#define IO_AWAITING_WRITE 4
#define IO_CLEANUP_IN_PROGRESS 5
#define IO_HANDLE_PAGE_PROT 6

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
extern unsigned long pagechecksum(struct page * page);
extern unsigned long checkpagecopy;
extern unsigned long checkpage;
#endif

/*
 * This is our sync function. With this solution we probably won't sleep
 * but that should not be a problem since tasks are stopped..
 */

void do_suspend_sync(void)
{
	while (1) {
		run_task_queue(&tq_disk);
		if (!TQ_ACTIVE(tq_disk))
			break;
		printk(KERN_ERR "Hm, tq_disk is not empty after run_task_queue\n");
	}
}

/* We memorize in swapfile_used what swap devices are used for suspension */
#define SWAPFILE_UNUSED    0
#define SWAPFILE_SUSPEND   1	/* This is the suspending device */
#define SWAPFILE_IGNORED   2	/* Those are other swap devices ignored for suspension */

static unsigned short swapfile_used[MAX_SWAPFILES];
static unsigned short root_swap;

static int swsusp_writepage(struct page *page);

static inline int sync_page(struct page *page)
{
	struct address_space *mapping = page->mapping;

	if (mapping && mapping->a_ops && mapping->a_ops->sync_page)
		return mapping->a_ops->sync_page(page);
	return 0;
}

static struct address_space_operations swsusp_aops = {
	writepage: swsusp_writepage,
	sync_page: block_sync_page,
};

struct address_space swsusp_space = {
	LIST_HEAD_INIT(swsusp_space.clean_pages),
	LIST_HEAD_INIT(swsusp_space.dirty_pages),
	LIST_HEAD_INIT(swsusp_space.locked_pages),
	0,				/* nrpages	*/
	&swsusp_aops,
};

int get_phys_params(struct pbe * pbe)
{
	unsigned long offset;
	struct inode *swapf = 0;

	get_swaphandle_info(pbe->swap_address, &offset, &pbe->swap_bdev, &swapf);
	if (pbe->swap_bdev) {
		pbe->block = offset;
		/* We are assuming this.
		io_info->zones_used = 1;
		io_info->block_size = PAGE_SIZE;
		*/
	} else if (swapf) {
#if 0
		int i, j;
		unsigned int block = offset
			<< (PAGE_SHIFT - swapf->i_sb->s_blocksize_bits);
#endif
		if (swapf->i_sb->s_blocksize != PAGE_SIZE) {
			abort_suspend("Get_phys_params: Error! Blocksize (%d) is not pagesize!", swapf->i_sb->s_blocksize);
			return 0;
		}
		if (!(pbe->block = bmap(swapf,0))) {
			abort_suspend("Get_phys_params: bad swap file!");
			return 0;
		}
#if 0
		io_info->block_size = swapf->i_sb->s_blocksize;
		for (i=0, j=0; j< PAGE_SIZE ; i++, j += io_info->block_size)
			if (!(io_info->zones[i] = bmap(swapf,block++))) {
				abort_suspend("get_phys_params: bad swap file");
				return 0;
			}
		io_info->zones_used = i;
		io_info->dev = swapf->i_dev;
		check_shift_keys(1, "Using swapfile entry %lx -> %lx + %lx\n",
				pbe->swap_address,
				pbe->swap_bdev,
				pbe->block);
		block_dump = 1;
#endif
	} else {
		return 0;
	}

	return 1;
}

int get_brw_params(struct io_info * io_info)
{
	unsigned long offset;
	struct inode *swapf = 0;

	get_swaphandle_info(io_info->swap_address, &offset, &io_info->dev, &swapf);
	if (io_info->dev) {
		io_info->zones[0] = offset;
		io_info->zones_used = 1;
		io_info->block_size = PAGE_SIZE;
	} else if (swapf) {
		int i, j;
		unsigned int block = offset
			<< (PAGE_SHIFT - swapf->i_sb->s_blocksize_bits);

		io_info->block_size = swapf->i_sb->s_blocksize;
		for (i=0, j=0; j< PAGE_SIZE ; i++, j += io_info->block_size)
			if (!(io_info->zones[i] = bmap(swapf,block++))) {
				abort_suspend("get_brw_params: bad swap file");
				return 0;
			}
		io_info->zones_used = i;
		io_info->dev = swapf->i_dev;
	} else {
		return 0;
	}

	return 1;
}

extern void end_buffer_io_sync(struct buffer_head * bh, int uptodate);

void cleanup_write_swsusp_page_async(struct io_info * io_info)
{
	struct buffer_head *bh = io_info->bh;
	struct page * page = bh->b_page;
	int origfree = nr_free_pages();

	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_HIGH, "Cleanup_write: [%p]\n", io_info);
	if (test_and_set_bit(IO_CLEANUP_IN_PROGRESS, &io_info->flags))
		return;

	TryLockPage(page); /* Ensure it's locked */

	if (!block_flushpage(page, 0))
		PAGE_BUG(page);
	
	page->mapping = NULL;
	page->index = 0;
	UnlockPage(page);

	if (atomic_read(&page->count) != 2)
		printk("Cleanup write: Page count is %d. Not good!\n", atomic_read(&page->count));
	atomic_dec(&page->count);
	__free_pages(page, 0);

	outstanding_io--;
	io_info->bh = NULL;
	io_info->flags = 0;
	pages_freed += nr_free_pages() - origfree;
	if ((origfree + 1) != nr_free_pages())
		printk("Cleanup write: Didn't free 1 page (%d->%d).\n", origfree, nr_free_pages());
}

void cleanup_read_swsusp_page_async(struct io_info * io_info)
{
	struct page *data_page = io_info->data_page, * bpage;
	struct buffer_head * bh = io_info->bh;
	pte_t * pte = NULL;
	int pageprot = 0;
	char * destaddress;
	
	if (test_and_set_bit(IO_CLEANUP_IN_PROGRESS, &io_info->flags))
		return;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_HIGH, "Cleanup_read: [%p:%p %p]", io_info, data_page, bh);
	clear_bit(IO_RESTORE_PAGE_PROT, &io_info->flags);
	
	// Remove any page protection while we restore contents
	if (test_bit(IO_HANDLE_PAGE_PROT, &io_info->flags)) {
		pte = lookup_address((unsigned long) page_address(io_info->data_page));
		printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "pte %p", pte);
		if (pte) {
			pageprot = pte->pte_low & ~_PAGE_CHG_MASK;
			if (pageprot != pgprot_val(PAGE_KERNEL_NOCACHE)) {
				set_pte(pte, pte_modify(*pte, PAGE_KERNEL_NOCACHE));	
				set_bit(IO_RESTORE_PAGE_PROT, &io_info->flags);
			}
		}
	}
	io_info->pte = pte;
	io_info->pageprot = pageprot;
	
	destaddress=kmap(data_page);
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "destaddr: %p", destaddress);
	memcpy(destaddress, bh->b_data, PAGE_SIZE);
	flush_dcache_page(data_page);
	kunmap(data_page);

	if (test_bit(IO_RESTORE_PAGE_PROT, &io_info->flags))
		set_pte(io_info->pte, pte_modify(*(io_info->pte), __pgprot(io_info->pageprot)));

	put_bh(bh);
	brelse(bh);
	bpage = bh->b_page;
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "bpage: %p", bpage);

	if (!buffer_busy(bh))
		try_to_free_buffers(bpage, __GFP_FAST);
	free_suspend_cache_page(io_info->cache_page);
	io_info->cache_page = NULL;

	io_info->bh = NULL;
	io_info->flags = 0;
	outstanding_io--;
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_HIGH, "\n");
}

int swsusp_brw_page(int rw, struct page *page, kdev_t dev, int b[], int size)
{
	struct buffer_head *head, *bh;

	if (!PageLocked(page))
		panic("swsusp_brw_page: page not locked for I/O");

	if (!page->buffers)
 		create_empty_buffers(page, dev, size);

	head = bh = page->buffers;

	do {
		lock_buffer(bh);
		bh->b_blocknr = *(b++);
		if (rw == WRITE)
			__mark_buffer_dirty(bh);
		set_bit(BH_Mapped, &bh->b_state);
		set_bit(BH_Uptodate, &bh->b_state) ;
		set_buffer_async_io(bh);
		bh = bh->b_this_page;
	} while (bh != head);

	do {
		struct buffer_head *next = bh->b_this_page;
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUMS
		if ((unsigned long) page_address(page) == checkpagecopy)
			printk(KERN_ERR "Page checksum in swsusp_brw_page is %lx.\n",
				pagechecksum(page));
#endif
		submit_bh(rw, bh);
		bh = next;
	} while (bh != head);

	return 0;
}

void cleanup_finished_swsusp_io(void)
{
	struct io_info * this = io_info, * firstfree = NULL;

	this = io_info;
	firstfree = NULL;
	
	/* We clean up all we can at once so we don't only batch the first lot */
	while (this) {
		if ((this->bh) && (this->bh->b_page) && (!buffer_locked(this->bh)) && (!PageLocked(this->bh->b_page))) {
			if (test_bit(IO_AWAITING_READ, &this->flags)) 
				cleanup_read_swsusp_page_async(this);
			else if (test_bit(IO_AWAITING_WRITE, &this->flags))
				cleanup_write_swsusp_page_async(this);
		}
		if ((!test_bit(IO_IN_USE, &this->flags)) && (!firstfree))
				firstfree = this;
		this = this->next;
	}
}

static struct io_info * get_io_info_struct(void)
{
	struct io_info * this = io_info, * last = NULL, * firstfree = NULL;
	unsigned long newpage = 0;
	int iteration = 0;

try_again:
	this = io_info;
	last = NULL;
	firstfree = NULL;
	if (!(iteration%8))
		run_task_queue(&tq_disk);
	iteration++;
	
	/* We clean up all we can at once so we don't only batch the first lot */
	while (this) {
		if ((this->bh) && (this->bh->b_page) && (!buffer_locked(this->bh)) && (!PageLocked(this->bh->b_page))) {
			if (test_bit(IO_AWAITING_READ, &this->flags)) 
				cleanup_read_swsusp_page_async(this);
			else if (test_bit(IO_AWAITING_WRITE, &this->flags))
				cleanup_write_swsusp_page_async(this);
		}
		if ((!test_bit(IO_IN_USE, &this->flags)) && (!firstfree)) 
			firstfree = this;
		last = this;
		this = this->next;
	}

	if ((max_async_ios) && (outstanding_io >= max_async_ios))
		goto try_again;

	if (firstfree) {
		this = firstfree;
		goto finish;
	}

	/* If not, is there more space in the page already allocated? */
	if (last) {
		int remaining = PAGE_SIZE - ((unsigned long) last & ~PAGE_MASK) - sizeof(struct io_info);
		if (remaining >= (sizeof(struct io_info))) {
			this = (struct io_info *) ((unsigned long) last + sizeof(struct io_info));
			last->next = this;
			this->next=NULL;
			goto finish;
		}
	}

	/* Need to allocate a new page */
	newpage = get_zeroed_page(GFP_ATOMIC);
	printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE, "[NewIOPage %lx]", newpage);
	if (!newpage)
		goto try_again;
	infopages++;
	this = (struct io_info *) newpage;
	this->next = NULL;
	SetPageNosave(virt_to_page(newpage));
	if (last)
		last->next = this;
	else
		io_info = this;
finish:
	if (test_and_set_bit(IO_IN_USE, &this->flags))
		goto try_again;
	return this;
}

/*
 * Returns zero when no io_info structs left.
 */
static int free_io_info_structs(void)
{
	struct io_info * this = io_info, * last = NULL, * pageprev = NULL;
	unsigned long pageaddress = 0;
	int pageused = 0, numleft = 0;

	pageaddress = (unsigned long) this;
	while (this) {
		do { /* Forward at least one */
			if ((this->bh) && (this->bh->b_page) && (!buffer_locked(this->bh)) && (!PageLocked(this->bh->b_page))) {
				if (test_bit(IO_AWAITING_READ, &this->flags)) 
					cleanup_read_swsusp_page_async(this);
				else if (test_bit(IO_AWAITING_WRITE, &this->flags))
					cleanup_write_swsusp_page_async(this);
			}
			if (test_bit(IO_IN_USE, &this->flags)) {
				pageused = 1;
				numleft++;
			}
			last = this;
			this = this->next;
		} while ((this) && (pageused) && ((((unsigned long) this) & PAGE_MASK) == pageaddress)); 
		if ((((unsigned long) this) & PAGE_MASK) != pageaddress) { /* At start of new page? */
			if (!pageused && (!in_interrupt())) {
				if (!pageprev) {
					io_info = this;
				}
				else
					pageprev->next = this;
				ClearPageNosave(virt_to_page(pageaddress));
				free_pages(pageaddress, 0);
				infopages--;
			} else
				pageprev = last;
			pageaddress = (unsigned long) this;
		}
	}
	return numleft;
}

void __inline reset_io_stats(void)
{
	outstanding_io = 0;
	infopages = 0;
	pages_allocated = 0;
	pages_freed = 0;
}

void check_io_stats(void)
{
	if (outstanding_io)
		printk("Outstanding_io after writing is %d.\n", outstanding_io);
	if (infopages)
		printk("Info pages is %d.\n", infopages);
	if (pages_allocated != pages_freed)
		printk("Pages allocated is %d. Freed is %d.\n", pages_allocated, pages_freed);
}

void sync_swap_partitions(void)
{
	int i;
	
	swap_list_lock();
	for(i=0; i<MAX_SWAPFILES; i++) {
		if (swapfile_used[i] == SWAPFILE_SUSPEND) {
			sync_buffers(swap_info[i].swap_device, 2);
		}
	}
	swap_list_unlock();
}

void finish_all_io(void)
{
	int numleft;

	check_shift_keys(0, NULL);
	while ((numleft = free_io_info_structs())) {
		sync_swap_partitions();
		do_suspend_sync();
	}
}

static int swsusp_writepage(struct page *page)
{
	struct io_info * io_info = (struct io_info *) page->index;
	
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "Writepage %p.", page);
#if 1
	if (remove_exclusive_swap_page(page)) {
		printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "Unlocking page %p. (1)", page);
		UnlockPage(page);
		return 0;
	}
#endif
	get_brw_params(io_info);
 	swsusp_brw_page(WRITE, page, io_info->dev, io_info->zones, io_info->block_size);
	return 0;
}

static int start_write_swsusp_page_async(swp_entry_t swap_address, struct page * data_page)
{
	struct io_info * io_info = get_io_info_struct();
	unsigned long buffer_virt;
	char * to, * from;
	struct page * buffer_page;

	if (!io_info)
		return 1;

	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_HIGH, "Start_write: [%p]", io_info);
	while (!(buffer_virt = get_zeroed_page(GFP_ATOMIC)))
		cleanup_finished_swsusp_io();
	pages_allocated++;
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "[WRITE BUFFER]->%d", nr_free_pages());
	buffer_page = virt_to_page(buffer_virt);
	
	io_info->swap_address = swap_address;
	io_info->data_page = data_page;
	io_info->page = buffer_page;

	set_bit(IO_WRITING, &io_info->flags);

	to = (char *) buffer_virt;
	from = kmap(data_page);
	memcpy(to, from, PAGE_SIZE);
	flush_dcache_page(data_page);
	flush_dcache_page(buffer_page);
	kunmap(data_page);

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	if (from == (char *) checkpagecopy)
		printk(KERN_ERR "Page checksum for copy of page %lx (%p) is %lx. For buffer (%p) it is %lx.\n",
			checkpage,
			data_page,
			pagechecksum(data_page),
			buffer_page,
			pagechecksum(buffer_page));
#endif

	atomic_inc(&buffer_page->count);
	lock_page(buffer_page);

	BUG_ON(buffer_page->mapping);
	buffer_page->mapping = &swsusp_space;
	buffer_page->index = (unsigned long) io_info;

	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_VERBOSE, "-> (PRE BRW) %d\n", nr_free_pages());
	get_brw_params(io_info);
 	swsusp_brw_page(WRITE, buffer_page, io_info->dev, io_info->zones, io_info->block_size);
	
	io_info->bh = buffer_page->buffers;
	io_info->cache_page = last_suspend_cache_page;
	last_suspend_cache_page = NULL;
	
	set_bit(IO_AWAITING_WRITE, &io_info->flags);
	outstanding_io++;

	return 0;
}
	
static int start_read_async(kdev_t bdev, long pos, struct page * data_page)
{
	struct io_info * io_info = get_io_info_struct();
	struct buffer_head *bh = NULL;
	unsigned int major;
	int correct_size;
	unsigned long flags;

	if (!io_info)
		return 1;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_HIGH, "Start_read: [%p]", io_info);

	io_info->data_page = data_page;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, " {%x %lx <-> %p ", bdev, pos, data_page);
 	bh = getblk(bdev, pos, PAGE_SIZE);
	
	io_info->bh = bh;
	
	flags = bh->b_page->flags & ~(1 << BH_Uptodate | 1 << BH_Dirty);

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "via %p/%p ", io_info, bh->b_page);
	bh->b_page->index = (unsigned long) io_info;
	
	major = MAJOR(bdev);

	/* Determine correct block size for this device. */
	correct_size = get_hardsect_size(bdev);

	/* Verify requested block sizes. */
	if (bh->b_size % correct_size) {
		printk(KERN_EMERG "start_read_swsusp_page: device %s: "
		       "only %d-char blocks implemented (%u)\n",
		       kdevname(bdev),
		       correct_size, bh->b_size);
		goto sorry;
	}

	lock_buffer(bh);
	atomic_inc(&bh->b_count);
	set_buffer_async_io(bh);
	atomic_set_buffer_clean(bh);
	__mark_buffer_clean(bh);
	
	lock_page(bh->b_page);

	submit_bh(READ, bh);

	if (last_suspend_cache_page)
		io_info->cache_page = last_suspend_cache_page;
	else
		io_info->cache_page = 0;
	last_suspend_cache_page = NULL;

	set_bit(IO_AWAITING_READ, &io_info->flags);
	outstanding_io++;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_HIGH, "\n");

	return 0;

sorry:
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_ERROR, "Start_read: Sorry!\n");
	/* Make sure we don't get infinite dirty retries.. */
	mark_buffer_clean(bh);

	clear_bit(IO_AWAITING_READ, &io_info->flags);

	return -EIO;
}

/*
 * swsusp_rw_swap_page:
 * Based on __rw_swap_page_nolock.
 * Asynchronously reads or writes a page to swap.
 * Used to read/write swap header.
 */
int swsusp_rw_swap_page(int rw, struct pbe * pbe)
{
	struct page * page = pbe->address, * cache_page;
	swp_entry_t entry = pbe->swap_address;

	atomic_inc(&page->count);
	lock_page(page);
	if (page->mapping)
		PAGE_BUG(page);

	/* needs sync_page to wait I/O completation */
	page->mapping = &swapper_space;
	last_suspend_cache_page = NULL;

	if (rw_swap_page_base(rw, entry, page))
		lock_page(page);
	else
		return 1;	/* Invalid entry */
	
	if (!block_flushpage(page, 0))
		PAGE_BUG(page);
	page->mapping = NULL;
	UnlockPage(page);

	if (last_suspend_cache_page)
		cache_page = last_suspend_cache_page;
	else
		cache_page = 0;

	sync_page(page);
	if ((rw == READ) && (cache_page))
		free_suspend_cache_page(cache_page);
	atomic_dec(&page->count);

	return 0;
}

void read_swapfiles(void) /* This is called before saving image */
{
	int i, len;
	
	len=strlen(resume_file);
	root_swap = 0xFFFF;
	
	swap_list_lock();
	for(i=0; i<MAX_SWAPFILES; i++) {
		if (swap_info[i].flags == 0) {
			swapfile_used[i]=SWAPFILE_UNUSED;
		} else {
			if (!len) {
	    			printk(KERN_ERR "resume= option should be used to set suspend device" );
				if (root_swap == 0xFFFF) {
					swapfile_used[i] = SWAPFILE_SUSPEND;
					root_swap = i;
				} else
					swapfile_used[i] = SWAPFILE_IGNORED;				  
			} else {
	  			/* we ignore all swap devices that are not the resume_file */
				if (resume_device == swap_info[i].swap_device) {
					swapfile_used[i] = SWAPFILE_SUSPEND;
					root_swap = i;
				} else {
				  	swapfile_used[i] = SWAPFILE_IGNORED;
				}
			}
		}
	}
	swap_list_unlock();
}

/*
 * mark_swapfiles:
 * Change the swapfile signature as appropriate.
 */

void mark_swapfiles(swp_entry_t prev, int mode)
{
	struct pbe pbe;
	char * signature;
	unsigned long pageaddress;

	PRINTFREEMEM("at start of mark_swapfiles");
	
	pageaddress = get_zeroed_page(GFP_ATOMIC);
	pbe.address = virt_to_page(pageaddress);
	signature = ((union p_diskpage) pageaddress).pointer->swh.magic.magic;

	pbe.swap_address = SWP_ENTRY(root_swap, 0);
	if (swsusp_rw_swap_page(READ, &pbe)) {
		free_pages(pageaddress, 0);
		return;
	}

	switch(mode) {
	case MARK_SWAP_RESUME:
		if (!memcmp("1R",signature,2))
			memcpy(signature,"SWAP-SPACE",10);
		else if (!memcmp("2R",signature,2))
			memcpy(signature,"SWAPSPACE2",10);
		else printk(KERN_ERR name_resume "Unable to find suspended-data signature (%.10s - misspelled?\n", signature);
		break;
	case MARK_SWAP_RECOVER:
		if (!memcmp("1R",signature,2))
			memcpy(signature,"SWAP-SPACE",10);
		else if (!memcmp("2R",signature,2))
			memcpy(signature,"SWAPSPACE2",10);
	  	else if ((!memcmp("SWAP-SPACE",signature,10))
			 &&(!memcmp("SWAPSPACE2",signature,10))) {
			abort_suspend("\nSwapspace is altered (%.10s)\n", signature);
			free_pages(pageaddress, 0);
			return;
		}
		break;
	case MARK_SWAP_SUSPEND:
		printk(KERN_ERR "Signature is %.10s\n", signature);
	  	if ((!memcmp("SWAP-SPACE",signature,10)))
		  	memcpy(signature,"1R",2);
		else if ((!memcmp("SWAPSPACE2",signature,10)))
			memcpy(signature,"2R",2);
		else {
			abort_suspend("\nSwapspace is not swapspace (%.10s)\n", signature);
			free_pages(pageaddress, 0);
			return;
		}

		/* Put bdev of suspend header in swap header (unsigned short) */
		{
			kdev_t * header_ptr = (kdev_t *) (&(signature[2]));
			*header_ptr = swap_info[SWP_TYPE(prev)].swap_device;
		}
		((union p_diskpage) pageaddress).pointer->link.next = prev; /* prev is the first/last swap page of the resume area */
		/* link.next lies in last 4 bytes of magic */
		break;
	}
	flush_dcache_page(virt_to_page(pageaddress)) ;
	swsusp_rw_swap_page(WRITE, &pbe);
	PRINTFREEMEM("before freeing pbe page in mark_swapfiles");
	free_pages(pageaddress, 0);
	PRINTFREEMEM("after freeing pbe page in mark_swapfiles");
	sync_swap_partitions();
	do_suspend_sync();
}

int bdev_read_page(kdev_t bdev, long pos, struct page * page)
{
	struct buffer_head *bh;
	char * destaddress;
	BREAD_BUFFER_HEAD(bdev, pos);
	destaddress = kmap(page);
	memcpy(destaddress, bh->b_data, PAGE_SIZE);
	flush_dcache_page(page) ;
	kunmap(page);
	if (!buffer_uptodate(bh)) {
		abort_suspend("buffer_uptodate false in bdev_read_page!\n");
		brelse(bh);
		BFREE(bh);
		return -1;
	}
	brelse(bh);
	BFREE(bh);
	return 0;
} 

static int rw_swap_page_sync(int rw, swp_entry_t swap_address, struct page * data_page)
{
	struct io_info io_info;
	unsigned long buffer_address;
	struct page *page = NULL;
	void * pageaddress;
	pte_t * pte;
	int pageprot = 0;
	int restore = 0;

	buffer_address = get_zeroed_page(__GFP_FAST);
	page = virt_to_page(buffer_address);

	io_info.page = page;

	atomic_inc(&page->count);
	atomic_inc(&page->count);
	lock_page(page);

	BUG_ON(page->mapping);
	page->mapping = &swapper_space;

	if (rw == WRITE) {
		pageaddress = kmap(data_page);
		memcpy((char *) buffer_address, (char *) pageaddress, PAGE_SIZE);
		flush_dcache_page(data_page) ;
		kunmap(data_page);
	}

	if (rw_swap_page_base(rw, swap_address, page))
		lock_page(page);

	if (last_suspend_cache_page)
		io_info.cache_page = last_suspend_cache_page;
	else
		io_info.cache_page = 0;
	last_suspend_cache_page = NULL;

	if (rw == WRITE) {
		if (!block_flushpage(page, 0))
			PAGE_BUG(page);
	}
	sync_page(page);

	page->mapping = NULL;
	UnlockPage(page);

	if (rw == READ) {
		pageaddress = kmap(data_page);

		// Remove any page protection while we restore contents
		pte = lookup_address((unsigned long) pageaddress);
		if (pte) {
			pageprot = pte->pte_low & ~_PAGE_CHG_MASK;
			if (pageprot != pgprot_val(PAGE_KERNEL_NOCACHE)) {
				set_pte(pte, pte_modify(*pte, PAGE_KERNEL_NOCACHE));	
				restore = 1;
			}
		}
		
		memcpy((char *) pageaddress, (char *) buffer_address, PAGE_SIZE);
		flush_dcache_page(virt_to_page(data_page));

		if (restore)
			set_pte(pte, pte_modify(*(pte), __pgprot(pageprot)));

		kunmap(data_page);
	}

	if ((rw == READ) && (io_info.cache_page))
		free_suspend_cache_page(io_info.cache_page);
	atomic_dec(&page->count);
	atomic_dec(&page->count);
	free_pages(buffer_address, 0);

	return 0;
}

int bdev_read_swap(struct pbe * pbe, int asyncio)
{
#if 0
	kdev_t bdev = swap_info[SWP_TYPE(pbe->swap_address)].swap_device;
	int pos = SWP_OFFSET(pbe->swap_address); 
#endif
	if ((TEST_ACTION_STATE(SUSPEND_NO_ASYNC_READS)) || (!asyncio))
		return bdev_read_page(pbe->swap_bdev, pbe->block, pbe->address);
	else
		return start_read_async(pbe->swap_bdev, pbe->block, pbe->address);
} 

int write_swap_page(swp_entry_t posn, struct page * page, int asyncio)
{
	if ((TEST_ACTION_STATE(SUSPEND_NO_ASYNC_WRITES)) || (!asyncio))
		return rw_swap_page_sync(WRITE, posn, page);
	else
		return start_write_swsusp_page_async(posn, page);
} 

int noresume_fix_signature(union p_diskpage cur)
{
	/* 
	 * We don't do a sanity check here: we want to restore the swap 
	 * whatever version of kernel made the suspend image
	 */
	struct buffer_head *bh;
	printk(name_resume "%s: Fixing swap signatures...\n", resume_file);
				/* We need to write swap, but swap is *not* enabled so
				   we must write the device directly */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE,"Swap signature will be set to %10s\n",cur.pointer->swh.magic.magic);
	BREAD_BUFFER_HEAD(resume_device, 0);
	if (is_read_only(bh->b_dev)) {
		printk(KERN_ERR name_resume "Can't write to read-only device %s\n",
		       kdevname(bh->b_dev));
		BFREE(bh);
		return -EIO;
	}
	
	//flush_dcache_page(virt_to_page(bh->b_data));
	memcpy(bh->b_data, cur.ptr, PAGE_SIZE);
	//generic_make_request(WRITE, bh);
	mark_buffer_dirty(bh);
	ll_rw_block(WRITE, 1, &bh);
	wait_on_buffer(bh);
	sync_swap_partitions();
	do_suspend_sync();
	if (buffer_uptodate(bh)) {
		brelse(bh);
		BFREE(bh);
#if 0
		//free_suspend_cache_page(last_suspend_cache_page);
		/* As with fix_signature, try to read to ensure its written. */
		BREAD_BUFFER_HEAD(resume_device, 0);
		brelse(bh);
		BFREE(bh);
#endif
		return 0;
	}
	printk(name_resume "Warning %s: Fixing swap signatures unsuccessful...\n", resume_file);
	bforget(bh);
	BFREE(bh);
	return -EINVAL;		/* non fatal error */
}

