/*
 * linux/kernel/suspend.c
 *
 * This file is to realize architecture-independent
 * machine suspend feature using pretty near only high-level routines
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * We'd like to thank the following people for their work:
 * 
 * Pavel Machek <pavel@ucw.cz>:
 * Modifications, defectiveness pointing, being with me at the very beginning,
 * suspend to swap space, stop all tasks. Port to 2.4.18-ac and 2.5.17.
 *
 * Steve Doddi <dirk@loth.demon.co.uk>: 
 * Support the possibility of hardware state restoring.
 *
 * Raph <grey.havens@earthling.net>:
 * Support for preserving states of network devices and virtual console
 * (including X and svgatextmode)
 *
 * Kurt Garloff <garloff@suse.de>:
 * Straightened the critical function in order to prevent compilers from
 * playing tricks with local variables.
 *
 * Andreas Mohr <a.mohr@mailto.de>
 *
 * Alex Badea <vampire@go.ro>:
 * Fixed runaway init
 *
 * Jeff Snyder <je4d@pobox.com>
 * ACPI patch
 *
 * Nathan Friess <natmanz@shaw.ca>
 * Some patches.
 *
 * More state savers are welcome. Especially for the scsi layer...
 *
 * For TODOs,FIXMEs also look in Documentation/swsusp.txt
 */

#define SWSUSP_MAIN_C

#include "common.h"
#include <asm/i387.h> /* for kernel_fpu_end */

unsigned char software_suspend_enabled = 0;
unsigned int suspend_task = 0;
unsigned int nr_suspends = 0;
struct sysinfo swapinfo;
unsigned char trap_page_allocs = 0; /* code to catch a hard-to-find bug */

/* Variables to be preserved over suspend */
static int resume_status = 0;
int pageset2_sizelow = 0;
int expected_size1, expected_size2; 
struct pagedir __nosavedata pagedir_nosave;

static int pm_suspend_state = 0;
int pagedir1_size = 0, pagedir2_size = 0;

int swsusp_act_used = 0;
int swsusp_lvl_used = 0;
int swsusp_dbg_used = 0;

int now_resuming = 0;
char resume_file[256] = "";	/* For resume= kernel option */

unsigned long orig_mem_free = 0;
unsigned long base_mem_free = 0;
unsigned long swapforimage = 0;
int orig_loglevel = 0;

static void ide_unsuspend(void);
static void ide_suspend(void);
static int drivers_suspend(void);
static void drivers_resume(int);
unsigned long * inusemap = NULL;
unsigned long * pageset2map = NULL;
unsigned long * pagedirmap = NULL;

#ifdef DEFAULT_SUSPEND_CONSOLE
extern void reset_terminal(int currcons, int do_clear);
#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
extern unsigned long checkpage;
#endif

kdev_t resume_device;

/* Suspend pagedir is allocated before final copy, and saved with data
 * as well as separately, so it can be used to load pageset2 at resume.
 *
 * At resume, the original pagedir is loaded as pagedir_nosave and then
 * moved to a place where it doesn't collide with the data to be copied
 * back. Then the data is read (again into spots that don't collide) and
 * then copied back, giving us access to the saved pagedir again and
 * forgetting the loaded pagedir at the same time. We then used the saved
 * pagedir to load pageset2 (if necessary) before freeing that pagedir.
 */

struct pagedir pagedir1 = { 0, 0, 0, 0}, pagedir2 = {0, 0, 0, 0};

unsigned long swsusp_debug_state = 0;
int swsusp_default_console_level = 0;

unsigned long swsusp_action = 0;
unsigned long swsusp_result = 0;

int image_size_limit = 0;
int max_async_ios = 100;
int currentbeep = 260;

/* Locks */
spinlock_t suspend_irq_lock __nosavedata = SPIN_LOCK_UNLOCKED;
__nosavedata unsigned long pagedirflags;

/* Pagedir.c */
extern int calcpagedirsize(int size);
extern pageset_sizes_t count_data_pages(int whichpagedir);
extern int create_pagedir(struct pagedir * p, int pageset_size, int alloc_from);
extern void copy_pageset1(void);
extern int allocatemap(unsigned long ** pagemap, int setnosave);
extern void show_checksum_errors(struct pagedir * p);
extern void check_pagedir_checksums(struct pagedir * p, int at_orig_address);
extern void free_pagedir(struct pagedir * p);
extern int freemap(unsigned long ** pagemap);
extern void store_pagedir_checksums(struct pagedir * p);

/* Freeze_and_free.c */

extern void recalculate_stats(int amount_eaten);
extern void display_stats(unsigned int eaten);
extern int eat_memory_and_freeze_processes(void);
extern void free_grabbed_memory(void);

/* proc.c */

extern int swsusp_init_proc(void);
extern int swsusp_cleanup_proc(void);

__inline__ int fill_suspend_header(struct suspend_header *sh)
{
	int i;

	memset((char *)sh, 0, sizeof(*sh));

	sh->version_code = LINUX_VERSION_CODE;
	sh->num_physpages = num_physpages;
	sh->base_mem_free = orig_mem_free;
	sh->swapforimage = swapforimage;
	sh->expected_size1 = expected_size1;
	sh->expected_size2 = expected_size2;
	strncpy(sh->machine, system_utsname.machine, 65);
	strncpy(sh->version, system_utsname.version, 65);
	sh->num_cpus = smp_num_cpus;
	sh->page_size = PAGE_SIZE;
	sh->pagedir = pagedir1;
	sh->param0 = swsusp_result;
	sh->param1 = swsusp_action;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	sh->param2 = swsusp_debug_state;
#endif
	sh->param3 = console_loglevel;
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	sh->param4 = checkpage;
#endif
#if defined(CONFIG_FBCON_SPLASHSCREEN)
	sh->progress0 = fb_splash_progress[0];
	sh->progress1 = fb_splash_progress[1];
	sh->progress2 = fb_splash_progress[2];
	sh->progress3 = fb_splash_progress[3];
	sh->progress4 = fb_splash_progress[4];
#endif

	/* Save kdevs so we can handle multiple swap files on resume */
	for (i = 0; i < MAX_SWAPFILES; i++)
		if (swap_info[i].swap_device) {
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "Swap device %d is %x.\n", i, swap_info[i].swap_device);
			sh->kdevs[i] = swap_info[i].swap_device;
		}

	/* TODO: needed? mounted fs' last mounted date comparison
	 * [so they haven't been mounted since last suspend.
	 * Maybe it isn't.] [we'd need to do this for _all_ fs-es]
	 */
	return 0;
}

/* Make disk drivers accept operations, again */
static void ide_unsuspend(void)
{
#ifdef CONFIG_BLK_DEV_IDE
	ide_disk_unsuspend(0);
#endif
}

static void ide_suspend(void)
{
#ifdef CONFIG_BLK_DEV_IDE
	unsigned long flags;
	
	spin_lock_irqsave(&io_request_lock, flags);

	do_suspend_sync();

	ide_disk_suspend();
	
	spin_unlock_irqrestore(&io_request_lock, flags);
#endif
}

#define RESUME_PHASE1 1 /* Called from interrupts disabled */
#define RESUME_PHASE2 2 /* Called with interrupts enabled */
#define RESUME_ALL_PHASES (RESUME_PHASE1 | RESUME_PHASE2)

/* Called from process context */
static int drivers_suspend(void)
{
#ifdef CONFIG_BLK_DEV_MD
	md_notify_reboot(NULL, SYS_HALT, NULL);
#endif
	ide_suspend();
	if (!pm_suspend_state) {
		if (pm_send_all(PM_SUSPEND,(void *)3)) {
			printk(name_suspend "Problem while sending suspend event\n");
			drivers_resume(RESUME_ALL_PHASES);
			return(1);
		}
		pm_suspend_state=1;
	}
	  
	return(0);
}

void drivers_resume(int flags)
{
  	if(flags & RESUME_PHASE2) {
#ifdef CONFIG_BLK_DEV_HD
		do_reset_hd();			/* Kill all controller state */
#endif
	}
  	if (flags & RESUME_PHASE1) {
#ifdef CONFIG_BLK_DEV_IDE
		ide_disk_unsuspend(1);		
#endif
#ifdef CONFIG_BLK_DEV_MD
		md_autostart_arrays();
#endif
	}
  	if (flags & RESUME_PHASE2) {
		if (pm_suspend_state) {
			if (pm_send_all(PM_RESUME,(void *)0))
				printk(name_resume "Problem while sending resume event\n");
			pm_suspend_state=0;
		} else
			printk(name_resume "PM suspend state wasn't raised\n");

#ifdef DEFAULT_SUSPEND_CONSOLE
		update_screen(fg_console);
#endif
	}
}

/*
 * save_image
 * Result code (int): Zero on success, non zero on failure.
 * Functionality    : High level routine which performs the steps necessary
 *                    to prepare and save the image after preparatory steps
 *                    have been taken.
 * Key Assumptions  : Processes frozen, sufficient memory available, drivers
 *                    suspended.
 * Called from      : do_magic_suspend_2
 */
static int save_image(void)
{
	pageset_sizes_t result;
	int temp_result, alloc_from, i;
	int orig_num_range_pages, iteration = 0;

	read_swapfiles();

	/* 
	 * FIXME.
	 * We trap anything else that tries to free or allocate memory from now on.
	 * Processes not in the fridge can still be triggered by timers. We should
	 * probably stop this happening, but for now we just trap them if they try
	 * to free or allocate memory on us. This is really only critical when saving
	 * pageset 2 & copying pageset 1 (which is done with IRQs disabled anyway),
	 * but we trap also trap them while trying to calculate the image stats so
	 * that our task isn't further complicated.
	 */

	trap_page_allocs = 1;

	/* Now that we're using range pages, we need to do this iteratively. We go
	 * through the original process (get addresses for basic data to be saved,
	 * swap addresses to store that data in and any extra pages needed to copy
	 * pageset1 data), then check whether we added new range pages in doing so.
	 * If we did, we go back to the start and update the ranges for pagesets 1
	 * & 2, get any extra swap that's needed and any extra pages to be allocated.
	 * When we stop getting extra range pages, we will have gotten stable values
	 * for the other variables.
	 */

	do {
		iteration++;

		orig_num_range_pages = num_range_pages;
		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Iteration %d.\n", 
				iteration);

		free_pagedir(&pagedir1);
		free_pagedir(&pagedir2);

		recalculate_stats(0);
		display_stats(0);
	
		expected_size1 = pageset1_size;
		expected_size2 = pageset2_size;

		PRINTPREEMPTCOUNT("At entry to save_image.");

		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) 
			goto abort_saving;

		if ((RAM_TO_SUSPEND > max_mapnr) || (RAM_TO_RESUME > max_mapnr)) {
			printk(KERN_CRIT name_suspend "Couldn't get enough free pages, on %ld pages short\n",
				 MAX(RAM_TO_RESUME, RAM_TO_SUSPEND) - max_mapnr);
			prepare_status(1, 1, "Couldn't get enough free pages, on %ld pages short\n",
				 MAX(RAM_TO_RESUME, RAM_TO_SUSPEND) - max_mapnr);
			goto abort_saving;
		}
		if (swapinfo.freeswap < SWAP_NEEDED)  {
			printk(KERN_CRIT name_suspend "There's not enough swap space available (%ld pages available, need %d)\n",
				 swapinfo.freeswap, SWAP_NEEDED);
			prepare_status(1, 1, "There's not enough swap space available (%ld pages available, need %d)\n",
				 swapinfo.freeswap, SWAP_NEEDED);
			SET_RESULT_STATE(SUSPEND_ABORTED);
			SET_RESULT_STATE(SUSPEND_INSUFFICIENT_SWAP);
			goto abort_saving;
		}

		swapforimage = swapinfo.freeswap;
	
		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Creating pagedir1 and allocating extra pages (if any)\n");

		alloc_from = (pageset1_size > pageset2_sizelow ? pageset2_sizelow : pageset1_size);
	
		ide_unsuspend();

		if (create_pagedir(&pagedir1, pageset1_size, alloc_from))
			goto abort_saving;

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Creating pagedir2\n");

		if (create_pagedir(&pagedir2, pageset2_size, pageset2_size))
			goto abort_saving;
	
		ide_suspend();

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Calling count_data_pages to recalculate ranges.\n");

		result = count_data_pages(0);

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Number of range pages is now: %d.\n\n", 
				num_range_pages);

	} while (orig_num_range_pages < num_range_pages);

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Calling count_data_pages to set pageset addresses.\n");

	result = count_data_pages(1);
	
	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_saving;

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"--  Result of count_data_pages: %d and %d\n", result.size1, result.size2);
	pageset2_sizelow = result.size2low;
	pageset2_size = result.size2;
	pageset1_size = result.size1;
	
#if 0
	{
		int i, unsetstart = -1;
		struct pbe * pbe = pagedir1.data;
		for (i = 0; i < pageset1_size; i++) {
			if (pbe->origaddress) {
				if (unsetstart != -1) {
					abort_suspend(" ---  Pagedir1 has unset original address entries from %d-%d.\n",
						unsetstart, i);
					unsetstart = -1;
				}
			} else {
				if (unsetstart != -1)
					unsetstart = i;
			}
			pbe = get_next_pbe(pbe);
		}
		
		if (unsetstart != -1) 
			abort_suspend(" ---  Pagedir1 has unset original address entries from %d-%d.\n",
						unsetstart, i);
		unsetstart = -1;
		
		pbe = pagedir1.data;
		for (i = 0; i < pageset1_size; i++) {
			if (pbe->address) {
				if (unsetstart != -1) {
					abort_suspend(" ---  Pagedir1 has unset destination address entries from %d-%d.\n",
						unsetstart, i);
					unsetstart = -1;
				}
			} else {
				if (unsetstart != -1)
					unsetstart = i;
			}
			pbe = get_next_pbe(pbe);
		}
		
		if (unsetstart != -1) 
			abort_suspend(" ---  Pagedir1 has unset entries from %d-%d.\n",
						unsetstart, i);
		unsetstart = -1;

		pbe = pagedir2.data;
		for (i = 0; i < pageset2_size; i++) {
			if (pbe->address) {
				if (unsetstart != -1) {
					abort_suspend(" ---  Pagedir2 has unset destination address entries from %d-%d.\n",
						unsetstart, i);
					unsetstart = -1;
				}
			} else {
				if (!unsetstart)
					unsetstart = i;
			}
			pbe = get_next_pbe(pbe);
		}
		
		if (unsetstart !=-1)
			abort_suspend(" ---  Pagedir2 has unset desination addresses entries from %d-%d.\n",
						unsetstart, i);
	}
#endif

		if (TEST_RESULT_STATE(SUSPEND_ABORTED))
			goto abort_saving;

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Copying pageset2 destination addresses from 0 to %d\n", alloc_from - 1);

		{
			/* 
			 * We know that we don't need to check that we have enough valid pagedir2
			 * entries because this was handled above
			 */

			struct pbe * pbe1 = pagedir1.data;
			struct pbe * pbe2 = pagedir2.data;
			for(i=0; i < alloc_from; i++) {
				while (PageHighMem(pbe2->address))
					pbe2 = get_next_pbe(pbe2);
				pbe1->address = pbe2->address;
				printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "   Entry %d: %lx", 
						i, 
						(unsigned long) pbe1->address);
				pbe1 = get_next_pbe(pbe1);
				pbe2 = get_next_pbe(pbe2);
			}
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 0, "\n");
		}

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Actual values: %d and %d.\n", 
			pageset1_size, 
			pageset2_size);

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	prepare_status(1, 1, "Storing pagedir 2 checksums");
	store_pagedir_checksums(&pagedir2);
#endif

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Preparing to write pages\n");

	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, pagedirflags); 

	ide_unsuspend();

	check_shift_keys(1, "About to write pagedir2. ");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_saving;

	temp_result = write_pageset(&pagedir2, 2);
	
	if (temp_result || TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		goto abort_saving;
	}

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Written pageset2\n");
	
	check_shift_keys(1, "About to copy pageset 1. ");

	ide_suspend();

	prepare_status(1, 1,"Copying pageset1 & storing checksums");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_saving;

	drivers_suspend();

	swsusp_spin_lock_irqsave(&suspend_irq_lock, pagedirflags); 

	PRINTPREEMPTCOUNT("Prior to storing checksums & copying pageset 1.");

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	store_pagedir_checksums(&pagedir1);
#endif

	copy_pageset1();
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Done\n");
	
	trap_page_allocs = 0;

	/*
	 *  ---------------------   FROM HERE ON, NEED TO REREAD PAGESET2 IF ABORTING!!! -----------------
	 *  (We need to ensure saved pages are freed even if memory is still correct).
	 */
	
	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, pagedirflags);
	
	ide_unsuspend();

	check_shift_keys(1, "About to write pageset1. ");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	/*
	 * End of critical section. From now on, we can write to memory,
	 * but we should not touch disk. This specially means we must _not_
	 * touch swap space! Except we must write out our image of course.
	 *
	 */
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Writing pageset1\n");

	temp_result = write_pageset(&pagedir1, 1);

	if (temp_result || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Done\n");

	check_shift_keys(1, "About to write pagedir. ");

	if (temp_result || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto abort_reloading_pagedir_two;

	temp_result = write_pagedir_and_suspend_header();

	if (temp_result || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto abort_reloading_pagedir_two;

	si_swapinfo_devsonly(&swapinfo);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, "Finished writing image. %ld pages of swap left.\n", swapinfo.freeswap);
	ide_suspend();
	swsusp_spin_lock_irqsave(&suspend_irq_lock, pagedirflags);
	PRINTPREEMPTCOUNT("At exit from save_image.");
	return 0;

abort_reloading_pagedir_two:
	trap_page_allocs = 0;

	printk(KERN_ERR "Reloading secondary pagedir.\n");
	temp_result = read_secondary_pagedir(1);
	if (temp_result)
		panic("Attempt to reload pagedir 2 while aborting a suspend failed.");
abort_saving:
	trap_page_allocs = 0;

	free_pagedir(&pagedir2);
	free_pagedir(&pagedir1);

	{
		int was_locked = (spin_is_locked(&suspend_irq_lock));
		if (!was_locked)
			swsusp_spin_lock_irqsave(&suspend_irq_lock, pagedirflags);
	}

	ide_suspend();
	PRINTPREEMPTCOUNT("At exit from save_image.");
	return -1;		

}

/*
 * suspend_power_down
 * Functionality   : Powers down or reboots the computer once the image
 *                   has been written to disk.
 * Key Assumptions : Able to reboot/power down via code called or that
 *                   the warning emitted if the calls fail will be visible
 *                   to the user (ie printk resumes devices).
 * Called From     : do_magic_suspend_2
 */

extern void (*pm_power_off)(void);
extern void swsusp_power_off(void);
void suspend_power_down(void)
{
	C_A_D = 0;
	/* No delay should be needed - ide_disk_suspend purges cache now */
	printk(name_suspend "Trying to power down.\n");
	beepOK;			/* last beep */
	prepare_status(1, 1, "Ready to power down.");
#ifdef CONFIG_VT
	if (TEST_ACTION_STATE(SUSPEND_REBOOT) ^ (!!(shift_state & (1 << KG_CTRL))))
		machine_restart(NULL);
	else
#endif
	{
		if(!pm_power_off) {
			pm_power_off=swsusp_power_off;
			printk(name_suspend "Trying to use an apm bios call (you should enable apm or acpi)\n"); 
		}
		machine_power_off();
	}

	printk(KERN_EMERG name_suspend "Probably not capable for powerdown.\n");
	machine_halt();
	printk(KERN_EMERG name_suspend "System is now halted.\n");
	while (1)
		cpu_relax();
	/* NOTREACHED */
}

/*
 * do_magic_resume_1
 * Functionality   : Preparatory steps for copying the original kernel back.
 * Called From     : include/asm/suspend.h:do_suspend_lowlevel
 */

void do_magic_resume_1(void)
{
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	if (console_loglevel >= SUSPEND_MEDIUM) {
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Checking pagedir1 checksums...\n");
		check_pagedir_checksums(&pagedir_nosave, 0);
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "(There should be no differences here.)\n");
		show_checksum_errors(&pagedir1);
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Done.");
		check_shift_keys(1, "Completed checking pagedir. ");
	}
#endif

	drivers_suspend();
	barrier();
	mb();
	swsusp_spin_lock_irqsave(&suspend_irq_lock, pagedirflags);	/* Done to disable interrupts */ 
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_resume "Waiting for DMAs to settle down...\n");
	mdelay(1000);	/* We do not want some readahead with DMA to corrupt our memory, right?
			   Do it with disabled interrupts for best effect. That way, if some
			   driver scheduled DMA, we have good chance for DMA to finish ;-). */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_resume "About to copy pageset1 back...\n");
}

/*
 * do_magic_resume_2
 * Functionality   : Steps taken after copying back the original kernel at
 *                   resume.
 * Key Assumptions : Will be able to read back secondary pagedir (if 
 *                   applicable).
 * Called From     : include/asm/suspend.h:do_suspend_lowlevel
 */

inline void restore_processor_context (void);
extern void unblank_screen(void);
extern int console_blanked;

void do_magic_resume_2(void)
{

	__flush_tlb_global();
	trap_page_allocs = 0;
	
#ifdef DEFAULT_SUSPEND_CONSOLE
	reset_terminal(fg_console, 1);
	update_screen(fg_console);
	console_blanked = fg_console + 1;
	unblank_screen();
#endif

	PRINTPREEMPTCOUNT("In resume_2 after copy back.");

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	if (console_loglevel >= SUSPEND_MEDIUM) {
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Checked pagedir1 checksums after copying the image back...\n");
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "(Remember that having a few differences here is normal and okay.)\n");
		show_checksum_errors(&pagedir1);
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Done.\n");
		check_shift_keys(1, "Ready to read pageset 2 back. ");
	}
#endif

	tainted |= 4; /* Taint the kernel because we have suspended */
	now_resuming = 1;

	ide_unsuspend();
	
	read_secondary_pagedir(0);
	
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	if (console_loglevel >= SUSPEND_MEDIUM) {
		check_pagedir_checksums(&pagedir2, 1);
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Checking pagedir2 checksums...\n");
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "(No differences should be found here.)\n");
		show_checksum_errors(&pagedir2);
		printlog(SUSPEND_CHECKSUMS, SUSPEND_MEDIUM, "Done.\n");
		check_shift_keys(1, "Ready to cleanup. ");
	}

#endif

	prepare_status(0, 0, "Cleaning up...");
	update_status(0, 100, NULL);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, name_resume "Freeing page directories.\n");
	free_pagedir(&pagedir2);
	update_status(25, 100, NULL);
	free_pagedir(&pagedir1);
	update_status(50, 100, NULL);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, name_resume "Free pagedir lock.\n");
	PRINTPREEMPTCOUNT("Before unlocking pagedir.");
	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, pagedirflags);
	PRINTPREEMPTCOUNT("After unlocking pagedir.");
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, name_resume "Resume drivers.\n");
	drivers_resume(RESUME_ALL_PHASES);
	update_status(75, 100, NULL);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, name_resume "Fixing swap signatures... ");
	mark_swapfiles(((swp_entry_t) {0}), MARK_SWAP_RESUME);
	update_status(100, 100, NULL);
	schedule(); /* Give user time to see we do reach 100! */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, "ok\n");
}

/*
 * do_magic_suspend_1
 * Functionality   : Steps taken prior to saving CPU state and the image
 *                   itself.
 * Called From     : include/asm/suspend.h:do_suspend_lowlevel
 */
void do_magic_suspend_1(void)
{
	mb();
	barrier();
	swsusp_spin_lock_irqsave(&suspend_irq_lock, pagedirflags);
}

/*
 * do_magic_suspend_2
 * Functionality   : Steps taken after saving CPU state to save the
 *                   image and powerdown/reboot or recover on failure.
 * Key Assumptions : save_image returns zero on success; otherwise we need to
 *                   clean up and exit. The state on exiting this routine 
 *                   should be essentially the same as if we have suspended,
 *                   resumed and reached the end of do_magic_resume_2.
 * Called From     : include/asm/suspend.h:do_suspend_lowlevel
 */
void do_magic_suspend_2(void)
{
	if (!save_image())
		suspend_power_down();
	kernel_fpu_end();
	
	if (!TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED))
		printk(KERN_EMERG name_suspend "Suspend failed, trying to recover...\n");
	MDELAY(1000); /* So user can wait and report us messages if armageddon comes :-) */

	barrier();
	mb();
	drivers_resume(RESUME_ALL_PHASES);
	mark_swapfiles(((swp_entry_t) {0}), MARK_SWAP_RECOVER);
	PRINTPREEMPTCOUNT("Before unlocking pagedir.");
	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, pagedirflags);
	PRINTPREEMPTCOUNT("After unlocking pagedir.");
}

/*
 * credit for this code to kdb - but not everyone has it applied.
 */
#define MODLIST_SIZE 4096

static void local_print_modules(void)
{
	static char modlist[MODLIST_SIZE];
	struct module *this_mod;
	int pos = 0;

	this_mod = module_list;
	while (this_mod) {
		if (this_mod->name)
			pos += snprintf(modlist+pos, MODLIST_SIZE-pos-1, 
					"%s ", this_mod->name);
		this_mod = this_mod->next;
	}
	printk("%s\n",modlist);
}

/*
 * display_debug_info
 * Functionality   : At the end of resuming, displays information that may be
 *                   helpful in debugging software suspend.
 * Called From     : do_software_suspend
 */
static void display_debug_info(void)
{
	si_swapinfo_devsonly(&swapinfo);
	printk("\n\nPlease include the following information in any bug report:\n");
	printk("- SWSUSP Version : %s\n", swsusp_version);
	printk("- Kernel Version : %s\n", UTS_RELEASE);
	printk("- Compiler vers. : %d.%d\n", __GNUC__, __GNUC_MINOR__);
#ifdef CONFIG_MODULES	
	printk("- Modules loaded : ");
	local_print_modules();
#else
	printk("- No module support.\n");
#endif
	printk("- Attempt number : %d\n", nr_suspends);
	printk("- Swap available : %ld (amount unused when preparing image).\n", swapforimage);
	printk("- Pageset sizes  : %d and %d (%d low). (Pagedir sizes: %d and %d)\n", pageset1_size, pageset2_size,
			pageset2_sizelow,
			calcpagedirsize(pageset1_size), calcpagedirsize(pageset2_size));
	printk("- Expected sizes : %d and %d.\n", expected_size1, expected_size2);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	printk("- Parameters     : %ld %ld %ld %d %d\n", swsusp_result, swsusp_action, swsusp_debug_state, image_size_limit, max_async_ios);
#else
	printk("- Parameters     : %ld %ld %d %d\n", swsusp_result, swsusp_action, image_size_limit, max_async_ios);
#endif
	printk("- Calculations   : Image size: %d. Ram to suspend: %ld. To resume: %ld.\n", SWAP_NEEDED, RAM_TO_SUSPEND, RAM_TO_RESUME);
	printk("- Limits         : %lu pages RAM. Initial boot: %lu. Current boot: %lu.\n", max_mapnr, orig_mem_free, base_mem_free);
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	printk("- Image compression option compiled in.\n");
#endif
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	printk("- Checksumming enabled.\n");
#endif
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	printk("- Debugging compiled in.\n");
#endif
#ifdef CONFIG_PREEMPT
	printk("- Preemptive kernel.\n");
#endif
	printk("- Max ranges used: %d\n", max_ranges_used);
}
	
#define SUSPEND_C
#include <asm/suspend.h>

/*
 * do_software_suspend
 * Functionality   : First level of code for software suspend invocations.
 *                   Stores and restores load averages (to avoid a spike),
 *                   allocates bitmaps, freezes processes and eats memory
 *                   as required before suspending drivers and invoking
 *                   the 'low level' code to save the state to disk.
 *                   By the time we return from do_suspend_lowlevel, we
 *                   have either failed to save the image or successfully
 *                   suspended and reloaded the image. The difference can
 *                   be discerned by checking SUSPEND_ABORTED.
 * Called From     : 
 */
extern inline int notify_resume(void);
#ifdef CONFIG_MAGIC_SYSRQ
extern int sysrq_pressed;
#endif

static void do_software_suspend(void)
{
	unsigned long avenrun_save[3];
	int i;

	if (!software_suspend_enabled) {
		printk(name_swsusp "Software suspend is disabled\n");
		return;
	}

	printk(name_swsusp "Initiating a software_suspend cycle.\n");
	BUG_ON(in_interrupt());
	software_suspend_enabled = 0;
	swsusp_result = 0;
	max_ranges_used = 0;
	nr_suspends++;
	
	/*
	 * Running swsusp makes for a very high load average. I'm told that
	 * sendmail and crond check the load average, so in order for them
	 * not to be unnecessarily affected by the operation of swsusp, we
	 * store the avenrun values prior to suspending and restore them
	 * at the end of the resume cycle. Thus, the operation of swsusp
	 * should be invisible to them. Thanks to Marcus Gaugusch and Bernard
	 * Blackham for noticing the problem and suggesting the solution.
	 */
	
	for (i = 0; i < 3; i++)
		avenrun_save[i] = avenrun[i];

	PRINTFREEMEM("at start of do_software_suspend");
	si_swapinfo_devsonly(&swapinfo); 
	if (!swapinfo.totalswap) {
		printk(KERN_ERR "You need some swap space on a swap partition to be able to suspend to disk.\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_NOSWAP_AVAILABLE);
		goto out;
	}

	now_resuming = 0;
	last_suspend_cache_page = NULL;

	/* Suspend console switch (if necessary) */
	if (prepare_suspend_console())
		printk(name_suspend "Can't allocate a console... proceeding\n");
	PRINTFREEMEM("after preparing suspend_console");
	beepOK;			/* first beep */

	/* Allocate bitmaps for suspend page flags */
	printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE, "Allocating inusemap\n");
	if (allocatemap(&inusemap, 1))
		goto out;
	PRINTFREEMEM("after allocating inusemap");
	if (allocatemap(&pageset2map, 1)) {
		freemap(&inusemap);
		goto out;
	}
	PRINTFREEMEM("after allocating pageset2 map");

	/* Allocate bitmaps for ignoring pagedir in checksumming */
	if (allocatemap(&pagedirmap, 0)) {
		freemap(&pageset2map);
		freemap(&inusemap);
		goto out;
	}

	PRINTFREEMEM("after allocating pagedir map");
 
	set_chain_names(&pagedir1);
	set_chain_names(&pagedir2);
	/*
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM, "nr_unused_buffer_heads is %d.\n", nr_unused_buffer_heads);
	while ((!nr_unused_buffer_heads) && (bh_tries < 50)) {
		yield();
		bh_tries++;
	}
	if (!nr_unused_buffer_heads) {
		abort_suspend("No free buffer heads for 50 yields! I won't be able to save anything.");
		goto out;
	}
	*/

	/* Free up memory if necessary */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, "Eating memory.\n");
	PRINTPREEMPTCOUNT("Before eat_memory_and_freeze_processes.");
	if (eat_memory_and_freeze_processes() || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto out;

	PRINTPREEMPTCOUNT("After eat_memory_and_freeze_processes.");
	PRINTFREEMEM("after eating memory");

	prepare_status(1, 1, "Freeing eaten memory.");

	/* Release memory that has been eaten */
	free_grabbed_memory();

	beepOK;		/* second beep */

#ifdef CONFIG_PREEMPT
	PRINTPREEMPTCOUNT("Before preempt_disable: Preempt count %d->");
	preempt_disable();
	PRINTPREEMPTCOUNT("After preempt_disable: Preempt count %d->");
#endif

	if (!TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		prepare_status(1, 0, "Suspending drivers.");
		if (drivers_suspend()==0) { 
			prepare_status(1, 0, "Starting low level suspend...");
			PRINTPREEMPTCOUNT("Before do_suspend_lowlevel.");
			do_suspend_lowlevel(0);
			PRINTPREEMPTCOUNT("After do_suspend_lowlevel.");
			beepOK;
		} 
	}
	
	if (SUSPEND_VERBOSE)
		display_debug_info();
	
	check_shift_keys(1, NULL);

	finish_all_io();
	PRINTFREEMEM("after finishing all IO");

#ifdef CONFIG_PREEMPT
	PRINTPREEMPTCOUNT("Before preempt_enable: Preempt count %d->");
	preempt_enable();
	PRINTPREEMPTCOUNT("After preempt_disable: Preempt count %d->");
#endif

out:
	PRINTFREEMEM("at 'out'");

	free_grabbed_memory();

	PRINTFREEMEM("after freeing grabbed memory");
	if (pagedirmap)
		freemap(&pagedirmap);
	PRINTFREEMEM("after freeing pagedir map");
	if (pageset2map)
		freemap(&pageset2map);
	PRINTFREEMEM("after freeing pageset2 map");
	if (inusemap)
		freemap(&inusemap);
	PRINTFREEMEM("after freeing inuse map");

	free_ranges();
	
	PRINTFREEMEM("after freeing ranges");

	if (TEST_RESULT_STATE(SUSPEND_UNEXPECTED_ALLOC)) {
		int orig_console_loglevel = console_loglevel;
		console_loglevel = 5;
		printk("\n\n");
		printk("Congratulations. You have managed to get an unexpected page allocation.\n");
		printk("Nigel really wants to know about this. Please send your log of this\n");
		printk("attempted suspend to him (ncunningham@users.sourceforge.net) so he can\n");
		printk("squash the bug.\n");
		printk("It should be safe for you to try again to suspend.\n");
		printk("Please press shift to carry on.\n");
		while (shift_state & (1 << KG_SHIFT))
			yield();
		while (!(shift_state & (1 << KG_SHIFT)))
			yield();
		console_loglevel = orig_console_loglevel;
	}
	
	for (i = 0; i < 3; i++)
		avenrun[i] = avenrun_save[i];

#ifdef CONFIG_MAGIC_SYSRQ
	sysrq_pressed = 0;
#endif

	thaw_processes();

	MDELAY(1000);
	check_shift_keys(1, "About to restore original console. ");
	restore_console();
#ifdef DEFAULT_SUSPEND_CONSOLE
	update_screen(fg_console);
#endif
	PRINTFREEMEM("at end of do_software_suspend");

	if (notify_resume())
		printk(KERN_EMERG "Failed to notify resume chain.\n");

	software_suspend_enabled = 1;
#ifdef CONFIG_PREEMPT
	PRINTPREEMPTCOUNT("Exiting with preempt count");
#endif
}

/*
 * This is main interface to the outside world. It needs to be
 * called from process context. It sets the active flag, 
 * notifies kswsuspd thread through the wake up and waits 
 * for ACTIVE flag to be cleared.
 */

static DECLARE_WAIT_QUEUE_HEAD(swsusp_wait);
static DECLARE_COMPLETION(cycle_complete);

void software_suspend_pending(void)
{
	if (waitqueue_active(&swsusp_wait)) {
		wake_up_interruptible(&swsusp_wait);
		init_completion(&cycle_complete);
		wait_for_completion(&cycle_complete);
	}
}

/*
 * This is kswsuspd, the kernel daemon which waits to be woken by a userspace
 * daemon (via software_suspend_pending). It signals completion (failure or
 * the end of a full suspend/resume cycle) by resetting cycle_unfinished. In
 * this way, the userspace process blocks until the cycle is complete and a
 * script doesn't need to poll the proc entry (there is now no way to tell
 * from /proc/sys/kernel/swsusp whether the cycle has completed anyway).
 */
static int swsusp_mainloop(void *unused)
{
  	printk(name_swsusp "kswsuspd starting\n"); /* This will print the current version on boot */
	daemonize();

	swsusp_init_proc();

	strcpy(current->comm, "kswsuspd");
	sigfillset(&current->blocked);

	for(;;) {
		interruptible_sleep_on(&swsusp_wait);
		swsusp_result = 0;
		do_software_suspend();
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
		currentbeep = 100; /* not the original one, so that we know if this is first or subsequent suspend */
#endif
		complete(&cycle_complete);
	}

	swsusp_cleanup_proc();
	
	return(0);
}

/*
 * Called from init kernel_thread.
 * We check if we have an image and if so we try to resume.
 * We also start kswsuspd if configuration looks right.
 */

void software_resume(void)
{
	orig_loglevel = console_loglevel;

#ifdef CONFIG_SMP
	printk(KERN_ERR name_swsusp "SMP not supported. Disabled.\n");
	return;
#else
	if (!(cpu_has_pse||cpu_has_pse36)) {
		printk(KERN_ERR name_swsusp "pse or pse36 is required. Disabled.");
		return;
	}
	if (sizeof(swp_entry_t) != sizeof(long)) {
		printk(KERN_WARNING name_swsusp "Size of swp_entry_t != size of long. Disabled.\n");
		return;
	}
	if (PAGE_SIZE % sizeof(struct pbe)) {
		printk(KERN_WARNING name_swsusp "Page size (%ld) mod size of struct pbe (%d) is not 0 (%ld). Disabled.\n", PAGE_SIZE, sizeof(struct pbe), PAGE_SIZE % sizeof(struct pbe));
		return;
	}
#endif
	if (!(resume_status & RESUME_SPECIFIED)) {
		/* Without a pointer to the swap partition we can do nothing - even if noresume is given */
		printk(KERN_ALERT name_swsusp "Missing or invalid swap partition location (resume= parameter). Disabled.\n");
		return;
	}

	/* We enable the possibility of machine suspend */
	software_suspend_enabled = 1;
	orig_mem_free = nr_free_pages();

	suspend_task = current->pid;

	if (resume_status & NORESUME) {
		printk(KERN_WARNING name_swsusp "Resuming disabled as requested.\n");
		if (resume_file[0])
			read_primary_suspend_image(resume_file,1); /* non fatal error ignored */
		suspend_task = 0;
		kernel_thread(swsusp_mainloop, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND | SIGCHLD);
		console_loglevel = orig_loglevel;
		return;
	}

	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, name_swsusp "Resuming from %s\n", resume_file);
	if (read_primary_suspend_image(resume_file, 0)) { /* non fatal error (normal swap space) */
		suspend_task = 0;
		kernel_thread(swsusp_mainloop, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND | SIGCHLD);
		console_loglevel = orig_loglevel;
		return;
	}		

	prepare_status(0, 0, "Copying pages back (no status - sensitive!)...");
	
	do_suspend_lowlevel(1);
	BUG();
}

/*
 * Resume setup: obtain the swap device in which the signature will be used
 * to store a pointer to the suspend header.
 */
static int __init resume_setup(char *str)
{
	if( (str == NULL) || (strncmp(str, "/dev/h", 6)) ) {
		printk(name_swsusp "Currently only compatible with IDE /dev/h?? swap partitions.\n");
		resume_status |= COMMANDLINE_ERROR;
		software_suspend_enabled = 0;
		return 1;
	}

	strncpy(resume_file, str, 255);
	resume_status |= RESUME_SPECIFIED;
	return 0;
}
/*
 * Allow the user to set the action parameter from lilo, prior to resuming.
 */
static int __init swsusp_act_setup(char *str)
{
	if(str)
		swsusp_action=simple_strtol(str,NULL,0);
	printk(name_swsusp "Debug level parameter set : %lx\n",swsusp_action);
	if (TEST_ACTION_STATE(SUSPEND_PAUSE))
		printk("Pausing enabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_SLOW))
		printk("Slow down enabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_NO_ASYNC_READS))
		printk("Asynchronous reads disabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_NO_ASYNC_WRITES))
		printk("Asynchronous writes disabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_NOPAGESET2))
		printk("Pageset2 enabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_LOGALL))
		printk("Logging all messages enabled.\n");
	if (TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION))
		printk("Image compression disabled.\n");
	swsusp_act_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug parameter from lilo, prior to resuming.
 */
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
static int __init swsusp_dbg_setup(char *str)
{
	if(str)
		swsusp_debug_state=simple_strtol(str,NULL,0);
	printk(name_swsusp "Debug parameter set to %lx\n",swsusp_debug_state);
	printlog(SUSPEND_FREEZER, SUSPEND_ERROR,"Freezer messages enabled\n");
	printlog(SUSPEND_EAT_MEMORY, SUSPEND_ERROR,"Memory freeing enabled\n");
	printlog(SUSPEND_PAGESETS, SUSPEND_ERROR,"Pageset debugging enabled\n");
	printlog(SUSPEND_MARK_PAGESET2, SUSPEND_ERROR,"Marking pageset2 debugging enabled\n");
	printlog(SUSPEND_COPY_PAGESET, SUSPEND_ERROR,"Copying pageset debugging enabled\n");
	printlog(SUSPEND_WRITE_PAGESETS, SUSPEND_ERROR,"Writing pagesets debugging enabled\n");
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_ERROR,"Reading pagesets debugging enabled\n");
	printlog(SUSPEND_CHECKSUMS, SUSPEND_ERROR,"Checksumming debugging enabled\n");
	printlog(SUSPEND_SWAP, SUSPEND_ERROR,"Swap allocation debugging enabled\n");
	printlog(SUSPEND_MEMORY, SUSPEND_ERROR,"Memory debugging enabled\n");
	swsusp_dbg_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug level parameter from lilo, prior to
 * resuming.
 */
static int __init swsusp_lvl_setup(char *str)
{
	if(str) {
		swsusp_default_console_level= simple_strtol(str,NULL,0);
		console_loglevel = swsusp_default_console_level;
	}
	printk(name_swsusp "Debug parameter set to %d\n",swsusp_default_console_level);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"Low level debugging messages enabled\n");
	printlog(SUSPEND_ANY_SECTION, SUSPEND_MEDIUM,"Medium level debugging messages enabled\n");
	printlog(SUSPEND_ANY_SECTION, SUSPEND_HIGH,"High level debugging messages enabled\n");
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE,"Verbose debugging enabled\n");
	swsusp_lvl_used = 1;
	return 0;
}
#endif

/*
 * Allow the user to specify that we should ignore any image found and restore
 * the swap header to its normal value. This is equivalent to running the
 * task queue (but not a sync) and then turning off the power. The same
 * precautions should be taken: fsck if you're not journalled.
 */
static int __init noresume_setup(char *str)
{
	resume_status |= NORESUME;
	/* Message printed later */
	return 0;
}

__setup("resume=", resume_setup);
__setup("swsusp_act=", swsusp_act_setup);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
__setup("swsusp_dbg=", swsusp_dbg_setup);
__setup("swsusp_lvl=", swsusp_lvl_setup);
#endif
__setup("noresume", noresume_setup);

EXPORT_SYMBOL(software_suspend_pending);
EXPORT_SYMBOL(refrigerator);
EXPORT_SYMBOL(suspend_task);
