/*
 * linux/kernel/suspend.c
 *
 */

#define SWSUSP_PAGEDIR_C

#include "common.h"
extern unsigned long * inusemap;
extern unsigned long * pageset2map;
extern unsigned long * pagedirmap;
extern struct pagedir pagedir1, pagedir2, pagedir_nosave;

int get_phys_params(struct pbe * pbe);

inline void set_next_pagedir_page(struct pbe * pbe, struct pbe * next)
{
	struct pbelink * link = (struct pbelink *) PBEPAGE_START(pbe);
	link->next = next;
}

inline struct pbe * get_next_pagedir_page(struct pbe * pbe)
{
	struct pbelink * link = (struct pbelink *) PBEPAGE_START(pbe);
	return link->next;
}

inline struct pbe * get_next_pbe(struct pbe * pbe)
{
	int index_within_page = (pbe - ((struct pbe *) PBEPAGE_START(pbe)));
	struct pbe * result;
	if (index_within_page < (PBES_PER_PAGE - 1)) {
		result = pbe;
		result++;
	} else
		result = get_next_pagedir_page(pbe);
	return result;
}

static void clear_map(unsigned long * pagemap)
{
	int i;
	for(i=0; i < (max_mapnr / (8 * sizeof(unsigned long))); i++) {
		pagemap[i]=0;
	}
}

int allocatemap(unsigned long ** pagemap, int setnosave)
{
	void * check;
	int i;
	if (*pagemap) {
		printk("Error. Pagemap already allocated.\n");
	} else {
		check = (void *) __get_free_pages(__GFP_FAST, BITMAP_ORDER);
		if (!check) {
			abort_suspend("Error. Unable to allocate memory for pagemap.\n");
			return 1;
		}
		if (setnosave)
			for (i = 0; i < (1 << BITMAP_ORDER); i++)
				SetPageNosave(virt_to_page((unsigned long) check + i * PAGE_SIZE));
		*pagemap = (unsigned long *) check;
	}
	clear_map(*pagemap);
	return 0;
}

int freemap(unsigned long ** pagemap)
{
	int i;
	if (!*pagemap) {
		printk("Error. Pagemap not allocated.\n");
		return 1;
	} else {
		for (i = 0; i < (1 << BITMAP_ORDER); i++)
			ClearPageNosave(virt_to_page((unsigned long) *pagemap + i * PAGE_SIZE));
		free_pages((unsigned long) *pagemap, BITMAP_ORDER);
		*pagemap = NULL;
		return 0;
	}
}

#define PAGENUMBER(page) (page-mem_map)
#define PAGEINDEX(page) ((PAGENUMBER(page))/(8*sizeof(unsigned long)))
#define PAGEBIT(page) ((int) ((PAGENUMBER(page))%(8 * sizeof(unsigned long))))

/* 
 * freepagesmap is used in two ways: 
 * - During suspend, to tag pages which are not used (to speed up count_data_pages);
 * - During resume, to tag pages which are is pagedir1. This does not tag pagedir2
 *   pages, so !== first use.
 */
#define PageInUse(page)		test_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define SetPageInUse(page)	set_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define ClearPageInUse(page)	clear_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])

#define PagePageset2(page)	test_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define SetPagePageset2(page)	set_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define ClearPagePageset2(page)	clear_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])

#define PagePagedir(page)	test_bit(PAGEBIT(page), &pagedirmap[PAGEINDEX(page)])
#define SetPagePagedir(page)	set_bit(PAGEBIT(page), &pagedirmap[PAGEINDEX(page)])
#define ClearPagePagedir(page)	clear_bit(PAGEBIT(page), &pagedirmap[PAGEINDEX(page)])

#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
unsigned long pagechecksum(struct page * page)
{
	char * address = kmap_nonblock(page);
	unsigned long checksum = 0;
	int i = 0;

	if (!address) { /* Failed to kmap */
		printk("Failed to kmap page %p.\n", page);
		return 0;
	}

	for (i = 0; i < PAGE_SIZE; i++)
		checksum+= (int) (*(address + i) * i);
	kunmap(page);
	return checksum;
}

__nosavedata void check_pagedir_checksums(struct pagedir * p, int at_orig_address)
{
	int i;
	struct pbe * pbe;
	unsigned long current_checksum;

	pbe = p->data;
	for(i = 0; i < p->pageset_size; i++) {
		if (!PagePagedir(pbe->origaddress)) { 	/* If not marked as a pagedir page */
			if (at_orig_address)
				current_checksum = pagechecksum(pbe->origaddress);
			else
				current_checksum = pagechecksum(pbe->address);
			pbe->checksum_error =(current_checksum != pbe->checksum);
		}
		pbe = get_next_pbe(pbe);
	}
}

void show_checksum_errors(struct pagedir * p)
{
	int i;
	struct pbe * pbe;

	pbe = p->data;
	for(i = 0; i < p->pageset_size; i++) {
		if (!PagePagedir(pbe->origaddress)) { 	/* If not marked as a pagedir page */
			if (pbe->checksum_error) {
				char * pageaddress = kmap_nonblock(pbe->origaddress);
				printlog(SUSPEND_CHECKSUMS, SUSPEND_LOW, 
					"Page checksum for page %d/%d doesn't match. ",
					i+1,
					p->pageset_size);
				printlog(SUSPEND_CHECKSUMS, SUSPEND_LOW,
					"Original Address: %p (%p). ",
					pbe->origaddress,
					pageaddress);
				kunmap(pbe->origaddress);
				pageaddress = kmap_nonblock(pbe->address);
				printlog(SUSPEND_CHECKSUMS, SUSPEND_LOW,
					"Copy Address: %p (%p).\n",
					pbe->address,
					pageaddress);
				kunmap(pbe->address);
			} 
		}
		pbe = get_next_pbe(pbe);
	}
}

__nosavedata void store_pagedir_checksums(struct pagedir * p)
{
	int i;
	struct pbe * pbe;

	pbe = p->data;
	for(i = 0; i < p->pageset_size; i++) {
		/* 
		 * For pagedir1, checksum calculated before pages are copied 
		 * but while interrupts are disabled.
		 * For pagedir2, checksum actual page (no copy made) 
		 */
		if (!PagePagedir(pbe->origaddress)) /* If not marked as a pagedir page */
			pbe->checksum = pagechecksum(pbe->origaddress);
		pbe = get_next_pbe(pbe);
	}
}
#endif

#define __calcpagedirsize(size) (SUSPEND_PD_PAGES(size))
int calcpagedirsize(int size)
{
	/*
	 * If adding the pagedir and header will mean we need to store more pages 
	 * than fit, we need to expand the size until it does fit (the pagedir itself 
	 * might require more than one pagedir to store, hence the while).
	 */
	int result;

	result = __calcpagedirsize(size);
	while (__calcpagedirsize(size + result) > result) {
		result++;
	}
	return result;
}

static void generate_free_page_map(void) 
{
	int i, loop;
	struct page * page;
	pg_data_t *pgdat = pgdat_list;
	unsigned type;
	unsigned long flags;

	for(i=0; i < max_mapnr; i++)
		SetPageInUse(mem_map+i);
	
	for (type=0;type < MAX_NR_ZONES; type++) {
		zone_t *zone = pgdat->node_zones + type;
		int order = MAX_ORDER - 1;
		free_area_t *area;
		struct list_head *head, *curr;
		swsusp_spin_lock_irqsave(&zone->lock, flags);
		do {
			int first_entry = 1;
			area = zone->free_area + order;
			head = &area->free_list;
			curr = head;

			for(;;) {
				if(!curr) {
//					printk(KERN_ERR name_suspend "FIXME: this should not happen but it does!!!");
					break;
				}
				if (first_entry) 
					first_entry--;
				else {
					page = list_entry(curr, struct page, list);
					for(loop=0; loop < (1 << order); loop++)
						ClearPageInUse(page+loop);
				}

				curr = curr->next;
				if (curr == head)
					break; 				
			}
		} while(order--);
		swsusp_spin_unlock_irqrestore(&zone->lock, flags);
	}
}

static int is_head_of_free_region(struct page * page)
{
	struct page * posn = page;

	while (((posn-mem_map) < max_mapnr) && (!PageInUse(posn))) 
		posn++;
	return (posn - page);
}

static int append_to_range_chain(int chain, unsigned long min, unsigned long max)
{
	int result = 0;

	switch (chain) {
		case 0:
			return 0;
		case 1:
			result = append_range_to_range_chain(&pagedir1.origranges, min, max);
			break;
		case 2:
			result = append_range_to_range_chain(&pagedir2.origranges, min, max);
			if (!result)
				result = append_range_to_range_chain(&pagedir1.destranges, min, max);
	}

	check_shift_keys(0, NULL);

	return result;
}
pageset_sizes_t count_data_pages(int whichpagedir)
{
	int chunk_size, loop, numnosave = 0;
	int ranges = 0, currentrange = 0;
	struct pbe * pbe1 = NULL;
	struct pbe * pbe2 = NULL;
	int usepagedir2;
	int rangemin = 0;

	pageset_sizes_t result;
	result.size1 = 0;
	result.size2 = 0;
	result.size2low = 0;

	if (whichpagedir) {
		pbe1 = pagedir1.data;
		pbe2 = pagedir2.data;
	}

	put_range_chain(&pagedir1.origranges);
	put_range_chain(&pagedir1.destranges);
	put_range_chain(&pagedir2.origranges);

	generate_free_page_map();
	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		result.size1 = -1;
		result.size2 = -1;
		result.size2low = -1;
		return result;
	}

	if (max_mapnr != num_physpages) {
		abort_suspend("mapnr is not expected");
		result.size1 = -1;
		result.size2 = -1;
		result.size2low = -1;
		return result;
	}
	for (loop = 0; loop < max_mapnr; loop++) {
		if (!PageReserved(mem_map+loop)) {
			if (PageNosave(mem_map+loop)) {
				numnosave++;
				append_to_range_chain(currentrange, rangemin, loop - 1);
				rangemin = loop;
				currentrange = 0;
				continue;
			}

			if ((chunk_size=is_head_of_free_region(mem_map+loop))!=0) {
				append_to_range_chain(currentrange, rangemin, loop - 1);
				rangemin = loop;
				loop += chunk_size - 1;
				currentrange = 0;
				continue;
			}
		} else {
			if (PageNosave(mem_map+loop)) {
				abort_suspend("Reserved page marked nosave! (%lx)\n", ADDRESS(loop));
				result.size1 = 0;
				result.size2 = 0;
				result.size2low = 0;
				return result;
			}
			/*
			 * Just copy whole code segment. Hopefully it is not that big.
			 */
			if (ADDRESS(loop) >= (unsigned long)
				&__nosave_begin && ADDRESS(loop) < 
				(unsigned long)&__nosave_end) {
				numnosave++;
				append_to_range_chain(currentrange, rangemin, loop - 1);
				rangemin = loop;
				currentrange = 0;
				continue;
			}
			/* Hmm, perhaps copying all reserved pages is not too healthy as they may contain 
			   critical bios data? */

		};

		usepagedir2 = (PageHighMem(mem_map+loop) || PagePageset2(mem_map+loop));
		
		if (currentrange != (1 + usepagedir2)) {
			append_to_range_chain(currentrange, rangemin, loop - 1);
			currentrange = usepagedir2 + 1;
			rangemin = loop;
			ranges++;
		}
		
		if (whichpagedir) {
			if (usepagedir2) {
				if (result.size2 <= pagedir2.pageset_size) {
					pbe2->origaddress = mem_map+loop;
					pbe2->address = mem_map+loop;
					pbe2 = get_next_pbe(pbe2);
				}
			} else {
				if (result.size1 <= pagedir1.pageset_size) {
					pbe1->origaddress = mem_map+loop;
					pbe1 = get_next_pbe(pbe1);
				}
			}
		}

		if (PagePageset2(mem_map+loop)) {
			result.size2++;
			if (!PageHighMem(mem_map+loop))
				result.size2low++;
		} else
			result.size1++;
	}
	
	append_to_range_chain(currentrange, rangemin, loop - 1);

	/*
	 * printlog(SUSPEND_VERBOSE, "\nResults: %d and %d(%d low). %d marked Nosave\n", 
	 *		result.size1, result.size2, result.size2low, numnosave);
	 */
	if (whichpagedir) {
		printnolog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, 0, "\n");
		if (result.size1 > pagedir1.pagedir_capacity)
			abort_suspend("%d more pages to be copied than were allowed for! Pagedir1 capacity is %d (counted %d)\n", 
					result.size1 - pagedir1.pagedir_capacity, 
					pagedir1.pagedir_capacity,
					result.size1);
		if (result.size2 > pagedir2.pagedir_capacity)
			abort_suspend("%d more pages to be copied than were allowed for! Pagedir1 capacity is %d (counted %d)\n", 
					result.size2 - pagedir2.pagedir_capacity,
					pagedir2.pagedir_capacity,
					result.size2);
	}
	printnolog(SUSPEND_RANGES, SUSPEND_MEDIUM, 0, "Counted %d ranges.\n", ranges);
	return result;
}

__nosavedata void copy_pageset1(void)
{
	int i = 0;
	struct pbe * pbe = pagedir1.data;

	for (i = 0; i < pageset1_size; i++) {
#if 0
		printnolog(SUSPEND_COPY_PAGESET, SUSPEND_VERBOSE, 1, "!%d! ", i);
		printnolog(SUSPEND_COPY_PAGESET, SUSPEND_VERBOSE, 0, "[1]%lx ->", (unsigned long) page_address(pbe->origaddress)); 
		printnolog(SUSPEND_COPY_PAGESET, SUSPEND_VERBOSE, 0, "[2]%lx.", (unsigned long) page_address(pbe->address));
#endif
		copy_page((char *) page_address(pbe->address), 
				(char *) page_address(pbe->origaddress));
		pbe = get_next_pbe(pbe);
	}

#if 0
	printnolog(SUSPEND_COPY_PAGESET, SUSPEND_VERBOSE, 0, "\n");
#endif
}

void free_pagedir(struct pagedir * p)
{
	int i, swapcount = 0;
	struct pbe * pbe;

	PRINTFREEMEM("At start of free_pagedir");

	if (p->pagedir_size == 0)
		return;

	put_range_chain(&p->origranges);
	put_range_chain(&p->destranges);
	put_range_chain(&p->swapranges);
	put_range_chain(&p->blockranges);
	put_range_chain(&p->allocdranges);

	/* Entries can validly be NULL. We might be called if creating a pagedir fails to allocate a page */
	if (p->first_allocated) {
		pbe = p->first_allocated;
		for(i = p->alloc_from; i < p->pagedir_capacity; i++) {
			if (pbe->address) {
				ClearPageNosave(pbe->address);
				free_page((unsigned long) page_address(pbe->address));
			}
			pbe = get_next_pbe(pbe);
		}
	}
	
	pbe = p->data;
	for(i = 0; i < p->pagedir_capacity; i++) {
		swp_entry_t entry = pbe->swap_address;
		if (entry.val) {
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "(Free %d %p:%lx)", i,
				pbe,
				entry.val);
			swap_free(entry);
			swapcount++;
			pbe->swap_address.val = 0;
		}
		pbe = get_next_pbe(pbe);
	}
		
	pbe = p->data;
	for(i = 0; i < p->pagedir_size; i++) {
		struct pbe * next = get_next_pagedir_page(pbe);
		free_page((unsigned long) pbe);
		pbe = next;
	}

	p->pagedir_size = 0;
	p->data = 0;
	PRINTFREEMEM("At end of free_pagedir");
	printlog(SUSPEND_SWAP, SUSPEND_HIGH, "Freed %d swap pages in free_pagedir. Pageset size was %d.\n", swapcount, p->pageset_size);
}

int create_pagedir(struct pagedir * p, int pageset_size, int alloc_from)
{
	int pagedir_size = calcpagedirsize(pageset_size);
	struct pbe * next;
	struct pbe * last = NULL;
	struct pbe * pbe;
	int swapcount = 0;

#undef abort_and_exit
#define abort_and_exit(f, a...) do { \
			abort_suspend(f, ## a); \
			free_pagedir(p); \
			return 0;\
	} while(0)

	prepare_status(0, 0, "Preparing page directory.");
	if (pageset_size == 0) {
		p->pagedir_size = p->pageset_size = p->pagedir_capacity = 0;
		p->alloc_from = alloc_from;
		return 0;
	}

	PRINTFREEMEM("at start of create_pagedir");

	{
		int i;
		for (i = 0; i < pagedir_size; i++) {
			pbe = (struct pbe *) get_zeroed_page(GFP_ATOMIC);
			if (!pbe) {
				int j;
				pbe = p->data;
				for (j = 0; j < i; j++) {
					next = get_next_pagedir_page(pbe);
					free_page((unsigned long) pbe);
					pbe = next;
				}
				abort_suspend("Unable to allocate a pagedir.\n");
				return 1;
			}
			ClearPageNosave(virt_to_page(pbe));
			ClearPagePageset2(virt_to_page(pbe));
			SetPagePagedir(virt_to_page(pbe));
			if (i == 0)
				p->data = pbe;
			else
				set_next_pagedir_page(last, pbe);
			last = pbe;
		}
	}
	
	{
		int i;
		pbe = p->data;
		
		for(i=0; i < pageset_size; i++) {
			swp_entry_t entry;
			pbe->origaddress = 0;
			pbe->address = 0;
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "");
			entry = get_swap_page(1);
			swapcount++;
			if (!entry.val)
				abort_and_exit("Not enough swapspace (this shouldn't happen!)\n");
			pbe->swap_address = entry;
			add_to_range_chain(&p->swapranges, swap_entry_to_range_val(entry));
			get_phys_params(pbe);
			if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
				p->pagedir_capacity = pageset_size;
				p->pagedir_size = pagedir_size;
				free_pagedir(p);
				return 0;
			}
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 0, "(Get %d %p -> %lx)", i, 
				pbe,
				pbe->swap_address.val);
			if (i == alloc_from)
				p->first_allocated = pbe;
			pbe = get_next_pbe(pbe);
		}
		printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 0, "\n");
	}
	
	{
		int i, numnosaveallocated=0;
		struct pbe * pbe = p->first_allocated;
	
		for(i=alloc_from; i < pageset_size; i++) {
			pbe->address = virt_to_page(get_zeroed_page(GFP_ATOMIC));
			if (!pbe->address)
				abort_and_exit("Unable to allocate pages for pagedir!\n");
			SetPageNosave(pbe->address);
			add_to_range_chain(&p->allocdranges, pbe->address - mem_map);
			add_to_range_chain(&p->destranges, pbe->address - mem_map);
			numnosaveallocated++;
			pbe = get_next_pbe(pbe);
		}
		if (pageset_size > alloc_from) {
			printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"   Allocated memory for pages from %d-%d (%d pages).\n", alloc_from, pageset_size - 1, pageset_size - alloc_from);
		}
	}

	print_chain(6, &p->swapranges, 1);
	p->pagedir_size = calcpagedirsize(pageset_size);
	p->pageset_size = p->pagedir_capacity = pageset_size;
	p->alloc_from = alloc_from;

	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"   Created pagedir of %d pages at %p to store %d pages.\n",
			p->pagedir_size,
			p->data,
			p->pageset_size);

	PRINTFREEMEM("at end of create_pagedir");
	printlog(SUSPEND_SWAP, SUSPEND_HIGH, "Allocated %d swap pages in create_pagedir.\n", swapcount);
	return 0;
}

/*
 * Mark unshared pages in processes not needed for suspend as being able to be written
 * out in a separate pagedir. 
 * HighMem pages are simply marked as pageset2. They won't be needed during suspend.
 * Since we don't know which highmem pages are free, we can't return a count any more.
 */

void markpagesforpageset2(void)
{
	struct list_head * entry;
	int i;

	if (max_mapnr != num_physpages) {
		abort_suspend("mapnr is not expected");
		return;
	}
	
	for (i = 0; i < max_mapnr; i++)
		ClearPagePageset2(mem_map+i);

	if (TEST_ACTION_STATE(SUSPEND_NOPAGESET2)) {
#ifdef CONFIG_HIGHMEM
		printk("Ignoring request for no pageset2 - HighMem support is configured.\n");
#else
		return;
#endif
	}

	/*
	 * Mark page cache pages as pageset2.
	 */
	
	/*
	 * Inactive list
	 */
	entry = inactive_list.prev;
	while (entry != &inactive_list) {
		struct page * page;

		page = list_entry(entry, struct page, lru);
		SetPagePageset2(page);
		entry = entry->prev;
	}

	/*
	 * Active list
	 */
	entry = active_list.prev;
	while (entry != &active_list) {
		struct page * page;

		page = list_entry(entry, struct page, lru);
		SetPagePageset2(page);
		entry = entry->prev;
	}

	/*
	 * Next, ensure that pagedir pages aren't marked as pageset 2
	 */

	if (pagedir1.pageset_size) {
		struct pbe * pbe = pagedir1.data;
		for (i = 0; i < pagedir1.pagedir_size; i++) {
			struct page * page = virt_to_page(pbe);
			if (PagePageset2(page)) {	// Must be assigned by the time recalc stats is called
				printlog(SUSPEND_MARK_PAGESET2, SUSPEND_VERBOSE, "Pagedir[%d] was marked as pageset2 - unmarking.\n", i);
				ClearPagePageset2(page);
			}
			pbe = get_next_pbe(pbe);
		}
	}

	/*
	 * Finally, ensure HighMem pages are pageset 2
	 */

	for (i = 0; i < max_mapnr; i++)
		if (PageHighMem(mem_map+i))
			SetPagePageset2(mem_map+i);
}

void warmup_collision_cache(void) {
	int i;
	struct pbe * pbe = pagedir_nosave.data;
	
	allocatemap(&inusemap, 0);/* Doesn't get deallocated because forgotten
				   when we copy PageDir1 back. Doesn't matter
				   if collides because not used during copy back.
				 */
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "Setting up pagedir cache...");
	for (i = 0; i < max_mapnr; i++)
		ClearPageInUse(mem_map+i);

	for(i=0; i < pagedir_nosave.pageset_size; i++) {
		SetPageInUse(pbe->origaddress);
		printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 1, "Marking %p as in use.", pbe->origaddress);
		if (!(i%800)) {
			printlog(SUSPEND_READ_PAGESETS, SUSPEND_MEDIUM, ".");
		}
		pbe = get_next_pbe(pbe);
	}
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "%d", i);
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, "|\n");
}


/*
 * We check here that pagedir & pages it points to won't collide with pages
 * where we're going to restore from the loaded pages later
 */

int check_pagedir(void)
{
	int i, nrdone = 0;
	void **eaten_memory = NULL;
	void **c = eaten_memory, *f, *addr;
	struct page * pageaddr = NULL;
	struct pbe * pbe = pagedir_nosave.data;

	// Because we're trying to make this work when we're saving as much memory as possible
	// we need to remember the pages we reject here and then free them when we're done.
	
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\nAllocating memory.\n");
	for(i=0; i < pagedir_nosave.pageset_size; i++) {
		while ((addr = (void *) get_zeroed_page(GFP_ATOMIC))) {
			memset(addr, 0, PAGE_SIZE);
			pageaddr = virt_to_page(addr);
			if (!PageInUse(pageaddr)) {
				break;
			}
			eaten_memory = addr;
			*eaten_memory = c;
			c = eaten_memory;
		}
		printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 1, "%d/%d (%p = %p)", i, pagedir_nosave.pageset_size, pbe, pageaddr);
		pbe->address = pageaddr;
		pbe = get_next_pbe(pbe);
		nrdone++;
	}

	// Free unwanted memory
	c = eaten_memory;
	while(c) {
		//printlog(SUSPEND_LOW, ":");
		f = c;
		c = *c;
		if (f)
			free_page((unsigned long) f);
	}
	eaten_memory = NULL;
	
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"Check_pagedir: Prepared %d of %d pages.\n", nrdone, pagedir_nosave.pageset_size);
	
	check_shift_keys(1, "Pagedir prepared. ");

	return 0;
}

int relocate_pagedir(void)
{
	void **eaten_memory = NULL;
	void **c = eaten_memory, *m = NULL, *f;
	int oom = 0, i, numeaten = 0;
	struct pbe * pbe = pagedir_nosave.data;
	struct pbe * last = NULL;

	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"Relocating conflicting parts of pagedir.\n");

	for (i = 0; i < pagedir_nosave.pagedir_size; i++) {
		int this_collides = 0;

		this_collides = PageInUse(virt_to_page(pbe));

		if (this_collides) {
			printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\nRelocating pbe pagedir %d.\n", i);
			while ((m = (void *) get_zeroed_page(GFP_ATOMIC))) {
				memset(m, 0, PAGE_SIZE);
				if (!PageInUse(virt_to_page(m))) {
					copy_page(m, (void *) pbe);
					free_page((unsigned long) pbe);
					if (i == 0)
						pagedir_nosave.data = m;
					else 
						set_next_pagedir_page(last, m);
					pbe = (struct pbe *) m;
					break;
				}
				numeaten ++;
				eaten_memory = m;
				printnolog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE, 0, "Eaten: %d. Still to try:%d\r", numeaten, nr_free_pages()); 
				*eaten_memory = c;
				c = eaten_memory;
			}

			if (!m) {
				printk("\nRan out of memory trying to relocate pagedir (tried %d pages).\n", numeaten);
				oom = 1;
				break;
			}
		}
		SetPagePagedir(virt_to_page(pbe));
		last = pbe;
		pbe = get_next_pagedir_page(pbe);
	}
		
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\nFreeing rejected memory locations...");

	c = eaten_memory;
	while(c) {
		f = c;
		c = *c;
		if (f)
			free_pages((unsigned long) f, 0);
	}
	eaten_memory = NULL;
	
	printlog(SUSPEND_READ_PAGESETS, SUSPEND_VERBOSE,"\n");
	
	check_shift_keys(1, "Pagedir relocated. ");

	if (oom) 
		return -ENOMEM;
	else
		return 0;
}

char* origrangesname = "original addresses";
char* destrangesname = "destination addresses";
char* swaprangesname = "swap addresses";
char* blockrangesname = "block addresses";
char* allocdrangesname = "allocated addresses";

void set_chain_names(struct pagedir * p)
{
	p->origranges.name = origrangesname;
	p->destranges.name = destrangesname;
	p->swapranges.name = swaprangesname;
	p->blockranges.name = blockrangesname;
	p->allocdranges.name = allocdrangesname;
}
