/*
 * /drivers/swsusp/proc.c
 *
 * This file is to realize proc entries for tuning Software Suspend.
 *
 * Versions:
 * 1: /proc/sys/kernel/swsusp the only tuning interface
 * 2: Initial version of this file
 * 3: Removed KDB parameter. 
 *    Added checkpage parameter (for checking checksum of a page over time).
 * 4: Added entry for maximum granularity in splash screen progress bar.
 *    (Progress bar is slow, but the right setting will vary with disk & processor
 *    speed and the user's tastes).
 * 5: Added enable_escape to control ability to cancel aborting by pressing
 *    ESC key.
 */

#define SWSUSP_PROC_C

int swsusp_proc_version = 5;

#include "common.h"
#include <linux/proc_fs.h>

extern int swsusp_default_console_level;
extern unsigned long checkpage;
extern int bar_granularity_limit;

struct proc_param {
	int key;
	char * filename;
	void * variable;
	int bit;
	int permissions; /* file permissions */
};

#define PROC_WRITEONLY 0200
#define PROC_READONLY 0400
#define PROC_RW 0600

/*
 * There's nothing special about the 'key' field. It just needs to match
 * the entry you want run in the case statement below. I've made them
 * increasing numbers simply to make the pattern clearer.
 * 1-5 are straight integers.
 * 10  is the activation case.
 * 20+ are setting/reading bits.
 */
static struct proc_param proc_params[] = {
	{ 1, "version", NULL, 0, PROC_READONLY },
	{ 2, "last_result", &swsusp_result, 0, PROC_READONLY },
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	{ 3, "debug_sections", &swsusp_debug_state, 0, PROC_RW },
#endif
	{ 4, "default_console_level", &swsusp_default_console_level, 0, PROC_RW },
	{ 5, "image_size_limit", &image_size_limit, 0, PROC_RW },
	{ 6, "async_io_limit", &max_async_ios, 0, PROC_RW },
	{ 7, "interface_version", &swsusp_proc_version, 0, PROC_READONLY },
#ifdef CONFIG_SOFTWARE_SUSPEND_CHECKSUM
	{ 8, "checkpage", &checkpage, 0, PROC_RW },
#endif
#ifdef CONFIG_FBCON_SPLASHSCREEN
	{ 9, "progressbar_granularity_limit", &bar_granularity_limit, 0, PROC_RW },
#endif
	{ 10, "activate", NULL, 0, PROC_WRITEONLY },
	{ 20, "reboot", &swsusp_action, SUSPEND_REBOOT, PROC_RW },
	{ 21, "no_output", &swsusp_action, SUSPEND_NO_OUTPUT, PROC_RW },
	{ 22, "pause_between_steps", &swsusp_action, SUSPEND_PAUSE, PROC_RW },
	{ 23, "slow", &swsusp_action, SUSPEND_SLOW, PROC_RW },
	{ 24, "beeping", &swsusp_action, SUSPEND_BEEP, PROC_RW },
	{ 25, "no_async_reads", &swsusp_action, SUSPEND_NO_ASYNC_READS, PROC_RW },
	{ 26, "no_async_writes", &swsusp_action, SUSPEND_NO_ASYNC_WRITES, PROC_RW },
	{ 27, "nopageset2", &swsusp_action, SUSPEND_NOPAGESET2, PROC_RW },
	{ 28, "log_everything", &swsusp_action, SUSPEND_LOGALL, PROC_RW },
#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
	{ 29, "disable_compression", &swsusp_action, SUSPEND_NO_COMPRESSION, PROC_RW },
#endif
	{ 30, "enable_escape", &swsusp_action, SUSPEND_CAN_CANCEL, PROC_RW },
};

static struct proc_dir_entry *swsusp_dir, *swsusp_files[sizeof(proc_params) / sizeof(struct proc_param)];
static struct proc_dir_entry *compat_entry, *compat_parent;

int swsusp_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;
	struct proc_param * index = (struct proc_param *) data;
	
	switch (index->key) {
		case 1:
			len = sprintf(page, "%s\n", swsusp_version);
			break;
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 9:
			len = sprintf(page, "%ld\n", *((unsigned long *) index->variable));
			break;
		case 8:
			len = sprintf(page, "%lx\n", *((unsigned long *) index->variable));
			break;
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30:
			len = sprintf(page, "%d\n", test_bit(index->bit, ((unsigned long *) index->variable))!=0);
			break;
	}

	*eof = 1;
	return len;
}

int swsusp_read_compat_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	len = sprintf(page, "%ld %ld %ld %d %d %d\n",
			swsusp_result,
			swsusp_action,
			swsusp_debug_state,
			swsusp_default_console_level,
			image_size_limit,
			max_async_ios);
#else
	len = sprintf(page, "%ld %ld %d %d\n",
			swsusp_result,
			swsusp_action,
			image_size_limit,
			max_async_ios);
#endif
	*eof = 1;
	return len;
}
int swsusp_write_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	struct proc_param * index = (struct proc_param *) data;
	char * buf = (char *) get_zeroed_page(GFP_ATOMIC);
	
	if (!buf)
		return -ENOMEM;
	
	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	buf[count] = 0;

	switch (index->key) {
		case 1:
		case 2:
			break;
		case 3:
		case 4:
		case 5:
		case 6:
		case 8:
		case 9:
			if (buf)
				*((unsigned long *) index->variable) = simple_strtoul(buf, NULL, 0);
			break;
		case 10:
			software_suspend_pending();
			break;
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30:
			{
			int value = simple_strtoul(buf, NULL, 0);
			if (value)
				set_bit(index->bit, ((unsigned long *) index->variable));
			else 
				clear_bit(index->bit, ((unsigned long *) index->variable));
			}
			break;
	}

	free_pages((unsigned long) buf, 0);
	return count;
}

int swsusp_write_compat_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	char * buf = (char *) get_zeroed_page(GFP_ATOMIC), *lastbuf;
	int i;
	unsigned long nextval;
	
	if (!buf)
		return -ENOMEM;
	
	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	for (i = 0; i < 6; i++) {
		if (!buf)
			break;
		lastbuf = buf;
		nextval = simple_strtoul(buf, &buf, 0);
		if (buf == lastbuf)
			break;
		switch (i) {
			case 0:
				swsusp_result = nextval;
				break;
			case 1:
				swsusp_action = nextval;
#ifndef CONFIG_SOFTWARE_SUSPEND_DEBUG
				i++;
#endif
				break;
			case 2:
				swsusp_debug_state = nextval;
				break;
			case 3:
				swsusp_default_console_level = nextval;
				break;
			case 4:
				image_size_limit = nextval;
				break;
			case 5:
				max_async_ios = nextval;
				break;

		}
		buf++;
		while (*buf == ' ')
			buf++;
	}
	buf[count] = 0;
	free_pages((unsigned long) buf, 0);
	return count;
}


/* Based on remove_proc_entry */

struct proc_dir_entry * find_proc_dir_entry(const char *name, struct proc_dir_entry *parent)
{
	struct proc_dir_entry **p;
	int len;

	len = strlen(name);
	for (p = &parent->subdir; *p; p=&(*p)->next ) {
		if (proc_match(len, name, *p)) {
			return *p;
		}
	}
	return NULL;
}

int swsusp_init_proc(void)
{
	int i;
	int numfiles = sizeof(proc_params) / sizeof(struct proc_param);
	
	swsusp_dir = proc_mkdir("swsusp", NULL);
	
	if (swsusp_dir == NULL) {
		printk(KERN_ERR "Failed to create /proc/swsusp.\n");
		return 1;
	}

	for (i=0; i<numfiles; i++) {
		swsusp_files[i] = create_proc_entry(proc_params[i].filename,
				proc_params[i].permissions, 
				swsusp_dir);
		swsusp_files[i]->data = &proc_params[i];
		swsusp_files[i]->read_proc = swsusp_read_proc;
		swsusp_files[i]->write_proc = swsusp_write_proc;
	}

	compat_parent = find_proc_dir_entry("sys", &proc_root);
	if (compat_parent) {
		compat_parent = find_proc_dir_entry("kernel", compat_parent);
		if (compat_parent) {
			compat_entry = create_proc_entry("swsusp",
				0x600, 
				compat_parent);
			compat_entry->read_proc = swsusp_read_compat_proc;
			compat_entry->write_proc = swsusp_write_compat_proc;
		}
	}

	return 0;
}

int swsusp_cleanup_proc(void)
{
	int i;
	int numfiles = sizeof(proc_params) / sizeof(struct proc_param);
	
	for (i = 0; i < numfiles; i++)
		remove_proc_entry(proc_params[i].filename, swsusp_dir);
	return 0;
	
	if (compat_parent && compat_entry)
		remove_proc_entry("swsusp", compat_parent);
}
