
#include "common.h"

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define RPRINTNOLOG(f, a...) \
	do { if (1 || debuggingchain) \
		printnolog(f, ##a); \
	} while (0)
#else
#define RPRINTNOLOG(f, a...) do { } while(0)
#endif
int debuggingchain = 0;

struct range * unused_ranges = NULL;
int nr_unused_ranges = 0;
int max_ranges_used = 0;
int num_range_pages = 0;
static unsigned long ranges_allocated = 0;

/* Add_range_page
 * Allocates new pages for storing ranges.
 * Returns 1 on failure to get a page.
 * Otherwise adds the new pages to the unused_ranges pool.
 */

int add_range_pages(int number_requested)
{
	int i, j;
	struct range * ranges;
	unsigned long next;

	for (j = 0; j < number_requested; j++) {
		next = get_zeroed_page(GFP_ATOMIC);
		RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_LOW, 0, "Add range page %lx.\n", next);
		if (!next) {
			if (!nr_unused_ranges)
				return 1;
			else
				return 0;
		}

		num_range_pages++;
		ranges = (struct range *) next;
		for (i = 0; i < (PAGE_SIZE / sizeof(struct range) - 1); i++)
			(ranges+i)->next = (ranges+i+1);
		(ranges + i - 1)->next = unused_ranges;
		unused_ranges = ranges;
		nr_unused_ranges += i;
		if (console_loglevel > 5) {
			for (i = 0; i < (PAGE_SIZE / sizeof(struct range) - 1); i++)
				RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Ranges %d = %p. Next = %p\n", 
					i, 
					(ranges + i),
					(ranges + i)->next);
		}
	}
	return 0;
}


/* 
 * Free ranges.
 *
 * Frees pages allocated by add_range_pages()
 *
 * Checks that all ranges allocated have been freed and aborts
 * if this is not true.
 *
 * Ranges may not be in memory order but we don't
 * mind. We just look for a range that is on a
 * page boundary. That gives us the pages to be
 * freed. As we find them, we link them together
 * into a new chain (we're not going to use the
 * other ranges anyway) and then free the chain.
 */

int free_ranges(void)
{
	int numpages = 0, i = 0;
	struct range * first = NULL, * last = NULL;

	if (ranges_allocated)
		printk(" *** Warning: %ld ranges still allocated when free_ranges() called.\n", ranges_allocated);

	while (unused_ranges) {
		if (!(((unsigned long) unused_ranges) & (PAGE_SIZE - 1))) {
			if (!first) {
				first = last = unused_ranges;
				printnolog(SUSPEND_RANGES, SUSPEND_HIGH, 0, "Setting firstpage to %p.\n", first);
			} else {
				printnolog(SUSPEND_RANGES, SUSPEND_HIGH, 0, "Linking in page # %d: %p.\n", numpages + 1, unused_ranges);
				last->next = unused_ranges;
				last = unused_ranges;
			}
			numpages++;
			unused_ranges= unused_ranges->next;
			last->next = NULL;
		} else
			unused_ranges= unused_ranges->next;
	}
	nr_unused_ranges=0;

	while (first) {
		last = first;
		first = first->next;
		i++;
		printnolog(SUSPEND_RANGES, SUSPEND_LOW, 0, "Freeing page %d (%lx).\n", i, last);
		free_pages((unsigned long) last, 0);
		num_range_pages--;
	}
	return 0;
}

/* get_range
 *
 * Returns a free range, having removed it from the
 * unused list and having incremented the usage counter.
 * May imply allocating a new page and may therefore 
 * fail, returning NULL instead.
 *
 * No locking. This is because we are only called 
 * from swsusp, which is single threaded
 */

static struct range * get_range(void)
{
	struct range * result;
	
	if (!unused_ranges)
		if (add_range_pages(1))
			return NULL;

	result = unused_ranges;
	unused_ranges = unused_ranges->next;
	nr_unused_ranges--;
	ranges_allocated++;
	if (ranges_allocated > max_ranges_used)
		max_ranges_used++;
	result->minimum = 0;
	result->maximum = 0;
	result->next = NULL;
	return result;
}

/*
 * put_range.
 *
 * Returns a range to the pool of unused pages and
 * decrements the usage counter.
 */
static void put_range(struct range * range)
{
	if (!range) {
		printk("Error! put_range called with NULL range.\n");
		return;
	}
	range->next = unused_ranges;
	unused_ranges = range;
	ranges_allocated--;
	nr_unused_ranges++;
}

/*
 * put_range_chain.
 *
 * Returns a whole chain of ranges to the
 * unused pool.
 */
void put_range_chain(struct rangechain * chain)
{
	struct range * this = chain->first, *next = NULL;
	int count = 0;

	debuggingchain = chain->debug;
	RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0, "\n");
	
	while (this) {
		count++;
		next = this->next;
		RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "%s: Putting range %p.\n", chain->name, this);
		if (!next) {
			this->next = unused_ranges;
			unused_ranges = chain->first;
			chain->first = NULL;
			chain->frees+= count;
			ranges_allocated -= count;
			nr_unused_ranges += count;
		}
		this = next;
	}
	
	RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_MEDIUM, 0, "Emptied %s. For this chain, allocated %d and freed %d. Used optimisation %d times.\n",
			chain->name,
			chain->allocs,
			chain->frees,
			chain->timesusedoptimisation);
	chain->allocs = 0;
	chain->frees = 0;
	chain->timesusedoptimisation = 0;
	chain->lastaccessed = NULL;	/* Invalidate optimisation info */
}

void print_chain(int debuglevel, struct rangechain * chain, int printasswap)
{
	struct range * this = chain->first;
	int count = 0;
	
	debuggingchain = chain->debug;

	if (console_loglevel < debuglevel)
		return;
	
	RPRINTNOLOG(SUSPEND_RANGES, debuglevel, 1, "%s:", chain->name);
	
	while (this) {
		if (printasswap)
			RPRINTNOLOG(SUSPEND_RANGES, debuglevel, 0, "%lx-%lx ", 
					range_val_to_swap_entry(this->minimum), 
					range_val_to_swap_entry(this->maximum));
		else
			RPRINTNOLOG(SUSPEND_RANGES, debuglevel, 0, "%lu-%lu ", this->minimum, this->maximum);
		this = this->next;
		count++;
	}
	RPRINTNOLOG(SUSPEND_RANGES, debuglevel, 0, "\n%s: %d entries/%ld allocated. Allocated %d and freed %d.",
			chain->name,
			count, 
			ranges_allocated,
			chain->allocs,
			chain->frees);
	if (count != (chain->allocs - chain->frees)) {
		chain->debug = 1;
		check_shift_keys(1, "Discrepancy in chain.");
	}
}
/*
 * add_to_range_chain.
 *
 * Takes a value to be stored and a pointer
 * to a chain and adds the value to the range
 * chain, merging with an existing range or 
 * adding a new entry as necessary. Ranges 
 * are stored in increasing order.
 *
 * Values should be consecutive, and so may
 * need to be transformed first. (eg for
 * pages, would want to call with page-mem_map).
 *
 * Important optimisation:
 * We store in the chain info the location of
 * the last range accessed or added (and its
 * previous). If the next value is outside this
 * range by one, we start from the previous
 * entry instead of the start of the chain.
 * In cases of heavy fragmentation, this saves
 * a lot of time searching.
 * 
 * Returns:
 * 0 if successful
 * 1 if the value is already included.
 * 2 if unable to allocate memory.
 * 3 if fall out bottom (shouldn't happen).
 */

int add_to_range_chain(struct rangechain * chain, unsigned long value)
{
	struct range * this, * prev = NULL, * prevtoprev = NULL;
	int usedoptimisation = 0;
	
	debuggingchain = chain->debug;
	RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0, "\n");
	
	chain->size++;

	//printk("Adding %ld to %s.\n", value, chain->name);
	//print_chain(0, chain, 0);

	//RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0, "%s:", chain->name);

	if (!chain->first) {	/* Empty */
		chain->first = get_range();
		chain->allocs++;
		if (!chain->first)
			return 2;
		chain->first->maximum = value;
		chain->first->minimum = value;
		RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0, "%s: Added initial entry %p to chain.\n", 
				chain->name,
				chain->first);
		print_chain(6, chain, 0);
		return 0;
	}
	
	this = chain->first;

	if (chain->lastaccessed && chain->prevtolastaccessed && chain->prevtoprev) {
		if ((value + 1) == chain->lastaccessed->minimum) {
			prev = chain->prevtoprev;
			this = chain->prevtolastaccessed;
			usedoptimisation = 1;
		} else if ((chain->lastaccessed) && ((value - 1) == chain->lastaccessed->maximum)) {
			prev = chain->prevtolastaccessed;
			this = chain->lastaccessed;
			usedoptimisation = 1;
		}
		
		/* prev wrong but doesn't matter */
		if (usedoptimisation)
			RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0,
				"Adding %ld, using last accessed: (%ld-%ld)(%ld-%ld)\n",
				value,
				prev->minimum,
				prev->maximum,
				this->minimum,
				this->maximum);
		chain->timesusedoptimisation++;
	}

	while (this) {
		/* Need new entry prior to this? */
		if ((value + 1) < this->minimum) {
			struct range * new = get_range();
			if (!new)
				return 2;
			chain->allocs++;
			new->minimum = value;
			new->maximum = value;
			new->next = this;
			/* Prior to start of chain? */
			if (!prev)
				chain->first = new;
			else
				prev->next = new;
			RPRINTNOLOG(SUSPEND_RANGES, SUSPEND_HIGH, 0, 
					"%s: Adding new range %p prior to existing range.\n",
					chain->name,
					new);
			print_chain(6, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = new;
			}
			return 0;
		}

		if ((this->minimum <= value) && (this->maximum >= value)) {
			printk("%s: Trying to add a value (%ld) already included in chain.\n",
					chain->name,
					value);
			print_chain(0, chain, 0);
			chain->size--;	/* Undo addition done above */
			return 1;
		}
		if ((value + 1) == this->minimum) {
			this->minimum = value;
			print_chain(7, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = this;
			}
			return 0;
		}
		if ((value - 1) == this->maximum) {
			if ((this->next) && (this->next->minimum == value + 1)) {
				struct range * oldnext = this->next;
				this->maximum = this->next->maximum;
				this->next = this->next->next;
				put_range(oldnext);
				chain->lastaccessed = NULL;	/* Invalidate optimisation info */
				chain->frees++;
				RPRINTNOLOG(SUSPEND_RANGES,
						SUSPEND_HIGH,
						0,
						"%s: Putting merged range %p.\n", 
						chain->name,
						oldnext);
				print_chain(6, chain, 0);
				if (!usedoptimisation) {
					chain->prevtoprev = prevtoprev;
					chain->prevtolastaccessed = prev;
					chain->lastaccessed = this;
				}
				return 0;
			} 
			this->maximum = value;
			print_chain(7, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = this;
			}
			return 0;
		}
		if (!this->next) {
			struct range * new = get_range();
			if (!new)
				return 2;
			chain->allocs++;
			new->minimum = value;
			new->maximum = value;
			new->next = NULL;
			this->next = new;
			RPRINTNOLOG(SUSPEND_RANGES,
					SUSPEND_HIGH,
					0,
					"%s: Got new range %p for end of chain.\n",
					chain->name,
					new);
			print_chain(6, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prev;
				chain->prevtolastaccessed = this;
				chain->lastaccessed = new;
			}
			return 0;
		}
		prevtoprev = prev;
		prev = this;
		this = this->next;
	}
	printk("\nFell out the bottom of add_to_range_chain. This shouldn't happen!\n");
	SET_RESULT_STATE(SUSPEND_ABORTED);
	return 3;
}

/* append_range
 * Used where we know a range is to be added to the end of the list
 * and does not need merging with the current last range.
 * (count_data_pages only at the moment)
 */

int append_range_to_range_chain(struct rangechain * chain, unsigned long minimum, unsigned long maximum)
{
	struct range * newrange, * oldlast = NULL;

	if (chain->first) {
		oldlast = chain->first;
		while (oldlast->next)
			oldlast = oldlast->next;
	}

	newrange = get_range();
	if (!newrange)
		return 2;

	chain->allocs++;
	newrange->minimum = minimum;
	newrange->maximum = maximum;
	newrange->next = NULL;

	if (oldlast)
		oldlast->next = newrange;
	else
		chain->first = newrange;
	RPRINTNOLOG(SUSPEND_RANGES,
		SUSPEND_HIGH,
		0,
		"%s: Got new range %p for end of chain.\n",
		chain->name,
		newrange);
		print_chain(6, chain, 0);
	/* No need to reset optimisation info since added to end */
	return 0;
}

/* get_next_range_entry
 * Given info on the previous entry in a list, 
 * update the info & return the next entry
 */
unsigned long get_next_range_entry(struct range_entry * range_entry)
{
	unsigned long thisvalue;

	/* We are already pointing at the entry to return */
	if (!range_entry->rangeptr) 
		return -1;
	
	thisvalue = range_entry->thisvalue;
	
	if (range_entry->thisvalue == range_entry->rangeptr->maximum) {
		if (!range_entry->rangeptr->next)
			range_entry->rangeptr = NULL;
		else {
			range_entry->rangeptr = range_entry->rangeptr->next;
			range_entry->thisvalue = range_entry->rangeptr->minimum;
		}
	} else
		range_entry->thisvalue++;
	
	return thisvalue;
}

/* Initialise range entry */
void init_range_entry(struct range_entry * range_entry, struct rangechain rangechain)
{
	range_entry->rangeptr = rangechain.first;
	range_entry->thisvalue = range_entry->rangeptr->minimum;
}
